//---------------------------------------------------------------------------

#include "DebugEdit.h"
#include "Utility.h"
//---------------------------------------------------------------------------
/////////////////////////////////////////////////
/// Class TDebugVariable
/////////////////////////////////////////////////
TDebugVariable::TDebugVariable(int line_no, int iteration,
	std::string var_name, std::string var_value, int is_string_var)
{
	FLineNumber = line_no;
	FIteration = iteration;
	FVariableName.assign(var_name);
	FVariableValue.assign(var_value);
	FIsStringVar = (is_string_var == 0 ? 0 : 1); // make the meaning boolean
}
//---------------------------------------------------------------------------
int TDebugVariable::GetLineNumber()
{
	return FLineNumber;
}
//---------------------------------------------------------------------------
int TDebugVariable::GetIteration()
{
	return FIteration;
}
//---------------------------------------------------------------------------
std::string TDebugVariable::GetVariableName()
{
	return FVariableName;
}
//---------------------------------------------------------------------------
std::string TDebugVariable::GetVariableValue()
{
	return FVariableValue;
}
//---------------------------------------------------------------------------
int TDebugVariable::GetIsStringVar()
{
	return FIsStringVar;
}
//---------------------------------------------------------------------------
void TDebugVariable::SetLineNumber( int line_no )
{
	FLineNumber = line_no;
}
//---------------------------------------------------------------------------
void TDebugVariable::SetIteration( int iteration )
{
	FIteration = iteration;
}
//---------------------------------------------------------------------------
void TDebugVariable::SetVariableName( std::string var_name )
{
	FVariableName.assign(var_name);
}
//---------------------------------------------------------------------------
void TDebugVariable::SetVariableValue( std::string var_value )
{
	FVariableValue.assign(var_value);
}
//---------------------------------------------------------------------------
void TDebugVariable::SetIsStringVar( int is_string_var )
{
	// Stored as int, but evaluated as a bool
	FIsStringVar = (is_string_var == 0 ? 0 : 1);
}
//---------------------------------------------------------------------------

/////////////////////////////////////////////////
/// Class TDebugLine
/////////////////////////////////////////////////
TDebugLine::TDebugLine(int line_no, bool line_executed, std::string logic)
{
	FLineNumber = line_no;
	FLineExecuted = line_executed;
	FLogic.assign(logic);
} // constructor
//---------------------------------------------------------------------------
TDebugLine::~TDebugLine()
{
	clear_variables_map();
} // destructor
//---------------------------------------------------------------------------
void TDebugLine::clear_variables_map()
{
	// FVariablesMap is keyed on iteration, so in a loop there may be
	// multiple entries, each containing a std::map of objects to be freed.
	if (!FVariablesMap.empty())
	{
		VariablesMap::iterator it = FVariablesMap.begin();

		while (it != FVariablesMap.end())
		{
			VariableObjectsMap t_vars = (*it).second;
			// t_vars is keyed on variable name, and contains TDebugVariable
			// objects.
			VariableObjectsMap::iterator it_obj = t_vars.begin();
			while (it_obj != t_vars.end())
			{
				TDebugVariable* t_bugvar = (*it_obj).second;
				delete t_bugvar;

				it_obj++;
			}

			it++;
		}
		FVariablesMap.clear();
	}
} // clear_variables_map
//---------------------------------------------------------------------------
void TDebugLine::AddDebugVariable(TDebugVariable* bugvar)
{
	int t_iteration = bugvar->GetIteration();
	std::string t_var_name = bugvar->GetVariableName();

	// Retrieve the VariableObjectsMap for this iteration
	// (std::map will create one if it does not exist yet)
	VariableObjectsMap t_vars = FVariablesMap[t_iteration];
	// Add the bugvar to this map (that is, it has the value
	// of the variable as of the FLineNumber of this object
	// and the iteration of the loop (if in a loop; there will
	// be only one entry in FVariablesMap if the line is not
	// in a loop).
	t_vars[t_var_name] = bugvar;

	// In case it didn't exist before, add the VariableObjectsMap back
	// to FVariablesMap
	FVariablesMap[t_iteration] = t_vars;
} // AddDebugVariable (1)
//---------------------------------------------------------------------------
void TDebugLine::AddDebugVariable(int iteration,
	std::string var_name, std::string value, int is_string_var)
{
	TDebugVariable* t_bugvar = new TDebugVariable(FLineNumber, iteration,
		var_name, value, is_string_var);
	AddDebugVariable(t_bugvar);
} // AddDebugVariable (2)
//---------------------------------------------------------------------------
std::string TDebugLine::GetLogic()
{
	return FLogic;
}
//---------------------------------------------------------------------------
int TDebugLine::GetLineNumber()
{
	return FLineNumber;
}
//---------------------------------------------------------------------------
bool TDebugLine::GetLineExecuted()
{
	return FLineExecuted;
}
//---------------------------------------------------------------------------
int TDebugLine::GetVariablesCount()
{
	// FVariablesMap is keyed on iteration. Within each iteration is a
	// COMPLETE LIST of the variables as of this line of executing logic.
	// So just look at the first entry in FVariablesMap and return
	// the count of VariableObjectsMap objects.
	int t_count = 0;
	if (!FVariablesMap.empty())
	{
		// Get all of the variables for the first (and possibly only)
		// iteration. Note that the key is not necessarily 1 or 0, but
		// whatever the iteration loop counter provided when the variable
		// object was created. Use an iterator!
		VariablesMap::iterator it = FVariablesMap.begin();

		if (it != FVariablesMap.end())
		{
			VariableObjectsMap t_vars = (*it).second;

			t_count = (int)t_vars.size();
		}
	}
	return t_count;
} // GetVariableCount()
//---------------------------------------------------------------------------
int TDebugLine::GetVariableIterationsCount()
{
	// FVariablesMap stores one entry for each iteration of the
	// line of logic. So the size of FVariablesMap is the number
	// of iterations.
	//return (int)FVariablesMap.size();
	int t_count = 0;
	if (!FVariablesMap.empty())
	{
		VariablesMap::iterator it = FVariablesMap.begin();
		while (it != FVariablesMap.end())
		{
			t_count++;
			it++;
		}
	}
	return t_count;
} // GetVariableIterationsCount()
//---------------------------------------------------------------------------
TDebugVariable* TDebugLine::GetDebugVariable(int iteration, int index)
{
	// Note: parameter "index" is zero-based
	TDebugVariable* t_bugvar = NULL;

	// The VariableObjectsMap is keyed on variable name (so they would
	// automatically sort). This function is responding to a caller
	// requesting the TDebugVariable object at a specific line number
	// and iteration level. That is the caller's loop might look like
	//
	// t_count = GetVariablesCount(iteration=0);
	// for (t_idx = 0; t_idx < t_count; t_idx++)
	//    t_bugvar = GetDebugVariable(0, t_idx);
	if (!FVariablesMap.empty())
	{
		VariableObjectsMap t_vars = FVariablesMap[iteration];
		if (index < (int)t_vars.size())
		{
			int t_check = 0;
			VariableObjectsMap::iterator it_obj = t_vars.begin();
			while (it_obj != t_vars.end())
			{
				if (t_check == index)
				{
					t_bugvar = (*it_obj).second;
					// got it!
					break;
				}
				t_check++;
				it_obj++;
			}
		}
	}
	return t_bugvar;
} // GetDebugVariable
//---------------------------------------------------------------------------
int TDebugLine::GetVariablesIteration(int iteration_idx)
{
	// FVariablesMap is keyed on iteration, but that is not the same thing
	// as an incrementing index. Depending upon conditions in the code,
	// there might be three elements, with keys 3,5,7. In order for GetDebugVariableName
	// and GetDebugVariableValue to return the correct information, we have to
	// provide a way to know what the iteration_idx'th actual iteration (the key) is.
	int t_iteration = -1; // error flag
	VariablesMap::iterator it = FVariablesMap.begin();
	int t_check_idx = 0;
	while (it != FVariablesMap.end())
	{
		if (t_check_idx == iteration_idx)
		{
			t_iteration = (*it).first;
			break;
		}

		t_check_idx++;
		it++;
	}

	return t_iteration;
} // GetVariableIteration
//---------------------------------------------------------------------------
/////////////////////////////////////////////////
/// Class TDebugEdit
/////////////////////////////////////////////////
TDebugEdit::TDebugEdit()
{
	// constructor
}
//---------------------------------------------------------------------------
TDebugEdit::~TDebugEdit()
{
	clear_debug_lines();
	FPosAtLine.clear();
} // destructor
//---------------------------------------------------------------------------
void TDebugEdit::clear_debug_lines()
{
	if (!FDebugLines.empty())
	{
		std::map<int, TDebugLine*>::iterator it = FDebugLines.begin();
		while (it != FDebugLines.end())
		{
			TDebugLine* t_debug_line = (*it).second;
			delete t_debug_line;

			it++;
		}

		FDebugLines.clear();
	}
} // clear_debug_lines
//---------------------------------------------------------------------------
void TDebugEdit::ParseCompiledLogic(const std::string compiled)
{
	// TRunEdits::debug_edit uses this to compare the parameter "pos"
	// to the contents of this list in order to find the corresponding
	// line of code in the vector of edit logic lines. See also
	// FindCompiledLogicLine in this class.
	std::string t_parse;
	std::string t_eol_mark = "\n";

	t_parse.assign(compiled);
	// Make sure this ends with a line feed (it simplifies the loop, below)
	t_parse += t_eol_mark;

	// Start fresh!
	FPosAtLine.clear();

	if (compiled.empty())
	{
		// Nothing to do!
		return;
	}
	else
	{
		// The first line starts at pos 0
		FPosAtLine.push_back(0);
	}

	size_t t_len = t_parse.length();

	for (size_t t_pos = 0; t_pos < t_len; t_pos++)
	{
		if (t_parse[t_pos] == '\n')
		{
			FPosAtLine.push_back((int)t_pos +1);
		}
	}

} // ParseCompiledLogic
//---------------------------------------------------------------------------
int TDebugEdit::FindCompiledLogicLine(const size_t pos)
{
	// This is called from TRunEdits::debug_edit
	// We have a position in the compiled logic (param pos)
	// Find the index in FPosAtLine where pos is >= to the value,
	// and < the value in the next highest element.
	size_t t_max = FPosAtLine.size();
	int t_index = -1;

	for (size_t t_idx = 0; t_idx < t_max; t_idx++)
	{
		if (pos >= FPosAtLine[t_idx])
			t_index = (int)t_idx;
		else
			break;
	}

	return t_index;
} // FindCompiledLogicLine
//---------------------------------------------------------------------------
void TDebugEdit::ParseEditLogic(std::string edit_logic)
{
	std::string t_parse,
				t_line;

	t_parse.assign(edit_logic);

	std::string eol_mark;
	eol_mark.assign("\r\n");

	// Let's make the parsing a little easier, and ensure the edit_logic
	// finishes with "\r\n"
	t_parse += eol_mark;

	FParsedEditLogic.clear();

	size_t t_len = t_parse.length();
	size_t t_pos = 0;
	size_t t_eol;

	while (t_pos < t_len )
	{
		t_eol = t_parse.find(eol_mark);
		if (t_eol != std::string::npos)
		{
			t_line = t_parse.substr(0, t_eol);
			// Is this an artifact of SQLizing single quotes for storage
			// in the SQLite database?
			t_line = stringReplace(t_line, "\'\'", "'");
			// change tabs to spaces
			t_line = stringReplace(t_line, "\t", "    ");
			FParsedEditLogic.push_back(t_line);
			t_pos = t_eol+2;
			t_parse = t_parse.substr(t_pos, t_len);
		}
		else
		{
			break;
		}
	}
} // ParseEditLogic
//---------------------------------------------------------------------------
size_t TDebugEdit::GetEditLogicLinesCount()
{
	return FParsedEditLogic.size();
} // GetEditLogicLinesCount
//---------------------------------------------------------------------------

std::string TDebugEdit::GetEditLogicAtLine(int line_no)
{
	std::string t_logic;

	if (line_no < (int)FParsedEditLogic.size())
	{
    	t_logic.assign(FParsedEditLogic[line_no]);
	}

	return t_logic;
}
//---------------------------------------------------------------------------
bool TDebugEdit::HasDebugLine(int line_no)
{
	// Tell the caller whether any debug information was created for line_no
	bool t_has_debug_line = false;

	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_has_debug_line = true;
		}
	}

	return t_has_debug_line;
} // HasDebugLine
//---------------------------------------------------------------------------
void TDebugEdit::AddDebugLine(int line_no, bool line_executed, std::string logic)
{
	TDebugLine* t_line = new TDebugLine(line_no, line_executed, logic);
	FDebugLines[line_no] = t_line;
}
//---------------------------------------------------------------------------
int TDebugEdit::GetLineNumberAtMapIndex(int index)
{
	int t_line_no = -1;

	// The FDebugLines map is keyed on line_no, and the entries may skip
	// some lines in the code. Therefore, the FDebugLines.size() may be less
	// than the total number of lines in the compiled logic.
	// TRunEdits::send_debug_to_callbacks needs to get the correct line number.

	if (index < (int)FDebugLines.size())
	{
		int t_check = 0;
		std::map<int, TDebugLine*>::iterator it = FDebugLines.begin();
		while (it != FDebugLines.end())
		{
			if (t_check == index)
			{
				t_line_no = (*it).first;
				break;
			}

			t_check++;
			it++;
		}
	}

	return t_line_no;
} // GetLineNumberAtMapIndex
//---------------------------------------------------------------------------
int TDebugEdit::GetPosAtLine(int line_no)
{
	if (line_no < (int)FPosAtLine.size())
	{
		return (int)FPosAtLine[line_no];
	}
	else
	{
		return -1; // error condition
	}
}
//---------------------------------------------------------------------------
std::string TDebugEdit::GetLogic(int line_no)
{
	std::string t_logic;
	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_logic.assign(t_line->GetLogic());
		}

	}

	return t_logic;
} // GetLogic
//---------------------------------------------------------------------------
bool TDebugEdit::GetLineExecuted(int line_no)
{
	bool t_executed = false;
	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_executed = t_line->GetLineExecuted();
		}
	}

	return t_executed;
} // GetLineExecuted
//---------------------------------------------------------------------------
int TDebugEdit::GetDebugLinesCount()
{
	return (int)FDebugLines.size();
}
//---------------------------------------------------------------------------
int TDebugEdit::GetDebugVariablesCount(int line_no)
{
	// How many variables exist at this line (remember,
	// each could have multiple iterations)
	int t_count = 0;
	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_count = t_line->GetVariablesCount();
		}
	}

	return t_count;
}
//---------------------------------------------------------------------------
int TDebugEdit::GetDebugVariableIterationsCount(int line_no)
{
	int t_count = 0;

	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_count = t_line->GetVariableIterationsCount();
		}
	}
	return t_count;
}
//---------------------------------------------------------------------------
int TDebugEdit::GetDebugVariablesIteration(int line_no, int iteration_idx)
{
	int t_iteration = -1;   // this is an error flag... means the params are invalid

	// FVariablesMap is keyed on iteration, but that is not the same thing
	// as an incrementing index. Depending upon conditions in the code,
	// there might be three elements, with keys 3,5,7. In order for GetDebugVariableName
	// and GetDebugVariableValue to return the correct information, we have to
	// provide a way to know what the iteration_idx'th actual iteration (the key) is.
	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_iteration = t_line->GetVariablesIteration(iteration_idx);
		}
	}

	return t_iteration;
} // GetDebugVariablesIteration
//---------------------------------------------------------------------------
std::string TDebugEdit::GetDebugVariableName(int line_no, int iteration,
	int var_idx)
{
	std::string t_var_name;

	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			TDebugVariable* t_bugvar = t_line->GetDebugVariable(iteration, var_idx);
			if (t_bugvar)
			{
				t_var_name.assign(t_bugvar->GetVariableName());
			}
		}
	}
	return t_var_name;
} // GetDebugVariableName
//---------------------------------------------------------------------------
std::string TDebugEdit::GetDebugVariableValue(int line_no, int iteration,
	int var_idx)
{
	std::string t_var_value;

	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			TDebugVariable* t_bugvar = t_line->GetDebugVariable(iteration, var_idx);
			if (t_bugvar)
			{
				t_var_value.assign(t_bugvar->GetVariableValue());
			}
		}
	}
	return t_var_value;
} // GetDebugVariableValue
//---------------------------------------------------------------------------
int TDebugEdit::GetDebugVariableIsStringVar(int line_no, int iteration, int var_idx)
{
	int t_is_string_var = 0;

	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			TDebugVariable* t_bugvar = t_line->GetDebugVariable(iteration, var_idx);
			if (t_bugvar)
			{
				t_is_string_var = t_bugvar->GetIsStringVar();
			}
		}
	}
	return t_is_string_var;
} // GetDebugVariableIsStringVar
//---------------------------------------------------------------------------
void TDebugEdit::AddDebugVariable(int line_no, int iteration,
	std::string var_name, std::string value, int is_string_var)
{
	if (!FDebugLines.empty())
	{
		TDebugLine* t_line = FDebugLines[line_no];
		if (t_line)
		{
			t_line->AddDebugVariable(iteration, var_name, value, is_string_var);
		}
	}
} // AddDebugVariable (2)
//---------------------------------------------------------------------------


