//---------------------------------------------------------------------------

#ifndef DebugEditH
#define DebugEditH
//---------------------------------------------------------------------------
#include <string>
#include <map>
#include <vector>

// TDebugVariable hold ONE instance of a variable and its value
// as of a specific line number and loop iteration. These
// objects are created ONLY for lines that have executed.
class TDebugVariable
{
private:
	int FLineNumber;
	int FIteration;
	std::string FVariableName;
	std::string FVariableValue;
	// Added for the EditWriter test bench, so it will know whether to
	// decorate variable values with enclosing quotation marks... lets the
	// user see when a variable value is padded with blank spaces
	int FIsStringVar;
public:
	TDebugVariable(int line_no, int iteration, std::string var_name, std::string value, int is_string_var);

	int GetLineNumber();
	int GetIteration();
	std::string GetVariableName();
	std::string GetVariableValue();
	int GetIsStringVar();

	void SetLineNumber( int line_no );
	void SetIteration( int iteration );
	void SetVariableName( std::string var_name );
	void SetVariableValue( std::string var_value );
	void SetIsStringVar(int is_string_var);
};

// TDebugLine holds ONE instance of a line of code and nests a list of
// TDebugVariable objects (including all loop iterations) sorted by
// iteration, and then by variable name

typedef std::map<std::string, TDebugVariable*> VariableObjectsMap;
typedef std::map<int, VariableObjectsMap> VariablesMap;

class TDebugLine
{
private:
	int FLineNumber;
	bool FLineExecuted;
	std::string FLogic;
	VariablesMap FVariablesMap;

	void clear_variables_map();

public:
	TDebugLine(int line_no, bool line_executed, std::string logic);
	~TDebugLine();

	void AddDebugVariable(TDebugVariable* bugvar);
	void AddDebugVariable(int iteration, std::string var_name, std::string value, int is_string_var);
	int GetLineNumber();
	bool GetLineExecuted();
	std::string GetLogic();
	int GetVariablesCount();
	int GetVariableIterationsCount();
	TDebugVariable* GetDebugVariable(int iteration, int var_idx);
	int GetVariablesIteration(int iteration_idx);
};

// Lastly, TDebugEdit is the uber container for the results of a call
// to API function Edit_DebugEdit
class TDebugEdit
{
private:
	std::map<int, TDebugLine*> FDebugLines;
	// This will be the pos (position) of each
	// line of compiled code. Used by TRunEdits::debug_edit
	// to find the correct corresponding line in edit logic
	std::vector<size_t> FPosAtLine;
	// This is the original edit logic, parsed into individual lines,
	// and used by TRunEdits::debug_edit to return code to the
	// runtime debugger.
	std::vector<std::string> FParsedEditLogic;

	void clear_debug_lines();

public:
	TDebugEdit();
	~TDebugEdit();

	void ParseCompiledLogic(const std::string compiled);
	int FindCompiledLogicLine(const size_t pos);

	void ParseEditLogic(std::string edit_logic);
	std::string GetEditLogicAtLine(int line_no);
	size_t GetEditLogicLinesCount();

	bool HasDebugLine(int line_no);
	void AddDebugLine(int line_no, bool line_executed, std::string logic);
	void AddDebugVariable(int line_no, int iteration, std::string var_name, std::string value, int is_string_var);

	int GetLineNumberAtMapIndex(int idx);
	int GetPosAtLine(int line_no);
	std::string GetLogic(int line_no);
	bool GetLineExecuted(int line_no);

	int GetDebugLinesCount();
	int GetDebugVariableIterationsCount(int line_no);
	int GetDebugVariablesCount(int line_no);
	std::string GetDebugVariableName(int line_no, int iteration, int var_idx);
	std::string GetDebugVariableValue(int line_no, int iteration, int var_idx);
	int GetDebugVariableIsStringVar(int line_no, int iteration, int var_idx);
	int GetDebugVariablesIteration(int line_no, int iteration_idx);
};


#endif
