//---------------------------------------------------------------------------

#include <stdexcept>
#include <vector>
#include <algorithm>
// For the Borland compilers, use #pragma hdrstop for precompiled headers (pch)
#ifdef EE_WINDOWS
	#pragma hdrstop
#endif
#include "EditAPI.h"
#include "SMF.h"
#include "ExternalDLLTracker.h"
//---------------------------------------------------------------------------
//   We start with a capacity for 20 concurrently open metafiles. Edit_Initialize
//	 will allocate additional elements as needed. DllMain will ensure
//   that the contents of the vector are freed at shut-down.
static std::vector <TSMF*> SMFList(20);

// Startup and Shutdown are called from DLLMain. InitToNull and DeleteSMF
// operate on each element of SMFList.

void InitToNull(TSMF* smf)
{
	smf = NULL;
}

void DeleteSMF(TSMF* smf)
{
	delete smf;
}

void Startup()
{
	std::for_each(SMFList.begin(), SMFList.end(), InitToNull);
} // Startup

void Shutdown()
{
	std::for_each(SMFList.begin(), SMFList.end(), DeleteSMF);
	// ClearExternalDLLs is in ExternalDLLTracker.cpp
	ClearExternalDLLs();
} // Shutdown
//---------------------------------------------------------------------------

// InitCheck is used only from within these API interface functions.
static void InitCheck( const int smfID )
{
	if ((int)SMFList.size() < smfID || !SMFList[smfID])
		throw std::invalid_argument("Invalid metafile smfID!");
	// Clear the Engine's error message pump before calling
	// the next SMF functions.
	SMFList[smfID]->ClearEngineErrorsPump();
} // InitCheck
//---------------------------------------------------------------------------
/*   Function:  Edit_Initialize
 *
 *   Description: Opens a metafile and populates a token for use with all other
 *                Engine functions.
 *
 *   Parameters:
			  smfID - Output. Receives the index into the Edit Engine's internal
					  list of open metafiles. The value will be in the range of
					  [0..<n>]. The smfID is then used as the first argument to
					  all other Edit_ functions.

			  smf_name - Input. The fully qualified file name (i.e., includes
					  network/local drive identifier plus path plus file name)
					  of the metafile to be opened.

			  smf_init_error - Output. Buffer to receive error information
					  during initialization (since there will be no TSMF
					  object to query for EngineErrorPump entries).

			  error_buffer_size - Input. The number of bytes in the
					  smf_init_error buffer.

			  version_compare - Output. Receives the comparison value of the metafile
					  version vs. what the Engine expects:
					  0 = metafile is the right version for this version of the Engine
					  1 = metafile is out of date; use Edit_UpdateMetafileStructure
					 -1 = metafile version newer than Engine version; update Engine!

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_Initialize(
		 int* smfID,
		 const char* smf_name,
		 char* smf_init_error,
		 const int error_buffer_size,
		 int* version_compare)
{
	int t_result = 0;
	size_t t_smfIdx;
	size_t t_max = SMFList.size();
	// Initialize *smfID to -1 as a flag that we have not found
	// the metafile in the SMFList yet... and possibly will not
	// be able to open the metafile.
	*smfID = -1;

	TSMF* t_smf = NULL;

	// Look for an open place in vector SMFList
	for (t_smfIdx = 0; t_smfIdx < t_max; t_smfIdx++)
	{
		// Found an unused slot? Great... get outta here.
		if (!SMFList[t_smfIdx])
			break;
	}

	try
	{
		t_smf = new TSMF();
		t_result = t_smf->InitializeMetafile(smf_name,
						 smf_init_error,
						 error_buffer_size,
						 version_compare );

		if (t_result == 0)
		{
			// If we found an open slot in SMFList, re-use it for this new
			// metafile. Otherwise, add a slot to the vector.
			if (t_smfIdx < t_max)
				SMFList[t_smfIdx] = t_smf;
			else
				SMFList.insert(SMFList.end(), t_smf);

			// Return the smfID to the caller
			*smfID = (int)t_smfIdx;
		}
		else
		{
			if (t_smf)
			{
				delete t_smf;
				t_smf = NULL;
			}
		}
	}
	catch (...)
	{
		t_result = -1; // Unable to initialize metafile
		if (t_smf)
		{
			delete t_smf;
			t_smf = NULL;
		}
	}
	return t_result;
} // Edit_Initialize
//---------------------------------------------------------------------------

/*   Function:  Edit_UpdateMetafile
 *
 *   Description: Opens a metafile and evaluates the control tables (EE_*)
 *                making structural changes as needed.
 *
 *   Parameters:
			  smf_name - Input. The fully qualified file name (i.e., includes
					  network/local drive identifier plus path plus file name)
					  of the metafile to be opened and updated.

			  smf_init_error - Output. Buffer to receive error information
					  during initialization (since there will be no TSMF
					  object to query for EngineErrorPump entries).

			  error_buffer_size - Input. The number of bytes in the
					  smf_init_error buffer. Recommend making this 1000 bytes.


 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_UpdateMetafile(
		 const char* smf_name,
		 char* smf_init_error,
		 const int error_buffer_size)
{
	int t_result = 0;
	// We don't use t_version_compare in this context, but it's needed for
	// TSMF constructor. Set t_version_check to 0 (false) so that the ctor
	// does not abort this file-open process.
	int t_version_compare,
		t_version_check = 0;

	TSMF* t_smf = NULL;

	try
	{
		t_smf = new TSMF();
		t_result = t_smf->InitializeMetafile(smf_name,
							   smf_init_error,
							   error_buffer_size,
							   &t_version_compare,
							   t_version_check );

		if (t_result != 0)
		{
			// Close the metafile now
			delete t_smf;
			t_smf = NULL;
		}
		else
		{
			// so far, so good... let's update the metafile
			t_result = t_smf->UpdateMetafile(smf_init_error, error_buffer_size);

			// and now, regardless of the outcome, close this TSMF object
			delete t_smf;
			t_smf = NULL;

		}

	}
	catch (...)
	{
		t_result = -1; // Unable to update metafile
		if (t_smf)
		{
			delete t_smf;
			t_smf = NULL;
		}
	}
	return t_result;
} // Edit_UpdateMetafile
//---------------------------------------------------------------------------

/*   Function:  Edit_Exit
 *
 *   Description: This function is called when EDITS processing is complete. It
 *                closes the metafile and releases all memory that has been
 *                allocated within the Edit Engine.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_Exit(
		 const int smfID )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		delete SMFList[smfID];
		SMFList[smfID] = NULL;
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_Exit
//---------------------------------------------------------------------------

/*   Function:  Edit_CreateNewMetafile
 *
 *   Description: Creates a new *.smf containing just the tables that are
 *                required for all metafiles.
 *
 *                NOTE: This function is called WITHOUT an smfID. It can be
 *                used to set up a brand new metafile, so the smfID is unknown
 *                until the caller subsequently invokes Edit_Initialize.
 *
 *   Parameters:
			  smf_name - Input. The fully qualified file name (i.e., includes
					  network/local drive identifier plus path plus file name)
					  of the metafile to be opened.

 *
 *   Return value: Zero (0) normally, error code (-2) if a system error has occurred.
 */
int EDIT_API Edit_CreateNewMetafile(
		 const char* smf_name)
{
	int t_result = 0;

	try
	{
		// This function is implemented in TSMF, although it is not instantiated
		// as a member of that class.
		t_result = CreateNewMetafile(smf_name);
	}
	catch (...)
	{
		// Failure, NOS.
		t_result = -2;
	}
	return t_result;

} // Edit_CreateNewMetafile
//---------------------------------------------------------------------------
/*   Function:    Edit_CompactMetafile
 *
 *   Description: This function rebuilds the SMF to "reclaim" space left
 *                following extensive deletions.
 *
 *                NOTE: This function is called WITHOUT an smfID. It must be
 *                run when there are no other connections to the metafile
 *                (i.e., this process requires exclusive access).
 *
 *   Parameters:
			  smf_name - Input. The fully qualified file name (i.e., includes
					  network/local drive identifier plus path plus file name)
					  of the metafile to be opened.

 *
 *   Return value: Zero (0) normally, error code (-2) if a system error has occurred.
 */
int EDIT_API Edit_CompactMetafile(
		 const char* smf_name)
{
	int t_result = 0;

	try
	{
		// This function is implemented in TSMF, although it is not instantiated
		// as a member of that class.
		t_result = CompactMetafile(smf_name);
	}
	catch (...)
	{
		// Failure, NOS.
		t_result = -2;
	}
	return t_result;

} // Edit_CompactMetafile
//---------------------------------------------------------------------------
/*   Function:    Edit_GetEngineVersion
 *
 *   Description:  Returns the current version of the Edit Engine (EDITS50). The
 *                 version will take the form [Major].[Minor].[Release].[Build]
 *
 *                 NOTE: This function is called WITHOUT an smfID. A caller should
 *                 always be able to determine the Engine version without having
 *                 to open a metafile.
 *
 *   Parameters:
			  version_buffer - Output. The buffer to receive the version number.
				  We recommend making it 20 bytes, which is overkill but virtually
				  ensures it will always be large enough. The buffer will be returned
				  with a NULL terminator. See top of SMF.cpp for
				  ENGINE_VERSION constant.

			  buffer_length - Input. The actual size of the version_buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetEngineVersion(
		 char* version_buffer,
		 int buffer_length )
{
	int t_result = 0;

	try
	{
		// This function is implemented in TSMF, although it is not instantiated
		// as a member of that class.
		t_result = GetEngineVersion(version_buffer, buffer_length);
	}
	catch (...)
	{
		// Failure, NOS.
		t_result = -2;
	}
	return t_result;

} // Edit_GetEngineVersion
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
/*   Function:  Edit_GetMetafileStructureVersion
 *
 *   Description: Retrieves the metafile STRUCTURE version (a string that should be
 *                large enough to accommodate version changes over time),
 *                set by the maintainer of the Engine. This value is stored in
 *                a metafile and checked when the metafile is opened to make sure
 *                the table structures conform to what the current version of the
 *                Engine expects to see.
 *
 *                NOTE: This does not require a smfID. It is returning information
 *                applicable to ALL metafiles.
 *
 *   Parameters:
			  smf_structure_version - Output. The buffer to receive the metafile
					  structure version number. See top of SMF.cpp for
					  METAFILE_STRUCTURE_VERSION constant.

			  smf_version_size - Input. The number of bytes in the smf_version
					  buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetMetafileStructureVersion(
		 char* smf_structure_version,
		 int smf_version_size)
{
	int t_result = 0;

	try
	{
		// Note: This is called WITHOUT an smfID
		t_result = GetMetafileStructureVersion(smf_structure_version, smf_version_size);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}
	return t_result;

} // Edit_GetMetafileStructureVersion
//---------------------------------------------------------------------------



/*   Function:  Edit_GetMetafileVersion
 *
 *   Description: Retrieves the metafile version (a string that could be
 *                of any size), set by the metafile author.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  smf_version - Output. The buffer to receive the metafile's
					  version identifier.

			  smf_version_size - Input. The number of bytes in the smf_version
					  buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetMetafileVersion(
		 int smfID,
		 char* smf_version,
		 int smf_version_size)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetMetafileVersion(smf_version, smf_version_size);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}
	return t_result;

} // Edit_GetMetafileVersion
//---------------------------------------------------------------------------

/*   Function:  Edit_GetMetafileComment
 *
 *   Description: Retrieves the metafile comment (a string that could be
 *                of any size), set by the metafile author.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  smf_comment - Output. The buffer to receive the metafile's comment.

			  smf_comment_size - Input. The number of bytes in the smf_comment
					  buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetMetafileComment(
		 int smfID,
		 char* smf_comment,
		 int smf_comment_size)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetMetafileComment(smf_comment, smf_comment_size);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}
	return t_result;

} // Edit_GetMetafilecomment
//---------------------------------------------------------------------------

/*   Function:    Edit_GetEngineErrorsCount
 *
 *   Description: Retrieves the number of system errors that occurred during
 *                the execution of the most recent call to an EDITS API function.
 *
 *                For all API functions (except this one and Edit_GetEngineErrorByIndex)
 *                if the Engine encounters a condition it cannot handle, it will
 *                store a message in the "Engine errors pump", and return a
 *                non-zero result code. The Engine client should then call
 *                this function to find out how many errors there are, and
 *                Edit_GetEngineErrorByIndex to retrieve each of the messages.
 *
 *                NOTE: The next call to an EDITS API function will invoke
 *                InitCheck() (see above in this unit), which will clear out
 *                the Engine errors pump.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  how_many_errors - Output. Receives the number of error messages
				 currently stored in the Engine errors pump.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred,
 *   error code (-2) if the smfID parameter is invalid.
 */
int EDIT_API Edit_GetEngineErrorsCount(
		 const int smfID,
		 int* how_many_errors )
{
	// Note: Do NOT call InitCheck() from here! That routine clears
	// the error message pump.
	if ((int)SMFList.size() < smfID || !SMFList[smfID])
	{
		return -2;
	}

	int t_result = 0;

	try
	{
		t_result = SMFList[smfID]->GetEngineErrorsCount(how_many_errors);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetEngineErrorsCount
//---------------------------------------------------------------------------
/*   Function:     Edit_GetEngineErrorByIndex
 *
 *   Description:  Retrieves the idx'th error message from the Engine message
 *                 pump.
 *
 *                For all API functions (except this one and Edit_GetEngineErrorsCount)
 *                if the Engine encounters a condition it cannot handle, it will
 *                store a message in the "Engine errors pump", and return a
 *                non-zero result code. The Engine client should then call
 *                Edit_GetEngineErrorsCount to find out how many errors there are,
 *                and then this function to retrieve each of the messages.
 *
 *                NOTE: The next call to an EDITS API function will invoke
 *                InitCheck() (see above in this unit), which will clear out
 *                the Engine errors pump.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  idx - Input. Index into list of error messages in the Engine errors
				 pump. NOTE: This is a 1-based list; that is, the first message
				 should be retrieved by passing idx=1, and the last message should
				 use idx=[return value from EditGetEngineErrorsCount].

			  error_message - Output. Receives the idx'th error message. NOTE:
				 Error messages are constructed "on the fly" so that they may
				 contain context-sensitive information. This function will
				 retrieve only as many bytes of the message as indicated by
				 parameter message_length. Therefore, messages could be truncated
				 if the error_message buffer is not large enough. We are using
				 buffers of 200 bytes; your results may vary.

			  message_length - Input. The size of the error_message buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred,
 *   error code (-2) if the smfID is invalid.
 */
int EDIT_API Edit_GetEngineErrorByIndex(
		 const int smfID,
		 const int idx,
		 char* error_message,
		 const int message_length)
{
	// Note: Do NOT call InitCheck() from here! That routine clears
	// the error message pump.
	if ((int)SMFList.size() < smfID || !SMFList[smfID])
	{
		return -2;
	}

	int t_result = 0;

	try
	{
		t_result = SMFList[smfID]->GetEngineErrorByIndex(idx, error_message, message_length);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetEngineErrorByIndex
//---------------------------------------------------------------------------

/*   Function:  Edit_GetPKeys
 *
 *   Description: The EDITS tools require that every table in a metafile has
 *                as its first column an auto-incrementing integer field
 *                that is always named PKey (shorthand for Primary Key).
 *
 *                Retrieves the primary keys for all rows in the designated
 *                table. The caller states the sort order, and passes in a
 *                pointer to a callback function to receive the list. Useful
 *                for applications such as EditWriter, which needs to populate
 *                a Navigator with all of the objects in a table.
 *
 *                For each PKey, the caller invokes Edit_GetDataByFieldName
 *                or Edit_GetDataByColumnOrder to retrieve the data from the
 *                columns in the associated row of the table.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table from which to retrieve PKeys.

			  order_by - Input. The field name(s) on which the primary keys should sort.

			  owner - Input. this pointer to an object (e.g., a Windows form) that
				 will handle the primary key values being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like CallbackPKeys.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 *   If the result set is empty (that is, nothing is sent to the callback function),
 *   this function returns -2.
 */
int EDIT_API Edit_GetPKeys(
		 const int smfID,
		 const char* table_name,
		 const char* order_by,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetPKeys( table_name, order_by, owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}
	return t_result;

} // Edit_GetPKeys
//---------------------------------------------------------------------------

/*   Function:      Edit_CheckEditSetAgainstLayout
 *
 *   Description:   When the Engine loads an edit set for processing, it analyzes
 *                  the field objects referenced by each edit in the edit set to
 *                  verify that the field is included in specified record layout.
 *                  For any edits where referenced fields are not in the layout,
 *                  the Engine sets a flag. When the edit set is run against a
 *                  data record (e.g., a call to Edit_RunEdits), any edits with
 *                  the flag set will be skipped.
 *
 *                  This behavior occurs silently; the caller receives no feedback
 *                  that edits may have been skipped. This is not a bug; this design
 *                  decision comes to us from the earliest versions of the Engine.
 *
 *                  EditWriter v5 provides a feature that asks the user to identify
 *                  an edit set and a layout, and makes a call to this function.
 *                  Any edits that will be flagged to skip, and the fields that
 *                  cause this action, are returned to EditWriter so that it can
 *                  present a human-readable list to the user.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_pkey - Input. The primary key (PKey) that uniquely identifies
				 the edit set.

			  layout_pkey - Input. The primary key (PKey) that uniquely identifies
				 the layout.

			  owner - Input. this pointer to an object (e.g., a Windows form) that
				 will handle the values being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackCheckEditSetAgainstLayout.
 *
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 *   If the result set is empty (that is, all fields referenced by all edits exist in
 *   the layout), the function returns -2 as a notification to the caller, so that
 *   a suitable "no worries!" message may be displayed to the user.
 */
int EDIT_API Edit_CheckEditSetAgainstLayout(
		 const int smfID,
		 const int edit_set_pkey,
		 const int layout_pkey,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CheckEditSetAgainstLayout(
		    edit_set_pkey,
			layout_pkey,
			owner,
			callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}
	return t_result;

} // Edit_CheckEditSetAgainstLayout
//---------------------------------------------------------------------------

/*   Function: Edit_GetTableNames
 *
 *   Description: Sends the list of all tables in the metafile to the caller
 *                via the callback function. This includes the MCO tables as
 *                well as lookup tables used in edit logic.
 *
 *                For each table, the caller then retrieves the data structure
 *                using Edit_GetTableStructure.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  owner - Input. this pointer to an object (e.g., a Windows form) that
				 will handle the data being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackString.
 *
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetTableNames(
		 const int smfID,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetTableNames( owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetTableNames
//---------------------------------------------------------------------------
/*   Function:  Edit_GetTablesCount
 *
 *   Description: Edit_GetTablesCount is used with Edit_GetTableNameByIndex
 *              to retrieve the names of all the tables in the metafile
 *              without resorting to a callback function.
 *
 *              See also: Edit_GetTableNames (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  how_many_tables - Output. Receives the total number of tables in
				 the metafile.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetTablesCount(
		 const int smfID,
		 int* how_many_tables )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetTablesCount( how_many_tables );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetTablesCount
//---------------------------------------------------------------------------
/*   Function:  Edit_GetTableNameByIndex
 *
 *   Description: Used with Edit_GetTablesCount to retrieve the names of all
 *              of the tables in the metafile without resorting to using a
 *              callback function.
 *
 *              See also: Edit_GetTableNames (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  idx - Input. Index into the Edit Engine's internal list of all
				 tables in the metafile.

			  buffer - Output. The buffer to receive the data.

			  buff_size - Input. The size of buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetTableNameByIndex(
		 const int smfID,
		 const int idx,
		 char* buffer,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );
		t_result = SMFList[smfID]->GetTableNameByIndex( idx, buffer, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetTableNameByIndex
//---------------------------------------------------------------------------
/*   Function:  Edit_GetTableStructure
 *
 *   Description: Retrieves the definitions for all of the fields in the
 *                designated table: field name, data type (and size, for
 *                char and varchar data types) and whether the field
 *                will allow NULL values.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  owner - Input. this pointer to an object (e.g., a Windows form) that
				 will handle the data being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.
 *
 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackTableStructure.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetTableStructure(
		 const int smfID,
		 const char* table_name,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetTableStructure( table_name, owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetTableStructure
//---------------------------------------------------------------------------

/*   Function: Edit_GetColumnsCount
 *
 *   Description:  Used with Edit_GetColumnStructure to retrieve the
 *                 definitions of fields (name, data type, whether it
 *                 can accept NULL values), without resorting to a callback.
 *
 *                 See also: Edit_GetTableStructure (which uses a callback)
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  how_many_columns - Output. Receives the number of columns in
				 the table structure. Use this value for a loop controller
				 to make calls to Edit_GetColumnStructure.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetColumnsCount(
		 const int smfID,
		 const char* table_name,
		 int* how_many_columns )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetColumnsCount( table_name, how_many_columns );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetColumnCount
//---------------------------------------------------------------------------
/*   Function:  Edit_GetColumnStructure
 *
 *   Description:  Used with Edit_GetColumnCount to retrieve the
 *                 definitions of fields (name, data type, whether it
 *                 can accept NULL values), without resorting to a callback.
 *
 *                 See also: Edit_GetTableStructure (which uses a callback)
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  column_order - Input. The nth column of the total number of columns
				as reported by the call to Edit_GetColumnCount. NOTE: This
				function expects the column_order to be 1-base, not 0-base;
				i.e., the first column_order should be 1 and the last should
				be the column count returned by Edit_GetColumnCount.

			  field_name - Output. Receives the name of the field.

			  fldname_size - Input. Size of the field_name buffer.

			  data_type - Output. Receives the data type (and size, if data
				type is char or varchar).

			  type_size - Input. Size of the data_type buffer.

			  nullability - Output. Receives the NULL vs. NOT NULL attribute.

			  null_size - Input. Size of the nullability buffer.

			  is_trigger_field - Output. Receives 1 for "true, is a trigger field"
				  or 0 for "false, is NOT a trigger field"

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetColumnStructure(
		 const int smfID,
		 const char* table_name, // input
		 const int column_order, // input, assumes 1-base, not 0-base
		 char* field_name,       // output
		 const int fldname_size, // input (size of field_name array)
		 char* data_type,        // output
		 const int type_size,    // input (size of data_type array)
		 char* nullability,      // output
		 const int null_size,    // input (size of nullability array)
		 int* is_trigger_field)  // output
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetColumnStructure( table_name, column_order,
			field_name, fldname_size,
			data_type, type_size,
			nullability, null_size,
			is_trigger_field );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}


	return t_result;

} // Edit_GetColumnStructure
//---------------------------------------------------------------------------

/*   Function: Edit_GetColumnOrder
 *
 *   Description: Given a table name and field name, this function reports
 *            the field's column order. NOTE: Column order is 1-based,
 *            not 0-based.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  field_name - Input. Name of the field in the table's structure.

			  col_order - Output. Receives the column order of the field.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetColumnOrder(
		 const int smfID,
		 const char* table_name,
		 const char* field_name,
		 int* col_order )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetColumnOrder( table_name, field_name, col_order );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}


	return t_result;
} // Edit_GetColumnOrder
//---------------------------------------------------------------------------

/*   Function: Edit_GetFieldName
 *
 *   Description: Given a table name and column order, this function
 *            reports the column's field name.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  col_order - Input. The 1-based column number of the field.

			  field_name - Output. Receives the name of the field.

			  fldname_size - Input. Size of the field_name buffer.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetFieldName(
		 const int smfID,
		 const char* table_name,
		 const int col_order,
		 char* field_name,
		 const int fldname_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetFieldName( table_name, col_order,
			field_name, fldname_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}


	return t_result;
} // Edit_GetFieldName
//---------------------------------------------------------------------------
/*   Function:  Edit_GetIndexCount
 *
 *   Description:  Used with Edit_GetIndex to retrieve the sql statement
 *                 that created the index definitions.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  how_many_indexes - Output. Receives the number of index definitions
				 for the named table. Use this value for a loop controller
				 to make calls to Edit_GetIndex.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetIndexCount(
		 const int smfID,
		 const char* table_name,
		 int* how_many_indexes )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetIndexCount( table_name, how_many_indexes );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetIndexCount
//---------------------------------------------------------------------------
/*   Function:  Edit_GetIndex
 *
 *   Description:  Retrieves the sql statement used to create the idx'th
 *                 index on this table.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  idx - Input. This is retrieved in a loop, so its the idx'th index
				 out of a total returned from Edit_GetIndexCount.

			  index_sql - Output. Receives the sql statement that was used to
				 create the index. NOTE: Can we just set a recommended size here,
				 please? How about we say make the index_sql buffer 200 bytes?

			  buff_size - Input. The size of the index_sql buffer.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetIndex(
		 const int smfID,
		 const char* table_name,
		 const int idx,
		 char* index_sql,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetIndex( table_name, idx, index_sql, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetIndex
//---------------------------------------------------------------------------
/*   Function:  Edit_GetTriggerCount
 *
 *   Description:  Used with Edit_GetTrigger to retrieve the sql statement
 *                 that created the trigger definitions.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  how_many_triggers - Output. Receives the number of trigger definitions
				 for the named table. Use this value for a loop controller
				 to make calls to Edit_GetTrigger.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetTriggerCount(
		 const int smfID,
		 const char* table_name,
		 int* how_many_triggers )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetTriggerCount( table_name, how_many_triggers );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetTriggerCount
//---------------------------------------------------------------------------
/*   Function:  Edit_GetTrigger
 *
 *   Description:  Retrieves the sql statement used to create the idx'th
 *                 trigger on this table.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table in the metafile.

			  idx - Input. This is retrieved in a loop, so its the idx'th index
				 out of a total returned from Edit_GetTriggerCount.

			  trigger_sql - Output. Receives the sql statement that was used to
				 create the trigger. NOTE: Can we just set a recommended size here,
				 please? How about we say make the trigger_sql buffer 1000 bytes?

			  buff_size - Input. The size of the trigger_sql buffer.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetTrigger(
		 const int smfID,
		 const char* table_name,
		 const int idx,
		 char* trigger_sql,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetTrigger( table_name, idx, trigger_sql, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetTrigger
//---------------------------------------------------------------------------

/*   Function: Edit_ReadData
 *
 *   Description: Given a table name and primary key (PKey), this function
 *            sends the contents of every field in the tuple via a callback.
 *
 *
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  pkey - Input. Primary key value to locate the row in the table.

			  owner - Input. this pointer to an object (e.g., a Windows form)
				 that will handle the data being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackReadData.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_ReadData(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ReadData( table_name, pkey, owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_ReadData
//---------------------------------------------------------------------------

int EDIT_API Edit_RefreshReadData(
		 const int smfID,
		 const char* table_name,
		 const int pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->RefreshReadData( table_name, pkey );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_RefreshReadData
//---------------------------------------------------------------------------

/*   Function:  Edit_GetDataSizeByColumnOrder
 *
 *   Description: Used with Edit_GetDataByColumnOrder. Given a table name and
 *            primary key value, this function reports the size of the buffer
 *            needed to retrieve all of the data in the field designated by
 *            column order.
 *
 *            The caller then allocates a buffer of data_size bytes and
 *            uses it to call Edit_GetDataByColumnOrder, thereby retrieving
 *            the data from the metafile table.
 *
 *            This process is for users who do not want to (or cannot) use
 *            a callback. See also Edit_ReadData (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  pkey - Input. Primary key for the tuple from which the data is to
				 be retrieved.

			  col_order - Input. The column order number of the field.

			  data_size - Output. The size of the data in the designated column.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetDataSizeByColumnOrder(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 const int col_order,
		 int* data_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetDataSizeByColumnOrder( table_name, pkey, col_order, data_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}


	return t_result;

} // Edit_GetDataSizeByColumnOrder
//---------------------------------------------------------------------------
/*   Function: Edit_GetDataByColumnOrder
 *
 *   Description: Used with Edit_GetDataSizeByColumnOrder. Given a table name
 *            and primary key value, plus an appropriately-sized buffer, this
 *            function retrieves the data from the field designated by
 *            column order.
 *
 *            This process is for users who do not want to (or cannot) use
 *            a callback. See also Edit_ReadData (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  pkey - Input. Primary key for the tuple from which the data is to
				 be retrieved.

			  col_order - Input. The column order number of the field.

			  buffer - Output. The buffer to receive the data.

			  buff_size - Input. The size of buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetDataByColumnOrder(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 const int col_order,
		 char* buffer,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetDataByColumnOrder( table_name, pkey, col_order, buffer, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetDataByColumnOrder
//---------------------------------------------------------------------------
/*   Function:  Edit_GetDataSizeByFieldName
 *
 *   Description: Used with Edit_GetDataByFieldName. Given a table name and
 *            primary key value, this function reports the size of the buffer
 *            needed to retrieve all of the data in the field designated by
 *            a field name.
 *
 *            The caller then allocates a buffer of data_size bytes and
 *            uses it to call Edit_GetDataByFieldName, thereby retrieving
 *            the data from the metafile table.
 *
 *            This process is for users who do not want to (or cannot) use
 *            a callback. See also Edit_ReadData (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  pkey - Input. Primary key for the tuple from which the data is to
				 be retrieved.

			  field_name - Input. Name of the field in the table.

			  data_size - Output. The size of the data in the designated column.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetDataSizeByFieldName(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 const char* field_name,
		 int* data_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetDataSizeByFieldName( table_name, pkey, field_name, data_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}


	return t_result;

} // Edit_GetDataSizeByFieldName
//---------------------------------------------------------------------------
/*   Function: Edit_GetDataByFieldName
 *
 *   Description: Used with Edit_GetDataSizeByFieldName. Given a table name
 *            and primary key value, plus an appropriately-sized buffer, this
 *            function retrieves the data from the field designated by
 *            the field name.
 *
 *            This process is for users who do not want to (or cannot) use
 *            a callback. See also Edit_ReadData (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. Name of the table in the metafile.

			  pkey - Input. Primary key for the tuple from which the data is to
				 be retrieved.

			  field_name - Input. Name of the field in the table.

			  buffer - Output. The buffer to receive the data.

			  buff_size - Input. The size of buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetDataByFieldName(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 const char* field_name,
		 char* buffer,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetDataByFieldName( table_name, pkey, field_name, buffer, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetDataByColumnOrder
//---------------------------------------------------------------------------

/*   Function: Edit_OpenCursor
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function opens a cursor based on the user-provided
			   SQL statement. The cursor remains "live" until Edit_CloseCursor
			   is invoked. The user provides a name for the cursor, and
			   uses that name whenever any of the cursor access functions
			   are called.

			   NOTE: SQLite cursors are forward-navigable only. To get to the
			   first tuple in the result set, use this function. Then step
			   through the cursor using Edit_CursorNext until it returns an error
			   code indicating it has reached the end. The is no way to
			   step BACK to a previous tuple in the result set. That's just
			   how it works.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor. This can
				 be any alphanumeric combination (A-Z, a-z, 0-9) plus underscore;
				 no other punctuation permitted. The name can be any length.
				 Names might be: LayoutFields, edits_in_edit_set, 1_Agency,
				 or scrap_99. The Engine will store the name for case-insensitive
				 referencing (i.e., we're going to lowercase it for our internal
				 list controller).

			  sql - Input. We will allow only SELECT statements; statements that
				 begin with anything other than SELECT will generate an error.
				 Users are expected to know how to write SQL that can be understood
				 by SQLite, so if your sole experience with SQL is Microsoft Access,
				 you'll want to brush up on more standard syntax.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_OpenCursor(
		 const int smfID,
		 const char* cursor_name,
		 const char* sql )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->OpenCursor( cursor_name, sql );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_OpenCursor
//---------------------------------------------------------------------------
/*   Function: Edit_CloseCursor
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function closes a cursor that was opened using
			   Edit_OpenCursor. Please close cursors when you don't need
			   them anymore.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_CloseCursor(
		 const int smfID,
		 const char* cursor_name )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CloseCursor( cursor_name );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_CloseCursor
//---------------------------------------------------------------------------

/*   Function: Edit_CursorRowCount
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function reports the number of rows in the cursor.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  row_count - Output. The number of rows (tuples) in the cursor.

 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_CursorRowCount(
		 const int smfID,
		 const char* cursor_name,
		 int* row_count )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CursorRowCount( cursor_name, row_count );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_CursorRowCount
//---------------------------------------------------------------------------
/*   Function: Edit_CursorColCount
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function reports the number of columns in a tuple of the cursor.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  col_count - Output. The number of columns in a tuple of the cursor.

 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_CursorColCount(
		 const int smfID,
		 const char* cursor_name,
		 int* col_count )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CursorColCount( cursor_name, col_count );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_CursorColCount
//---------------------------------------------------------------------------
/*   Function: Edit_CursorNext
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function steps the cursor to the next row (tuple).
			   If the cursor is at the end of file (Eof), this function
			   returns an error code.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.


 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_CursorNext(
		 const int smfID,
		 const char* cursor_name )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CursorNext( cursor_name );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_CursorNext
//---------------------------------------------------------------------------
/*   Function: Edit_CursorEof
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
 			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function reports whether the cursor is at end of its
			   result set.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  is_eof - Output. Receives 0 (false) or 1 (true).


 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_CursorEof(
		 const int smfID,
		 const char* cursor_name,
		 int* is_eof )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CursorEof( cursor_name, is_eof );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_CursorEof
//---------------------------------------------------------------------------
/*   Function: Edit_ReadCursorData
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			   This function reports data for all columns of a tuple
			   by sending the information to a callback function.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  owner - Input. this pointer to an object (e.g., a Windows form) that
				 will handle the data being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackReadData.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_ReadCursorData(
		 const int smfID,
		 const char* cursor_name,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ReadCursorData( cursor_name, owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit ReadCursorData
//---------------------------------------------------------------------------
/*   Function:  Edit_GetCursorDataSizeByNumber
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

 *            Used with Edit_GetCursorDataByNumber. Given a cursor name
 *            this function reports the size of the buffer needed to
 *            retrieve all of the data in the field designated by
 *            column order.
 *
 *            The caller then allocates a buffer of data_size bytes and
 *            uses it to call Edit_GetCursorDataByNumber, thereby retrieving
 *            the data from the named cursor.
 *
 *            This process is for users who do not want to (or cannot) use
 *            a callback. See also Edit_ReadCursorData (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  col_number - Input. The column order number of the field.

			  data_size - Output. The size of the data in the designated column.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetCursorDataSizeByNumber(
		 const int smfID,
		 const char* cursor_name,
		 const int col_number,
		 int* data_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetCursorDataSizeByNumber( cursor_name, col_number, data_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetCursorDataSizeByNumber
//---------------------------------------------------------------------------
/*   Function:  Edit_GetCursorDataByNumber
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			  Used with Edit_GetCursorDataByNumber. Given a cursor name and
			  an appropriately-sized buffer, this function retrieves the data
			  from the field designated by column number.

			  This process is for users who do not want to (or cannot) use
			  a callback. See also Edit_ReadCursorData (which uses a callback).

 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  col_number - Input. The column order number of the field.

			  buffer - Output. The buffer to receive the data.

			  buff_size - Input. The size of buffer.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetCursorDataByNumber(
		 const int smfID,
		 const char* cursor_name,
		 const int col_number,
		 char* buffer,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetCursorDataByNumber( cursor_name, col_number, buffer, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetCursorDataByNumber
//---------------------------------------------------------------------------

/*   Function:  Edit_GetCursorDataSizeByName
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

 *            Used with Edit_GetCursorDataByName. Given a cursor name
 *            this function reports the size of the buffer needed to
 *            retrieve all of the data in the field designated by
 *            field name.
 *
 *            The caller then allocates a buffer of data_size bytes and
 *            uses it to call Edit_GetCursorDataByName, thereby retrieving
 *            the data from the named cursor.
 *
 *            This process is for users who do not want to (or cannot) use
 *            a callback. See also Edit_ReadCursorData (which uses a callback).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  field_name - Input. The name of the field.

			  data_size - Output. The size of the data in the designated column.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetCursorDataSizeByName(
		 const int smfID,
		 const char* cursor_name,
		 const char* field_name,
		 int* data_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetCursorDataSizeByName( cursor_name, field_name, data_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetCursorDataSizeByName
//---------------------------------------------------------------------------

/*   Function:  Edit_GetCursorDataByName
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			  Used with Edit_GetCursorDataByName. Given a cursor name and
			  an appropriately-sized buffer, this function retrieves the data
			  from the field designated by field name.

			  This process is for users who do not want to (or cannot) use
			  a callback. See also Edit_ReadCursorData (which uses a callback).

 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  field_name - Input. The name of the field.

			  buffer - Output. The buffer to receive the data.

			  buff_size - Input. The size of buffer.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetCursorDataByName(
		 const int smfID,
		 const char* cursor_name,
		 const char* field_name,
		 char* buffer,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetCursorDataByName( cursor_name, field_name, buffer, buff_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetCursorDataByName
//---------------------------------------------------------------------------
/*   Function:  Edit_GetCursorIntByNumber
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			  Given a cursor name, retrieves the integer data in the field
			  identified by column number.

			  This process is for users who do not want to (or cannot) use
			  a callback. See also Edit_ReadCursorData (which uses a callback).

 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  col_number - Input. The column order number of the field.

			  data - Output. Receives the integer value.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetCursorIntByNumber(
		 const int smfID,
		 const char* cursor_name,
		 const int col_number,
		 int* data )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetCursorIntByNumber( cursor_name, col_number, data );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetCursorIntByNumber
//---------------------------------------------------------------------------

/*   Function:  Edit_GetCursorIntByName
 *
 *   Description: The direct cursor access functions include:
			   Edit_OpenCursor
			   Edit_CursorRowCount
			   Edit_CursorColCount
			   Edit_ReadCursorData
			   Edit_CursorNext
			   Edit_CursorEof
			   Edit_CloseCursor
			   // for data retrieval without use of callbacks, try:
			   Edit_GetCursorDataSizeByNumber
			   Edit_GetCursorDataByNumber
			   Edit_GetCursorDataSizeByName
			   Edit_GetCursorDataByName
			   // to retrieve integer values (particularly PKey):
			   Edit_GetCursorIntByNumber
			   Edit_GetCursorIntByName

			  Given a cursor name, retrieves the integer data in the field
			  identified by field name.

			  This process is for users who do not want to (or cannot) use
			  a callback. See also Edit_ReadCursorData (which uses a callback).

 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  field_name - Input. The name of the field.

			  data - Output. Receives the integer value.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetCursorIntByName(
		 const int smfID,
		 const char* cursor_name,
		 const char* field_name,
		 int* data )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetCursorIntByName( cursor_name, field_name, data );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetCursorIntByName
//---------------------------------------------------------------------------

/****************************************************************************/
/*************        Functions for Running Edits              **************/
/****************************************************************************/
/*   Function:     Edit_UpdateRecordByTag
 *
 *   Description:  The caller passes in the Record Layout Tag and a data record.
 *                 The Engine parses the data according to the Record Layout
 *                 to prepare it for running edits.
 *
 *                 NOTE: This function is used with flat-file data. A separate
 *                 interface will need to be developed to support multi-dimensional
 *                 data, such as XML or HL7.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  layout_tag - Input. The layout tag uniquely identifies a record
				 layout in the metafile, and is less susceptible to errors
				 from typographical mistakes than a full layout name. The tag
				 can be any alphanumeric value, of any length, but we assume
				 it will be something quite brief.

			  data - Input. The data to be edited.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_UpdateRecordByTag(
		 const int smfID,
		 const char* layout_tag,
		 const char* data )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->UpdateRecordByTag( layout_tag, data );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_UpdateRecordByTag
//---------------------------------------------------------------------------

/*   Function:     Edit_UpdateRecordByName
 *
 *   Description:  The caller passes in the Record Layout Name and a data record.
 *                 The Engine parses the data according to the Record Layout
 *                 to prepare it for running edits.
 *
 *                 NOTE: This function is used with flat-file data. A separate
 *                 interface will need to be developed to support multi-dimensional
 *                 data, such as XML or HL7.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  layout_name - Input. The layout name uniquely identifies a record
				 layout in the metafile. The odd naming conventions traditionally
				 used have often led to programmers misspelling the layout name.
				 We recommend using function Edit_UpdateRecordByTag, which takes
				 the much briefer Layout Tag as the unique identifier.

			  data - Input. The data to be edited.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_UpdateRecordByName(
		 const int smfID,
		 const char* layout_name,
		 const char* data )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->UpdateRecordByName( layout_name, data );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_UpdateRecordByName
//---------------------------------------------------------------------------

/*   Function:     Edit_DisplayDataRecordByTag
 *
 *   Description:  Following a call to Edit_UpdateRecordByTag or
 *                 Edit_UpdateRecordByName, a call to this function will
 *                 display the data record parsed into the record layout.
 *                 The description of each field, including the data value
 *                 taken from the "data" argument to Edit_UpdateRecordBy[Tag][Name],
 *                 is returned via the callback function.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  layout_tag - Input. The layout tag uniquely identifies a record
				 layout in the metafile, and is less susceptible to errors
				 from typographical mistakes than a full layout name. The tag
				 can be any alphanumeric value, of any length, but we assume
				 it will be something quite brief.

			  owner - Input. this pointer to an object (e.g., a Windows form)
				 that will handle the data being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackDisplayDataRecord.
 *
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_DisplayDataRecordByTag(
		 const int smfID,
		 const char* layout_tag,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DisplayDataRecordByTag( layout_tag, owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_DisplayDataRecordByTag
//---------------------------------------------------------------------------

/*   Function:     Edit_DisplayDataRecordByName
 *
 *   Description:  Following a call to Edit_UpdateRecordByTag or
 *                 Edit_UpdateRecordByName, a call to this function will
 *                 display the data record parsed into the record layout.
 *                 The description of each field, including the data value
 *                 taken from the "data" argument to Edit_UpdateRecordBy[Tag][Name],
 *                 is returned via the callback function.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  layout_name - Input. The layout name uniquely identifies a record
				 layout in the metafile. The odd naming conventions traditionally
				 used have often led to programmers misspelling the layout name.
				 We recommend using function Edit_DisplayDataRecordByTag, which takes
				 the much briefer Layout Tag as the unique identifier.

			  owner - Input. this pointer to an object (e.g., a Windows form)
				 that will handle the data being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackDisplayDataRecord.
 *
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_DisplayDataRecordByName(
		 const int smfID,
		 const char* layout_name,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DisplayDataRecordByName( layout_name, owner, callback_func );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_DisplayDataRecordByName
//---------------------------------------------------------------------------

/*   Function:    Edit_DisplayFieldInMessagesAs
 *
 *   Description: For vendors who frequently use different names for data
				  items from the official names given by standard setters.
				  This function attaches a "display as" name to each field
				  in a record layout, so that the edit message can identify
				  the item by the name used in the vendor's software. For
				  example:

				  02 is not a valid value for Radiation Type [Rad--Regional RX Modality]

				  NAACCR data item 1570 is "Rad--Regional RX Modality", but a vendor
				  instead calls it "Radiation Type". By using this API function
				  after preparing the layout and before running edits,
				  the Engine will display BOTH names in error messages.

				  NOTE: The Engine prepares a layout through a call to
				  Edit_UpdateRecordByTag or Edit_UpdateRecordByName. Therefore,
				  it is necessary to invoke one of these functions, either
				  passing in the first data record to process or passing NULL
				  for the data parameter. Then, call
				  Edit_DisplayFieldInMessagesAs for EACH data item you are
				  renaming prior to running edits. You have to do this ONLY
				  once per layout; that is, you do NOT have to repeat the
				  calls to Edit_DisplayFieldInMessagesAs every time you load
				  another data record with Edit_UpdateRecordByTag.

				  So your processing loop might look like this (pseudo code):

				  // Prepare the layout by invoking this API, passing in
				  // EITHER an actual data record, or NULL.
				  Edit_UpdateRecordByTag(smfID, layout_tag, NULL);

				  // For each data item you are renaming (and ONLY for those;
				  // you don't do anything for fields you are not renaming)
				  for ( < condition >; < limit >; < increment > )
				  {
					  Edit_DisplayFieldInMessagesAs(...) ;
				  }

				  while (!data_source->eof())
				  {
					  // This time, pass in the data record
					  t_data = data_source->GetData() ;

					  Edit_UpdateRecordByTag(smfID, layout_tag, t_data) ;

					  // Run edits
					  Edit_RunEdits(...) ;

					  // Advance to next data record, and load it into
					  // the Engine
					  data_source->next() ;
				  }
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  field_number - Input. The field number that uniquely
				 identifies the data item. This number is determined by the
				 standard setter (e.g., NAACCR Volume II).

			  display_name - Input. The name for the data item to be used
				 in edit error messages.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_DisplayFieldInMessagesAs(
		 const int smfID,
		 const int field_number,
		 const char* display_name )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DisplayFieldInMessagesAs( field_number, display_name );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_DisplayFieldInMessagesAs
//---------------------------------------------------------------------------

/*   Function:     Edit_NeverSkipEmptyField
 *
 *   Description:  This function selectively overrides an edit or edit set option
 *                 to skip an edit if any of its fields are empty. (See
 *                 Edit_RunEdits, parameter edit_options, for setting edit set
 *                 options, and Edit_SetEditOptions, parameter skip_if_empty
 *                 for setting edit options.)
 *
 *                 A frequent use case for invoking this function occurs in
 *                 GenEDITS Plus, when a user has indicated the SKIP EMPTY option
 *                 prior to running edits. Many NAACCR edits reference one or more
 *                 fields with "Over-ride" in their names. Those fields are likely
 *                 to be empty even when all other referenced fields are not. By
 *                 calling Edit_NeverSkipEmptyField for the "Over-ride" fields,
 *                 if case data shows that ONLY the "Over-ride" field is empty,
 *                 the edit will run; however, if at least one other field is
 *                 also empty, then the SKIP EMPTY option is enforced, and the
 *                 edit will not run.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  field_number - Input. The field number that uniquely
				 identifies the data item. This number is determined by the
				 standard setter (e.g., NAACCR Volume II).

			  never_skip_empty - Input. Treat as a boolean; 0 (zero) means false,
				 do not override the SKIP EMPTY edit or edit set option. Any non-zero
				 value means true; edits will override the SKIP EMPTY edit or edit set
				 option if this field is the only empty field referenced by the
				 edit for the current case.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_NeverSkipEmptyField(
		 const int smfID,
		 const int field_number,
		 const int never_skip_empty )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->NeverSkipEmptyField( field_number, never_skip_empty );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_NeverSkipEmptyField
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
/*   Function:     Edit_FieldDataByNumber
 *
 *   Description:  This function retrieves everything there is to be told about
				   a field in a layout. We agree that a function which takes
				   fourteen parameters is ridiculous, but on the other hand,
				   we couldn't convince ourselves that making separate API
				   functions to get each of the elements was all that great a
				   solution, either.

				   So this is a discussion topic for the EDITS Programmers
				   Workgroup. What do YOU want the Engine to do to let you get
				   this information?
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  layout_tag - Input. The EE_Layouts.LayoutTag of the layout containing
				 the field.

			  field_number - Input. The field number as defined by the standard setter
				 (e.g., the NAACCR number).

			  pkey - Output. Receives the primary key of the field in EE_Fields table.

			  start_pos - Output. Receives the starting position of the field in the layout.

			  field_length - Output. Receives the length of the field.

			  field_name - Output. Receives the name of the field. We recommend making
				 it 200 bytes, to accommodate even the most descriptive field name.

			  field_name_len - Input. Size of the field_name buffer.

			  display_name - Output. Receives the display name used in edit messages.
				 We recommend making it 200 bytes.

			  display_name_len - Input. Size of the display_name buffer.

			  data_type - Output. Receives the data type. This is probably either
				 "S " or "D1", so make the buffer at least 3 bytes.

			  data_type_len - Input. The size you made the data_type buffer.

			  value - Output. Receives the data presently loaded into this field.
				 We recommend making it the size of the longest data item in the
				 layout, plus 1 for a NULL terminator. In the NAACCR 14 layout,
				 that would be a buffer of 2001 bytes.

			  value_len - Input. The size you made the value buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_FieldDataByNumber(
		 const int smfID,
		 const char* layout_tag,
		 const int field_number,
		 int* pkey,
		 int* start_pos,
		 int* field_length,
		 char* field_name,
		 const int field_name_len,
		 char* display_name,
		 const int display_name_len,
		 char* data_type,
		 const int data_type_len,
		 char* value,
		 const int value_len )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->FieldDataByNumber(
			layout_tag,
			field_number,
			pkey, start_pos, field_length,
			field_name, field_name_len,
			display_name, display_name_len,
			data_type, data_type_len,
			value, value_len );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_FieldDataByNumber
//---------------------------------------------------------------------------

/*   Function:     Edit_LoadEditSet
 *
 *   Description:  Loads an edit set into memory, and evaluates the fields
 *                 referenced by the edits in that edit set against the
 *                 fields in the designated layout.
 *
 *                 NOTE 1: If your program does not need to suppress edits
 *                 in an edit set prior to running edits, you may skip this
 *                 call and go directly to Edit_RunEdits. That function performs
 *                 this same step the first time the Engine detects a new
 *                 edit set/layout combination.
 *
 *                 NOTE 2: You may have multiple concurrent combinations of
 *                 edit sets and layouts. The Engine will perform the
 *                 field analysis only once per combination, and then hold
 *                 the results of the analysis in memory until the metafile
 *                 is closed.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set to be run.

			  layout_tag - Input. Identifies the record layout that will be
			     used with this edit set.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_LoadEditSet(
		 const int smfID,
		 const char* edit_set_tag,
		 const char* layout_tag )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->LoadEditSet(
			edit_set_tag,
			layout_tag);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_LoadEditSet
//---------------------------------------------------------------------------
/*   Function:     Edit_UnLoadEditSet
 *
 *   Description:  Unloads an edit set/layout pair previously loaded and prepared
 *                 either through a call to Edit_LoadEditSet or Edit_RunEdits.
 *                 This function is needed ONLY for programs such as EditWriter,
 *                 where a user may run edits, then modify the edit set or layout,
 *                 and run the edits again. As noted in the description of
 *                 Edit_LoadEditSet, the Engine performs its analysis of the
 *                 edit set/layout combination at the time the pairing is created,
 *                 and does not revisit that process. It also loads the edits into
 *                 memory, and does not reassess whether any changes to the list
 *                 have occurred between calls to Edit_RunEdits.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set that was previously
				 loaded and paired with the layotu.

			  layout_tag - Input. Identifies the record layout used with this
				 edit set.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_UnloadEditSet(
		 const int smfID,
		 const char* edit_set_tag,
		 const char* layout_tag )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->UnloadEditSet(
			edit_set_tag,
			layout_tag);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_UnloadEditSet
//---------------------------------------------------------------------------
/*   Function:     Edit_GetRunEditsSummary
 *
 *   Description:  For applications that run edits against a collection of data
 *                 records (e.g., GenEDITS Plus), the caller should instruct the
 *                 Engine to reset the error/warning/skip counts to 0 (zero) for
 *                 all of the edits in an edit set/layout pair.
 *
 *                 After iterating through all of the data records and reporting
 *                 results for cases with edit errors, the caller can retrieve
 *                 the summary information needed by invoking Edit_GetRunEditsSummary.
 *                 This function will provide all of these statistics for each
 *                 edit:
 *                 - identifiers Edit Tag, Edit Name and Agency Code
 *                 - total counts of errors, warnings and skips
 *                 - whether the edit always skipped because the layout did not
 *                   contain at least one of the referenced fields
 *                 - whether the edit always skipped because the edit logic did
 *                   not compile (either missing or contains a syntax error)
 *                 - whether the edit was suppressed
 *
 *                 NOTE: Between collections of data files, the caller must
 *                 invoke Edit_UnloadEditSet followed by Edit_LoadEditSet to
 *                 clear all counters and edit-level options. If edit-level
 *                 options are required, the caller must call Edit_SetEditOptions
 *                 prior to invoking Edit_RunEdits on each record in the data
 *                 collection.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set.

			  layout_tag - Input. Identifies the layout.

			  include_zero_counts - Input. Treat as a boolean: 0 (zero) means
				 false; do not report on edits that have zero values for errors,
				 warnings and skips. Any non-zero value means true; report summary
				 data even if the edit did not fail/warn/skip for any data record.

			  total_errors, total_warnings, total_skips, total_messages,
				 total_warning_messages - Output.
				 The total of these values for all edits in the edit set.

			  owner - Input. this pointer to an object (e.g., a Windows form) that
				 will handle the summary values being returned to it through
				 the callback function. (In C++ and Java, the pointer is called
				 "this". In Delphi/Object Pascal, it is called "Self". In Visual
				 Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the calling program
				 that will receive the data lobbed at it from this function call.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackGetRunEditsSummary.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
 int EDIT_API Edit_GetRunEditsSummary(
		  const int smfID,
		  const char* edit_set_tag,
		  const char* layout_tag,
		  const int include_zero_counts,
		  int* total_errors,
		  int* total_warnings,
		  int* total_skips,
		  int* total_messages,
		  int* total_warning_messages,
		  void* owner,
		  void* callback_func )
 {
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetRunEditsSummary(
			edit_set_tag,
			layout_tag,
			include_zero_counts,
			total_errors,
			total_warnings,
			total_skips,
			total_messages,
			total_warning_messages,
			owner,
			callback_func);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetRunEditsSummary
//---------------------------------------------------------------------------


/*   Function:     Edit_SetEditOptions
 *
 *   Description:  The caller may set processing instructions for the entire
 *                 edit set using the edit_options parameter to Edit_RunEdits.
 *
 *
 *                 * A user may wish to suppress (skip) an edit in an edit set
 *                 from running, to reduce the "noise" on the edit report.
 *                 This is useful when a user is aware of a chronic problem
 *                 in the data, and wants to see what other issues exist
 *                 prior to taking remedial action.
 *
 *                 * A user may wish to selectively flag a particular edit
 *                 in an edit set to be skipped whenever the data includes
 *                 a <BLANK> value for one or more referenced fields.
 *
 *                 * A user may wish to selectively flag a particular edit
 *                 in an edit set to be skipped whenever the a referenced
 *                 field failed its single-field edit.
 *
 *                 * A user may wish to override an edit set option to skip
 *                 for empty or failed conditions. That is, regardless of the
 *                 edit set options, the named edit should be run, always.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set that contains the
				 edit to be suppressed.

			  edit_tag - Input. Identifies the edit that is to have its "suppress"
				 state toggled.

			  suppress_edit - Input. Treated as a boolean value. 0 (zero) means
				 false, do not suppress (the default). Any non-zero value indicates
				 true, suppress the edit.

			  skip_if_empty - Input. Treated as a boolean value. 0 (zero) means
				 false, do not skip for empty fields (the default). Any non-zero value
				 indicates true, skip the edit if any of the referenced fields
				 contains a <BLANK> value.

			  skip_if_failed - Input. Treated as a boolean value. 0 (zero) means
				 false, do not skip even if a field failed its single-field edit
				 (the default). Any non-zero value indicates true, skip the edit
				 if any of the referenced fields failed its single-field edit.

			  no_skips_override - Input. Treated as a boolean value. 0 (zero) means
				 false, do not override the edit set options about skipping for
				 empty or failed conditions (the default). Any non-zero value
				 indicates true, i.e., regardless of instructions in the edit set
				 options about skipping, always run this edit.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_SetEditOptions(
		 const int smfID,
		 const char* edit_set_tag,
		 const char* edit_tag,
		 const int suppress_edit,
		 const int skip_if_empty,
		 const int skip_if_failed,
		 const int no_skips_override )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->SetEditOptions(
			edit_set_tag,
			edit_tag,
			suppress_edit,
			skip_if_empty,
			skip_if_failed,
			no_skips_override);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_SetEditOptions
//---------------------------------------------------------------------------

/*   Function:     Edit_GetEditOptions
 *
 *   Description:  The caller uses this function to determine if the edit
 *                 in the designated edit set was previously suppressed through
 *                 a call to Edit_SuppressEdit.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set that contains the
				 edit.

			  edit_tag - Input. Identifies the edit about which the caller is
				 requesting the "suppressed" state.

			  is_suppressed - Output. Receives the current suppressed state of
				 the edit within this edit set. Treat as a boolean value:
				 0 (zero) means false, the edit is not suppressed. 1 (one) means
				 true, the edit is suppressed.

			  is_skip_empty - Output. Receives the current skip-empty rule for
				 the edit within this edit set. Treat as a boolean value:
				 0 (zero) means false, the edit does not skip if a field is empty.
				 1 (one) means true, the edit will be skipped if one of the
				 referenced fields is empty (contains <BLANK> value).

			  is_skip_failed - Output. Receives the current skip-failed rule for
				 the edit within this edit set. Treat as a boolean value:
				 0 (zero) means false, the edit does not skip if a field failed
				 its single-field edit. 1 (one) means true, the edit will be
				 skipped if one of the referenced fields failed its
				 single-field edit.

			  is_no_skips_override - Output. Receives the current skip-override rule for
				 the edit within this edit set. Treat as a boolean value:
				 0 (zero) means false, the edit does not override the edit set options.
				 1 (one) means true, the edit overrides edit set options about
				 skip empty/failed, and will always run (unless suppressed, of course).

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetEditOptions(
		 const int smfID,
		 const char* edit_set_tag,
		 const char* edit_tag,
		 int* is_suppressed,
		 int* is_skip_empty,
		 int* is_skip_failed,
		 int* is_no_skips_override )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetEditOptions(
			edit_set_tag,
			edit_tag,
			is_suppressed,
			is_skip_empty,
			is_skip_failed,
			is_no_skips_override);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_GetEditOptions
//---------------------------------------------------------------------------

/*   Function:     Edit_ClearEditOptions
 *
 *   Description:  Resets the edit options for all edits in the edit set identified
 *                 by the edit_set_tag to false (0).
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_ClearEditOptions(
		 const int smfID,
		 const char* edit_set_tag )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ClearEditOptions(edit_set_tag);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_ClearEditOptions
//---------------------------------------------------------------------------


/*   Function:     Edit_RunEdits
 *
 *   Description:  This function runs all of the edits in the designated edit set
 *                 against the data record laid out according to the designated
 *                 layout. The callback parameters return edit error messages
 *                 to the caller.
 *
 *                 The first time this function is invoked for an open metafile,
 *                 the Engine will evaluate the fields referenced in edits
 *                 against the fields in the record layout. For any edits that
 *                 reference fields not in the layout, the edit will be flagged
 *                 to "skip" silently.
 *
 *                 Note that when using this function, it is NOT necessary to
 *                 use the Edit_UpdateRecordByTag or Edit_UpdateRecordByName
 *                 functions to pre-load the case data.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_set_tag - Input. Identifies the edit set to be run.

			  layout_tag - Input. Identifies the record layout that maps
				 the data record.

			  data - Input. The data record to be edited.

			  edit_options - Input. Fine-tunes the behavior or processing.
				 CONSTANT        HEX     DEC  MEANING
				 EE_NOSKIP       0x0000    0  All edits run.
				 EE_SKIPEMPTY    0x0001    1  Edit skips if any field is empty.
				 EE_SKIPFAIL     0x0002    2  Multi-field edits are skipped when
											  any of the referenced fields
											  failed its single-field edit.
				 EE_SUPPRESSWARN 0x0080  128  Suppresses warnings (the edit will
											  run, but the warning will not be
											  returned to the caller).

			  errors_count - Output. Receives the number of data errors found.
				 Note that this count may include multiple "hits" from edits
				 that generate more than one error message. Therefore, the
				 errors_count may be higher than the number of edits that fail.

			  owner - Input. OPTIONAL. this pointer to an object (e.g., a Windows
				 form) that will handle the primary key values being returned to
				 it through the callback function. (In C++ and Java, the pointer
				 is called "this". In Delphi/Object Pascal, it is called "Self".
				 In Visual Basic, it is called "Myself".)

			  callback_func - Input. OPTIONAL. Pointer to a function in the
				 calling program that will receive the data lobbed at it from
				 this function call.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_RunEdits(
		 const int smfID,
		 const char* edit_set_tag,
		 const char* layout_tag,
		 const char* data,
		 const int edit_options,
		 int* errors_count,
		 void* owner,
		 void* callback_func )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->RunEdits(
			edit_set_tag, layout_tag, data,
			edit_options, errors_count,
			owner, callback_func);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_RunEdits
//---------------------------------------------------------------------------

/*   Function:     Edit_GetEditDetails
 *
 *   Description:  During processing of a call to Edit_RunEdits, when an edit
 *                 error has been reported to the message callback function,
 *                 use this function to retrieve the fields referenced by
 *                 this edit and the values for those fields in
 *                 the currently-loaded layout.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_tag - Input. Identifies the edit object

			  owner - Input. "this" pointer to an object (e.g., a Windows
				 form) that will handle the primary key values being returned to
				 it through the callback function. (In C++ and Java, the pointer
				 is called "this". In Delphi/Object Pascal, it is called "Self".
				 In Visual Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the
				 calling program that will receive the data lobbed at it from
				 this function call.
 *
 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects a callback function structured like
 *           CallbackDisplayDataRecord.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_GetEditDetails(
		 const int smfID,
		 const char* edit_tag,
		 void* owner,
		 void* callback_func)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetEditDetails(
			edit_tag, owner, callback_func);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_GetEditDetails
//---------------------------------------------------------------------------

/*   Function:     Edit_DebugEdit
 *
 *   Description:  Called from the EditWriter test bench on the Edit form,
 *                 as well as from any report writer that chooses to offer
 *                 this feature. This function runs the edit identified by
 *                 edit_tag and sends back processing information to the caller
 *                 through the callback functions.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_tag - Input. Identifies the edit to debug.

			  layout_tag - Input. Identifies the layout to use.

			  data - Input. The data to be edited by this edit, laid out
				 according to the record layout identified by layout_tag.

			  owner - Input. Object (e.g., a form in the calling application)
				 that will know what to do with the information being lobbed
				 back via the callback functions.

			  logic_callback - Input. Optional callback function to receive
				 lines of executing code as it is processing.

			  variables_callback - Input. Optional callback function to receive
				 changing values of variables during edit processing.

			  fields_callback - Input. Optional callback function to receive
				 fields referenced by the edit, and the values in the fields
				 (parsed out of data parameter according to positions described
				 in the layout)

			  notes_callback - Input. Optional callback function to receive
				 edit results as they are generated by the Engine.

 *   Note:   Refer to edit_external.h for the definitions of the callback functions.
 *           This function expects callbacks to take these forms:
 *           logic_callback:      CallbackDebugFunction
 *           variables_callback:  CallbackDebugVarsFunction
 *           fields_callback:     CallbackDebugFunction
 *           notes_callback:      CallbackRunEdits
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_DebugEdit(
		 const int smfID,
		 const char* edit_tag,
		 const char* layout_tag,
		 const char* data,
		 void* owner,
		 void* logic_callback,
		 void* variables_callback,
		 void* fields_callback,
		 void* notes_callback)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DebugEdit(
			edit_tag, layout_tag, data,
			owner, logic_callback, variables_callback,
			fields_callback, notes_callback);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_DebugEdit
//---------------------------------------------------------------------------
/*   Function:      Edit_RecompileEditLogic
 *
 *   Description:   This function recompiles the human-readable edit logic,
 *                  and writes it to EE_Edits.CompiledLogic. Recompiling
 *                  edit logic needn't require a user to have access privileges
 *                  to modify an object
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_tag - Input. Uniquely identifies the edit object within
				 this metafile.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has
 *            occurred, -2 if the edit logic did NOT compile without errors,
 *            -3 if the edit logic compiled without errors, but had warnings,
 *            -4 if the edit logic compiled with BOTH errors and warnings.
 *            (user should be advised to run the syntax checker in EditWriter)
 */
int EDIT_API Edit_RecompileEditLogic(
		 const int smfID,
		 const char* edit_tag )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->RecompileEditLogic(edit_tag);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_RecompileEditLogic
//---------------------------------------------------------------------------
/*   Function:      Edit_UpdateAllAccessField
 *
 *   Description:   The following table/field combinations do not require
 *                  a user to be logged in with any access privileges:
 *
 *                  EE_Fields.LocalName
 *
 *                  This function lets EditWriter and its ilk send through
 *                  updates to any of these fields bypassing the access rights
 *                  verification that occurs for all other metafile constituent
 *                  objects.
 *
 *                  NOTE: This function will return a failure if the value
 *                  violates the table's UNIQUE constraint for the field.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  table_name - Input. EE_Edits, EE_EditSets, EE_Layouts or EE_Fields.

			  pkey - Input. The primary key for the record to be updated.

			  field_name - Input. EditTag, EditSetTag, LayoutTag or LocalName.

			  value - Input. The new value to be written to the field.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */

int EDIT_API Edit_UpdateAllAccessField(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 const char* field_name,
		 const char* value)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->UpdateAllAccessField(
			table_name, pkey, field_name, value);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Edit_UpdateAllAccessField
//---------------------------------------------------------------------------

/****************************************************************************/
/*************     API Functions for Parameterized Queries     **************/
/****************************************************************************/
/*   Function:    Edit_PrepareParameterizedCursor
 *
 *   Description: Creates a query object and places it in a prepared state
 *                in SQLite. The sql argument will be a valid SELECT statement
 *                that includes placeholders for values that will later be
 *                bound to the query. For example:
 *
 *                SELECT * FROM TRAC2010 WHERE PKey = ?;
 *
 *                The "?" is the parameter placeholder. See Edit_BindIntParameter
 *                to see how a PKey value can be provided.
 *
 *                NOTE: sql can contain multiple parameter placeholders, of
 *                data type int or text (char*). See Edit_BindTextParameter
 *                to see how to bind a text parameter.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  sql - Input. We will allow only SELECT statements; statements that
				 begin with anything other than SELECT will generate an error.
				 Users are expected to know how to write SQL that can be understood
				 by SQLite, so if your sole experience with SQL is Microsoft Access,
				 you'll want to brush up on more standard syntax.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_PrepareParameterizedCursor(
		 const int smfID,
		 const char* cursor_name,
		 const char* sql )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->PrepareParameterizedCursor( cursor_name, sql );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_PrepareParameterizedCursor
//---------------------------------------------------------------------------

/*   Function:    Edit_BindIntParameter
 *
 *   Description: This function should be run for each integer parameter in
 *                the query. It assigns a value to the parameter, which will
 *                be applied by the SQLite engine when Edit_OpenParameterizedQuery
 *                is called.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  param_num - Input. The parameter to be bound. NOTE: SQLite uses
				 1-based access for the parameters. So a query with one parameter
				 would use 1 for this argument; if it has more than one, it should
				 call this function iteratively, starting from 1 to the total
				 number of parameters.

			  value - Input. The value to be bound to the param_num'th parameter.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_BindIntParameter(
		 const int smfID,
		 const char* cursor_name,
		 const int param_num,
		 const int value )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->BindIntParameter( cursor_name, param_num, value );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_BindIntParameter
//---------------------------------------------------------------------------

/*   Function:    Edit_BindTextParameter
 *
 *   Description: This function should be run for each text parameter in
 *                the query. It assigns a value to the parameter, which will
 *                be applied by the SQLite engine when Edit_OpenParameterizedQuery
 *                is called.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

			  param_num - Input. The parameter to be bound. NOTE: SQLite uses
				 1-based access for the parameters. So a query with one parameter
				 would use 1 for this argument; if it has more than one, it should
				 call this function iteratively, starting from 1 to the total
				 number of parameters.

			  value - Input. The value to be bound to the param_num'th parameter.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */

int EDIT_API Edit_BindTextParameter(
		 const int smfID,
		 const char* cursor_name,
		 const int param_num,
		 const char* value )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->BindTextParameter( cursor_name, param_num, value );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_BindTextParameter
//---------------------------------------------------------------------------

/*   Function:    Edit_OpenParameterizedCursor
 *
 *   Description: Opens the cursor described by the call to
 *                Edit_PrepareParameterizedCursor with its parameters bound by
 *                calls to Edit_BindIntParameter and/or Edit_BindTextParameter.
 *                The resulting cursor may contain zero or several rows.
 *                The data in those rows may be retrieved using
 *                Edit_GetCursorDataByName and/or Edit_GetCursorDataByNumber.
 *
 *                NOTE: Parameterized queries are expected to be called
 *                repeatedly, usually binding new values to the parameters.
 *                Unlike Edit_OpenCursor, you do not want to close a parameterized
 *                query to rebind the parameters. Instead, call
 *                Edit_ResetParameterizedCursor after you have retrieved all the
 *                values, then the appropriate Edit_Bind*Parameter functions,
 *                and then Edit_OpenParameterizedCursor.
 *
 *                When you are finished with the query, call
 *                Edit_ReleaseParameterizedCursor.
 *
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Edit_OpenParameterizedCursor(
		 const int smfID,
		 const char* cursor_name )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->OpenParameterizedCursor( cursor_name );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_OpenParameterizedCursor
//---------------------------------------------------------------------------


/*   Function:    Edit_ResetParameterizedCursor
 *
 *   Description: The SQLite engine requires that a parameterized query be
 *                be reset prior to binding new values to the parameters.
 *                Failure to make this call between a call to Edit_OpenParameterizedCursor
 *                and the next call to Edit_Bind*Parameter will generate an error
 *                from SQLite.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */

int EDIT_API Edit_ResetParameterizedCursor(
		 const int smfID,
		 const char* cursor_name )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ResetParameterizedCursor( cursor_name );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_ResetParameterizedCursor
//---------------------------------------------------------------------------

/*   Function:    Edit_ReleaseParameterizedCursor
 *
 *   Description: This function closes a parameterized query that was created
 *                using Edit_PrepareParameterizedCursor. It is essential that
 *                queries which have been prepared in the SQLite engine be
 *                released (finalized) when you're done with them, or your
 *                operating system may not be able to obtain unique access
 *                to the SQLite database file (*.smf), which can lead to many
 *                headaches.
 *
 *                Just remember these matched sets:
 *
 *                - If you use Edit_OpenCursor, match it with Edit_CloseCursor.
 *                - If you use Edit_PrepareParameterizedQuery, match it with
 *                  Edit_ReleaseParameterizedQuery.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  cursor_name - Input. The user-assigned name for the cursor.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */

int EDIT_API Edit_ReleaseParameterizedCursor(
		 const int smfID,
		 const char* cursor_name )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ReleaseParameterizedCursor( cursor_name );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Edit_ReleaseParameterizedCursor
//---------------------------------------------------------------------------

#ifdef __SQLEW__

/****************************************************************************/
/*************        API Functions for EditWriter             **************/
/****************************************************************************/
/*   Function:    Ew_Login
 *
 *   Description: Determines whether the agency_code and password are a
 *                valid match for what is stored in the EE_Agencies.AdminCode
 *                and EE_Agencies.Password fields.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  agency_code - Input. The login agency code (e.g., NAACCR, SEER).
				 May be blank or NULL.

			  password - Input. Plain-text password, just as the user typed it.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_Login(
		 const int smfID,
		 const char* agency_code,
		 const char* password )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->Login( agency_code, password );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Ew_Login
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLoginsCount
 *
 *   Description:  Retrieves the number of agency codes used as logins to the
 *                 metafile.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  how_many_logins - Output. Receives the number of logins maintained
				 by the access controller (TEw_AccessController).

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLoginsCount(
		 const int smfID,
		 int* how_many_logins )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLoginsCount( how_many_logins );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Ew_GetLoginsCount
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLoginByIndex
 *
 *   Description:  Retrieves the idx'th agency code used to log into the metafile.
 *
 *                 A user may log into a metafile under multiple IDs (agency codes)
 *                 so as to be able to modify objects owned by those agencies.
 *
 *                 EditWriter calls this API function (and Ew_GetLoginsCount)
 *                 in order to display the user's current logins on the login
 *                 dialog.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  idx - Input. Index into the list of logins maintained by the
				 access controller (TEw_AccessController).

			  agency_pkey - Output. Receives the EE_Agencies.PKey for the idx'th agency

			  agency_code - Output. Receives the idx'th agency code.

			  agency_code_size - Input. Size of the agency_code buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLoginByIndex(
		 const int smfID,
		 const int idx,
		 int* agency_pkey,
		 char* agency_code,
		 const int agency_code_size)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLoginByIndex( idx, agency_pkey, agency_code, agency_code_size );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Ew_GetLoginByIndex
//---------------------------------------------------------------------------

/*   Function:     Ew_ValidateUnique
 *
 *   Description:  Validates that the data value in the specified Table.Field
 *                 is unique within this metafile.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The table (generally an EE_* table).

			  pkey - Input. The primary key (PKey) for the record being validated.

			  field_name - Input. The field being validated.

			  value - Input. The proposed contents of the field.

			  data_is_unique - Output. Receives the ruling:
				 1 means "the proposed data is valid/unique within this
					metafile... go ahead and write it to the metafile";
				 0 means "the proposed data matches the value in another
					record; attempt to write will generate constraint failure error"

			  dupe_pkey - Output. In the event the data is NOT unique, this variable
				 receives the PKey value of the duplicate record. The caller should
				 take appropriate steps to inform the user of WHICH record in the table
				 contains the duplicate information.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_ValidateUnique(
		 const int smfID,
		 const char* table_name,
		 const int pkey,
		 const char* field_name,
		 const char* value,
		 int* data_is_unique,
		 int* dupe_pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ValidateUnique( table_name, pkey, field_name,
			value, data_is_unique, dupe_pkey );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;

} // Ew_ValidateUnique
//---------------------------------------------------------------------------


/*   Function:     Ew_BeginTrans()
 *
 *   Description:  Starts a transaction. Use this whenever modifying the metafile
				   in ways that require everything-or-nothing.

				   Write instructions (see Ew_Post) will implicitly wrap a transaction
				   around individual statements, unless the metafile is "in transaction
				   mode" (i.e., Ew_BeginTrans() has been called without a matching
				   Ew_CommitTrans() or Ew_RollbackTrans()).

				   This API does not support nested transactions. (Although
				   SQLite does.)

				   See also Ew_CommitTrans() and Ew_RollbackTrans()
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_BeginTrans(
		 const int smfID  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->BeginTrans();
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_BeginTrans
//---------------------------------------------------------------------------

/*   Function:     Ew_CommitTrans()
 *
 *   Description:  Commits a transaction. When Ew_BeginTrans() has been
				   invoked, Ew_CommitTrans must be called to "close out"
				   the transaction and write the changes to the metafile.

				   Write instructions (see Ew_Post) will implicitly wrap a transaction
				   around individual statements, unless the metafile is "in transaction
				   mode" (i.e., Ew_BeginTrans() has been called without a matching
				   Ew_CommitTrans() or Ew_RollbackTrans()).

				   This API does not support nested transactions. (Although
				   SQLite does.)

				   See also Ew_BeginTrans() and Ew_RollbackTrans()
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_CommitTrans(
		 const int smfID  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CommitTrans();
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_CommitTrans
//---------------------------------------------------------------------------

/*   Function:     Ew_RollbackTrans()
 *
 *   Description:  Rolls back (cancels) a transaction. When Ew_BeginTrans() has been
				   invoked, Ew_RollbackTrans must be called to "close out"
				   the transaction and roll back any changes made by statements
				   contained within the transaction.

				   Write instructions (see Ew_Post) will implicitly wrap a transaction
				   around individual statements, unless the metafile is "in transaction
				   mode" (i.e., Ew_BeginTrans() has been called without a matching
				   Ew_CommitTrans() or Ew_RollbackTrans()).

				   This API does not support nested transactions. (Although
				   SQLite does.)

				   See also Ew_BeginTrans() and Ew_CommitTrans()
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_RollbackTrans(
		 const int smfID  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->RollbackTrans();
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_RollbackTrans
//---------------------------------------------------------------------------

/*   Function:     Ew_SavePoint()
 *
 *   Description:  Starts a transaction. Use this whenever modifying the metafile
				   in ways that require everything-or-nothing.

				   Use this to begin a transaction nested within another transaction.


				   See also Ew_ReleaseSavePoint() and Ew_RollbackToSavePoint()
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  savepoint_name - Input. The name by which this nested transaction will
				 be known to SQLite. Please use alphanumeric characters for the name,
				 no embedded spaces or punctuation. IMPORTANT: You will use this name
				 for a call to Ew_ReleaseSavePoint or Ew_RollbackToSavePoint, so
				 be careful to be consistent about capitalization!

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_SavePoint(
		 const int smfID,
		 const char* savepoint_name  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->SavePoint(savepoint_name);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_SavePoint
//---------------------------------------------------------------------------

/*   Function:     Ew_ReleaseSavePoint()
 *
 *   Description:  Commits and releases a transaction.

				   Use this to complete a transaction nested within another
				   transaction. Note that if an outer transaction invokes
				   ROLLBACK, the updates within the save point transaction
				   will also be rolled back, despite having been committed
				   at this point.

				   See also Ew_SavePoint() and Ew_RollbackToSavePoint()
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  savepoint_name - Input. The name by which this nested transaction will
				 be known to SQLite. This must be EXACTLY the name used when you
				 called Ew_SavePoint

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_ReleaseSavePoint(
		 const int smfID,
		 const char* savepoint_name  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ReleaseSavePoint(savepoint_name);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_ReleaseSavePoint
//---------------------------------------------------------------------------

/*   Function:     Ew_ReleaseSavePoint()
 *
 *   Description:  Rolls back and releases a transaction.

				   Use this to roll back a transaction nested within another
				   transaction. Note that if the outer transaction COMMITs,
				   all updates within its block WILL be saved, and just the
				   updates within the SavePoint block will be rolled back.

				   See also Ew_SavePoint() and Ew_ReleaseSavePoint()
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  savepoint_name - Input. The name by which this nested transaction is
				 known to SQLite. This must be EXACTLY the name used when you
				 called Ew_SavePoint.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_RollbackToSavePoint(
		 const int smfID,
		 const char* savepoint_name  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->RollbackToSavePoint(savepoint_name);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_RollbackToSavePoint
//---------------------------------------------------------------------------

/*   Function:     Ew_Edit
 *
 *   Description:  In order to update the fields in a record of a table, that
				   table must be placed "in edit mode"; i.e., ready to receive
				   modifications. This function determines that the table is not
				   already in edit mode, and then sets a flag to indicate that
				   the table is ready for update.

				   After setting the table into edit mode, use Ew_SetFieldByName
				   or Ew_SetFieldByNumber to prepare the record for writing. Then
				   call Ew_Post to write the data to the table.

				   See also Ew_Cancel, to take the table out of edit mode and
				   cancel pending writes.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.

			  pkey - Input. The row in the table that is being modified. This value
				 will be used in the execution of Ew_Post to build a WHERE clause
				 to the SQL UPDATE statement.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_Edit(
		 const int smfID,
		 const char* table_name,
		 const int pkey  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->Edit(table_name, pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_Edit
//---------------------------------------------------------------------------

/*   Function:     Ew_Insert()
 *
 *   Description:  In order to add a row to a table, the table must be placed
				   "in insert mode".

				   After setting the table into insert mode, use Ew_SetFieldByName
				   or Ew_SetFieldByNumber to prepare the record for writing. Then
				   call Ew_Post to write the data to the table.

				   See also Ew_Cancel, to take the table out of insert mode and
				   cancel pending writes.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_Insert(
		 const int smfID,
		 const char* table_name  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->Insert(table_name);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_Insert
//---------------------------------------------------------------------------

/*   Function:     Ew_Delete()
 *
 *   Description:  Deletes the row in the named table with a PKey
 *                 value of delete_pkey.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.

			  delete_pkey - Input. The PKey value of the row to be deleted.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_Delete(
		 const int smfID,
		 const char* table_name,
		 const int delete_pkey  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->Delete(table_name, delete_pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_Delete
//---------------------------------------------------------------------------

/*   Function:     Ew_Post()
 *
 *   Description:  Writes pending changes to the named table. If the post completes
				   modifications begun with Ew_Edit(), output variable new_pkey
				   will contain the same value as was passed in via Ew_Edit().
				   If the post completes modifications begun with Ew_Insert(),
				   output variable new_pkey will contain the PKey (primary key)
				   value for the newly-added row.

				   See also Ew_Cancel, to take the table out of edit or insert
				   mode and cancel pending writes.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.

			  new_pkey - Output. The primary key for the row that was updated
				 or added.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_Post(
		 const int smfID,
		 const char* table_name,
		 int* new_pkey   )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->Post(table_name, new_pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_Post
//---------------------------------------------------------------------------

/*   Function:     Ew_Cancel()
 *
 *   Description:  Cancels update/insert mode for the named table. Clears
				   pending writes.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_Cancel(
		 const int smfID,
		 const char* table_name  )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->Cancel(table_name);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_Cancel
//---------------------------------------------------------------------------

/*   Function:     Ew_DeleteEditSetsPivot
 *
 *   Description:  Deletes the record in table EE_EditSetEditsPivot for the
				   edit set/edit primary keys passed as parameters.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  editset_pkey - Input. The primary key value to locate in
				 EE_EditSetsEditsPivot.EditSetPKey field.

			  edit_pkey - Input. The primary key value to locate in
				 EE_EditSetsEditsPivot.EditPKey field.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_DeleteEditSetsPivot(
		 const int smfID,
		 const int editset_pkey,
		 const int edit_pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DeleteEditSetsPivot(editset_pkey, edit_pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_DeleteEditSetsPivot
//---------------------------------------------------------------------------

/*   Function:     Ew_AddEditSetsPivot
 *
 *   Description:  Adds a record to table EE_EditSetEditsPivot with the
				   edit set/edit primary keys passed as parameters.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  editset_pkey - Input. The primary key value to use in
				 EE_EditSetsEditsPivot.EditSetPKey field.

			  edit_pkey - Input. The primary key value to use in
				 EE_EditSetsEditsPivot.EditPKey field.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_AddEditSetsPivot(
		 const int smfID,
		 const int editset_pkey,
		 const int edit_pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->AddEditSetsPivot(editset_pkey, edit_pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_AddEditSetsPivot
//---------------------------------------------------------------------------

/*   Function:     Ew_DeleteLayoutsPivot
 *
 *   Description:  Deletes the record in table EE_LayoutFieldsPivot for the
				   layout/field primary keys passed as parameters.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  layout_pkey - Input. The primary key value to locate in
				 EE_LayoutFieldsPivot.LayoutPKey field.

			  field_pkey - Input. The primary key value to locate in
				 EE_LayoutFieldsPivot.FieldPKey field.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_DeleteLayoutsPivot(
		 const int smfID,
		 const int layout_pkey,
		 const int field_pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DeleteLayoutsPivot(layout_pkey, field_pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_DeleteLayoutsPivot
//---------------------------------------------------------------------------

/*   Function:     Ew_AddLayoutsPivot
 *
 *   Description:  Adds a record to table EE_LayoutFieldsPivot with the
				   edit set/edit primary keys passed as parameters.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  layout_pkey - Input. The primary key value to use in
				 EE_LayoutFieldsPivot.LayoutPKey field.

			  field_pkey - Input. The primary key value to use in
				 EE_LayoutFieldsPivot.FieldPKey field.

			  start_pos - Input. The starting position of the field within the layout.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_AddLayoutsPivot(
		 const int smfID,
		 const int layout_pkey,
		 const int field_pkey,
		 const int start_pos )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->AddLayoutsPivot(layout_pkey, field_pkey, start_pos);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_AddLayoutsPivot
//---------------------------------------------------------------------------

/*   Function:     Ew_SetFieldByName()
 *
 *   Description:  For a table in update/insert mode (set via calls to Ew_Edit()
				   or Ew_Insert()), the new values of fields are established
				   through calls to this function, or its counterpart,
				   Ew_SetFieldByNumber().

				   Note that you do not have to provide a value for every field.
				   Calls to this function set an internal field-level property to
				   indicate that the value was modified. If that flag is not set,
				   the field will be excluded from the SQL INSERT or UPDATE
				   statement, thus leaving intact the existing values of fields
				   not specifically marked for change.

				   After all of the field changes have been made, call Ew_Post()
				   to write the changes to the table. See also Ew_Cancel() to
				   clear pending writes.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.

			  field_name - Input. The name of the field.

			  value - Input. The value (data) to be written to the named field
				 in the named table.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_SetFieldByName(
		 const int smfID,
		 const char* table_name,
		 const char* field_name,
		 const char* value )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->SetFieldByName(table_name, field_name, value);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_SetFieldByName
//---------------------------------------------------------------------------

/*   Function:     Ew_SetFieldByNumber()
 *
 *   Description:  For a table in update/insert mode (set via calls to Ew_Edit()
				   or Ew_Insert()), the new values of fields are established
				   through calls to this function, or its counterpart,
				   Ew_SetFieldByName(). The "by number" option is handy for
				   abstract data processing, although admittedly not all that
				   useful in the context of EditWriter.

				   Note that you do not have to provide a value for every field.
				   Calls to this function set an internal field-level property to
				   indicate that the value was modified. If that flag is not set,
				   the field will be excluded from the SQL INSERT or UPDATE
				   statement, thus leaving intact the existing values of fields
				   not specifically marked for change.

				   After all of the field changes have been made, call Ew_Post()
				   to write the changes to the table. See also Ew_Cancel() to
				   clear pending writes.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to Edit_Initialize.

			  table_name - Input. The name of the table.

			  column_number - Input. The ordinal position of the column. Note that
				 this is a 1-based list; that is, the first column in the table
				 should be identified as column_number 1, not 0.

			  value - Input. The value (data) to be written to the named field
				 in the named table.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_SetFieldByNumber(
		 const int smfID,
		 const char* table_name,
		 const int column_number,
		 const char* value )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->SetFieldByNumber(table_name, column_number, value);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_SetFieldByNumber
//---------------------------------------------------------------------------
/*   Function:      Ew_CheckSyntax
 *
 *   Description:   Compiles the human-readable edit logic into tokenized code,
 *                  and evaluates the tokenized code to determine that it is
 *                  syntactically correct.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_logic - Input. The edit logic, as written by the edit author.

			  how_many_errors - Output. The number of syntax errors found in the
				 compiled (tokenized) logic. If this number is greater than 0 (zero),
				 use it as a loop controller for calls to Ew_GetSyntaxError to
				 retrieve each of the syntax errors. An edit that generates 1+
				 errors will not write the tokenized code to EE_Edits.CompiledLogic,
				 and will therefore not run when Edit_RunEdits or its like is invoked.

			  how_many_warnings - Output. In some situations, the compiler may
				 warn the user that syntax is technically correct, but inadvisable.
				 (Refer to FixEdits #4796 for a scenario described by Alan Houser.)
				 If edit_logic generates 0 errors but 1+ warnings, the tokenized
				 code will nevertheless be written to EE_Edits.CompiledLogic, and
				 the edit will run.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_CheckSyntax(
		 const int smfID,
		 const char* edit_logic,
		 int* how_many_errors,
		 int* how_many_warnings )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CheckSyntax(edit_logic,
			how_many_errors, how_many_warnings);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_CheckSyntax
//---------------------------------------------------------------------------

/*   Function:      Ew_GetSyntaxError
 *
 *   Description:   Following a call to Ew_CheckSyntax, the caller should
 *                  retrieve the list of syntax errors using this function.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  error_idx - Input. Index into the list of syntax errors discovered
				 during the call to Ew_CheckSyntax. This is a zero-based iteration;
				 the caller's loop controller should count from 0 to (how_many_errors -1).

			  line_number - Output. EditWriter uses this to enable a user to
				 double-click on a syntax error, and the line highlighter in the
				 edit logic widget positions on the line that generated the error.
				 This line_number will be 0-based, to correspond to how lines are
				 counted in EditWriter's widget library; note however that the
				 line number displayed in syntax_error will be 1-based, because
				 that's what humans expect.

			  syntax_error - Output. Receives the syntax error, prefaced by the
				 line number (1-based) and a colon (" 3: closing parenthesis expected").
				 We recommend making this buffer large enough to contain a descriptive
				 error message; e.g., 200 bytes should do it.

			  syntax_error_len - Input. The size of the buffer supplied for receiving
				 the syntax_error.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetSyntaxError(
		 const int smfID,
		 const int error_idx,
		 int* line_number,
		 char* syntax_error,
		 const int syntax_error_len)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetSyntaxError(error_idx, line_number,
			syntax_error, syntax_error_len);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetSyntaxError
//---------------------------------------------------------------------------

/*   Function:      Ew_GetSyntaxWarning
 *
 *   Description:   Following a call to Ew_CheckSyntax, and if argument
 *                  how_many_warnings > 0, the caller should
 *                  retrieve the list of syntax warnings using this function.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  warning_idx - Input. Index into the list of syntax warnings discovered
				 during the call to Ew_CheckSyntax. This is a zero-based iteration;
				 the caller's loop controller should count from 0 to (how_many_warnings -1).

			  line_number - Output. EditWriter uses this to enable a user to
				 double-click on a syntax warning, and the line highlighter in the
				 edit logic widget positions on the line that generated the warning.
				 This line_number will be 0-based, to correspond to how lines are
				 counted in EditWriter's widget library; note however that the
				 line number displayed in syntax_warning will be 1-based, because
				 that's what humans expect.

			  syntax_warning - Output. Receives the syntax warning, prefaced by the
				 line number (1-based) and a colon (" 5: ambiguous call to INLIST").
				 We recommend making this buffer large enough to contain a descriptive
				 warning message; e.g., 200 bytes should do it.

			  syntax_warning_len - Input. The size of the buffer supplied for receiving
				 the syntax_warning.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetSyntaxWarning(
		 const int smfID,
		 const int warning_idx,
		 int* line_number,
		 char* syntax_warning,
		 const int syntax_warning_len)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetSyntaxWarning(warning_idx, line_number,
			syntax_warning, syntax_warning_len);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetSyntaxWarning
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLogicListCount
 *
 *   Description:  Ew_CheckSyntax creates lists of Field, Table and Message
 *                 objects referenced by the edit logic. Immediately following
 *                 a call to Ew_CheckSyntax, the calling application should
 *                 invoke Ew_GetLogicListCount for each of the object lists,
 *                 and for those with counts greater than 0, retrieve each
 *                 of the identifying items using Ew_GetLogicListItem.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  which_list - Input. Token to identify which referenced list:
				 'F' = Fields
				 'M' = numbered Messages
				 'E' = text Messages
				 'T' = Tables

			  how_many_items - Output. Receives the number of items in the list.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLogicListCount(
		 const int smfID,
		 const char which_list,
		 int* how_many_items )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLogicListCount(which_list, how_many_items);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetLogicListCount
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLogicFieldNumber
 *
 *   Description:  Ew_CheckSyntax creates lists of Field, Table and Message
 *                 objects referenced by the edit logic. Immediately following
 *                 a call to Ew_CheckSyntax, the calling application should
 *                 invoke Ew_GetLogicListCount for each of the object lists,
 *                 and for those with counts greater than 0, retrieve each
 *                 of the identifying items using Ew_GetLogicFieldNumber,
 *                 Ew_GetLogicMessageNumber, Ew_GetLogicMessageText, and
 *                 Ew_GetLogicTableName.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  idx - Input. Index into the list.

			  field_number - Output. Receives the field_number at the idx'th
				 position in the list.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLogicFieldNumber(
		 const int smfID,
		 const int idx,
		 int* field_number )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLogicFieldNumber(idx, field_number);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetLogicFieldNumber
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLogicTableName
 *
 *   Description:  Ew_CheckSyntax creates lists of Field, Table and Message
 *                 objects referenced by the edit logic. Immediately following
 *                 a call to Ew_CheckSyntax, the calling application should
 *                 invoke Ew_GetLogicListCount for each of the object lists,
 *                 and for those with counts greater than 0, retrieve each
 *                 of the identifying items using Ew_GetLogicFieldNumber,
 *                 Ew_GetLogicMessageNumber, Ew_GetLogicErrorMessage, and
 *                 Ew_GetLogicTableName.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  idx - Input. Index into the list.

			  table_name - Output. Receives the table_name at the idx'th
				 position in the list. We recommend making this buffer
				 31 bytes; we'll be imposing table name lengths of
				 30 bytes, and you'll want to leave room for the NULL terminator.

			  buff_size - Input. The size of the table_name buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLogicTableName(
		 const int smfID,
		 const int idx,
		 char* table_name,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLogicTableName(idx, table_name, buff_size);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetLogicTableName
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLogicErrorMessage
 *
 *   Description:  Ew_CheckSyntax creates lists of Field, Table and Message
 *                 objects referenced by the edit logic. Immediately following
 *                 a call to Ew_CheckSyntax, the calling application should
 *                 invoke Ew_GetLogicListCount for each of the object lists,
 *                 and for those with counts greater than 0, retrieve each
 *                 of the identifying items using Ew_GetLogicFieldNumber,
 *                 Ew_GetLogicMessageNumber, Ew_GetLogicErrorMessage, and
 *                 Ew_GetLogicTableName.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  idx - Input. Index into the list.

			  msg_func_token - Output. Receives a token identifying the
				 EDITS Language function that was used for the message.
				 The tokens are:

				 int  char  EDITS Language function
				 ---  ----  -----------------------
				  69    E   ERROR_TEXT
				  83    S   SAVE_TEXT

			  error_msg - Output. Receives the error message at the idx'th
				 position in the list. We recommend making this buffer
				 500 bytes; that should be overkill for any text message
				 returned from edit logic. (EDITS Language functions
				 ERROR_TEXT and SAVE_TEXT return strings; use
				 Ew_GetLogicMessageNumber to retrieve numbered messages.

			  buff_size - Input. The size of the error_msg buffer.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLogicErrorMessage(
		 const int smfID,
		 const int idx,
		 int* msg_func_token,
		 char* error_msg,
		 const int buff_size )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLogicErrorMessage(idx,
			msg_func_token, error_msg, buff_size);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetLogicErrorMessage
//---------------------------------------------------------------------------
/*   Function:     Ew_GetLogicMessageNumber
 *
 *   Description:  Ew_CheckSyntax creates lists of Field, Table and Message
 *                 objects referenced by the edit logic. Immediately following
 *                 a call to Ew_CheckSyntax, the calling application should
 *                 invoke Ew_GetLogicListCount for each of the object lists,
 *                 and for those with counts greater than 0, retrieve each
 *                 of the identifying items using Ew_GetLogicFieldNumber,
 *                 Ew_GetLogicMessageNumber, Ew_GetLogicErrorMessage, and
 *                 Ew_GetLogicTableName.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  idx - Input. Index into the list.

			  msg_func_token - Output. Receives a token identifying the
				 EDITS Language function that was used for the message.
				 The tokens are:

				 int  char  EDITS Language function
				 ---  ----  -----------------------
				  65    A   SET_WARNING
				  68    D   Default error message (EE_Edits.MessagePKey
							foreign key to EE_Messages.MessageNumber)
				  69    E   ERROR_MSG
				  73    I   SET_ERROR
				  83    S   SAVE_ERROR_TEXT
				  85    U   USR2 - note this is a deprecated EDITS Language function
				  87    W   SAVE_WARNING_TEXT


			  msg_number - Output. Receives the EE_Messages.MessageNmber referenced
				  in the EDITS Language function identified by the above token.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetLogicMessageNumber(
		 const int smfID,
		 const int idx,
		 int* msg_func_token,
		 int* msg_number)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetLogicMessageNumber(idx,
			msg_func_token, msg_number);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetLogicMessageNumber
//---------------------------------------------------------------------------

/*   Function:     Ew_CopyLookupTable
 *
 *   Description:  Copying a lookup table is a really complicated process,
 *                 so we'll let the Engine handle all the details.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  source_table_name - Input. Name of the source table.

			  target_table_name - Input. Name of the new table.

			  agency_pkey - Input. The primary key (PKey) value of the agency
				 that will be affiliated with this new table.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_CopyLookupTable(
		 const int smfID,
		 const char* source_table_name,
		 const char* target_table_name,
		 const int agency_pkey)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		// The penultimate parameter, is_renaming_table, ensures we only copy
		// the table under its new name, and retain the source table.
		t_result = SMFList[smfID]->RenameLookupTable( source_table_name,
			target_table_name, false, agency_pkey );

	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_CopyLookupTable
//---------------------------------------------------------------------------

/*   Function:     Ew_OpenImportMetafile
 *
 *   Description:  Used primarily by the EditWriter Import Metafile module,
 *                 this function temporarily attaches to a source metafile
 *                 so that SELECT statements can be run on tables JOINed
 *                 to identify quickly the differences between source and target,
 *                 and to run INSERT and UPDATE queries pulling from source
 *                 into target.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  source_smf_name - Input. The fully-qualified path and name of the
				 metafile that contains the table we are copying. NOTE: It may
				 be a good idea on Windows systems to ensure there are no
				 embedded blanks in the path; use Win API function GetShortPathName().

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_OpenImportMetafile( const int smfID, const char* source_smf_name)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->OpenImportMetafile( source_smf_name );

	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_OpenImportMetafile
//---------------------------------------------------------------------------

/*   Function:     Ew_CloseImportMetafile
 *
 *   Description:  Used primarily by the EditWriter Import Metafile module,
 *                 this function detaches from the source metafile that was
 *                 attached in a call to Ew_OpenImportMetafile.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_CloseImportMetafile( const int smfID)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CloseImportMetafile();

	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_CloseImportMetafile
//---------------------------------------------------------------------------

/*   Function:     Ew_GetImportCandidates
 *
 *   Description:  Used primarily by the EditWriter Import Metafile module,
 *                 this function executes queries on the joined Source and
 *                 Target metafiles to identify objects that are New, Changed,
 *                 Conflicts on New (e.g., Edit Tag matches another edit in Target),
 *                 and Conflicts on Changed.
 *
 *                 The template for the callback is CallbackImportMetafile,
 *                 where parameter target_pkey will be -1 (meaning "not applicable")
 *                 for the query identifying new candidates, and parameter
 *                 status will be one of:
 *                      1 - New Candidate
 *                      2 - Changed Candidate
 *                      3 - New Candidate with a property that conflicts
 *                          with an existing object
 *                      4 - Updated Candidate with a property that conflicts
 *                          with another existing object
 *
 *                 Therefore, the callback will be run four times for each
 *                 MCO to cover all of the possible status considerations.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  table_name - Input. One of the control tables: EE_Agencies,
				 EE_Edits, EE_EditSets, EE_Fields, EE_Layouts, EE_Messages
				 or EE_LookupTables.

			  match_on_field_name - Input. The column to use to look for a match
				 between the source and target metafiles:
				 EE_Agencies - AdminCode
				 EE_Edits - EditName or EditTag (user preference)
				 EE_EditSets - EditSetName or EditSetTag (user preference)
				 EE_Fields - FieldNumber
				 EE_Layouts - LayoutName or LayoutTag (user preference)
				 EE_Messages - MessageNumber
				 EE_LookupTables - TableName

			  owner - Input. "this" pointer to an object (e.g., a Windows
				 form) that will handle the primary key values being returned to
				 it through the callback function. (In C++ and Java, the pointer
				 is called "this". In Delphi/Object Pascal, it is called "Self".
				 In Visual Basic, it is called "Myself".)

			  callback_func - Input. Pointer to a function in the
				 calling program that will receive the data lobbed at it from
				 this function call. Use template for CallbackImportMetafile.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_GetImportCandidates(
	const int smfID,
	const char* table_name,
	const char* match_on_field_name,
	void* owner,
	void* callback_func)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->GetImportCandidates(table_name,
			match_on_field_name, owner, callback_func);

	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_GetImportCandidates
//---------------------------------------------------------------------------

/*   Function:     Ew_ImportLookupTable
 *
 *   Description:  Used primarily by the EditWriter Import Metafile module,
 *                 this function temporarily attaches to a source metafile
 *                 from which it extracts:
 *                 - the CREATE TABLE structure
 *                 - all CREATE INDEX definitions
 *                 - all CREATE TRIGGER definitions
 *                 and uses these to create the table in the currently open
 *                 metafile. Then, in single SELECT statement, it imports all
 *                 of the data from the source metafile. This process is
 *                 astoundingly faster than any other means of moving this
 *                 data around.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  table_name - Input. The name of the table in the source database.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_ImportLookupTable( const int smfID, const char* table_name)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ImportLookupTable( table_name );

	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_ImportLookupTable
//---------------------------------------------------------------------------

/*   Function:     Ew_ImportWritePassword
 *
 *   Description:  Used exclusively by the EditWriter Import Metafile module,
 *                 this function is invoked when a new Agency is imported.
 *                 The Engine does not provide a mechanism for writing to the
 *                 EE_Agencies.Password directly; instead, it accepts the
 *                 plain-text password and runs it through a one-way hash to
 *                 generate the munged password. But for the Import Metafile
 *                 process, we need to be able to bring over a new Agency record
 *                 without knowing the plain-text password. Thus, this workaround.
 *                 I do not plan to put these Import functions into the EDITS API
 *                 documentation; an NPCR programmer will have to discover these
 *                 by reading the source code.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  agency_pkey - Input. The primary key for the newly-added Agency
				 in the Target metafile. Everything *except* the password has
				 been written to that record; now we'll transfer the password
				 from the Source metafile.

			  the_password - Input. The hashed/munged password taken from the
				 source metafile's EE_Agencies record.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_ImportWritePassword(
		 const int smfID,
		 const int agency_pkey,
		 const char* the_password)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ImportWritePassword( agency_pkey, the_password );

	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_ImportWritePassword
//---------------------------------------------------------------------------
/*   Function:     Ew_ClearEditLogicPivots
 *
 *   Description:  This function accompanies Edit_RecompileEditLogic. After the
 *                 EE_Edits.EditLogic has been compiled, and regardless of
 *                 whether it results in valid EE_Edits.CompiledLogic, you
 *                 will want to get rid of any pivot table links between
 *                 the Edit object and the Fields, Messages and Lookup Tables
 *                 referenced from the logic. (These are in tables
 *                 EE_EditFieldsPivot, EE_EditTablesPivot,
 *                 EE_EditMessagesPivot (numbered messages in EE_Messages) and
 *                 EE_EditErrorsPivot (not really a pivot; these are the ad hoc
 *                 text messages strewn throughout edit logic).
 *
 *                 Following this function, the caller (usually EditWriter, but
 *                 you're welcome to try this yourself) will re-add the Field,
 *                 Message and Table references from the information retrieved
 *                 through calls to Ew_GetLogic* functions. Then use Ew_Insert,
 *                 Ew_SetFieldByName and Ew_Post to add the pivot data back.
 *
 *                 I really need to figure out where I'm going to explain how
 *                 all of this works, with screen shots, circles and arrows.
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  edit_pkey - Input. The primary key for the edit.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_ClearEditLogicPivots( const int smfID, const int edit_pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->ClearEditLogicPivots( edit_pkey );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_ClearEditLogicPivots
//---------------------------------------------------------------------------

/*   Function:     Ew_DefineTableStructure
 *
 *   Description:  Support for creating new lookup tables, or restructuring
 *                 existing ones.
 *
 *                 This function initializes an internal structure, making it
 *                 ready to receive properties describing column definitions
 *                 and index instructions. Once all of those are in place,
 *                 call the "make it so" function.
 *
 *                 NOTE 1: If parameter restructure_existing is true, this function
 *                 verifies that a table exists in the metafile by that name. If
 *                 restructure_existing is false, this function verifies that a
 *                 table DOES NOT exist in the metafile by that name. Errors will
 *                 be raised for violations.
 *
 *                 NOTE 2: This function ensures that the table name does not
 *                 begin with "EE_" (regardless of case), does not exceed
 *                 thirty bytes, begins with an alpha character (a-Z, A-Z) and
 *                 thereafter consists of alpha characters, digits 0-9 or an
 *                 underscore. We are arbitrarily declaring that any other
 *                 combination is illegal, and an error will be raised to tell
 *                 you so!
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  table_name - Input. Name of the table.

			  restructure_existing - input. Treated as a boolean. 0 (zero) means
				 false, this is a definition of a new table. Non-zero means true,
				 the table already exists and its properties are being modified.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_DefineTableStructure(
		 const int smfID,
		 const char* table_name,
		 const int restructure_existing )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->DefineTableStructure( table_name, restructure_existing );
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_DefineTableStructure
//---------------------------------------------------------------------------
/*   Function:     Ew_AddColumnDef
 *
 *   Description:  Following the initialization call to Ew_DefineTableStructure,
 *                 and prior to the "make it so" call to Ew_CreateTable, call
 *                 this function to create the column definitions for the table.
 *
 *                 NOTE 1: Do not send col_name PKey. Ew_CreateTable will add that,
 *                 and this function will return an error if you try to add it
 *                 yourself.
 *
 *                 NOTE 2: This function ensures that the col_name does not exceed
 *                 thirty bytes, begins with an alpha character (a-Z, A-Z) and
 *                 thereafter consists of alpha characters, digits 0-9 or an
 *                 underscore. We are arbitrarily declaring that any other
 *                 combination is illegal, and an error will be raised to tell
 *                 you so!
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  col_name - Input. Required. The name of the column.

			  copy_from_col_name - Input. Optional. Used when restructuring a table. It is
				 possible that the change to the structure involves renaming the
				 column, which can make it tricky to write SQL to port the data.
				 If you want the data to move from the old structure to the new,
				 state the old name of the column here. If this parameter is
				 NULL or an empty string, even if the table is being restructured,
				 the data will not be ported.

			  data_type - Input. Required.
				 0 - Character
				 1 - Integer

				 Any other value will generate an error.

			  data_length - Input. Optional (sort of). Applies only to character-type fields. This is
				 the intended size of the data (although SQLite will allow it to
				 be of any length). EditWriter relies upon this value to set
				 limits on the edit control in the table data entry form.

			  is_nullable - Input. Required. If 0 (zero), the column will be defined NOT NULL.
				 If non-zero (true), the column will be defined to allow NULL values.

			  is_trigger_field - Input. Required. Do not attempt to modify a trigger-maintained
				 field. Ideally, these will eventually be dropped from lookup tables
				 altogether. If an existing column is a trigger field, please say
				 so here, so that the triggers can be ported to the new table.

			  CHECK CONSTRAINTS:
			  check_limit_to_list - Input. Optional. A comma-separated list of valid optons
				 which can be used in a check constraint such as this:

				 CHECK(Behavior IN ('0','1','2','3'))

				 This parameter should consist of only this portion of the above
				 example: '0','1','2','3'
				 That is, a string of options, each delimited with an apostrophe
				 (character 39 on the ASCII chart), separated by commas.

			  check_range_low and check_range_high - Input. Optional.
				 If either is not NULL and not an empty string, then both
				 must have a non-NULL, non-empty value. These values are used
				 to create a column check constraint such as this:

				 CHECK(SITECODE BETWEEN 'C000' AND 'C999')

			  default_value - Input. Optional. If a row is added to the table without
				 designating a value, this default value will be automatically
				 written to the column.

			  index_status - Input. Required. Choices are:
				 0 - Not indexed
				 1 - Indexed, duplicates allowed
				 2 - Indexed, duplicates NOT allowed

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_AddColumnDef(
		 const int smfID,
		 const char* col_name,
		 const char* copy_from_col_name,
		 const int data_type,
		 const int data_length,
		 const int is_nullable,
		 const int is_trigger_field,
		 const char* check_limit_to_list,
		 const char* check_range_low,
		 const char* check_range_high,
		 const char* default_value,
		 const int index_status)
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->AddColumnDef(col_name,
												copy_from_col_name,
												data_type,
												data_length,
												is_nullable,
												is_trigger_field,
												check_limit_to_list,
												check_range_low,
												check_range_high,
												default_value,
												index_status);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_AddColumnDef
//---------------------------------------------------------------------------
/*   Function:   Ew_CreateTable
 *
 *   Description:  After a lookup table definition has been created, this is the
 *                 "make it so" function. Here is what happens:
 *
 *                 - If restructuring an existing table (i.e., parameter
 *                   restructure_existing to function Ew_DefineTableStructure
 *                   is true):
 *                   - retrieve the trigger definitions from the existing table
 *                     and store them in the internal structure, and
 *                   - rename the source table to a temporary/backup name.
 *
 *                 - Create a new table according to the instructions in the
 *                   one-to-many calls to Ew_AddColumnDef.
 *                 - If restructuring, port the data from the source table into
 *                   the appropriate columns of the new table.
 *                 - Drop the old (source) table.
 *                 - Build indexes on the new table.
 *                 - If trigger-maintained fields exist on a restructured table,
 *                   reapply the triggers to the newly created table.
 *                 - If NOT restructuring, this is a new table that will be added
 *                   to EE_LookupTables.
 *
 *                 See also: Ew_DefineTableStructure, Ew_AddColumnDef
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

			  lookup_pkey - Output. The primary key for the row that was updated
				 or added in EE_LookupTables.
 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */
int EDIT_API Ew_CreateTable(
		 const int smfID,
		 int* lookup_pkey )
{
	int t_result = 0;

	try
	{
		InitCheck( smfID );

		t_result = SMFList[smfID]->CreateTable(lookup_pkey);
	}
	catch (...)
	{
		// Ask the TEngineErrorsPump for the error message(s)
		t_result = -1;
	}

	return t_result;
} // Ew_CreateTable
//---------------------------------------------------------------------------

#endif  // __SQLEW__

// Header template
/*   Function:
 *
 *   Description:
 *
 *   Parameters:
			  smfID - Input. Index into Edit Engine's internal list of open
				 metafiles. This parameter is initialized by a call to
				 Edit_Initialize.

 *
 *   Return value: Zero (0) normally, error code (-1) if a system error has occurred.
 */

//---------------------------------------------------------------------------

