//---------------------------------------------------------------------------

#ifndef EditCompilerH
#define EditCompilerH
//---------------------------------------------------------------------------
#ifdef __SQLEW__
#include <string>
#include <map>
#include <vector>

#include "CompilerTokens.h" // used by TTokenMaster and TEditCompiler

// Forward declarations
class TSQLDB;
class TSQLQuery;
class TEngineErrorsPump;
class TTablesList;
class TTableDef;

typedef enum
{
	 crNormal,
	 crEmptyAPICode,
	 crUnclosedComments,
	 crError,
	 crWarning
} TCompilerResults;

typedef struct {
		const char *funcname;
		char token;
		int returnType;   // EEi_CHAR, EEi_INT or NOTHING (0, procedure calls)
} EEt_IntFuncListType;

typedef struct
{
	  char orig[15];
	  TTokenID tok_name;
	  TTokenType tok_type;
	  char compiled;
} EEt_TokType;

/*   For the user-defined functions look-up table, FUDFDefList.
 *   Note that this lives and grows regardless of how many edits
 *   are being compiled-- one at a time, as from the EW3 Syntax
 *   Checker button, or all of them, as when generating an RMF.
 *   The difference is that, unlike EW2, we'll reset the
 *   is_used_in_edit flag to false with the compile of each edit,
 *   until such time as we determine the UDF is being used in
 *   THIS edit. This is an experiment; we may have to resort to
 *   the EW2 way of doing things if the engine misbehaves. KB 06/16/05
 */
typedef struct {
		char         func_name[33];
		unsigned int func_addr;
        bool         is_used_in_edit;
} EEt_UDFDef;

// For walk_code_block(), to keep track of nested CONSTRUCTs.
typedef struct {
    char token;
    //int  open_braces;
    int  construct_line_no;
    int  open_brace_line_no;
    int  close_brace_line_no;
} EEt_NestedConstruct;

/***************************************
 ***  TVariableDef
 ***************************************/
class TVariableDef
{
private:
	std::string FVarName;
	char   FVarType;  // EEi_INT or EEi_CHAR
	size_t FVarSize;
	std::string FVarValue;
	bool   FIsReported;
public:
	TVariableDef();
	~TVariableDef();

	std::string GetVarName();
	void SetVarName(std::string var_name);

	char GetVarType();
	void SetVarType(char var_type);

	size_t GetVarSize();
	void SetVarSize(size_t var_size);

	std::string GetVarValue();
	void SetVarValue(std::string var_value);

	bool GetIsReported();
	void SetIsReported(bool is_reported);
};

class TVariablesMap
{
private:
	std::map<std::string, TVariableDef*> FVariables;
public:
	TVariablesMap();
	~TVariablesMap();

	void ClearVariables();

	TVariableDef* GetVariable(std::string var_name);
	bool AddVariable(TVariableDef* var, std::string& err_msg);
};

class TLogicMessageObject
{
public:
	int FuncToken;
	int MessageNumber;
	std::string MessageText;

	// default constructor
	TLogicMessageObject() {};
};

/***************************************
 ***  TEditCompiler
 ***************************************/
class TEditCompiler
{
private:
	TSQLDB* FSQLDB;
	TSQLQuery* FQry;
	TEngineErrorsPump* FEngineErrorsPump;
	TTablesList* FTablesList;

	std::string FAPICode;
	std::string FCompiledCode;
	std::string FCurrentVarName;  // for parse_assign()
	std::string FCurrentString; // also for parse_assign()
	int FCurrentNumber;

	// The lists of Fields, Tables and Messages are cleared in
	// Compile() and populated in tokenize(). The calling program
	// (EditWriter) will request these lists through APIs
	// Ew_LogicListCount and Ew_LogicListItem.
	// FLogicFieldsList contains the FieldNumber for each field
	// FLogicMessagesList contains the MessageNumber for each message
	// FLogicErrorsList contains the Error text for each non-numbered message
	// FLogicTablesList contains the ?? for each table
	std::vector<int> FLogicFieldsList;
	std::vector<std::string> FLogicTablesList;
	std::vector<TLogicMessageObject*> FLogicMessageNumbersList;
	std::vector<TLogicMessageObject*> FLogicErrorMessagesList;

	size_t FAPIptr;
	size_t FCompiledPtr;
	size_t FCompLineNo;
	unsigned int FNumTokens;
	char FCurrentToken;

	TTokenType FCurrentTokenType;
	std::string field_name_to_field_num(std::string fname);

	std::vector<std::string> FErrorMessages;
	std::vector<std::string> FWarningMessages;
	TVariablesMap* FVariablesMap;
	std::vector<EEt_UDFDef*> FUDFDefList;

	void clear_logic_messages_lists();
	// Called from tokenize() to populate FLogicFieldsList
	void add_to_logic_fields(int item);
	// Called from Compile() to parse out items for FLogicTablesList
	void update_logic_tables();
	// Called from Compile() to parse out items for
	// FLogicMessageNumbersList and FLogicErrorMessagesList
	void update_logic_messages();

	void comp_sntx_err(const char *errmsg);
	void comp_sntx_warn(const char *warnmsg);
	int strip_comments();
	void findUDFs();
	int tokenize(int& inlist_param);
	int get_next_token();
	bool balance_parens();
	void parse_block(int required);
	std::string extract_var_name(unsigned int& var_length);
	std::string get_intern_func_name(char func_code);
	char get_intern_func_code(std::string func_name);
	EEt_IntFuncListType* get_intern_func_def(char func_code);

	void  check_params(unsigned char func_code, std::string the_code);

	bool parse_params(std::string the_code, std::vector<std::string>& params);
	bool verify_param_count(std::string func_name, int param_count, int low_expected, int hi_expected);
	bool check_param_type(std::string func_name, std::string full_param, int param_num, int expected_type);
	bool message_number_is_valid(std::string func_name, std::string full_param);
	// These two are for LOOKUP, ILOOKUP and RLOOKUP
	void find_table(std::string func_name, std::string table_name, std::string extension);
	bool find_index_name(TTableDef* table, std::string compare_idx_name);
	// These two for SQLLookup and SQLRangeLookup
	TTableDef* find_sql_table(std::string table_name);
	void find_sql_fields(TTableDef* table, std::string fields_list);

	bool check_date_type(std::string func_name, std::string param);
	void no_such_variable(int param_num, std::string func_name, std::string var_name);
	TVariableDef* find_variable(std::string var_name);
	void walk_code_block(char token, int start_block, int& IF_end_block);
	void walk_expr(std::string the_code);

	void parse_if(int& IF_end_block);
	void parse_else(int& IF_end_block);
	void parse_while();
	void parse_for();
	void parse_for_prep();
	void parse_assign();
	void parse_vardecl(char var_type);
	void parse_fndecl();
	void parse_fncall();
	void parse_return();

public:
	TEditCompiler(TSQLDB* database, TEngineErrorsPump* pump, TTablesList* tables_list);
	~TEditCompiler();

	int Compile(std::string edit_logic, int* how_many_errors, int* how_many_warnings);

	// Processing error messages
	char * GetErrorMessage(int index);
	int GetErrorCount();

	char * GetWarningMessage(int index);
	int  GetWarningCount();

	std::string GetCompiledCode();

	int GetLogicListCount(char which_list);
	int GetLogicFieldNumber(int idx);
	std::string GetLogicTableName(int idx);
	std::string GetLogicErrorMessage(int idx, int* msg_func_token);
	int GetLogicMessageNumber(const int idx, int* msg_func_token);
};

#endif // __SQLEW
#endif
