//---------------------------------------------------------------------------
#include <memory>  // std::unique_ptr
#include <stdexcept>
#include <cstdio>  // snprint
#include <cstdlib> // std::atoi

#include "EditErrors.h"
#include "SQLDB.h"
#include "FastSQLQuery.h"
#include "EngineErrorsPump.h"
#include "EditSet.h"
#include "Layout.h"
#include "Utility.h"
#include "LibFuncs.h"
//---------------------------------------------------------------------------
/*
 *    class TErrorObject
 */
TErrorObject::TErrorObject(std::string error_type, std::string error_message,
	std::string func_name)
{
	// E=error, W=warning, M=message
	FErrorType.assign(error_type);
	// The message after processing by TEditErrors::expand_msg_macros
	FErrorMessage.assign(error_message);
	// The EDITS Language function that created the message, or "Default"
	FFunctionName.assign(func_name);
} // constructor
//---------------------------------------------------------------------------
/*
 *    class TEditErrors
 */
TEditErrors::TEditErrors(TSQLDB* database, TEngineErrorsPump* pump)
{
	FSQLDB = database;
	FMessageQry = NULL;        // ditto

	FEngineErrorsPump = pump;

	FCallbackOwner = NULL;
	FCallbackFunction = NULL;
	FLayout = NULL;
	FLibFuncs = NULL; // see SetLibFuncs

	FAltMessageNumber = 0;

	// Initialize these to empty values. They'll be read in ReportDefaultError
	FError_Text_Edit_PKey = 0;
	FError_Text_Message.clear();
} // constructor
//---------------------------------------------------------------------------
TEditErrors::~TEditErrors()
{
	clear_errors_map();

	if (FMessageQry)
	{
        FMessageQry->ReleaseParameterizedQuery();
		delete FMessageQry;
	}
} // destructor
//---------------------------------------------------------------------------
void TEditErrors::SetLibFuncs(TLibFuncs* lib_funcs)
{
	// This is so we can grab the current value of DateErrorCode for expanding
	// message macros.
	FLibFuncs = lib_funcs;
}
//---------------------------------------------------------------------------
void TEditErrors::clear_errors_map()
{
	// This will be invoked from Initialize for each new case processed, whether
	// from TRunEdits::RunEdits or TRunEdits::DebugEdit.
	// Also called from this class' destructor.
	std::map< int, std::vector<TErrorObject*> >::iterator it = FErrorsMap.begin();
	size_t t_idx;
	while (it != FErrorsMap.end())
	{
		std::vector<TErrorObject*> t_err_objs = (*it).second;
		for (t_idx = 0; t_idx < t_err_objs.size(); t_idx++)
		{
			TErrorObject* t_obj = t_err_objs[t_idx];
			delete t_obj;
			t_obj = NULL;
		}

		t_err_objs.clear();

		it++;
	}

	FErrorsMap.clear();

} // clear_errors_map
//---------------------------------------------------------------------------

std::string TEditErrors::expand_msg_macros(std::string raw_message,
	TEdit* running_edit)
{
	std::string t_expanded_msg,
				t_parse,
				t_flag,        // 'F', 'V' or 'T'
				t_str_macro_order, // '1', '13', etc.
				t_delimiters = " []{}()!@#$&*-_+=\\|;:'\"<>,.?/"; // all but '%'

	t_parse.assign(raw_message);
	size_t t_pos = t_parse.find_first_of("%");
	size_t t_pos2 = 0;

	if (t_pos == std::string::npos)
	{
		// Nothing to expand... just return what we were given
		return t_parse;
	}

	t_expanded_msg.assign(t_parse.substr(0, t_pos));
	// Advance t_parse to the first position following the "%"
	// 'F'=Field, 'V'=Value, 'T'=Title... which we may not support, so we'll
	// turn it back into 'F' until somebody yipes at us
	t_pos++;
	t_parse.assign(t_parse.substr(t_pos, t_parse.length()));

	while (t_pos > 0)
	{
		t_flag = t_parse.substr(0,1);
		uppercase(t_flag);

		if (t_flag == "D")
		{
			switch (FLibFuncs->GetDateErrorCode())
			{
				case DT_VALID:
					t_expanded_msg += "no error / other";
					break;
				case DT_MISSING:
					t_expanded_msg += "missing";
					break;
				case DT_ERROR:
					t_expanded_msg += "invalid";
					break;
				case DT_UNKNOWN:
					t_expanded_msg += "unknown";
					break;
				case DT_ERROR_DA: // 4L
					t_expanded_msg += "invalid as to day";
					break;
				case DT_ERROR_MO: // 8L
					t_expanded_msg += "invalid as to month";
					break;
				case DT_ERROR_YR: // 16L
					t_expanded_msg += "invalid as to year";
					break;
				case DT_MISSING_DA: // 32L
					t_expanded_msg += "missing the day";
					break;
				case DT_MISSING_MO: // 64L
					t_expanded_msg += "missing the month";
					break;
				case DT_MISSING_YR: // 128L
					t_expanded_msg += "missing the year";
					break;
				case DT_UNKNOWN_DA: // 256L
					t_expanded_msg += "unknown as to day";
					break;
				case DT_UNKNOWN_MO: // 512L
					t_expanded_msg += "unknown as to month";
					break;
				case DT_UNKNOWN_YR: // 1024L
					t_expanded_msg += "unknown as to year";
					break;
				case DT_EMPTY_STRING: // 4096L
					t_expanded_msg += "date is an empty string";
					break;
				default:
					break;

			} // switch (FLibFuncs->GetDateErrorCode())

			// t_parse is sitting just past the "%" of (e.g.) "%D1" or "%DC".
			// Let's move off of that.
			t_parse.assign(t_parse.substr(2, t_parse.length()));

		}
		else
		{
			for (t_pos2 = 1; t_pos2 < t_parse.length(); t_pos2++)
			{
				if (t_delimiters.find(t_parse.substr(t_pos2,1)) != std::string::npos)
				{
					break;
				}
			}

			if (t_pos2 -1 > 0)
				t_str_macro_order.assign(t_parse.substr( 1, t_pos2));
			else
				t_str_macro_order.assign(t_parse.substr( 1, 1 ));

			int t_macro_order = std::atoi(t_str_macro_order.c_str());
			// The macro order is 1-based; the vector in TEdit is 0-based. Adjust.
			int t_fld_num = running_edit->GetFieldByMacroOrder(t_macro_order-1);
			if (t_fld_num == -1)
			{
				// This is an error condition; t_macro_order out of range
				std::string t_err_msg = "Field macro order (";
				t_err_msg += t_str_macro_order;
				t_err_msg += ") out of range for edit tag \"";
				t_err_msg += running_edit->GetEditTag();
				t_err_msg += "\"";
				FEngineErrorsPump->AddError("TEditErrors", "expand_msg_macros",
					t_err_msg);

				throw std::runtime_error("");
			}
			else
			{
				TLayoutField* t_layout_field = FLayout->GetFieldByFieldNumber(t_fld_num);

				std::string t_field_name,
							t_display_name,
							t_field_value,
							t_data_type;

				if (t_layout_field)
				{
					// Despite what may have been specified in the edit logic,
					// we are supporting only %F<n> and %V<n>... if the logic
					// specifies %S, %L or %T I'm giving them the field name
					if ((t_flag.compare("F") == 0) ||
						(t_flag.compare("S") == 0) ||
						(t_flag.compare("L") == 0) ||
						(t_flag.compare("T") == 0))
					{
						// Get field name
						t_field_name.assign(t_layout_field->GetFieldName());
						t_display_name.assign(t_layout_field->GetDisplayName());

						if ((t_field_name.compare(t_display_name) == 0) ||
							(t_display_name.empty()))
						{
							// If the display name is empty, or if it exactly
							// matches the field name, just use the field name.
							t_expanded_msg += t_field_name;
						}
						else
						{
							// But if the display name is different from field name,
							// show both... display name first.
							t_expanded_msg += t_display_name;
							t_expanded_msg += " [";
							t_expanded_msg += t_field_name;
							t_expanded_msg += "]";
						}
					}
					else if (t_flag.compare("V") == 0)
					{
						// Get field value
						t_field_value.assign(t_layout_field->GetValue());
						t_data_type.assign(t_layout_field->GetDataType());
						if (t_data_type.compare("D1") == 0)
						{
							// Display the date as Y:0000 M:00 D:00
							std::string t_d1_date = (std::string)"Y:" + t_field_value.substr(0,4);
							t_d1_date += (std::string)" M:" + t_field_value.substr(4,2);
							t_d1_date += (std::string)" D:" + t_field_value.substr(6,2);
							t_field_value.assign(t_d1_date);
						}
						else
						{
							// let's report <BLANK> in the message where applicable
							if (trim(t_field_value).empty())
							{
                            	t_field_value.assign("<BLANK>");
							}
						}
						t_expanded_msg += t_field_value;
					}
				} // if (t_layout_field)

			} // if (t_fld_num == -1)
		} // if (t_flag == "D")

		// Onward! See if we have more macros to expand
		t_pos = t_parse.find_first_of("%");
		if (t_pos != std::string::npos)
		{
			// Pick up whatever preceded this new instance of '%'
			t_expanded_msg += t_parse.substr(t_pos2, t_pos-t_pos2);
			// Advance past the '%'
			t_pos++;
			// and parse the string again
			t_parse.assign(t_parse.substr(t_pos, t_parse.length()));
		}
		else
		{
			// No more macros... just pick up the rest of the string
			if (t_pos2 < t_parse.length())
			{
				t_expanded_msg += t_parse.substr(t_pos2, t_parse.length());
			}
			break;
		}
	} // while (t_pos > 0)

	return t_expanded_msg;
} // expand_msg_macros
//---------------------------------------------------------------------------
void TEditErrors::add_error_to_map(int edit_pkey, TErrorObject* error_object)
{
	// Called by ReportDefaultError, this puts the error message object
	// into the map that is collecting error messages for a batch of edits
	// (or for a single edit, as when Edit_DebugEdit is invoked).
	std::map<int, std::vector<TErrorObject*> >::iterator it =
		FErrorsMap.find(edit_pkey);
	if (it != FErrorsMap.end())
	{
		it->second.push_back(error_object);
	}
	else
	{
		std::vector<TErrorObject*> t_errors;
		t_errors.reserve(20);
		t_errors.push_back(error_object);
		FErrorsMap[edit_pkey] = t_errors;
	}
} // add_error_to_map
//---------------------------------------------------------------------------
void TEditErrors::Initialize(void* owner, void* callback_func,
	TLayout* layout, int suppress_warnings)
{
	// We must have BOTH an owner and a function; otherwise, we'll
	// set them both to NULL
	if (owner && callback_func)
	{
		FCallbackOwner = owner;
		FCallbackFunction = (CallbackRunEdits)callback_func;
	}
	else
	{
		FCallbackOwner = NULL;
		FCallbackFunction = NULL;
	}

	clear_errors_map();

	// Assign the TLayout
	FLayout = layout;

	// Clear out previous error text message;
	FError_Text_Edit_PKey = 0;
	FError_Text_Message.clear();

	// Store the edit set option "suppress warnings"
	FSuppressWarnings = suppress_warnings;

} // Initialize
//---------------------------------------------------------------------------
int TEditErrors::FailedEditsCount()
{
	// This returns the number of edits in the batch that failed. It is
	// not the same thing as the number of error messages generated, because
	// an edit can generate more than one error.
	return (int)FErrorsMap.size();
} // FailedEditsCount
//---------------------------------------------------------------------------
bool TEditErrors::create_message_query()
{
	if (!FMessageQry)
	{
		std::string t_sql;
		FMessageQry = new TFastSQLQuery(FSQLDB);

		t_sql.assign("SELECT [Message] ");
		t_sql += "FROM [EE_Messages] ";
		t_sql += "WHERE [MessageNumber] = ?;";

		FMessageQry->SetSQLText(t_sql);

		if (!FMessageQry->PrepareParameterizedQuery())
		{
			FEngineErrorsPump->AddError("TEditErrors", "create_message_query",
				"Failure to prepare FMessageQry");
			std::string t_err_msg = FMessageQry->LastErrorMessage();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TEditErrors", "create_message_query", t_err_msg);
			}
			throw std::runtime_error("");
		}

	}
	return true;
} // create_message_query
//---------------------------------------------------------------------------
void TEditErrors::ReportDefaultError(TEdit* running_edit, std::string error_type)
{
	// We'll look up and process the default error message only if it is needed.
	// Parameter error_type will be "E" for edits that RETURN FAIL and
	// "W" for those that RETURN WARN.
	std::string t_message,
	            t_error_message,
				t_error_func,
				t_error_type,
				t_edit_tag,
				t_edit_name,
				t_admin_code;

	int t_edit_pkey = running_edit->GetPKey();

	t_edit_tag.assign(running_edit->GetEditTag());
	t_edit_name.assign(running_edit->GetEditName());
	t_admin_code.assign(running_edit->GetAdminCode());

	if (t_edit_pkey == FError_Text_Edit_PKey)
	{
		t_error_message.assign(FError_Text_Message);
		t_error_func.assign("ERROR_TEXT");
		// Regardless of what was passed in the parameter list, ERROR_TEXT
		// ALWAYS returns a FAIL.
		t_error_type.assign("E");
	}
	else
	{
		// We have moved on to a different edit. Clear these now.
		FError_Text_Edit_PKey = 0;
		FError_Text_Message.clear();

		t_error_type.assign(error_type);


		try
		{
			// We need to retrieve the message to be used
			if (FAltMessageNumber == 0)
			{
				// Using the edit's default message number
				t_message.assign(running_edit->GetDefaultErrorMessage());
				t_error_message = expand_msg_macros(t_message, running_edit);
			}
			else
			{

				if (create_message_query())
				{
					// Bind the FAltMessageNumber to the query
					FMessageQry->BindIntParameter(1, FAltMessageNumber);

					t_error_func.assign("ERROR_MSG");

					if (FMessageQry->OpenParameterizedQuery())
					{
						if (!FMessageQry->Eof())
						{
							std::string t_message = FMessageQry->GetFieldByNumber(0);

							t_error_message = expand_msg_macros(t_message, running_edit);
						}

						// Reset for re-use
						FMessageQry->ResetParameterizedQuery();
					}
				} // if (create_message_query)
			} // if (FAltMessageNumber == 0)
		}
		catch (std::runtime_error& e)
		{
			std::string t_err_msg = e.what();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TEditErrors", "ReportDefaultError", t_err_msg);
			}
			if (FMessageQry)
			{
				t_err_msg.assign(FMessageQry->LastErrorMessage());
				if (!t_err_msg.empty())
				{
					FEngineErrorsPump->AddError("TEditErrors", "ReportDefaultError", t_err_msg);
				}
			}
			throw;
		}

	} // if (t_edit_pkey == FError_Text_Edit_PKey)


	// Yow: Don't report the default error message if it is "0",
	// the message that complains "This edit has no default error message"... that
	// is, unless there is literally NOTHING else to report.
	std::string t_test = running_edit->GetDefaultMessageNumber();

	if (t_test.compare("0") == 0)
	{
		// If absolutely nothing else has been reported for this edit,
		// go ahead and yelp about there being no default error message.
		// This is a situation that should not happen, so if it draws
		// the user's attention, so much the better.
		if ((running_edit->GetErrorMessagesCount() == 0) &&
			(running_edit->GetWarningMessagesCount() == 0))
		{
			(*FCallbackFunction)(FCallbackOwner, t_edit_tag.c_str(),
				t_edit_name.c_str(), t_admin_code.c_str(),
				t_error_type.c_str(), t_error_message.c_str());
		}
	}
	else
	{
		// For edits that have a non-zero default error message, go ahead
		// and report it to the caller.
		if (FCallbackOwner && FCallbackFunction)
		{
			(*FCallbackFunction)(FCallbackOwner, t_edit_tag.c_str(),
				t_edit_name.c_str(), t_admin_code.c_str(),
				t_error_type.c_str(), t_error_message.c_str());
		}
	}
	// Even if we have already reported this to the callback function,
	// we'll put it into the list of messages
	TErrorObject* t_err_obj = new TErrorObject(t_error_type, t_error_message, t_error_func);
	add_error_to_map( t_edit_pkey, t_err_obj );

	// Reporting summary information: The count of error MESSAGES may exceed
	// the count of ERRORS (i.e., the count of cases that failed the edit)
	// We increment TEditObject::IncrErrorsCount() from TRunEdits::interpret_block
	// at the RETURN handler, but we need to count the actual messages from
	// this class.

	if (t_test.compare("0") == 0)
	{
		if ((t_error_type[0] == 'E') &&
			(running_edit->GetErrorMessagesCount() == 0))
		{
			running_edit->IncrMessagesCount();
		}
		else if ((t_error_type[0] == 'W') &&
		         (running_edit->GetWarningMessagesCount() == 0))
		{
			running_edit->IncrWarningMessagesCount();
		}
	}
	else
	{
		if (t_error_type[0] == 'E')
		{
			running_edit->IncrMessagesCount();
		}
		else if (t_error_type[0] == 'W')
		{
			running_edit->IncrWarningMessagesCount();
		}
	}
} // ReportDefaultError
//---------------------------------------------------------------------------
void TEditErrors::ReportSaveErrorText(TEdit* running_edit, int msg_num)
{
	// Called from TLibFuncs::call_save_error_text
	std::string t_error_message,
		t_error_func,
		t_error_type,
		t_edit_tag,
		t_edit_name,
		t_admin_code;

	int t_edit_pkey = running_edit->GetPKey();

	t_edit_tag.assign(running_edit->GetEditTag());
	t_edit_name.assign(running_edit->GetEditName());
	t_admin_code.assign(running_edit->GetAdminCode());

	t_error_type.assign("E"); // Error

	t_error_func.assign("SAVE_ERROR_TEXT");

	try
	{
		if (create_message_query())
		{
			// Bind the msg_num to the query
			FMessageQry->BindIntParameter(1, msg_num);

			if (FMessageQry->OpenParameterizedQuery())
			{
				if (!FMessageQry->Eof())
				{
					std::string t_message = FMessageQry->GetFieldByNumber(0);

					t_error_message = expand_msg_macros(t_message, running_edit);
				}
				else
				{
					t_error_message.assign("Engine error: Message number ");
					t_error_message += intToString(msg_num);
					t_error_message += " not found";
				}

				// Reset for re-use
				FMessageQry->ResetParameterizedQuery();
			}
		} // if (create_message_query)

	}
	catch (std::runtime_error& e)
	{
		std::string t_err_msg = e.what();
		if (!t_err_msg.empty())
		{
			FEngineErrorsPump->AddError("TEditErrors", "ReportSaveErrorText", t_err_msg);
		}
		t_err_msg.assign(FMessageQry->LastErrorMessage());
		if (!t_err_msg.empty())
		{
			FEngineErrorsPump->AddError("TEditErrors", "ReportSaveErrorText", t_err_msg);
		}
		throw std::runtime_error("");
	}

	if (FCallbackOwner && FCallbackFunction)
	{
	  (*FCallbackFunction)(FCallbackOwner,
		t_edit_tag.c_str(), t_edit_name.c_str(), t_admin_code.c_str(),
		t_error_type.c_str(), t_error_message.c_str());
	}
	// Even if we have already reported this to the callback function,
	// we'll put it into the list of messages
	TErrorObject* t_err_obj = new TErrorObject(t_error_type, t_error_message, t_error_func);
	add_error_to_map( t_edit_pkey, t_err_obj );

	// Reporting summary information: The count of error MESSAGES may exceed
	// the count of ERRORS (i.e., the count of cases that failed the edit)
	// We increment TEditObject::IncrErrorsCount() from TRunEdits::interpret_block
	// at the RETURN handler, but we need to count the actual messages from
	// this class.

	running_edit->IncrMessagesCount();

} // ReportSaveErrorText
//---------------------------------------------------------------------------
void TEditErrors::ReportSetWarning(TEdit* running_edit, int msg_num)
{
	// Report warnings only if the edit set option does NOT include
	// EE_SUPPRESSWARN (passed into this class from TRunEdits::RunEdits
	// via TEditErrors::Initialize
	if (FSuppressWarnings == 0)
	{
		// Called from TLibFuncs::call_set_warning
		std::string t_error_message,
					t_error_func,
					t_error_type,
					t_edit_tag,
					t_edit_name,
					t_admin_code;

		int t_edit_pkey = running_edit->GetPKey();

		t_edit_tag.assign(running_edit->GetEditTag());
		t_edit_name.assign(running_edit->GetEditName());
		t_admin_code.assign(running_edit->GetAdminCode());

		t_error_type.assign("W");  // Warning

		t_error_func.assign("SET_WARNING");

		try
		{
			if (create_message_query())
			{
				// Bind the msg_num to the query
				FMessageQry->BindIntParameter(1, msg_num);

				if (FMessageQry->OpenParameterizedQuery())
				{
					if (!FMessageQry->Eof())
					{
						std::string t_message = FMessageQry->GetFieldByNumber(0);

						t_error_message = expand_msg_macros(t_message, running_edit);
					}
					else
					{
						t_error_message.assign("Engine error: Message number ");
						t_error_message += intToString(msg_num);
						t_error_message += " not found";
					}

					// Reset for re-use
					FMessageQry->ResetParameterizedQuery();
				}
			} // if (create_message_query)
		}
		catch (std::runtime_error& e)
		{
			std::string t_err_msg = e.what();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TEditErrors", "ReportSetWarning", t_err_msg);
			}
			t_err_msg.assign(FMessageQry->LastErrorMessage());
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TEditErrors", "ReportSetWarning", t_err_msg);
			}
			throw;
		}

		if (FCallbackOwner && FCallbackFunction)
		{
		  (*FCallbackFunction)(FCallbackOwner,
			t_edit_tag.c_str(), t_edit_name.c_str(), t_admin_code.c_str(),
			t_error_type.c_str(), t_error_message.c_str());
		}

		// Even if we have already reported this to the callback function,
		// we'll put it into the list of messages
		TErrorObject* t_err_obj = new TErrorObject(t_error_type, t_error_message, t_error_func);
		add_error_to_map( t_edit_pkey, t_err_obj );

		running_edit->IncrWarningMessagesCount();

	} // if (FSuppressWarnings == 0)
} // ReportSetWarning
//---------------------------------------------------------------------------
void TEditErrors::ReportSaveWarningText(TEdit* running_edit, int msg_num)
{
	// Report warnings only if the edit set option does NOT include
	// EE_SUPPRESSWARN (passed into this class from TRunEdits::RunEdits
	// via TEditErrors::Initialize
	if (FSuppressWarnings == 0)
	{
		// Called from TLibFuncs::call_save_warning_text
		std::string t_error_message,
					t_error_func,
					t_error_type,
					t_edit_tag,
					t_edit_name,
					t_admin_code;

		int t_edit_pkey = running_edit->GetPKey();

		t_edit_tag.assign(running_edit->GetEditTag());
		t_edit_name.assign(running_edit->GetEditName());
		t_admin_code.assign(running_edit->GetAdminCode());

		t_error_type.assign("W");  // Warning

		t_error_func.assign("SAVE_WARNING_TEXT");

		try
		{
			if (create_message_query())
			{
				// Bind the msg_num to the query
				FMessageQry->BindIntParameter(1, msg_num);

				if (FMessageQry->OpenParameterizedQuery())
				{
					if (!FMessageQry->Eof())
					{
						std::string t_message = FMessageQry->GetFieldByNumber(0);

						t_error_message = expand_msg_macros(t_message, running_edit);
					}
					else
					{
						t_error_message.assign("Engine error: Message number ");
						t_error_message += intToString(msg_num);
						t_error_message += " not found";
					}

					// Reset for re-use
					FMessageQry->ResetParameterizedQuery();
				}
			} // if (create_message_query)
		}
		catch (std::runtime_error& e)
		{
			std::string t_err_msg = e.what();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TEditErrors", "ReportSaveWarningText", t_err_msg);
			}
			t_err_msg.assign(FMessageQry->LastErrorMessage());
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TEditErrors", "ReportSaveWarningText", t_err_msg);
			}
			throw;
		}

		if (FCallbackOwner && FCallbackFunction)
		{
		  (*FCallbackFunction)(FCallbackOwner,
			t_edit_tag.c_str(), t_edit_name.c_str(), t_admin_code.c_str(),
			t_error_type.c_str(), t_error_message.c_str());
		}

		// Even if we have already reported this to the callback function,
		// we'll put it into the list of messages
		TErrorObject* t_err_obj = new TErrorObject(t_error_type, t_error_message, t_error_func);
		add_error_to_map( t_edit_pkey, t_err_obj );

		running_edit->IncrWarningMessagesCount();
	} // if (FSuppressWarnings == 0)
} // ReportSaveWarningText
//---------------------------------------------------------------------------
void TEditErrors::SetErrorText(TEdit* running_edit, std::string error_message)
{
	// Called from TLibFuncs::call_error_text

	// ReportDefaultError will check these values to see if it should be
	// using the last message called from ERROR_TEXT when the edit logic encounters
	// a RETURN FAIL.
	FError_Text_Edit_PKey = running_edit->GetPKey();
	std::string t_error_message = expand_msg_macros(error_message, running_edit);
	FError_Text_Message.assign(t_error_message);

} // SetErrorText
//---------------------------------------------------------------------------
void TEditErrors::SetSaveText(TEdit* running_edit, std::string save_message)
{
	// Called from TLibFuncs::call_save_text
	int t_edit_pkey = running_edit->GetPKey();
	std::string t_message = expand_msg_macros(save_message, running_edit);
	std::string t_error_type = "M", // "E",
				t_error_func = "SAVE_TEXT";

	std::string t_edit_tag,
				t_edit_name,
				t_admin_code;

	t_edit_tag.assign(running_edit->GetEditTag());
	t_edit_name.assign(running_edit->GetEditName());
	t_admin_code.assign(running_edit->GetAdminCode());

	if (FCallbackOwner && FCallbackFunction)
	{
	  (*FCallbackFunction)(FCallbackOwner,
		t_edit_tag.c_str(), t_edit_name.c_str(), t_admin_code.c_str(),
		t_error_type.c_str(), t_message.c_str());
	}
	// Even if we have already reported this to the callback function,
	// we'll put it into the list of messages
	TErrorObject* t_err_obj = new TErrorObject(t_error_type, t_message, t_error_func);
	add_error_to_map( t_edit_pkey, t_err_obj );
} // SetSaveText
//---------------------------------------------------------------------------
void TEditErrors::SetAltMessageNumber(int alt_message_number)
{
	// EDITS Language function ERROR_MSG (TLibFuncs::call_error_msg) changes
	// the default error message of an edit, allowing an edit to return a
	// context-specific error message. TRunEdits::ProcessEdit invokes this
	// method at the start of each edit run to reset the value to 0, so that
	// ReportDefaultError knows what to do on a failure.
	if (alt_message_number >= 0)
	{
		FAltMessageNumber = alt_message_number;
	}
}
//---------------------------------------------------------------------------

