//---------------------------------------------------------------------------
#include <memory>
#include <algorithm> // std::transform
#include <stdexcept>
#include <cstdlib>  // std::atoi
#include <cstdio>   // std::snprintf

#include "EditSet.h"
#include "SQLDB.h"        // for TSQLDB
#include "FastSQLQuery.h" // for faster queries (no "field-by-name" support)
#include "Utility.h"
#include "EngineErrorsPump.h"

//---------------------------------------------------------------------------
/*****************************************************************************
 *** class TEdit
 ****************************************************************************/
//---------------------------------------------------------------------------
TEdit::TEdit(TSQLDB* database, TEngineErrorsPump* pump,
		const int edit_pkey, const std::string edit_name,
		const std::string edit_tag, const std::string admin_code,
		const std::string default_message_number,
		const std::string default_error_message)
{
	FSQLDB = database;
	FEngineErrorsPump = pump;
	FPKey = edit_pkey;
	FEditName.assign(edit_name);
	FEditTag.assign(edit_tag);
	FAdminCode.assign(admin_code);
	FDefaultMessageNumber.assign(default_message_number);
	FDefaultErrorMessage.assign(default_error_message);
	FErrorsCount = 0;
	FWarningsCount = 0;
	FSkipsCount = 0;
	FErrorMessagesCount = 0;
	FWarningMessagesCount = 0;
	FTallyErrorsCount = 0;
	FTallyWarningsCount = 0;
	FTallySkipsCount = 0;
	FTallyErrorMessagesCount = 0;
	FTallyWarningMessagesCount = 0;
	FIsSuppressed = false;
	FSkipIfEmpty = false;
	FSkipIfFailed = false;
	FNoSkipsOverride = false;
	FCompiled.clear();

	FFieldsInMacroOrder.reserve(20);
	FFieldsNotInLayout.reserve(20);

} // constructor
//---------------------------------------------------------------------------
TEdit::~TEdit()
{
	// If I need it
} // destructor
//---------------------------------------------------------------------------

void TEdit::BuildEditFieldsList(TFastSQLQuery* query)
{
	// Need this as a convenience for TRunEdits::DebugEdit
	char t_fldbuff[500];

	int t_pkey = this->GetPKey();

	// Finally, retrieve the referenced fields in field macro order
	std::snprintf(t_fldbuff, 500,
		"SELECT f.FieldNumber FROM EE_EditFieldsPivot piv, EE_Fields f WHERE f.PKey = piv.FieldPKey AND piv.EditPKey = %d ORDER BY piv.FieldOrder;",
		t_pkey);

	query->SetIsActive(false);
	query->SetSQLText(t_fldbuff);
	try
	{
		query->SetIsActive(true);
		while (!query->Eof())
		{
			int t_fldnum = query->GetIntFieldByNumber(0);
			this->AddField(t_fldnum);

			query->FetchNext();
		}
		query->SetIsActive(false);
	}
	catch( std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TEdit", "BuildEditFieldsList", e.what());
		std::string t_error = (std::string)"SQL: " + (std::string)t_fldbuff;
		FEngineErrorsPump->AddError("TEdit", "BuildEditFieldsList", t_error);
		throw;
	}

} // BuildEditFieldsList
//---------------------------------------------------------------------------

std::string TEdit::GetCompiledLogic(TFastSQLQuery* edit_logic_qry)
{
	std::string t_compiled;

	// If we have been here before, we already have the compiled code.
	if (!FCompiled.empty())
	{
		t_compiled.assign(FCompiled);
	}
	else
	{
		edit_logic_qry->BindIntParameter(1, FPKey);
		try
		{
			if (edit_logic_qry->OpenParameterizedQuery())
			{
				FCompiled.assign(edit_logic_qry->GetFieldByNumber(0));
				t_compiled.assign(FCompiled);
			}

			edit_logic_qry->ResetParameterizedQuery();
		}
		catch (...)
		{
			edit_logic_qry->ResetParameterizedQuery();
			throw;
		}

	}
	return t_compiled;
} // GetCompiledLogic()
//---------------------------------------------------------------------------
void TEdit::ClearCounts()
{
	FErrorsCount = 0;
	FWarningsCount = 0;
	FSkipsCount = 0;
	FErrorMessagesCount = 0;
	FWarningMessagesCount = 0;
} // ClearCounts
//---------------------------------------------------------------------------
void TEdit::IncrErrorsCount()
{
	FErrorsCount += 1;
	FTallyErrorsCount += 1;
}
//---------------------------------------------------------------------------
void TEdit::IncrWarningsCount()
{
	FWarningsCount += 1;
	FTallyWarningsCount += 1;
}
//---------------------------------------------------------------------------
void TEdit::IncrWarningMessagesCount()
{
	FWarningMessagesCount += 1;
	FTallyWarningMessagesCount += 1;
}
//---------------------------------------------------------------------------
void TEdit::IncrSkipsCount()
{
	FSkipsCount += 1;
	FTallySkipsCount += 1;
}
//---------------------------------------------------------------------------
void TEdit::IncrMessagesCount()
{
	FErrorMessagesCount += 1;
	FTallyErrorMessagesCount += 1;
}
//---------------------------------------------------------------------------
int TEdit::GetFieldsCount()
{
	return (int)FFieldsInMacroOrder.size();
}
//---------------------------------------------------------------------------
int TEdit::GetFieldByMacroOrder(int idx)
{
	int t_count = GetFieldsCount();
	int t_field_number = -1; // error condition
	if (idx >= 0 && idx < t_count)
	{
		t_field_number = FFieldsInMacroOrder[idx];
	}
	return t_field_number;
} // GetFieldByMacroOrder
//---------------------------------------------------------------------------
void TEdit::AddField(int field_number)
{
	bool t_add_fld = true;

	for (size_t t_idx = 0; t_idx < FFieldsInMacroOrder.size(); t_idx++)
	{
		if (FFieldsInMacroOrder[t_idx] == field_number)
		{
			t_add_fld = false;
			break;
		}
	}

	if (t_add_fld)
	{
		FFieldsInMacroOrder.push_back(field_number);
	}

} // AddField
//---------------------------------------------------------------------------
void TEdit::SetFieldNotInLayout(int field_number)
{
	if (field_number > 0)
	{
		bool t_add_fld = true;

		for (size_t t_idx = 0; t_idx < FFieldsNotInLayout.size(); t_idx++)
		{
			if (FFieldsNotInLayout[t_idx] == field_number)
			{
				t_add_fld = false;
				break;
			}
		}

		if (t_add_fld)
		{
			FFieldsNotInLayout.push_back(field_number);
		}
	}
} // SetFieldNotInLayout
//---------------------------------------------------------------------------
bool TEdit::SkipEditSilently()
{
	size_t t_list_size = FFieldsNotInLayout.size();
	return (t_list_size > 0);
}
//---------------------------------------------------------------------------
/*****************************************************************************
 *** class TEditSet
 ****************************************************************************/
//---------------------------------------------------------------------------
TEditSet::TEditSet(TSQLDB* database, TEngineErrorsPump* pump)
{
	FSQLDB = database;
	FEngineErrorsPump = pump;
	// Initialize the "suppress warnings" flag to false.
	// It may be set to true by, e.g., Edit_RunEdits
	FSuppressWarnings = 0;

	FEditsInRunOrder.reserve(1000);
} // constructor
//---------------------------------------------------------------------------
TEditSet::~TEditSet()
{
	// FEditsByEditName contains ALL of the TEdit objects that were
	// created in LoadEditSet. We'll use it to free allocated memory.
	std::map<std::string, TEdit*>::iterator it;
	for (it = FEditsByEditName.begin(); it != FEditsByEditName.end(); it++)
	{
		delete(*it).second;
	}

} // destructor
//---------------------------------------------------------------------------
void TEditSet::ClearCounts()
{
	std::map<std::string, TEdit*>::iterator it;
	for (it = FEditsByEditName.begin(); it != FEditsByEditName.end(); it++)
	{
		TEdit* t_edit = (*it).second;
		t_edit->ClearCounts();
	}
}
//---------------------------------------------------------------------------
int TEditSet::LoadEditSet(const std::string edit_set_tag)
{
	int t_result = 0;
	// This is the edits control query... contains the Edit PKeys
	// Note that it's parameterized to run twice with IsMultiField set
	// to 0 (single-field edits), then 1 (multi-field edits)
	std::unique_ptr<TFastSQLQuery> t_pkeys_qry(new TFastSQLQuery(FSQLDB));
	// Parameterized query for single-field edit details
	std::unique_ptr<TFastSQLQuery> t_edit_qry(new TFastSQLQuery(FSQLDB));
	// The field numbers referenced by this edit; also parameterized
	std::unique_ptr<TFastSQLQuery> t_fld_qry(new TFastSQLQuery(FSQLDB));

	std::string t_edit_set_tag = stringReplace(edit_set_tag, "'", "''");
	char t_buffer[500];
	std::snprintf(t_buffer, 500,
		"SELECT PKey, EditSetName FROM EE_EditSets WHERE EditSetTag = '%s';",
		t_edit_set_tag.c_str());

	t_pkeys_qry->SetSQLText(t_buffer);

	try
	{
		t_pkeys_qry->SetIsActive(true);
		FPKey = t_pkeys_qry->GetIntFieldByNumber(0);
		FEditSetName.assign( t_pkeys_qry->GetFieldByNumber(1) );
		FEditSetTag.assign( edit_set_tag );

		t_pkeys_qry->SetIsActive(false);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "SQL: ";
		t_error += (std::string)t_buffer;

		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", e.what());
		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", t_error);

		t_result = -1;
	}

	// Re-use the pkeys query, but this time we'll hold the EE_EditSetEditsPivot.EditPKeys
	// We're also parameterizing it so we can change the value of the IsMultiField param,
	// below
	std::string t_sql =
		"SELECT piv.EditPKey "\
		"FROM EE_EditSetEditsPivot piv, EE_Edits a " \
		"WHERE a.PKey = piv.EditPKey " \
		"AND a.IsMultiField = ? " \
		"AND piv.EditSetPKey = ? " \
		"ORDER BY a.EditName;";

	t_pkeys_qry->SetSQLText(t_sql);

	if (!t_pkeys_qry->PrepareParameterizedQuery())
	{
		std::string t_error = "SQL: ";
		t_error += t_sql;

		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", "Failure to prepare edit PKeys parameterized query");
		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", t_error);

		t_result = -1;
	}

	// Create parameterized query for edit details
	t_sql =
		"SELECT a.EditName, a.EditTag, b.AdminCode, " \
		"m.MessageNumber, m.Message " \
		"FROM EE_Edits a, EE_Agencies b, EE_Messages m " \
		"WHERE b.PKey = a.AgencyPKey " \
		"AND m.PKey = a.MessagePKey " \
		"AND a.PKey = ?;";

	t_edit_qry->SetSQLText(t_sql);

	if (!t_edit_qry->PrepareParameterizedQuery())
	{
		std::string t_error = "SQL: ";
		t_error += (std::string)t_buffer;

		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", "Failure to prepare edit details parameterized query");
		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", t_error);

		t_result = -1;
	}

	// Create parameterized query for fields referenced by edit
	t_sql =
		"SELECT f.FieldNumber " \
		"FROM EE_EditFieldsPivot piv, EE_Fields f " \
		"WHERE f.PKey = piv.FieldPKey " \
		"AND piv.EditPKey = ? " \
		"ORDER BY piv.FieldOrder;";

	t_fld_qry->SetSQLText(t_sql);

	if (!t_fld_qry->PrepareParameterizedQuery())
	{
		std::string t_error = "SQL: ";
		t_error += (std::string)t_buffer;

		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", "Failure to prepare edit fieldss parameterized query");
		FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", t_error);

		t_result = -1;
	}

	// If nothing bad happened with preparing the queries, let's go!
	int t_loop = 0;
	while (t_result == 0 && t_loop < 2)
	{
		try
		{
			// t_loop is for IsMultiField=0, single-field edits
			//               IsMultiField=1, multi_field edits
			t_pkeys_qry->BindIntParameter(1, t_loop);
			t_pkeys_qry->BindIntParameter(2, FPKey); // edit set PKey

			if (t_pkeys_qry->OpenParameterizedQuery())
			{
				int t_pkey;
				std::string t_edit_name,
							t_edit_tag,
							t_admin_code,
							t_message_number,
							t_message;
				do
				{
					t_pkey = t_pkeys_qry->GetIntFieldByNumber(0);

					// 06/29/2016 KB - If edit set does not have any single- or
					//                 multi-field edits, there won't be
					//                 anything to create here.
					if (t_pkey > 0)
					{
						t_edit_qry->BindIntParameter(1, t_pkey);
						if (t_edit_qry->OpenParameterizedQuery())
						{
							t_edit_name      = t_edit_qry->GetFieldByNumber(0);
							t_edit_tag       = t_edit_qry->GetFieldByNumber(1);
							t_admin_code     = t_edit_qry->GetFieldByNumber(2);
							t_message_number = t_edit_qry->GetFieldByNumber(3);
							t_message        = t_edit_qry->GetFieldByNumber(4);

							TEdit* t_edit = FEditsByEditTag[t_edit_tag];
							if (!t_edit)
							{
								t_edit =
								new TEdit(FSQLDB, FEngineErrorsPump, t_pkey,
								t_edit_name, t_edit_tag, t_admin_code,
								t_message_number, t_message);
							}
							t_fld_qry->BindIntParameter(1, t_pkey);
							if (t_fld_qry->OpenParameterizedQuery())
							{
								int t_fldnum;
								do
								{
									t_fldnum = t_fld_qry->GetIntFieldByNumber(0);

									if (t_fldnum > 0)
									{
										t_edit->AddField(t_fldnum);
									}

									t_fld_qry->FetchNext();

								} while (!t_fld_qry->Eof());
							}

							t_fld_qry->ResetParameterizedQuery();

							FEditsInRunOrder.push_back(t_edit);
							FEditsByEditName[t_edit_name] = t_edit;
							FEditsByEditTag[t_edit_tag] = t_edit;

						} // if (t_edit_qry->OpenParameterizedQuery())
					} // if (t_pkey > 0)

					t_edit_qry->ResetParameterizedQuery();

					// On to the next EE_Edits PKey
					t_pkeys_qry->FetchNext();

				} while (!t_pkeys_qry->Eof());
			} // if (t_pkeys_qry->OpenParameterizedQuery())

			t_pkeys_qry->ResetParameterizedQuery();

			t_loop++;
		}
		catch (std::runtime_error& e)
		{
			FEngineErrorsPump->AddError("TEditSet", "LoadEditSet", e.what());
			t_result = -1;
		}
	} // while (t_result == 0 && t_loop < 2)

	t_pkeys_qry->ReleaseParameterizedQuery();
	t_edit_qry->ReleaseParameterizedQuery();
	t_fld_qry->ReleaseParameterizedQuery();

	return t_result;
} // LoadEditSet
//---------------------------------------------------------------------------

TEdit* TEditSet::GetEditByRunOrderIndex( int idx )
{
	int t_count = GetEditsCount();
	TEdit* t_edit = NULL;
	if (idx < t_count)
	{
		t_edit = FEditsInRunOrder[idx];
	}

	return t_edit;
} // GetEditByRunOrderIndex
//---------------------------------------------------------------------------
int TEditSet::SetEditOptions( std::string edit_tag, bool suppress_edit,
	bool skip_if_empty, bool skip_if_failed, bool no_skips_override )
{
	int t_return = 0;

	TEdit* t_edit = FEditsByEditTag[ edit_tag ];
	if (t_edit)
	{
		t_edit->SuppressEdit( suppress_edit );
		t_edit->SkipIfEmpty( skip_if_empty );
		t_edit->SkipIfFailed( skip_if_failed );
		t_edit->NoSkipsOverride( no_skips_override );
	}
	else
	{
		char t_buffer[1000];
		std::snprintf(t_buffer, 1000,
			"Cannot find edit tag \"%s\" in edit set \"%s\" [\"%s\"]",
			edit_tag.c_str(), FEditSetName.c_str(), FEditSetTag.c_str());
		FEngineErrorsPump->AddError("TEditSet", "SetEditOptions", t_buffer );
		t_return = -1;
	}
	return t_return;
} // SetEditOptions
//---------------------------------------------------------------------------
int TEditSet::ClearEditOptions()
{
	int t_return = 0;
	size_t t_count = FEditsInRunOrder.size();
	for (size_t t_idx = 0; t_idx < t_count; t_idx++)
	{
		TEdit* t_edit = FEditsInRunOrder[t_idx];
		t_edit->SuppressEdit( false );
		t_edit->SkipIfEmpty( false );
		t_edit->SkipIfFailed( false );
		t_edit->NoSkipsOverride( false );
	}

	return t_return;
} // ClearEditOptions
//---------------------------------------------------------------------------
int TEditSet::GetEditOptions( std::string edit_tag, int* is_suppressed,
	int* is_skip_if_empty, int* is_skip_if_failed, int* is_no_skips_override )
{
	int t_return = 0;

	TEdit* t_edit = FEditsByEditTag[ edit_tag ];
	if (t_edit)
	{
		*is_suppressed = (t_edit->IsSuppressed() ? 1 : 0);
		*is_skip_if_empty = (t_edit->IsSkipIfEmpty() ? 1 : 0);
		*is_skip_if_failed = (t_edit->IsSkipIfFailed() ? 1 : 0);
		*is_no_skips_override = (t_edit->IsNoSkipsOverride() ? 1 : 0);
	}
	else
	{
		char t_buffer[1000];
		std::snprintf(t_buffer, 1000,
			"Cannot find edit tag \"%s\" in edit set \"%s\" [\"%s\"]",
			edit_tag.c_str(), FEditSetName.c_str(), FEditSetTag.c_str());
		FEngineErrorsPump->AddError("TEditSet", "GetEditOptions", t_buffer );
		t_return = -1;
	}

	return t_return;
} // GetEditOptions
//---------------------------------------------------------------------------




