//---------------------------------------------------------------------------

#ifndef EditSetH
#define EditSetH
//---------------------------------------------------------------------------
/* This class is for referencing Edit Sets for the Run Edits processing.
 */
#include <string>
#include <map>
#include <vector>

// Forward declarations
class TSQLDB;
class TFastSQLQuery;
class TEngineErrorsPump;

class TEdit
{
private:
	TSQLDB* FSQLDB;
	TEngineErrorsPump* FEngineErrorsPump;

	std::string FEditName;
	std::string FEditTag;
	std::string FAdminCode;
	std::string FDefaultMessageNumber; // we'll use it as a string, so get it that way
	std::string FDefaultErrorMessage;
	int FPKey;
	// These three are reset to 0 for each new data record processed through
	// TRunEdits::RunEdits
	int FErrorsCount;
	int FWarningsCount;
	int FSkipsCount;
	int FErrorMessagesCount;
	int FWarningMessagesCount;

	// These three are set to 0 when the TEdit object is instantiated. They
	// maintain the tally of all FErrorsCount, FWarningsCount and FSkipsCount
	int FTallyErrorsCount;
	int FTallyWarningsCount;
	int FTallySkipsCount;
	int FTallyErrorMessagesCount;
	int FTallyWarningMessagesCount;

	// FIsSuppressed means the edit is not to be run
	bool FIsSuppressed;
	// FSkipIfEmpty and FSkipIfFailed could be set either by calling
	// Edit_SetEditSetOptions OR Edit_SetEditOptions.
	bool FSkipIfEmpty;
	bool FSkipIfFailed;
	// FNoSkipsOverride means that the Edit_RunEdits edit_options had instructions
	// for skipping if empty or failed single-field edit, but the user wants
	// THIS edit to run regardless.
	bool FNoSkipsOverride;
	std::string FCompiled; // compiled edit logic
	// This vector lists all of the fields referenced by the edit,
	// in the "field macro order".
	std::vector<int> FFieldsInMacroOrder;
	// This vector holds the EE_Fields.FieldNumber for a field that is referenced
	// by this edit, but is NOT in the layout associated with the edit set
	// that contains this edit. It will be updated by TRunEdits::PrepareEditSet.
	std::vector<int> FFieldsNotInLayout;

public:
	TEdit(TSQLDB* database, TEngineErrorsPump* pump, const int edit_pkey,
		const std::string edit_name, const std::string edit_tag,
		const std::string admin_code,
		const std::string default_message_number,
		const std::string default_error_message);
	~TEdit();

	// Need this as a convenience for TRunEdits::DebugEdit
	void BuildEditFieldsList(TFastSQLQuery* query);

	// property getters
	int GetPKey() { return FPKey; }
	std::string GetEditName() { return FEditName; }
	std::string GetEditTag() { return FEditTag; }
	std::string GetAdminCode() { return FAdminCode; }
	std::string GetDefaultMessageNumber() { return FDefaultMessageNumber; }
	std::string GetDefaultErrorMessage() { return FDefaultErrorMessage; }

	int GetErrorsCount() { return FErrorsCount; }
	int GetWarningsCount() { return FWarningsCount; }
	int GetSkipsCount() { return FSkipsCount; }
	int GetErrorMessagesCount() { return FErrorMessagesCount; }
	int GetWarningMessagesCount() { return FWarningMessagesCount; }

	int GetTallyErrorsCount() { return FTallyErrorsCount; }
	int GetTallyWarningsCount() { return FTallyWarningsCount; }
	int GetTallySkipsCount() { return FTallySkipsCount; }
	int GetTallyErrorMessagesCount() { return FTallyErrorMessagesCount; }
	int GetTallyWarningMessagesCount() { return FTallyWarningMessagesCount; }

	std::string GetCompiledLogic(TFastSQLQuery* edit_logic_qry);

	// Incrementers for the count fields
	void IncrErrorsCount();
	void IncrWarningsCount();
	void IncrSkipsCount();
	void IncrMessagesCount();
	void IncrWarningMessagesCount();

	void ClearCounts();

	void SuppressEdit( bool suppress_edit ) { FIsSuppressed = suppress_edit; }
	bool IsSuppressed() { return FIsSuppressed; } ;

	void SkipIfEmpty( bool skip_if_empty ) { FSkipIfEmpty = skip_if_empty; }
	bool IsSkipIfEmpty() { return FSkipIfEmpty; }

	void SkipIfFailed( bool skip_if_failed ) { FSkipIfFailed = skip_if_failed; }
	bool IsSkipIfFailed() { return FSkipIfFailed; }

	void NoSkipsOverride( bool no_skips_override ) { FNoSkipsOverride = no_skips_override; }
	bool IsNoSkipsOverride() { return FNoSkipsOverride; }

	// Fields referenced by the edit logic, added in "field macro order"
	void AddField(int field_number);
	int GetFieldsCount();
	int GetFieldByMacroOrder(int idx);

	// Called from TRunEdits::PrepareEditSet
	void SetFieldNotInLayout(int field_number);

	bool SkipEditSilently();

};

class TEditSet
{
private:
	TSQLDB* FSQLDB;
	TEngineErrorsPump* FEngineErrorsPump;

	std::string FEditSetName;
	std::string FEditSetTag;
	int FPKey;
	int FSuppressWarnings;

	std::map<std::string, TEdit*> FEditsByEditName;
	std::map<std::string, TEdit*> FEditsByEditTag;
	// FEditsInRunOrder lists the edits this way:
	//  * Single-field or no-field edits first, sorted alphabetically by edit name
	//  * Multi-field edits, sorted alphabetically by edit name
	std::vector<TEdit*> FEditsInRunOrder;

public:
	TEditSet(TSQLDB* database, TEngineErrorsPump* pump);
	~TEditSet();

	// getters
	std::string GetEditSetName() { return FEditSetName; }
	std::string GetEditSetTag() { return FEditSetTag; }
	int GetPKey() { return FPKey; };

	int LoadEditSet(const std::string edit_set_tag);
	void ClearCounts();

	int GetEditsCount() { return (int)FEditsInRunOrder.size(); }
	TEdit* GetEditByRunOrderIndex( int idx );

	// Set/Get/Clear edit options
	int SetEditOptions( std::string edit_tag, bool suppress_edit, bool skip_if_empty, bool skip_if_failed, bool no_skips_override );
	int GetEditOptions( std::string edit_tag, int* is_suppressed, int* is_skip_if_empty, int* is_skip_if_failed, int* is_no_skips_override );
	int ClearEditOptions();

	// The RunEdits() functions also allow "suppress warnings", which
	// we need to have available when we are generating results back
	// to a caller. Get and set it here.
	int GetSuppressWarnings() { return FSuppressWarnings; }
	void SetSuppressWarnings(int suppress_warnings) { FSuppressWarnings=suppress_warnings; }
};

#endif
