//---------------------------------------------------------------------------

#ifndef ExternalDLLTrackerH
#define ExternalDLLTrackerH
//---------------------------------------------------------------------------
#include <string>
#include <vector>

/*   EDITS may call functions in external libraries via EDITS Language
     function ExternalDLL(). That process involves calling Windows API
     LoadLibrary(). In order to minimize performance hits, we'll load
     a library only once, then hold onto the pointers to the library
     and the EditsInterface function within that library, and not
     FreeLibrary() until the Engine doesn't need it anymore.

     The trick is this: Multiple instances of calls to LoadLibrary
     for the same library will return the same hLib, a Microsoft
     efficiency. It is therefore possible for two metafiles to load
     the same libary, each thereby getting the same hLib, and then
     both of them try to FreeLibrary(hLib)... causing the second
     attempt to throw an error.

     We therefore moved the maintenance of a std::map of these
	 pointers outside the TSMF class, so that the DllMain
     will call the destructor function when the entire Engine
     shuts down.

     NOTE: This does mean that users will not be able to change
     a called DLL while the Engine is running, but we regard that
     need to be slight and manageable (stop the Engine first!).
	 -KBeaumont 02/07/2014
 */
#ifdef EE_WINDOWS
#include <windows.h>  // HINSTANCE, LoadLibrary, FreeLibrary
#else
#include <dlfcn.h>    // Linux: dlopen, dlsym, dlclose
#endif

typedef struct
{
	std::string DLLname;
#ifdef EE_WINDOWS
	HINSTANCE hLib;             // result of LoadLibrary()
#else  // Linux
	void* hLib;                 // result of dlopen("CStage.so", RTLD_LAZY)
#endif
	void* EditsInterfaceFunc;   // hLib->GetProcAddress("EditsInterface")
								// dlsym(hLib, "EditsInterface");
} EEt_ExternalDLL;

// For TLibFuncs::call_externaldll
static std::vector<EEt_ExternalDLL*> FExternalDLLs;

// Note: This is called from the DLL_PROCESS_DETACH in DllMain, via
// ShutDown() in EditAPI.cpp
void ClearExternalDLLs();

EEt_ExternalDLL* GetExternalDLL(std::string dll_name);
#ifdef EE_WINDOWS
void AddExternalDLL(std::string dll_name, HINSTANCE hLib, void* func_ptr);
#else
void AddExternalDLL(std::string dll_name, void* hLib, void* func_ptr);
#endif

//---------------------------------------------------------------------------

#endif
