#include <algorithm> // std::transform
#include <cstring>   // std::strlen
#include <stdexcept>

#ifdef EE_WINDOWS
#include <windows.h> // Sleep()
#else
#include <unistd.h>  // sleep()
#endif

#include "FastSQLQuery.h"
#include "SQLDB.h" // for TSQLDB

/*****************************************
 ***  TFastSQLQuery
 *****************************************/
TFastSQLQuery::TFastSQLQuery(TSQLDB* database)
{
	SQLDB = database;
	FSQLText.clear();

	FIsActive = false;
	FStatement = NULL;
	FColCount = 0;

	FIsParameterized = false; // SetSQLText may change this to true
	FPreparedCount = 0;

} // constructor
//---------------------------------------------------------------------------
TFastSQLQuery::~TFastSQLQuery()
{
	force_finalize_query();
} // destructor
//---------------------------------------------------------------------------
void TFastSQLQuery::force_finalize_query()
{
	// called from the destructor
	int t_count = get_prepared_count();
	int t_rc = SQLITE_OK, t_idx;
	for (t_idx = 0; ((t_idx < t_count) && (t_rc == SQLITE_OK)); t_idx++)
	{
		try
		{
			t_rc = sqlite3_finalize(FStatement);
		}
		catch (...)
		{
			std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

			FLastErrorMessage.assign(t_err_msg);
			throw std::runtime_error("TFastSQLQuery::force_finalize_query");
		}
	}
} // force_finalize_query
//---------------------------------------------------------------------------
int TFastSQLQuery::get_prepared_count()
{
	return FPreparedCount;
}
//---------------------------------------------------------------------------
void TFastSQLQuery::incr_prepared_count()
{
	FPreparedCount += 1;
}
//---------------------------------------------------------------------------
void TFastSQLQuery::decr_prepared_count()
{
	if (FPreparedCount > 0)
	{
		FPreparedCount -= 1;
	}
}
//---------------------------------------------------------------------------
bool TFastSQLQuery::Eof()
{
	return FEof;
}
//---------------------------------------------------------------------------
int TFastSQLQuery::ColCount()
{
	return FColCount;
}
//---------------------------------------------------------------------------
bool TFastSQLQuery::PrepareParameterizedQuery()
{
	if (!reset_statement())
	{
		FLastErrorMessage.assign("Failure to reset statement object");
	}
	else
	{
		FColCount = 0;
		FEof = false;
		const char* t_tail;
		int t_rc;

		// NB: See comments in TFastSQLQuery::close_query about
		// the requirement to balance calls to sqlite3_prepare_v2
		// with corresponding calls to sqlite3_finalize... lest
		// you leave the database in a state that it will refuse to
		// close.
		t_rc = sqlite3_prepare_v2(
			SQLDB->Database(),        // Database handle
			(char*)FSQLText.c_str(),  // SQL text
			(int)FSQLText.length(),   // Length of SQL Text in bytes
			&FStatement,              // OUT: statement handle
			&t_tail);                 // OUT: pointer to unused portion of SQL Text

		if (t_rc == SQLITE_OK)
		{
			incr_prepared_count();
			return true;
		}
		else
		{
			std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

			FLastErrorMessage.assign(t_err_msg);
		}
	} // if (!reset_statement())
	// If we got to here, we failed
	throw std::runtime_error("TFastSQLQuery::PrepareParameterizedQuery");
}  // PrepareParameterizedQuery
//---------------------------------------------------------------------------
bool TFastSQLQuery::reset_statement()
{
	// NOTE: This is called for BOTH parameterized and non-parameterized
	// queries. For the former, we are handling these calls internally, so
	// we don't need to bark at the user... see note in ResetParameterizedQuery
	// Clear this for re-use
	int t_rc = sqlite3_reset(FStatement);

	if (t_rc != SQLITE_OK)
	{
		FLastErrorMessage.assign("Failure to reset statement object");
		throw std::runtime_error("TFastSQLQuery::reset_statement");
	}

	return true;

} // reset_statement
//---------------------------------------------------------------------------
bool TFastSQLQuery::ResetParameterizedQuery()
{
	// The caller uses this to reset a previously-prepared query so that
	// new parameter values may be bound and it can be re-run. If the
	// query is not parameterized, there is no need to call this. True,
	// there is no harm, either, but I want programmers to think of these
	// functions in the context of sets and use them appropriately.
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: inappropriate call");
		throw std::runtime_error("TFastSQLQuery::ResetParameterizedQuery");
	}

	return reset_statement();

} // ResetParameterizedQuery
//---------------------------------------------------------------------------
bool TFastSQLQuery::BindIntParameter(const int param_num, const int value)
{

	int t_rc = sqlite3_bind_int(FStatement, param_num, value);

	if (t_rc == SQLITE_OK)
	{
		return true;
	}
	else if (t_rc == SQLITE_MISUSE)
	{
		std::string t_err_msg = "unknown error (SQLITE_MISUSE)";
		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TFastSQLQuery::BindIntParameter");
	}
	else
	{
		std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TFastSQLQuery::BindIntParameter");
	}
} // BindIntParameter
//---------------------------------------------------------------------------
bool TFastSQLQuery::BindTextParameter(const int param_num, const char* value)
{
	int t_rc = sqlite3_bind_text(FStatement, param_num,
		value, std::strlen(value), SQLITE_TRANSIENT);

	if (t_rc == SQLITE_OK)
	{
		return true;
	}
	else if (t_rc == SQLITE_RANGE)
	{
		std::string t_err_msg;
		char t_buffer[100];
		std::snprintf(t_buffer, 100,
			"parameter index [%d] out of range (SQLITE_RANGE error)", param_num);
		FLastErrorMessage.assign(t_buffer);
		throw std::runtime_error("TFastSQLQuery::BindTextParameter");
	}
	else if (t_rc == SQLITE_MISUSE)
	{
		std::string t_err_msg = "unknown error (SQLITE_MISUSE)";
		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TFastSQLQuery::BindTextParameter");
	}
	else
	{
		std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TFastSQLQuery::BindTextParameter");
	}
} // BindTextParameter
//---------------------------------------------------------------------------
bool TFastSQLQuery::OpenParameterizedQuery()
{
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: use open_query instead");
		throw std::runtime_error("TFastSQLQuery::OpenParameterizedQuery");
	}

	std::string t_column_name;

	int t_param_count = sqlite3_bind_parameter_count(FStatement);
	if (t_param_count > 0)
	{
		// Position on the first row in the result set
		int t_rc = sqlite3_step(FStatement);

		if (t_rc == SQLITE_ROW)
		{
			FEof = false;
		}
		else // empty result set/end of file
		{
			FEof = true;
		}

		// Regardless whether we have an empty result set,
		// get the number of columns in the result set
		FColCount = sqlite3_column_count(FStatement);

		// Clear any previous error message
		FLastErrorMessage.clear();

		// Set the public flag
		FIsActive = true;

		return true;
	}
	// t_param_count is not greater than 0
	FLastErrorMessage.assign("Parameterized query has no bound parameters");
	throw std::runtime_error("TFastSQLQuery::OpenParameterizedQuery");
} // OpenParameterizedQuery
//---------------------------------------------------------------------------
bool TFastSQLQuery::ReleaseParameterizedQuery()
{
	// The caller uses this to finalize a previously-prepared query. If the
	// query is not parameterized, there is no need to call this. True,
	// there is no harm, either, but I want programmers to think of these
	// functions in the context of sets and use them appropriately.
	// Therefore, even though this function itself calls SetIsActive(false),
	// it will bark at the programmer to use the functions that belong
	// together.
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: use SetIsActive instead");
		throw std::runtime_error("TFastSQLQuery::ReleaseParameterizedQuery");
	}

	// This function exists ONLY to provide a "set"... goes with
	// PrepareParameterizedQuery
	return finalize_query();
}
//---------------------------------------------------------------------------
bool TFastSQLQuery::open_query()
{
	if (!reset_statement())
	{
		FLastErrorMessage.assign("TFastSQLQuery::open_query failed to reset statement object");
	}
	else
	{
		FColCount = 0;
		FEof = false;
		const char* t_tail;
		int t_rc;

		// NB: See comments in TFastSQLQuery::close_query about
		// the requirement to balance calls to sqlite3_prepare_v2
		// with corresponding calls to sqlite3_finalize... lest
		// you leave the database in a state that it will refuse to
		// close.
		t_rc = sqlite3_prepare_v2(
			SQLDB->Database(),        // Database handle
			(char*)FSQLText.c_str(),  // SQL text
			(int)FSQLText.length(),   // Length of SQL Text in bytes
			&FStatement,              // OUT: statement handle
			&t_tail);                 // OUT: pointer to unused portion of SQL Text

		if (t_rc == SQLITE_OK)
		{

			incr_prepared_count();

			// Position on the first row in the result set
			t_rc = sqlite3_step(FStatement);

			if (t_rc != SQLITE_ROW) // empty result set/end of file
			{
				FEof = true;
			}

			// Regardless whether we have an empty result set,
			// get the number of columns in the result set
			FColCount = sqlite3_column_count(FStatement);


			// Clear any previous error message
			FLastErrorMessage.clear();
			// Set the public flag
			FIsActive = true;

			return true;
		}
		else
		{
			std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

			FLastErrorMessage.assign(t_err_msg);
			throw std::runtime_error("TFastSQLQuery::open_query");
		}
	} // if (!reset_statement)
	// If we (somehow) got to here, we do not have an active query
	FIsActive = false;
	return false;
} // open_query
//---------------------------------------------------------------------------
bool TFastSQLQuery::close_query()
{
	return finalize_query();

} // close_query
//---------------------------------------------------------------------------
bool TFastSQLQuery::finalize_query()
{
	// NOTE: Every call to sqlite3_prepare_v2 MUST HAVE a corresponding
	// call to sqlite3_finalize. If not, the "close database" call to
	// sqlite3_close will fail with an error message mumbling about
	// unfinalized statements.
	//
	// The trick was figuring out how to balance these without trashing
	// the FStatement while a cursor query is open.

	// Assume the best...
	int t_rc = SQLITE_OK;

	if (FIsActive)
	{
		FIsActive = false;
		/* Refer to documents for sqlite3_finalize, and note particularly
		   this:
		   It is a grievous error for the application to try to use a prepared
		   statement after it has been finalized. Any use of a prepared
		   statement after it has been finalized can result in undefined and
		   undesirable behavior such as segfaults and heap corruption.

		   It is therefore critical to assign FStatement = NULL after
		   invoking sqlite3_finalize.
		*/
		try
		{
			t_rc = sqlite3_finalize(FStatement);
			if (t_rc == SQLITE_OK)
			{
				decr_prepared_count();
			}
		}
		catch (...)
		{
			FStatement = NULL;

			std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

			FLastErrorMessage.assign(t_err_msg);
			throw std::runtime_error("TFastSQLQuery::finalize_query");
		}

		FStatement = NULL;
	}
	return (t_rc == SQLITE_OK);
} // finalize_query
//---------------------------------------------------------------------------
bool TFastSQLQuery::FetchNext()
{
	int t_rc = -1; // initialize to meaningless value
    int t_retries = 0;

    FLastErrorMessage.clear();

    if (FEof)
    {
        FLastErrorMessage.assign("End of result set; no more records");
        return false;
    }
    else
    {
		while ((t_rc != SQLITE_ROW) && (!FEof) && (t_retries < 3))
        {
            t_rc = sqlite3_step(FStatement);

            switch (t_rc)
            {
                case SQLITE_DONE:
					FEof = true;
                    break;

                case SQLITE_BUSY:
#ifdef EE_WINDOWS
                    Sleep(1000); // sleep for 1 second
#else
		            sleep(1);
#endif
                    if (++t_retries > 3)
                        FLastErrorMessage.assign("Database busy; cannot fetch next record");
                    break;

                case SQLITE_MISUSE:
                    FLastErrorMessage.assign("SQLITE_MISUSE: Statement is bad");
                    // force out of this loop
                    t_retries = 4;
                    break;

                case SQLITE_ERROR:
					FLastErrorMessage.assign(SQLDB->GetLastDBErrorMessage());
                    // force out of this loop
                    t_retries = 4;
                    break;

                default:
                    // I'll assume (for starters) that we get here because
                    // t_rc == SQLITE_ROW
                    if (t_rc != SQLITE_ROW)
                    {
                        FLastErrorMessage.assign(SQLDB->GetLastDBErrorMessage());
                        // force out of this loop
                        t_retries = 4;
                    }
                    break;
            }  // switch (t_rc)
        } // while ((t_rc != SQLITE_ROW) && (!FEof) && (t_retries < 3))
    } // if (!FEof)

	return (t_rc == SQLITE_ROW || t_rc == SQLITE_DONE);
}  // FetchNext()
//---------------------------------------------------------------------------
char* TFastSQLQuery::GetFieldByNumber(int column_number)
{
	const char* t_value = "";
	if (FIsActive)
	{
		t_value = (char*)sqlite3_column_text(FStatement, column_number);
		// sqlite3_column_text will return NULL if there is no data in the
		// column, which can lead to an exception when a calling routine
		// assumes the value will be a NULL-terminated string. Check this
		// value, and fix it if necessary.
		if (!t_value)
			t_value = "";
	}
	return (char*)t_value;
} // GetFieldByNumber
//---------------------------------------------------------------------------
int TFastSQLQuery::GetIntFieldByNumber(int column_number)
{
	int t_value = 0;
	if (FIsActive)
	{
		t_value = sqlite3_column_int(FStatement, column_number);
	}
	return t_value;
} // GetFieldByNumber
//---------------------------------------------------------------------------
bool TFastSQLQuery::GetIsActive()
{
	return FIsActive;
}
//---------------------------------------------------------------------------
bool TFastSQLQuery::SetIsActive(bool is_active)
{
	if (is_active)
	{
		// if the query is already active, close it and reopen it
		if (FIsActive)
			close_query();

		return open_query();
	}
	else
	{
		return close_query();
	}
}
//---------------------------------------------------------------------------
bool TFastSQLQuery::GetIsParameterized()
{
	return FIsParameterized;
}
//---------------------------------------------------------------------------
void TFastSQLQuery::set_is_parameterized(bool is_parameterized)
{
	FIsParameterized = is_parameterized;
}
//---------------------------------------------------------------------------
char* TFastSQLQuery::GetSQLText()
{
	return (char*)FSQLText.c_str();
}
//---------------------------------------------------------------------------
void TFastSQLQuery::SetSQLText(std::string sql_text, bool check_parameterized /* = true */)
{
	// Until proven otherwise...
	set_is_parameterized(false);

	if (!sql_text.empty())
	{
		if (check_parameterized)
		{
			// Check to see if this is a parameterized query, and set the
			// FIsParameterized property accordingly.

			// Note that check_parameterized is false when called from
			// TLibFuncs::internal_lookup. We may find other reasons
			// why we should make this option more public, but let's
			// wait to see...
			size_t t_pos = sql_text.find("?");
			if (t_pos != std::string::npos)
			{
				set_is_parameterized(true);
			}
		}
	}
	// Empty string or not, assign the string the user provided.
	FSQLText.assign(sql_text);
} // SetSQLText
//---------------------------------------------------------------------------
char* TFastSQLQuery::LastErrorMessage()
{
	// Support for property LastErrorMessage, a read-only property
	return (char*)FLastErrorMessage.c_str();
}
//---------------------------------------------------------------------------

