//---------------------------------------------------------------------------
#include <memory>
#include <algorithm> // std::transform
#include <stdexcept>
#include <cstdlib>  // std::atoi, std::malloc, std::free
#include <cstdio>   // std::snprintf
#include <cstring>  // memset

#include "Layout.h"
#include "edit_external.h" // callback function templates
#include "SQLDB.h"
#include "FastSQLQuery.h"
#include "Utility.h"
#include "EngineErrorsPump.h"

//---------------------------------------------------------------------------
/*****************************************************************************
 *** class TLayoutField
 ****************************************************************************/
//---------------------------------------------------------------------------
TLayoutField::TLayoutField(TLayout* layout, TSQLDB* database, TEngineErrorsPump* pump,
	int field_pkey, int start_pos)
{
    FLayout = layout;
	FSQLDB = database;
	FSetupFieldQry = NULL;
	FEngineErrorsPump = pump;

	// Initialize the members, assigning only the FPKey and FStartPos
	FPKey = field_pkey;
	FStartPos = start_pos;

	FFieldNumber = 0;
	FFieldName.clear();
	FDisplayName.clear();
	FFieldLength = 0;
	FDataType.clear();

	FFailedSingleFieldEdit = false;
	FNeverSkipEmpty = false;

	setup_layout_field();
} // constructor
//---------------------------------------------------------------------------
TLayoutField::~TLayoutField()
{
	if (FSetupFieldQry)
	{
		FSetupFieldQry->ReleaseParameterizedQuery();
		delete FSetupFieldQry;
	}
} // destructor
//---------------------------------------------------------------------------
bool TLayoutField::create_setup_field_query()
{
	if (!FSetupFieldQry)
	{
		std::string t_sql;
		FSetupFieldQry = new TFastSQLQuery(FSQLDB);

		t_sql.assign("SELECT [FieldNumber], [FieldName], [FieldLength], [DataType] ");
		t_sql += "FROM [EE_Fields] WHERE [PKey] = ?;";


		FSetupFieldQry->SetSQLText(t_sql);

		if (!FSetupFieldQry->PrepareParameterizedQuery())
		{
			FEngineErrorsPump->AddError("TLayoutField", "create_setup_field_query",
				"Failure to prepare FSetupFieldQry");
			std::string t_err_msg = FSetupFieldQry->LastErrorMessage();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TLayoutField", "create_setup_field_query", t_err_msg);
			}
			throw std::runtime_error("");
		}

	}
	return true;
} // create_setup_field_query
//---------------------------------------------------------------------------

void TLayoutField::setup_layout_field()
{
	if (create_setup_field_query())
	{
		FSetupFieldQry->BindIntParameter(1, FPKey);

		if (FSetupFieldQry->OpenParameterizedQuery())
		{
			if (!FSetupFieldQry->Eof())
			{
				FFieldNumber = FSetupFieldQry->GetIntFieldByNumber(0);
				FFieldName = FSetupFieldQry->GetFieldByNumber(1);
				FFieldLength = FSetupFieldQry->GetIntFieldByNumber(2);
				FDataType = FSetupFieldQry->GetFieldByNumber(3);
			}
			FSetupFieldQry->ResetParameterizedQuery();
		}
	}
} // setup_layout_field
//---------------------------------------------------------------------------
int TLayoutField::GetPKey()
{
	return FPKey;
}
//---------------------------------------------------------------------------
int TLayoutField::GetStartPos()
{
	return FStartPos;
}
//---------------------------------------------------------------------------
int TLayoutField::GetFieldNumber()
{
	return FFieldNumber;
}
//---------------------------------------------------------------------------
int TLayoutField::GetFieldLength()
{
	return FFieldLength;
}
//---------------------------------------------------------------------------
char* TLayoutField::GetFieldName()
{
	return std::strncpy(FLayout->LayoutFieldBuffer,
						(char*)FFieldName.c_str(),
						FLayout->LayoutFieldBuffSize);
}
//---------------------------------------------------------------------------
char* TLayoutField::GetDisplayName()
{
	return std::strncpy(FLayout->LayoutFieldBuffer,
						(char*)FDisplayName.c_str(),
						FLayout->LayoutFieldBuffSize);
}
//---------------------------------------------------------------------------
int TLayoutField::SetDisplayName(const char* display_name)
{
	// We don't want to set the display name unless it is different from
	// the field's actual name. This is so we don't display error messages
	// that redundantly list the name (e.g.,
	// "Date of Diagnosis [Date of Diagnosis]: invalid as to day"
	std::string t_field_name,
				t_display_name;

	t_field_name.assign(FFieldName);
	t_display_name.assign(display_name);

	uppercase(t_field_name);
	uppercase(t_display_name);

	if (t_field_name.compare(t_display_name) == 0)
		FDisplayName.clear();
	else
		FDisplayName.assign(display_name);

	return 0;
}
//---------------------------------------------------------------------------
char* TLayoutField::GetDataType()
{
	return std::strncpy(FLayout->LayoutFieldBuffer,
						(char*)FDataType.c_str(),
						FLayout->LayoutFieldBuffSize);
}
//---------------------------------------------------------------------------
char* TLayoutField::GetValue()
{
	return FLayout->FillFieldValue(FStartPos, FFieldLength);
}
//---------------------------------------------------------------------------
/*****************************************************************************
 *** class TLayout
 ****************************************************************************/
//---------------------------------------------------------------------------
TLayout::TLayout(TSQLDB* database, TEngineErrorsPump* pump)
{
	FSQLDB = database;
	FEngineErrorsPump = pump;
	FData = NULL;
	FDataLength = 0;

	LayoutFieldBuffSize = 0;
	LayoutFieldBuffer = NULL;

	SetupLayoutFieldBuffer(2500);


} // constructor
//---------------------------------------------------------------------------
TLayout::~TLayout()
{
	// FFieldsByStartPos contains ALL of the TLayoutField objects that were
	// created in create_layout_fields; FFieldsByNumber holds the same objects.
	// So we're using FFieldsByStartPos to ensure that we free all of the
	// allocated memory.
	size_t t_count = FFieldsByStartPos.size();
	for (size_t t_idx = 0; t_idx < t_count; t_idx++)
	{
		delete FFieldsByStartPos[t_idx];
	}

	if (LayoutFieldBuffer)
	{
		std::free(LayoutFieldBuffer);
		LayoutFieldBuffer = NULL;
	}

	if (FData)
	{
		std::free(FData);
		FData = NULL;
	}
} // destructor
//---------------------------------------------------------------------------
void TLayout::SetupLayoutFieldBuffer(size_t buff_size)
{
	// LayoutFieldBuffer exists as a scratch pad so that TLayoutField
	// can write temporary values to return to callers. This is all to
	// eliminate passing std::string reference parameters to GetFieldName
	// et. al., trying to improve performance by nanoseconds.

	if (LayoutFieldBuffer)
	{
		// we're resizing; delete the previous allocation
		std::free(LayoutFieldBuffer);
	}
	// Set the LayoutFieldBuffSize to the new size;
	// the ctor defaults it to 2500
	LayoutFieldBuffSize = buff_size;

	LayoutFieldBuffer = (char*)std::malloc(LayoutFieldBuffSize+1);
} // SetupLayoutFieldBuffer
//---------------------------------------------------------------------------
void TLayout::create_layout_fields()
{
	std::unique_ptr<TFastSQLQuery> t_qry(new TFastSQLQuery(FSQLDB));
	char t_buffer[200];
	std::string t_sql =
		"SELECT p.FieldPKey, p.StartPos FROM EE_LayoutFieldsPivot p, EE_Fields f ";
	t_sql += "WHERE p.FieldPKey = f.PKey AND p.LayoutPKey = %d " \
			 "ORDER BY p.StartPos, f.FieldLength DESC;";

	std::snprintf(t_buffer, 200, t_sql.c_str(), FPKey);

	t_qry->SetSQLText(t_buffer);
	try
	{
		t_qry->SetIsActive(true);
		int t_field_pkey,
		    t_start_pos;

		while (!t_qry->Eof())
		{
			t_field_pkey = t_qry->GetIntFieldByNumber(0);
			t_start_pos  = t_qry->GetIntFieldByNumber(1);

			TLayoutField* t_fld =
				new TLayoutField(this, FSQLDB, FEngineErrorsPump,
					t_field_pkey, t_start_pos);

			FFieldsByStartPos.push_back(t_fld);

			FFieldsByNumber[t_fld->GetFieldNumber()] = t_fld;

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);
	}
	catch(std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLayout","create_layout_fields", e.what());
		std::string t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TLayout","create_layout_fields", t_error);
		FEngineErrorsPump->AddError("TLayout","create_layout_fields", t_buffer);
		throw std::runtime_error("Failure to create layout field objects");
	}

} // create_layout_fields
//---------------------------------------------------------------------------
int TLayout::LoadLayout(const std::string layout_tag)
{
	int t_result = 0;
	// Find the layout in the metafile

	std::unique_ptr<TFastSQLQuery> t_qry(new TFastSQLQuery(FSQLDB));

	char t_buffer[500];
	std::string t_sql =
		"SELECT LayoutName, PKey FROM EE_Layouts WHERE LayoutTag = '%s'";

	std::string t_layout_tag = stringReplace(layout_tag, "'", "''");
	int t_count =
		std::snprintf(t_buffer, 500, t_sql.c_str(), t_layout_tag.c_str());
	if (t_count < 500)
	{
		t_qry->SetIsActive(false);
		t_qry->SetSQLText(t_buffer);
		try
		{
			t_qry->SetIsActive(true);

			FLayoutName = t_qry->GetFieldByNumber(0);
			FPKey = t_qry->GetIntFieldByNumber(1);
			FLayoutTag = layout_tag;

			t_qry->SetIsActive(false);

		}
		catch (std::runtime_error& e)
		{
			FEngineErrorsPump->AddError("TLayout","LoadLayout", e.what());
			std::string t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
			FEngineErrorsPump->AddError("TLayout","LoadLayout", t_error);
			FEngineErrorsPump->AddError("TLayout","LoadLayout", t_buffer);

			t_qry->SetIsActive(false);

			t_result = -1;
		}
	}
	else
	{
		std::string t_error = "Layout Tag is much too long! (";
		t_error += layout_tag;
		t_error +  ")\nWhen used in this SQL the total exceeds 500 bytes:\n";
		t_error += t_sql;
		FEngineErrorsPump->AddError("TLayout","LoadLayout", t_error);
		t_result = -1;
	}

	create_layout_fields();

	return t_result;
} // LoadLayout
//---------------------------------------------------------------------------
int TLayout::DisplayFieldInMessagesAs( const int field_number,
	const std::string display_name )
{
	int t_result = 0;

	TLayoutField* t_fld = FFieldsByNumber[ field_number ];
	if (t_fld)
	{

		t_fld->SetDisplayName(display_name.c_str());
	}
	else
	{
		// Actually, just because a field is not found in a particular
		// layout does not mean we have an error condition. So we'll just
		// let this pass silently.
	}
	return t_result;

} // DisplayFieldInMessagesAs
//---------------------------------------------------------------------------
int TLayout::NeverSkipEmptyField( const int field_number,
	const int never_skip_empty )
{
	int t_result = 0;

	TLayoutField* t_fld = FFieldsByNumber[ field_number ];
	if (t_fld)
	{
		bool t_never_skip_empty = (never_skip_empty != 0);

		t_fld->SetNeverSkipEmpty(t_never_skip_empty);
	}
	else
	{
		// Actually, just because a field is not found in a particular
		// layout does not mean we have an error condition. So we'll just
		// let this pass silently.
	}
	return t_result;

} // NeverSkipEmptyField
//---------------------------------------------------------------------------
int TLayout::LoadData(const std::string data)
{
	size_t t_data_len = data.length();
	if (t_data_len > FDataLength || !FData)
	{
		if (FData)
		{
			std::free(FData);
			FData = NULL;
		}

		FDataLength = t_data_len;

		FData = (char*)std::malloc(t_data_len +1);
	}

	std::memset(FData, ' ', FDataLength);

	std::strncpy(FData, data.c_str(), t_data_len);
	FData[FDataLength] = '\x0';

	TLayoutField* t_fld;

	size_t t_count = FFieldsByStartPos.size();

	for (size_t t_idx = 0; t_idx < t_count; t_idx++)
	{
		t_fld = FFieldsByStartPos[t_idx];

		if (t_fld)
		{
			// Clear the FFailedSingleFieldEdit flag
			// If the Single-field edit fails for this case,
			// the flag will be reset to true
			t_fld->SetFailedSingleFieldEdit(false);

		} // if (t_fld)
	} // for...
	return 0;
} // LoadData
//---------------------------------------------------------------------------
char* TLayout::FillFieldValue(const int start_pos, const int field_length)
{
	// Field values could really get us... if someday some fool decides
	// to edit a memo field (we already have 2K fields in the layout),
	// it may be necessary to increase the size of LayoutFieldBuffer.
	if (field_length >= (int)LayoutFieldBuffSize)
	{
		// Make the buffer the field_length plus a cushion
		// Note that this resizing will remain for the lifetime
		// of the Layout being in memory
		SetupLayoutFieldBuffer(field_length + 500);
	}

	if ((start_pos-1) + field_length <= (int)FDataLength)
	{
		std::strncpy(LayoutFieldBuffer,	&FData[start_pos-1], field_length);
		LayoutFieldBuffer[field_length] = '\x0';
	}
	else
	{
		LayoutFieldBuffer[0] = '\x0';
	}


	return LayoutFieldBuffer;
} // FillFieldValue
//---------------------------------------------------------------------------
int TLayout::DisplayOneField(int field_number, int field_macro_order,
	void* owner, void* callback_func)
{
	// The field_macro_order parameter is the order the edit author
	// established for the field within the context of the edit under
	// consideration. Taken from EE_EditFieldsPivot.FieldOrder.
	int t_result = 0;

	CallbackDisplayDataRecord t_callback_func =
		(CallbackDisplayDataRecord)callback_func;


	int t_pkey,
		t_start_pos,
		t_field_length;

	std::string t_field_name,
				t_display_name,
				t_data_type,
				t_value,
				t_trim_value;

	TLayoutField* t_fld = GetFieldByFieldNumber(field_number);
	if (t_fld)
	{
		t_pkey = t_fld->GetPKey();
		t_start_pos = t_fld->GetStartPos();
		t_field_length = t_fld->GetFieldLength();

		t_field_name.assign(t_fld->GetFieldName());
		t_display_name.assign(t_fld->GetDisplayName());
		t_data_type.assign(t_fld->GetDataType());
		t_value.assign(t_fld->GetValue());

		t_trim_value = trim(t_value);
		if (t_trim_value.empty())
		{
			t_value.assign("<BLANK>");
		}

		// Send it all to the callback function
		(*t_callback_func)(owner, t_pkey, t_start_pos, field_number, field_macro_order,
			t_field_name.c_str(), t_display_name.c_str(), t_field_length,
			t_data_type.c_str(), t_value.c_str());

	}
	else
	{
		char t_error[100];
		std::snprintf(t_error, 100, "Field number %d not in Layout with tag %s",
			field_number, FLayoutTag.c_str());
		FEngineErrorsPump->AddError("TLayout","DisplayOneField", t_error);
		t_result = -1;
	}

	return t_result;
} // DisplayOneField
//---------------------------------------------------------------------------
int TLayout::DisplayDataRecord( void* owner, void* callback_func )
{
	int t_result = 0;

	CallbackDisplayDataRecord t_callback_func =
		(CallbackDisplayDataRecord)callback_func;

	int t_pkey,
		t_start_pos,
		t_field_number,
		t_field_length;

	// Field macro order is the value in EE_EditFieldsPivot.FieldOrder, which
	// is inapplicable in this context. But we're sharing a function template
	// (CallbackDisplayDataRecord) with the run-edits function,
	// TLayout::DisplayOneField, so we need a dummy to be a placeholder.
	// Programmers calling this function should ignore the field_macro_order
	// parameter for this context.
	int t_field_macro_order = 0;

	std::string t_field_name,
				t_display_name,
				t_data_type,
				t_value,
				t_trim_value;

	TLayoutField* t_fld;
	size_t t_count = FFieldsByStartPos.size();

	for (size_t t_idx = 0; t_idx < t_count; t_idx++)
	{
		t_fld = FFieldsByStartPos[t_idx];

		t_pkey = t_fld->GetPKey();
		t_start_pos = t_fld->GetStartPos();
		t_field_number = t_fld->GetFieldNumber();
		t_field_length = t_fld->GetFieldLength();

		t_field_name.assign(t_fld->GetFieldName());
		t_display_name.assign(t_fld->GetDisplayName());
		t_data_type.assign(t_fld->GetDataType());
		t_value.assign(t_fld->GetValue());

		t_trim_value = trim(t_value);
		if (t_trim_value.empty())
		{
			t_value.assign("<BLANK>");
		}

		// Send it all to the callback function
		(*t_callback_func)(owner, t_pkey, t_start_pos, t_field_number, t_field_macro_order,
			t_field_name.c_str(), t_display_name.c_str(), t_field_length,
			t_data_type.c_str(), t_value.c_str());
	}

	return t_result;
} // DisplayDataRecord
//---------------------------------------------------------------------------
int TLayout::FieldDataByNumber(
		 const int field_number,
		 int* pkey,
		 int* start_pos,
		 int* field_length,
		 char* field_name,
		 const int field_name_len,
		 char* display_name,
		 const int display_name_len,
		 char* data_type,
		 const int data_type_len,
		 char* value, const int value_len )
{
	int t_result = 0;

	TLayoutField* t_fld = FFieldsByNumber[ field_number ];
	if (t_fld)
	{
		*pkey = t_fld->GetPKey();
		*start_pos = t_fld->GetStartPos();
		*field_length = t_fld->GetFieldLength();

		if (field_name_len > 0)
		{
			std::snprintf(field_name, field_name_len, "%s", t_fld->GetFieldName());
		}

		if (display_name_len > 0)
		{
			std::snprintf(display_name, display_name_len, "%s", t_fld->GetDisplayName());
		}

		if (data_type_len > 0)
		{
			std::snprintf(data_type, data_type_len, "%s", t_fld->GetDataType());
		}

		if (value_len > 0)
		{
			std::snprintf(value, value_len, "%s", t_fld->GetValue());
		}

	}
	else
	{
		char t_buffer[500];
		std::snprintf(t_buffer, 500,
			"Field number %d not found in layout named \"%s\"/Tag \"%s\"",
			field_number, FLayoutName.c_str(), FLayoutTag.c_str());

		FEngineErrorsPump->AddError("TLayout","FieldDataByNumber", t_buffer);

		t_result = -1;
	}
	return t_result;
} // FieldDataByNumber
//---------------------------------------------------------------------------
TLayoutField* TLayout::GetFieldByFieldNumber( int field_number )
{
	return FFieldsByNumber[field_number];
}
//---------------------------------------------------------------------------

