//---------------------------------------------------------------------------
#include <stdexcept>
#include <cstdlib> // std::malloc, std::free
#include <memory>
#include <time.h> // for SetSystemDate
#include <stdint.h> // intptr_t
#include <cctype> // isdigit

#include "LibFuncs.h"
#include "RunEdits.h"
#include "TokenMaster.h"
#include "Variable.h"
#include "EngineErrorsPump.h"
#include "Utility.h"
#include "TableStats.h"
#include "EditErrors.h"
#include "SQLDB.h"
#include "FastSQLQuery.h"
#include "ExternalDLLTracker.h"
#include "Utility.h"

// day of year, days in month
// Used by the date handling functions
static int MonthDaysOfYear[13][2] =
{
	{   0,  0 }, // skip this... it makes referencing easier in the code
	{   1, 31 },
	{  32, 28 }, // Yes we know: leap years. Fear not, we accommodate those.
	{  60, 31 },
	{  91, 30 },
	{ 121, 31 },
	{ 152, 30 },
	{ 182, 31 },
	{ 213, 31 },
	{ 244, 30 },
	{ 274, 31 },
	{ 305, 30 },
	{ 335, 31 }
};


//---------------------------------------------------------------------------
TLookupQuery::TLookupQuery(TSQLDB* database, TTableDef* table_def)
{
	FQry = new TFastSQLQuery(database);
	FTableDef = table_def;
	FPadTo = 0;
} // ctor
//---------------------------------------------------------------------------
TLookupQuery::~TLookupQuery()
{
	FQry->ReleaseParameterizedQuery();
	delete FQry;
} // dtor
//---------------------------------------------------------------------------
TFastSQLQuery* TLookupQuery::GetQuery()
{
	return FQry;
}
//---------------------------------------------------------------------------
TTableDef* TLookupQuery::GetTableDef()
{
	return FTableDef;
}
//---------------------------------------------------------------------------
void TLookupQuery::AddColumnOrder(int column_order)
{
	// The SQLLookup and SQLRangeLookup functions may have more than one
	// column to reference. We're pushing the column orders onto the stack
	// in the order in which they appeared in the edit logic. See internal_sqllookup
	// for details.
	if (column_order > 0 && column_order <= FTableDef->GetColumnCount())
	{
		FColumnOrderList.push_back(column_order);
	}
}
//---------------------------------------------------------------------------
int TLookupQuery::GetColumnsCount()
{
	return (int)FColumnOrderList.size();
}
//---------------------------------------------------------------------------
int TLookupQuery::GetColumnOrder(int idx)
{
	if (idx >= 0 && idx < GetColumnsCount())
		return FColumnOrderList[idx];
	else
		return -1;
}
//---------------------------------------------------------------------------
void TLookupQuery::SetPadTo(int pad_size)
{
	if (pad_size > 0)
	{
		FPadTo = pad_size;
	}
}
//---------------------------------------------------------------------------
int TLookupQuery::GetPadTo()
{
	return FPadTo;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
TLibFuncs::TLibFuncs(TRunEdits* caller, TEngineErrorsPump* pump,
	TSQLDB* database, TTablesList* tables_list, TEditErrors* edit_errors_manager,
	TTokenMaster* token_master)
{
	FRunEdits = caller;
	FEngineErrorsPump = pump;
	FSQLDB = database;
	FTablesList = tables_list;
	FEditErrors = edit_errors_manager;
	FTokenMaster = token_master;

	setup_function_pointers();

} // constructor
//---------------------------------------------------------------------------
TLibFuncs::~TLibFuncs()
{
	{
		// parameterized queries created by internal_lookup
		std::map< std::string, TLookupQuery*>::iterator it;
		for (it = FInternalLookupQueries.begin();
			 it != FInternalLookupQueries.end(); it++)
		{
			TLookupQuery* t_lookup_qry = (*it).second;
			delete t_lookup_qry;
		}
	}

	{
		// parameterized queries created by internal_rlookup
		std::map< std::string, TLookupQuery*>::iterator it;
		for (it = FInternalRLookupQueries.begin();
			 it != FInternalRLookupQueries.end(); it++)
		{
			TLookupQuery* t_lookup_qry = (*it).second;
			delete t_lookup_qry;
		}
	}

	{
		// parameterized queries created by internal_sqllookup
		std::map< std::string, TLookupQuery*>::iterator it;
		for (it = FInternalSQLLookupQueries.begin();
			 it != FInternalSQLLookupQueries.end(); it++)
		{
			TLookupQuery* t_lookup_qry = (*it).second;
			delete t_lookup_qry;
		}
	}

	{
		// parameterized queries created by internal_sqlrangelookup
		std::map< std::string, TLookupQuery*>::iterator it;
		for (it = FInternalSQLRLookupQueries.begin();
			 it != FInternalSQLRLookupQueries.end(); it++)
		{
			TLookupQuery* t_lookup_qry = (*it).second;
			delete t_lookup_qry;
		}
	}

} // destructor
//---------------------------------------------------------------------------
void TLibFuncs::setup_function_pointers()
{
	for (size_t t_idx = 0; t_idx < 256; t_idx++)
	{
		FFuncPointers[t_idx] = NULL;
	}

	// Compiled logic references EDITS Language Functions in this manner:
	// Edit logic:      strcpy(str1, "002");
	// Compiled logic:  Mo(N04str1,"03002");
	// The token following the 'M' is used to find the function pointer in
	// this list.

	// Note that by subtracting 65 from the value of the token we can instantly
	// find the index into array FFuncPointers holding the function address.
	// The arithmetic works like this: If the int value of the char token is
	// greater than 0, subtract 65 from the token to find the slot in the array
	// that will hold the pointer. But if the int value of the char token is
	// less than 0, it's an "Alt+" character. If you subtract token from 256,
	// you'll find the appropriate spot in the array. Of course, since the token
	// value is less than 0, you get there by adding it to 256. Oh, what fun we have!

																  // 65+
	FFuncPointers['A' - 65] = &TLibFuncs::call_at;                //  0, 'A'
	FFuncPointers['K' - 65] = &TLibFuncs::call_empty;             // 10, 'K'
	FFuncPointers['L' - 65] = &TLibFuncs::call_error_msg;         // 11, 'L'
	FFuncPointers['M' - 65] = &TLibFuncs::call_error_text;        // 12, 'M'
	FFuncPointers['N' - 65] = &TLibFuncs::call_fmtstr;            // 13, 'N'
	FFuncPointers['S' - 65] = &TLibFuncs::call_ilookup;           // 18, 'S'
	FFuncPointers['U' - 65] = &TLibFuncs::call_inlist;            // 20, 'U'
	FFuncPointers['V' - 65] = &TLibFuncs::call_justified;         // 21, 'V'
	FFuncPointers['W' - 65] = &TLibFuncs::call_left;              // 22, 'W'
	FFuncPointers['X' - 65] = &TLibFuncs::call_lookup;            // 23, 'X'
	FFuncPointers['Y' - 65] = &TLibFuncs::call_lower;             // 24, 'Y'
	FFuncPointers['Z' - 65] = &TLibFuncs::call_match;             // 25, 'Z'
	FFuncPointers[']' - 65] = &TLibFuncs::call_nameexpr;          // 28  '['
	FFuncPointers['e' - 65] = &TLibFuncs::call_right;             // 36, 'e'
	FFuncPointers['f' - 65] = &TLibFuncs::call_rlookup;           // 37, 'f'
	FFuncPointers['g' - 65] = &TLibFuncs::call_save_error_text;   // 38, 'g'
	FFuncPointers['h' - 65] = &TLibFuncs::call_save_text;         // 39, 'h'
	FFuncPointers['m' - 65] = &TLibFuncs::call_strcat;            // 44, 'm'
	FFuncPointers['n' - 65] = &TLibFuncs::call_strcmp;            // 45, 'n'
	FFuncPointers['o' - 65] = &TLibFuncs::call_strcpy;            // 46, 'o'
	FFuncPointers['p' - 65] = &TLibFuncs::call_strlen;            // 47, 'p'
	FFuncPointers['q' - 65] = &TLibFuncs::call_substr;            // 48, 'q'
	FFuncPointers['s' - 65] = &TLibFuncs::call_trim;              // 50, 's'
	FFuncPointers['t' - 65] = &TLibFuncs::call_upper;             // 51, 't'
	FFuncPointers['{' - 65] = &TLibFuncs::call_sqllookup;         // 123, '{'
	FFuncPointers['}' - 65] = &TLibFuncs::call_sqlrangelookup;     // 125, '}'
	FFuncPointers[256 + ''] = &TLibFuncs::call_save_warning_text; // 68, ''
	FFuncPointers[256 + ''] = &TLibFuncs::call_set_warning;       // 67, ''
	FFuncPointers[256 + ''] = &TLibFuncs::call_externaldll;       // ???, ''
	FFuncPointers[256 + ''] = &TLibFuncs::call_val;               // ???, ''
	FFuncPointers[256 + ''] = &TLibFuncs::call_allow_future_date_iop; // ???, ''
	FFuncPointers[256 + ''] = &TLibFuncs::call_date_day_iop;      // 136, '' (Alt+0136)
	FFuncPointers[256 + ''] = &TLibFuncs::call_date_month_iop;    // 137, '' (Alt+0137)
	FFuncPointers[256 + ''] = &TLibFuncs::call_date_year_iop;     // 138, '' (Alt+0138)
	FFuncPointers[256 + ''] = &TLibFuncs::call_datecmp_iop;       // 139, ''  (Alt+0139)
	FFuncPointers[256 + ''] = &TLibFuncs::call_daydiff_iop;       // 140, '' (Alt+0140)
	FFuncPointers[256 + ''] = &TLibFuncs::call_monthdiff_iop;     // 77, ''
	FFuncPointers[256 + ''] = &TLibFuncs::call_dayinterval_iop;   // 141, '' (Alt+0141)
	FFuncPointers[256 + ''] = &TLibFuncs::call_monthinterval_iop; // 143, '' (Alt+0143)
	FFuncPointers[256 + ''] = &TLibFuncs::call_valid_date_iop;    // 144, '' (Alt+0144)
	FFuncPointers[256 + ''] = &TLibFuncs::call_yeardiff_iop;      // 145, ''  (Alt+0145)
	FFuncPointers[256 + ''] = &TLibFuncs::call_yearinterval_iop;  // 146, ''  (Alt+0146)
/*
	// We are removing these from the language. Susan has long since
	// eliminated them from edit logic, so they should not exist
	// anywhere in anybody's metafile.
	FFuncPointers[''] = &TLibFuncs::call_copyrecord;        // 69, ''
	FFuncPointers['B'] = &TLibFuncs::call_binlookup;          1, 'B'
	FFuncPointers['C'] = &TLibFuncs::call_check_flag;        2, 'C'
	FFuncPointers['D'] = &TLibFuncs::call_clear_all_flags;   3, 'D'
	FFuncPointers['E'] = &TLibFuncs::call_date_day;          //  4, 'E'
	FFuncPointers['F'] = &TLibFuncs::call_date_month;        //  5, 'F'
	FFuncPointers['G'] = &TLibFuncs::call_date_year;         //  6, 'G'
	FFuncPointers['H'] = &TLibFuncs::call_datecmp;           //  7, 'H'
	FFuncPointers['I'] = &TLibFuncs::call_daydiff;           //  8, 'I'
	FFuncPointers['J'] = &TLibFuncs::call_dayinterval;       //  9, 'J'
	FFuncPointers['P'] = &TLibFuncs::call_getfield;          // 15, 'P'
	FFuncPointers['^'] = &TLibFuncs::call_nameexpr;          // 28, '^'
	FFuncPointers['`'] = &TLibFuncs::call_putfield;          // 31, '`'
	FFuncPointers['i'] = &TLibFuncs::call_set_error;         // 40, 'i'
	FFuncPointers['O'] = &TLibFuncs::call_getche;            14, 'O'
	FFuncPointers['Q'] = &TLibFuncs::call_getnum;            16, 'Q'
	FFuncPointers['T'] = &TLibFuncs::call_increment_flag;    19, 'T'
	FFuncPointers['^'] = &TLibFuncs::call_print;             29, '^'
	FFuncPointers['_'] = &TLibFuncs::call_putch;             30, '_'
	FFuncPointers['a'] = &TLibFuncs::call_puts;              32, 'a'
	FFuncPointers['b'] = &TLibFuncs::call_quit;              33, 'b'
	FFuncPointers['c'] = &TLibFuncs::call_read;              34, 'c'
	FFuncPointers['d'] = &TLibFuncs::call_recordempty;       35, 'd'
	FFuncPointers['j'] = &TLibFuncs::call_set_flag;          41, 'j'
	FFuncPointers['l'] = &TLibFuncs::call_show;              43, 'l'
	FFuncPointers['r'] = &TLibFuncs::call_trace              49, 'r'
	FFuncPointers['R'] = &TLibFuncs::call_getvar;            // 17, 'R'
	FFuncPointers['k'] = &TLibFuncs::call_setvar;            // 42, 'k'
	FFuncPointers['u'] = &TLibFuncs::call_use;               52, 'u'
	FFuncPointers['v'] = &TLibFuncs::call_usr1         53, 'v'
	FFuncPointers['x'] = &TLibFuncs::call_usr2;              // 55, 'x'
	FFuncPointers['z'] = &TLibFuncs::call_usr4;              // 57, 'z'
	FFuncPointers['w'] = &TLibFuncs::call_usr10        54, 'w'
	FFuncPointers['y'] = &TLibFuncs::call_usr3         56, 'y'
	FFuncPointers['|'] = &TLibFuncs::call_usr6         59, '|',
	FFuncPointers['~'] = &TLibFuncs::call_usr8         61, '~'
	FFuncPointers[''] = &TLibFuncs::call_usr9         62, ''
	FFuncPointers[''] = &TLibFuncs::call_valid_date;        // 64, '' (Alt+0129)
	FFuncPointers['['] = &TLibFuncs::call_monthdiff;         // 26, '['
	FFuncPointers['\\'] = &TLibFuncs::call_monthinterval;     // 27, '\\'
	FFuncPointers[''] = &TLibFuncs::call_yeardiff;          // 65, ''
	FFuncPointers[''] = &TLibFuncs::call_yearinterval;      // 66, ''
 */
} // setup_function_pointers
//---------------------------------------------------------------------------
FCallFuncPtr TLibFuncs::GetFuncPointer(char token)
{
	if (token > 0)
		return FFuncPointers[token - 65];
	else
		return FFuncPointers[256 + token];
}
//---------------------------------------------------------------------------
void TLibFuncs::check_open_paren()
{
	// We should be on the opening parenthesis of the parameters list
	TToken* t_token = FTokenMaster->GetHardToken(LPAREN);

	if (!t_token)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "check_open_paren",
			"syntax error: expected opening parenthesis");
		throw std::runtime_error("");
	}
} // check_open_paren
//---------------------------------------------------------------------------

void TLibFuncs::check_close_paren()
{
	// In the rewrite of TRunEdits::EvaluateExpression, we sometimes eat closing
	// parentheses as we encounter them. Therefore, this method will
	// frequently crash. So we are changing the philosophy here: IF the next
	// token is RPAREN, eat it. Otherwise, leave it alone, because something
	// else already dealt with it. Big assumption, probably dangerous, but
	// we need to move on!
	FTokenMaster->SavePos();
	TToken* t_token = FTokenMaster->GetCompiledToken();
	if (t_token->GetTokenID() != RPAREN)
	{
		FTokenMaster->RestorePos();
	}
} // check_close_paren
//---------------------------------------------------------------------------

std::string TLibFuncs::translate_mask_to_regex(std::string mask)
{
	std::string t_pattern = "";
	std::string t_mask_char;
	std::string t_read_ahead;
	std::string t_token;

	int t_repeats;

	// Flag to tell the "t_token" processor at the bottom of this
	// function when NOT to encase a token in surrounding square
	// braces. That is, when to do this: [[:digit::] instead of
	// this: [:digit:]. Turned on (true) when t_mask_char == "{" or "[",
	// and off (false) when t_mask_char == "}" or "]".
	bool t_token_in_subpattern_braces = false;
	bool t_token_in_optional_brackets = false;

	char t_buffer[20];

	size_t t_len = mask.length();

	for (size_t t_idx = 0; t_idx < t_len; t_idx++)
	{
		// Get the character at the t_idx'th position
		t_mask_char.assign(mask.substr(t_idx,1));

		// Looking for patterns that repeat, such as:
		// "dddddddd", a string of exactly nine digits
		// Those require syntax like this:
		// [[:digit:]]{9,9}
		t_token.clear();
		t_read_ahead.clear();
		t_repeats = 0;
		if (t_idx +1 < t_len)
		{
			t_read_ahead.assign(mask.substr(t_idx+1,1));
			t_repeats = 1;
		}
		// Run the gauntlet to see what we have
		if (t_mask_char.compare("x") == 0)
		{
			// Specifies any alphanumeric character
			t_token.assign("[:alnum:]");
		}
		else if (t_mask_char.compare("b") == 0)
		{
			// Specifies a blank
			t_token.assign( "[:blank:]" );
		}
		else if (t_mask_char.compare("@") == 0)
		{
			// Specifies any character EXCEPT blank

			if (t_token_in_optional_brackets)
			{
				t_token.assign( "[^[:blank:]]" );
			}
			else if (t_token_in_subpattern_braces)
			{
				std::string t_bad_token = "Syntax error in MATCH mask: \"" + mask + "\": ";
				t_bad_token += "\"Not-blank\" character \"@\" inside sub-pattern \"{}\" does not process properly. ";
				t_bad_token += "Try \"[@]\" or \"[@]*\" instead.";
				throw std::runtime_error(t_bad_token);
			}
			else
			{

				t_token.assign( "^[:blank:]" );
			}
		}
		else if (t_mask_char.compare("a") == 0)
		{
			// Specifies any alphabetic character
			t_token.assign( "[:alpha:]" );
		}
		else if (t_mask_char.compare("u") == 0)
		{
			// Specifies any uppercase alphabetic character
			t_token.assign( "[:upper:]" );
		}
		else if (t_mask_char.compare("l") == 0)
		{
			// Specifies any lowercase alphabetic character
			t_token.assign( "[:lower:]" );
		}
		else if (t_mask_char.compare("d") == 0)
		{
			// Specifies a digit 0-9
			t_token.assign( "[:digit:]" );
		}
		else
		{
			if (t_mask_char.compare("?") == 0)
			{
				// Specifies any character
				if (t_token_in_subpattern_braces)
				{
					std::string t_prev_char = mask.substr(t_idx-1,1);
					if ((t_prev_char.compare(",") == 0) ||
						(t_read_ahead.compare(",") == 0))
					{
						std::string t_bad_token = "Syntax error in MATCH mask: \"" + mask;
						t_bad_token += "\". The \"any\" character \"?\"";
						t_bad_token += " inside sub-pattern \"{}\" cannot be combined with other mask characters. ";
						t_bad_token += "Try optional sub-pattern syntax like \"[?,a,b,d]\" instead. ";
						t_bad_token += "(Reminder: The \"?\" means ANYTHING. All other tokens are redundant.)";
						throw std::runtime_error(t_bad_token);
					}
				}
				t_token.assign(".");
			}
			else if (t_mask_char.compare("!") == 0)
			{
				// Specifies that following character is to be
				// treated as a literal, and not as a mask token
				if (t_idx +1 < t_len)
				{
					t_read_ahead.assign( mask.substr(t_idx+1,1) );
					if ((t_read_ahead.compare("+") == 0) ||
						(t_read_ahead.compare("-") == 0) ||
						(t_read_ahead.compare("{") == 0) ||
						(t_read_ahead.compare("?") == 0))
					{
						// Most tokens do not need to be escaped, and
						// using the escape indiscriminately can run
						// afoul of regex shorthand... e.g., escaping
						// "d" just because the mask author said to
						// would yield "\d", which is the token for digit.
						t_pattern += "\\";
						t_idx++;
						t_mask_char.assign(mask.substr(t_idx,1));
						t_pattern += t_mask_char;
					}
					else
					{
						t_pattern += t_read_ahead;
						t_idx++;
					}
				}
			}
			else if (t_mask_char.compare(":") == 0)
			{
				if (t_token_in_subpattern_braces)
				{
					// If we're inside {}, just change the ":" to "-"
					t_pattern += "-";
				}
				else
				{
					// Specifies a range for a character position.
					// E.g., 1:5 allows only values from 1 to 5.
					// The range operator applies only to a single
					// character position.
					// For this, we need to back up to just prior to last character
					// written to t_pattern and insert "[", that character,
					// this hyphen, the next character and the closing "]"
					// Re-using t_token and t_read_ahead for this purpose...
					t_read_ahead.assign(t_pattern.substr(t_pattern.length()-1,1));
					t_token.assign(t_pattern.substr(0, t_pattern.length()-1));
					t_pattern.assign(t_token);
					t_pattern += "[" + t_read_ahead + "-";
					if (t_idx+1 < t_len)
					{
						t_read_ahead.assign(mask.substr(t_idx+1,1));
						t_pattern += t_read_ahead + "]";
					}
					t_idx++;
					// Clear t_token and t_read_ahead so they don't get used below.
					t_token.clear();
					t_read_ahead.clear();
				}
			}

			else if (t_mask_char.compare("{") == 0)
			{
				// NOTE: EDITS MASK uses curly braces {} the
				// way std::regex uses parentheses ()

				// {} encloses a subpattern. Can be followed by
				// repetition characters and control strings.
				// Without repetition or control strings, pattern
				// within braces must be met only once. (See "*")
				t_pattern += "[(";
				t_token_in_subpattern_braces = true;
			}
			else if (t_mask_char.compare("}") == 0)
			{
				// NOTE: Again, EDITS MASK uses curly braces {} the
				// way std::regex uses parentheses().
				t_pattern += ")]";
				/*
				if (t_read_ahead.compare("*") == 0)
				{
					t_pattern += "]";
				}
				*/
				t_token_in_subpattern_braces = false;
			}
			else if (t_mask_char.compare("[") == 0)
			{
				// Encloses an optional sub-pattern, meaning the
				// match is valid with or without the enclosed
				// pattern.
				t_pattern += "(";
				t_token_in_optional_brackets = true;
			}
			else if (t_mask_char.compare("]") == 0)
			{
				// Close optional sub-pattern
				t_pattern += ")";
				t_token_in_optional_brackets = false;
			}
			else if (t_mask_char.compare(",") == 0)
			{
				// Not documented, but a comma is used in EDITS MATCH mask
				// to separate optional characters. std::regex uses
				// a bitwise OR
				t_pattern += "|";
			}
			else if (t_mask_char.compare("-") == 0)
			{
				// hyphen must be escaped
				t_pattern += "\\" + t_mask_char;
			}
			else
			{
				// All other characters (including "*") just
				// transfer to t_pattern.
				t_pattern += t_mask_char;
			}
		}

		// Depending upon where the t_token appears in the mask,
		// we may need to decorate it a little.
		if (!t_token.empty())
		{
			if (t_mask_char.compare(t_read_ahead) == 0)
			{
				// The next character is the same as this character;
				// that is, we have a repeating pattern. Count up all
				// of the repeats and report it like this:
				// match mask: "dddd"
				// regex pattern: "[[:digit:]]{4,4}"
				t_pattern += "["+t_token+"]";
				while ((t_idx < t_len) &&
					   (t_mask_char.compare(t_read_ahead) == 0))
				{
					t_repeats++;
					t_idx++;
					t_read_ahead.assign(mask.substr(t_idx+1,1));
				}
				std::snprintf(t_buffer, 20, "{%d,%d}", t_repeats, t_repeats);
				t_pattern += (std::string)t_buffer;
			}
			else
			{
				if (t_token_in_subpattern_braces || t_token_in_optional_brackets)
				{
					// This token is inside a pair of braces {} or []
					// Don't decorate it!
					t_pattern += t_token;
				}
				else
				{
					// This token is NOT inside {} or []
					// Surround it with square braces
					t_pattern += "["+t_token+"]";
				}
			}
		 }
	}  // for (t_idx...

	// This is ghastly, but... the above would construct this
	// mask: "@{?}*" as pattern "[^[:blank:]{1,1}][(.)]*", and
	// while THIS would be valid: "[^[:blank:]{1,1}][([:alnum:])]*",
	// subexpression "[(.)]" is not. We need to remove the outer
	// square braces for this, so... we'll stringreplace it.

	if (t_pattern.find("[(.)]") != std::string::npos)
	{
		t_pattern = stringReplace(t_pattern, "[(.)]", "(.)");
	}

	return t_pattern;
} // translate_mask_to_regex
//---------------------------------------------------------------------------


//////////////////////////////////////////////////////////////////////////////
///            SUPPORT FOR EDITS LANGUAGE DATE FUNCTIONS
//////////////////////////////////////////////////////////////////////////////
int TLibFuncs::GetDateErrorCode()
{
	// This is called from TRunEdits::debug_showvars
	return FDateErrorCode;
}
//---------------------------------------------------------------------------
void TLibFuncs::ClearDateErrorCode()
{
	// This is called from TRunEdits::ProcessBlock right before a LIBFUNC call.
	// It means the function is not a date-handling function. The date-handling
	// functions will reset the value to DT_VALID or one of the other values.
	FDateErrorCode = 0;
}
//---------------------------------------------------------------------------
void TLibFuncs::SetSystemDate()
{
	// Set by TRunEdits::RunEdits and TRunEdits::DebugEdit
	// to make sure that the system date referenced by the date
	// functions is current.
	time_t	t_timer;
	struct tm *t_tblock;

	t_timer = time(&t_timer);
	t_tblock = localtime(&t_timer);
	FSystemDate = (t_tblock->tm_year+1900L)*10000 +
				  (t_tblock->tm_mon+1)*100 + t_tblock->tm_mday;
}
//---------------------------------------------------------------------------
int TLibFuncs::GetSystemDate()
{
	// TRunEdits::eval7_atom needs to be able to assign this
	return FSystemDate;
}
//---------------------------------------------------------------------------
void TLibFuncs::SetYearsIntoFuture()
{
	// Reset to 0 at the start of each edit (TRunEdits::ProcessEdit)
	// This may be given a non-zero value by a call to
	// ALLOW_FUTURE_DATE_IOP.
	FYearsIntoFuture = 0;
}
//---------------------------------------------------------------------------
int TLibFuncs::date_is_valid(const char* the_date)
{
	int t_year, t_month, t_day;
	int t_high_year, t_high_month, t_high_day;

	if (!check_date_format(the_date))
	{
		return 0; // false
	}
	else if (FDateErrorCode == DT_EMPTY)
	{
		return 1L; // true
	}
	else
	{
		t_year = get_int_val(the_date, 0, 4);
		t_month = get_int_val(the_date,4, 2);
		t_day = get_int_val(the_date,6,2);

		// Range checking: Our lowest allowable date is January 1, 1850,
		// and highest allowable date is today

		// Get today's date and extract the year, month and day portions
		t_high_year = (int)(FSystemDate / 10000L);
		t_high_month = (int)((FSystemDate / 100) - t_high_year * 100);
		t_high_day   = (int)(FSystemDate - (t_high_month * 100 + t_high_year * 10000));

		// It's zero unless the edit invoked ALLOW_FUTURE_DATE_IOP
		if (FYearsIntoFuture > 0)
		{
			t_high_year += (int)FYearsIntoFuture;
		}

		// Evaluate the date parts
		if (! ((t_year >= 1850) && (t_year <= t_high_year)))
		{
			FDateErrorCode = DT_ERROR_YR;
			return 0;
		}
		else if (!( (t_month >= 0) && (t_month <= 12) ) )
		{
			// Note: Month and Day may be 0 if they were left blank
			// (that is, only the Year portion of the date is provided)
			FDateErrorCode = DT_ERROR_MO;
			return 0;
		}
		else if (! (t_day <= 31) )
		{
			FDateErrorCode = DT_ERROR_DA;
			return 0;
		}

		// Still good? Then check MonthDaysOfYear array (see top of this file)
		// to make sure the day-of-month does not exceed allowable range.
		int t_max_days = MonthDaysOfYear[t_month][1];
		if (t_day > t_max_days)
		{
			// For for February 29 in non-leap years
			if (t_day == 29 && t_month == 2)
			{
				if (!is_leap_year(t_year))
				{
					// That's right... it's a DAY failure,
					// despite all of this analysis of the year.
					FDateErrorCode = DT_ERROR_DA;
					return 0;
				}
			}
			else
			{
				// day exceeds max days and it isn't February
				FDateErrorCode = DT_ERROR_DA;
				return 0;
			}
		} // if (t_day > t_max_days)

		// We checked t_high_year, but failed to make sure
		// t_year + t_month + day are ALL LESS than their t_high_*
		// counterparts.
		int t_high_date = (t_high_year * 10000) +
						   (t_high_month * 100) +
						   (t_high_day);

		int t_chk_date = (t_year * 10000) +
						  (t_month * 100) +
						  (t_day);

		if (t_chk_date > t_high_date)
		{
			if (t_year > t_high_year)
				FDateErrorCode = DT_ERROR_YR;
			else if (t_month > t_high_month)
				FDateErrorCode = DT_ERROR_MO;
			else
				FDateErrorCode = DT_ERROR_DA;

			return 0;

		}

		// If we make it to here, we have a valid date. However, it may have
		// some missing parts (month or day), so let's check for that.
		// Functions such as call_datecmp_iop will need this information to
		// properly handle minmax flags.
		if (t_month == 0)
		{
			FDateErrorCode = DT_MISSING_MO;
		}
		else if (t_day == 0)
		{
			FDateErrorCode = DT_MISSING_DA;
		}

	} // if (check_date_format(the_date))

	return true;
} // date_is_valid
//---------------------------------------------------------------------------
bool TLibFuncs::check_date_format(const char* the_date)
{
	FDateErrorCode = DT_VALID; // assume all will be well

	// Dates must be 8-characters long, consisting of digits and spaces
	if (std::strlen(the_date) != 8)
	{
		// Warning: EditWriter 4 user documentation states explicitly
		// that valid_date_iop does NOT set this value, but I can't think
		// why we would want that.
		FDateErrorCode = DT_ERROR;
	}
	else
	{
		// Dates are processed in D1 format: ccyymmdd, where any blank
		// values will be at the rightmost position.
		size_t t_first_blank_pos = 8;
		for (size_t t_idx = 0; t_idx < t_first_blank_pos; t_idx++)
		{
			if (the_date[t_idx] == ' ')
			{
				t_first_blank_pos = t_idx;
				break;
			}
		}
		// Possible results:
		//  * We have no blanks; check to be sure we have all digits
		//  * We have some blanks; check to be sure they occur at the
		//    month or day positions and run to the end of the_date
		//  * We have ALL blanks; check to be sure that is, in fact, true!
		if (t_first_blank_pos == 8)
		{
			// No blanks... make sure we have all digits
			for (size_t t_idx = 0; t_idx < 8; t_idx++)
			{
				if (!std::isdigit(the_date[t_idx]))
				{
					// Note the location of the first bad digit
					if (t_idx < 4)
						FDateErrorCode = DT_ERROR_YR;
					else if (t_idx < 6)
						FDateErrorCode = DT_ERROR_MO;
					else
						FDateErrorCode = DT_ERROR_DA;

					break;
				} // if (!isdigit...)
			}
		}
		else
		{
			// At least one blank... first make sure it's on one of
			// of the legitimate boundaries for year, month, day
			if ((t_first_blank_pos > 0) && (t_first_blank_pos < 4))
				FDateErrorCode = DT_ERROR_YR;
			else if (t_first_blank_pos == 5)
				FDateErrorCode = DT_ERROR_MO;
			else if (t_first_blank_pos == 7)
				FDateErrorCode = DT_ERROR_DA;
			else
			{
				// Okay, now make sure that everything from the first
				// blank to the end of the_date is blank.
				for (size_t t_idx = t_first_blank_pos+1;
					 t_idx < 8; t_idx++)
				{
					if (the_date[t_idx] != ' ')
					{
						// Note the location of the first non-blank
						if (t_idx < 4)
							FDateErrorCode = DT_ERROR_YR;
						else if (t_idx < 6)
							FDateErrorCode = DT_ERROR_MO;
						else
							FDateErrorCode = DT_ERROR_DA;

						break;
					}
				}
				// This is a little tricky... if we are still DT_VALID
				// at this point, but we know the t_first_blank_pos
				// was 0, then we have an empty string. Tell the caller.
				if ((FDateErrorCode == DT_VALID) &&
					(t_first_blank_pos == 0))
				{
					FDateErrorCode = DT_EMPTY;
				}
			}

		}
	}

	return ((FDateErrorCode == DT_VALID) || (FDateErrorCode == DT_EMPTY));
} // check_date_format
//---------------------------------------------------------------------------
int TLibFuncs::calculate_daydiff(const char* start_date,
	const char* end_date, intptr_t min_max_flag)
{
	// Called from call_daydiff_iop and call_dayinterval_iop
	// calculate_monthdiff, calculate_yeardiff
	int t_result = 0;
	try
	{
		int t_date_integrity =
			check_date_integrity(start_date, min_max_flag);
		if (t_date_integrity != DT_VALID)
		{
			t_result = t_date_integrity;
		}
		else
		{
			t_date_integrity = check_date_integrity(end_date, min_max_flag);
			if (t_date_integrity != DT_VALID)
			{
				t_result = t_date_integrity;
			}
		}

		if (t_result == 0)
		{
			// If we make it to here, we have valid
			// date objects. Now perform the evaluation
			int t_start_date, t_end_date,
				t_start_day_of_year, t_end_day_of_year,
				t_start_year, t_end_year;

			int t_ord_flag;

			t_start_date = complete_date(start_date, min_max_flag, end_date, true);

			t_end_date = complete_date(end_date, min_max_flag, start_date, false);

			if (t_start_date <= t_end_date)
			{
				t_ord_flag = 0;

				t_start_day_of_year = get_day_of_year(t_start_date);
				t_start_year = (int)(t_start_date / 10000);

				t_end_day_of_year = get_day_of_year(t_end_date);
				t_end_year = (int)(t_end_date / 10000);

			}
			else
			{
				// If the begin date is greater than the end date, flip 'em
				// for the rest of this evaluation
				t_ord_flag = 1;

				t_start_day_of_year = get_day_of_year(t_end_date);
				t_start_year = (int)(t_end_date / 10000);

				t_end_day_of_year = get_day_of_year(t_start_date);
				t_end_year = (int)(t_start_date / 10000);
			}

			// Calculate the end date day
			while (t_start_year < t_end_year)
			{
				t_end_day_of_year += 365;
				if (is_leap_year(t_start_year))
					t_end_day_of_year++;

				t_start_year++;
			}

			if (t_ord_flag == 0)
			{
				t_result = (t_end_day_of_year - t_start_day_of_year);
			}
			else
			{
				t_result = (t_start_day_of_year - t_end_day_of_year);
			}

		 } // if (t_result == 0)
	}
	catch (...)
	{
		std::string t_err_msg;
		t_err_msg.assign("failure in call to calculate_daydiff");
		FEngineErrorsPump->AddError("TLibFuncs", "calculate_daydiff", t_err_msg);
		throw std::runtime_error("");
	}

	return t_result;

} // calculate_daydiff
//---------------------------------------------------------------------------
int TLibFuncs::calculate_monthdiff(const char* start_date,
	const char* end_date, intptr_t min_max_flag)
{
	// Called from call_monthdiff_iop and call_monthinterval_iop
	int t_result = 0;
	try
	{
		int t_days_between = calculate_daydiff(start_date, end_date, min_max_flag);
		if ((t_days_between == DT_ERROR) ||
			(t_days_between == DT_EMPTY) ||
			(t_days_between == DT_UNKNOWN))
		{
			t_result = t_days_between;
		}
		else
		{
			// If t_days_between is a number, divide it into 30-day months
			t_result = ( t_days_between / 30L );
		}
	}
	catch (...)
	{
		std::string t_err_msg;
		t_err_msg.assign("failure in call to calculate_monthdiff");
		FEngineErrorsPump->AddError("TLibFuncs", "calculate_monthdiff", t_err_msg);
		throw std::runtime_error("");
	}

	return t_result;

} // calculate_monthdiff
//---------------------------------------------------------------------------
int TLibFuncs::calculate_yeardiff(const char* start_date,
	const char* end_date, intptr_t min_max_flag)
{
	// Called from call_yeardiff_iop and call_yearinterval_iop
	int t_result = 0;
	try
	{
		int t_days_between = calculate_daydiff(start_date, end_date, min_max_flag);
		if ((t_days_between == DT_ERROR) ||
			(t_days_between == DT_EMPTY) ||
			(t_days_between == DT_UNKNOWN))
		{
			t_result = t_days_between;
		}
		else
		{
			// If t_days_between is a number, divide it into 365-day years
			t_result = ( t_days_between / 365L );
		}
	}
	catch (...)
	{
		std::string t_err_msg;
		t_err_msg.assign("failure in call to calculate_yeardiff");
		FEngineErrorsPump->AddError("TLibFuncs", "calculate_yeardiff", t_err_msg);
		throw std::runtime_error("");
	}

	return t_result;

} // calculate_yeardiff
//---------------------------------------------------------------------------
int TLibFuncs::check_date_integrity(const char* the_date,
	intptr_t min_max_flag)
{
	// check_date_integrity is used by EVERY function that takes a MinMaxFlag parameter
	// call_datecmp_iop
	// call_daydiff_iop, call_monthdiff_iop, call_yeardiff_iop
	// call_dayinterval_iop, call_monthinterval_iop, call_yearinterval_iop

	int t_result = 0;

	int t_date_is_valid = date_is_valid(the_date);

	if (t_date_is_valid)
	{
		if (FDateErrorCode == DT_EMPTY)
		{
			t_result = DT_EMPTY;
		}
		else
		{
			if (min_max_flag == DT_EXACT)
			{
				// If the date has any missing parts, the result
				// is UNKNOWN
				if ((FDateErrorCode == DT_MISSING_MO) ||
					(FDateErrorCode == DT_MISSING_DA))
				{
					t_result = DT_UNKNOWN;
				}
			}
		}

		if (t_result == 0)
		{
			t_result = DT_VALID;
		}
	}
	else
	{
		t_result = DT_ERROR;
	}

	return t_result;

} // check_date_integrity
//---------------------------------------------------------------------------
void TLibFuncs::parse_date(const char* the_date, int& the_year,
	int& the_month, int& the_day)
{
	// Turns a string representing a date (CCYYMMDD format) into
	// its parts.
	the_year = get_int_val(the_date,0,4);
	the_month = get_int_val(the_date,4,2);
	the_day = get_int_val(the_date,6,2);
} // parse_date
//---------------------------------------------------------------------------
int TLibFuncs::complete_date(const char* the_date, intptr_t min_max_flag,
	const char* the_other_date, bool is_first_date)
{
	// Using the values in min_max_flag, this function parses the_date,
	// provides default values to missing portions, and returns the
	// "complete" date as an int
	//
	// the_other_date is needed for properly handling the min_max_flag

	int t_year, t_month, t_day;
	int t_other_year = 0, t_other_month = 0, t_other_day = 0;

	parse_date(the_date, t_year, t_month, t_day);
	parse_date(the_other_date, t_other_year, t_other_month, t_other_day);

	// Fill in the 0 values for the date

	if (t_month == 0)
	{
		if (min_max_flag == DT_MIN)
		{
			if (is_first_date)
			{
				// If the year portion is the same for both dates,
				// assume the months are the same.
				if ((t_year == t_other_year) &&
					(t_other_month > 0))
				{
					t_month = t_other_month;
				}
				else
				{
					// Assume December
					t_month = 12;
				}
			}
			else
			{
				if (t_other_month == 0)
				{
					// If the other month portion is empty, then when it
					// came through as the first date, it calculated to 12.
					// Do that here, too. (This is insane, you know.)
					t_other_month = 12;
				}

				// If year portion is the same for the_date and
				// the_other_date, assume the months are the same.
				if (t_year == t_other_year)
				{
					t_month = t_other_month;
				}
				else
				{
					// Otherwise, assume January
					t_month = 1;
				}
			}
		}
		else if (min_max_flag == DT_MAX)
		{
			if (is_first_date)
			{
				// Assume January
				t_month = 1;
			}
			else
			{
				// Assume December
				t_month = 12;
			}
		}
	} // checking if t_month is 0

	if (t_day == 0)
	{
		if (min_max_flag == DT_MIN)
		{
			if (is_first_date)
			{
				// If the year and month portions are the same
				// for both dates, assume the days are the same, too
				if ((t_year == t_other_year) &&
					(t_month == t_other_month) &&
					(t_other_day > 0))
				{
					t_day = t_other_day;
				}
				else
				{
					// Assume last day of the month
					t_day = MonthDaysOfYear[t_month][1];
				}
			}
			else
			{
				// SECOND date in function parameter list
				if (t_other_month == 0)
				{
					// force a match
					t_other_month = t_month;
				}
				// If the year and month portions are the same
				// for both dates...
				if ((t_year == t_other_year) &&
					(t_month == t_other_month))
				{
					// Good grief... if the day portion of the first date
					// is not blank, default the second date's day to
					// the first date's day
					if (t_other_day > 0)
					{
						t_day = t_other_day;
					}
					else
					{
						// Otherwise, the first date's day portion is ALSO
						// blank, and it calculated to the last day of the
						// month when it came through. SO to make these guys
						// match, we'll
						t_day = MonthDaysOfYear[t_other_month][1];
					}
				}
				else
				{
					// Otherwise, assume first day of the month
					t_day = 1;
				}
			}
		}
		else if (min_max_flag == DT_MAX)
		{
			if (is_first_date)
			{
				// Assume first day of the month
				t_day = 1;
			}
			else
			{
				// Assume last day of the month
				t_day = MonthDaysOfYear[t_month][1];
			}
		}
	} // checking if t_day is 0

	int t_complete_date = (t_year * 10000) +
						   (t_month * 100) +
						   (t_day);

	return t_complete_date;
} // complete_date
//---------------------------------------------------------------------------
bool TLibFuncs::is_leap_year(int the_year)
{
	bool t_is_leap_year = true;

	// If the_year is evenly divisible by 4, continue evaluation
	if ((the_year % 4) == 0)
	{
		// If the year is evenly divisible by 100, continue evaluation.
		// Most century years are NOT leap years, except for those that
		// are evenly divisible by 400. So, these century years are NOT
		// leap years: 1700, 1800, 1900, 2100
		// These century years ARE leap years: 1600, 2000, 2400
		// (Admit it: You didn't know these rules off the top of your head,
		// did you? Aren't you glad to have it demystified?)

		// If it's a century year, keep evaluating...
		if ((the_year % 100) == 0)
		{
			// If the year is NOT evenly divisible by 400, it is NOT
			// a leap year
			if ((the_year % 400) != 0)
			{
				t_is_leap_year = false;
			}
		}
	}
	else
	{
		t_is_leap_year = false;
	}
	return t_is_leap_year;
} // is_leap_year
//---------------------------------------------------------------------------
int TLibFuncs::get_day_of_year(int the_date)
{
	int t_year, t_month, t_day;
	t_year = (int)(the_date / 10000);
	t_month = (int)((the_date / 100) - t_year * 100);
	t_day = (int)(the_date - (t_month * 100 + t_year * 10000));

	int t_day_of_year = MonthDaysOfYear[t_month][0] + t_day -1;
	if (t_month > 2 && is_leap_year(t_year))
	{
		t_day_of_year++;
	}

	return t_day_of_year;
} // get_day_of_year
//---------------------------------------------------------------------------
int TLibFuncs::internal_lookup(const char* lookup_value,
	const char* table_specs, bool populate_tablevars)
{
	int t_result = 0,
		t_pad_to = 0;
	int t_column_number = -1; // needed for retrieving from the correct column of table
	std::string t_table_specs,
				t_table_name,
				t_column,
				t_scratch,
				t_sql,
				t_sqlize_value;

	// Save this as the key to the std::map of parameterized queries
	t_table_specs.assign(table_specs);

	// uppercase everything
	uppercase(t_table_specs);

	TLookupQuery* t_lookup_qry = FInternalLookupQueries[t_table_specs];
	TFastSQLQuery* t_qry = NULL;

	if (t_lookup_qry)
	{
		t_qry = t_lookup_qry->GetQuery();
		t_pad_to = t_lookup_qry->GetPadTo();

		t_column_number = t_lookup_qry->GetColumnOrder(0);
	}
	else
	{
		// This is the first data record that makes a call from the
		// currently running edit to perform a lookup into the specified
		// table and column. Prepare a parameterized query for it.

		// table_specs is likely to arrive with decorations... particularly,
		// at least for the near term, it will look like "BPLACE.DBF.CODE",
		// i.e., with ".DBF" stuck in the middle of it. We will have to parse
		// this down to [table_name[ [throw_away_DBF] and [column_name] ]
		//
		// BACKGROUND: The EmfToSmf converter created trigger-maintained fields
		// for what used to be concatenated-field indexed items. We'll want to
		// replace these with something more adaptable to SQL, but that will come
		// later.

		TTableDef* t_table = NULL;

		size_t t_pos = t_table_specs.find_first_of(".");
		if (t_pos != std::string::npos)
		{
			t_table_name.assign(t_table_specs.substr(0,t_pos));
			// Lop off the table name just past the ".", which may or may not
			// be "DBF"
			t_scratch.assign(t_table_specs.substr(t_pos+1,t_table_specs.length()));
			// Look for another delimiter indicating where the column specifier
			// begins
			t_pos = t_scratch.find_first_of(".");
			if (t_pos != std::string::npos)
			{
				t_column.assign(t_scratch.substr(t_pos+1, t_scratch.length()));
			}
			else
			{
				// If there are no more "." delimiters, we MAY have received
				// t_table_specs like "BPLACE.CODE"... that is, no DBF.
				// OR we may have received only "BPLACE.DBF", no column identifier...
				// which would be a syntax error, but we have to check.
				if (t_scratch.substr(0,3).compare("DBF") == 0)
				{
					t_scratch.assign("Table spec to LOOKUP function needs column identifier; received (\"");
					t_scratch += t_table_specs;
					t_scratch += "\")";
					FEngineErrorsPump->AddError("TLibFuncs", "internal_lookup",
						t_scratch);
					throw std::runtime_error("syntax error");
				}
				// All we have left in the string is the column name
				t_column.assign(t_scratch);
			}
		}

		if (t_table_name.empty() || t_column.empty())
		{
			// Throw a fit
			FEngineErrorsPump->AddError("TLibFuncs", "internal_lookup", "Cannot find lookup table or column");
			FEngineErrorsPump->AddError("TLibFuncs", "internal_lookup", t_table_specs);
			throw std::runtime_error("syntax error");
		}
		else
		{
			// We have to know the column number in order to retrieve
			// the correct data, but TFastSQLQuery doesn't "do" GetFieldByName
			// for efficiency, so we'll use a different route to the information.
			t_table = FTablesList->GetTable(t_table_name);

			TFieldDef* t_fld = t_table->GetColumn(trim(t_column).c_str());
			if (t_fld)
			{
				t_column_number = t_fld->GetColumnOrder();
				if (t_fld->GetIsTriggerField())
				{
					t_pad_to = get_trigger_padded_size(t_table_name.c_str(), t_column.c_str());
				}
			}
		}

		t_lookup_qry = new TLookupQuery(FSQLDB, t_table);
		t_lookup_qry->SetPadTo(t_pad_to);
		t_qry = t_lookup_qry->GetQuery();

		t_sql.assign("SELECT * FROM [");
		t_sql += t_table_name;
		t_sql += "] WHERE [";
		t_sql += t_column;
		t_sql += "] = ? ;";

		t_qry->SetSQLText(t_sql, true);

		if (t_qry->PrepareParameterizedQuery())
		{
			t_lookup_qry->AddColumnOrder(t_column_number);
			FInternalLookupQueries[t_table_specs] = t_lookup_qry;
		}
		else
		{
			FEngineErrorsPump->AddError("TLibFuncs", "internal_lookup",
				"Failure to prepare parameterized query");
			throw std::runtime_error("something bad has happened");
		}
	}

	try
	{
		// Escape parameters containing apostrophes ("SQLize" them)
		if (t_pad_to > 0)
		{
			// Make sure lookups into trigger-maintained fields are
			// padded to their full column width.
			t_sqlize_value.assign(stringReplace(padright(lookup_value, t_pad_to), "'", "''"));
		}
		else
		{
			// This is worrisome: Due to lookup table STREET having a record
			// containing an empty string (and not what it really wanted, a
			// blank space), we have to RTRIM the lookup_value. This is the fault
			// of EditWriter 4, which was trimming everything the user typed on
			// its way into lookup tables. Oh, what a tangled web we weave...
			t_sqlize_value.assign(stringReplace(rtrim(lookup_value), "'", "''"));
		}

		t_qry->BindTextParameter( 1, t_sqlize_value.c_str() );

		if (t_qry->OpenParameterizedQuery())
		{
			// EDITS enforce case-sensitivity, which is difficult to accomplish
			// with SQLite's collating sequences. So we'll retrieve data in
			// the loosest possible sense (COLLATE NOCASE is defined on text
			// fields), and then evaluate what we get.
			std::string t_compare;

			do
			{
				// If we found more than one match, this may be a
				// case-sensitivity issue. Compare the actual values
				// to the t_trim_lookup_value
				t_compare.assign(t_qry->GetFieldByNumber(t_column_number));
				t_compare.assign(trim(t_compare));
				if (t_compare.compare(trim(lookup_value)) == 0)
				{
					// This isn't quite as stupid as it looks. The NAACCR
					// metafile has a lookup table (STREET) with a one-character
					// entry that is an empty string. It is meant to indicate
					// that a blank character is allowed. So the code for that
					// table will get here, having found a match for '',
					// will NOT be at end-of-file, and therefore, this is a
					// successful lookup. Return TRUE.
					if (!t_qry->Eof())
					{
						t_result = 1L;
					}
					break;
				}
				t_qry->FetchNext();
			} while (!t_qry->Eof());

			if (t_result == 1L)
			{
				if (populate_tablevars)
				{
					std::string t_fld_name,
								t_fld_value;

					int t_int_fld_value;

					// Retrieve the TTableDef for this lookup table,
					// and iterate through its columns to find any
					// that were declared TABLEVAR.
					TTableDef* t_table = t_lookup_qry->GetTableDef(); // FTablesList->GetTable(t_table_name);
					if (!t_table)
					{
						t_scratch.assign("TablesList does not have a table named \"");
						t_scratch += t_table_name;
						t_scratch += "\"";
						FEngineErrorsPump->AddError("TLibFuncs", "internal_lookup", t_scratch);
						throw std::runtime_error("syntax error");
					}
					int t_count = t_table->GetColumnCount();
					// The first column in EVERY metafile table is
					// PKEY... a fact we needn't share with the user.
					// So start with the second column (t_idx = 1)
					for (int t_idx = 1; t_idx < t_count; t_idx++)
					{
						TFieldDef* t_fld = t_table->GetColumn(t_idx);
						if (t_fld)
						{
							t_fld_name.assign(t_fld->GetFieldName());
							TVariable* t_var =
								FRunEdits->GetVariable(t_fld_name.c_str(), false);
							if (t_var)
							{
								if (t_var->GetVariableType() & EEVT_CHAR)
								{
									t_fld_value = t_qry->GetFieldByNumber(t_idx);
									int t_fld_size = t_fld->GetColumnSize();
									if (t_fld->GetIsTriggerField())
									{
										t_fld_size = (int)t_fld_value.length();
									}
									t_var->SetTableVarValue(t_fld_value.c_str(), t_fld_size);
								}
								else if (t_var->GetVariableType() & EEVT_INT)
								{
									t_int_fld_value = t_qry->GetIntFieldByNumber(t_idx);
									t_var->SetValue(0,t_int_fld_value);
								}
							}
						}
						else
						{
							t_scratch.assign("Table named \"");
							t_scratch += t_table_name;
							t_scratch += "\" states column count = ";
							t_scratch += intToString(t_count);
							t_scratch += " but cannot return field at index ";
							t_scratch += intToString(t_idx+1);

							FEngineErrorsPump->AddError("TLibFuncs", "internal_lookup", t_scratch);
							throw std::runtime_error("EditEngine error");
						}
					}

				} // if (populate_tablevars)
			} // if (t_result == 1L)

			t_qry->ResetParameterizedQuery();

		} // if (t_qry->OpenParameterizedQuery)

	}
	catch (...)
	{
		std::string t_error = t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TLibFuncs","internal_lookup",t_error);
		char t_buffer[200];
		std::snprintf(t_buffer,200,"SQL: %s", t_buffer);
		// Report the SQL that failed
		t_error.assign(t_buffer);
		FEngineErrorsPump->AddError("TLibFuncs","internal_lookup",t_error);
		throw std::runtime_error("SQL error");
	}

	return t_result;
} // internal_lookup
//---------------------------------------------------------------------------

int TLibFuncs::internal_rlookup(const char* lookup_value,
	const char* table_specs)
{
	int t_result = 0,
		t_pad_to = 0;

	// table_specs is likely to arrive with decorations... particularly,
	// at least for the near term, it will look like "CS_OBS.DBF.INDEX1",
	// i.e., with ".DBF" stuck in the middle of it. We will have to parse
	// this down to [table_name[ [throw_away_DBF] and [column_name]
	//
	// BACKGROUND: The EmfToSmf converter created trigger-maintained fields
	// for what used to be concatenated-field indexed items. We'll want to
	// replace these with something more adaptable to SQL, but that will come
	// later.

	std::string t_table_specs,
				t_table_name,
				t_column,
				t_scratch,
				t_sql,
				t_sqlize_value;

	// Save this as the key to the std::map of parameterized queries
	t_table_specs.assign(table_specs);

	// uppercase everything
	uppercase(t_table_specs);

	TLookupQuery* t_lookup_qry = FInternalRLookupQueries[t_table_specs];
	TFastSQLQuery* t_qry = NULL;

	if (t_lookup_qry)
	{
		t_qry = t_lookup_qry->GetQuery();
		t_pad_to = t_lookup_qry->GetPadTo();
	}
	else
	{
		// This is the first data record that makes a call from the
		// currently running edit to perform a lookup into the specified
		// table and column. Prepare a parameterized query for it.

		TTableDef* t_table = NULL;

		// Get the table name
		size_t t_pos = t_table_specs.find_first_of(".");
		if (t_pos != std::string::npos)
		{
			t_table_name.assign(t_table_specs.substr(0,t_pos));
			// Lop off the table name just past the ".", which may or may not
			// be "DBF"
			t_scratch.assign(t_table_specs.substr(t_pos+1,t_table_specs.length()));
			// Look for another delimiter indicating where the column specifier
			// begins
			t_pos = t_scratch.find_first_of(".");
			if (t_pos != std::string::npos)
			{
				t_column.assign(t_scratch.substr(t_pos+1, t_scratch.length()));
			}
			else
			{
				// If there are no more "." delimiters, we MAY have received
				// t_table_specs like "BPLACE.CODE"... that is, no DBF.
				// OR we may have received only "BPLACE.DBF", no column identifier...
				// which would be a syntax error, but we have to check.
				if (t_scratch.substr(0,3).compare("DBF") == 0)
				{
					t_scratch.assign("Table spec to LOOKUP function needs column identifier; received (\"");
					t_scratch += t_table_specs;
					t_scratch += "\")";
					FEngineErrorsPump->AddError("TLibFuncs", "internal_rlookup",
						t_scratch);
					throw std::runtime_error("syntax error");
				}
				// All we have left in the string is the column name
				t_column.assign(t_scratch);
			}
		}

		if (t_table_name.empty() || t_column.empty())
		{
			// Throw a fit
			FEngineErrorsPump->AddError("TLibFuncs", "internal_rlookup", "Cannot find lookup table or column");
			FEngineErrorsPump->AddError("TLibFuncs", "internal_rlookup", t_table_specs);
			throw std::runtime_error("syntax error");
		}
		else
		{
			// Unlike internal_lookup, we do NOT have to know the
			// column number in order to retrieve the correct data,
			// because RLOOKUP is used for a "find closest match"
			// lookup.
			// HOWEVER, TLookupQuery likes a pointer to the TTableDef,
			// so let's just give it one
			t_table = FTablesList->GetTable(t_table_name);
			// Find the field
			TFieldDef* t_fld = t_table->GetColumn(trim(t_column).c_str());
			if (t_fld && t_fld->GetIsTriggerField())
			{
				t_pad_to = get_trigger_padded_size(t_table_name.c_str(), t_column.c_str());
			}
		}

		t_lookup_qry = new TLookupQuery(FSQLDB, t_table);
		t_lookup_qry->SetPadTo(t_pad_to);
		t_qry = t_lookup_qry->GetQuery();

		t_sql.assign("SELECT * FROM [");
		t_sql += t_table_name;
		t_sql += "] WHERE [";
		t_sql += t_column;
		t_sql += "] <= ? ORDER BY [";
		t_sql += t_column;
		t_sql += "] DESC LIMIT 1;";

		t_qry->SetSQLText(t_sql, true);

		if (t_qry->PrepareParameterizedQuery())
		{
			// Note that we don't touch the ColumnOrderList vector
			// in TLookupQuery... for RLOOKUP it can remain empty
			FInternalRLookupQueries[t_table_specs] = t_lookup_qry;
		}
		else
		{
			FEngineErrorsPump->AddError("TLibFuncs", "internal_rlookup",
				"Failure to prepare parameterized query");
			throw std::runtime_error("something bad has happened");
		}
	}

	try
	{
		// Escape parameters containing apostrophes ("SQLize" them)
		if (t_pad_to > 0)
		{
			t_sqlize_value.assign(stringReplace(padright(lookup_value, t_pad_to), "'", "''"));
		}
		else
		{
			t_sqlize_value.assign(stringReplace(lookup_value, "'", "''"));
		}

		t_qry->BindTextParameter( 1, t_sqlize_value.c_str() );

		if (t_qry->OpenParameterizedQuery())
		{
			if (!t_qry->Eof())
			{
				t_result = 1L;

				// RLOOKUP always populates tablevars...
				//if (populate_tablevars)
				{
					std::string t_fld_name,
								t_fld_value;

					int t_int_fld_value;

					// Retrieve the TTableDef for this lookup table,
					// and iterate through its columns to find any
					// that were declared TABLEVAR.
					TTableDef* t_table = t_lookup_qry->GetTableDef(); // FTablesList->GetTable(t_table_name);
					if (!t_table)
					{
						t_scratch.assign("TablesList does not have a table named \"");
						t_scratch += t_table_name;
						t_scratch += "\"";
						FEngineErrorsPump->AddError("TLibFuncs", "internal_rlookup", t_scratch);
						throw std::runtime_error("syntax error");
					}
					int t_count = t_table->GetColumnCount();

					// The first column in EVERY metafile table is
					// PKEY... a fact we needn't share with the user.
					// So start with the second column (t_idx = 1)
					for (int t_idx = 1; t_idx < t_count; t_idx++)
					{
						TFieldDef* t_fld = t_table->GetColumn(t_idx);
						if (t_fld)
						{
							t_fld_name.assign(t_fld->GetFieldName());
							TVariable* t_var = FRunEdits->GetVariable(t_fld_name.c_str(), false);
							if (t_var)
							{
								if (t_var->GetVariableType() & EEVT_CHAR)
								{
									t_fld_value = t_qry->GetFieldByNumber(t_idx);
									int t_fld_size = t_fld->GetColumnSize();
									if (t_fld->GetIsTriggerField())
									{
										t_fld_size = (int)t_fld_value.length();
									}
									t_var->SetTableVarValue(t_fld_value.c_str(), t_fld_size);
								}
								else if (t_var->GetVariableType() & EEVT_INT)
								{
									t_int_fld_value = t_qry->GetIntFieldByNumber(t_idx);
									t_var->SetValue(0,t_int_fld_value);
								}
							}
						}
						else
						{
							t_scratch.assign("Table named \"");
							t_scratch += t_table_name;
							t_scratch += "\" states column count = ";
							t_scratch += intToString(t_count);
							t_scratch += " but cannot return field at index ";
							t_scratch += intToString(t_idx+1);

							FEngineErrorsPump->AddError("TLibFuncs", "internal_rlookup", t_scratch);
							throw std::runtime_error("EditEngine error");
						}
					}

				}
			} // if (!t_qry->Eof())

		} // if (t_qry->OpenParameterizedQuery())

		t_qry->ResetParameterizedQuery();

	}
	catch (...)
	{
		std::string t_error = t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TLibFuncs","internal_rlookup",t_error);
		char t_buffer[200];
		std::snprintf(t_buffer,200,"SQL: %s", t_qry->GetSQLText());
		// Report the SQL that failed
		FEngineErrorsPump->AddError("TLibFuncs","internal_rlookup",t_buffer);
		throw std::runtime_error("SQL error");
	}

	return t_result;
} // internal_rlookup
//---------------------------------------------------------------------------
int TLibFuncs::get_trigger_padded_size(const char* table_name, const char* column_name)
{
	int t_pad_to = 0;

	std::unique_ptr<TFastSQLQuery> t_pad_qry(new TFastSQLQuery(FSQLDB));
	char t_buffer[500];

	std::snprintf(t_buffer, 500,
		"SELECT LENGTH([%s]) AS PadTo FROM [%s] LIMIT 1;",
		column_name, table_name);

	try
	{
		t_pad_qry->SetSQLText(t_buffer, true);
		t_pad_qry->SetIsActive(true);
		t_pad_to = t_pad_qry->GetIntFieldByNumber(0);
		t_pad_qry->SetIsActive(false);
	}
	catch (...)
	{
		std::string t_error = t_pad_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TLibFuncs","get_trigger_padded_size",t_error);
		char t_buffer2[500];
		std::snprintf(t_buffer2, 500, "SQL: %s", t_buffer);
		FEngineErrorsPump->AddError("TLibFuncs","get_trigger_padded_size",t_buffer2);
		throw std::runtime_error("SQL error");

	}
	return t_pad_to;
} // get_trigger_padded_size
//---------------------------------------------------------------------------

int TLibFuncs::internal_sqllookup(const char* table_name,
								  const char* columns_list,
								  const char* lookup_value,
								  bool populate_tablevars)
{
	int t_result = 0;

	std::string t_table_specs = std::string(table_name) +
								std::string(columns_list);

	std::string t_sqlize_lookup_value = stringReplace(lookup_value, "'", "''");
	std::string t_scratch;

	TLookupQuery* t_lookup_qry = FInternalSQLLookupQueries[t_table_specs];
	TFastSQLQuery* t_qry = NULL;

	if (t_lookup_qry)
	{
		t_qry = t_lookup_qry->GetQuery();
	}
	else
	{
		// This is the first data record that makes a call from the
		// currently running edit to perform a lookup into the specified
		// table and column. Prepare a parameterized query for it.

		TTableDef* t_table = FTablesList->GetTable(table_name);
		if (!t_table)
		{
			t_scratch.assign("TablesList does not have a table named \"");
			t_scratch += table_name;
			t_scratch += "\"";
			FEngineErrorsPump->AddError("TLibFuncs", "internal_sqllookup", t_scratch);
			throw std::runtime_error("syntax error");
		}

		// columns_list may consist of just one column name, or
		// a comma-separated list of multiple columns. For the latter
		// condition, our SQL will will concatenate the columns.
		//
		// Examples:
		//     SQLLookup("STATEZIP", "ZipHigh", t_hi_zip);
		//     yields this query:
		//     SELECT * FROM [STATEZIP] WHERE [ZipHigh] = '71449';
		//
		//     SQLLookup("STATEZIP", "State, ZipLow", t_search_str);
		//     yields
		//     SELECT * FROM [STATEZIP] WHERE [State] || [ZipLow] = 'KS66000';
		//

		// let's squeeze out unneeded embedded spaces: "State,ZipLow"
		std::string t_columns_list = stringReplace(trim(columns_list), " ", "");
		std::vector<std::string> t_column_names;
		size_t t_pos = t_columns_list.find_first_of(",");
		while (t_pos != std::string::npos)
		{
			t_column_names.push_back(t_columns_list.substr(0, t_pos));
			t_columns_list = t_columns_list.substr(t_pos+1, t_columns_list.length());
			t_pos = t_columns_list.find_first_of(",");
		}

		// now get the only or last column name in the list
		if (!t_columns_list.empty())
		{
			t_column_names.push_back(t_columns_list.substr(0, t_pos));
		}

		// Create a new TLookupQuery instance and grab its TFastSQLQuery()
		t_lookup_qry = new TLookupQuery(FSQLDB, t_table);
		t_qry = t_lookup_qry->GetQuery();

		// We have to know the column order number so that we can retrieve
		// the correct data, but TFastSQLQuery doesn't "do" GetFieldByName
		// for reasons of efficiency. So we'll take a different route to
		// the information.

		for (size_t t_idx = 0; t_idx < t_column_names.size(); t_idx++)
		{
			TFieldDef* t_fld = t_table->GetColumn(t_column_names[t_idx].c_str());
			if (t_fld)
			{
				t_lookup_qry->AddColumnOrder(t_fld->GetColumnOrder());
			}
			else
			{
				// Uh-oh... no such field in this table. Scream!
				t_scratch.assign("The table named \"");
				t_scratch += table_name;
				t_scratch += "\" does not have a column named \"";
				t_scratch += t_column_names[t_idx];
				t_scratch += "\"";
				FEngineErrorsPump->AddError("TLibFuncs", "internal_sqllookup", t_scratch);
				throw std::runtime_error("syntax error");

			}
		}

		// Now build the parameterized SQL
		std::string t_sql = "SELECT * FROM [";
		t_sql += table_name;
		t_sql += "] WHERE ";
		for (size_t t_idx = 0; t_idx < t_column_names.size(); t_idx++)
		{
			if (t_idx > 0)
			{
				t_sql += " || ";
			}
			t_sql += std::string("[") + t_column_names[t_idx] + std::string("]");
		}

		t_sql += " = ? ;";

		t_qry->SetSQLText(t_sql, true);

		if (t_qry->PrepareParameterizedQuery())
		{
			FInternalSQLLookupQueries[t_table_specs] = t_lookup_qry;
		}
		else
		{
			FEngineErrorsPump->AddError("TLibFuncs", "internal_sqllookup",
				"Failure to prepare parameterized query");
			throw std::runtime_error("something bad has happened");
		}

	} // if (t_lookup_qry == NULL)

	// If we get to here, we have a query we can run
	try
	{
		t_qry->BindTextParameter(1, t_sqlize_lookup_value.c_str() );

		if (t_qry->OpenParameterizedQuery())
		{
			// EDITS enforce case-sensitivity, which is difficult to accomplish
			// with SQLite's collating sequences. So we'll retrieve data in
			// the loosest possible sense (COLLATE NOCASE is defined on text
			// fields), and then evaluate what we get.
			std::string t_compare = "";

			do
			{
				// If we found more than one match, this may be a
				// case-sensitivity issue. Compare the actual values
				// to the t_sqlize_lookup_value
				for (int t_idx = 0; t_idx < t_lookup_qry->GetColumnsCount(); t_idx++)
				{
					t_compare += t_qry->GetFieldByNumber(t_lookup_qry->GetColumnOrder(t_idx));
				}

				if (t_compare.compare(t_sqlize_lookup_value) == 0)
				{
					// This isn't quite as stupid as it looks. The NAACCR
					// metafile has a lookup table (STREET) with a one-character
					// entry that is an empty string. It is meant to indicate
					// that a blank character is allowed. So the code for that
					// table will get here, having found a match for '',
					// will NOT be at end-of-file, and therefore, this is a
					// successful lookup. Return TRUE.
					if (!t_qry->Eof())
					{
						t_result = 1L;
					}
					break;
				}
				t_qry->FetchNext();
			} while (!t_qry->Eof());

			if (t_result == 1L)
			{
				if (populate_tablevars)
				{
					std::string t_fld_name,
								t_fld_value;

					int t_int_fld_value;

					// Retrieve the TTableDef for this lookup table,
					// and iterate through its columns to find any
					// that were declared TABLEVAR.
					TTableDef* t_table = t_lookup_qry->GetTableDef(); // FTablesList->GetTable(t_table_name);
					if (!t_table)
					{
						t_scratch.assign("TablesList does not have a table named \"");
						t_scratch += t_table->GetTableName();
						t_scratch += "\"";
						FEngineErrorsPump->AddError("TLibFuncs", "internal_sqllookup", t_scratch);
						throw std::runtime_error("syntax error");
					}
					int t_count = t_table->GetColumnCount();
					// The first column in EVERY metafile table is
					// PKEY... a fact we needn't share with the user.
					// So start with the second column (t_idx = 1)
					for (int t_idx = 1; t_idx < t_count; t_idx++)
					{
						TFieldDef* t_fld = t_table->GetColumn(t_idx);
						if (t_fld)
						{
							t_fld_name.assign(t_fld->GetFieldName());
							TVariable* t_var =
								FRunEdits->GetVariable(t_fld_name.c_str(), false);
							if (t_var)
							{
								if (t_var->GetVariableType() & EEVT_CHAR)
								{
									t_fld_value = t_qry->GetFieldByNumber(t_idx);
									int t_fld_size = t_fld->GetColumnSize();
									if (t_fld->GetIsTriggerField())
									{
										t_fld_size = (int)t_fld_value.length();
									}
									t_var->SetTableVarValue(t_fld_value.c_str(), t_fld_size);
								}
								else if (t_var->GetVariableType() & EEVT_INT)
								{
									t_int_fld_value = t_qry->GetIntFieldByNumber(t_idx);
									t_var->SetValue(0,t_int_fld_value);
								}
							}
						}
						else
						{
							t_scratch.assign("Table named \"");
							t_scratch += t_table->GetTableName();
							t_scratch += "\" states column count = ";
							t_scratch += intToString(t_count);
							t_scratch += " but cannot return field at index ";
							t_scratch += intToString(t_idx+1);

							FEngineErrorsPump->AddError("TLibFuncs", "internal_sqllookup", t_scratch);
							throw std::runtime_error("EditEngine error");
						}
					}

				} // if (populate_tablevars)
			} // if (t_result == 1L)

			t_qry->ResetParameterizedQuery();

		} // if (t_qry->OpenParameterizedQuery)
	}
	catch (...)
	{
		std::string t_error = t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TLibFuncs","internal_sqllookup",t_error);
		char t_buffer[200];
		std::snprintf(t_buffer,200,"SQL: %s", t_buffer);
		// Report the SQL that failed
		t_error.assign(t_buffer);
		FEngineErrorsPump->AddError("TLibFuncs","internal_sqllookup",t_error);
		throw std::runtime_error("SQL error");
	}

	return t_result;
} // internal_sqllookup
//---------------------------------------------------------------------------
int TLibFuncs::internal_sqlrlookup(const char* table_name,
								   const char* columns_list,
								   const char* lookup_value)
{
	int t_result = 0;
	bool populate_tablevars = true;

	std::string t_table_specs = std::string(table_name) +
								std::string(columns_list);

	std::string t_sqlize_lookup_value = stringReplace(lookup_value, "'", "''");
	std::string t_scratch;

	TLookupQuery* t_lookup_qry = FInternalSQLRLookupQueries[t_table_specs];
	TFastSQLQuery* t_qry = NULL;

	if (t_lookup_qry)
	{
		t_qry = t_lookup_qry->GetQuery();
	}
	else
	{
		// This is the first data record that makes a call from the
		// currently running edit to perform a lookup into the specified
		// table and column. Prepare a parameterized query for it.

		TTableDef* t_table = FTablesList->GetTable(table_name);
		if (!t_table)
		{
			t_scratch.assign("TablesList does not have a table named \"");
			t_scratch += table_name;
			t_scratch += "\"";
			FEngineErrorsPump->AddError("TLibFuncs", "internal_sqlrlookup", t_scratch);
			throw std::runtime_error("syntax error");
		}

		// columns_list may consist of just one column name, or
		// a comma-separated list of multiple columns. For the latter
		// condition, our SQL will will concatenate the columns.
		//
		// Examples:
		//     SQLLookup("STATEZIP", "ZipHigh", t_hi_zip);
		//     yields this query:
		//     SELECT * FROM [STATEZIP] WHERE [ZipHigh] = '71449';
		//
		//     SQLLookup("STATEZIP", "State, ZipLow", t_search_str);
		//     yields
		//     SELECT * FROM [STATEZIP] WHERE [State] || [ZipLow] = 'KS66000';
		//

		// let's squeeze out unneeded embedded spaces: "State,ZipLow"
		std::string t_columns_list = stringReplace(trim(columns_list), " ", "");
		std::vector<std::string> t_column_names;
		size_t t_pos = t_columns_list.find_first_of(",");
		while (t_pos != std::string::npos)
		{
			t_column_names.push_back(t_columns_list.substr(0, t_pos));
			t_columns_list = t_columns_list.substr(t_pos+1, t_columns_list.length());
			t_pos = t_columns_list.find_first_of(",");
		}

		// now get the only or last column name in the list
		if (!t_columns_list.empty())
		{
			t_column_names.push_back(t_columns_list.substr(0, t_pos));
		}

		// Create a new TLookupQuery instance and grab its TFastSQLQuery()
		t_lookup_qry = new TLookupQuery(FSQLDB, t_table);
		t_qry = t_lookup_qry->GetQuery();

		// We have to know the column order number so that we can retrieve
		// the correct data, but TFastSQLQuery doesn't "do" GetFieldByName
		// for reasons of efficiency. So we'll take a different route to
		// the information.

		for (size_t t_idx = 0; t_idx < t_column_names.size(); t_idx++)
		{
			TFieldDef* t_fld = t_table->GetColumn(t_column_names[t_idx].c_str());
			if (t_fld)
			{
				t_lookup_qry->AddColumnOrder(t_fld->GetColumnOrder());
			}
			else
			{
				// Uh-oh... no such field in this table. Scream!
				t_scratch.assign("The table named \"");
				t_scratch += table_name;
				t_scratch += "\" does not have a column named \"";
				t_scratch += t_column_names[t_idx];
				t_scratch += "\"";
				FEngineErrorsPump->AddError("TLibFuncs", "internal_sqlrlookup", t_scratch);

				delete t_lookup_qry;

				throw std::runtime_error("syntax error");
			}
		}

		// Now build the parameterized SQL
		std::string t_sql = "SELECT * FROM [";
		t_sql += table_name;
		t_sql += "] WHERE ";
		for (size_t t_idx = 0; t_idx < t_column_names.size(); t_idx++)
		{
			if (t_idx > 0)
			{
				t_sql += " || ";
			}
			t_sql += std::string("[") + t_column_names[t_idx] + std::string("]");
		}

		t_sql += " <= ? ORDER BY ";

		for (size_t t_idx = 0; t_idx < t_column_names.size(); t_idx++)
		{
			if (t_idx > 0)
			{
				t_sql += " || ";
			}
			t_sql += std::string("[") + t_column_names[t_idx] + std::string("]");
		}

		t_sql += " DESC LIMIT 1;";

		t_qry->SetSQLText(t_sql, true);

		if (t_qry->PrepareParameterizedQuery())
		{
			FInternalSQLRLookupQueries[t_table_specs] = t_lookup_qry;
		}
		else
		{
			FEngineErrorsPump->AddError("TLibFuncs", "internal_sqlrlookup",
				"Failure to prepare parameterized query");

			delete t_lookup_qry;

			throw std::runtime_error("cannot prepare parameterized query");
		}

	} // if (t_lookup_qry == NULL)

	// If we get to here, we have a query we can run
	try
	{
		t_qry->BindTextParameter(1, t_sqlize_lookup_value.c_str() );

		if (t_qry->OpenParameterizedQuery())
		{
			if (!t_qry->Eof())
			{
				t_result = 1L;

				std::string t_fld_name,
							t_fld_value;

				int t_int_fld_value;

				// Retrieve the TTableDef for this lookup table,
				// and iterate through its columns to find any
				// that were declared TABLEVAR.
				TTableDef* t_table = t_lookup_qry->GetTableDef();
				if (!t_table)
				{
					// This REALLY shouldn't happen here, but you never know...
					t_scratch.assign("TablesList does not have a table named \"");
					t_scratch += t_table->GetTableName();
					t_scratch += "\"";
					FEngineErrorsPump->AddError("TLibFuncs", "internal_sqlrlookup", t_scratch);
					throw std::runtime_error("syntax error");
				}
				int t_count = t_table->GetColumnCount();
				// The first column in EVERY metafile table is
				// PKEY... a fact we needn't share with the user.
				// So start with the second column (t_idx = 1)
				for (int t_idx = 1; t_idx < t_count; t_idx++)
				{
					TFieldDef* t_fld = t_table->GetColumn(t_idx);
					if (t_fld)
					{
						t_fld_name.assign(t_fld->GetFieldName());
						TVariable* t_var =
							FRunEdits->GetVariable(t_fld_name.c_str(), false);
						if (t_var)
						{
							if (t_var->GetVariableType() & EEVT_CHAR)
							{
								t_fld_value = t_qry->GetFieldByNumber(t_idx);
								int t_fld_size = t_fld->GetColumnSize();
								if (t_fld->GetIsTriggerField())
								{
									t_fld_size = (int)t_fld_value.length();
								}
								t_var->SetTableVarValue(t_fld_value.c_str(), t_fld_size);
							}
							else if (t_var->GetVariableType() & EEVT_INT)
							{
								t_int_fld_value = t_qry->GetIntFieldByNumber(t_idx);
								t_var->SetValue(0,t_int_fld_value);
							}
						}
					}
					else
					{
						t_scratch.assign("Table named \"");
						t_scratch += t_table->GetTableName();
						t_scratch += "\" states column count = ";
						t_scratch += intToString(t_count);
						t_scratch += " but cannot return field at index ";
						t_scratch += intToString(t_idx+1);

						FEngineErrorsPump->AddError("TLibFuncs", "internal_sqlrlookup", t_scratch);
						throw std::runtime_error("EditEngine error");
					}
				}
			} // if (!t_qry->Eof())

			t_qry->ResetParameterizedQuery();

		} // if (t_qry->OpenParameterizedQuery)
	}
	catch (...)
	{
		std::string t_error = t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TLibFuncs","internal_sqlrlookup",t_error);
		char t_buffer[200];
		std::snprintf(t_buffer,200,"SQL: %s", t_buffer);
		// Report the SQL that failed
		t_error.assign(t_buffer);
		FEngineErrorsPump->AddError("TLibFuncs","internal_sqlrlookup",t_error);
		throw std::runtime_error("SQL error");
	}

	return t_result;
} // internal_sqlrlookup
//---------------------------------------------------------------------------
void TLibFuncs::find_regex_pattern(const char* match_mask)
{
	// Called by call_match and call_inlist to locate regex patterns
	// for re-use, or create if they don't already exist
	FRegexIterator = FRegexPatterns.find(match_mask);
	if (FRegexIterator != FRegexPatterns.end())
	{
		// found it! let's use it.
		FTheRegexPattern = FRegexIterator->second;
	}
	else
	{
		std::string t_match_mask = translate_mask_to_regex(match_mask);
		// create a new pattern
#ifdef EE_WINDOWS
#ifdef BCB
		std::tr1::regex t_regex_pattern(t_match_mask.c_str());
#else // MinGW
		std::regex t_regex_pattern(t_match_mask.c_str());
#endif
#else // g++
		boost::regex t_regex_pattern(t_match_mask.c_str());
#endif

		// and store it in the map
		// note: the key is the original/untranslated match_mask,
		// not the one we'll actually run through std::regex
		FRegexPatterns[match_mask] = t_regex_pattern;

		FTheRegexPattern = t_regex_pattern;
	}

} // find_regex_pattern
//---------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////
////////////////             EDITS Language Functions          //////////////
/////////////////////////////////////////////////////////////////////////////
intptr_t TLibFuncs::call_allow_future_date_iop()
{
	// void allow_future_date_iop( years_into_future );
	// Sets state flag in Engine for duration of the currently running edit
	// to permit future dates to be evaluated.

	int t_result = 0;

	int t_years;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_years = (int)FRunEdits->EvaluateExpression();

		FYearsIntoFuture = (size_t)t_years;
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_allow_future_date", e.what());
		throw;
	}

	// Note: the return value of this function is immaterial; the point is
	// to set FYearsIntoFuture to a new value.
	return t_result;

} // call_allow_future_date
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_at()
{
	// int at(<str1>, <str2> [, <width_byte_boundary>]);
	// AT searches character expression <str2> for the first occurence of
	// character expression <str1>. If <str1> is found, its position in <str2>
	// is returned. If not found, zero is returned. <width_byte_boundary> is
	// optional, but if included specifies that only occurences on a
	// width byte boundary will be found.

	int t_result = 0;
	std::string t_str1_value,
				t_str2_value;
	int t_width = 1; // default width_byte_boundary to 1 character

	TTokenID t_token_id;

	try
	{
		// We should be on the opening parenthesis of the parameters list
		check_open_paren();

		// We're expecting a string parameter
		t_str1_value.assign((char*)FRunEdits->EvaluateExpression());

		// Get the next string parameter
		////////// TEMPORARY
		intptr_t eval_result = FRunEdits->EvaluateExpression();
		char* my_temp_value = (char*)eval_result;
		t_str2_value.assign(my_temp_value);
//		t_str2_value.assign((char*)FRunEdits->EvaluateExpression());

		// So we have our strings. Now we need to find the width_byte_boundary.
		// Note that this is an optional third parameter.
		t_token_id = FRunEdits->GetEvalTokenID();
		if (t_token_id == COMMA)
		{
			// Get the width_byte_boundary parameter
			t_width = (int)FRunEdits->EvaluateExpression();
		}
		else if (t_token_id != RPAREN)
		{
			FEngineErrorsPump->AddError("TLibFuncs", "call_at",
				"COMMA or RPAREN expected");
			throw std::runtime_error("");
		}

		if (t_width == 1)
		{
			// This is easy... find the first string anywhere in the
			// second string
			size_t t_found = t_str2_value.find(t_str1_value);
			if (t_found != std::string::npos)
			{
				t_result = (int)(t_found+1);
			}
		}
		else
		{
			// Less easy... we have to carve up t_str2_value into chunks
			// of t_width bytes. A vector, perhaps?
			std::vector<std::string> t_chunks;

			t_chunks.reserve(20);

			size_t t_len = t_str2_value.length();
			for (size_t t_idx = 0; t_idx < t_len; t_idx += t_width)
			{
				t_chunks.push_back(t_str2_value.substr(t_idx, t_width));
			}

			// Now let's see if we can find t_str1_value in the vector
			size_t t_found;
			std::string t_one_chunk;
			for (size_t t_idx = 0; t_idx < t_chunks.size(); t_idx++)
			{
				t_one_chunk = t_chunks[t_idx];
				t_found = t_one_chunk.find(t_str1_value);
				if (t_found != std::string::npos)
				{
					t_result = (int)(t_idx+1);
					break;
				}
			}
		}

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_at", e.what());
		throw;
	}

	// Return the verdict.
	return t_result;
} // call_at
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_empty()
{
	// empty( <str> | <field> );
	// Returns true if string is empty (composed entirely of 0 or more
	// blank spaces).

	int t_result = 1;  // assume it is empty until proven otherwise
	intptr_t t_source;

	try
	{
		check_open_paren();

		t_source = FRunEdits->EvaluateExpression();
		char* temp = (char*)t_source;
		size_t t_len = std::strlen((char*)t_source);
		for (size_t t_idx = 0; t_idx < t_len; t_idx++)
		{
			if ( ((char*)t_source)[t_idx] != ' ')
			{
				t_result = 0; // not empty
				break;
			}
		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_empty", e.what());
		throw;
	}


	return t_result;
} // call_empty
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_error_msg()
{
	// void error_msg( number );
	// changes default error message number

	int t_msg_num;

	try
	{
		check_open_paren();

		t_msg_num = (int)FRunEdits->EvaluateExpression();

		FEditErrors->SetAltMessageNumber(t_msg_num);

		// Note: unlike call_error_text, this function should NOT
		// force the edit to FAIL.
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_error_msg", e.what());
		throw;
	}

	// However, this function should always return failure code 0 so that
	// it can be used in a RETURN statement, like this:
	// return ERROR_MSG( <num> );

	return 0;
} // call_error_msg
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_error_text()
{
	// void error_text( string );
	// changes error message to a string to be reported if the edit fails.

	TEdit* t_running_edit = FRunEdits->GetRunningEdit();

	try
	{
		check_open_paren();

		FEditErrors->SetErrorText(t_running_edit,
			(char*)FRunEdits->EvaluateExpression());

		// Make sure the edit returns FAIL;
		FRunEdits->SetReturnFail();
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_error_text", e.what());
		throw;
	}

	// NOTE: ERROR_TEXT is defined ALWAYS to return FAIL. This is so that
	// it can be used with a RETURN statement, like this:
	// RETURN ERROR_TEXT("This is a failure message.");
	return 0;
} // call_error_text
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_save_error_text()
{
	// void save_error_text( msg1 [, msg2 [,...msgn]]);
	// Generates error messages but does not save flags.
	// NOTE: These message are generated as soon as this function is called, even
	// if the edit continues running.
	std::vector<int> t_messages;

	t_messages.reserve(20);

	int t_msg_num;
	TEdit* t_running_edit = FRunEdits->GetRunningEdit();
	TTokenID t_token_id;

	std::string t_scratch;

	try
	{
		check_open_paren();

		t_msg_num = (int)FRunEdits->EvaluateExpression();

		t_messages.push_back(t_msg_num);

		t_token_id = FRunEdits->GetEvalTokenID();
		if (t_token_id != RPAREN)
		{
			do
			{
				if (t_token_id == COMMA)
				{
					// Get the next message number
					t_msg_num = (int)FRunEdits->EvaluateExpression();
					t_messages.push_back(t_msg_num);

					t_token_id = FRunEdits->GetEvalTokenID();
				}
				else if (t_token_id != RPAREN)
				{
					FEngineErrorsPump->AddError("TLibFuncs", "call_save_error_text",
						"COMMA or RPAREN expected");
					throw std::runtime_error("");
				}
			} while (t_token_id != RPAREN);
		}
		// Make sure the edit returns FAIL;
		FRunEdits->SetReturnFail();

		// Now process the list of message numbers
		size_t t_max_messages = t_messages.size();
		for (size_t t_idx = 0; t_idx < t_max_messages; t_idx++)
		{
			FEditErrors->ReportSaveErrorText(t_running_edit, t_messages[t_idx]);
		}
		// Suppress the default error message... this edit will return FAIL now,
		// but we have already reported the errors we want.
		FRunEdits->SetOverrideErrorMessage(true);
	}
	catch (std::runtime_error& e)
	{
		std::string t_err = e.what();
		if (!t_err.empty())
		{
			FEngineErrorsPump->AddError("TLibFuncs", "call_save_error_text", e.what());
		}
		else
		{
			FEngineErrorsPump->AddError("TLibFuncs", "call_save_error_text",
				"unspecified/unexpected error");
		}
		throw;
	}

	// NOTE: SAVE_ERROR_TEXT is defined ALWAYS to return FAIL. This is so that
	// it can be used with a RETURN statement, like this:
	// RETURN SAVE_ERROR_TEXT(293, 297, 288);
	return 0;
} // call_save_error_text
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_save_text()
{
	// void save_text( string );
	// Causes a message to be displayed, and ensures the edit returns FAIL.

	TEdit* t_running_edit = FRunEdits->GetRunningEdit();
	try
	{
		check_open_paren();

		FEditErrors->SetSaveText(t_running_edit, (char*)FRunEdits->EvaluateExpression());

		// Make sure the edit returns FAIL;
		FRunEdits->SetReturnFail();
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_save_text", e.what());
		throw;
	}

	// NOTE: SAVE_TEXT is defined ALWAYS to return FAIL.
	return 0;
} // call_save_text
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_externaldll()
{
	// int ExternalDLL(dll_name, func_name [, parm1 [,parm2 [,...parm10])
	// Makes calls to functions in external Dynamic Link Libraries (DLLs)
	// from within EDITS.

	// WARNING: This code is a lot more delicate than it looks.
	int t_result = 0;

	// These arrays track the parameters passed to ExternalDLL. Array t_params
	// contains the parameter values (might be taken from variables or literals)

	// If running on Linux, this will hold the library name reformatted
	// (e.g., "CStage0205.dll" will be "libcstage0205.so")
	std::string t_so;

	// Initialize to empty values.
	intptr_t t_params[10];

	for (size_t t_param_idx = 0; t_param_idx < 10; t_param_idx++)
	{
		t_params[t_param_idx] = 0L;
	}

	int t_param_ctr = 0;

	TToken* t_token;
	TTokenID t_token_id;
	TTokenType t_token_type;

	std::string t_dll_name,
				t_err_msg;

	intptr_t t_func_name;

	try
	{
		check_open_paren();

		// Get the DLL name
		t_dll_name.assign((char*)FRunEdits->EvaluateExpression());

		// Now see what comes next. We're expecting a COMMA
		t_token_id = FRunEdits->GetEvalTokenID();
		if (t_token_id != COMMA)
		{
			throw std::runtime_error("syntax error: expected comma between first and second parameters");
		}

		// The second parameter is the function to call in the DLL
		//t_func_name.assign((char*)FRunEdits->EvaluateExpression());
		t_func_name = FRunEdits->EvaluateExpression();

		// Find out if we have more parameters, or if we hit the end
		t_token_id = FRunEdits->GetEvalTokenID();
		if (t_token_id == COMMA)
		{
			// The remaining parameters are optional, so we just crawl along to the
			// end of the parameter list to see what has been provided.

			do {
				t_token = FTokenMaster->GetCompiledToken();
				t_token_id = t_token->GetTokenID();

				if (t_token_id == COMMA)
				{
					continue;
				}

				if (t_token_id != RPAREN)
				{
					t_token_type = t_token->GetTokenType();

					if (t_token_type == STRING)
					{
						t_params[t_param_ctr] = (intptr_t)FTokenMaster->GetCallCargo();
					}
					else if (t_token_type == NUMBER)
					{
						t_params[t_param_ctr] = std::atol((char*)FTokenMaster->GetCallCargo());
					}
					else if ((t_token_type == IDENTIFIER) &&
							 (t_token_id == VAR))
					{
						TVariable* t_var =
							FRunEdits->GetVariable((char*)FTokenMaster->GetCallCargo());
						if (t_var)
						{
							int t_var_type = t_var->GetVariableType();
							if (t_var_type & EEVT_CHAR)
							{
								t_params[t_param_ctr] = (intptr_t)t_var->GetStringValueAsIntPtr();
							}
							else if (t_var_type & EEVT_INT)
							{
								 t_params[t_param_ctr] = t_var->GetValue(0);
							}
						}
						else
						{
							t_err_msg.assign("unknown variable: ");
							t_err_msg += (char*)FTokenMaster->GetCallCargo();
							throw std::runtime_error(t_err_msg);
						}
					}
					else
					{
						// Any other token type is invalid (this is in keeping
						// with how we processed parameters in EDITS40)
					}
					// We have only 10 possible parameter options
					if (++t_param_ctr > 9)
						break;;
				}
			} while (t_token_id != RPAREN);
		} // if (!RPAREN)
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_externaldll", e.what());
		throw;
	}

	// Okay, we have everything we need from the compiled code. Now to get
	// the DLL information
#ifdef EE_WINDOWS
	HINSTANCE hLib = NULL;
#else
	void* hLib = NULL;
#endif
	typedef int (*EditsInterfacePtr)(char *func_name, intptr_t parm1,
		intptr_t parm2, intptr_t parm3, intptr_t parm4, intptr_t parm5,
		intptr_t parm6, intptr_t parm7, intptr_t parm8, intptr_t parm9,
		intptr_t parm10);


	// If the library has not previously been loaded, do so now

	// On Linux, most files use lowercase naming convention,
	// and Windows doesn't care, so we'll always store the name lowercased
	// for easier retrieval from GetExternalDLL
	lowercase(t_dll_name);

	EditsInterfacePtr pEditsInterface = NULL;
	EEt_ExternalDLL* t_dll = GetExternalDLL(t_dll_name);

	if (t_dll)
	{
		hLib = t_dll->hLib;
		pEditsInterface = (EditsInterfacePtr)t_dll->EditsInterfaceFunc;
	}
	else
	{

#ifdef EE_WINDOWS
		hLib = LoadLibrary((char*)t_dll_name.c_str());
#else
		// Authors of NAACCR edits will be writing for the Windows-using
		// audience. t_dll_name is most likely to be in the form "CSTAGE0205.DLL"
		// Change the name to "libcstage0205.so"
		// (on Linux, shared library names start with "lib", have extension ".so",
		// and are all lowercase).
		t_so.assign("lib");
		t_so += t_dll_name;
		size_t t_pos = t_so.rfind(".");
		if (t_pos != std::string::npos)
		{
			t_so = t_so.substr(0, t_pos);
		}
		t_so += ".so";

		hLib = dlopen(t_so.c_str(), RTLD_LAZY);
#endif
		if (hLib)
		{
			// Now we have to find the EditsInterface function in the external
			// library. Reminder: Called libraries MUST have an API function
			// named "EditsInterface" which works as the traffic manager to
			// all other functions. Refer to EDITS documentation.
#ifdef EE_WINDOWS
			pEditsInterface = (EditsInterfacePtr)GetProcAddress(hLib, "EditsInterface");
			if (!pEditsInterface)
			{
				// Don't give up yet! Look for a name-mangled variation. (This can
				// happen when a C-style DLL is built usng Borland tools, and the
				// calling convention does not include __stdcall. This may be
				// subject to further modification if anybody ever uses this feature
				// for libraries other than CStage!
				pEditsInterface = (EditsInterfacePtr)GetProcAddress(hLib, "_EditsInterface");

			}
#else
			// For Linux
			// clear any existing error
			dlerror();
			// and look for the EditsInterface function
			pEditsInterface = (EditsInterfacePtr)dlsym(hLib, "EditsInterface");
			if (dlerror() != NULL)
			{
				// Chances are, pEditsInterface is already NULL, but let's
				// force the issue so that the error handling can kick in, below.
				pEditsInterface = NULL;
			}

#endif
			if (!pEditsInterface)
			{
				t_err_msg.assign("Cannot find EditsInterface function in external dll: ");
				t_err_msg += t_dll_name;
				FEngineErrorsPump->AddError("TLibFuncs", "call_externaldll", t_err_msg);
				throw std::runtime_error("");
			}

			// Add the library to ExternalDLLTracker list for future reference
			AddExternalDLL(t_dll_name, hLib, (void*)pEditsInterface);
		}
		else
		{
#ifdef EE_WINDOWS
			int t_error_num = GetLastError();
			if (t_error_num == 193)
			{
				t_err_msg.assign("Cannot load a 32-bit dll (");
				t_err_msg += t_dll_name;
				t_err_msg += ") into a 64-bit process";
			}
			else
			{
				t_err_msg.assign(system_error_message(GetLastError()));
			}
			FEngineErrorsPump->AddError("TLibFuncs", "call_externaldll", t_err_msg);
			t_err_msg.assign("Failure to load external dll: ");
			t_err_msg += t_dll_name;
#else
			t_err_msg.assign(dlerror());
			t_err_msg += t_so;
#endif
			FEngineErrorsPump->AddError("TLibFuncs", "call_externaldll", t_err_msg);
			throw std::runtime_error("");
		}
	} // if (!hLib)

	// By golly, we should be good to go!
	if (hLib && pEditsInterface)
	{
		try
		{
			t_result = (pEditsInterface)((char*)t_func_name,
										 t_params[0],
										 t_params[1],
										 t_params[2],
										 t_params[3],
										 t_params[4],
										 t_params[5],
										 t_params[6],
										 t_params[7],
										 t_params[8],
										 t_params[9]);

		}
		catch (std::runtime_error& e)
		{
			t_err_msg.assign(e.what());
			if (t_err_msg.empty())
			{
				t_err_msg.assign("failure in call to external dll");
			}
			FEngineErrorsPump->AddError("TLibFuncs", "call_externaldll", t_err_msg);
			throw;
		}
	}

	return t_result;
} // call_externaldll
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_fmtstr()
{
	// fmtstr( <var_name>, <format_str>, < number > );
	// Creates a formatted ASCII string representation of a number.
	// Mercifully, this function formats only ONE number per call. That is,
	// this is legal: fmtstr(str, "3+6=%ld", 3+6);
	// this is not:   fmtstr(str, "%ld+%ld=%ld, 3, 6, 3+6);
	std::string t_fmtstr,            // the second parameter
				t_format_specifier;  // the part between "%" and "ld"

	TToken* t_token;
	TTokenID t_token_id;
	// We should be on the opening parenthesis of the parameters list
	check_open_paren();

	// The first parameter MUST be a variable
	t_token = FTokenMaster->GetHardToken(VAR);

	if (!t_token)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: first parameter MUST be a variable");
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}

	TVariable* t_dest_var = FRunEdits->GetVariable((char*)FTokenMaster->GetCallCargo());

	if (!t_dest_var)
	{
		std::string t_err_msg = "variable named '";
		t_err_msg += (char*)FTokenMaster->GetCallCargo();
		t_err_msg += "' not found (undeclared?)";
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}
	else if (!(t_dest_var->GetVariableType() & EEVT_CHAR))
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: first parameter must be a CHAR[n] variable");
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}

	// Now we're expecting a COMMA
	t_token = FTokenMaster->GetHardToken(COMMA);
	if (!t_token)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: expected comma between first and second parameters");
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}

	// The second parameter may be a STRING literal, or a VAR value, or the
	// return from a function call (e.g., TRIM, UPPER, etc.)
	t_fmtstr.assign((char*)FRunEdits->EvaluateExpression());

	// Another COMMA, please
	t_token_id = FRunEdits->GetEvalTokenID();
	if (t_token_id != COMMA)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: expected comma between second and third parameters");
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}

	// Now we need to get our number
	int t_value = (int)FRunEdits->EvaluateExpression();

	// We have all we need... some assembly required...

	// Parse out the format specifier
	size_t t_pos = t_fmtstr.find("%");
	if (t_pos == std::string::npos)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: format string must include '%'");
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}

	// terrible names for these variables!!
	std::string t_scratch;
	std::string t_cargo = "";

	if (t_pos > 0)
	{
		// Store the first part of the format string in the
		// reference parameter, cargo.
		t_cargo += t_fmtstr.substr(0, t_pos);
	}

	// Lop off the string up through the "%" from t_fmtstr and
	// store it in t_format_specifier
	t_format_specifier = t_fmtstr.substr(t_pos+1, t_fmtstr.length());

	// Mark the end position of the format specifier ("ld")
	t_pos = t_format_specifier.find("ld");
	if (t_pos == std::string::npos)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: format string must include 'ld'");
		FEngineErrorsPump->AddError("TLibFuncs", "call_fmtstr", t_err_msg);
		throw std::runtime_error("");
	}

	// Hold onto whatever trails the "ld" in our scratch variable
	t_scratch.assign(t_format_specifier.substr(t_pos+2, t_format_specifier.length()));
	// and lop off the "ld" portion from t_format_specifier
	t_format_specifier = t_format_specifier.substr(0, t_pos);

	char t_buffer[100];
	std::snprintf(t_buffer, 100, "%d", t_value);
	std::string t_num_str;
	int t_howmany;
	if (t_format_specifier.length() > 0)
	{
		if (t_format_specifier[0] == '0')
		{
			// left pad, zero-fill for <n> bytes
			t_format_specifier = t_format_specifier.substr(1, std::string::npos);
			t_howmany = std::atoi(t_format_specifier.c_str());
			t_num_str = padleft(t_buffer, t_howmany, '0');
		}
		else if (t_format_specifier[0] == '-')
		{
			// pad right to <n> bytes
			t_format_specifier = t_format_specifier.substr(1, std::string::npos);
			t_howmany = std::atoi(t_format_specifier.c_str());
			t_num_str = padright(t_buffer, t_howmany);
		}
		else
		{
			// pad left with blanks to <n> bytes
			t_howmany = std::atoi(t_format_specifier.c_str());
			t_num_str = padleft(t_buffer, t_howmany);
		}
	}
	else
	{
		t_num_str.assign(t_buffer);
	}

	t_cargo += t_num_str;
	t_cargo += t_scratch;

	// Assign the source value to the variable.
	t_dest_var->SetValue(t_cargo.c_str());

	// No return value needed from this call.
	return 0;
} // call_fmtstr
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_ilookup()
{
	// int ILOOKUP( <string>, <tablename[.dbf[.tag]])
	// Looks up a value in a table, and populates TABLEVARs if they
	// have been declared.

	int t_result = 0;

	intptr_t t_lookup_value,
			 t_table_specs;

	check_open_paren();

	t_lookup_value = FRunEdits->EvaluateExpression();

	t_table_specs = FRunEdits->EvaluateExpression();

	t_result = internal_lookup((char*)t_lookup_value, (char*)t_table_specs, true);

	return t_result;

} // call_ilookup
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_inlist()
{
	// int INLIST( <string>, <list of values or ranges>
	//             [, <match_pattern> [, <start_pos> [, <length>]]])
	// Is a value contained in this list?

	int t_result = 0;

	TTokenID t_token_id;
	std::string t_source,
				t_values,
				t_match_mask = ""; // match pattern is optional

	std::string t_rtrim_str; // scratch var

	// If the user does not provide start pos or length, we'll default
	// these appropriately. (User may provide start_pos without length.)
	int t_start_pos = 0,
		t_length = 0;

	bool t_good_mask = true; // unless std::regex says otherwise

	try
	{
		check_open_paren();

		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// The second parameter is the values list to check
		t_values.assign((char*)FRunEdits->EvaluateExpression());

		// The remaining parameters are optional, so we just crawl along to the
		// end of the parameter list to see what has been provided.
		int t_param_idx = 3;

		t_token_id = FRunEdits->GetEvalTokenID();
		do {
			if (t_token_id == COMMA)
			{
				if (t_param_idx == 3)
				{
					t_match_mask.assign((char*)FRunEdits->EvaluateExpression());
				}
				else if (t_param_idx == 4)
				{
					t_start_pos = (int)FRunEdits->EvaluateExpression();
				}
				else
				{
					t_length = (int)FRunEdits->EvaluateExpression();
				}

				t_token_id = FRunEdits->GetEvalTokenID();
			}
			else if (t_token_id != RPAREN)
			{
				throw std::runtime_error("syntax error: expected COMMA or RPAREN");
			}
			t_param_idx++;
		} while (t_token_id != RPAREN);

		// First up: If we have a match pattern, do the match of the entire
		// source string (the match pattern must cover the entire field or string)
		if (!t_match_mask.empty())
		{
			// Find the appropriate regex object for FTheRegexPattern
			find_regex_pattern(t_match_mask.c_str());

#ifdef EE_WINDOWS
#ifdef BCB
			// The CodeGear compiler uses std::tr1
			try
			{
				if (!std::tr1::regex_match((char*)t_source.c_str(), FTheRegexPattern))
				{
					t_good_mask = false;
				}
			}
			catch (std::tr1::regex_error& e)
			{
				// Let's provide as much feedback as we can.
				std::string t_error = "INLIST match pattern invalid: \"" + t_match_mask + "\"";
				FEngineErrorsPump->AddError("TLibFuncs", "call_inlist", t_error);
				throw;
			}
#else
			// Mingw uses std::regex, though this may break if we're not on v4.9
			// (i.e., it may still rely upon boost::regex)
			try
			{
				if (!std::regex_match(t_source, FTheRegexPattern))
				{
					t_good_mask = false;
				}
			}
			catch (...)
			{
				// Let's provide as much feedback as we can.
				std::string t_error = "INLIST match pattern invalid: \"" + t_match_mask + "\"";
				FEngineErrorsPump->AddError("TLibFuncs", "call_inlist", t_error);
				throw;
			}
#endif
#else
			// g++ before v4.9 does not implement std::regex; use boost::regex instead
			try
			{
				if (!boost::regex_match(t_source, FTheRegexPattern))
				{
					t_good_mask = false;
				}
			}
			catch (boost::regex_error& e)
			{
				// Let's provide as much feedback as we can.
				std::string t_error = "INLIST match pattern invalid: \"" + t_match_mask + "\"";
				FEngineErrorsPump->AddError("TLibFuncs", "call_inlist", t_error);
				throw;
			}
#endif
			if (t_good_mask)
			{
				// Sometimes, I think they're trying to kill me. At least one
				// edit used this syntax:
				// INLIST(#S"Diagnostic Proc 73-87", "00-91", "dd,bb" )
				// i.e., return TRUE if the source contains a value between
				// 00-91, OR if it's empty. We just confirmed the mask is good,
				// so if the source IS empty, we're done. Get outta here!
				t_rtrim_str = rtrim(t_source);
				if (t_rtrim_str.empty())
				{
					t_result = 1L;
				}

			}
		} // if (!t_match_mask.empty())

		if (t_good_mask && t_result == 0)
		{
			// If the user specified a starting position, she used 1-based
			// indexing; decrement by one.
			if (t_start_pos > 0)
			{
			  if (t_start_pos -1 < (int)t_source.length() -1)
					t_start_pos--;
				else
				{
					char t_buffer[100];
					std::snprintf(t_buffer, 100,
					 "INLIST start pos (%d) is invalid compared to length of source string (%d)",
					 (int)t_start_pos, (int)t_source.length());
					FEngineErrorsPump->AddError("TLibFuncs", "call_inlist", t_buffer);
					throw;
				}
			}

			// If the user did NOT specify a length, we'll make it everything
			// from the t_start_pos'th position to the end of t_source.
			if (t_length == 0)
			{
				// Trim trailing blanks (this came up for postal code lists
				// that allowed, e.g., "ddddd,dddddddddd", and the value
				// was "94411    ").

				// Of course, this bollixes the list of options
				// if that list consists of blank spaces.
				t_rtrim_str = rtrim(t_source);
				if (!t_rtrim_str.empty())
				{
					t_source = rtrim(t_source);
				}
				t_length = (int)(t_source.length()) - t_start_pos;
			}

			std::string t_compare = t_source.substr(t_start_pos, t_length);

			// Carve up t_values into segments, breaking on the commas
			std::vector<std::string> t_segments;

			t_segments.reserve(20);

			size_t t_pos = t_values.find_first_of(",");
			while (t_pos != std::string::npos)
			{
				t_segments.push_back(t_values.substr(0, t_pos));
				// Lop off the segment we just processed, and continue.
				t_values.assign(t_values.substr(t_pos+1, t_values.length()));
				t_pos = t_values.find_first_of(",");
			}
			// Place whatever is left as the last segment
			// (test to make sure the last character was not a comma)
			if (!t_values.empty())
			{
				t_segments.push_back(t_values);
			}

			// FINALLY!! We can evaluate the t_values against t_compare
			std::string t_low_value, t_high_value;
			size_t t_seg_size = t_segments.size();
			std::string t_segment;

			// Live and learn... there exist edits that use INLIST to test
			// whether an input value (the "source") is an empty string.
			// For some reason, the comparison test doesn't return zeroes.
			// Working around...
			t_rtrim_str = rtrim(t_compare);
			bool t_compare_to_empty = t_rtrim_str.empty();
			for (size_t t_idx = 0; t_idx < t_seg_size; t_idx++)
			{
				t_segment.assign(t_segments[t_idx]);

				if (t_compare_to_empty)
				{
					t_rtrim_str = rtrim(t_segment);
					if (t_rtrim_str.empty())
					{
						// Found it! We have our match... urgh, that is,
						// as long as t_compare and t_segment are the
						// exact same length.
						if (t_compare.length() == t_segment.length())
						{
							// Okay, NOW we found it...
							t_result = 1L;
							break;
						}
					}
				}
				else
				{
					// Now we're looking for hyphens that denote a range
					t_pos = t_segment.find_first_of("-");
					if (t_pos == std::string::npos)
					{
						// No hyphens; place the same value into the low and high
						t_low_value.assign(t_segment);
						t_high_value.assign(t_segment);
					}
					else
					{
						t_low_value.assign(t_segment.substr(0, t_pos));
						t_high_value.assign(t_segment.substr(t_pos+1, t_segment.length()));
					}

					//if ((t_compare.compare(t_low_value) >= 0) &&
					//	(t_compare.compare(t_high_value) <= 0))
					if ((t_rtrim_str.compare(t_low_value) >= 0) &&
						(t_rtrim_str.compare(t_high_value) <= 0))
					{
						// FOUND IT! We have a match...
						t_result = 1L;
						break;
					}
                } // if (t_compare_to_empty
			} // for... (segment analysis)
		} // if (t_good_mask)
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_inlist", e.what());
		throw;
	}
	return t_result;
} // call_inlist
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_justified()
{
	// int JUSTIFIED(<string>|<field>, LEFT | RIGHT)
	// JUSTIFIED tests a field or string for left or right justification.
	// A string or field is considered left justified if there is a non-blank
	// in the first character. A string or field is considered right justified
	// if the last character is non-blank.

	// Returns t_result = 1 for true, 0 for false

	int t_result = 0;
	TToken* t_token;
	std::string t_source;

	try
	{
		check_open_paren();

		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// Now we're expecting a NUMCONST
		t_token = FTokenMaster->GetCompiledToken();

		if (t_token->GetTokenType() == NUMCONST)
		{
			TTokenID t_token_id = t_token->GetTokenID();
			if (t_token_id == (TTokenID) '0')  // token is 'Q'
			{
				if (t_source[t_source.length() -1] != ' ')
				{
					t_result = 1L;
				}
			}
			else if (t_token_id == (TTokenID) '1')  // token is 'R'
			{
				if (t_source[0] != ' ')
				{
					t_result = 1L;
				}
			}
			else
			{
				throw std::runtime_error("syntax error: expected trim flag (RIGHT or LEFT)");
			}
		}
		else
		{
			throw std::runtime_error("syntax error: expected trim flag (RIGHT or LEFT)");
		}

		check_close_paren();
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_justified", e.what());
		throw;
	}

	return t_result;
} // call_justified
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_left()
{
	std::string t_source;
	int t_count;
	intptr_t t_retval = 0;
	try
	{
		check_open_paren();

		// The first parameter is the source string
		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// The second parameter is the number of characters to return
		t_count = (int)FRunEdits->EvaluateExpression();

		// Store the result in cargo, to send it back to the caller
		t_retval = FTokenMaster->CopyToCargo((char*)t_source.substr(0, t_count).c_str());
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_left", e.what());
		throw;
	}
	// parameter cargo is actually a pointer to TRunEdits->FScratchBuffer buffer.
	// return it as an int to TRunEdits::eval7_atom for appropriate processing
	// of functions that return strings
	return t_retval;
} // call_left
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_lookup()
{
	// int LOOKUP( <string>, <tablename[.dbf[.tag]])
	// Looks up a value in a table, and populates TABLEVARs if they
	// have been declared.

	int t_result = 0;

	intptr_t t_lookup_value,
			 t_table_specs;

	try
	{
		check_open_paren();

		t_lookup_value = FRunEdits->EvaluateExpression();

		t_table_specs = FRunEdits->EvaluateExpression();

		t_result = internal_lookup((char*)t_lookup_value, (char*)t_table_specs, true);
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_lookup", e.what());
		throw;
	}

	return t_result;
} // call_lookup
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_lower()
{
	// lower( <str> | <field> );
	// Converts <str> to lowercase.

	std::string t_source;
	intptr_t t_retval = 0;

	try
	{
		check_open_paren();

		t_source.assign((char*)FRunEdits->EvaluateExpression());

		lowercase(t_source);

		t_retval = FTokenMaster->CopyToCargo((char*)t_source.c_str());
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_lower", e.what());
		throw;
	}

	// return cargo as an int to TRunEdits::eval7_atom for appropriate processing
	// of functions that return strings
	return t_retval;

} // call_lower
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_match()
{
	// int MATCH( <string>, <pattern> )
	// Test string for a match against a pattern
	int t_result = 0;

	intptr_t t_target,
			 t_match_mask;

	try
	{
		check_open_paren();

		// The first parameter is the source string
		t_target = FRunEdits->EvaluateExpression();

		// The second parameter is the match pattern (mask)
		t_match_mask = FRunEdits->EvaluateExpression();

		// Find the appropriate regex object for FTheRegexPattern
		find_regex_pattern((char*)t_match_mask);


#ifdef EE_WINDOWS
		try
		{
#ifdef BCB
			// The CodeGear compiler uses std::tr1
			if (std::tr1::regex_match((char*)t_target, FTheRegexPattern))
			{
				t_result = 1L;
			}
		}
		catch (std::tr1::regex_error& e)
#else
			  // Mingw uses std::regex, though this may break if we're not on v4.9
			  // (i.e., it may still rely upon boost::regex)
			if (std::regex_match((char*)t_target, FTheRegexPattern))
			{
				t_result = 1L;
			}
		}
		catch (...)
#endif
		{
			// Let's provide as much feedback as we can.
			std::string t_error = "MATCH pattern invalid: \"";
			t_error += (char*)t_match_mask;
			t_error += "\"";
			FEngineErrorsPump->AddError("TLibFuncs", "call_match", t_error);
			throw;
		}
#else
		try
		{
			// g++ v4.7 relies on Boost. v4.9 and later supports std::regex
			if (boost::regex_match((char*)t_target, FTheRegexPattern))
			{
				t_result = 1L;
			}
		}
		catch (boost::regex_error& e)
		{
			// Let's provide as much feedback as we can.
			std::string t_error = "MATCH pattern invalid: \"";
			t_error += (char*)t_match_mask;
			t_error += "\"";
			FEngineErrorsPump->AddError("TLibFuncs", "call_match", t_error);
			throw;
		}
#endif

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_match", e.what());
		throw;
	}
	// This return value not needed. Cargo has the result
	return t_result;
} // call_match
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_nameexpr()
{
	// nameexpr( <str> )
	// Provides indirect access to variables by their name.
	// Its primary use is the construction of variables.

	intptr_t t_result = 0;
	//TToken* t_token;
	intptr_t t_var_name;

	try
	{
		check_open_paren();

		t_var_name = FRunEdits->EvaluateExpression();

		// does the variable exist?
		TVariable* t_dest_var = FRunEdits->GetVariable((char*)t_var_name);
		if (t_dest_var)
		{
			// call_strcpy and call_strcat will need to retrieve the
			// variable name to locate the dest_var to receive a value
			FTokenMaster->CopyToCargo((char*)t_var_name);
			// all other uses are reading the variable contents
			t_result = (intptr_t)t_dest_var->GetStringValueAsIntPtr();
		}
		else
		{
			char t_err_msg[200];
			std::snprintf(t_err_msg, 200,
				"Variable named \"%s\" not found (is it undeclared?)",
				(char*)FTokenMaster->GetCallCargo());
			throw std::runtime_error(t_err_msg);
		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_nameexpr", e.what());
		throw;
	}

	// This return value not needed. Cargo has the result
	return t_result;
} // call_nameexpr
//---------------------------------------------------------------------------

intptr_t TLibFuncs::call_right()
{
	std::string t_source;
	size_t t_count;
	intptr_t t_retval = 0;
	try
	{
		check_open_paren();

		// The first parameter is the source string
		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// The second parameter is the number of characters to return
		t_count = (int)FRunEdits->EvaluateExpression();

		if (t_count > t_source.length())
		{
			t_count = t_source.length();
		}

		// Store the result in cargo, to send it back to the caller
		t_retval = FTokenMaster->CopyToCargo(
			(char*)t_source.substr(t_source.length() - t_count, t_count).c_str());

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_right", e.what());
		throw;
	}
	// return cargo as an int to TRunEdits::eval7_atom for appropriate processing
	// of functions that return strings
	return t_retval;
} // call_right
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_rlookup()
{
	// int RLOOKUP( <string>, <table_specs>)
	// Range lookup. Searches an indexed table for highest key that is less than
	// or equal to the search key.

	int t_result = 0;

	intptr_t t_lookup_value, t_table_specs;

	try
	{
		check_open_paren();

		t_lookup_value = FRunEdits->EvaluateExpression();

		t_table_specs = FRunEdits->EvaluateExpression();

		t_result = internal_rlookup((char*)t_lookup_value, (char*)t_table_specs);
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_rlookup", e.what());
		throw;
	}

	return t_result;
} // call_rlookup
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_sqllookup()
{
	// int SQLLOOKUP( <table_name>,<columns_list>,<string>)
	// Looks up a value in a table, and populates TABLEVARs

	int t_result = 0;

	intptr_t t_table_name,
			 t_columns_list,
			 t_lookup_value;

	check_open_paren();

	t_table_name = FRunEdits->EvaluateExpression();

	t_columns_list = FRunEdits->EvaluateExpression();

	t_lookup_value = FRunEdits->EvaluateExpression();

	t_result = internal_sqllookup((char*)t_table_name, (char*)t_columns_list,
		(char*)t_lookup_value, true);

	return t_result;

} // call_sqllookup
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_sqlrangelookup()
{
	int t_result = 0;
	intptr_t t_table_name,
			 t_columns_list,
			 t_lookup_value;

	check_open_paren();

	t_table_name = FRunEdits->EvaluateExpression();

	t_columns_list = FRunEdits->EvaluateExpression();

	t_lookup_value = FRunEdits->EvaluateExpression();

	t_result = internal_sqlrlookup((char*)t_table_name, (char*)t_columns_list,
		(char*)t_lookup_value);

	return t_result;
} // call_sqllookup
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_set_warning()
{
	// int set_warning( msg1 [, msg2 [, ...msgn]]);
	// Sets specified flags (message numbers) and generates corresponding
	// error messages. Function causes WARN to be returned from the edit.
	// Can be used to return multiple warning or error messages from the
	// same edit, unlike ERROR_MSG which changes the default error message only.
	// Warnings reported by SET_WARNING are in addition to any reported by
	// ERROR_MSG or other returns from the edit.

	std::vector<int> t_messages;

	t_messages.reserve(20);

	int t_msg_num;
	TEdit* t_running_edit = FRunEdits->GetRunningEdit();
	TTokenID t_token_id;

	std::string t_scratch;

	check_open_paren();

	t_msg_num = (int)FRunEdits->EvaluateExpression();

	t_messages.push_back(t_msg_num);

	t_token_id = FRunEdits->GetEvalTokenID();

	do
	{
		if (t_token_id == COMMA)
		{
			// There is another message number... go get it
			t_msg_num = (int)FRunEdits->EvaluateExpression();

			t_messages.push_back(t_msg_num);

			t_token_id = FRunEdits->GetEvalTokenID();
		}
		else if (t_token_id != RPAREN)
		{
			FEngineErrorsPump->AddError("TLibFuncs", "call_set_warning",
				"syntax error: expected RPAREN");
			throw std::runtime_error("");
		}

	} while (t_token_id != RPAREN);

	// Ensure this edit will return WARN regardless of whatever else happens
	FRunEdits->SetReturnWarn();

	// Now process the list of message numbers
	size_t t_max_messages = t_messages.size();
	for (size_t t_idx = 0; t_idx < t_max_messages; t_idx++)
	{
		FEditErrors->ReportSetWarning(t_running_edit, t_messages[t_idx]);
	}
	// NOTE: SET_WARNING is defined ALWAYS to return WARN (a failure code)
	return 0;
}  // call_set_warning
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_save_warning_text()
{
	// int save_warning_text( msg1 [, msg2 [, ...msgn]]);
	// NOTE: This is a deprecated function
	std::vector<int> t_messages;

	t_messages.reserve(20);

	int t_msg_num;
	TEdit* t_running_edit = FRunEdits->GetRunningEdit();
	TTokenID t_token_id;

	std::string t_scratch;

	check_open_paren();

	t_msg_num = (int)FRunEdits->EvaluateExpression();

	t_messages.push_back(t_msg_num);

	t_token_id = FRunEdits->GetEvalTokenID();

	do
	{
		if (t_token_id == COMMA)
		{
			// There is another message number... go get it
			t_msg_num = (int)FRunEdits->EvaluateExpression();

			t_messages.push_back(t_msg_num);

			t_token_id = FRunEdits->GetEvalTokenID();
		}
		else if (t_token_id != RPAREN)
		{
			FEngineErrorsPump->AddError("TLibFuncs", "call_save_warning_text",
				"syntax error: expected RPAREN");
			throw std::runtime_error("");
		}

	} while (t_token_id != RPAREN);

	// Ensure this edit will return WARN regardless of whatever else happens
	FRunEdits->SetReturnWarn();

	// Now process the list of message numbers
	size_t t_max_messages = t_messages.size();
	for (size_t t_idx = 0; t_idx < t_max_messages; t_idx++)
	{
		FEditErrors->ReportSaveWarningText(t_running_edit, t_messages[t_idx]);
	}
	// NOTE: SAVE_WARNING is defined ALWAYS to return WARN (a failure code)
	return 0;
} // call_save_warning_text
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_strcat()
{
	// strcat(<dest>, <source> [, num]);
	// The <dest> must always be a string variable. However, the source
	// could be any of these:
	//  - string variable
	//  - string literal
	//  - APPFIELD
	//  - LIBFUNC (e.g., trim(), substr(), upper(), lower(), etc.)
	// STRCAT concatenates num characters of <source> onto the end of <dest>;
	// num specifies how many characters to concatenate. If num is not passed,
	// all of <source> is concatenated.
	//
	// Note that neither parameter cargo nor return value are needed by the caller.

	std::string t_scratch;
	std::string t_src_value, t_dest_value;
	int t_count = 0;
	TToken* t_token;
	TTokenID t_token_id;
	TVariable* t_dest_var = NULL;
	try
	{
		// We should be on the opening parenthesis of the parameters list
		check_open_paren();
		// If the first parameter is a call to NAMEEXPR, we'll need to put
		// the FTokenMaster pos back... tell it to save the current position
		FTokenMaster->SavePos();

		// The first parameter MUST be a variable
		t_token = FTokenMaster->GetCompiledToken();
		if (t_token)
		{
			t_token_id = t_token->GetTokenID();
			if (t_token_id == VAR)
			{
				t_dest_var = FRunEdits->GetVariable((char*)FTokenMaster->GetCallCargo());
			}
			else if (t_token_id == LIBFUNC)
			{
				// Put the token back for processing the library call
				FTokenMaster->RestorePos();
				// Run NAMEEXPR
				FRunEdits->EvaluateExpression();
				// and retrieve the name of the variable from cargo
				intptr_t t_varname = FTokenMaster->GetCallCargo();

				t_dest_var = FRunEdits->GetVariable((char*)t_varname);
			}
		}
		else
		{
			throw std::runtime_error("syntax error: invalid first parameter");
		}


		if (!t_dest_var)
		{
			std::string t_err_msg = "Variable named \"";
			t_err_msg += (char*)FTokenMaster->GetCallCargo();
			t_err_msg += "\" not found (undeclared?)";
			throw std::runtime_error(t_err_msg.c_str());
		}
		else if (!(t_dest_var->GetVariableType() & EEVT_CHAR))
		{
			throw std::runtime_error("syntax error: first parameter must be a CHAR[n] variable");
		}
		else
		{
			// Now see what comes next. We're expecting a COMMA
			t_token = FTokenMaster->GetHardToken(COMMA);
			if (!t_token)
			{
				throw std::runtime_error("syntax error: expected comma between first and second parameters");
			}

			// Get the source parameter
			t_src_value.assign((char*)FRunEdits->EvaluateExpression());

			// The next token will be either a COMMA or RPAREN
			t_token_id = FRunEdits->GetEvalTokenID();

			if (t_token_id == COMMA)
			{
				// optional parameter to limit the number of bytes to
				// concatenate from <source>
				t_count = (int)FRunEdits->EvaluateExpression();

				t_token_id = FRunEdits->GetEvalTokenID();
			}

			// Let's verify that we got RPAREN
			if (t_token_id != RPAREN)
			{
				throw std::runtime_error("syntax error: closing parenthesis expected");
			}

			// Get whatever is presently in the <dest> variable
			t_dest_value = (char*)t_dest_var->GetStringValueAsIntPtr();

			if (t_count > 0)
			{
				// Caller wants only the first <n> bytes of the <source>
				t_dest_value += t_src_value.substr(0, t_count);
			}
			else
			{
				// Caller wants to concatenate all of the <source>
				t_dest_value += t_src_value;
			}
			t_dest_var->SetValue(t_dest_value.c_str());
		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_strcat", e.what());
		throw;
	}
	// No return value needed from this call.
	return 0;
} // call_strcat
//---------------------------------------------------------------------------

// idx 45, token 'n'
intptr_t TLibFuncs::call_strcmp()
{
	// int strcmp( <str1>, <str2> [, <num]);
	// STRCMP compares the contents of the first <num> characters of <str1>
	// and <str2>. The comparison is case-sensitive. If <num> is not passed,
	// the full strings are compared.
	//
	// Return value t_result will hold one of these:
	//  0: means <str1> == <str2> (exactly equals)
	// -1: means <str1> < <str2>  (less than)
	//  1: means <str1> > <str2>  (greater than)
	std::string t_str1_value, t_str2_value;
	TTokenID t_token_id;

	// t_result contains the function return value, which is passed up the
	// chain to EvaluateExpression.

	int t_result = 0;
	try
	{
		// We should be on the opening parenthesis of the parameters list
		check_open_paren();

		// Let's see what our <str1> is
		t_str1_value.assign((char*)FRunEdits->EvaluateExpression());

		// Let's see what our <str2> is
		t_str2_value.assign((char*)FRunEdits->EvaluateExpression());

		// So we have our strings. Now we need to find the number of
		// characters to compare. Note that we have an optional third parameter,
		// so we'll loop until we hit the closing parenthesis
		int t_number = 0;

		t_token_id = FRunEdits->GetEvalTokenID();

		if (t_token_id == COMMA)
		{
			t_number = (int)FRunEdits->EvaluateExpression();

			t_token_id = FRunEdits->GetEvalTokenID();
		}
		// Let's verify that we got RPAREN
		if (t_token_id != RPAREN)
		{
			throw std::runtime_error("syntax error: closing parenthesis expected");
		}

		// If the optional <num> parameter was provided, we'll compare just
		// the first <num> bytes of the strings.
		int t_compare;
		if (t_number > 0)
		{
			t_compare = t_str1_value.compare(0, t_number, t_str2_value.substr(0, t_number));
		}
		else
		{
			// If the optional <num> parameter was NOT provided,
			// we'll compare the entire strings.
			t_compare = t_str1_value.compare(t_str2_value);
		}

		if (t_compare < 0)
			t_result = -1;
		else if (t_compare > 0)
			t_result = 1;
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_strcmp", e.what());
		throw;
	}

	// Return the verdict.
	return t_result;
} // call_strcmp
//---------------------------------------------------------------------------

intptr_t TLibFuncs::call_strcpy()
{
	// strcpy( <dest>, <source> );
	// The <dest> must always be a string variable. However, the source could be
	// another string variable, a string literal, the contents of a field in
	// the data buffer, or the return value from an EDITS Language function call.
	//
	// Note that either string variable could be any of these:
	//   - a string literal
	//   - a string variable
	//   - an APPFIELD
	//   - the return value of a LIBFUNC call (e.g., trim(), upper(), substr(), etc.)

	std::string t_src_value;
	TToken* t_token;
	TTokenID t_token_id;
	TVariable* t_dest_var = NULL;
	try
	{
		// We should be on the opening parenthesis of the parameters list
		check_open_paren();

		// If the first parameter is a call to NAMEEXPR, we'll need to put
		// the FTokenMaster pos back... tell it to save the current position
		FTokenMaster->SavePos();

		// The first parameter MUST be a variable
		t_token = FTokenMaster->GetCompiledToken();
		if (t_token)
		{
			t_token_id = t_token->GetTokenID();
			if (t_token_id == VAR)
			{
				t_dest_var = FRunEdits->GetVariable((char*)FTokenMaster->GetCallCargo());
			}
			else if (t_token_id == LIBFUNC)
			{
				// Put the token back for processing the library call
				FTokenMaster->RestorePos();
				// Run NAMEEXPR
				FRunEdits->EvaluateExpression();
				// and retrieve the name of the variable from cargo
				intptr_t t_varname = FTokenMaster->GetCallCargo();

				t_dest_var = FRunEdits->GetVariable((char*)t_varname);
			}
		}
		else
		{
			throw std::runtime_error("syntax error: invalid first parameter");
		}

		if (!t_dest_var)
		{
			char t_err_msg[200];
			std::snprintf(t_err_msg, 200,
				"Variable named \"%s\" not found (is it undeclared?)",
				(char*)FTokenMaster->GetCallCargo());
			throw std::runtime_error(t_err_msg);
		}
		else if (!(t_dest_var->GetVariableType() & EEVT_CHAR))
		{
			throw std::runtime_error("syntax error: first parameter must be a CHAR[n] variable");
		}
		else
		{
			if (t_token_id == VAR)
			{
				// Now see what comes next. We're expecting a COMMA
				t_token = FTokenMaster->GetHardToken(COMMA);
				if (!t_token)
				{
					throw std::runtime_error("syntax error: expected comma between first and second parameters");
				}
			}

			// Get the source parameter
			t_src_value.assign((char*)FRunEdits->EvaluateExpression());

			// The next token will be either a COMMA or RPAREN
			t_token_id = FRunEdits->GetEvalTokenID();
			if (t_token_id == COMMA)
			{
				// optional parameter to limit the number of bytes to
				// concatenate from <source>
				int t_count;

				// strcpy() accepts an optional negation character ("-")
				// before the optional final parameter. This character
				// instructs the function to pad with blank spaces out
				// to the number in this last parameter
				FTokenMaster->SavePos();
				TToken* t_sub_token = FTokenMaster->GetHardToken(SUB);
				if (!t_sub_token)
				{
					// Put the pos back for the next call
					FTokenMaster->RestorePos();
				}

				t_count = (int)FRunEdits->EvaluateExpression();

				// Update the t_token_id for the RPAREN test, below.
				t_token_id = FRunEdits->GetEvalTokenID();

				if (t_count > 0)
				{
					if (t_sub_token)
					{
					  while ((int)t_src_value.length() < t_count)
					  {
							t_src_value += " ";
					  }
					}
					t_src_value.assign(t_src_value.substr(0, t_count));
				}

			}
			// Let's verify that we got RPAREN instead
			if (t_token_id != RPAREN)
			{
				throw std::runtime_error("syntax error: closing parenthesis expected");
			}

			// Assign the source value to the variable.
			t_dest_var->SetValue(t_src_value.c_str());

		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_strcpy", e.what());
		throw;
	}
	// No return value needed from this call.
	return 0;
} // call_strcpy
//---------------------------------------------------------------------------

// idx 47, token 'p'
intptr_t TLibFuncs::call_strlen()
{
	// int strlen( <string> | <field> );
	// t_result will be passed up the chain through TRunEdits::EvaluateExpression
	// to the caller for whatever purpose the caller needs it.
	// STRLEN returns how many characters are in string. (Trailing blanks are not
	// counted.)

	intptr_t t_result = 0;
	// n = strlen(<source>)
	// Source can be any of these:
	//   - a string variable
	//   - a string literal
	//   - an APPFIELD
	//   - the result of another function call, e.g.: strlen(substr(trim("  foo  ", 2))) yields 2
	std::string t_scratch;
	std::string t_source;

	try
	{
		check_open_paren();

		// Get the source parameter
		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// The documentation for strlen says that trailing blanks are
		// not counted, so we'll use our rtrim() function (see Utility.cpp)
		// Hold it! The documentation was wrong... EDITS40 did NOT
		// trim off trailing blanks (nor should it... let the user apply
		// a TRIM function if that is what they want).
		// t_source = rtrim(t_source);

        // And assign the string length to t_result
		t_result = (int)t_source.length();
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_strlen", e.what());
		throw;
	}

	return t_result;
} // call_strlen
//---------------------------------------------------------------------------

// idx 48, token 'q'
intptr_t TLibFuncs::call_substr()
{
	// cargo = substr(<source>, start_pos [, num_chars]);
	// Note that cargo does not have to be a declared entity, like a variable.
	// The return value of substr() generally goes as a source parameter to
	// strcpy, strcat, trim, etc.
	// The source could be any of these:
	//   - a string variable
	//   - a string literal
	//   - an APPFIELD
	//   - the result of another function call, e.g.: substr(trim("  foo  ", 2)) yields "oo"

	std::string t_source;
	TTokenID t_token_id;
	intptr_t t_retval = 0;
	try
	{
		check_open_paren();

		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// So we have our source string. Now we need to find the
		// starting position. Note that we have an optional third parameter,
		// so we'll loop until we hit the closing parenthesis
		int t_start_pos = -1,
			t_count = 0,
			t_number;

		t_token_id = FRunEdits->GetEvalTokenID();
		do {
			if (t_token_id == COMMA)
			{
				// Get the t_start_pos or t_count, depending on the iteration
				// we're in for this loop
				t_number = (int)FRunEdits->EvaluateExpression();

				t_token_id = FRunEdits->GetEvalTokenID();

				if (t_start_pos == -1)
					t_start_pos = t_number-1;
				else
					t_count = t_number;
			}
			else if (t_token_id != RPAREN)
			{
				FEngineErrorsPump->AddError("TLibFuncs", "call_substr",
					"COMMA or RPAREN expected");
				throw std::runtime_error("");
			}
		} while (t_token_id != RPAREN);

		if (t_start_pos >= 0)
		{
			if (t_count > 0)
			{
				t_retval = FTokenMaster->CopyToCargo((char*)t_source.substr(t_start_pos, t_count).c_str());
			}
			else
			{
				t_retval = FTokenMaster->CopyToCargo( (char*)t_source.substr(t_start_pos).c_str() );
			}

		}
		else
		{
			t_retval = FTokenMaster->CopyToCargo("");
		}

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_substr", e.what());
		throw;
	}
	// return cargo as an int to TRunEdits::eval7_atom for appropriate processing
	// of functions that return strings
	return t_retval;

} // call_substr
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_trim()
{
	TToken* t_token;
	TTokenID t_token_id;

	intptr_t t_retval = 0;

	std::string t_source;
	try
	{
		check_open_paren();

		// The first parameter is the source string
		t_source.assign((char*)FRunEdits->EvaluateExpression());

		// Now see what comes next. We're expecting a COMMA
		t_token_id = FRunEdits->GetEvalTokenID();
		if (t_token_id != COMMA)
		{
			throw std::runtime_error("syntax error: expected comma between first and second parameters");
		}

		// Now we're expecting a NUMCONST to indicate how to trim t_source...
		// LEFT, RIGHT or BOTH
		t_token = FTokenMaster->GetCompiledToken();

		// The second parameter is the "trim flag"
		if (t_token->GetTokenType() == NUMCONST)
		{
			TTokenID t_token_id = t_token->GetTokenID();

			size_t t_start_pos,
				   t_end_pos;

			if (t_token_id == (TTokenID) '0') // token is 'Q'
			{
				// t_trim_flag = TRIMRIGHT;
				t_start_pos = 0;
				t_end_pos = t_source.length() -1;
				while ((t_end_pos > t_start_pos) &&
					   (t_source[t_end_pos] == ' '))
				{
					t_end_pos--;
				}

				// Store the result in cargo, to send it back to the caller
				t_retval = FTokenMaster->CopyToCargo((char*)t_source.substr(t_start_pos, (t_end_pos - t_start_pos)+1).c_str());
			}
			else if (t_token_id == (TTokenID) '1') // token is 'R'
			{
				// t_trim_flag = TRIMLEFT;
				t_start_pos = 0;
				t_end_pos = t_source.length();
				while ((t_source[t_start_pos] == ' ') &&
					   (t_start_pos < t_end_pos))
				{
					t_start_pos++;
				}

				// Store the result in cargo, to send it back to the caller
				t_retval = FTokenMaster->CopyToCargo((char*)t_source.substr(t_start_pos, t_end_pos).c_str());
			}
			else if (t_token_id == (TTokenID) '2') // token is 'S'
			{
				// t_trim_flag == TRIMBOTH
				t_start_pos = 0;
				t_end_pos = t_source.length();
				while ((t_source[t_start_pos] == ' ') &&
					   (t_start_pos < t_end_pos))
				{
					t_start_pos++;
				}
				t_end_pos = t_source.length() -1;
				while ((t_end_pos > t_start_pos) &&
					   (t_source[t_end_pos] == ' '))
				{
					t_end_pos--;
				}

				// Store the result in cargo, to send it back to the caller
				t_retval = FTokenMaster->CopyToCargo((char*)t_source.substr(t_start_pos, (t_end_pos - t_start_pos)+1).c_str());

			}
			else
			{
				throw std::runtime_error("syntax error: expected trim flag (RIGHT or LEFT)");
			}
		}
		else
		{
			throw std::runtime_error("syntax error: expected trim flag (RIGHT or LEFT)");
		}

		// That should be the end of the call to TRIM
		check_close_paren();
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_trim", e.what());
		throw;
	}
	// parameter cargo is actually a pointer to TRunEdits->FScratchBuffer buffer.
	// return it as an int to TRunEdits::eval7_atom for appropriate processing
	// of functions that return strings
	return t_retval;
} // call_trim
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_upper()
{
	// upper( <str> | <field> );
	// Converts <str> to uppercase.

	std::string t_source;

	intptr_t t_retval = 0;

	try
	{
		check_open_paren();

		t_source.assign((char*)FRunEdits->EvaluateExpression());

		uppercase(t_source);

		t_retval = FTokenMaster->CopyToCargo((char*)t_source.c_str());

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_upper", e.what());
		throw;
	}

	// parameter cargo is actually a pointer to TRunEdits->FScratchBuffer buffer.
	// return it as an int to TRunEdits::eval7_atom for appropriate processing
	// of functions that return strings
	return t_retval;

} // call_upper
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_val()
{
	// int val( <str> | <field> );
	// Returns the numeric value of an ASCII string.

	int t_result = 0;

	try
	{
		check_open_paren();

		/*  Note this odd (but documented) behavior of atol:
		 *  (from cplusplus.com)
		 *    The string can contain additional characters after those that form
		 *    the integral number, which are ignored and have no effect on the
		 *    behavior of this function.
		 *
		 *  Thus, std::atol("344-7349") returns a value of 344. This is how
		 *  it worked in EDITS40, so we won't take Herculean steps to change it.
		 *  It is documented in EditWriter v5 Help for this function.
		 */
		t_result = std::atol((char*)FRunEdits->EvaluateExpression());

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_val", e.what());
		throw;
	}

	return t_result;
} // call_val
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_valid_date_iop()
{
	// int valid_date_iop( <str> | <field> );
	// Checks date for validity

	int t_result = 0;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_result = (date_is_valid((char*)FRunEdits->EvaluateExpression()) ? 1 : 0);

		if (FDateErrorCode == DT_EMPTY)
		{
			// From EditWriter Help regarding this specific date function and
			// empty dates:
			//
			// "Note that if the input date is an empty string this function will
			//  return FALSE. In some circumstances it is valid for a date field
			//  to be empty. However, it is the responsibility of the author of
			//  the edit to write the code to test for EMPTY and take appropriate action."
			//
			// There are edits out in the field that depend upon this loopy
			// behavior, so we're slavishly following a questionable design
			// from prior versions of the Engine.
			t_result = 0;
		}

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_valid_date_iop", e.what());
		throw;
	}

	return t_result;

} // call_valid_date_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_date_day_iop()
{
	// int date_day_iop( date )
	// Returns day portion of date
	int t_result = 0;

	std::string t_date;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date.assign((char*)FRunEdits->EvaluateExpression());

		if (date_is_valid(t_date.c_str()))
		{
			if (FDateErrorCode == DT_EMPTY)
			{
				t_result = DT_EMPTY;
			}
			else if ((FDateErrorCode == DT_MISSING_DA) ||
					 (FDateErrorCode == DT_MISSING_MO))
			{
				t_result = DT_DAY_EMPTY;
			}
			else
			{
				t_result = std::atoi(t_date.substr(6,2).c_str());;
			}
		}
		else
		{
			t_result = DT_ERROR;
		}

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_date_day_iop", e.what());
		throw;
	}

	return t_result;
} // call_date_day_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_date_month_iop()
{
	// int date_month_iop( date )
	// Returns month portion of date
	int t_result = 0;

	std::string t_date;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date.assign((char*)FRunEdits->EvaluateExpression());

		if (date_is_valid(t_date.c_str()))
		{
			if (FDateErrorCode == DT_EMPTY)
			{
				t_result = DT_EMPTY;
			}
			else if (FDateErrorCode == DT_MISSING_MO)
			{
				t_result = DT_MONTH_EMPTY;
			}
			else
			{
				t_result = std::atoi(t_date.substr(4,2).c_str());
			}
		}
		else
		{
			t_result = DT_ERROR;
		}

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_date_month_iop", e.what());
		throw;
	}

	return t_result;
} // call_date_month_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_date_year_iop()
{
	// int date_year_iop( date )
	// Returns year portion of date
	int t_result = 0;
	std::string t_date;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date.assign((char*)FRunEdits->EvaluateExpression());

		if (date_is_valid(t_date.c_str()))
		{
			if (FDateErrorCode == DT_EMPTY)
			{
				t_result = DT_EMPTY;
			}
			else
			{
				t_result = std::atoi(t_date.substr(0,4).c_str());
			}
		}
		else
		{
			t_result = DT_ERROR;
		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_date_year_iop", e.what());
		throw;
	}

	return t_result;
} // call_date_year_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_datecmp_iop()
{
	// int datecmp_iop( date1, date2, min_max_flag)
	// Compares two dates by subtracting date2 from date1

	int t_result = 0;

	std::string t_date1, t_date2;
	intptr_t t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1.assign((char*)FRunEdits->EvaluateExpression());

		t_date2.assign((char*)FRunEdits->EvaluateExpression());

		t_min_max_flag = FRunEdits->EvaluateExpression();

		int t_date_integrity =
			check_date_integrity(t_date1.c_str(), t_min_max_flag);

		if (t_date_integrity != DT_VALID)
		{
			t_result = t_date_integrity;
		}
		else
		{
			t_date_integrity = check_date_integrity(t_date2.c_str(), t_min_max_flag);
			if (t_date_integrity != DT_VALID)
			{
				t_result = t_date_integrity;
			}
		}

		if (t_result == 0)
		{
			// If we make it to here, we have valid
			// date objects. Now perform the evaluation
			int t_start_date = complete_date(t_date1.c_str(), t_min_max_flag, t_date2.c_str(), true);

			int t_end_date = complete_date(t_date2.c_str(), t_min_max_flag, t_date1.c_str(), false);

			if (t_start_date == t_end_date)
				t_result = 0;
			else if (t_start_date > t_end_date)
				t_result = 1;
			else // (t_start_date < t_end_date)
				t_result = -1;

		 } // if (t_result == 0)

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TLibFuncs", "call_datecmp_iop", e.what());
		throw;
	}

	return t_result;
} // call_datecmp_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_daydiff_iop()
{
	// int daydiff_iop( date1, date2, min_max_flag)
	// finds the difference between two dates, counted in days
	int t_result = 0;

	std::string t_date1, t_date2;
	intptr_t t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1.assign((char*)FRunEdits->EvaluateExpression());

		t_date2.assign((char*)FRunEdits->EvaluateExpression());

		t_min_max_flag = (int)FRunEdits->EvaluateExpression();

		t_result = calculate_daydiff(t_date1.c_str(), t_date2.c_str(), t_min_max_flag);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		if (t_error.empty())
		{
			t_error.assign("Something bad has happened.");
		}
		FEngineErrorsPump->AddError("TLibFuncs", "call_daydiff_iop", t_error);
		throw;
	}

	return t_result;
} // call_daydiff_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_dayinterval_iop()
{
	// int dayinterval_iop(start_date, end_date,
	//     min_interval, max_interval, min_max_flag);
	// Checks to see if two dates fall within an interval, counted in days,
	// of each other.
	int t_result = 0;
	intptr_t t_date1, t_date2;
	intptr_t t_min_interval, t_max_interval, t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1 = FRunEdits->EvaluateExpression();

		t_date2 = FRunEdits->EvaluateExpression();

		t_min_interval = FRunEdits->EvaluateExpression();
		t_max_interval = FRunEdits->EvaluateExpression();
		t_min_max_flag = FRunEdits->EvaluateExpression();

		int t_days_between =
			calculate_daydiff((char*)t_date1, (char*)t_date2, t_min_max_flag);

		if ((t_days_between == DT_ERROR) ||
			(t_days_between == DT_EMPTY) ||
			(t_days_between == DT_UNKNOWN))
		{
			t_result = t_days_between;
		}
		else
		{
			// If t_days_between is a number, and that number falls within
			// the range specified, return TRUE
			if ((t_days_between >= (int)t_min_interval) &&
				(t_days_between <= (int)t_max_interval))
			{
				t_result = 1L;
			}
		}

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		if (t_error.empty())
		{
			t_error.assign("Something bad has happened.");
		}
		FEngineErrorsPump->AddError("TLibFuncs", "call_dayinterval_iop", t_error);
		throw;
	}

	return t_result;
} // call_dayinterval_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_monthdiff_iop()
{
	// int monthdiff_iop(start_date, end_date, min_max_flag);
	// Find the difference between two dates, counted in months.

	int t_result = 0;

	intptr_t t_date1, t_date2;
	intptr_t t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1 = FRunEdits->EvaluateExpression();

		t_date2 = FRunEdits->EvaluateExpression();

		t_min_max_flag = (int)FRunEdits->EvaluateExpression();

		t_result =
			calculate_monthdiff((char*)t_date1, (char*)t_date2, t_min_max_flag);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		if (t_error.empty())
		{
			t_error.assign("Something bad has happened.");
		}
		FEngineErrorsPump->AddError("TLibFuncs", "call_monthdiff_iop", t_error);
		throw;
	}

	return t_result;
} // call_monthdiff_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_monthinterval_iop()
{
	// int monthinterval_iop(start_date, end_date,
	//     min_interval, max_interval, min_max_flag);
	// Checks to see if two dates fall within an interval, counted in months,
	// of each other.
	int t_result = 0;

	intptr_t t_date1, t_date2;
	intptr_t t_min_interval, t_max_interval, t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1 = FRunEdits->EvaluateExpression();

		t_date2 = FRunEdits->EvaluateExpression();

		t_min_interval = FRunEdits->EvaluateExpression();
		t_max_interval = FRunEdits->EvaluateExpression();
		t_min_max_flag = FRunEdits->EvaluateExpression();

		int t_months_between =
			calculate_monthdiff((char*)t_date1, (char*)t_date2, t_min_max_flag);

		if ((t_months_between == DT_ERROR) ||
			(t_months_between == DT_EMPTY) ||
			(t_months_between == DT_UNKNOWN))
		{
			t_result = t_months_between;
		}
		else
		{
			// If t_months_between is a number, and that number falls within
			// the range specified, return TRUE
			if ((t_months_between >= (int)t_min_interval) &&
				(t_months_between <= (int)t_max_interval))
			{
				t_result = 1L;
			}
		}

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		if (t_error.empty())
		{
			t_error.assign("Something bad has happened.");
		}
		FEngineErrorsPump->AddError("TLibFuncs", "call_monthinterval_iop", t_error);
		throw;
	}

	return t_result;

} // call_monthinterval_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_yeardiff_iop()
{
	// int yeardiff_iop(start_date, end_date, min_max_flag);
	// Find the difference between two dates, counted in years.

	int t_result = 0;

	intptr_t t_date1, t_date2;
	intptr_t t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1 = FRunEdits->EvaluateExpression();

		t_date2 = FRunEdits->EvaluateExpression();

		t_min_max_flag = (int)FRunEdits->EvaluateExpression();

		t_result =
			calculate_yeardiff((char*)t_date1, (char*)t_date2, t_min_max_flag);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		if (t_error.empty())
		{
			t_error.assign("Something bad has happened.");
		}
		FEngineErrorsPump->AddError("TLibFuncs", "call_yeardiff_iop", t_error);
		throw;
	}

	return t_result;
} // call_yeardiff_iop
//---------------------------------------------------------------------------
intptr_t TLibFuncs::call_yearinterval_iop()
{
	// int yearinterval_iop(start_date, end_date,
	//     min_interval, max_interval, min_max_flag);
	// Checks to see if two dates fall within an interval, counted in years,
	// of each other.
	int t_result = 0;
	intptr_t t_date1, t_date2;
	intptr_t t_min_interval, t_max_interval, t_min_max_flag;

	ClearDateErrorCode();

	try
	{
		check_open_paren();

		t_date1 = FRunEdits->EvaluateExpression();

		t_date2 = FRunEdits->EvaluateExpression();

		t_min_interval = (int)FRunEdits->EvaluateExpression();
		t_max_interval = (int)FRunEdits->EvaluateExpression();
		t_min_max_flag = (int)FRunEdits->EvaluateExpression();

		int t_years_between =
			calculate_yeardiff((char*)t_date1, (char*)t_date2, t_min_max_flag);

		if ((t_years_between == DT_ERROR) ||
			(t_years_between == DT_EMPTY) ||
			(t_years_between == DT_UNKNOWN))
		{
			t_result = t_years_between;
		}
		else
		{
			// If t_years_between is a number, and that number falls within
			// the range specified, return TRUE
			if (((int)t_years_between >= t_min_interval) &&
				((int)t_years_between <= t_max_interval))
			{
				t_result = 1L;
			}
		}

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		if (t_error.empty())
		{
			t_error.assign("Something bad has happened.");
		}
		FEngineErrorsPump->AddError("TLibFuncs", "call_yearinterval_iop", t_error);
		throw;
	}

	return t_result;

} // call_yearinterval_iop
//---------------------------------------------------------------------------

