//---------------------------------------------------------------------------

#ifndef LibFuncsH
#define LibFuncsH
//---------------------------------------------------------------------------

#include <string>
#include <map>
#ifdef EE_WINDOWS
#include <regex>
#else
// std::regex not implemented until v4.9 of gcc; debian distributes v4.7
#include <boost/regex.hpp>
#endif

// forward declarations
class TLibFuncs;
class TRunEdits;
class TEngineErrorsPump;
class TSQLDB;
class TFastSQLQuery;
class TTablesList;
class TEditErrors;
class TTokenMaster;

// Pointer to EDITS Language functions
typedef intptr_t (TLibFuncs::*FCallFuncPtr)();

// date minmax flags
#define DT_MIN 	  				0x0001 // 0x0052   // Dec 82, char 'R' was: 0x0001
#define DT_MAX 	  				0x0002 // 0x0053   // Dec 83, char 'S' was: 0x0002
#define DT_EXACT	  			0x0003 // 0x0054   // Dec 84, char 'T' was: 0x0003

// These first four are returned from VALID_DATE. The "high sentinel value"
// makes it possible for the debugger to report a meaningful value to
// the user (edit author).

#define DT_VALID				0x4FFFFFFF // was: 0
#define DT_MISSING		 		0x5FFFFFFF // was: 1L
#define DT_ERROR				0x6FFFFFFF // was: 2L
#define DT_UNKNOWN				0x7FFFFFFF // no change

// DT_EMPTY is new with EDITS40. It is a return value from date IOP functions
#define DT_EMPTY				0x3FFFFFFF
#define DT_DAY_EMPTY            0x2FFFFFFF // call_date_day_iop
#define DT_MONTH_EMPTY          0x1FFFFFFF // call_date_month_iop

#define DT_ERROR_DA		 		4L
#define DT_ERROR_MO		 		8L
#define DT_ERROR_YR	 			16L
#define DT_MISSING_DA  			32L
#define DT_MISSING_MO  			64L
#define DT_MISSING_YR 			128L
#define DT_UNKNOWN_DA 			256L
#define DT_UNKNOWN_MO 			512L
#define DT_UNKNOWN_YR 			1024L
//#define DT_MUST_BE_D1           2048L  // EDITS40
#define DT_EMPTY_STRING         4096L

// forward declaration
class TTableDef;

// For the FInternal?LookupQueries maps
class TLookupQuery
{
private:
	TFastSQLQuery* FQry;
	TTableDef* FTableDef;
	std::vector<int> FColumnOrderList;
	// For concatenated-index fields, called from RLOOKUP
	int FPadTo;
public:
	TLookupQuery(TSQLDB* database, TTableDef* table_def);
	~TLookupQuery();

	TFastSQLQuery* GetQuery();
	TTableDef* GetTableDef();
	int GetColumnsCount();
	int GetColumnOrder(int idx);
	void AddColumnOrder(int column_order);

	void SetPadTo(int pad_size);
	int GetPadTo();
};

class TLibFuncs
{
private:
	TRunEdits* FRunEdits;
	TTokenMaster* FTokenMaster;
	TEngineErrorsPump* FEngineErrorsPump;
	TSQLDB* FSQLDB;
	TTablesList* FTablesList;
	TEditErrors* FEditErrors;

    // set to 0 at start of an Edit run. Can be reset via call_allow_future_date_iop
	size_t FYearsIntoFuture;

	// Holds the DT_ values defined above
	int FDateErrorCode;

	FCallFuncPtr FFuncPointers[256];

	// parameterized queries used by internal_lookup, internal_rlookup
	std::map< std::string, TLookupQuery* > FInternalLookupQueries;
	std::map< std::string, TLookupQuery* > FInternalRLookupQueries;
	std::map< std::string, TLookupQuery* > FInternalSQLLookupQueries;
	std::map< std::string, TLookupQuery* > FInternalSQLRLookupQueries;

	// The never-ending quest to squeak out improved performance. AQTime
	// reports that the call to create a std::regex object is costly. Let's
	// try doing that only once for each unique pattern, and then re-using
	// them for all subsequent case records that need them.

#ifdef EE_WINDOWS
#ifdef BCB
	// CodeGear uses std::tr1
	std::map< std::string, std::tr1::regex> FRegexPatterns;
	std::map< std::string, std::tr1::regex>::iterator FRegexIterator;
	std::tr1::regex FTheRegexPattern;
#else // MinGW
	// MinGW uses std::regex, but it's a stub until v4.9... you'll still
	// need Boost
	std::map< std::string, std::regex> FRegexPatterns;
	std::map< std::string, std::regex>::iterator FRegexIterator;
	std::regex FTheRegexPattern;
#endif // BCB or MinGW
#else
	// g++ 4.7 requires Boost; 4.9 will implement std::regex
	std::map< std::string, boost::regex> FRegexPatterns;
	std::map< std::string, boost::regex>::iterator FRegexIterator;
	boost::regex FTheRegexPattern;
#endif // EE_WINDOWS

	// for MATCH and INLIST
	std::string translate_mask_to_regex(std::string mask);

	int FSystemDate;

	void setup_parameters_buff(size_t buff_size);
	void setup_function_pointers();

	void check_open_paren();
	void check_close_paren();

	int date_is_valid(const char* the_date);
	bool check_date_format(const char* the_date);

	// the "diff" and "interval" functions share these
	int calculate_daydiff(const char* start_date, const char* end_date, intptr_t min_max_flag);
	int calculate_monthdiff(const char* start_date, const char* end_date, intptr_t min_max_flag);
	int calculate_yeardiff(const char* start_date, const char* end_date, intptr_t min_max_flag);

	// These functions are used by EVERY function that takes a MinMaxFlag parameter
	// call_datecmp_iop
	// call_daydiff_iop, call_monthdiff_iop, call_yeardiff_iop
	// call_dayinterval_iop, call_monthinterval_iop, call_yearinterval_iop
	int check_date_integrity(const char* the_date, intptr_t min_max_flag);
	void parse_date(const char* the_date, int& the_year, int& the_month, int& the_day);
	int complete_date(const char* the_date, intptr_t min_max_flag, const char* the_other_date, bool is_first_date);

	// for the "diff" and "interval" functions
	bool is_leap_year(int the_year);
	int get_day_of_year(int the_date);

	// for LOOKUP and ILOOKUP
	int internal_lookup(const char* lookup_value, const char* table_specs, bool populate_tablevars);
	// for RLOOKUP (range lookup)
	int internal_rlookup(const char* lookup_value, const char* table_specs);
	int get_trigger_padded_size(const char* table_name, const char* column_name);

	// for SQLLookup
	int internal_sqllookup(const char* table_name, const char* columns_list, const char* lookup_value, bool populate_tablevars);
	// for SQLRLookup (the range-lookup version)
	int internal_sqlrlookup(const char* table_name, const char* columns_list, const char* lookup_value);

	void find_regex_pattern(const char* match_mask);
public:
	TLibFuncs(TRunEdits* caller, TEngineErrorsPump* pump, TSQLDB* database,
		TTablesList* tables_list, TEditErrors* edit_errors_manager,
		TTokenMaster* token_master);
	~TLibFuncs();

	//size_t Count() { return FFuncPointersCount; };
	FCallFuncPtr GetFuncPointer(char token);

	int GetDateErrorCode();
	void ClearDateErrorCode();

	void SetSystemDate();
	int GetSystemDate();
	void SetYearsIntoFuture();

	// EDITS Language functions
	intptr_t call_allow_future_date_iop();
	intptr_t call_at();
	intptr_t call_empty();
	intptr_t call_error_msg();
	intptr_t call_error_text();
	intptr_t call_externaldll();
	intptr_t call_fmtstr();
	intptr_t call_ilookup();
	intptr_t call_inlist();
	intptr_t call_justified();
	intptr_t call_left();
	intptr_t call_lookup();
	intptr_t call_lower();
	intptr_t call_match();
	intptr_t call_nameexpr();
	intptr_t call_right();
	intptr_t call_rlookup();
	intptr_t call_sqllookup();
	intptr_t call_sqlrangelookup();
	intptr_t call_save_error_text();
	intptr_t call_save_text();
	intptr_t call_strcat();
	intptr_t call_strcmp();
	intptr_t call_strcpy();
	intptr_t call_strlen();
	intptr_t call_trim();
	intptr_t call_val();
	intptr_t call_substr();
	intptr_t call_upper();
	intptr_t call_set_warning();
	intptr_t call_save_warning_text();
	intptr_t call_date_day_iop();
	intptr_t call_date_month_iop();
	intptr_t call_date_year_iop();
	intptr_t call_datecmp_iop();
	intptr_t call_daydiff_iop();
	intptr_t call_dayinterval_iop();
	intptr_t call_monthdiff_iop();
	intptr_t call_monthinterval_iop();
	intptr_t call_valid_date_iop() ;
	intptr_t call_yeardiff_iop();
	intptr_t call_yearinterval_iop();
	/* We are removing these; most were deprecated in EDITS40.dll
	int call_datecmp(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_date_day(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_date_month(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_date_year(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_daydiff(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_dayinterval(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_monthdiff(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_monthinterval(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_nameexpr(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_valid_date(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_yeardiff(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_yearinterval(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_getfield(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_putfield(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_getvar(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_setvar(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_usr2(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_usr4(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_usr(int i);
	int call_usr1() { return call_usr(0);}
	int call_usr2() { return call_usr(1);}
	int call_usr3() { return call_usr(2);}
	int call_usr4() { return call_usr(3);}
	int call_usr5() { return call_usr(4);}
	int call_usr6() { return call_usr(5);}
	int call_usr7() { return call_usr(6);}
	int call_usr8() { return call_usr(7);}
	int call_usr9() { return call_usr(8);}
	int call_usr10() { return call_usr(9);}
	int call_binlookup(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_copyrecord(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_ftranslate(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_getche(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_getnum(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_increment_flag(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_putch(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_puts(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_print(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_quit(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_read(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_set_error(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_show(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_trace(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_translate(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_use(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_set_flag(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_check_flag(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_clear_all_flags(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };
	int call_recordempty(const std::string compiled, size_t& pos, std::string& cargo) { return 0; };

	*/

};
#endif
