//---------------------------------------------------------------------------
#include <memory>    // smart pointers
#include <algorithm> // std::transform
#include <stdexcept>
#include <cstdlib>  // std::atoi
#include <cstdio>   // std::snprintf
#include <cstring>  // std::strncpy
#include <climits>  // INT_MAX

#include "RunEdits.h"
#include "SQLDB.h"        // for TSQLDB
#include "FastSQLQuery.h" // for faster queries (no "get field by name" capability)
#include "Utility.h"
#include "EngineErrorsPump.h"
#include "Layout.h"
#include "EditSet.h"
#include "Variable.h"
#include "LibFuncs.h"
#include "EditErrors.h"
#include "DebugEdit.h"

//---------------------------------------------------------------------------
TRunEdits::TRunEdits(TSQLDB* database, TEngineErrorsPump* pump, TTablesList* tables_list)
{
	FSQLDB = database;
	FEditDetailsQry = NULL;
	FEditLogicQry = NULL;

	FEngineErrorsPump = pump;
	FTablesList = tables_list;

	FUDFPointers.reserve(20);
	FDebugLoopTopLineNumber.reserve(100); // unlikely, but let's been generous
	FDebugLoopIteration.reserve(100);
	FLayoutEditSetPairs.reserve(100);

	FEditErrors = new TEditErrors(FSQLDB, pump);
	FTokenMaster = new TTokenMaster(pump);

	FLibFuncs = new TLibFuncs(this, pump, database, tables_list,
		FEditErrors, FTokenMaster);

	// Oh, the incestuousness!! TRunEdits, TLibFuncs and TEditErrors all need
	// to be able to talk to one another. Let's establish the connection
	// between the last two.
	FEditErrors->SetLibFuncs(FLibFuncs);

	// Default the TestBench debugger vars to NULL
	FDebugCallbackOwner = NULL;
	FDebugLogicCallback = NULL;
	FDebugFieldsCallback = NULL;
	FDebugNotesCallback = NULL;
	FDebugVarsCallback = NULL;

	FDebugEditObject = NULL;

	FIsDebugging = false;

	// see declare_variables
	FVariablesVectorSize = 0;
	FTopOfVariablesStack = 0;

} // constructor
//---------------------------------------------------------------------------
TRunEdits::~TRunEdits()
{
	delete FLibFuncs;

	clear_variables();

	// Delete the TLayout objects created in this instance, if any
	std::map<std::string, TLayout*>::iterator it;

	for (it = FLayouts.begin(); it != FLayouts.end(); it++)
	{
		delete (*it).second;
	}

	// Delete the TLayoutEditSetPair objects
	for (size_t t_idx = 0; t_idx < FLayoutEditSetPairs.size(); t_idx++)
	{
		// This is where we delete the TEditSet objects. Remember,
		// more than one instance of an edit set may be in memory,
		// each associated with a different TLayout object. The
		// Layouts are created only once, tracked in map FLayouts,
		// and deleted in the loop just above this one.
		TLayoutEditSetPair* t_pair = FLayoutEditSetPairs[t_idx];
		delete t_pair->EditSet;
		delete t_pair;
	}

	if (FTokenMaster)
	{
		delete FTokenMaster;
		FTokenMaster = NULL;
	}

	if (FEditErrors)
	{
		delete FEditErrors;
		FEditErrors = NULL;
	}

	if (FEditLogicQry)
	{
		FEditLogicQry->ReleaseParameterizedQuery();
		delete FEditLogicQry;
	}

	if (FEditDetailsQry)
	{
		FEditDetailsQry->ReleaseParameterizedQuery();
		delete FEditDetailsQry;
	}
} // destructor
//---------------------------------------------------------------------------
int TRunEdits::UpdateRecord(std::string layout_tag, std::string data)
{
	int t_result = 0;
	TLayout* t_layout = NULL;

	// standardize the case of the layout_tag; it could be alphanumeric
	uppercase(layout_tag);

	// If we have already instantiated this layout, go get it.
	TLayout* t_test = FLayouts[layout_tag];
	if (t_test)
	{
		t_layout = t_test;
		if (t_layout->LoadData(data) == 0)
		{
			// That may be all we have to do from here
		}
		else
		{
			t_result = -1;
		}
	}
	else
	{
		t_layout = new TLayout(FSQLDB, FEngineErrorsPump);
		try
		{
			if (t_layout->LoadLayout(layout_tag) == 0)
			{
				// Add it to the list of layouts
				FLayouts[layout_tag] = t_layout;

				t_layout->LoadData(data);
			}
			else
			{
				delete t_layout;
				t_layout = NULL;
				// TLayout::LoadLayout will have told the errors pump
				// what the problem is; just return.
				t_result = -1;
			}

		}
		catch(std::runtime_error& e)
		{
			std::string t_what = e.what();
			if (!t_what.empty())
			{
				t_what.assign("Failure in TRunEdits::UpdateRecord");
			}
			FEngineErrorsPump->AddError("TRunEdits","UpdateRecord", t_what);

			delete t_layout;
			t_layout = NULL;

			t_result = -1;
		}
	}
	return t_result;
} // UpdateRecord
//---------------------------------------------------------------------------
int TRunEdits::DisplayDataRecord( std::string layout_tag,
	void* owner, void* callback_func )
{
	int t_result = 0;

	// standardize the case of the layout_tag; it could be alphanumeric
	uppercase(layout_tag);

	TLayout* t_layout = FLayouts[layout_tag];
	if (t_layout)
	{
		t_layout->DisplayDataRecord(owner, callback_func);
	}
	else
	{
		char t_buffer[500];
		std::snprintf(t_buffer, 500,
			"Cannot find record layout with tag \"%s\"",
			layout_tag.c_str());
		FEngineErrorsPump->AddError("TRunEdits", "DisplayDataRecord", t_buffer);
		t_result = -1;
	}

	return t_result;
} // DisplayDataRecord
//---------------------------------------------------------------------------
int TRunEdits::DisplayFieldInMessagesAs( const int field_number,
	std::string display_name )
{
	TLayout* t_layout;
	int t_result = 0;

	if (!FLayouts.empty())
	{
		// Iterate through the loaded layouts and for each one,
		// update the DisplayName of the field designated by field_number
		std::map<std::string, TLayout*>::iterator it;
		for (it = FLayouts.begin(); it != FLayouts.end(); it++)
		{
			t_layout = (*it).second;

			t_result = t_layout->DisplayFieldInMessagesAs( field_number, display_name );
		}
	}
	else
	{
		std::string t_error = "No layouts in memory. Call Edit_UpdateRecordByTag before calling Edit_DisplayFieldInMessagesAs.";
		FEngineErrorsPump->AddError("TRunEdits", "DisplayFieldInMessagesAs", t_error);
		t_result = -1;
	}

	return t_result;
} // DisplayFieldInMessagesAs
//---------------------------------------------------------------------------
int TRunEdits::NeverSkipEmptyField( const int field_number,
	const int never_skip_empty )
{
	TLayout* t_layout;
	int t_result = 0;

	if (!FLayouts.empty())
	{
		// Iterate through the loaded layouts and for each one,
		// update the NeverSkipEmpty of the field designated by field_number
		std::map<std::string, TLayout*>::iterator it;
		for (it = FLayouts.begin(); it != FLayouts.end(); it++)
		{
			t_layout = (*it).second;

			t_result = t_layout->NeverSkipEmptyField( field_number, never_skip_empty );
		}
	}
	else
	{
		std::string t_error = "No layouts in memory. Call Edit_UpdateRecordByTag before calling Edit_NeverSkipEmptyField.";
		FEngineErrorsPump->AddError("TRunEdits", "NeverSkipEmptyField", t_error);
		t_result = -1;
	}

	return t_result;
} // NeverSkipEmptyField
//---------------------------------------------------------------------------
int TRunEdits::FieldDataByNumber(
		 const char* layout_tag,
		 const int field_number,
		 int* pkey,
		 int* start_pos,
		 int* field_length,
		 char* field_name,
		 const int field_name_len,
		 char* display_name,
		 const int display_name_len,
		 char* data_type,
		 const int data_type_len,
		 char* value, const int value_len )
{
	int t_result = 0;

	// standardize layout_tag (could be mixed case)
	std::string t_layout_tag = layout_tag;
	uppercase(t_layout_tag);

	TLayout* t_layout = FLayouts[t_layout_tag];
	if (t_layout)
	{
		t_result = t_layout->FieldDataByNumber(
			field_number,
			pkey, start_pos, field_length,
			field_name, field_name_len,
			display_name, display_name_len,
			data_type, data_type_len,
			value, value_len);
	}
	else
	{
		char t_buffer[500];
		std::snprintf(t_buffer, 500,
			"Cannot find record layout with tag \"%s\"",
			layout_tag);
		FEngineErrorsPump->AddError("TRunEdits", "FieldDataByNumber", t_buffer);
		t_result = -1;
	}
	return t_result;
} // FieldDataByNumber
//---------------------------------------------------------------------------
TLayoutEditSetPair* TRunEdits::find_layout_editset_pair( const std::string layout_tag,
	const std::string edit_set_tag)
{
	TLayoutEditSetPair* t_pair = NULL,
					  * t_test_pair;

	std::string t_edit_set_tag = edit_set_tag;
	std::string t_layout_tag = layout_tag;

	uppercase(t_edit_set_tag);
	uppercase(t_layout_tag);

	TLayout* t_layout = FLayouts[t_layout_tag];

	// If we have these two objects, are they already paired for
	// running edits?
	if (t_layout)
	{
		std::string t_compare_tag;
		for (size_t t_idx = 0; t_idx < FLayoutEditSetPairs.size(); t_idx++)
		{
			t_test_pair = FLayoutEditSetPairs[t_idx];
			t_compare_tag = t_test_pair->EditSet->GetEditSetTag();
			if ((t_test_pair->Layout == t_layout) &&
				(t_compare_tag.compare(t_edit_set_tag) == 0))
			{
				t_pair = t_test_pair;
				break;
			}
		}
	}
	return t_pair;
} // find_layout_editset_pair
//---------------------------------------------------------------------------
TEditSet* TRunEdits::find_editset( const std::string edit_set_tag,
	const size_t start_idx, bool& more )
{
	TLayoutEditSetPair* t_pair;
	TEditSet* t_edit_set = NULL;

	std::string t_compare_tag;
	std::string t_edit_set_tag = edit_set_tag;
	uppercase(t_edit_set_tag);
	size_t t_idx = start_idx;

	for (; t_idx < FLayoutEditSetPairs.size(); t_idx++)
	{
		t_pair = FLayoutEditSetPairs[t_idx];
		t_compare_tag = t_pair->EditSet->GetEditSetTag();
		if (t_compare_tag.compare(t_edit_set_tag) == 0)
		{
			t_edit_set = t_pair->EditSet;
			break;
		}
	}
	// There could be more pairs than the one we found... if we found one.
	// If we DID find one, and more edit set/layout pairs exist, tell the
	// caller to come back and get them all. The caller will probably be
	// TRunEdits::SetEditOptions
	more = (t_edit_set && (t_idx+1 < FLayoutEditSetPairs.size()));

	return t_edit_set;
} // find_editset
//---------------------------------------------------------------------------

int TRunEdits::LoadEditSet( const std::string edit_set_tag,
	const std::string layout_tag )
{
	int t_result = 0;

	std::string t_edit_set_tag = edit_set_tag;
	std::string t_layout_tag = layout_tag;

	uppercase(t_edit_set_tag);
	uppercase(t_layout_tag);

	TLayoutEditSetPair* t_pair = find_layout_editset_pair( t_layout_tag, t_edit_set_tag );

	if (!t_pair)
	{
		TLayout* t_layout = FLayouts[t_layout_tag];
		if (!t_layout)
		{
			t_layout = new TLayout(FSQLDB, FEngineErrorsPump);
			if (t_layout->LoadLayout(t_layout_tag) == 0)
			{
				// Add it to the list of layouts
				FLayouts[t_layout_tag] = t_layout;
			}
			else
			{
				delete t_layout;
				t_layout = NULL;
				// TLayout::LoadLayout will have told the errors pump
				// what the problem is; just return.
				t_result = -1;
			}
		} // if (!t_layout)

		if (t_layout)
		{
			TEditSet* t_edit_set = prepare_editset(t_edit_set_tag, t_layout_tag);
			if (!t_edit_set)
			{
				char t_buffer[1000];
				std::snprintf(t_buffer, 1000,
					"Failure to load edit set tag \"%s\" with layout tag \"%s\";",
					edit_set_tag.c_str(), layout_tag.c_str());
				FEngineErrorsPump->AddError("TRunEdits", "LoadEditSet", t_buffer);
				t_result = -1;
			}
		}
	}

	return t_result;
} // LoadEditSet
//---------------------------------------------------------------------------

int TRunEdits::UnloadEditSet( const std::string edit_set_tag,
	const std::string layout_tag )
{
	int t_result = 0;

	std::string t_edit_set_tag = edit_set_tag;
	std::string t_layout_tag = layout_tag;

	uppercase(t_edit_set_tag);
	uppercase(t_layout_tag);

	TLayout* t_layout = FLayouts[t_layout_tag];

	// Is the layout already loaded? If not, we can stop now.
	if (t_layout)
	{
		TLayoutEditSetPair* t_test_pair;
		std::string t_compare_tag;

		for (size_t t_idx = 0; t_idx < FLayoutEditSetPairs.size(); t_idx++)
		{
			t_test_pair = FLayoutEditSetPairs[t_idx];
			t_compare_tag = t_test_pair->EditSet->GetEditSetTag();

			if ((t_test_pair->Layout == t_layout) &&
				(t_compare_tag.compare(t_edit_set_tag) == 0))
			{
				// Found it! Release the edit set and remove the pair from the list
				delete t_test_pair->EditSet;
				delete t_test_pair;

				FLayoutEditSetPairs.erase(FLayoutEditSetPairs.begin() + t_idx);
				break;
			}
		}

		// We also have to delete the layout so that it will be re-loaded
		// and evaluated when needed.
		delete t_layout;

		FLayouts.erase(t_layout_tag);
	}

	return t_result;
} // UnloadEditSet
//---------------------------------------------------------------------------

int TRunEdits::SetEditOptions( const std::string edit_set_tag,
	const std::string edit_tag,
	const int suppress_edit,
	const int skip_if_empty,
	const int skip_if_failed,
	const int no_skips_override )
{
	int t_result = 0;
	size_t t_start_idx = 0;
	bool t_more_edit_sets = true;


	while (t_more_edit_sets)
	{
		TEditSet* t_edit_set = find_editset( edit_set_tag, t_start_idx, t_more_edit_sets );

		if (t_edit_set)
		{
			t_result = t_edit_set->SetEditOptions(edit_tag,
				(suppress_edit ? true : false),
				(skip_if_empty ? true : false),
				(skip_if_failed ? true : false),
				(no_skips_override ? true : false));
		}
		else
		{
			char t_buffer[1000];
			std::snprintf(t_buffer, 1000,
				"Edit set with tag \"%s\" has not been initialized. Run Edit_LoadEditSet first.",
				edit_set_tag.c_str());
			FEngineErrorsPump->AddError("TRunEdits", "SetEditOptions", t_buffer);
			t_result = -1;

			break;
		}
		t_start_idx += 1;
	}
	return t_result;
} // SetEditOptions
//---------------------------------------------------------------------------

int TRunEdits::GetEditOptions( const std::string edit_set_tag,
	const std::string edit_tag,
	int* is_suppressed,
	int* is_skip_if_empty,
	int* is_skip_if_failed,
	int* is_no_skips_override )
{
	int t_result = 0;

	// Taking a chance here... we'll assume that SetEditOptions correctly
	// set the options on ALL edit set/layout pairs. Therefore, any we
	// select will tell us the current values of edit options.
	// Also therefore, t_start_pos and t_more_edit_sets are throwaway
	// values when the call is from this function.
	size_t t_start_pos = 0;
	bool t_more_edit_sets = true;
	TEditSet* t_edit_set = find_editset( edit_set_tag, t_start_pos, t_more_edit_sets );

	if (t_edit_set)
	{
		t_result = t_edit_set->GetEditOptions(edit_tag, is_suppressed,
			is_skip_if_empty, is_skip_if_failed, is_no_skips_override);
	}
	else
	{
		char t_buffer[1000];
		std::snprintf(t_buffer, 1000,
			"Edit set with tag \"%s\" has not been initialized. Run Edit_LoadEditSet first.",
			edit_set_tag.c_str());
		FEngineErrorsPump->AddError("TRunEdits", "GetEditOptions", t_buffer);
		t_result = -1;
	}

	return t_result;
} // GetEditOptions
//---------------------------------------------------------------------------
int TRunEdits::ClearEditOptions( const std::string edit_set_tag )
{
	int t_result = 0;

	// Taking a chance here... we'll assume that SetEditOptions correctly
	// set the options on ALL edit set/layout pairs. Therefore, any we
	// select will tell us the current values of edit options.
	// Also therefore, t_start_pos and t_more_edit_sets are throwaway
	// values when the call is from this function.
	size_t t_start_pos = 0;
	bool t_more_edit_sets = true;
	TEditSet* t_edit_set = find_editset( edit_set_tag, t_start_pos, t_more_edit_sets );

	if (t_edit_set)
	{
		t_result = t_edit_set->ClearEditOptions();
	}
	else
	{
		char t_buffer[1000];
		std::snprintf(t_buffer, 1000,
			"Edit set with tag \"%s\" has not been initialized. Run Edit_LoadEditSet first.",
			edit_set_tag.c_str());
		FEngineErrorsPump->AddError("TRunEdits", "ClearEditOptions", t_buffer);
		t_result = -1;
	}

	return t_result;

} // ClearEditOptions
//---------------------------------------------------------------------------

void TRunEdits::clear_variables()
{
	for (size_t t_idx = 0; t_idx < FVariablesVectorSize; t_idx++)
	{
		TVariable* t_var = FVariables[t_idx];
		if (t_var)
		{
			delete t_var;
		}
	}

	FVariables.resize(0);
} // clear_variables
//---------------------------------------------------------------------------
TToken* TRunEdits::GetCompiledToken()
{
	TToken* t_token;
	// Public function to allow calls from TLibFuncs class. Also, it simplifies
	// calls from this class.
	try
	{
		t_token = FTokenMaster->GetCompiledToken();
	}
	catch (...)
	{
		// TTokenMaster::GetCompiledToken will have added its own
		// message to the message pump. But we want to include the
		// name of the running edit.
		std::string t_err = "Edit: ";
		t_err += FRunningEdit->GetEditName();
		FEngineErrorsPump->AddError("TRunEdits", "GetCompiledToken", t_err);

		// now throw the error up to the caller
		throw;
	}

	return t_token;
} // GetCompiledToken
//---------------------------------------------------------------------------
TToken* TRunEdits::GetHardToken( const TTokenID token_id )
{
	// Public function to allow calls from TLibFuncs class. Also, it simplifies
	// calls from this class.
	return FTokenMaster->GetHardToken(token_id);
} // GetHardToken
//---------------------------------------------------------------------------
int TRunEdits::RunEdits(const std::string edit_set_tag, const std::string layout_tag,
	const std::string data, const int edit_options, int* errors_count,
	void* owner, void* callback_func)
{
	if (!create_edit_logic_query())
	{
		return -1;
	}
	// For each new case, clear the count of errors to 0 before running edits
	// We'll do this regardless of what else may happen here, so that the caller
	// does not end up with a bogus value "left over" from a previous run.
	*errors_count = 0;

	// First, let's update the layout with the new data record
	// Reminder: UpdateRecord will create a TLayout object for this
	// layout_tag if one does not already exist.
	int t_result = UpdateRecord(layout_tag, data);

	if (t_result == 0)
	{
		// Next, find out if this edit set and layout have already
		// been paired... that is, make sure that all of the fields
		// referenced by edits in the edit set are represented in
		// the layout. (Or those edits are flagged to skip if any
		// referenced fields are NOT in the layout.)
		TLayoutEditSetPair* t_pair =
			find_layout_editset_pair( layout_tag, edit_set_tag);

		if (!t_pair)
		{
			// If the pair does not already exist, try to create it
			// with the information at hand.
			std::string t_layout_tag = layout_tag;
			uppercase(t_layout_tag);

			std::string t_edit_set_tag = edit_set_tag;
			uppercase(t_edit_set_tag);

			// Next, prepare the edit set: Run the evaluation
			// of edits with fields not in layout (if it has not
			// already been done by a previous call to this
			// function using these same tag values).
			TEditSet* t_edit_set =
				prepare_editset(t_edit_set_tag, t_layout_tag);

			if (t_edit_set)
			{
				// Note: If prepare_editset returns a pointer to an edit set,
				// it was able to set up a new TLayoutEditSetPair
				// So NOW get the pair
				t_pair = find_layout_editset_pair( layout_tag, edit_set_tag);
			}
			else
			{
				char t_buffer[1000];
				std::snprintf(t_buffer, 1000,
					"Cannot find edit set tag \"%s\" in this metafile",
					edit_set_tag.c_str());
				FEngineErrorsPump->AddError("TRunEdits", "RunEdits", t_buffer);

				return -1;
			}
		}

		if (t_pair)
		{
			FRunningEditSet = t_pair->EditSet;
			FRunningLayout  = t_pair->Layout;
			FRunningEdit    = NULL;

			try
			{
                // Set the "suppress warnings" option
				FRunningEditSet->SetSuppressWarnings(
					(edit_options & EE_SUPPRESSWARN ? 1 : 0));


				// Clear the errors, warnings and skips from a previous run.
				FRunningEditSet->ClearCounts();

				// We also clear out/re-initialize FEditErrors manager.
				FEditErrors->Initialize(owner, callback_func, FRunningLayout,
					FRunningEditSet->GetSuppressWarnings());


				// Make sure our system date is current
				FLibFuncs->SetSystemDate();

				//  Iterate the edits in FRunningEditSet, retrieving the
				// CompiledLogic and processing it.
				int t_count = FRunningEditSet->GetEditsCount();

				std::string t_compiled;
				int t_fields_count,
					t_field_number,
					t_fld_idx;

				for (int t_idx = 0; t_idx < t_count; t_idx++)
				{
					FRunningEdit = FRunningEditSet->GetEditByRunOrderIndex(t_idx);

					if (FRunningEdit->SkipEditSilently())
					{
						// If the edit references fields that are not in the
						// currently-loaded layout, we don't want to run it.
						// Skip silently (i.e., don't send up a flare)
						continue;
					}

					if (FRunningEdit->IsSuppressed())
					{
						// The user doesn't want to see this edit...
						// move along!
						continue;
					}

					// If the user did not specify that the edit should not skip
					// for empty or failed conditions...
					if (!FRunningEdit->IsNoSkipsOverride())
					{
						bool t_skip_edit = false;
						std::string t_value;
						t_fields_count = FRunningEdit->GetFieldsCount();
						// ... then see if the SKIP EMPTY option was set at
						// EITHER the edit or edit set level.
						if (FRunningEdit->IsSkipIfEmpty() ||
							edit_options & EE_SKIPEMPTY)
						{
							for (t_fld_idx = 0; t_fld_idx < t_fields_count; t_fld_idx++)
							{
								t_field_number = FRunningEdit->GetFieldByMacroOrder(t_fld_idx);
								if (t_field_number > 0)
								{
									TLayoutField* t_layout_field =
										FRunningLayout->GetFieldByFieldNumber(t_field_number);

									if (t_layout_field)
									{
										// If the caller invoked Edit_NeverSkipEmptyField
										// and set the value to true for this field, then
										// do NOT consider it when evaluating whether
										// the edit should skip on empty.
										if (!t_layout_field->GetNeverSkipEmpty())
										{
											t_value.assign(t_layout_field->GetValue());
											if (t_value.find_first_not_of(" ",0) == std::string::npos)
											{
												// Field value is all spaces (char)32;
												t_skip_edit = true;
												break;
											}
										}
									}
								} // if (t_field_number > 0)

							}

							if (t_skip_edit)
							{
								// Increment the counter that is tracking
								// how many times this edit was skipped owing
								// to an empty field value
								FRunningEdit->IncrSkipsCount();
								continue;
							}
						}

						if (FRunningEdit->IsSkipIfFailed() ||
						    edit_options & EE_SKIPFAIL)
						{
							for (t_fld_idx = 0; t_fld_idx < t_fields_count; t_fld_idx++)
							{
								t_field_number = FRunningEdit->GetFieldByMacroOrder(t_fld_idx);
								if (t_field_number > 0)
								{
									TLayoutField* t_layout_field =
										FRunningLayout->GetFieldByFieldNumber(t_field_number);
									if (t_layout_field->GetFailedSingleFieldEdit())
									{
										// Field failed SF edit for this case
										t_skip_edit = true;
										break;
									}
								} // if (t_field_number > 0)

							}

							if (t_skip_edit)
							{
								// Increment the counter that is tracking
								// how many times this edit was skipped owing
								// to a field value that failed its SF edit
								FRunningEdit->IncrSkipsCount();
								continue;
							}
						}
					}

					try
					{
						t_compiled.assign(FRunningEdit->GetCompiledLogic(FEditLogicQry));
						// CompiledLogic could be empty if the EditLogic failed
						// to compile. Just skip silently; there are other ways
						// to notify the caller that a compile has failed.
						if (!t_compiled.empty())
						{

							FTokenMaster->SetCompiledLogic(t_compiled.c_str());

							process_edit();
						}
					}
					catch (...)
					{
						// Let's not abort the entire edit set. The routine
						// that threw this error will have written to the
						// engine errors pump. Just note that t_result is -1
						// and move on to the next edit.

						FRunningEdit = NULL;
						// Let the caller know there are errors to be retrieved
						t_result = -1;

						// I AM exiting the loop
						break;
					}
				} // for (t_idx... iterating edits)

			}
			catch (...)
			{
				t_result = -1;
			}

			// Clear the pointers so that we don't have debris floating around
			FRunningEditSet = NULL;
			FRunningLayout  = NULL;
			FRunningEdit    = NULL;
		} // if (t_pair)
	} // if call to UpdateRecord succeeded

	// Notify the caller of the total number of edits that failed
	*errors_count = FEditErrors->FailedEditsCount();

	return t_result;
} // RunEdits
//---------------------------------------------------------------------------
int TRunEdits::GetRunEditsSummary(const char* edit_set_tag, const char* layout_tag,
	const int include_zero_counts, int* total_errors, int* total_warnings,
	int* total_skips, int* total_messages, int* total_warning_messages,
	void* owner, void* callback_func)
{
	int t_result = 0;

	std::unique_ptr<TFastSQLQuery> t_syntax_qry(new TFastSQLQuery(FSQLDB));
	std::vector<std::string> t_syntax_edit_tags;

	std::string t_edit_set_tag = stringReplace(edit_set_tag,"'","''");
	char t_buffer[1000];
	std::snprintf(t_buffer, 1000,
		"SELECT c.EditName, c.EditTag " \
		"FROM EE_EditSetEditsPivot a, EE_EditSets b, EE_Edits c " \
		"WHERE a.EditSetPKey = b.PKey " \
		"AND a.EditPKey = c.PKey " \
		"AND b.EditSetTag = '%s' " \
		"AND IFNULL(c.CompiledLogic,'') = '' " \
		"ORDER BY 1;", t_edit_set_tag.c_str());
	t_syntax_qry->SetSQLText(t_buffer);
	try
	{
		t_syntax_qry->SetIsActive(true);
		while (!t_syntax_qry->Eof())
		{
			t_syntax_edit_tags.push_back(t_syntax_qry->GetFieldByNumber(1));

			t_syntax_qry->FetchNext();
		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TRunEdits", "GetRunEditsSummary", e.what());
		FEngineErrorsPump->AddError("TRunEdits", "GetRunEditsSummary", t_syntax_qry->LastErrorMessage());
		FEngineErrorsPump->AddError("TRunEdits", "GetRunEditsSummary", t_syntax_qry->GetSQLText());
		return -1;
	}

	// Initialize these to zero
	*total_errors = 0;
	*total_warnings = 0;
	*total_skips = 0;
	*total_messages = 0;
	*total_warning_messages = 0;

	TLayoutEditSetPair* t_pair =
		find_layout_editset_pair( layout_tag, edit_set_tag);

	if (t_pair)
	{
		CallbackGetRunEditsSummary t_callback_func =
			(callback_func ? (CallbackGetRunEditsSummary)callback_func : NULL);

		TEditSet* t_edit_set = t_pair->EditSet;
		int t_edits_count = t_edit_set->GetEditsCount();

		std::string t_edit_tag,
					t_edit_name,
					t_admin_code;
		int t_tally_errors,
			t_tally_warnings,
			t_tally_skips,
			t_tally_messages,
			t_tally_warning_messages,
			t_skipped_silently,
			t_is_suppressed,
			t_syntax_error_in_logic;

		size_t t_syntax_vector_size = t_syntax_edit_tags.size();

		for (int t_idx = 0; t_idx < t_edits_count; t_idx++)
		{
			// NOTE: This returns the edits in "run order"; that is,
			// single-field edits first, sorted alphabetically, followed by
			// multi-field edits, sorted alphabetically
			TEdit* t_edit = t_edit_set->GetEditByRunOrderIndex(t_idx);
			if (t_edit)
			{
				t_skipped_silently = (t_edit->SkipEditSilently() ? 1 : 0);
				t_is_suppressed = (t_edit->IsSuppressed() ? 1 : 0);

				t_syntax_error_in_logic = 0; // assume the logic is good-- it usually is
				if (t_syntax_vector_size > 0)
				{
					for (size_t t_vdx = 0; t_vdx < t_syntax_vector_size; t_vdx++)
					{
						if (t_syntax_edit_tags[t_vdx].compare(t_edit->GetEditTag()) == 0)
						{
							t_syntax_error_in_logic = 1;
							break;
						}
					}
				}

				t_tally_errors = t_edit->GetTallyErrorsCount();
				t_tally_warnings = t_edit->GetTallyWarningsCount();
				t_tally_skips = t_edit->GetTallySkipsCount();
				t_tally_messages = t_edit->GetTallyErrorMessagesCount();
				t_tally_warning_messages = t_edit->GetTallyWarningMessagesCount();

				if (include_zero_counts == 0)
				{
					if (t_tally_errors == 0 &&
						t_tally_warnings == 0 &&
						t_tally_skips == 0 &&
						t_skipped_silently == 0 &&
						t_is_suppressed == 0 &&
						t_syntax_error_in_logic == 0)
					{
						continue;
					}
				}

				// Increment the output parameters with the tally values
				*total_errors += t_tally_errors;
				*total_warnings += t_tally_warnings;
				*total_skips += t_tally_skips;
				*total_messages += t_tally_messages;
				*total_warning_messages += t_tally_warning_messages;

				if (t_callback_func)
				{
					t_edit_tag.assign(t_edit->GetEditTag());
					t_edit_name.assign(t_edit->GetEditName());
					t_admin_code.assign(t_edit->GetAdminCode());

					(*t_callback_func)(owner,
						t_edit_tag.c_str(),
						t_edit_name.c_str(),
						t_admin_code.c_str(),
						t_tally_errors,
						t_tally_warnings,
						t_tally_skips,
						t_tally_messages,
						t_tally_warning_messages,
						t_skipped_silently,
						t_is_suppressed,
						t_syntax_error_in_logic);
				} // if (t_callback_func
			}
		}

	}
	else
	{
		char t_buffer[1000];
		std::snprintf(t_buffer, 1000,
			"Edit set tag [%s] / Layout tag [%s] pair not loaded for RunEdits processing",
				edit_set_tag, layout_tag);
		FEngineErrorsPump->AddError("TRunEdits", "GetRunEditsSummary", t_buffer);

		t_result = -1;
	}
	return t_result;
} // GetRunEditsSummary
//---------------------------------------------------------------------------

bool TRunEdits::create_edit_logic_query()
{
	if (!FEditLogicQry)
	{
		FEditLogicQry = new TFastSQLQuery(FSQLDB);
		std::string t_sql =
			"SELECT CompiledLogic FROM EE_Edits WHERE PKey = ?;" ;

		FEditLogicQry->SetSQLText(t_sql);

		if (!FEditLogicQry->PrepareParameterizedQuery())
		{
			FEngineErrorsPump->AddError("TRunEdits",
				"create_edit_logic_query",
				"Failure to prepare FEditLogicQry");
			std::string t_err_msg = FEditLogicQry->LastErrorMessage();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TRunEdits",
					"create_edit_logic_query", t_err_msg);
			}

			throw std::runtime_error("");
		}
	}

	return true;
} // create_edit_logic_query
//----------------------------------------------------------------------------
bool TRunEdits::create_edit_details_query()
{
	if (!FEditDetailsQry)
	{
		std::string t_sql;
		FEditDetailsQry = new TFastSQLQuery(FSQLDB);

		t_sql.assign("SELECT f.FieldNumber, piv.FieldOrder ");
		t_sql += "FROM EE_EditFieldsPivot piv, EE_Fields f, EE_Edits e ";
		t_sql += "WHERE f.PKey = piv.FieldPKey ";
		t_sql += "AND e.PKey = piv.EditPKey ";
		t_sql += "AND e.EditTag = ? ";
		t_sql += "ORDER BY piv.FieldOrder;";

		FEditDetailsQry->SetSQLText(t_sql);

		if (!FEditDetailsQry->PrepareParameterizedQuery())
		{
			FEngineErrorsPump->AddError("TRunEdits",
				"create_edit_details_query",
				"Failure to prepare FEditDetailsQry");
			std::string t_err_msg = FEditDetailsQry->LastErrorMessage();
			if (!t_err_msg.empty())
			{
				FEngineErrorsPump->AddError("TRunEdits",
					"create_edit_details_query", t_err_msg);
			}

			throw std::runtime_error("");
		}

	}

	return true;
}  // create_edit_details_query
//---------------------------------------------------------------------------
int TRunEdits::GetEditDetails(const char* edit_tag, void* owner, void* callback_func)
{
	int t_result = 0;
	if (FRunningLayout)
	{
		try
		{
			if (create_edit_details_query())
			{
				FEditDetailsQry->BindTextParameter(1, edit_tag);

				if (FEditDetailsQry->OpenParameterizedQuery())
				{
					int t_field_num,
						t_field_order;
					do
					{
						t_field_num = FEditDetailsQry->GetIntFieldByNumber(0);
						if (t_field_num > 0)
						{
							t_field_order = FEditDetailsQry->GetIntFieldByNumber(1);

							t_result = FRunningLayout->DisplayOneField(
								t_field_num, t_field_order, owner, callback_func);

							if (t_result != 0)
							{
								break;
							}
						}

						FEditDetailsQry->FetchNext();
					} while (!FEditDetailsQry->Eof());

					FEditDetailsQry->ResetParameterizedQuery();

				}
			}
		}
		catch (...)
		{
			std::string t_what;
			t_what.assign( FEditDetailsQry->LastErrorMessage() );
			FEngineErrorsPump->AddError("TRunEdits", "GetEditDetails", t_what);
			t_what.assign( FEditDetailsQry->GetSQLText() );
			FEngineErrorsPump->AddError("TRunEdits", "GetEditDetails", t_what);

			t_result = -1;
		}
	}
	else
	{
		FEngineErrorsPump->AddError("TRunEdits","GetEditDetails",
			"No Layout is loaded!");
		t_result = -1;
	}

	return t_result;
} // GetEditDetails
//---------------------------------------------------------------------------
int TRunEdits::DebugEdit(const std::string edit_tag,
	const std::string layout_tag, const std::string data,
	void* owner, void* logic_callback, void* variables_callback,
	void* fields_callback, void* notes_callback)
{
	// First, let's update the layout with the new data record
	// Reminder: UpdateRecord will create a TLayout object for this
	// layout_tag if one does not already exist.

	// NOTE: This function runs without an edit set.
	FRunningEditSet = NULL;

	std::string t_layout_tag = layout_tag;
	uppercase(t_layout_tag);

	// If we have already instantiated this layout, delete it!
	// This function is called from the EditWriter test bench, which
	// creates unique layouts for each edit... but always names them
	// the same thing. From one edit to the next, that layout is going
	// to be completely different.
	TLayout* t_test = FLayouts[t_layout_tag];
	if (t_test)
	{
		// Delete the layout so that it will be re-loaded
		// and evaluated by UpdateRecord
		delete t_test;

		FLayouts.erase(t_layout_tag);
	}

	int t_result = UpdateRecord(t_layout_tag, data);

	if (t_result != 0)
	{
		return t_result;
	}
	else
	{
		FRunningLayout = FLayouts[t_layout_tag];
	}

	// We want to clear out/re-initialize FEditErrors manager.
	// Default the last parameter, suppress_warnings, to false (zero)
	FEditErrors->Initialize(owner, notes_callback, FRunningLayout, 0);

	FDebugCallbackOwner = owner;
	FDebugLogicCallback = (CallbackDebugFunction)logic_callback;
	FDebugFieldsCallback = (CallbackDebugFieldsFunction)fields_callback;
	FDebugNotesCallback = (CallbackRunEdits)notes_callback;
	FDebugVarsCallback = (CallbackDebugVarsFunction)variables_callback;

	FIsDebugging = true;

	// Retrieve the edit by edit_tag, get its PKey, and retrieve the
	// EditLogic. Parse the EditLogic into lines of source code
	std::unique_ptr<TFastSQLQuery> t_qry(new TFastSQLQuery(FSQLDB));
	char t_buffer[500];
	std::snprintf(t_buffer, 500,
		"SELECT a.PKey, a.EditName, b.AdminCode, m.MessageNumber, m.Message, a.EditLogic, a.CompiledLogic " \
		"FROM EE_Edits a, EE_Agencies b, EE_Messages m " \
		"WHERE b.PKey = a.AgencyPKey " \
		"AND m.PKey = a.MessagePKey " \
		"AND a.EditTag = '%s';", edit_tag.c_str());

	t_qry->SetSQLText(t_buffer);
	try
	{
		t_qry->SetIsActive(true);

		int t_pkey = t_qry->GetIntFieldByNumber(0);
		std::string t_edit_name = t_qry->GetFieldByNumber(1);
		std::string t_admin_code = t_qry->GetFieldByNumber(2);
		std::string t_default_message_number = t_qry->GetFieldByNumber(3);
		std::string t_default_error_message = t_qry->GetFieldByNumber(4);
		std::string t_edit_logic = t_qry->GetFieldByNumber(5);

		std::string t_compiled = t_qry->GetFieldByNumber(6);
		// Watch out for edits that don't compile!
		if (t_compiled.empty())
		{
			std::string t_error = std::string("Edit \"") + t_edit_name +
				std::string("\" does not compile. Run the syntax checker in EditWriter to find and fix the problem.");

			throw std::runtime_error(t_error.c_str());
		}

		FRunningEdit = new TEdit(FSQLDB, FEngineErrorsPump, t_pkey,
			t_edit_name, edit_tag, t_admin_code, t_default_message_number,
			t_default_error_message);

		t_qry->SetIsActive(false);

		FRunningEdit->BuildEditFieldsList(t_qry.get());

		// Make sure our "today" date is current
		FLibFuncs->SetSystemDate();

		FDebugEditObject = new TDebugEdit();
		try
		{
			FTokenMaster->SetCompiledLogic(t_compiled.c_str());

			FDebugEditObject->ParseEditLogic(t_edit_logic);
			FDebugEditObject->ParseCompiledLogic(t_compiled);

			process_edit();

		}
		catch(...)
		{
			delete FDebugEditObject;
			FDebugEditObject = NULL;
			throw;
		}

		delete FDebugEditObject;
		FDebugEditObject = NULL;
	}
	catch (std::runtime_error& e)
	{
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TRunEdits", "DebugEdit", t_what);
		}
		std::string t_error = "SQL: ";
		t_error += (std::string)t_buffer;

		FEngineErrorsPump->AddError("TRunEdits", "DebugEdit", t_error);

		t_result = -1;

	}

	FIsDebugging = false;

	return t_result;
} // DebugEdit
//---------------------------------------------------------------------------
TEditSet* TRunEdits::prepare_editset(const std::string edit_set_tag,
	const std::string layout_tag)
{

	// Edit set objects are created each time they are paired with
	// a different layout. In practice, it will be a rare occurrence
	// to have an edit set loaded more than once, but it is necessary
	// so that we can properly track edits which need to skip silently
	// because they reference a field that is not in the associated
	// layout.

	std::string t_edit_set_tag = edit_set_tag;
	std::string t_layout_tag = layout_tag;

	uppercase(t_edit_set_tag);
	uppercase(t_layout_tag);

	TLayout* t_layout = FLayouts[t_layout_tag];
	TEditSet* t_edit_set = new TEditSet(FSQLDB, FEngineErrorsPump);
	if (t_edit_set)
	{
		try
		{
			int t_result = t_edit_set->LoadEditSet(t_edit_set_tag);

			// Now evaluate the fields referenced by the edits in the edit set
			// against the fields in the layout, and set flags appropriately.
			if (t_result == 0)
			{
				int t_count = t_edit_set->GetEditsCount();

				for (int t_idx = 0; t_idx < t_count; t_idx++)
				{
					TEdit* t_edit = t_edit_set->GetEditByRunOrderIndex(t_idx);
					if (t_edit)
					{
						int t_fld_count = t_edit->GetFieldsCount();
						for (int t_jdx = 0; t_jdx < t_fld_count; t_jdx++)
						{
							int t_fnum = t_edit->GetFieldByMacroOrder(t_jdx);
							if (t_fnum > 0)
							{
								TLayoutField* t_layout_field =
									t_layout->GetFieldByFieldNumber(t_fnum);
								// The field is not in the record layout;
								// add it to the list that tells the edit
								// it should "skip silently" during the
								// RunEdits process.
								if (!t_layout_field)
								{
									t_edit->SetFieldNotInLayout(t_fnum);
								}
							}
							else
							{
								// Same answer, different spot!
								t_edit->SetFieldNotInLayout(t_fnum);
							} // if (t_fnum > 0)
						} // for (t_jdx to t_fld_count)
					} // if (t_edit)
				} // for (t_idx to t_count)
			} // if t_result == 0)
			else
			{
				delete t_edit_set;
				t_edit_set = NULL;
			}
		}
		catch(...)
		{
			// FEngineErrorsPump has been updated elsewhere.
			// Just clean up and get out.
			delete t_edit_set;
			t_edit_set = NULL;
		}

		if (t_layout && t_edit_set)
		{
			// And now make sure we have this pair in our collection
			// for the next time
			TLayoutEditSetPair* t_pair = find_layout_editset_pair(t_layout_tag, t_edit_set_tag);
			if (!t_pair)
			{
				t_pair = new TLayoutEditSetPair();
				t_pair->EditSet = t_edit_set;
				t_pair->Layout = t_layout;
				FLayoutEditSetPairs.push_back(t_pair);
			}
		}

	} // if (t_edit_set)

	return t_edit_set;
} // prepare_editset
//---------------------------------------------------------------------------
void TRunEdits::process_edit()
{
	// Tell the TEditErrors manager to use the edit's default error message
	// unless the edit logic invokes ERROR_MSG to change it.
	FEditErrors->SetAltMessageNumber(0);

	// Default FReturnFail to FALSE until ERROR_TEXT, SAVE_TEXT, etc. change it
	FReturnFail = false;
	// Ditto for FReturnWarn... until SET_WARNING or SAVE_WARNING_TEXT change it
	FReturnWarn = false;

	// Reset the variables tracker to 0
	FTopOfVariablesStack = 0;

	if (FIsDebugging)
	{
		// Report the fields and values of the fields referenced by FRunningEdit
		// to the FDebugFieldsCallback
		debug_fields();
	}
	// We'll keep going until interpret_block() sets this to true, or
	// we run out of compiled logic to process
	FFinished = false;

	size_t t_pos = setup_UDF_pointers();

	FTokenMaster->MovePosBy((int)t_pos);

	try
	{
		// Reset the FYearsIntoFuture value that may be altered by
		// a call to ALLOW_FUTURE_DATE_IOP
		FLibFuncs->SetYearsIntoFuture();

		// This is referenced by interpret_block, case RETURN FAIL.
		// The ONLY time this is reset to true is by TLibFuncs::call_save_error_text,
		// which will report all its errors, continue processing, and ultimately
		// trigger return FAIL from the edit. However, historically, it has
		// also suppressed the default error message. So this flag lets us
		// accommodate that behavior.
		FOverrideErrorMessage = false;

		// When processing the entire edit, don't send in
		// an "end of block" marker. See ExecuteIF, for example
		interpret_block((TTokenID) 0);

		// NOTE BENE: There exist EDITS Programming Language functions which can
		// set FReturnFail and FReturnWarn to true, and yet the edit itself may
		// simply peter out without actually stating RETURN. When that happens,
		// the code in TEditErrors will have properly incremented the
		// the FRunningEdit->IncrMessagesCount(), but we must also increment
		// FRunningEdit->IncrErrorsCount(). Do that now, and be sure to disable
		// FReturnFail/FReturnWarn once you have counted them.
		if (FReturnFail)
		{
			FReturnFail = false;
			FRunningEdit->IncrErrorsCount();
		}
		if (FReturnWarn)
		{
			FReturnWarn = false;
			if (!FRunningEditSet->GetSuppressWarnings())
			{
				FRunningEdit->IncrWarningsCount();
			}
		}



		if (FIsDebugging)
		{
			send_debug_to_callbacks();
		}
	}
	catch (std::runtime_error& e)
	{
		// The routine that threw this error probably wrote to the Engine
		// pump already. Just add the identification of the edit (if possible)
		// and throw the error on up the chain.
		if (FRunningEdit)
		{
			std::string t_edit_error = "Edit: ";
			t_edit_error += FRunningEdit->GetEditName();
			t_edit_error += " [Tag: ";
			t_edit_error += FRunningEdit->GetEditTag();
			t_edit_error += "]";
			FEngineErrorsPump->AddError("TRunEdits","process_edit", t_edit_error);
		}
		else
		{
			std::string t_what = e.what();
			if (!t_what.empty())
			{
				FEngineErrorsPump->AddError("TRunEdits", "process_edit", t_what);
			}
		}

		// We still want to see what happened in the debugger!
		if (FIsDebugging)
		{
			send_debug_to_callbacks();
		}
		// This will toss us back to the loop in RunEdits()
		throw;
	}
} // process_edit
//---------------------------------------------------------------------------
void TRunEdits::SetOverrideErrorMessage( bool over_ride )
{
    // Called by TLibFuncs::call_save_error_text
	FOverrideErrorMessage = over_ride;
}
//---------------------------------------------------------------------------
int TRunEdits::interpret_block(TTokenID end_token)
{
	// end_token will be one of these:
	//  (TToken) 0 - The entire edit block is being processed.
	//   RBRACE    - Called from ExecuteIF and its like to indicate that
	//               a block of code within {...} is being evaluated, and
	//               we should get out when we encounter the closing RBRACE.
	//               Also used with calls from UDFs... which could contain
	//               many nested LBRACE/RBRACE pairs.
	//   SCOLON    - Called from ExecuteIF and its like to indicate the
	//               statement following the if/else/do/while/for statement
	//               is a single statement, not enclosed in LBRACE/RBRACE

	int t_result = 0;

	// controlling debug_edit for user-defined function calls
	bool t_is_udfunccall;

	// This is a scratch variable, which may be filled by GetCompiledToken
	std::string t_cargo;
	int t_jump;

	TToken* t_token;

	size_t t_compiled_length = FTokenMaster->GetCompiledLength();

	// support for the runtime debugger
	int t_debug_line_before,
		t_debug_line_after,
		t_debug_pos,
		t_max_loops,
		t_line_idx;

	do
	{
		// Reset the pointer into FTokenMaster::FStrings
		FTokenMaster->ResetTopOfStringsStack();

		FTokenMaster->SavePos();
		t_token = GetCompiledToken();

		// We could end up here if the edit has no default error message,
		// and the logic ends where an ELSE condition ended. In any case,
		// if we don't have a valid token, we're done here.
		if (!t_token)
			break;

		// Send off for processing based upon the token
		switch (t_token->GetTokenType())
		{
			case IDENTIFIER:
			case FUNCTION:
				FTokenMaster->RestorePos();

				t_is_udfunccall = (t_token->GetTokenID() == UDFUNCCALL);

				if (t_token->GetTokenID() == UDFUNCDECL)
				{
					// We just finished processing a UDF, and it left us
					// looking at the start of another. Return to caller!
					return t_result;
				}

				if (FIsDebugging)
				{
					// Skip over newline characters; otherwise the reporting
					// of variables gets whacked out.
					FTokenMaster->SkipNewLines();

					t_debug_line_before =
						FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
				}

				EvaluateExpression();

				// Regardless of the result of the function call, we DID execute the line(s)
				// that contained the call itself, so display it/them.
				if (FIsDebugging)
				{
					if (t_is_udfunccall)
					{
						t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before);
						debug_edit(t_debug_pos, true);
					}
					else
					{
						t_debug_line_after =
							FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
						t_max_loops = t_debug_line_after-t_debug_line_before;

						for (t_line_idx = 0; t_line_idx <= t_max_loops; t_line_idx++)
						{
							t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before+t_line_idx);
							debug_edit(t_debug_pos, true);
						}
					}
				}
				// Following the call to EvaluateLibFunc, eval7_atom advances
				// the token. We need to make sure we're not sitting on the
				// end_token.
				if ((FEvalToken->GetTokenID() == SCOLON) &&
					(end_token == SCOLON))
				{
					return t_result;
				}

				break;

			case BLOCK:
				// Display the block delimiter
				// If this is a loop iteration > 0, we'll just get the
				// variable updates
				if (FIsDebugging)
				{
					debug_edit(FTokenMaster->GetPos(), true);
				}
				switch (t_token->GetTokenID())
				{
					case LBRACE:
						t_result = interpret_block(RBRACE);
						return t_result;
						//break;

					case RBRACE:
						if (end_token != RBRACE)
						{
							// Had a long if..else if..else situation inside
							// a WHILE condition. The WHILE was terminated by
							// the RBRACE, but this call to process_block found
							// the RBRACE while processing the internal
							// conditions. The right thing to do is to back
							// the pos off of the RBRACE, return to the caller,
							// and let nested calls get back to the point where
							// the end_token is RBRACE.
							FTokenMaster->RestorePos();
						}
						return t_result;
					default:
						break;
				}
				break;

			case DELIMITER:
				if ((t_token->GetTokenID() == SCOLON) &&
					(end_token == SCOLON))
				{
					return t_result;
				}
				break;

			case CONSTRUCT:
				t_jump = std::atoi((char*)FTokenMaster->GetCallCargo());
				switch (t_token->GetTokenID())
				{
					case IF:
						t_result = ExecuteIF(t_jump);
						break;
					case ELSE:
						// ELSE conditions are being handled in ExecuteIF(), but
						// if the edit nests an if..else condition inside another
						// if..else condition, it's possible to encounter this
						// situation. The right thing to do is jump to the end
						// of the block.
						FTokenMaster->MovePosBy(t_jump);
						break;
					case WHILE:
						t_result = ExecuteWHILE(t_jump);
						break;

					case DO:
						t_result = ExecuteDO(t_jump);
						break;

					case FOR:
						t_result = ExecuteFOR(t_jump);
						break;

					default:
						// NOTHING, NOT, SSTRING, and the rest
						break;
				}

				break;

			case KEYWORD:
				switch (t_token->GetTokenID())
				{
					case EEi_CHAR:
					case EEi_INT:
						declare_variables(t_token->GetTokenID());
						if (FIsDebugging)
						{
							debug_edit(FTokenMaster->GetPos(), true);
						}
						break;

					case RETURN:
						// Find out what we're returning: PASS or FAIL

						// Save the line number in the compiled code
						if (FIsDebugging)
						{
							// Skip over newline characters; otherwise the reporting
							// of variables gets whacked out.
							FTokenMaster->SkipNewLines();

							t_debug_line_before =
								FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
						}

						t_result = (int)EvaluateExpression();

						// We're returning from the edit, so we need
						// to set the flag that lets us out.

						// When the compiled token reads "DS", it means
						// "return WARN". eval7_atom changes the 'S'
						// to 2. Note that LibFuncs can also activate FReturnWarn.
						if (t_result == 2 || FReturnWarn)
						{
							// Set FReturnWarn to false so we don't report
							// this warning count outside the loop, below.
							FReturnWarn = false;

							FTokenMaster->CopyToCargo("WARN");
							if (!FRunningEditSet)
							{
								// This call came from a function that does not
								// initialize an edit set (e.g., DebugEdit,
								// RunEditByName). Go ahead and report the warning
								FEditErrors->ReportDefaultError(FRunningEdit, "W");
								FRunningEdit->IncrWarningsCount();
							}
							else
							{
								if (FRunningEditSet->GetSuppressWarnings() == 0)
								{
									FEditErrors->ReportDefaultError(FRunningEdit, "W");
									FRunningEdit->IncrWarningsCount();
								}
							}
						}
						// Similarly, when the compiled token reads "DQ",
						// it means "return FAIL". eval7_atom turns the 'Q'
						// into 0. Note that some LibFuncs can also activate
						// FReturnFail.
						else if (t_result == 0 || FReturnFail)
						{
							// Set FReturnFail to false so we don't report
							// this error outside the loop, below.
							FReturnFail = false;

							FTokenMaster->CopyToCargo("FAIL");

							if (!FOverrideErrorMessage)
							{
								FEditErrors->ReportDefaultError(FRunningEdit, "E");
							}

							FRunningEdit->IncrErrorsCount();
							if (FRunningEdit->GetFieldsCount() == 1)
							{
								int t_field_number =
									FRunningEdit->GetFieldByMacroOrder(0);
								if (t_field_number > 0)
								{
									TLayoutField* t_layout_field =
										FRunningLayout->GetFieldByFieldNumber(t_field_number);
									t_layout_field->SetFailedSingleFieldEdit(true);
								}
							}
						}

						FFinished = true;

						if (FIsDebugging)
						{
							t_debug_line_after =
								FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
							t_max_loops = t_debug_line_after-t_debug_line_before;

							for (t_line_idx = 0; t_line_idx <= t_max_loops; t_line_idx++)
							{
								t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before+t_line_idx);
								debug_edit(t_debug_pos, true);
							}
						}
						break;

					default:
						break;
				}
				break;

			default:
				// Any other situation should simply advance the
				// pos through the compiled logic.
				break;
		} // switch (t_token->GetTokenType())

	} while ((FTokenMaster->GetPos() < t_compiled_length) && !FFinished);

	return t_result;
} // interpret_block
//---------------------------------------------------------------------------
void TRunEdits::SetReturnFail()
{
	// Called from TLibFuncs::call_error_text, call_save_text, etc.
	// to ensure that the edit will return FALSE and display messages.
	FReturnFail = true;
}
//---------------------------------------------------------------------------
void TRunEdits::SetReturnWarn()
{
	// Called from TLibFuncs::call_set_warn and call_save_warning_text
	// to ensure that the edit will return WARN and display messages.
	FReturnWarn = true;
}
//---------------------------------------------------------------------------
size_t TRunEdits::setup_UDF_pointers()
{
	// At the beginning of the encoded block is a string that looks
	// like this: I007350083300929014910186400843011370143002495. These
	// are string representations of address pointers (?) for user-
	// defined functions. This code "walks" that string, turning
	// 5-byte chunks into integers (735, 833, 929, 1491, 1864, 843, 1137,
	// 143 and 2495).
	//
	// The letter ("I" in this example) is used to show how many UDFs
	// are defined ("I" is the 9th letter of the alphabet, ASCII value
	// 73; "@" is the ASCII character directly preceding "A", value 64.
	// Thus, compiled[0] - '@' yields 9.
	//
	// See TMetaFile::call() for usage.
	/* From TMetaFile.cpp:
	for ( i = 0 ; i < *prog - '@'; ++i )
	{
		c = *(prog + 5 * i + 6);
		*(prog + 5 * i + 6) = 0;
		EEi_FuncAddressTable[i] =
		(unsigned int) EEi_atol((char *) prog + 5 * i + 1);
		*(prog + 5 * i + 6) = c;
	}

	//  Now jump to the start of the Edits Logic (that is, skip past
	//  all the UDF addresses).
	prog =  EEi_TempProgBuffer = prog + (*prog - '@') * 5 + 1;
	*/

	size_t t_pos = 0;
	size_t t_idx, t_limit;

	const char* t_compiled_logic = FTokenMaster->GetCompiledLogic();
	size_t t_compiled_length = FTokenMaster->GetCompiledLength();

	t_limit = (t_compiled_logic[0] - '@');
	std::string t_address;
	int t_addr_ptr;

	FUDFPointers.clear();
	char t_scratch[6];
	for (t_idx = 0; t_idx < t_limit && t_idx < t_compiled_length; t_idx++)
	{
		std::strncpy(t_scratch, &t_compiled_logic[(t_idx*5)+1], 5);
		t_scratch[5] = '\x0';
		t_addr_ptr = std::atoi(t_scratch);

		if (t_addr_ptr > 0)
		{
			// Add the address pointer to the list of UDFs
			FUDFPointers.push_back(t_addr_ptr);
		}

	}

	//  Now jump to the start of the Edits Logic (that is, skip past
	//  all the UDF addresses).
	t_pos = (t_limit *5) +1;

	return t_pos;

} // setup_UDF_pointers
//---------------------------------------------------------------------------
void TRunEdits::declare_variables(TTokenID token_id)
{
	// When we get here, the last call to GetCompiledToken retrieved
	// the token indicating EEi_INT ('I') or EEi_CHAR ('H'). Store that
	// information in the variable type
	int t_type = ( token_id == EEi_INT ? EEVT_INT : EEVT_CHAR );

	int t_size;

	// GetCompiledToken places relevant information into FTokenMaster::FStrings. Watch for
	// how it is used to retrieve the variable name, variable size, and the
	// do..while loop controller.
	intptr_t t_var_name;
	TToken* t_token;

	// We need to accommodate variable declarations that could look like
	// any of these... that is, the type declaration affects multiple variables:
	//    int num1, num2, num_array[7];
	//    long long1, long_array[2];
	//    char str1[5], mychar, str2[7];
	//    tablevar CODE, LABEL;
	do
	{
		t_var_name = 0;

		// Default the variable size (capacity) to 1.
		// We'll reassign it as appropriate in the array evaluation area, below.
		t_size = 1;

		t_token = GetCompiledToken();

		t_var_name = FTokenMaster->GetCallCargo();

		FTokenMaster->SavePos();
		t_token = GetCompiledToken();

		if (t_token->GetTokenID() == LBRACKET)
		{
			t_type |= EEVT_ARRAY;

			// Get the size of the array
			t_size = (int)EvaluateExpression();
		}
		else
		{
			FTokenMaster->RestorePos();
		}

		if (t_var_name > 0)
		{
			if (FTopOfVariablesStack >= FVariablesVectorSize)
			{
				// increment the possible number of variables by 20;
				// it's an arbitrary number, but serves as an acceptable
				// growth rate
				FVariablesVectorSize += 20;
				FVariables.resize(FVariablesVectorSize);
				for (size_t t_idx = FTopOfVariablesStack; t_idx < FVariablesVectorSize; t_idx++)
				{
					FVariables[t_idx] = NULL;
				}
			}

			TVariable* t_var = FVariables[FTopOfVariablesStack];
			if (!t_var)
			{
				t_var = new TVariable((char*)t_var_name, t_type, t_size);
				FVariables[FTopOfVariablesStack] = t_var;
			}
			else
			{
				// If there is a TVariable in this slot, we're
				// going to re-use it.
				t_var->Reset((char*)t_var_name, t_type, t_size);
			}

			FTopOfVariablesStack += 1;
		}
		t_token = GetCompiledToken();
	} while (t_token->GetTokenID() != SCOLON);
} // declare_variables
//---------------------------------------------------------------------------
intptr_t TRunEdits::EvaluateExpression()
{
	// The return value from this function is a mechanism for
	// returning values from LIBFUNCs such as AT().
	// The cargo parameter, passed by reference, may contain
	// a string value needed by the calling routine.
	intptr_t t_value = 0;

	// We are using quasi-global variables for the current TToken object
	// and the current position into the compiled logic. Assign parameter
	// "pos" to our global.


	FTokenMaster->SavePos();
	FEvalToken = GetCompiledToken();

	if (FEvalToken->GetTokenType() == DELIMITER &&
		FEvalToken->GetTokenID() == SCOLON)
	{
		// Let the caller deal with the SCOLON; it's a flag
		// in process_block
		return t_value;
	}
	// N.B.: The old Engine didn't get STRING or APPFIELD values here,
	// but from the call_* functions (see TLibFuncs). But I think this
	// is much cleaner, and it works... so let's go with it.
	else if (FEvalToken->GetTokenType() == STRING)
	{
		t_value = (intptr_t)FTokenMaster->GetCallCargo();
		// move off the STRING token
		FEvalToken = GetCompiledToken();

		return t_value;
	}
	else if (FEvalToken->GetTokenType() == APPFIELD)
	{
		int t_field_number = std::atol((char*)FTokenMaster->GetCallCargo());
		TLayoutField* t_layout_field =
			FRunningLayout->GetFieldByFieldNumber(t_field_number);
		t_value = FTokenMaster->CopyToCargo(t_layout_field->GetValue());

		// Move off the APPFIELD...
		FEvalToken = GetCompiledToken();

		return t_value;
	}

	eval0_variables( t_value );

/*
	// We do have one more condition to worry about... use of DT_TODAY.
	// It's neatly handled in eval7_atom, but we need to know about it
	// so that we return t_value, and not cast it to an intptr_t...
	// that way led to an access violation.
	TTokenType t_token_type = FEvalToken->GetTokenType();
	TTokenID t_token_id = FEvalToken->GetTokenID();
	// Continue to the next level of evaluation
	eval0_variables( &t_value );

	if (t_token_type == NUMCONST && t_token_id == 'O')
	{
		std::string t_str_value = intToString(t_value);
		t_value = FTokenMaster->CopyToCargo(t_str_value.c_str());
		//return t_value;
	}
*/
	return (intptr_t)t_value;
} // EvaluateExpression
//---------------------------------------------------------------------------
void TRunEdits::eval0_variables( intptr_t &value )
{
	// NOTE: the output parameter 'value' to these eval* functions may contain
	// a number, a boolean (0 or 1L), or a string value. Strings are returned
	// by assigning the pointer address to the int. The receiver must "know"
	// what it is expecting, and make the necessary conversions.

	if (FEvalToken->GetTokenType() == IDENTIFIER)
	{
		FTokenMaster->SavePos();
		TToken* t_save_token = (TToken*)0;
		intptr_t t_save_cargo = 0;

		// It's a variable. The name of the variable is stored in
		// the most recent entry to FTokenMaster::FStrings.
		TVariable* t_var = GetVariable((char*)FTokenMaster->GetCallCargo());

		if (t_var)
		{
			std::string t_str_value;

			// Cast the variable to the appropriate subclass, and
			// initialize the value parameter to whatever is presently
			// in it.
			if (t_var->GetVariableType() & EEVT_INT)
			{
				value = (int)t_var->GetValue(0);
			}
			else if (t_var->GetVariableType() & EEVT_CHAR)
			{
				value = (intptr_t)t_var->GetStringValueAsIntPtr();
			}

			// We have retrieved the current value of the variable, so move on.
			// This next call is a test... we may move the pointer back.
			// Watch for handling of t_old_pos, below.
			FTokenMaster->SavePos();
			t_save_token = FEvalToken;
			t_save_cargo = FTokenMaster->GetCallCargo();

			FEvalToken = GetCompiledToken();

			// Is this an array variable? Rarely used, but could happen.
			// Array variables are only for arrays of ints or longs.
			// Although strings are declared as arrays of char, they are never
			// referenced one byte at a time in the EDITS language.
			// Default t_idx to 0; a non-array int variable will store its value
			// in index 0
			if (FEvalToken->GetTokenType() == DELIMITER &&
				FEvalToken->GetTokenID() == LBRACKET &&
				t_var->GetVariableType() & EEVT_ARRAY)
			{
				// We're expecting an index into an array
				FEvalToken = GetCompiledToken();
				int t_idx = std::atoi((char*)FTokenMaster->GetCallCargo());
				FEvalToken = GetHardToken(RBRACKET);

				if (!FEvalToken)
				{
					FEngineErrorsPump->AddError("TRunEdits", "eval0_variables",
						"RBRACKET expected");
					throw std::runtime_error("");
				}

				if ((t_var->GetVariableType() & EEVT_INT) &&
					(t_idx >= (int)t_var->Count()))
				{
					char t_buffer[500];
					std::snprintf(t_buffer, 500,
						"syntax error: index %d is greater than number of elements in integer array named \"%s\"",
						t_idx, t_var->GetVariableName());
					FEngineErrorsPump->AddError("TRunEdits", "eval0_variables", t_buffer);
					throw std::runtime_error("");
				}

				// update our go-back-to-here variables, in case we need them.
				FTokenMaster->SavePos();
				t_save_token = FEvalToken;
				t_save_cargo = FTokenMaster->GetCallCargo();

				// Move past the RBRACKET
				FEvalToken = GetCompiledToken();
			}

			if (FEvalToken->GetTokenType() == OPERATOR &&
				FEvalToken->GetTokenID() == ASSIGN)
			{
				// Retrieve the value being assigned; set the
				// pos at the first byte past the assignment token,
				// which might be another variable. A recursive call
				// from here back into this same function will deal
				// with that if it needs to, or will move on to the
				// next level if the value is anything else

				// Move past the ASSIGN operator
				FEvalToken = GetCompiledToken();

				// Get the value to assign
				eval0_variables( value );

				// and update the value of the variable
				if (t_var->GetVariableType() & EEVT_INT)
					t_var->SetValue(0,(int)value);
				else if (t_var->GetVariableType() & EEVT_CHAR)
					t_var->SetValue((char*)value);

				return;

			}
			else
			{
				// It's not an assignment; reinstate our saved values.
				// Here's what happens next: This travels all the way
				// down to eval7_atom as IDENTIFIER token, where the
				// current value is assigned, then a new call to
				// GetCompiledToken occurs, and as it works it's way back
				// up the chain, the code deals with whatever OPERATORs
				// may follow this variable.
				FTokenMaster->RestorePos();
				FEvalToken = t_save_token;
				FTokenMaster->CopyToCargo( (char*)t_save_cargo );
			}
		} // if (we found the variable)
	} // if (token type is an IDENTIFIER)

	// Continue the evaluation
	eval1_logical_ops( value );

} // eval0_variables
//---------------------------------------------------------------------------
void TRunEdits::eval1_logical_ops( intptr_t &value )
{
	// At this level, we are looking out for logical operators AND and OR.
	// See note in eval0_variables that explains how the param named 'value' is
	// used by the caller.
	intptr_t t_partial_value;

	eval2_relational_ops( value );

	// After this call, we'll be positioned on the AND or OR operators
	// (if they exist).
	TTokenType t_token_type = FEvalToken->GetTokenType();
	TTokenID t_token_id = FEvalToken->GetTokenID();

	while (t_token_type == OPERATOR &&
		  (t_token_id == AND || t_token_id == OR))
	{
		// move off the boolean operator
		FEvalToken = GetCompiledToken();

		// and go get the boolean result of the next section of the expression
		eval2_relational_ops(t_partial_value);

		if (t_token_id == AND)
		{
			value = value && t_partial_value;
		}
		else if (t_token_id == OR)
		{
			value = value || t_partial_value;
		}

		t_token_type = FEvalToken->GetTokenType();
		t_token_id = FEvalToken->GetTokenID();
	}

} // eval1_logical_ops
//---------------------------------------------------------------------------
void TRunEdits::eval2_relational_ops( intptr_t &value )
{
	eval3_add_subtract( value );

	TTokenType t_token_type = FEvalToken->GetTokenType();
	TTokenID t_token_id = FEvalToken->GetTokenID();

	if ((t_token_type == OPERATOR)
		&& (t_token_id == EQ ||
			t_token_id == NE ||
			t_token_id == LT ||
			t_token_id == LE ||
			t_token_id == GT ||
			t_token_id == GE))
	{
		intptr_t t_partial_value = 0;

		// Move off the relational operator
		FEvalToken = GetCompiledToken();

		// and get the value being compared
		eval3_add_subtract( t_partial_value );

		switch (t_token_id)
		{
			case EQ:
				value = (value == t_partial_value);
				break;

			case NE:
				value = (value != t_partial_value);
				break;

			case LT:
				value = (value < t_partial_value);
				break;

			case LE:
				value = (value <= t_partial_value);
				break;

			case GT:
				value = (value > t_partial_value);
				break;

			case GE:
				value = (value >= t_partial_value);
				break;

			default:
				break;
		}
	} // if t_token_id is a relational op
} // eval2_relational_ops
//---------------------------------------------------------------------------
void TRunEdits::eval3_add_subtract( intptr_t &value )
{
	// add or subtract

	eval4_mult_div_mod( value );

	TTokenType t_token_type = FEvalToken->GetTokenType();
	TTokenID t_token_id = FEvalToken->GetTokenID();

	intptr_t t_partial_value;

	while (t_token_type == OPERATOR &&
		  (t_token_id == ADD ||
		   t_token_id == SUB))
	{

		// Move off the operator
		FEvalToken = GetCompiledToken();

		eval4_mult_div_mod( t_partial_value );

		if (t_token_id == ADD)
		{
			value = value + t_partial_value;
		}
		else if (t_token_id == SUB)
		{
			value = value - t_partial_value;
		}

		t_token_type = FEvalToken->GetTokenType();
		t_token_id = FEvalToken->GetTokenID();

	} // while (ADD or SUB)
} // eval3_add_subtract
//---------------------------------------------------------------------------
void TRunEdits::eval4_mult_div_mod( intptr_t &value )
{
	// multiply, divide or modulus

	eval5_unary_ops( value );

	TTokenType t_token_type = FEvalToken->GetTokenType();
	TTokenID t_token_id = FEvalToken->GetTokenID();

	intptr_t t_partial_value;

	while (t_token_type == OPERATOR &&
		  (t_token_id == MULT ||
		   t_token_id == DIV ||
		   t_token_id == MOD))
	{
		// Move off the operator
		FEvalToken = GetCompiledToken();

        // and continue the evaluation of the expression
		eval5_unary_ops( t_partial_value);

		if (t_token_id == MULT)
		{
			value = value * t_partial_value;
		}
		else if (t_token_id == DIV)
		{
			if (t_partial_value == 0)
			{
				std::string t_error = "divide by zero error";
				FEngineErrorsPump->AddError("TRunEdits", "eval4_mult_div_mod", t_error);
				throw std::runtime_error("");
			}
			else
			{
				value = value / t_partial_value;
			}
		}
		else if (t_token_id == MOD)
		{
			value = value % t_partial_value;
		}

		t_token_type = FEvalToken->GetTokenType();
		t_token_id = FEvalToken->GetTokenID();

	} // (while MULT || DIV || MOD)

} // eval4_mult_div_mod
//---------------------------------------------------------------------------
void TRunEdits::eval5_unary_ops( intptr_t &value )
{
	// unary operators are a plus or minus sign preceding a value,
	// or a NOT operator
	TTokenType t_save_token_type = FEvalToken->GetTokenType();
	TTokenID t_save_token_id = FEvalToken->GetTokenID();

	if ( t_save_token_type == OPERATOR &&
		(t_save_token_id == ADD ||
		 t_save_token_id == SUB ||
		 t_save_token_id == NOT))
	{
		// Move off the unary operator
		FEvalToken = GetCompiledToken();
	}
	else
	{
		t_save_token_type = (TTokenType)0;
		t_save_token_id = (TTokenID)0;
	}

	// Go deal with parentheses
	eval6_parens( value );

	if (t_save_token_type == OPERATOR)
	{
		if (t_save_token_id == SUB)
		{
			// invert the value
			//*value = -(*value);
			value = -value;
		}
		else if (t_save_token_id == NOT)
		{
			// negate the value
			//*value = !(*value);
			value = !(value);
		}
	}

} // eval5_unary_ops
//---------------------------------------------------------------------------
void TRunEdits::eval6_parens( intptr_t &value )
{
	if (FEvalToken->GetTokenType() == DELIMITER &&
		FEvalToken->GetTokenID() == LPAREN)
	{
		// Move off the LPAREN
		FEvalToken = GetCompiledToken();

		// Back to the top, to evaluate the subexpression
		eval0_variables( value );

		// Now advance past (what should be) the RPAREN
		FEvalToken = GetCompiledToken();
	}
	else
	{
		eval7_atom( value );
	}

} // eval6_parens
//---------------------------------------------------------------------------
void TRunEdits::eval7_atom( intptr_t &value)
{
	// Assign the value of a number, the contents of a variable,
	// or the result of a function call.
	TVariable* t_var;

	std::string t_str_value;

	TTokenType t_token_type = FEvalToken->GetTokenType();
	TTokenID t_token_id = FEvalToken->GetTokenID();

	char* t_call_cargo;

	switch (t_token_type)
	{
		case FUNCTION:
			if (t_token_id == UDFUNCCALL)
			{
				value = (int)evaluateUDF();
			}
			else
			{
				t_call_cargo = (char*)FTokenMaster->GetOneChar();
				value = EvaluateLibFunc(t_call_cargo[0]);
			}
			FEvalToken = GetCompiledToken();
			return;

		case IDENTIFIER:
			t_call_cargo = (char*)FTokenMaster->GetCallCargo();
			t_var = GetVariable(t_call_cargo);

			if (t_var->GetVariableType() & EEVT_INT)
			{
				value = (int)t_var->GetValue(0);
			}
			else if (t_var->GetVariableType() & EEVT_CHAR)
			{
				value = (intptr_t)t_var->GetStringValueAsIntPtr();
			}
			FEvalToken = GetCompiledToken();
			return;

		case NUMBER:  // is a numeric constant
			t_call_cargo = (char*)FTokenMaster->GetCallCargo();
			value = std::atol(t_call_cargo);
			FEvalToken = GetCompiledToken();
			return;

		case CHARCONST:
			value = FEvalToken->GetCompiled();
			FEvalToken = GetCompiledToken();
			return;

		case NUMCONST:
			switch (FEvalToken->GetCompiled())
			{
				case 'O':
					// Today's date as initialized by RunEdits()
					//*value = FLibFuncs->GetSystemDate();
					//t_str_value = intToString(*value);
					t_str_value = intToString(FLibFuncs->GetSystemDate());
					value = FTokenMaster->CopyToCargo(t_str_value.c_str());
					break;
				case 'P':
					value = DT_UNKNOWN;
					break;
				case 'W':
					value = DT_EMPTY;
					break;
				case 'X':
					value = DT_VALID;
					break;
				case 'Y':
					value = DT_MISSING;
					break;
				case 'Z':
					value = DT_ERROR;
					break;
				case 'y':
					value = DT_DAY_EMPTY;
					break;
				case 'z':
					value = DT_MONTH_EMPTY;
					break;

				default:    // Q=0 (FAIL), R=1 (PASS), S=2 (WARN, DT_MAX), T = 3
					t_call_cargo = (char*)FTokenMaster->GetOneChar();
					value = t_call_cargo[0]-'Q';
					break;
			}
			FEvalToken = GetCompiledToken();
			return;

		default:
			if (t_token_id != RPAREN)
			{
				FEngineErrorsPump->AddError("TRunEdits", "eval7_atom", "syntax error: expected RPAREN");
				throw std::runtime_error("");
			}
			break;
	} // switch (t_token_type)
} // eval7_atom
//---------------------------------------------------------------------------
TTokenID TRunEdits::GetEvalTokenID()
{
	// This exists to support TLibFuncs. Many of those functions need to know
	// what the last TTokenID was following a call to GetCompiledToken().
	return FEvalToken->GetTokenID();
}
//---------------------------------------------------------------------------
int TRunEdits::evaluateUDF()
{
	int t_result;

	if (!FUDFPointers.empty())
	{
		FTokenMaster->SavePos();
		char* t_cargo = (char*)FTokenMaster->GetOneChar();
		size_t t_idx = t_cargo[0] - '@' -1;
		if (t_idx < FUDFPointers.size())
		{
			size_t t_saved_pos = FTokenMaster->GetPos();
			size_t t_udf_pos = FUDFPointers[t_idx];

			// Advance FTokenMaster to the beginning of the UDF
			FTokenMaster->MovePosBy((int)(t_udf_pos - t_saved_pos));

			// print the function declaration ("function MyFunc()")
			debug_edit(FTokenMaster->GetPos(), true);

			t_result = interpret_block(RBRACE);

			// Return to the place in the main code that called this UDF
			FTokenMaster->MovePosTo(t_saved_pos);
		}
	}

	return t_result;
} // evaluateUDF
//---------------------------------------------------------------------------
intptr_t TRunEdits::EvaluateLibFunc(const char library_key)
{
	// The return value from this function is a mechanism for
	// returning values from LIBFUNCs such as AT().
	intptr_t t_result = 0;

	// Implementation of the EDITS Language functions is
	// found in class TLibFuncs (attempt to reduce the noise here).
	FCallFuncPtr t_func_ptr = FLibFuncs->GetFuncPointer(library_key);
	if (t_func_ptr)
	{
		t_result = (*FLibFuncs.*t_func_ptr)();
	}
	else
	{
		FEngineErrorsPump->AddError("TRunEdits", "EvaluateLibFunc",
			"syntax error: invalid/deprecated/removed function reference");
		throw std::runtime_error("");
	}

	return t_result;
} // EvaluateLibFunc
//---------------------------------------------------------------------------
TVariable* TRunEdits::GetVariable(const char* var_name, bool complain /* = true */)
{
	// The key into std::map FVariables is the lowercase variable name
	// The caller to this function may not necessarily have lowercased it
	TVariable* t_var = NULL;
	if (FTopOfVariablesStack > 0)
	{
		for (size_t t_idx = 0; t_idx < FTopOfVariablesStack; t_idx++)
		{
			TVariable* t_test = FVariables[t_idx];
			if (t_test)
			{
				if (posix_stricmp(t_test->GetVariableName(), var_name) == 0)
				{
					t_var = t_test;
					break;
				}
			}
		}
	}
	// 99.99% of the time, I want to know if the process is attempting to
	// locate a variable that isn't in the FVariables map. However,
	// TLibFuncs::internal_lookup and TLibFuncs::internal_rlookup
	// have reasons for attempting to locate
	// a tablevar which may or may not have been declared. So those
	// functions override the default value of parameter "complain"
	// to prevent this from happening...
	if (!t_var)
	{
		// Variable not found
		if (complain)
		{
			char t_buffer[100];
			std::snprintf(t_buffer, 100, "In Edit Tag \"%s\", Variable \"%s\" not found!",
				FRunningEdit->GetEditTag().c_str(), var_name);
			FEngineErrorsPump->AddError("TRunEdits", "GetVariable", t_buffer);
			throw std::runtime_error("");
		}
	}

	return t_var;
} // GetVariable
//---------------------------------------------------------------------------

int TRunEdits::ExecuteIF(const int jump)
{
	// We should be on the opening parenthesis of the parameters list
	TToken* t_token= GetHardToken(LPAREN);
	TTokenID t_end_token;
	if (!t_token)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: expected opening parenthesis");
		FEngineErrorsPump->AddError("TRunEdits", "ExecuteIF", t_err_msg);
		throw std::runtime_error("");
	}

	int t_debug_line_before,
		t_debug_line_after;

	// Save the line number in the compiled code
	if (FIsDebugging)
	{
		t_debug_line_before =
			FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
	}

	// Get the condition
	bool t_condition_result = (EvaluateExpression() ? true : false);

	// Regardless of the value of t_condition_result, we DID execute the line(s)
	// that contained the IF condition, so display it/them.
	if (FIsDebugging)
	{
		t_debug_line_after = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
		int t_debug_pos;
		int t_max_loops = t_debug_line_after-t_debug_line_before;

		for (int t_line = 0; t_line <= t_max_loops; t_line++)
		{
			t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before+t_line);
			debug_edit(t_debug_pos, true);
		}
	}

	if (t_condition_result)
	{
		// Condition is true... process the IF block
		// We need to know whether the block is delimited by LBRACE/RBRACE
		FTokenMaster->SavePos();
		t_token = GetHardToken(LBRACE);
		if (t_token)
		{
			t_end_token = RBRACE;
		}
		else
		{
			t_end_token = SCOLON;
			FTokenMaster->RestorePos();
		}

		if (FIsDebugging)
		{
			debug_edit(FTokenMaster->GetPos(), t_condition_result);
		}

		interpret_block(t_end_token);


		// Look to see if our next token is the ELSE to this IF,
		// and if it is, then we'll jump over it.
		FTokenMaster->SavePos();
		t_token = GetHardToken(ELSE);
		if (t_token)
		{
			int t_jump_over_else = std::atoi((char*)FTokenMaster->GetCallCargo());
			FTokenMaster->MovePosBy(t_jump_over_else);
		}
		else
		{
			// Put pos back and get out of this function
			FTokenMaster->RestorePos();
		}
	}
	else
	{
		// Condition is false... jump over the remainder of the block
		FTokenMaster->MovePosBy(jump);

		// We want to handle ELSE, if it exists, right here
		FTokenMaster->SavePos();
		t_token = GetHardToken(ELSE);
		if (t_token)
		{
			FTokenMaster->SavePos();
			t_token = GetHardToken(LBRACE);
			if (t_token)
			{
				t_end_token = RBRACE;
			}
			else
			{
				t_end_token = SCOLON;
				FTokenMaster->RestorePos();
			}

			// Update the debug info (display the line with the ELSE)
			if (FIsDebugging)
			{
				debug_edit(FTokenMaster->GetPos(), true);
			}

			interpret_block(t_end_token);

		}
		else
		{
			// Put pos back and get out of this function
			FTokenMaster->RestorePos();
		}
	}

	return 0;
} // ExecuteIF
//---------------------------------------------------------------------------
int TRunEdits::ExecuteWHILE(const int jump)
{
	// We should be on the opening parenthesis of the parameters list
	TToken* t_token = GetHardToken(LPAREN);
	if (!t_token)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: expected opening parenthesis");
		FEngineErrorsPump->AddError("TRunEdits", "ExecuteWHILE", t_err_msg);
		throw std::runtime_error("");
	}

	TTokenID t_end_token;

	int t_loop_idx = -1;   // This is reset only if FIsDebugging is true

	int t_debug_line_before,
		t_debug_line_after;

	// Save the line number in the compiled code
	if (FIsDebugging)
	{
		t_debug_line_before = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
	}

	// Save the pos; we'll be setting it back to this point within the processing
	// loop, below.
	size_t t_start_of_loop = FTokenMaster->GetPos();
	// Get the condition
	bool t_condition_result = (EvaluateExpression() ? true : false);

	// Regardless of the value of t_condition_result, we DID execute the line(s)
	// that contained the WHILE condition, so display it/them.
	if (FIsDebugging)
	{
		t_debug_line_after = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
		int t_debug_pos;
		int t_max_loops = t_debug_line_after-t_debug_line_before;

		for (int t_line = 0; t_line <= t_max_loops; t_line++)
		{
			t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before+t_line);
			debug_edit(t_debug_pos, true);
		}
	}

	// When loop processing is finished, this is where we want to end up
	size_t t_end_of_loop = FTokenMaster->GetPos() + jump;

	if (t_condition_result)
	{
		// Condition is true... process the WHILE block
		// We need to know whether the block is delimited by LBRACE/RBRACE
		// vs. a one-liner that is terminated with a SCOLON
		FTokenMaster->SavePos();
		t_token = GetHardToken(LBRACE);
		if (t_token)
		{
			t_end_token = RBRACE;
		}
		else
		{
			t_end_token = SCOLON;
			FTokenMaster->RestorePos();
		}

		if (FIsDebugging)
		{
			// Note: Looping routines (for, while, do..while) can be
			// nested. But once you're in a loop, you're there until
			// the loop condition fails or the code returns out of the
			// edit altogether.
			//
			// ExecuteWHILE differs from ExecuteFOR in that it returns to its
			// caller for every iteration of the loop (ExecuteFOR doesn't
			// return until the loop is finished, although nested for(a;b;c)
			// calls present interesting challenges).
			//
			// SO: Check to see if we're in a loop already
			// (if FDebugLoopTopLineNumber.empty(), we are NOT), and if we are,
			// see if it's the same loop being re-entered (the top line number
			// will be different if we're entering a nested while loop).
			int t_debug_line_number = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
			if ((FDebugLoopTopLineNumber.empty()) ||
				(FDebugLoopTopLineNumber.back() != t_debug_line_number))
			{
				// Save the line number, marking the top of this loop
				FDebugLoopTopLineNumber.push_back(t_debug_line_number);

				// And set the loop iteration counter to 1
				FDebugLoopIteration.push_back(1);
				t_loop_idx = (int)FDebugLoopIteration.size() -1;
			}
		} // if (FIsDebugging)

		// If we get to here, we are going to execute the "while" line,
		// so display it (allow for a while condition that spans
		// multiple source code lines). Note that if this is a repeat
		// pass through the loop, we'll only be updating the variables.
		do
		{
			interpret_block(t_end_token);

			if (FFinished)
			{
				break;
			}

			FTokenMaster->MovePosTo(t_start_of_loop);

			t_condition_result = (EvaluateExpression() ? true : false);

			// If the condition for the loop is still true, increment the
			// iteration tracker.
			if (FIsDebugging && t_condition_result)
			{
				// Increment the loop iteration tracker
				FDebugLoopIteration[t_loop_idx] += 1;
			}

		} while (t_condition_result);

		// At this point, we are outside the processing loop, but the pos
		// is positioned on the first token following the while(...).
		// Jump over the block because we're finished with it.
		FTokenMaster->MovePosTo(t_end_of_loop);

		if (FIsDebugging)
		{
			// We're done with this loop. Pop the stacks
			FDebugLoopIteration.pop_back();
			FDebugLoopTopLineNumber.pop_back();
		}

	} // if (t_condition_result)
	else
	{
		// Condition is false... jump over the remainder of the block
		FTokenMaster->MovePosBy( jump );

		// Display the entire contents of the jumped-over code
		// in grey font (or whatever you're using to distinguish
		// between executed and not-executed code).
		if (FIsDebugging)
		{
			debug_edit(FTokenMaster->GetPos(), t_condition_result);
		}
	} // if (!t_condition_result)

	return 0;
} // ExecuteWHILE
//---------------------------------------------------------------------------
int TRunEdits::ExecuteFOR(const int jump)
{
	// t_old_pos is a scratch variable for looking ahead
	// t_pos_condition is the start of the condition (middle) portion of the
	//                 for... loop
	// t_pos_adjuster  is the start of the third portion of the for... loop,
	//                 which contains an incrementer or decrementer affecting
	//                 condition
	//  for ( <assignment> ; <condition>; <adjuster>)
	size_t t_pos_condition,
		   t_pos_adjuster,
		   t_start_of_loop;

	// This helps us find the final closing parenthesis at the end of the
	// FOR... statement (and allows for nested parentheses in the adjuster portion)
	int t_parens = 0;

	// This is reset only if FIsDebugging is true
	int t_loop_idx = -1;

	std::string t_err_msg;

	// We should be on the opening parenthesis of the FOR
	TToken* t_token = GetHardToken(LPAREN);
	if (!t_token)
	{
		t_err_msg.assign("syntax error: expected opening parenthesis");
		FEngineErrorsPump->AddError("TRunEdits", "ExecuteFOR", t_err_msg);
		throw std::runtime_error("");
	}
	else
	{
		t_parens++;
	}

	TTokenID t_end_token;

	// for debug_edit
	int t_debug_line_before,
		t_debug_line_after;

	if (FIsDebugging)
	{
		t_debug_line_before = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
	}

	// Process the first third... it will set values in variables
	EvaluateExpression();

	// We should now be at the beginning of the second third, the condition.
	// Store this location, because we'll be coming back to it.
	t_pos_condition = FTokenMaster->GetPos();

	// Process the condition
	bool t_condition_result = (EvaluateExpression() ? true : false);

	// We should now be at the third portion, the loop adjuster
	// Save this position, but don't evaluate it until we have iterated
	// the loop at least once
	t_pos_adjuster = FTokenMaster->GetPos();;

	// Now let's get down to the block to be processed... scoot along
	// until we find the last closing parenthesis in the FOR loop
	while (t_parens > 0)
	{
		t_token = GetCompiledToken();
		if (t_token->GetTokenID() == LPAREN)
			t_parens++;
		else if (t_token->GetTokenID() == RPAREN)
			t_parens--;
	} // while (t_parens > 0)

	// We executed the line(s) that contained the FOR definition,
	// so display it/them.
	if (FIsDebugging)
	{
		t_debug_line_after = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
		int t_debug_pos;
		int t_max_loops = t_debug_line_after-t_debug_line_before;

		for (int t_line = 0; t_line <= t_max_loops; t_line++)
		{
			t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before+t_line);
			debug_edit(t_debug_pos, true);
		}
	}

	if (t_condition_result)
	{
		// Condition is true... process the FOR block
		// We need to know whether the block is delimited by LBRACE/RBRACE
		// vs. a one-liner that is terminated with SCOLON
		FTokenMaster->SavePos();
		t_token = GetHardToken(LBRACE);
		if (t_token)
		{
			t_end_token = RBRACE;
			if (FIsDebugging)
			{
				// print the line with the RBRACE, too
				debug_edit(FTokenMaster->GetPos(), true);
			}
		}
		else
		{
			t_end_token = SCOLON;
		}

		FTokenMaster->RestorePos();

		// Save the starting position of the block to be processed
		t_start_of_loop = FTokenMaster->GetPos();

		if (FIsDebugging)
		{
			// Note: Looping routines (for, while, do..while) can be
			// nested. But once you're in a loop, you're there until
			// the loop condition fails or the code returns out of the
			// edit altogether.
			//
			// ExecuteWHILE differs from ExecuteFOR in that it returns to its
			// caller for every iteration of the loop (ExecuteFOR doesn't
			// return until the loop is finished, although nested for(a;b;c)
			// calls present interesting challenges).
			//
			// SO: Check to see if we're in a loop already
			// (if FDebugLoopTopLineNumber.empty(), we are NOT), and if we are,
			// see if it's the same loop being re-entered (the top line number
			// will be different if we're entering a nested while loop).
			int t_debug_line_number = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
			if ((FDebugLoopTopLineNumber.empty()) ||
				(FDebugLoopTopLineNumber.back() != t_debug_line_number))
			{
				// Save the line number, marking the top of this loop
				FDebugLoopTopLineNumber.push_back(t_debug_line_number);

				// And set the loop iteration counter to 1
				FDebugLoopIteration.push_back(1);
				t_loop_idx = (int)FDebugLoopIteration.size() -1;
			}
		} // if (FIsDebugging)

		// If we get to here, we are going to execute the FOR block.
		do
		{
			interpret_block(t_end_token);

			if (FFinished)
			{
				break;
			}

			FTokenMaster->SavePos();

			// Set the pos to the start of the third portion of the FOR statement
			FTokenMaster->MovePosTo(t_pos_adjuster);
			// Run it to update variable values
			EvaluateExpression();

			// Now set the pos to the start of the condition portion
			FTokenMaster->MovePosTo(t_pos_condition);

			// Find out if we're still in the loop
			t_condition_result = (EvaluateExpression() ? true : false);

			if (t_condition_result)
			{
				// If the condition for the loop is still true, increment the
				// iteration tracker.
				if (FIsDebugging)
				{
					// Increment the loop iteration tracker
					FDebugLoopIteration[t_loop_idx] += 1;
				}

				// Reset to start of loop
				FTokenMaster->MovePosTo(t_start_of_loop);
			}
			else
			{
				//FTokenMaster->RestorePos();
				FTokenMaster->MovePosTo(t_start_of_loop + jump);
			}

		} while (t_condition_result);

		if (FIsDebugging)
		{
			// We're done with this loop; pop the stacks
			FDebugLoopIteration.pop_back();
			FDebugLoopTopLineNumber.pop_back();
		}

	} // if (t_condition_result)
	else
	{
		// Condition is false... jump over the remainder of the block
		FTokenMaster->MovePosBy( jump );
	}

	return 0;
} // ExecuteFOR
//---------------------------------------------------------------------------
int TRunEdits::ExecuteDO(const int jump)
{
	std::string t_scratch;
	int t_loop_idx = -1;   // This is reset only if FIsDebugging is true

	int t_debug_line_before,
		t_debug_line_after;

	// Save the line number in the compiled code
	if (FIsDebugging)
	{
		t_debug_line_before = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
	}

	// We should be on the opening brace of the do block
	TToken* t_token = GetHardToken(LBRACE);

	if (!t_token)
	{
		std::string t_err_msg;
		t_err_msg.assign("syntax error: expected opening brace \"{\"");
		FEngineErrorsPump->AddError("TRunEdits", "ExecuteDO", t_err_msg);
		throw std::runtime_error("");
	}

	TTokenID t_end_token = RBRACE;

	// Save the pos; we'll be setting it back to this point within the processing
	// loop, below.
	size_t t_start_of_loop = FTokenMaster->GetPos();

	// It's possible to write a do... block putting the opening brace on the
	// line below the "do", so let's go back and print what we have so far.
	if (FIsDebugging)
	{
		t_debug_line_after = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
		int t_debug_pos;
		int t_max_loops = t_debug_line_after-t_debug_line_before;

		for (int t_line = 0; t_line <= t_max_loops; t_line++)
		{
			t_debug_pos = FDebugEditObject->GetPosAtLine(t_debug_line_before+t_line);
			debug_edit(t_debug_pos, true);
		}
	}

	if (FIsDebugging)
	{
		// Note: Looping routines (for, while, do..while) can be
		// nested. But once you're in a loop, you're there until
		// the loop condition fails or the code returns out of the
		// edit altogether.
		//
		// ExecuteDO differs from ExecuteFOR in that it returns to its
		// caller for every iteration of the loop (ExecuteFOR doesn't
		// return until the loop is finished, although nested for(a;b;c)
		// calls present interesting challenges).
		//
		// SO: Check to see if we're in a loop already
		// (if FDebugLoopTopLineNumber.empty(), we are NOT), and if we are,
		// see if it's the same loop being re-entered (the top line number
		// will be different if we're entering a nested DO loop).
		int t_debug_line_number = FDebugEditObject->FindCompiledLogicLine(FTokenMaster->GetPos());
		if ((FDebugLoopTopLineNumber.empty()) ||
			(FDebugLoopTopLineNumber.back() != t_debug_line_number))
		{
			// Save the line number, marking the top of this loop
			FDebugLoopTopLineNumber.push_back(t_debug_line_number);

			// And set the loop iteration counter to 1
			FDebugLoopIteration.push_back(1);
			t_loop_idx = (int)FDebugLoopIteration.size() -1;
		}
	} // if (FIsDebugging)

	// Initialize the condition to true: We will run the loop AT LEAST once
	// before testing to see if we should loop back around
	bool t_condition_result = true;

	do {
		interpret_block(t_end_token);

		if (FFinished)
		{
			break;
		}

		// We should be on the WHILE now
		t_token = GetHardToken(WHILE);

		if (!t_token)
		{
			std::string t_err_msg;
			t_err_msg.assign("syntax error: expected WHILE");
			FEngineErrorsPump->AddError("TRunEdits", "ExecuteDO", t_err_msg);
			throw std::runtime_error("");
		}

		// We should be on the opening parenthesis of the parameters list
		t_token = GetHardToken(LPAREN);

		if (!t_token)
		{
			std::string t_err_msg;
			t_err_msg.assign("syntax error: expected LPAREN \"(\"");
			FEngineErrorsPump->AddError("TRunEdits", "ExecuteDO", t_err_msg);
			throw std::runtime_error("");
		}

		t_condition_result = (EvaluateExpression() ? true : false);

		if (t_condition_result)
		{
			// If the condition for the loop is still true, increment the
			// iteration tracker.
			if (FIsDebugging)
			{
				// Increment the loop iteration tracker
				FDebugLoopIteration[t_loop_idx] += 1;
			}

			// reset the pos to the start of the loop
			FTokenMaster->MovePosTo(t_start_of_loop);
		}

	} while (t_condition_result);

	if (FIsDebugging)
	{
		// We're done with this loop. Pop the stacks
		FDebugLoopIteration.pop_back();
		FDebugLoopTopLineNumber.pop_back();
	}

	return 0;
} // ExecuteDO
//---------------------------------------------------------------------------
//  debug_edit returns each line of edit logic as it is executing to the caller
//  through FDebugLogicCallback. Lines of code that execute are prefixed with
//  a guillemet '' (Alt+0187). Source lines are "jumped over" by block
//  processing functions like exec_if(); those lines of code are also returned
//  to the caller, but without the guillemet, so that the user can distinguish
//  what code ran vs. what was skipped. (EditWriter uses font colors to indicate
//  these differences.)
//
//  This function also calls debug_showvars to return the list of variables
//  and their current values for each line of code processed.
//
void TRunEdits::debug_edit(size_t pos, bool debug_execute)
{
	if (FIsDebugging)
	{
		int t_debug_line_number = FDebugEditObject->FindCompiledLogicLine(pos);
		std::string t_logic = FDebugEditObject->GetEditLogicAtLine(t_debug_line_number);

		if (debug_execute)
		{
			// Add to FDebugEditObject
			if (!FDebugEditObject->GetLineExecuted(t_debug_line_number))
			{
				FDebugEditObject->AddDebugLine(t_debug_line_number, true, t_logic);
			}

			// Display the variables
			debug_showvars(t_debug_line_number);
		} // if (debug_execute)
		else
		{
			// In a loop, a line of code may execute only once, then not again.
			// Don't overwrite a line saying it did not execute simply because
			// it did not execute in THIS pass through the loop.
			if (!FDebugEditObject->GetLineExecuted(t_debug_line_number))
			{
				FDebugEditObject->AddDebugLine(t_debug_line_number, false, t_logic);
			}
		}

	} // if (FIsDebugging)
} // debug_edit
//---------------------------------------------------------------------------
void TRunEdits::send_debug_to_callbacks()
{
	if (FDebugEditObject)
	{
		if (FDebugLogicCallback)
		{
			int t_max_lines = (int)FDebugEditObject->GetEditLogicLinesCount();
			int t_line_no,
				t_last_line = 1;
			std::string t_logic;

			for (int t_idx = 0; t_idx < t_max_lines; t_idx++)
			{
				t_line_no = FDebugEditObject->GetLineNumberAtMapIndex(t_idx);
				if (t_line_no >= 0)
				{
					// The debug_edit may not have been called for lines that
					// were skipped. Let's go fill in the blanks.
					while (t_last_line < t_line_no)
					{
						t_logic.assign(FDebugEditObject->GetEditLogicAtLine(t_last_line));

						(*FDebugLogicCallback)(FDebugCallbackOwner, t_last_line+1, 0, t_logic.c_str());

						t_last_line++;
					}

					if (FDebugEditObject->HasDebugLine(t_line_no))
					{
						bool t_executed = FDebugEditObject->GetLineExecuted(t_line_no);
						int t_line_executed = (t_executed ? 1 : 0);

						t_logic.assign(FDebugEditObject->GetLogic(t_line_no));

						(FDebugLogicCallback)(FDebugCallbackOwner, t_line_no+1, t_line_executed, t_logic.c_str());
					}
					else
					{
						t_logic.assign(FDebugEditObject->GetEditLogicAtLine(t_line_no));

						(*FDebugLogicCallback)(FDebugCallbackOwner, t_line_no+1, 0, t_logic.c_str());
					}
					t_last_line = t_line_no+1;

				}

				// If the caller provided FDebugVarsCallback, report
				// the variables as of this line of code
				if (FDebugVarsCallback)
				{
					std::string t_var_name,
								t_var_value;

					int t_iterations_count,
						t_vars_count,
						t_var_idx,
						t_iteration,
						t_iteration_idx,
						t_is_string_var;

					// Each line in FDebugEditObject stores a list of iterations.
					// Within each iteration is a list of all of the variables
					// and their values AS OF this line + iteration.
					// So we'll send to the caller, in order, all variables
					// (sorted alphabetically by FDebugEditObject) grouped
					// by iteration.
					t_iterations_count = FDebugEditObject->GetDebugVariableIterationsCount(t_line_no);
					t_vars_count = FDebugEditObject->GetDebugVariablesCount(t_line_no);
					for (t_iteration_idx = 0; t_iteration_idx < t_iterations_count; t_iteration_idx++)
					{
						t_iteration = FDebugEditObject->GetDebugVariablesIteration(
							t_line_no, t_iteration_idx);

						if (t_iteration < 0)
							continue;
						// This looks confusing, but the truth is that a LINE is
						// executed t_iterations_count times, which means that
						// the variables will EACH be updated t_iterations_count times.
						// So the t_vars_count is constant for the line and all of its
						// iterations.
						for (t_var_idx = 0; t_var_idx < t_vars_count; t_var_idx++)
						{
							t_var_name.assign(FDebugEditObject->GetDebugVariableName(
								t_line_no, t_iteration, t_var_idx));
							t_var_value.assign(FDebugEditObject->GetDebugVariableValue(
								t_line_no, t_iteration, t_var_idx));
							t_is_string_var = FDebugEditObject->GetDebugVariableIsStringVar(
								t_line_no, t_iteration, t_var_idx);

							(*FDebugVarsCallback)(FDebugCallbackOwner, t_line_no+1, t_iteration,
								t_var_name.c_str(), t_var_value.c_str(), t_is_string_var);
						} // for (t_var_idx...)
					} // for (t_iteration ...)
				} // if (FDebugVarsCallback)
			} // for (int t_idx...
		} // if (FDebugLogicCallback)
	} // if (FDebugEditObject)
} // send_debug_to_callbacks
//---------------------------------------------------------------------------
void TRunEdits::debug_showvars(int debug_line_number)
{
	if (FIsDebugging)
	{
		std::string t_var_name,
					t_var_value;
		int t_iteration = 0,
			t_is_string_var;
		if (!FDebugLoopIteration.empty())
		{
			t_iteration = FDebugLoopIteration.back();
		}
		// Loop through all of the currently known variables and
		// report their current state to TDebugEditObject

		for (size_t t_idx = 0; t_idx < FTopOfVariablesStack; t_idx++)
		{
			TVariable* t_var = FVariables[t_idx];
			std::string t_print;
			if (t_var)
			{
				t_var_name.assign(t_var->GetVariableName());
				t_is_string_var = 0; // default to false
				if (t_var->GetVariableType() & EEVT_INT)
				{
					int t_value = t_var->GetValue(0);
					if ((t_value >= DT_MONTH_EMPTY) &&
						(t_value <= DT_UNKNOWN))
					{
						switch (t_value)
						{
							case DT_MONTH_EMPTY:
								t_var_value.assign("DT_MONTH_EMPTY");
								break;
							case DT_DAY_EMPTY:
								t_var_value.assign("DT_DAY_EMPTY");
								break;
							case DT_EMPTY:
								t_var_value.assign("DT_EMPTY");
								break;
							case DT_VALID:
								t_var_value.assign("DT_VALID");
								break;
							case DT_MISSING:
								t_var_value.assign("DT_MISSING");
								break;
							case DT_ERROR:
								t_var_value.assign("DT_ERROR");
								break;
							case DT_UNKNOWN:
								t_var_value.assign("DT_UNKNOWN");
								break;
							default:
								t_var_value.assign(intToString(t_value)); // see Utility.cpp
								break;
						}
					}
					else
					{
						t_var_value.assign(intToString(t_value)); // see Utility.cpp
					}
				}
				else if (t_var->GetVariableType() & EEVT_CHAR)
				{
					intptr_t t_int_value = t_var->GetStringValueAsIntPtr();
					if (t_int_value > 0)
					{
						t_var_value.assign((char*)t_int_value);
					}
					t_is_string_var = 1; // string variables are decorated in EditWriter
				}

				FDebugEditObject->AddDebugVariable(debug_line_number, t_iteration,
					t_var_name, t_var_value, t_is_string_var);
			}
		}

	} // if (FDebugVarsCallback)
} // debug_showvars
//---------------------------------------------------------------------------
void TRunEdits::debug_fields()
{
	// Called from process_edit
	if (FIsDebugging && FDebugFieldsCallback)
	{
		int t_count = FRunningEdit->GetFieldsCount();
		int t_fld_num;
		TLayoutField* t_layout_fld;
		std::string t_field_name;
		std::string t_value;
		std::string t_print;
		for (int t_idx = 0; t_idx < t_count; t_idx++)
		{
			t_fld_num = FRunningEdit->GetFieldByMacroOrder(t_idx);
			t_layout_fld = FRunningLayout->GetFieldByFieldNumber(t_fld_num);
			if (t_layout_fld)
			{
				t_field_name.assign(t_layout_fld->GetFieldName());
				t_value.assign(t_layout_fld->GetValue());

				t_print.assign(t_field_name);
				t_print += " == ";
				t_print += t_value;

				(*FDebugFieldsCallback)(FDebugCallbackOwner, t_print.c_str());

			}
		} // for...
	} // if (FDebugFieldsCallback)
} // debug_fields
//---------------------------------------------------------------------------

