//---------------------------------------------------------------------------

#ifndef RunEditsH
#define RunEditsH
//---------------------------------------------------------------------------
#include <string>
#include <vector>
#include <map>
#include <stdint.h> // intptr_t

#include "TokenMaster.h"
#include "edit_external.h"

// Forward declarations
class TSQLDB;
class TFastSQLQuery;
class TEngineErrorsPump;
class TEditErrors;
class TLayout;
class TEdit;
class TEditSet;
class TVariable;
class TLibFuncs;
class TTablesList;
class TCondition;
class TDebugEdit;

typedef struct
{
	TLayout* Layout;
	TEditSet* EditSet;
} TLayoutEditSetPair;

class TRunEdits
{
private:
	TSQLDB* FSQLDB;
	TFastSQLQuery* FEditLogicQry;
	TFastSQLQuery* FEditDetailsQry;
	TEngineErrorsPump* FEngineErrorsPump;
	TEditErrors* FEditErrors;
	TTokenMaster* FTokenMaster;

	std::map<std::string, TLayout*> FLayouts;   // key is layout_tag
	std::vector<int> FUDFPointers;

	// FVariables vector never clears its contents (until the TRunEdits dtor)
	std::vector<TVariable*> FVariables;
	size_t FVariablesVectorSize;  // starts at 0, increments as needed by 20
	size_t FTopOfVariablesStack;  // reset to 0 in process_edit

	TEditSet* FRunningEditSet;
	TLayout*  FRunningLayout;
	TEdit*    FRunningEdit;

	// quasi-global variables used by the eval* functions (processing edit logic)
	TToken* FEvalToken;

	TLibFuncs* FLibFuncs;

	TTablesList* FTablesList;

	bool FFinished;
	bool FReturnFail; // set by ERROR_TEXT, SAVE_TEXT, ERROR_MSG, etc.
	bool FReturnWarn; // set by SET_WARNING, SAVE_WARNING_TEXT

	// SAVE_ERROR_TEXT suppresses a default error message (not the same thing
	// as what ERROR_MSG does, which is to replace the message). This value
	// is set to false in process_edit, and may be set to true from
	// TLibFuncs::call_set_error_text;
	bool FOverrideErrorMessage;

	// Support for Edit_DebugEdit

	void* FDebugCallbackOwner;
	CallbackDebugFunction FDebugLogicCallback;
	CallbackDebugFieldsFunction FDebugFieldsCallback;
	CallbackRunEdits FDebugNotesCallback;
	CallbackDebugVarsFunction FDebugVarsCallback;

	bool FIsDebugging;
	TDebugEdit* FDebugEditObject;

	// These FDebugLoop* variables are for updating variables
	// from within ExecuteFOR, ExecuteWHILE, ExecuteDO
	std::vector<int>  FDebugLoopTopLineNumber;
    std::vector<int>  FDebugLoopIteration;

	std::vector<TLayoutEditSetPair*> FLayoutEditSetPairs;

	TLayoutEditSetPair* find_layout_editset_pair( const std::string layout_tag, const std::string edit_set_tag);
	TEditSet* find_editset( const std::string edit_set_tag, size_t start_idx, bool& more );
	TEditSet* prepare_editset(const std::string edit_set_tag, const std::string layout_tag);

	void clear_variables();
	void process_edit();
	int interpret_block(TTokenID end_token);

	size_t setup_UDF_pointers();

	void declare_variables(TTokenID token_id);

	// emulating path of execution from EDITS40
	// it's a lot more obtuse, but should run much faster
	void eval0_variables( intptr_t &value );
	void eval1_logical_ops( intptr_t &value );
	void eval2_relational_ops( intptr_t &value );
	void eval3_add_subtract( intptr_t &value );
	void eval4_mult_div_mod( intptr_t &value );
	void eval5_unary_ops( intptr_t &value );
	void eval6_parens( intptr_t &value );
	void eval7_atom( intptr_t &value );

	int evaluateUDF();

	// These functions invoke the FDebug*Callback functions
	void debug_edit(size_t pos, bool debug_execute);
	void debug_fields();
	void debug_showvars(int debug_line_number);

	void send_debug_to_callbacks();

    bool create_edit_logic_query();
	bool create_edit_details_query();

public:
	TRunEdits(TSQLDB* database, TEngineErrorsPump* pump, TTablesList* tables_list);
	~TRunEdits();

	// getters
	TEditSet* GetRunningEditSet() { return FRunningEditSet; };
	TLayout*  GetRunningLayout()  { return FRunningLayout; };
	TEdit*    GetRunningEdit()    { return FRunningEdit; };

	int UpdateRecord(std::string layout_tag, std::string data);
	int DisplayDataRecord( std::string layout_tag, void* owner, void* callback_func );
	int DisplayFieldInMessagesAs( const int field_number, std::string display_name );

	int NeverSkipEmptyField( const int field_number, const int never_skip_empty );

	int FieldDataByNumber(
		const char* layout_tag, const int field_number, int* pkey,
		int* start_pos, int* field_length,
		char* field_name, const int field_name_len,
		char* display_name, const int display_name_len,
		char* data_type, const int data_type_len,
		char* value, const int value_len );


	int LoadEditSet( const std::string edit_set_tag, const std::string layout_tag );
	int UnloadEditSet( const std::string edit_set_tag, const std::string layout_tag );

	int SetEditOptions( const std::string edit_set_tag, const std::string edit_tag,
		const int suppress_edit, const int skip_if_empty, const int skip_if_failed,
		const int no_skips_override );
	int GetEditOptions( const std::string edit_set_tag, const std::string edit_tag,
		int* is_suppressed, int* is_skip_if_empty, int* is_skip_if_failed,
		int* is_no_skips_override );

	int ClearEditOptions( const std::string edit_set_tag );

	int RunEdits(const std::string edit_set_tag, const std::string layout_tag, const std::string data, const int edit_options, int* errors_count, void* owner, void* callback_func);
	int GetRunEditsSummary(const char* edit_set_tag, const char* layout_tag,
		const int include_zero_counts, int* total_errors, int* total_warnings,
		int* total_skips, int* total_messages, int* total_warning_messages,
		void* owner, void* callback_func);
	int GetEditDetails(const char* edit_tag, void* owner, void* callback_func);
	int DebugEdit(const std::string edit_tag, const std::string layout_tag, const std::string data, void* owner, void* logic_callback, void* variables_callback, void* fields_callback, void* notes_callback);

	intptr_t EvaluateExpression();
	TTokenID GetEvalTokenID();

	intptr_t EvaluateLibFunc(const char library_key);

	int ExecuteIF(const int jump);
	int ExecuteWHILE(const int jump);
	int ExecuteFOR(const int jump);
	int ExecuteDO(const int jump);

	// Support for calls from class TLibFuncs
	TVariable* GetVariable(const char* var_name, bool complain = true);
	TToken* GetCompiledToken();
	TToken* GetHardToken(const TTokenID token_id );
	char* GetScratchBuffer();

    // Called by ERROR_TEXT, SAVE_TEXT, etc. to ensure the edit FAILs
	void SetReturnFail() ;
	// Called by SET_WARNING and SAVE_WARNING_TEXT
	void SetReturnWarn();
    // Called by SAVE_ERROR_TEXT
	void SetOverrideErrorMessage( bool over_ride );

};

#endif
