#include <memory>    // smart pointers
#include <algorithm> // std::transform
#include <stdexcept>
#include <cstdlib>  // std::atoi
#include <cstring> // strncpy
#include <cstdio>  // sprintf, snprintf

//---------------------------------------------------------------------------
#include "SMF.h"
#include "SQLDB.h"
#include "SQLQuery.h"
#include "Utility.h"
#include "EngineErrorsPump.h"
#include "RunEdits.h"


#ifdef __SQLEW__
#include "Ew_AccessController.h"
#include "SMFWriter.h"
#include "EditCompiler.h"
#endif

// Needed by check_mf_structure_version and UpdateMetafile
// UPDATE THIS VALUE WHENEVER THE DATABASE STRUCTURE CHANGES
#define METAFILE_STRUCTURE_VERSION "5.01"

// Needed for GetEngineVersion
// UPDATE THIS VALUE WHENEVER THE ENGINE VERSION CHANGES, and keep it
// synchronized with the VersionInfo embedded in Project | Options | Version
#define ENGINE_VERSION "5.0.0.18"

//---------------------------------------------------------------------------
// Note: CreateNewMetafile can be called when no metafile is open, so
// it does not belong to an instance of TSMF... it's just a function.
#include "SMFTables.h"
int CreateNewMetafile(const char* smf_name)
{
	int t_result = 0;
	TSQLDB* t_sqldb = NULL;
	TSMFTables* t_tables = NULL;
	try
	{
		t_sqldb = new TSQLDB();
		if (t_sqldb)
		{
			t_sqldb->OpenDatabase(smf_name);
			t_tables = new TSMFTables(t_sqldb);
			if (t_tables)
			{
				if (!t_tables->CreateSMFTables(METAFILE_STRUCTURE_VERSION))
				{
					t_result = -2;
				}
			}
			delete t_tables;
			t_tables = NULL;
		}
		delete t_sqldb;
		t_sqldb = NULL;
	}
	catch(...)
	{
		if (t_tables)
		{
			delete t_tables;
			t_tables = NULL;
		}
		if (t_sqldb)
		{
			delete t_sqldb;
			t_sqldb = NULL;
		}
		t_result = -2;
	}

	return t_result;
} // CreateNewMetafile
//---------------------------------------------------------------------------
int CompactMetafile(const char* smf_name)
{
	int t_result = 0;

	TSQLDB* t_sqldb = NULL;

	try
	{
		t_sqldb = new TSQLDB();
		if (t_sqldb)
		{
			t_sqldb->OpenDatabase(smf_name);
			if (!t_sqldb->Vacuum())
			{
				t_result = -2;
			}
		}

		delete t_sqldb;
		t_sqldb = NULL;
	}
	catch (...)
	{
		if (t_sqldb)
		{
			delete t_sqldb;
			t_sqldb = NULL;
		}

		t_result = -2;
	}

	return t_result;
} // CompactMetafile
//---------------------------------------------------------------------------
int GetEngineVersion( char* version_buffer, int buffer_length)
{
	int t_result = 0;

	try
	{
		std::strncpy(version_buffer, ENGINE_VERSION, buffer_length);
	}
	catch (...)
	{
		t_result = -2;
	}

	return t_result;
} // GetEngineVersion
//---------------------------------------------------------------------------
int GetMetafileStructureVersion(char* smf_structure_version, int smf_version_size)
{
	int t_result = 0;

	try
	{
		std::strncpy(smf_structure_version, METAFILE_STRUCTURE_VERSION, smf_version_size);
	}
	catch (...)
	{
		t_result = -2;
	}

	return t_result;
} // GetMetafileStructureVersion
//---------------------------------------------------------------------------
/****************************************************************************
 ***       Callback functions invoked by build_tables_list                ***
 ****************************************************************************/
static int TableInfoCallback(void* callback_owner, int col_count,
	char** col_values, char** col_names)
{
	int t_result = 0;

	std::string t_field_name,
				t_type,
				t_default_value,
				t_col_name,
				t_col_value;

	/* We'll be receiving information that looks like this:
		cid = 0          -- column order number
		name = PKey      -- column name
		type = integer   -- data type
		notnull = 1      -- NOT nullable
		dflt_value =     -- default value
		pk = 1           -- position (if any) in the primary key list

		cid = 1
		name = Gesundheit
		type = varchar(10)
		notnull = 0
		dflt_value = 'sniff!'  -- example of a string default
		pk = 0

		cid = 2
		name = Counter
		type = integer
		notnull = 1
		dflt_value = 0         -- example of an integer default
		pk = 0

		We have to catch them into variables so that we can construct the
		string in the order of information we want to present as a foreign
		key definition.
	*/

	TSMF* t_smf = (TSMF*)callback_owner;
	if (t_smf)
	{
		try
		{
			TTableDef* t_table_def = t_smf->GetCurrentTableDef();
			if (t_table_def)
			{
				for (int t_idx = 0; t_idx < col_count; t_idx++)
				{
					t_col_name.assign(col_names[t_idx]);

					if (t_col_name.compare("name") == 0)
						t_field_name.assign(col_values[t_idx]);
					else if (t_col_name.compare("type") == 0)
						t_type.assign(col_values[t_idx]);
					else if (t_col_name.compare("dflt_value") == 0)
					{
						if (col_values[t_idx])
						{
							t_default_value.assign(col_values[t_idx]);
						}
					}
				}

				if (t_default_value.length() > 0)
				{
					std::string t_line = "default " + t_field_name;
					t_line += " TO ";
					if ((t_type.find("char") != std::string::npos) || (t_type.compare("text") == 0))
						t_line += "'" + t_default_value + "'";
					else
						t_line += t_default_value;


					t_table_def->AddDefault(t_line.c_str());
				}
			}
			else
			{
				// This should not happen... but if it does, let's stop
				// the sqlite3_exec processing to this callback.
				t_result = 1;
			}
		}
		catch (...)
		{
			// Something bad has happened; return a non-zero
			// value from the callback so that the sqlite3_exec will
			// return SQLITE_ABORT and not keep making calls into this
			// routine
			t_result = 1;
		}
	}

	return t_result;
} // TableInfoCallback
//---------------------------------------------------------------------------

static int ForeignKeyCallback(void* callback_owner, int col_count,
	char** col_values, char** col_names)
{
	int t_result = 0;
	std::string t_table,
				t_from,
				t_to,
				t_on_update,
				t_on_delete,
				t_match;

	std::string t_col_name, t_col_value;

	/* We'll be receiving information that looks like this:
		id = 0
		seq = 0
		table = EE_EDSET
		from = EdSetPKey
		to = PKey
		on_update = NO ACTION
		on_delete = CASCADE
		match = NONE

		We have to catch them into variables so that we can construct the
		string in the order of information we want to present as a foreign
		key definition.
	*/

	TSMF* t_smf = (TSMF*)callback_owner;
	if (t_smf)
	{
		try
		{
			TTableDef* t_table_def = t_smf->GetCurrentTableDef();
			if (t_table_def)
			{
				for (int t_idx = 0; t_idx < col_count; t_idx++)
				{
					t_col_name.assign(col_names[t_idx]);
					if (t_col_name.compare("table") == 0)
						t_table.assign(col_values[t_idx]);
					else if (t_col_name.compare("from") == 0)
						t_from.assign(col_values[t_idx]);
					else if (t_col_name.compare("to") == 0)
						t_to.assign(col_values[t_idx]);
					else if (t_col_name.compare("on_update") == 0)
					{
						t_col_value.assign(col_values[t_idx]);
						if (t_col_value.compare("NO ACTION") != 0)
							t_on_update.assign(t_col_value);
					}
					else if (t_col_name.compare("on_delete") == 0)
					{
						t_col_value.assign(col_values[t_idx]);
						if (t_col_value.compare("NO ACTION") != 0)
							t_on_delete.assign(t_col_value);
					}
					else if (t_col_name.compare("match") == 0)
					{
						t_col_value.assign(col_values[t_idx]);
						if (t_col_value.compare("NONE") != 0)
							t_match = t_col_value;
					}

				}
				std::string t_line = t_from + " REFERENCES ";
				t_line += t_table + "(" + t_to + ")";
				if (t_on_update.length() > 0)
					t_line += " ON UPDATE " + t_on_update;
				if (t_on_delete.length() > 0)
					t_line += " ON DELETE " + t_on_delete;
				if (t_match.length() > 0)
					t_line += " MATCH " + t_match;

				t_table_def->AddForeignKey(t_line.c_str());
			}
			else
			{
				// This should not happen... but if it does, let's stop
				// the sqlite3_exec processing to this callback.
				t_result = 1;
			}
		}
		catch (...)
		{
			// Something bad has happened; return a non-zero
			// value from the callback so that the sqlite3_exec will
			// return SQLITE_ABORT and not keep making calls into this
			// routine
			t_result = 1;
		}
	}

	return t_result;
} // ForeignKeyCallback
//---------------------------------------------------------------------------


TSMF::TSMF()
{
	FEngineErrorsPump = NULL;

	FTablesList = NULL;
	FTableDef = NULL;

	// Parameterized queries for build_tables_list
	FAddTableDefIndexesQry = NULL;
	FAddTableDefConstraintsQry = NULL;
	FAddTableDefTriggersQry = NULL;

	FSQLDB = NULL;

	FRunEdits = NULL;
	FEditCompiler = NULL;

#ifdef __SQLEW__
	FEw_AccessController = NULL;
	FSMFWriter = NULL;
#endif

} // constructor
//---------------------------------------------------------------------------
TSMF::~TSMF()
{
	if (FAddTableDefIndexesQry)
	{
		FAddTableDefIndexesQry->ReleaseParameterizedQuery();
		delete FAddTableDefIndexesQry;
		FAddTableDefIndexesQry = NULL;
	}

	if (FAddTableDefConstraintsQry)
	{
		FAddTableDefConstraintsQry->ReleaseParameterizedQuery();
		delete FAddTableDefConstraintsQry;
		FAddTableDefConstraintsQry = NULL;
	}

	if (FAddTableDefTriggersQry)
	{
		FAddTableDefTriggersQry->ReleaseParameterizedQuery();
		delete FAddTableDefTriggersQry;
		FAddTableDefTriggersQry = NULL;
	}

	if (FTablesList)
	{
		delete FTablesList;
		FTablesList = NULL;
	}

	std::map<std::string, TSQLQuery*>::iterator it = FQueriesList.begin();
	while (it != FQueriesList.end())
	{
		TSQLQuery* t_qry = (TSQLQuery*)(*it).second;
		delete t_qry;
		it++;
	}

	FQueriesList.clear();

	if (FRunEdits)
	{
		delete FRunEdits;
		FRunEdits = NULL;
	}

#ifdef __SQLEW__
	if (FImportMetafile)
	{
		delete FImportMetafile;
	}

	if (FEw_AccessController)
	{
		delete FEw_AccessController;
	}

	if (FSMFWriter)
	{
		delete FSMFWriter;
	}

#endif

	if (FEditCompiler)
	{
		delete FEditCompiler;
	}

	delete FEngineErrorsPump;

	if (FSQLDB)
	{
		delete FSQLDB;
		FSQLDB = NULL;
	}

} // destructor
//---------------------------------------------------------------------------
int TSMF::InitializeMetafile(const char* smf_name, char* smf_init_error,
	const int error_buffer_size, int* version_compare,
	const int check_version /* = 1 */)
{
	// Parameter check_version is a boolean, defaulting to 1.
	// When true, the call came from Edit_Initialize.
	// When false (0), the call came from Edit_UpdateMetafile... which
	// already found out from Edit_Initialize that the structure needs
	// adjusting. See below for how check_version is used to control
	// whether we call check_mf_structure_version.

	int t_result = -1; // change to 0 if everything works out okay

	// Set FLastEngineError to "no error"
	FEngineErrorsPump = new TEngineErrorsPump();

	// Open the database named in smf_name.
	// Locate the smf_version (wherever we decide to store it!)
	// Park a pointer to the instance of TSQLDB in vector SMFList
	// (see EditAPI.cpp)
	FSQLDB = new TSQLDB();

	// Default the smf_init_error to an empty string
	if (error_buffer_size > 0)
	{
		smf_init_error[0] = '\x0';
	}

	if (FSQLDB->OpenDatabase(smf_name))
	{
		// Set up FTablesList with all tables in the metafile
		if (!build_tables_list())
		{
			FSQLDB->CloseDatabase();
			strncpy(smf_init_error, "Unable to build tables list", error_buffer_size);
			return t_result;
		}

		// Make sure this metafile has a minimum allowable structure version
		if (check_version &&
			!check_mf_structure_version(smf_init_error, error_buffer_size, version_compare))
		{
			FSQLDB->CloseDatabase();
			return t_result;
		}

		FRunEdits = new TRunEdits(FSQLDB, FEngineErrorsPump, FTablesList);

#ifdef __SQLEW__
		FEw_AccessController = new TEw_AccessController(FEngineErrorsPump, FSQLDB);
		FSMFWriter = new TSMFWriter(this, FSQLDB, FEngineErrorsPump, FEw_AccessController, FTablesList);
		// This is not initialized until a user is actually running the Import Metafile
		// module from EditWriter
		FImportMetafile = NULL;
#endif


		FEditCompiler = new TEditCompiler(FSQLDB, FEngineErrorsPump, FTablesList);

		// If we make it to here, we're good to go
		t_result = 0;

	}

	return t_result;
} // InitializeMetafile
//---------------------------------------------------------------------------
bool TSMF::check_mf_structure_version(char* smf_init_error, const int error_buffer_size,
	int* version_compare)
{
	bool t_result = true;

	std::string t_mf_structure_version = METAFILE_STRUCTURE_VERSION;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

	std::string t_sql =
		"SELECT [ItemValue] FROM EE_Control WHERE [ItemName] = 'MFStructureVersion';";

	try
	{
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);
		try
		{

			if (!t_qry->Eof())
			{
				std::string t_stored_version = t_qry->GetFieldByName("ItemValue");
				*version_compare = t_mf_structure_version.compare(t_stored_version);

				if (*version_compare != 0)
				{
					t_result = false;

					std::string t_error = "Engine expects metafile structure version \"";
					t_error += t_mf_structure_version;
					t_error += "\", but metafile is version \"";
					t_error += t_stored_version;
					t_error += "\"";

					strncpy(smf_init_error, (char*)t_error.c_str(), error_buffer_size);
				}
			}
		}
		catch(...)
		{
			t_qry->SetIsActive(false);
			throw;
		}
		t_qry->SetIsActive(false);
	}
	catch (...)
	{
		std::string t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
		strncpy(smf_init_error, (char*)t_error.c_str(), error_buffer_size);
		t_result = false;
	}

	return t_result;
} // check_mf_structure_version
//---------------------------------------------------------------------------
int TSMF::UpdateMetafile(char* smf_init_error, const int error_buffer_size)
{
	int t_result = 0;

	TSMFTables* t_tables = NULL;

	try
	{
		t_tables = new TSMFTables(FSQLDB);
		if (t_tables)
		{
			t_result = t_tables->UpdateMetafile(METAFILE_STRUCTURE_VERSION,
												smf_init_error,
												error_buffer_size);

			delete t_tables;
			t_tables = NULL;

		}
	}
	catch (...)
	{
		if (t_tables)
		{
			delete t_tables;
			t_tables = NULL;
		}
	}

	return t_result;
} // UpdateMetafile
//---------------------------------------------------------------------------
int TSMF::GetMetafileVersion(char* smf_version, int smf_version_size)
{
	int t_result = 0;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT [ItemValue] FROM EE_Control ";
	t_sql += "WHERE [ItemName] = 'MFContentVersion';";

	try
	{
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);
		try
		{
			std::strncpy(smf_version, t_qry->GetFieldByName("ItemValue"), smf_version_size);
		}
		catch(...)
		{
			t_qry->SetIsActive(false);
			throw;
		}
		t_qry->SetIsActive(false);
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TSMF","GetMetafileVersion", e.what());
		std::string t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","GetMetafileVersion", t_error);
		t_result = -1;
	}

	return t_result;
} // GetMetafileVersion
//---------------------------------------------------------------------------
int TSMF::GetMetafileComment(char* smf_comment, int smf_comment_size)
{
	int t_result = 0;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT [ItemValue] FROM EE_Control ";
	t_sql += "WHERE [ItemName] = 'MFComment';";

	try
	{
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);
		try
		{
			strncpy(smf_comment, t_qry->GetFieldByName("ItemValue"), smf_comment_size);
		}
		catch(...)
		{
			t_qry->SetIsActive(false);
			throw;
		}
		t_qry->SetIsActive(false);
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TSMF","GetMetafileComment", e.what());
		std::string t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","GetMetafileComment", t_error);
		t_result = -1;
	}

	return t_result;
} // GetMetafileComment
//---------------------------------------------------------------------------
int TSMF::ClearEngineErrorsPump()
{
	return FEngineErrorsPump->ClearPump();
}
//---------------------------------------------------------------------------
int TSMF::GetEngineErrorsCount(int* how_many_errors)
{
	return FEngineErrorsPump->GetErrorsCount(how_many_errors);
}
//---------------------------------------------------------------------------
int TSMF::GetEngineErrorByIndex(int idx, char* error_message, int message_length)
{
	// FIX ME ? This is up for discussion, but I plan to document the error-retrieval
	// function to assume the index is 1-based. However, TEngineErrorsPump will
	// expect 0-based. Really, I'm up in the air about this issue and want to discuss
	// with programmers working in languages other than C++ to get their take.
	return FEngineErrorsPump->GetErrorByIndex(idx-1, error_message, message_length);
}
//---------------------------------------------------------------------------
TTableDef* TSMF::GetCurrentTableDef()
{
	// The callback functions need to populate properties of TTableDef
	// objects during processing of build_tables_list. This is how
	// those callbacks can get at that information.
	return FTableDef;
}
//---------------------------------------------------------------------------

bool TSMF::build_tables_list()
{
	bool t_result = true;

	if (FTablesList)
	{
		// For whatever reason, this was caught in a recursive call. Just delete
		// the old one and build a new one.
		delete FTablesList;
		FTablesList = NULL;
	}

	FTablesList = new TTablesList();

	std::unique_ptr<TSQLQuery> t_table_names_qry(new TSQLQuery(FSQLDB));

	std::string t_sql = "SELECT [tbl_name] FROM sqlite_master WHERE [type] LIKE 'table' ORDER BY 1;";

	std::string t_table_name;

	t_table_names_qry->SetSQLText((char*)t_sql.c_str());

	try
	{
		t_table_names_qry->SetIsActive(true);
		while (t_result && !t_table_names_qry->Eof())
		{
			t_table_name.assign(t_table_names_qry->GetFieldByName("tbl_name"));

			t_result = UpdateTableStats(t_table_name.c_str());

			t_table_names_qry->FetchNext();
		}

		if (!t_result)
		{
			FEngineErrorsPump->AddError("TSMF", "build_tables_list",
				"Failure to build list of tables");
		}

	}
	catch(std::runtime_error& e)
	{
		std::string t_error = e.what();
		FEngineErrorsPump->AddError("TSMF","build_tables_list", t_error);

		t_result = false;
	}

	// We won't need this anymore; let's just disable it
	FTableDef = NULL;

	return t_result;
} // build_tables_list
//---------------------------------------------------------------------------
bool TSMF::UpdateTableStats(const char* table_name)
{
	bool t_result = true;

	std::string t_table_name,
				t_col_name,
				t_col_type,
				t_str_col_size;

	int t_how_many, t_idx;

	std::unique_ptr<TSQLQuery> t_struct_qry(new TSQLQuery(FSQLDB));

	t_table_name.assign(table_name);

	// I need FTableDef to be a public property of the form so that
	// my calls to get foreign key constraints can update the
	// appropriate fields from within the callback. Function
	// create_table_def will create it, but we can destroy
	// it when we're done here.
	if ( !create_table_def(t_table_name, t_struct_qry.get()) )
	{
		FTableDef = NULL;
		return false;
	}

	t_table_name.assign(FTableDef->GetTableName());
	uppercase(t_table_name);


	if (!add_table_def_indexes(t_table_name))
	{
		delete FTableDef;
		FTableDef = NULL;
		return false;
	}

	if (!add_table_def_constraints(t_table_name))
	{
		delete FTableDef;
		FTableDef = NULL;
		return false;
	}

	if (!add_table_triggers(t_table_name))
	{
		delete FTableDef;
		FTableDef = NULL;
		throw std::runtime_error("Failure in call to add_table_triggers");
	}

	// We're going to need to see column defaults as part of the TFieldDef
	t_how_many = FTableDef->GetDefaultCount();
	if (t_how_many > 0)
	{
		char t_default_def[200];
		char t_field_name[31];
		char t_default_value[100];
		size_t t_pos, t_dest, t_def_len;
		for (t_idx = 0; t_idx < t_how_many; t_idx++)
		{
			std::memset(t_default_def, '\x0', 200);
			std::strncpy(t_default_def, FTableDef->GetDefault(t_idx), 199);
			t_def_len = std::strlen(t_default_def);
			std::memset(t_field_name, '\x0', 31);
			std::memset(t_default_value, '\x0', 100);

			// this will look like
			//   default MYLABEL TO "ABC" for character fields, or
			//   default MYCODE TO 99  (i.e., no quotes) for numeric fields
			// skip over the leading "default " to find the column name

			t_pos = 8;
			t_dest = 0;
			while (t_default_def[t_pos] != ' ')
			{
				t_field_name[t_dest++] = t_default_def[t_pos++];
			}

			// t_pos will be pointing to this:
			//  " TO ..."
			//  Skip over the TO and the leading/trailing blanks

			t_pos += 4;
			// For this purpose, we don't want the quotes surrounding
			// the default value.
			t_dest = 0;
			while (t_pos < t_def_len)
			{
				if (t_default_def[t_pos] != 39)  // Alt+0039 = ' (apostrophe)
				{
					t_default_value[t_dest++] = t_default_def[t_pos];
				}

				t_pos++;
			}

			// Get the TFieldDef;
			TFieldDef* t_fld = FTableDef->GetColumn(t_field_name);
			if (t_fld)
			{
				t_fld->SetDefaultValue(t_default_value);
			}
		} // for (t_idx... iterating defaults
	} // if (t_how_many > 0)

	// We also want to be able to return field-level CHECK
	// constraints, but ONLY if they are "limit to list"
	// such as "CHECK(EditsFunc IN ('E','S'))
	// or range controllers such as "CHECK(PrimarySite BETWEEN 'C000' AND 'C999')";
	t_how_many = FTableDef->GetCheckCount();
	if (t_how_many > 0)
	{
		std::string t_parse;
		char t_field_name[31];

		// Yes, I'm setting an artificial limit here, but
		// seriously... a thousand bites should be plenty.
		char t_check_constraint[1000];
		char t_limit_to_list[1000];
		std::string t_check_between;
		size_t t_pos,
			   t_range_pos,
			   t_dest,
			   t_check_len;

		for (t_idx = 0; t_idx < t_how_many; t_idx++)
		{
			std::memset(t_field_name, '\x0', 31);
			std::memset(t_check_constraint, '\x0', 1000);
			std::strncpy(t_check_constraint, FTableDef->GetCheck(t_idx), 999);
			t_check_len = std::strlen(t_check_constraint);

			// Look at the first characters following "check"
			// We want to advance to the beginning of the
			// field name
			t_pos = 5;
			while (t_pos < t_check_len)
			{
				char t_char = t_check_constraint[t_pos];
				if ((t_char == ' ') || (t_char == '('))
				{
					t_pos++;
				}
				else
				{
					break;
				}
			}

			if (t_pos >= t_check_len)
				continue;

			t_dest = 0;
			while (t_pos < t_check_len)
			{
				// Now we'll keep going until we have the field name
				char t_char = t_check_constraint[t_pos];
				if ((t_char == ' ') || (t_char == '('))
				{
					break;
				}

				t_field_name[t_dest++] = t_char;
				t_pos++;
			}

			// Look for a CHECK...BETWEEN constraint, first.
			t_check_between.assign(FTableDef->GetCheck(t_idx));
			lowercase(t_check_between);
			t_range_pos = t_check_between.find("between");
			if (t_range_pos != std::string::npos)
			{
				// Get the TFieldDef;
				TFieldDef* t_fld = FTableDef->GetColumn(t_field_name);
				if (t_fld)
				{
					char t_ranger[100];
					std::memset(t_ranger, '\x0', 100);
					size_t t_range_len = t_check_between.length(),
						   t_dest = 0,
						   t_and_pos = t_check_between.find(" and ");

					t_range_pos += 7; // first character past "between"

					// skip over white space and (if this is a character
					// field) the opening apostrophe to the first range
					while ((t_range_pos < t_range_len) &&
						   (t_check_between[t_range_pos] == 32 ||    // ' '
							t_check_between[t_range_pos] ==  9 ||    // '\t'
							t_check_between[t_range_pos] == 13 ||    // '\r'
							t_check_between[t_range_pos] == 10 ||    // '\n'
							t_check_between[t_range_pos] == 39))     // apostrophe
					{
						t_range_pos++;
					}

					// Now start loading t_ranger
					while ((t_range_pos < t_range_len) &&
						   (t_dest < 100) &&
						   (t_check_between[t_range_pos] != 39))
					{
						// If a between .. and .. check is in place for an integer
						// field, the values are not decorated with apostrophes.
						// instead, see if t_range_pos has caught up to t_and_pos,
						// and if it has... break!
						if (t_and_pos != std::string::npos)
						{
							if (t_range_pos == t_and_pos)
							{
                            	break;
							}
						}
						// t_check_between has been lowercased.
						// return the literal: t_check_constraint
						t_ranger[t_dest++] = t_check_constraint[t_range_pos++];
					}

					// Place guillemet between the low and high values
					// of the range (for easier parsing at the receiving end)
					if (t_dest < 100)
						t_ranger[t_dest++] = ''; // Alt+0187

					// Now we want to step over the closing apostrophe
					// to the low end of the range, the white space and " AND "
					while ((t_range_pos < t_range_len) &&
						   (t_check_between[t_range_pos] == 32 ||    // ' '
							t_check_between[t_range_pos] ==  9 ||    // '\t'
							t_check_between[t_range_pos] == 13 ||    // '\r'
							t_check_between[t_range_pos] == 10 ||    // '\n'
							t_check_between[t_range_pos] == 39))     // apostrophe
					{
						t_range_pos++;
					}

					t_range_pos += 3; // AND

					// Skip over white space and opening apostrophe
					// to the high end of the range
					while ((t_range_pos < t_range_len) &&
						   (t_check_between[t_range_pos] == 32 ||    // ' '
							t_check_between[t_range_pos] ==  9 ||    // '\t'
							t_check_between[t_range_pos] == 13 ||    // '\r'
							t_check_between[t_range_pos] == 10 ||    // '\n'
							t_check_between[t_range_pos] == 39))     // apostrophe
					{
						t_range_pos++;
					}

					// Finish loading t_ranger, looking out for the
					// closing apostrophe (character) or right parenthesis (integer)
					while ((t_range_pos < t_range_len) &&
						   (t_dest < 100) &&
						   (t_check_between[t_range_pos] != 39) &&
						   (t_check_between[t_range_pos] != 41))
					{
						// t_check_between has been lowercased.
						// return the literal: t_check_constraint
						t_ranger[t_dest++] = t_check_constraint[t_range_pos++];
					}

					t_fld->SetCheckBetween(t_ranger);
				}

				// POSSIBLY dangerous assumption, in the horrible event
				// somebody embeds the characters "between" in the middle
				// of a limit-to-list check constraint, but we'll nevertheless
				// make the assumption that isn't going to happen... and
				// say that this check constraint is done, done, done.
				continue;
			}

			t_parse.assign(FTableDef->GetCheck(t_idx));
			lowercase(t_parse);
			// Syntax might read " in(" or " in ("
			t_pos = t_parse.find(" in(");
			if (t_pos == std::string::npos)
			{
				t_pos = t_parse.find(" in ");
				if (t_pos == std::string::npos)
				{
					// This is not a limit-to-list check constraint;
					// move on
					continue;
				}
			}

			// Okay, we have one. Get it from FTableDef so that
			// we don't mess with capitalization of the list

			std::memset(t_limit_to_list, '\x0', 1000);

			// Advance t_pos to the first character following the
			// open parenthesis that follows "IN".

			while (t_pos < t_check_len)
			{
				if (t_check_constraint[t_pos] == '(')
				{
					t_pos++;
					break;
				}
				t_pos++;
			}
			t_dest = 0;

			while (t_pos < t_check_len)
			{
				if (t_check_constraint[t_pos] == ')')
				{
					// We're done. Get out.
					break;
				}
				t_limit_to_list[t_dest++] = t_check_constraint[t_pos++];
			}

			// Get the TFieldDef;
			TFieldDef* t_fld = FTableDef->GetColumn(t_field_name);
			if (t_fld)
			{
				t_fld->SetLimitToList(t_limit_to_list);
			}
		} // for (iterating check constraints)
	}

	if (t_result)
	{
		FTablesList->AddTable(FTableDef);
	}
	return t_result;
} // UpdateTableStats
//---------------------------------------------------------------------------
bool TSMF::create_table_def(const std::string table_name, TSQLQuery* qry)
{
	// Called from build_tables_list, CreateTable and ImportLookupTable
	// to set up TTableDef object
	bool t_result = true;

	FTableDef = new TTableDef(table_name.c_str());
	std::string t_sql;
	std::string t_col_name, t_col_type, t_str_col_size;
	TEditDataType t_enum_col_type;
	int t_col_size,
		t_col_count;
	size_t t_open_paren,
		   t_close_paren;
	int t_is_nullable,
		t_is_primary_key,
		t_is_auto_inc;

	// We don't want data, just the column definitions. So by asking for
	// a row that doesn't exist, we get just what we need. ("rowid" is
	// an element of SQLite database structures, so every table will have it.
	t_sql  = "SELECT * FROM ";
	t_sql += table_name;
	t_sql += " WHERE rowid = 0;";

	qry->SetSQLText((char*)t_sql.c_str());
	qry->SetIsActive(true);
	t_col_count = qry->ColCount();
	for (int t_col = 0; t_col < t_col_count; t_col++)
	{
		t_col_name.assign(qry->GetColumnName(t_col));
		t_col_type.assign(qry->GetColumnDataType(t_col));

		lowercase(t_col_type);
		t_enum_col_type = edtUnknown;
		if (t_col_type.compare("integer") == 0)
			t_enum_col_type = edtInteger;
		else if (t_col_type.compare("double") == 0)
			t_enum_col_type = edtDouble;
		else if (t_col_type.compare("text") == 0)
			t_enum_col_type = edtMemo;
		else if (t_col_type.compare("datetime") == 0)
			t_enum_col_type = edtDateTime;
		// Note to self: Be very careful with use of these
		// std::string::compare >= tests. I had these last two
		// higher up in this list, and the one for "char"
		// caught "text" and "datetime"!
		else if (t_col_type.compare("varchar") >= 0)
			t_enum_col_type = edtVarchar;
		else if (t_col_type.compare("char") >= 0)
			t_enum_col_type = edtChar;


		if ((t_enum_col_type == edtVarchar) ||
			(t_enum_col_type == edtChar))
		{
			t_open_paren = t_col_type.find("(");
			t_close_paren = t_col_type.find(")");
			if ((t_open_paren != std::string::npos) &&
				(t_close_paren != std::string::npos))
			{
				t_str_col_size = t_col_type.substr(t_open_paren +1, (t_close_paren - t_open_paren)-1);
				t_col_size = std::atoi(t_str_col_size.c_str());
			}
		}
		else
		{
			t_col_size = 0;
		}

		////////////////////////////////////////////////////////////////////
		/// Constraints are a bear to retrieve from SQLite, but here's how
		////////////////////////////////////////////////////////////////////

		// Get the [NOT] NULL and PRIMARY KEY constraints (I'm having no
		// luck with the auto-increment constraint... but I don't really
		// care that much.) See below for FOREIGN KEY and CHECK constraints
		if (!qry->GetColumnConstraints(table_name.c_str(), t_col_name.c_str(),
			t_is_nullable, t_is_primary_key, t_is_auto_inc))
		{
			// FIX ME: If this call returns false, it's because it failed
			// in SQLite; find out what the error was!
			t_is_nullable = 0;
			t_is_primary_key = 0;
			t_is_auto_inc = 0;
		}

		TFieldDef* t_field = new TFieldDef(t_col_name.c_str(),
			t_col, t_enum_col_type, t_col_size,
			(t_is_nullable != 0), (t_is_primary_key != 0), (t_is_auto_inc != 0));

		if ((t_enum_col_type == edtVarchar) && (t_col_size == 1))
		{
			t_field->SetIsTriggerField(true);
		}

		FTableDef->AddColumn(t_field);
	}

	return t_result;
} // create_table_def
//---------------------------------------------------------------------------
bool TSMF::add_table_def_indexes(const std::string table_name)
{
	bool t_result = true;

	if (!FAddTableDefIndexesQry)
	{
		// Called from build_tables_list and ImportLookupTable
		// to set up TTableDef object
		std::string t_sql  = "SELECT [sql] FROM sqlite_master ";
		t_sql += "WHERE [tbl_name] LIKE ? ";
		t_sql += "AND [type] LIKE 'index';";

		FAddTableDefIndexesQry = new TSQLQuery(FSQLDB);
		FAddTableDefIndexesQry->SetSQLText(t_sql);

		if (!FAddTableDefIndexesQry->PrepareParameterizedQuery())
		{
			delete FAddTableDefIndexesQry;
			FAddTableDefIndexesQry = NULL;
			return false;
		}
	}

	FAddTableDefIndexesQry->BindTextParameter(1, table_name.c_str());

	if (FAddTableDefIndexesQry->OpenParameterizedQuery())
	{
		do
		{
			FTableDef->AddIndex(FAddTableDefIndexesQry->GetFieldByNumber(0));
			FAddTableDefIndexesQry->FetchNext();
		} while (!FAddTableDefIndexesQry->Eof());
	}
	else
	{
		t_result = false;
	}

	FAddTableDefIndexesQry->ResetParameterizedQuery();


	return t_result;
} // add_table_def_indexes
//---------------------------------------------------------------------------
bool TSMF::add_table_def_constraints(const std::string table_name)
{
	// Called from build_tables_list, CreateTable and ImportLookupTable
	// to set up TTableDef object
	bool t_result = true;

	// DEFAULT constraints can be retrieved from PRAGMA table_info
	std::string t_pragma = "PRAGMA table_info('" + table_name + "');";
	FSQLDB->Pragma(t_pragma.c_str(), this, (void*)TableInfoCallback);

	// Foreign key constraints require calling PRAGMA foreign_key_list
	// and interpreting the results.
	t_pragma = "PRAGMA foreign_key_list('" + table_name + "');";
	// int t_result =
	FSQLDB->Pragma(t_pragma.c_str(), this, (void*)ForeignKeyCallback);

	// FAddTableDefConstraintsQry
	if (!FAddTableDefConstraintsQry)
	{
		// Emerging research suggests there is NO WAY to retrieve the
		// definition of CHECK constraints except to parse the CREATE TABLE
		// statement. grrrrrr!
		std::string t_sql  = "SELECT [sql] FROM sqlite_master ";
		t_sql += "WHERE [tbl_name] LIKE ? ";
		t_sql += "AND [type] LIKE 'table';";

		FAddTableDefConstraintsQry = new TSQLQuery(FSQLDB);
		FAddTableDefConstraintsQry->SetSQLText(t_sql);

		if (!FAddTableDefConstraintsQry->PrepareParameterizedQuery())
		{
			delete FAddTableDefConstraintsQry;
			FAddTableDefConstraintsQry = NULL;
			return false;
		}
	}

	FAddTableDefConstraintsQry->BindTextParameter(1, table_name.c_str());

	if (FAddTableDefConstraintsQry->OpenParameterizedQuery())
	{
		std::string t_check,
					t_pos_finder;
		std::string t_parse =
			FAddTableDefConstraintsQry->GetFieldByNumber(0);

		// I don't want to lowercase the entire constraint,
		// because it may contain limit-to-list information.
		// So let's try a dodge...
		t_pos_finder.assign(t_parse);

		lowercase(t_pos_finder);
		size_t t_pos = t_pos_finder.find("check");
		while (t_pos != std::string::npos)
		{
			// Careful! This could have found 'check' embedded within
			// a column name (e.g., 'checkmate', 'nochecking')
			// Walk the string to find ALL check constraints,
			// and pay attention to nested parentheses.
			size_t t_len = t_pos_finder.length();
			size_t t_parens = 0;
			size_t t_walk = t_pos + 5; // first character past "check"
			char t_char;
			while (t_walk < t_len)
			{
				// find first opening paren after the check (could have intervening
				// white space, tab, line feed...)
				t_char = t_pos_finder[t_walk++];
				if (!(t_char == ' '||
					  t_char == '\t'||
					  t_char == '\r'||
					  t_char == '\n'))
				{
					// If the character is not an open parenthesis,
					// this is not a check constraint
					if (t_char == '(')
					{
						t_parens++;
					}
					// in any case, we're not walking over whitespace,
					// so...
					break;
				}
			} // while (t_walk < t_len)

			if (t_parens == 1)
			{
				// We are now one place past the "check(", so keep walking
				// until we find the matching close parenthesis. Remember that
				// there could be nested open/close parentheses, so the
				// t_parens count could go up before it goes down.
				while (t_parens > 0 && t_walk < t_len)
				{
					t_char = t_pos_finder[t_walk++];
					if (t_char == '(')
						t_parens++;
					else if (t_char == ')')
						t_parens--;

					if (t_parens == 0)
					{
						t_check.assign(t_parse.substr(t_pos, t_walk-t_pos));
						FTableDef->AddCheck(t_check.c_str());

						t_pos_finder = t_pos_finder.substr(t_walk+1, t_pos_finder.length());
						t_parse = t_parse.substr(t_walk+1, t_parse.length());
						break;
					}
				} // while (t_parens > 0 && t_walk < t_len)
			} // if (t_parens == 1)
			else
			{
				// We get here if "check" was not actually a constraint, but
				// something else.
				t_pos_finder = t_pos_finder.substr(t_pos+5, t_pos_finder.length());
				t_parse = t_parse.substr(t_pos+5, t_parse.length());
			}

			// Look for the next instance of "check", if any
			t_pos = t_pos_finder.find("check");

		} // while (t_pos != std::string::npos)
	}
	else
	{
		t_result = false;
	}

	FAddTableDefConstraintsQry->ResetParameterizedQuery();

	return t_result;
} // add_table_def_constraints
//---------------------------------------------------------------------------
bool TSMF::add_table_triggers(const std::string table_name)
{
	// Called from build_tables_list and ImportLookupTable
	// to set up TTableDef object
	bool t_result = true;
	std::string t_sql;

	TSQLQuery* qry = new TSQLQuery(FSQLDB);

	// And finally, get the trigger definitions
	t_sql  = "SELECT [sql] FROM sqlite_master ";
	t_sql += "WHERE [tbl_name] LIKE '";
	t_sql += table_name;
	t_sql += "' ";
	t_sql += "AND [type] LIKE 'trigger';";
	qry->SetSQLText((char*)t_sql.c_str());
	qry->SetIsActive(true);
	try
	{
		while (!qry->Eof())
		{
			FTableDef->AddTrigger(qry->GetFieldByName("sql"));
			qry->FetchNext();
		}
	}
	catch(...)
	{
		qry->SetIsActive(false);
		throw;
	}

	qry->SetIsActive(false);

	return t_result;
} // add_table_triggers
//---------------------------------------------------------------------------
bool TSMF::copy_table_defaults(TTableDef* old_table_def,
	TTableDef* new_table_def)
{
	// Called from RenameLookupTable
	bool t_result = true;
	int t_how_many, t_idx;

	t_how_many = old_table_def->GetDefaultCount();
	if (t_how_many > 0)
	{
		char t_default_def[200];
		char t_field_name[31];
		char t_default_value[100];
		size_t t_pos, t_dest, t_def_len;
		for (t_idx = 0; t_idx < t_how_many; t_idx++)
		{
			std::memset(t_default_def, '\x0', 200);
			std::strncpy(t_default_def, old_table_def->GetDefault(t_idx), 199);
			t_def_len = std::strlen(t_default_def);
			std::memset(t_field_name, '\x0', 31);
			std::memset(t_default_value, '\x0', 100);

			// this will look like
			//   default MYLABEL TO "ABC" for character fields, or
			//   default MYCODE TO 99  (i.e., no quotes) for numeric fields
			// skip over the leading "default " to find the column name

			t_pos = 8;
			t_dest = 0;
			while (t_default_def[t_pos] != ' ')
			{
				t_field_name[t_dest++] = t_default_def[t_pos++];
			}

			// t_pos will be pointing to this:
			//  " TO ..."
			//  Skip over the TO and the leading/trailing blanks

			t_pos += 4;
			// For this purpose, we don't want the quotes surrounding
			// the default value.
			t_dest = 0;
			while (t_pos < t_def_len)
			{
				if (t_default_def[t_pos] != 39)  // Alt+0039 = ' (apostrophe)
				{
					t_default_value[t_dest++] = t_default_def[t_pos];
				}

				t_pos++;
			}

			// Get the TFieldDef for this renamed table;
			TFieldDef* t_fld = new_table_def->GetColumn(t_field_name);
			if (t_fld)
			{
				t_fld->SetDefaultValue(t_default_value);
			}
		} // for (t_idx... iterating defaults
	} // if (t_how_many > 0)

	return t_result;
} // copy_table_defaults
//---------------------------------------------------------------------------
bool TSMF::copy_table_checks(TTableDef* old_table_def,
	TTableDef* new_table_def)
{
	// Called from RenameLookupTable
	bool t_result = true;
	int t_how_many, t_idx;

	// We also want to be able to return field-level CHECK
	// constraints, but ONLY if they are "limit to list"
	// such as "CHECK(EditsFunc IN ('E','S'))
	// or range controllers such as "CHECK(PrimarySite BETWEEN 'C000' AND 'C999')";
	t_how_many = old_table_def->GetCheckCount();
	if (t_how_many > 0)
	{
		std::string t_parse;
		char t_field_name[31];

		// Yes, I'm setting an artificial limit here, but
		// seriously... a thousand bites should be plenty.
		char t_check_constraint[1000];
		char t_limit_to_list[1000];
		std::string t_check_between;
		size_t t_pos,
			   t_range_pos,
			   t_dest,
			   t_check_len;

		for (t_idx = 0; t_idx < t_how_many; t_idx++)
		{
			std::memset(t_field_name, '\x0', 31);
			std::memset(t_check_constraint, '\x0', 1000);
			std::strncpy(t_check_constraint, old_table_def->GetCheck(t_idx), 999);
			t_check_len = std::strlen(t_check_constraint);

			// Look at the first characters following "check"
			// We want to advance to the beginning of the
			// field name
			t_pos = 5;
			while (t_pos < t_check_len)
			{
				char t_char = t_check_constraint[t_pos];
				if ((t_char == ' ') || (t_char == '('))
				{
					t_pos++;
				}
				else
				{
					break;
				}
			}

			if (t_pos >= t_check_len)
				continue;

			t_dest = 0;
			while (t_pos < t_check_len)
			{
				// Now we'll keep going until we have the field name
				char t_char = t_check_constraint[t_pos];
				if ((t_char == ' ') || (t_char == '('))
				{
					break;
				}

				t_field_name[t_dest++] = t_char;
				t_pos++;
			}

			// Look for a CHECK...BETWEEN constraint, first.
			t_check_between.assign(old_table_def->GetCheck(t_idx));
			lowercase(t_check_between);
			t_range_pos = t_check_between.find("between");
			if (t_range_pos != std::string::npos)
			{
				// Get the TFieldDef;
				TFieldDef* t_old_fld = old_table_def->GetColumn(t_field_name);
				if (t_old_fld)
				{
					char t_ranger[100];
					std::memset(t_ranger, '\x0', 100);
					size_t t_range_len = t_check_between.length(),
						   t_dest = 0;

					t_range_pos += 7; // first character past "between"

					// skip over white space and (if this is a character
					// field) the opening apostrophe to the first range
					while ((t_range_pos < t_range_len) &&
						   (t_check_between[t_range_pos] == 32 ||    // ' '
							t_check_between[t_range_pos] ==  9 ||    // '\t'
							t_check_between[t_range_pos] == 13 ||    // '\r'
							t_check_between[t_range_pos] == 10 ||    // '\n'
							t_check_between[t_range_pos] == 39))     // apostrophe
					{
						t_range_pos++;
					}

					// Now start loading t_ranger
					while ((t_range_pos < t_range_len) &&
						   (t_dest < 100) &&
						   (t_check_between[t_range_pos] != 39))
					{
						// t_check_between has been lowercased.
						// return the literal: t_check_constraint
						t_ranger[t_dest++] = t_check_constraint[t_range_pos++];
					}

					// Place guillemet between the low and high values
					// of the range (for easier parsing at the receiving end)
					if (t_dest < 100)
						t_ranger[t_dest++] = ''; // Alt+0187

					// Now we want to step over the closing apostrophe
					// to the low end of the range, the white space and " AND "
					while ((t_range_pos < t_range_len) &&
						   (t_check_between[t_range_pos] == 32 ||    // ' '
							t_check_between[t_range_pos] ==  9 ||    // '\t'
							t_check_between[t_range_pos] == 13 ||    // '\r'
							t_check_between[t_range_pos] == 10 ||    // '\n'
							t_check_between[t_range_pos] == 39))     // apostrophe
					{
						t_range_pos++;
					}

					t_range_pos += 3; // AND

					// Skip over white space and opening apostrophe
					// to the high end of the range
					while ((t_range_pos < t_range_len) &&
						   (t_check_between[t_range_pos] == 32 ||    // ' '
							t_check_between[t_range_pos] ==  9 ||    // '\t'
							t_check_between[t_range_pos] == 13 ||    // '\r'
							t_check_between[t_range_pos] == 10 ||    // '\n'
							t_check_between[t_range_pos] == 39))     // apostrophe
					{
						t_range_pos++;
					}

					// Finish loading t_ranger
					while ((t_range_pos < t_range_len) &&
						   (t_dest < 100) &&
						   (t_check_between[t_range_pos] != 39))
					{
						// t_check_between has been lowercased.
						// return the literal: t_check_constraint
						t_ranger[t_dest++] = t_check_constraint[t_range_pos++];
					}

					// Get the TFieldDef for this renamed table;
					TFieldDef* t_fld = new_table_def->GetColumn(t_field_name);
					if (t_fld)
					{
						t_fld->SetCheckBetween(t_ranger);
					}
				}

				// POSSIBLY dangerous assumption, in the horrible event
				// somebody embeds the characters "between" in the middle
				// of a limit-to-list check constraint, but we'll nevertheless
				// make the assumption that isn't going to happen... and
				// say that this check constraint is done, done, done.
				continue;
			}

			t_parse.assign(old_table_def->GetCheck(t_idx));
			lowercase(t_parse);
			// Syntax might read " in(" or " in ("
			t_pos = t_parse.find(" in(");
			if (t_pos == std::string::npos)
			{
				t_pos = t_parse.find(" in ");
				if (t_pos == std::string::npos)
				{
					// This is not a limit-to-list check constraint;
					// move on
					continue;
				}
			}

			// Okay, we have one. Get it from old_table_def so that
			// we don't mess with capitalization of the list

			std::memset(t_limit_to_list, '\x0', 1000);

			// Advance t_pos to the first character following the
			// open parenthesis that follows "IN".

			while (t_pos < t_check_len)
			{
				if (t_check_constraint[t_pos] == '(')
				{
					t_pos++;
					break;
				}
				t_pos++;
			}
			t_dest = 0;

			while (t_pos < t_check_len)
			{
				if (t_check_constraint[t_pos] == ')')
				{
					// We're done. Get out.
					break;
				}
				t_limit_to_list[t_dest++] = t_check_constraint[t_pos++];
			}

			// Get the TFieldDef;
			TFieldDef* t_fld = new_table_def->GetColumn(t_field_name);
			if (t_fld)
			{
				t_fld->SetLimitToList(t_limit_to_list);
			}
		} // for (iterating check constraints)
	}

	return t_result;
} // copy_table_checks
//---------------------------------------------------------------------------

/*  Bizarre!! This runs once, then crashes on the BindTextParameter for the next table
bool TSMF::add_table_triggers(const std::string table_name)
{
	// Called from build_tables_list and ImportLookupTable
	// to set up TTableDef object

	bool t_result = true;

	// FAddTableDefTriggersQry
	if (!FAddTableDefTriggersQry)
	{
		// Called from build_tables_list and ImportLookupTable
		// to set up TTableDef object
		std::string t_sql  = "SELECT [sql] FROM sqlite_master ";
		t_sql += "WHERE [tbl_name] LIKE ? ";
		t_sql += "AND [type] LIKE 'trigger';";

		FAddTableDefTriggersQry = new TSQLQuery(FSQLDB);
		FAddTableDefTriggersQry->SetSQLText(t_sql);

		if (!FAddTableDefTriggersQry->PrepareParameterizedQuery())
		{
			delete FAddTableDefTriggersQry;
			FAddTableDefTriggersQry = NULL;
			return false;
		}
	}

	try
	{
		FAddTableDefTriggersQry->BindTextParameter(1, table_name.c_str());

		if (FAddTableDefTriggersQry->OpenParameterizedQuery())
		{
			do
			{
				FTableDef->AddTrigger(FAddTableDefTriggersQry->GetFieldByNumber(0));
				FAddTableDefTriggersQry->FetchNext();
			} while (!FAddTableDefTriggersQry->Eof());
		}
		else
		{
			t_result = false;
		}

		FAddTableDefIndexesQry->ResetParameterizedQuery();
	}
	catch(std::runtime_error& e)
	{
		std::string t_error = e.what();
		FEngineErrorsPump->AddError("TSMF","add_table_triggers", t_error);
		t_error = FAddTableDefTriggersQry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","add_table_triggers", t_error);

		t_result = false;
	}
	return t_result;
} // add_table_triggers
//---------------------------------------------------------------------------
*/
/******************************************************************************
 *** Retrieve primary keys using callback functions
 ******************************************************************************/
int TSMF::GetPKeys( const char* table_name, const char* order_by,
	void* owner, void* callback_func )
{
	int t_result = 0;

	std::string t_sql = "SELECT PKey FROM ";
	t_sql += table_name;
	t_sql += " ORDER BY ";
	if (order_by && (strlen(order_by) > 0))
	{
		t_sql += order_by;
	}
	else
	{
		t_sql += "PKey";
	}

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	t_qry->SetSQLText(t_sql);
	try
	{
		t_qry->SetIsActive(true);
		if (!t_qry->Eof())
		{
			CallbackPKeys t_callback_func = (CallbackPKeys)callback_func;
			int t_pkey;
			while (!t_qry->Eof())
			{
				t_pkey = t_qry->GetIntFieldByName("PKey");
				(*t_callback_func)(owner, table_name, t_pkey);
				t_qry->FetchNext();
			}
		}
		else
		{
        	t_result = -2; // notification: result set is empty
		}
	}
	catch(std::runtime_error& e)
	{
		std::string t_error = e.what();
		FEngineErrorsPump->AddError("TSMF","GetPKeys", t_error);
		t_error = (std::string)"SQL error for table " +
				  (std::string)table_name +
				  (std::string)": " +
				  (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","GetPKeys", t_error);

		t_result = -1;
	}

	return t_result;

} // GetPKeys
//---------------------------------------------------------------------------
int TSMF::CheckEditSetAgainstLayout( const int edit_set_pkey,
	const int layout_pkey, void* owner, void* callback_func )
{
	int t_result = 0;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[1000];
/*
SELECT efp.FieldPKey, f.FieldName, esep.EditPKey, e.EditName, e.EditTag
FROM EE_EditFieldsPivot efp, EE_Fields f, EE_EditSetEditsPivot esep, EE_Edits e
WHERE efp.FieldPKey = f.PKey
AND efp.EditPKey = esep.EditPKey
AND e.PKey = efp.EditPKey
AND esep.EditSetPKey = 10
AND efp.FieldPKey NOT IN (
   SELECT lfp.FieldPKey
   FROM EE_LayoutFieldsPivot lfp
   WHERE lfp.LayoutPKey = 1
)
*/
	std::snprintf(t_buffer, 1000,
		  "SELECT efp.FieldPKey, f.FieldName, esep.EditPKey, e.EditName, e.EditTag " \
		  "FROM EE_EditFieldsPivot efp, EE_Fields f, EE_EditSetEditsPivot esep, EE_Edits e " \
		  "WHERE efp.FieldPKey = f.PKey " \
		  "  AND efp.EditPKey = esep.EditPKey " \
		  "  AND e.PKey = efp.EditPKey " \
		  "  AND esep.EditSetPKey = %d " \
		  "  AND efp.FieldPKey NOT IN ( " \
		  "    SELECT lfp.FieldPKey " \
		  "	   FROM EE_LayoutFieldsPivot lfp " \
		  "	   WHERE lfp.LayoutPKey = %d); ",
		  edit_set_pkey, layout_pkey);

	try
	{
		t_qry->SetSQLText(t_buffer);

		t_qry->SetIsActive(true);
		if (!t_qry->Eof())
		{
			CallbackCheckEditSetAgainstLayout t_callback_func =
				(CallbackCheckEditSetAgainstLayout)callback_func;

			int t_field_pkey,
				t_edit_pkey;

			std::string t_field_name,
						t_edit_name,
						t_edit_tag;
			while (!t_qry->Eof())
			{
				t_field_pkey = t_qry->GetIntFieldByName("FieldPKey");
				t_field_name = t_qry->GetFieldByName("FieldName");
				t_edit_pkey = t_qry->GetIntFieldByName("EditPKey");
				t_edit_name = t_qry->GetFieldByName("EditName");
				t_edit_tag = t_qry->GetFieldByName("EditTag");

				(*t_callback_func)(owner, t_field_pkey, t_field_name.c_str(),
				    t_edit_pkey, t_edit_name.c_str(), t_edit_tag.c_str());
				t_qry->FetchNext();
			}
		}
		else
		{
        	t_result = -2; // notification: result set is empty
		}
	}
	catch(std::runtime_error& e)
	{
		std::string t_error = e.what();
		FEngineErrorsPump->AddError("TSMF","CheckEditSetAgainstLayout", t_error);
		t_error = (std::string)"SQL error: " +
				  (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","CheckEditSetAgainstLayout", t_error);

		t_result = -1;
	}


	return t_result;
} // CheckEditSetAgainstLayout
//---------------------------------------------------------------------------

int TSMF::GetTableNames( void* owner, void* callback_func )
{
	int t_result = 0;

	CallbackString t_callback_func = (CallbackString)callback_func;
	std::map<std::string, TTableDef*> t_tables_list =
		FTablesList->TablesList();

	std::map<std::string, TTableDef*>::iterator it = t_tables_list.begin();
	while (it != t_tables_list.end())
	{
		TTableDef* t_table = (TTableDef*)(*it).second;
		(*t_callback_func)(owner, t_table->GetTableName());
		it++;
	}
	return t_result;
} // GetTableNames
//---------------------------------------------------------------------------
int TSMF::GetTablesCount(int* how_many_tables)
{
	*how_many_tables = FTablesList->Count();
	return 0;
} // GetTablesCount()
//---------------------------------------------------------------------------
int TSMF::GetTableNameByIndex( const int idx, char* buffer,
	const int buff_size )
{
	int t_how_many = FTablesList->Count();

	if ((idx >= 0) && (idx < t_how_many))
	{
		int t_count = 0;
		std::map<std::string, TTableDef*> t_tables_list =
			FTablesList->TablesList();
		std::map<std::string, TTableDef*>::iterator it = t_tables_list.begin();
		while ((it != t_tables_list.end()) && (t_count <= idx))
		{
			if (t_count == idx)
			{
				TTableDef* t_table = (TTableDef*)(*it).second;

				std::strncpy(buffer, t_table->GetTableName(), buff_size);
				break;
			}
			it++;
			t_count++;
		}

		return 0;
	}
	else
	{
		std::string t_error;

		char t_buffer[100];
		std::snprintf(t_buffer, 100, "Parameter idx (%d) must be in range 0 - %d",
			idx, t_how_many);

		t_error.assign(t_buffer);

		FEngineErrorsPump->AddError("TSMF", "GetTableNameByIndex", t_error);
		return -1;
	}
} // GetTableNameByIndex
//---------------------------------------------------------------------------
int TSMF::GetIndexCount( const char* table_name, int* how_many_indexes )
{
	int t_result = 0;
	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		*how_many_indexes = t_table->GetIndexCount();
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += (std::string)table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetIndexCount", t_error);

		t_result = -1;
	}

	return t_result;
} // GetIndexCount
//---------------------------------------------------------------------------
int TSMF::GetIndex( const char* table_name, const int idx,
	char* index_sql, const int buff_size )
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	// let's at a minimum ensure the index_sql goes back to the
	// caller as a null-terminated string
	index_sql[0] = '\x0';
	if (t_table)
	{
		std::string t_sql;
		t_sql.assign(t_table->GetIndex(idx));
		std::strncpy(index_sql, (char*)t_sql.c_str(), buff_size);
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += (std::string)table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetIndex", t_error);

		t_result = -1;
	}

	return t_result;
} // GetIndex
//---------------------------------------------------------------------------
int TSMF::GetTriggerCount( const char* table_name, int* how_many_triggers )
{
	int t_result = 0;
	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		*how_many_triggers = t_table->GetTriggerCount();
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += (std::string)table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetTriggerCount", t_error);

		t_result = -1;
	}

	return t_result;
} // GetTriggerCount
//---------------------------------------------------------------------------
int TSMF::GetTrigger( const char* table_name, const int idx,
	char* trigger_sql, const int buff_size )
{
	int t_result = 0;
	TTableDef* t_table = FTablesList->GetTable(table_name);
	// let's at a minimum ensure the trigger_sql goes back to the
	// caller as a null-terminated string
	trigger_sql[0] = '\x0';
	if (t_table)
	{
		std::string t_sql;
		t_sql.assign(t_table->GetTrigger(idx));
		std::strncpy(trigger_sql, (char*)t_sql.c_str(), buff_size);
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += (std::string)table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetTrigger", t_error);

		t_result = -1;
	}

	return t_result;
} // GetTrigger
//---------------------------------------------------------------------------
std::string TSMF::field_data_type_as_string(TFieldDef* fld)
{
	std::string t_data_type;
	if (fld)
	{
		char t_buffer[20];
		switch (fld->GetDataType())
		{
			case edtInteger:
				t_data_type.assign("integer");
				break;
			case edtDouble:
				t_data_type.assign("double");
				break;
			case edtVarchar:
				std::sprintf(t_buffer, "varchar(%d)", fld->GetColumnSize());
				t_data_type.assign(t_buffer);
				break;
			case edtChar:
				std::sprintf(t_buffer, "char(%d)", fld->GetColumnSize());
				t_data_type.assign(t_buffer);
				break;
			case edtMemo:
				t_data_type.assign("text");
				break;
			case edtDateTime:
				t_data_type.assign("datetime");
				break;
			default:
				t_data_type.assign("unknown");
				break;
		}
	}

	return t_data_type;

} // field_data_type_as_string
//---------------------------------------------------------------------------

std::string TSMF::field_nullable_as_string(TFieldDef* fld)
{
	std::string t_nullable;

	if (fld)
	{
		if (fld->GetIsNullable())
			t_nullable.assign("NULL");
		else
			t_nullable.assign ("NOT NULL");
	}

	return t_nullable;
} // field_nullable_as_string
//---------------------------------------------------------------------------

int TSMF::GetTableStructure(const char* table_name,
	void* owner, void* callback_func )
{
	int t_result = 0;

	CallbackTableStructure t_callback_func = (CallbackTableStructure)callback_func;
	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		std::string t_data_type,
					t_nullable,
					t_default_value,
					t_limit_to_list,
					t_check_between;
		int t_col_count = t_table->GetColumnCount();
		for (int t_idx = 0; t_idx < t_col_count; t_idx++)
		{
			TFieldDef* t_fld = t_table->GetColumn(t_idx);

			t_data_type.assign(field_data_type_as_string(t_fld));

			t_nullable.assign(field_nullable_as_string(t_fld));

			int t_is_trigger = (t_fld->GetIsTriggerField() ? 1 : 0);

			t_default_value.assign(t_fld->GetDefaultValue());

			t_limit_to_list.assign(t_fld->GetLimitToList());

			t_check_between.assign(t_fld->GetCheckBetween());

			// Note that we return t_idx+1 for the column order so
			// that the first column (field) in the structure is
			// numbered 1, not 0.
			(*t_callback_func)(owner, table_name, t_idx+1,
				t_fld->GetFieldName(), t_data_type.c_str(), t_nullable.c_str(),
				t_is_trigger, t_default_value.c_str(), t_limit_to_list.c_str(),
				t_check_between.c_str());
		}
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "GetTableStructure", t_error);
		t_result = -1;
	}

	return t_result;
} // GetTableStructure
//---------------------------------------------------------------------------
int TSMF::GetColumnsCount(const char* table_name, int* how_many_columns)
{
	int t_result = 0;
	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		*how_many_columns = t_table->GetColumnCount();
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "GetColumnsCount", t_error);
		t_result = -1;
	}
	return t_result;
} // GetTablesCount()
//---------------------------------------------------------------------------
int TSMF::GetColumnStructure( const char* table_name, int column_order,
	char* field_name, int fldname_size,
	char* data_type, int type_size,
	char* nullability, int null_size,
	int* is_trigger_field)
{
	int t_result = 0;
	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		if ((column_order > 0) && (column_order <= t_table->GetColumnCount()))
		{
			// column_order is 1-based, meaning that the 0th field in the
			// t_table field list will be referenced as 1.
			TFieldDef* t_fld = t_table->GetColumn(column_order-1);
			if (t_fld)
			{
				std::string t_data_type, t_nullable;
				t_data_type.assign(field_data_type_as_string(t_fld));
				t_nullable.assign(field_nullable_as_string(t_fld));

				memset(field_name, '\x0', fldname_size);
				memset(data_type, '\x0', type_size);
				memset(nullability, '\x0', null_size);
				strncpy(field_name, t_fld->GetFieldName(), fldname_size-1);
				strncpy(data_type, (char*)t_data_type.c_str(), type_size-1);
				strncpy(nullability, (char*)t_nullable.c_str(), null_size-1);
				*is_trigger_field = (t_fld->GetIsTriggerField() ? 1 : 0);
			}
		}
		else
		{
			std::string t_error = (std::string)"In table " + (std::string)table_name +
			(std::string)", column_order " + intToString(column_order) +
			(std::string)" is out of range";
			FEngineErrorsPump->AddError("TSMF", "GetColumnStructure", t_error);
			t_result = -1;
		}
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "GetColumnStructure", t_error);
		t_result = -1;
	}
	return t_result; // error condition
} // GetColumnStructure
//---------------------------------------------------------------------------
// Other service functions:
// Give me a table and field name, and I'll tell you its column order
int TSMF::GetColumnOrder(const char* table_name, const char* field_name,
	int* col_order )
{
	int t_result = 0;

	// Default the col_order to zero
	*col_order = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		// The lookup into the fields list is to verify that the table
		// actually HAS a field by this name
		TFieldDef* t_fld = t_table->GetColumn(field_name);
		if (t_fld)
		{
			*col_order = t_fld->GetColumnOrder();
		}
		else
		{
			std::string t_error = (std::string)"Table \"" +
								  (std::string)table_name +
								  (std::string)"\" has no field named \"" +
								  (std::string)field_name +
								  (std::string)"\"";
			FEngineErrorsPump->AddError("TSMF", "GetColumnOrder", t_error);
			t_result = -1;
		}
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "GetColumnStructure", t_error);
		t_result = -1;
	}

	return t_result;

} // GetColumnOrder
//---------------------------------------------------------------------------
// Give me a table and column order, I'll tell you its field name
int TSMF::GetFieldName(const char* table_name, const int col_order,
	char* field_name, const int fldname_size )
{
	int t_result = 0;
	if (fldname_size > 0)
	{
		// default the field_name to an empty string
		field_name[0] = '\0';

		TTableDef* t_table = FTablesList->GetTable(table_name);
		if (t_table)
		{
			TFieldDef* t_fld = t_table->GetColumn(col_order);
			if (t_fld)
			{
				strncpy(field_name, t_fld->GetFieldName(), fldname_size-1);
			}
			else
			{
				std::string t_error = (std::string)"In table " + (std::string)table_name +
				(std::string)", col_order " + intToString(col_order) +
				(std::string)" is out of range";
				FEngineErrorsPump->AddError("TSMF", "GetFieldName", t_error);
				t_result = -1;
			}
		}
		else
		{
			std::string t_error = "Cannot find table named \"";
			t_error += table_name;
			t_error += "\"";
			FEngineErrorsPump->AddError("TSMF", "GetFieldName", t_error);
			t_result = -1;
		}
	}

	return t_result;
} // GetFieldName
//---------------------------------------------------------------------------

/*****************************************************************************
 *** Functions for retrieving data
 *****************************************************************************/
int TSMF::ReadData(const char* table_name, int pkey,
	void* owner, void* callback_func )
{
	int t_result = 0;
	char t_buffer[200];

	std::snprintf(t_buffer, 200, "SELECT * FROM [%s] WHERE PKey = %d;", table_name, pkey);
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	t_qry->SetSQLText(t_buffer);
	t_qry->SetIsActive(true);
	if (!t_qry->Eof())
	{
		TTableDef* t_table = FTablesList->GetTable(table_name);
		if (t_table)
		{
			CallbackReadData t_callback_func = (CallbackReadData)callback_func;

			std::string t_field_name, t_data_type, t_data_string;
			int t_col_count = t_table->GetColumnCount();
			for (int t_idx = 0; t_idx < t_col_count; t_idx++)
			{
				TFieldDef* t_fld = t_table->GetColumn(t_idx);
				if (t_fld)
				{
					// Meta data
					t_field_name.assign(t_fld->GetFieldName());
					t_data_type.assign(field_data_type_as_string(t_fld));
					// Data from the row identified by parameter pkey
					t_data_string.assign(t_qry->GetFieldByNumber(t_idx));

					// Send it off to the callback function. Increment
					// t_idx in the third parameter, because it represents
					// the "column order", which is 1-based for humans.
					(*t_callback_func)(owner, table_name, pkey, (t_idx+1),
						t_field_name.c_str(),
						t_data_type.c_str(),
						t_data_string.c_str());
				}
				else
				{
					std::string t_error = (std::string)"Error in Engine! Field index out of range for table " +
						(std::string)table_name;
					FEngineErrorsPump->AddError("TSMF", "ReadData", t_error);
					t_result = -1;
					break;
				}
			}
		}
		else
		{
			std::string t_error = "Cannot find table named \"";
			t_error += table_name;
			t_error += "\"";
			FEngineErrorsPump->AddError("TSMF", "ReadData", t_error);
			t_result = -1;
		}
	} // if (!t_qry->Eof())

	return t_result;
} // ReadData
//---------------------------------------------------------------------------

int TSMF::RefreshReadData( const char* table_name, const int pkey )
{
	int t_result = 0;

	char t_buffer[100];
	std::snprintf(t_buffer, 100, "SELECT * FROM [%s] WHERE PKey = %d;",
		table_name, pkey);
	// We will use lowercase table name in FQueriesList
	std::string t_table_name;
	t_table_name.assign(table_name);
	lowercase(t_table_name);

	TSQLQuery* t_qry = FQueriesList[t_table_name];

	if (!t_qry)
	{
		t_qry = new TSQLQuery(FSQLDB);
		FQueriesList[t_table_name] = t_qry;
	}
	else
	{
		// We have the query, but want to get a fresh snapshot.
		t_qry->SetIsActive(false);
	}

	try
	{
		t_qry->SetSQLText(t_buffer);
		t_qry->SetIsActive(true);
	}
	catch (...)
	{
		std::string t_error = (std::string)"SQL: " + (std::string)t_buffer;
		FEngineErrorsPump->AddError("TSMF","RefreshReadData", t_error);
		t_error = t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","RefreshReadData", t_error);
		t_result = -1;

	}

	return t_result;

} // RefreshReadData
//---------------------------------------------------------------------------

std::string TSMF::get_data_by_column_order( const char* table_name, int pkey,
	const int column_order)
{
	std::string t_data;
	TTableDef* t_table = FTablesList->GetTable(table_name);

	if (t_table)
	{
		if ((column_order > 0) && (column_order <= t_table->GetColumnCount()))
		{
			// column_order is 1-based, meaning that the 0th field in the
			// t_table field list will be referenced as 1.
			TFieldDef* t_fld = t_table->GetColumn(column_order-1);
			if (t_fld)
			{
				std::string t_field_name = t_fld->GetFieldName();
				char t_buffer[100];
				std::snprintf(t_buffer, 100, "SELECT * FROM [%s] WHERE PKey = %d;",
					table_name, pkey);

				// We will use lowercase table name in FQueriesList
				std::string t_table_name;
				t_table_name.assign(table_name);
				lowercase(t_table_name);

				TSQLQuery* t_qry = FQueriesList[t_table_name];

				if (!t_qry)
				{
					// We have to create a query into this table
					t_qry = new TSQLQuery(FSQLDB);

					// Add this to the FQueriesList so that the destructor
					// can be sure to delete it.
					FQueriesList[t_table_name] = t_qry;

					t_qry->SetSQLText(t_buffer);
					t_qry->SetIsActive(true);
				}
				else
				{
					// The query exists. Is it already holding the data
					// for the record we're interested in?
					int t_test = t_qry->GetIntFieldByName("PKey");
					if (t_test != pkey)
					{
						// The caller has changed the PKey. Reset the
						// query.
						std::snprintf(t_buffer, 100, "SELECT * FROM [%s] WHERE PKey = %d;",
							table_name, pkey);
						t_qry->SetIsActive(false);
						t_qry->SetSQLText(t_buffer);
						t_qry->SetIsActive(true);
					}
					// else { we're already on the right record }
				}
				if (!t_qry->Eof())
				{
					t_data.assign(t_qry->GetFieldByName(t_field_name));
				}
			}
		}
		else
		{
			std::string t_error = (std::string)"In table " + (std::string)table_name +
			(std::string)", column_order " + intToString(column_order) +
			(std::string)" is out of range";
			FEngineErrorsPump->AddError("TSMF", "get_data_by_column_order", t_error);
		}
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "get_data_by_column_order", t_error);
	}
	return t_data;
} // get_data_by_column_order
//---------------------------------------------------------------------------
int TSMF::GetDataSizeByColumnOrder( const char* table_name, const int pkey,
	const int column_order, int* data_size )
{
	int t_result = 0;

	std::string t_data = get_data_by_column_order(table_name, pkey, column_order);
	int t_data_size = (int)t_data.size();

	if (t_data_size > 0)
	{
		// The purpose of the call to this function is to find out
		// how big to make an array of char. I'm going to pad this by
		// one to ensure the caller creates a buffer that is big enough
		// to hold a null terminator.
		*data_size = t_data_size +1;
	}
	else
	{
		*data_size = 0;
	}

	return t_result;
} // GetDataSizeByColumnOrder
//---------------------------------------------------------------------------
int TSMF::GetDataByColumnOrder( const char* table_name, const int pkey,
	const int column_order, char* buffer, const int buff_size )
{
	int t_result = 0;
	if (buff_size > 0)
	{
		// Whatever else may happen here, let's ensure a null-terminated string
		// goes back to the caller.
		buffer[0] = '\x0';

		std::string t_data = get_data_by_column_order(table_name, pkey, column_order);
		size_t t_data_size = t_data.size();

		if (t_data_size > 0)
		{
			strncpy(buffer, (char*)t_data.c_str(), buff_size );
		}
		else
		{
			std::string t_error = "Failure to retrieve data from table \"";
			t_error += table_name;
			t_error += "\", column number ";
			t_error += intToString(column_order);
			FEngineErrorsPump->AddError("TSMF", "GetDataByColumnOrder", t_error);
			t_result = -1; // messages were set in get_data_by_column_order
		}
	}
	return t_result;
} // GetDataByColumnOrder
//---------------------------------------------------------------------------
std::string TSMF::get_data_by_field_name( const char* table_name, int pkey,
	const char* field_name)
{
	std::string t_data;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		// The lookup into the fields list is to verify that the table
		// actually HAS a field by this name
		TFieldDef* t_fld = t_table->GetColumn(field_name);
		if (t_fld)
		{
			char t_buffer[100];
			std::snprintf(t_buffer, 100, "SELECT * FROM [%s] WHERE PKey = %d;",
				table_name, pkey);
			// We will use lowercase table name in FQueriesList
			std::string t_table_name;
			t_table_name.assign(table_name);
			lowercase(t_table_name);

			TSQLQuery* t_qry = FQueriesList[t_table_name];

			if (!t_qry)
			{
				// We have to create a query into this table
				t_qry = new TSQLQuery(FSQLDB);

				FQueriesList[t_table_name] = t_qry;

				t_qry->SetSQLText(t_buffer);
				t_qry->SetIsActive(true);
			}
			else
			{
				// The query exists. Is it already holding the data
				// for the record we're interested in?
				int t_test = t_qry->GetIntFieldByName("PKey");
				if (t_test != pkey)
				{
					// The caller has changed the PKey. Reset the
					// query.
					std::snprintf(t_buffer, 100, "SELECT * FROM [%s] WHERE PKey = %d;",
						table_name, pkey);
					t_qry->SetIsActive(false);
					t_qry->SetSQLText(t_buffer);
					t_qry->SetIsActive(true);
				}
				// else { we're already on the right record }
			}
			if (!t_qry->Eof())
			{
				t_data.assign(t_qry->GetFieldByName(field_name));
			}

			// IMPORTANT: Finalize this query so that DDL can occur on the table
			// if necessary (see the DROP TABLE statement in RenameLookupTable).
			t_qry->SetIsActive(false);
		} // if (t_fld)
		// else { Write error message to errors pump saying cannot get t_fld? }
		else
		{
			std::string t_error = "Cannot find field named \"";
			t_error += field_name;
			t_error += "\" in table named \"";
			t_error += table_name;
			t_error += "\"";
			FEngineErrorsPump->AddError("TSMF", "get_data_by_field_name", t_error);
		}
	} // if (t_table)
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "get_data_by_field_name", t_error);
	}

	return t_data;
} // get_data_by_field_name
//---------------------------------------------------------------------------
int TSMF::GetDataSizeByFieldName( const char* table_name, const int pkey,
	const char* field_name, int* data_size )
{
	int t_result = 0;

	std::string t_data = get_data_by_field_name(table_name, pkey, field_name);
	int t_data_size = (int)t_data.size();

	if (t_data_size > 0)
	{
		// The purpose of the call to this function is to find out
		// how big to make an array of char. I'm going to pad this by
		// one to ensure the caller creates a buffer that is big enough
		// to hold a null terminator.
		*data_size = t_data_size +1;
	}
	else
	{
		*data_size = 0;
	}

	return t_result;
} // GetDataSizeByFieldName
//---------------------------------------------------------------------------
int TSMF::GetDataByFieldName( const char* table_name, const int pkey,
	const char* field_name, char* buffer, const int buff_size )
{
	int t_result = 0;
	if (buff_size > 0)
	{
		// Whatever else may happen here, let's ensure a null-terminated string
		// goes back to the caller.
		buffer[0] = '\x0';

		std::string t_data = get_data_by_field_name(table_name, pkey, field_name);
		size_t t_data_size = t_data.size();

		if (t_data_size > 0)
		{
			strncpy(buffer, (char*)t_data.c_str(), buff_size );
		}
		else
		{
			std::string t_err = "Failure to retrieve data from ";
			t_err += (std::string)table_name + "." + (std::string)field_name;
			FEngineErrorsPump->AddError("TSMF", "GetDataByFieldName", t_err);
			t_result = -1;
		}
	}
	return t_result;
} // GetDataByFieldName
//---------------------------------------------------------------------------
/*****************************************************************************
 *** Cursor-access functions
 *****************************************************************************/
 bool TSMF::sql_is_valid(std::string& sql, std::string& err_msg)
 {
	// Called from both OpenCursor and PrepareParameterizedCursor

	// Note 1: sql is passed by reference, so this is modifying the source
	uppercase(sql);

	// Note 2: It is legal, though not encouraged, to send sql
	// that contains multiple statements. The results are unpredictable,
	// but just in case somebody tries to sneak a write statement through
	// behind a SELECT statement, we search the WHOLE string for offending
	// terminology.
	size_t t_pos = sql.find("SELECT ");
	if (t_pos == std::string::npos)
	{
		err_msg.assign("SQL does not contain a SELECT statement");
		return false;
	}
	t_pos = sql.find("UPDATE ");
	if (t_pos != std::string::npos)
	{
		err_msg.assign("SQL contains an UPDATE statement; not allowed");
		return false;
	}
	t_pos = sql.find("INSERT ");
	if (t_pos != std::string::npos)
	{
		err_msg.assign("SQL contains an INSERT statement; not allowed");
		return false;
	}
	t_pos = sql.find("DELETE ");
	if (t_pos != std::string::npos)
	{
		err_msg.assign("SQL contains a DELETE statement; not allowed");
		return false;
	}
	// If we made it to here, we're good to continue
	return true;
 } // sql_is_valid
//---------------------------------------------------------------------------
bool TSMF::cursor_name_exists(std::string& cursor_name, std::string& err_msg)
{
	// Called from both OpenCursor and PrepareParameterizedCursor
	// Does a query already exist using this cursor_name? If so, should
	// we warn the user to close it, or just delete it and open a new one?

	// NOTE: Both parameters are passed by reference, and are altered here.
	TSQLQuery* t_qry = get_cursor_from_list(cursor_name);
	if (t_qry)
	{
		err_msg = "A cursor named \"";
		err_msg += cursor_name;
		err_msg += "\" already exists.";
		return true;
	}

	return false;
} // cursor_name_exists
//---------------------------------------------------------------------------
TSQLQuery* TSMF::get_cursor_from_list(std::string& cursor_name)
{
	// NOTE: cursor_name is passed by reference and altered here
	lowercase(cursor_name);

	TSQLQuery* t_qry = FCursorsList[cursor_name];

	return t_qry;

} // get_cursor_from_list
//---------------------------------------------------------------------------
int TSMF::OpenCursor( const char* cursor_name, const char* sql)
{
	int t_result = 0;
	std::string t_cursor_name, t_sql, t_error;

	t_sql.assign(sql);
	// Note: t_sql will return converted to UPPERCASE.
	if (!sql_is_valid(t_sql, t_error))
	{
		FEngineErrorsPump->AddError("TSMF","OpenCursor", t_error);
		FEngineErrorsPump->AddError("TSMF","OpenCursor", sql);
		return -1; // error: SQL is not valid
	}

	t_cursor_name.assign(cursor_name);
	// Note: t_cursor_name will return converted to LOWERCASE
	if (cursor_name_exists(t_cursor_name, t_error))
	{
		FEngineErrorsPump->AddError("TSMF","OpenCursor", t_error);
		FEngineErrorsPump->AddError("TSMF","OpenCursor", cursor_name);
		FEngineErrorsPump->AddError("TSMF","OpenCursor", sql);
		return -1; // error: SQL is not valid
	}

	// This is a new cursor...
	TSQLQuery* t_qry = new TSQLQuery(FSQLDB);
	t_qry->SetSQLText(t_sql.c_str());
	try
	{
		t_qry->SetIsActive(true);
		// if we don't crash, add this query to the FCursorsList
		FCursorsList[t_cursor_name] = t_qry;
	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		FEngineErrorsPump->AddError("TSMF","OpenCursor", t_error);
		t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","OpenCursor", t_error);
		FEngineErrorsPump->AddError("TSMF","OpenCursor", sql);

		delete t_qry;

		t_result = -1;
	}

	return t_result;
} // OpenCursor
//---------------------------------------------------------------------------
int TSMF::CloseCursor( const char* cursor_name )
{
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","CloseCursor", t_error);
		return -1; // cannot find cursor by that name
	}

	// Close and delete this cursor query
	t_qry->SetIsActive(false);
	delete t_qry;
	// and erase it from our map.
	FCursorsList.erase(t_cursor_name);

	return 0;
} // CloseCursor
//---------------------------------------------------------------------------
int TSMF::PrepareParameterizedCursor( const char* cursor_name, const char* sql)
{
	int t_result = 0;
	std::string t_cursor_name, t_sql, t_error;

	t_sql.assign(sql);
	// Note: t_sql will return converted to UPPERCASE.
	if (!sql_is_valid(t_sql, t_error))
	{
		FEngineErrorsPump->AddError("TSMF","PrepareParameterizedQuery", t_error);
		return -1; // error: SQL is not valid
	}

	t_cursor_name.assign(cursor_name);
	// Note: t_cursor_name will return converted to LOWERCASE
	if (cursor_name_exists(t_cursor_name, t_error))
	{
		FEngineErrorsPump->AddError("TSMF","PrepareParameterizedQuery", t_error);
		return -1; // error: SQL is not valid
	}

	// This is a new cursor...
	TSQLQuery* t_qry = new TSQLQuery(FSQLDB);
	t_qry->SetSQLText(t_sql.c_str());
	try
	{
		t_qry->PrepareParameterizedQuery();
		// If the call fails, an error will be thrown,
		// so the catch will clean up.
		FCursorsList[t_cursor_name] = t_qry;
	}
	catch (std::runtime_error& e)
	{
		std::string t_error = e.what();
		FEngineErrorsPump->AddError("TSMF","PrepareParameterizedQuery", t_error);
		t_error = "SQL error: " + (std::string)t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","PrepareParameterizedQuery", t_error);

		delete t_qry;

		t_result = -1;
	}

	return t_result;
} // PrepareParameterizedCursor
//---------------------------------------------------------------------------
int TSMF::BindIntParameter(const char* cursor_name, const int param_num,
	const int value)
{
	int t_result = 0;

	std::string t_cursor_name ;
	t_cursor_name.assign(cursor_name);

	TSQLQuery* t_qry = get_cursor_from_list(t_cursor_name);

	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","BindIntParameter", t_error);

		t_result = -1;
	}
	else
	{
		try
		{
			t_qry->BindIntParameter(param_num, value);
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMF", "BindIntParameter",
				t_qry->LastErrorMessage());
			t_result = -1;
		}
	}

	return t_result;
} // BindIntParameter
//---------------------------------------------------------------------------
int TSMF::BindTextParameter(const char* cursor_name, const int param_num,
	const char* value)
{
	int t_result = 0;

	std::string t_cursor_name ;
	t_cursor_name.assign(cursor_name);

	TSQLQuery* t_qry = get_cursor_from_list(t_cursor_name);

	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","BindTextParameter", t_error);

		t_result = -1;
	}
	else
	{
		try
		{
			t_qry->BindTextParameter(param_num, value);
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMF", "BindTextParameter",
				t_qry->LastErrorMessage());
			t_result = -1;
		}
	}

	return t_result;
} // BindTextParameter
//---------------------------------------------------------------------------
int TSMF::OpenParameterizedCursor(const char* cursor_name)
{
	int t_result = 0;

	std::string t_cursor_name ;
	t_cursor_name.assign(cursor_name);

	TSQLQuery* t_qry = get_cursor_from_list(t_cursor_name);

	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","OpenParameterizedCursor", t_error);

		t_result = -1;
	}
	else
	{
		try
		{
			t_qry->OpenParameterizedQuery();
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMF", "OpenParameterizedCursor",
				t_qry->LastErrorMessage());
			t_result = -1;
		}
	}

	return t_result;
} // OpenParameterizedCursor
//---------------------------------------------------------------------------
int TSMF::ResetParameterizedCursor(const char* cursor_name)
{
	int t_result = 0;

	std::string t_cursor_name ;
	t_cursor_name.assign(cursor_name);

	TSQLQuery* t_qry = get_cursor_from_list(t_cursor_name);

	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","ResetParameterizedCursor", t_error);

		t_result = -1;
	}
	else
	{
		try
		{
			t_qry->ResetParameterizedQuery();
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMF", "ResetParameterizedCursor",
				t_qry->LastErrorMessage());
			t_result = -1;
		}
	}

	return t_result;
} // ResetParameterizedCursor
//---------------------------------------------------------------------------
int TSMF::ReleaseParameterizedCursor(const char* cursor_name)
{
	int t_result = 0;

	std::string t_cursor_name ;
	t_cursor_name.assign(cursor_name);

	TSQLQuery* t_qry = get_cursor_from_list(t_cursor_name);

	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","ReleaseParameterizedCursor", t_error);

		t_result = -1;
	}
	else
	{
		try
		{
			t_qry->ReleaseParameterizedQuery();
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMF", "ReleaseParameterizedCursor",
				t_qry->LastErrorMessage());
			t_result = -1;
		}
		// Clean up...
		FCursorsList.erase(t_cursor_name);
		delete t_qry;
	}

	return t_result;
} // ReleaseParameterizedCursor
//---------------------------------------------------------------------------
int TSMF::CursorRowCount(const char* cursor_name, int* row_count)
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","CursorRowCount", t_error);

		return -1;
	}

	*row_count = t_qry->RowCount();

	return t_result;
} // CursorRowCount
//---------------------------------------------------------------------------
int TSMF::CursorColCount(const char* cursor_name, int* col_count)
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","CursorColCount", t_error);

		return -1;
	}

	*col_count = t_qry->ColCount();

	return t_result;
} // CursorColCount
//---------------------------------------------------------------------------
int TSMF::CursorNext(const char* cursor_name )
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","CursorNext", t_error);
		return -1; // cannot find cursor by that name
	}

	if (!t_qry->FetchNext())
	{
		// sqlite3_step failed; probably hit end-of-result-set
		std::string t_error = t_qry->LastErrorMessage();
		FEngineErrorsPump->AddError("TSMF","CursorNext", t_error);

		return -2;
	}
	return t_result;
} // CursorNext
//---------------------------------------------------------------------------
int TSMF::CursorEof(const char* cursor_name, int* is_eof)
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","CursorEof", t_error);
		return -1; // cannot find cursor by that name
	}

	if (t_qry->Eof())
	{
		*is_eof = 1; // true, cursor is at end of result set
	}
	else
	{
		*is_eof = 0; // false, not at end of result set
	}

	return t_result;
} // CursorEof
//---------------------------------------------------------------------------
int TSMF::ReadCursorData(const char* cursor_name,
	void* owner, void* callback_func )
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","ReadCursorData", t_error);
		t_result = -1; // cannot find cursor by that name
	}
	else
	{
		CallbackReadData t_callback_func = (CallbackReadData)callback_func;
		std::string t_field_name, t_data_type, t_data_string;
		int t_pkey = t_qry->GetIntFieldByNumber(0);
		int t_col_count = t_qry->ColCount();
		for (int t_col = 0; t_col < t_col_count; t_col++)
		{
			// Meta data
			t_field_name.assign(t_qry->GetColumnName(t_col));
			t_data_type.assign(t_qry->GetColumnDataType(t_col));
			t_data_string.assign(t_qry->GetFieldByNumber(t_col));

			// Send it off to the callback function.
			// Note that we increment t_col in the fourth parameter;
			// it represents the "column order", which is 1-based
			// for human beings.
			(*t_callback_func)(
				owner,
				t_cursor_name.c_str(),
				t_pkey,
				(t_col+1),
				t_field_name.c_str(),
				t_data_type.c_str(),
				t_data_string.c_str());
		}
	}

	return t_result;
} // ReadCursorData
//---------------------------------------------------------------------------
int TSMF::GetCursorDataSizeByNumber( const char* cursor_name,
	const int column_number, int* data_size )
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetCursorDataSizeByNumber", t_error);
		t_result = -1; // cannot find cursor by that name
	}

	else if ((column_number < 1) || column_number > t_qry->ColCount())
	{
		std::string t_error = "column_number is out of range";
		FEngineErrorsPump->AddError("TSMF","GetCursorDataSizeByNumber", t_error);

		t_result = -1;  // column_number is out of range
	}
	else
	{
		std::string t_data = t_qry->GetFieldByNumber(column_number-1);
		int t_data_size = (int)t_data.size();

		if (t_data_size > 0)
		{
			// The purpose of the call to this function is to find out
			// how big to make an array of char. I'm going to pad this by
			// one to ensure the caller creates a buffer that is big enough
			// to hold a null terminator.
			*data_size = t_data_size +1;
		}
		else
		{
			*data_size = 0;
		}
	}

	return t_result;
} // GetCursorDataSizeByNumber
//---------------------------------------------------------------------------
int TSMF::GetCursorDataByNumber( const char* cursor_name,
	const int column_number, char* buffer, const int buff_size )
{
	int t_result = 0;

	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetCursorDataByNumber", t_error);
		t_result = -1; // cannot find cursor by that name
	}

	else if ((column_number < 1) || column_number > t_qry->ColCount())
	{
		std::string t_error = "column_number is out of range";
		FEngineErrorsPump->AddError("TSMF","GetCursorDataSizeByNumber", t_error);

		t_result = -1;  // column_number is out of range
	}
	else
	{
		if (buff_size > 0)
		{
			// Whatever else may happen here, let's ensure a null-terminated string
			// goes back to the caller.
			buffer[0] = '\x0';

			std::string t_data = t_qry->GetFieldByNumber(column_number-1);
			size_t t_data_size = t_data.size();

			if (t_data_size > 0)
			{
				strncpy(buffer, (char*)t_data.c_str(), buff_size );
			}
		}
	}

	return t_result;
} // GetCursorDataByNumber
//---------------------------------------------------------------------------
int TSMF::GetCursorDataSizeByName( const char* cursor_name,
	const char* field_name, int* data_size )
{
	int t_result = 0;
	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetCursorDataSizeByName", t_error);
		t_result = -1; // cannot find cursor by that name
	}
	else
	{
		std::string t_data = t_qry->GetFieldByName(field_name);
		int t_data_size = (int)t_data.size();

		if (t_data_size > 0)
		{
			// The purpose of the call to this function is to find out
			// how big to make an array of char. I'm going to pad this by
			// one to ensure the caller creates a buffer that is big enough
			// to hold a null terminator.
			*data_size = t_data_size +1;
		}
		else
		{
			*data_size = 0;
		}
	}

	return t_result;
} // GetCursorDataSizeByName
//---------------------------------------------------------------------------
int TSMF::GetCursorDataByName( const char* cursor_name,
	const char* field_name, char* buffer, const int buff_size )
{
	int t_result = 0;

	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetCursorDataByName", t_error);
		t_result = -1; // cannot find cursor by that name
	}

	else
	{
		if (buff_size > 0)
		{
			// Whatever else may happen here, let's ensure a null-terminated string
			// goes back to the caller.
			buffer[0] = '\x0';

			std::string t_data = t_qry->GetFieldByName(field_name);
			size_t t_data_size = t_data.size();

			if (t_data_size > 0)
			{
				strncpy(buffer, (char*)t_data.c_str(), buff_size );
			}
		}
	}

	return t_result;
} // GetCursorDataByName
//---------------------------------------------------------------------------
int TSMF::GetCursorIntByNumber( const char* cursor_name,
	const int column_number, int* data )
{
	int t_result = 0;

	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetCursorIntByNumber", t_error);
		t_result = -1; // cannot find cursor by that name
	}
	else
	{
		*data = t_qry->GetIntFieldByNumber(column_number -1);
	}

	return t_result;
} // GetCursorIntByNumber
//---------------------------------------------------------------------------

int TSMF::GetCursorIntByName( const char* cursor_name,
	const char* field_name, int* data )
{
	int t_result = 0;

	std::string t_cursor_name;
	t_cursor_name.assign(cursor_name);
	lowercase(t_cursor_name);

	TSQLQuery* t_qry = FCursorsList[t_cursor_name];
	if (!t_qry)
	{
		std::string t_error = "Cannot find cursor named \"";
		t_error += cursor_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF","GetCursorIntByNumber", t_error);
		t_result = -1; // cannot find cursor by that name
	}
	else
	{
		*data = t_qry->GetIntFieldByName(field_name);
	}

	return t_result;
} // GetCursorIntByName
//---------------------------------------------------------------------------

/****************************************************************************/
/*************        API Functions for Running Edits          **************/
/****************************************************************************/
std::string TSMF::get_layout_tag_from_name(const char* layout_name)
{
	std::string t_layout_tag;
	char t_buffer[500];

	std::snprintf(t_buffer, 500, "SELECT LayoutTag FROM EE_Layouts WHERE LayoutName = '%s';", layout_name);
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	t_qry->SetSQLText(t_buffer);
	t_qry->SetIsActive(true);
	if (!t_qry->Eof())
	{
		t_layout_tag = t_qry->GetFieldByName("LayoutTag");
	}

	return t_layout_tag;
} // get_layout_tag_from_name
//---------------------------------------------------------------------------
int TSMF::UpdateRecordByTag( const char* layout_tag, const char* data)
{
	std::string t_data;
	if (data)
		t_data.assign(data);

	return FRunEdits->UpdateRecord(layout_tag, t_data);
}
//---------------------------------------------------------------------------
int TSMF::UpdateRecordByName( const char* layout_name, const char* data)
{
	std::string t_data,
				t_layout_tag;
	int t_result;

	if (data)
	{
		t_data.assign(data);
	}

	if (layout_name)
	{
		t_layout_tag = get_layout_tag_from_name(layout_name);
	}

	if (t_layout_tag.empty())
	{
		char t_buffer[500];
		std::snprintf(t_buffer, 500, "Layout name \"%s\" not found in metafile", layout_name);
		FEngineErrorsPump->AddError("TSMF","UpdateRecordByName", t_buffer);

		t_result = -1;
	}
	else
	{
		t_result = FRunEdits->UpdateRecord(t_layout_tag, t_data);
	}

	return t_result;
} // UpdateRecordByName
//---------------------------------------------------------------------------
int TSMF::DisplayDataRecordByTag( const char* layout_tag, void* owner,
	void* callback_func )
{
	return FRunEdits->DisplayDataRecord( layout_tag, owner, callback_func );
}
//---------------------------------------------------------------------------

int TSMF::DisplayDataRecordByName( const char* layout_name, void* owner,
	void* callback_func )
{
	int t_result = 0;

	std::string t_layout_tag;

	if (layout_name)
	{
		t_layout_tag = get_layout_tag_from_name(layout_name);
	}

	if (t_layout_tag.empty())
	{
		char t_buffer[500];
		std::snprintf(t_buffer, 500, "Layout name \"%s\" not found in metafile", layout_name);
		FEngineErrorsPump->AddError("TSMF","DisplayDataRecordByName", t_buffer);

		t_result = -1;
	}
	else
	{
		t_result = FRunEdits->DisplayDataRecord( t_layout_tag.c_str(), owner, callback_func );
	}

	return t_result;
} // DisplayDataRecordByName
//---------------------------------------------------------------------------
int TSMF::DisplayFieldInMessagesAs( const int field_number,
	const char* display_name )
{
	std::string t_display_name;

	if (display_name)
	{
		t_display_name.assign(display_name);
	}

	return FRunEdits->DisplayFieldInMessagesAs(field_number, t_display_name);
} // DisplayFieldInMessagesAs
//---------------------------------------------------------------------------

int TSMF::NeverSkipEmptyField( const int field_number,
	const int never_skip_empty )
{
	return FRunEdits->NeverSkipEmptyField(field_number, never_skip_empty);
} // DisplayFieldInMessagesAs
//---------------------------------------------------------------------------

int TSMF::FieldDataByNumber(
		 const char* layout_tag,
		 const int field_number,
		 int* pkey,
		 int* start_pos,
		 int* field_length,
		 char* field_name,
		 const int field_name_len,
		 char* display_name,
		 const int display_name_len,
		 char* data_type,
		 const int data_type_len,
		 char* value, const int value_len )
{
	int t_result = 0;

	if (layout_tag &&
		field_name &&
		display_name &&
		data_type &&
		value)
	{
		t_result = FRunEdits->FieldDataByNumber(
			layout_tag,
			field_number,
			pkey, start_pos, field_length,
			field_name, field_name_len,
			display_name, display_name_len,
			data_type, data_type_len,
			value, value_len );
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "DataByFieldNumber",
			"At least one of the char* parameters is NULL");
		t_result = -1;
	}

	return t_result;
} // FieldDataByNumber
//---------------------------------------------------------------------------

int TSMF::LoadEditSet( const char* edit_set_tag, const char* layout_tag )
{
	int t_result = 0;

	if (edit_set_tag && layout_tag)
	{
		t_result = FRunEdits->LoadEditSet(edit_set_tag, layout_tag);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "LoadEditSet",
			"At least one of the char* parameters is NULL");
		t_result = -1;
	}

	return t_result;

} // LoadEditSet
//---------------------------------------------------------------------------
int TSMF::UnloadEditSet( const char* edit_set_tag, const char* layout_tag )
{
	int t_result = 0;

	if (edit_set_tag && layout_tag)
	{
		t_result = FRunEdits->UnloadEditSet(edit_set_tag, layout_tag);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "UnloadEditSet",
			"At least one of the char* parameters is NULL");
		t_result = -1;
	}

	return t_result;

} // UnoadEditSet
//---------------------------------------------------------------------------
int TSMF::SetEditOptions( const char* edit_set_tag, const char* edit_tag,
	const int suppress_edit,
	const int skip_if_empty,
	const int skip_if_failed,
	const int no_skips_override )
{
	int t_result = 0;

	if (edit_set_tag && edit_tag)
	{
		t_result = FRunEdits->SetEditOptions(edit_set_tag, edit_tag, suppress_edit,
			skip_if_empty, skip_if_failed, no_skips_override);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "SetEditOptions",
			"At least one of the char* parameters is NULL");
		t_result = -1;
	}

	return t_result;
} // SetEditOptions
//---------------------------------------------------------------------------

int TSMF::GetEditOptions( const char* edit_set_tag, const char* edit_tag,
	int* is_suppressed, int* is_skip_if_empty, int* is_skip_if_failed,
	int* is_no_skips_override )
{
	int t_result = 0;
	if (edit_set_tag && edit_tag)
	{
		t_result = FRunEdits->GetEditOptions(edit_set_tag, edit_tag,
		is_suppressed, is_skip_if_empty, is_skip_if_failed,
		is_no_skips_override);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "GetEditOptions",
			"At least one of the char* parameters is NULL");
		t_result = -1;
	}

	return t_result;
} // GetEditOptions
//---------------------------------------------------------------------------
int TSMF::ClearEditOptions( const char* edit_set_tag )
{
	int t_result = 0;

	if (edit_set_tag)
	{
		t_result = FRunEdits->ClearEditOptions(edit_set_tag);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "ClearEditOptions",
			"Parameter edit_set_tag is NULL");
		t_result = -1;
	}

	return t_result;
} // ClearEditOptions
//---------------------------------------------------------------------------
int TSMF::RunEdits( const char* edit_set_tag, const char* layout_tag,
	const char* data, const int edit_options, int* errors_count,
	void* owner, void* callback_func)
{
	std::string t_edit_set_tag,
				t_layout_tag,
				t_data;
	std::vector<std::string> t_errors;

	if (edit_set_tag)
		t_edit_set_tag.assign(edit_set_tag);
	else
		t_errors.push_back("Parameter \"edit_set_tag\" must not be NULL");

	if (layout_tag)
		t_layout_tag.assign(layout_tag);
	else
		t_errors.push_back("Parameter \"layout_tag\" must not be NULL");

	if (data)
		t_data.assign(data);
	else
		t_errors.push_back("Parameter \"data\" must not be NULL");

	// owner and callback_func are a set: They must both be assigned, or both be NULL
	if (owner && (!callback_func))
		t_errors.push_back("Parameter \"owner\" is NOT NULL, but parameter \"callback_func\" IS NULL");

	if (callback_func && (!owner))
		t_errors.push_back("Parameter \"callback_func\" is NOT NULL, but parameter \"owner\" IS NULL");


	if (t_errors.size() > 0)
	{
		for (size_t t_idx = 0; t_idx < t_errors.size(); t_idx++)
		{
			FEngineErrorsPump->AddError("TSMF", "RunEdits",
				t_errors[t_idx]);
		}
		return -1;
	}

	return FRunEdits->RunEdits(t_edit_set_tag, t_layout_tag, t_data,
		edit_options, errors_count, owner, callback_func);

} // RunEdits
//---------------------------------------------------------------------------
int TSMF::GetRunEditsSummary(const char* edit_set_tag, const char* layout_tag,
	const int include_zero_counts, int* total_errors, int* total_warnings,
	int* total_skips, int* total_messages, int* total_warning_messages,
	void* owner, void* callback_func)
{
	std::string t_edit_set_tag,
				t_layout_tag;
	std::vector<std::string> t_errors;

	if (edit_set_tag)
		t_edit_set_tag.assign(edit_set_tag);
	else
		t_errors.push_back("Parameter \"edit_set_tag\" must not be NULL");

	if (layout_tag)
		t_layout_tag.assign(layout_tag);
	else
		t_errors.push_back("Parameter \"layout_tag\" must not be NULL");

	// owner and callback_func are a set: They must both be assigned, or both be NULL
	if (owner && (!callback_func))
		t_errors.push_back("Parameter \"owner\" is NOT NULL, but parameter \"callback_func\" IS NULL");

	if (callback_func && (!owner))
		t_errors.push_back("Parameter \"callback_func\" is NOT NULL, but parameter \"owner\" IS NULL");


	if (t_errors.size() > 0)
	{
		for (size_t t_idx = 0; t_idx < t_errors.size(); t_idx++)
		{
			FEngineErrorsPump->AddError("TSMF", "GetRunEditsSummary",
				t_errors[t_idx]);
		}
		return -1;
	}

	return FRunEdits->GetRunEditsSummary(t_edit_set_tag.c_str(), t_layout_tag.c_str(),
		include_zero_counts, total_errors, total_warnings, total_skips, total_messages,
		total_warning_messages, owner, callback_func);

} // GetRunEditsSummary
//---------------------------------------------------------------------------
int TSMF::GetEditDetails(const char* edit_tag, void* owner, void* callback_func)
{
	return FRunEdits->GetEditDetails(edit_tag, owner, callback_func);
}
//---------------------------------------------------------------------------
int TSMF::DebugEdit(const char* edit_tag, const char* layout_tag,
	const char* data, void* owner, void* logic_callback,
	void* variables_callback, void* fields_callback, void* notes_callback)
{
	std::string t_edit_tag,
				t_layout_tag,
				t_data;
	std::vector<std::string> t_errors;

	if (edit_tag)
		t_edit_tag.assign(edit_tag);
	else
		t_errors.push_back("Parameter \"edit_tag\" must not be NULL");

	if (layout_tag)
		t_layout_tag.assign(layout_tag);
	else
		t_errors.push_back("Parameter \"layout_tag\" must not be NULL");

	if (data)
		t_data.assign(data);
	else
		t_errors.push_back("Parameter \"data\" must not be NULL");

	// This function is pointless without at least one callback...
	// so we want to ensure that owner exists
	if (!owner)
		t_errors.push_back("Parameter \"owner\" must not be NULL");

	// Make sure we have at least one callback
	if (!(logic_callback ||
		variables_callback ||
		fields_callback ||
		notes_callback))
	{
		t_errors.push_back("Edit_DebugEdit is pointless without at least one callback function");
	}

	if (t_errors.size() > 0)
	{
		for (size_t t_idx = 0; t_idx < t_errors.size(); t_idx++)
		{
			FEngineErrorsPump->AddError("TSMF", "DebugEdit",
				t_errors[t_idx]);
		}
		return -1;
	}

	return FRunEdits->DebugEdit(t_edit_tag, t_layout_tag, t_data,
		owner, logic_callback, variables_callback, fields_callback,
		notes_callback);
} // DebugEdit
//---------------------------------------------------------------------------

int TSMF::UpdateAllAccessField( const char* table_name, const int pkey,
	const char* field_name, const char* value )
{
	int t_result = -1; // until we verify

	if ((posix_stricmp(table_name, "EE_Fields") == 0) &&
		(posix_stricmp(field_name, "LocalName") == 0))
	{
		t_result = 0; // good to go
	}

	if (t_result == 0)
	{
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		char t_buffer[500];
		std::snprintf(t_buffer, 500,
			"UPDATE [%s] SET [%s] = '%s' WHERE [PKey] = %d;",
			table_name, field_name, value, pkey);

		try
		{
			t_qry->SetSQLText(t_buffer);
			if (!t_qry->ExecSQL())
			{
				FEngineErrorsPump->AddError("TSMF", "UpdateAllAccessField",
					t_qry->LastErrorMessage());
			}

		}
		catch (std::runtime_error& e)
		{
			FEngineErrorsPump->AddError("TSMF", "UpdateAllAccessField",
				e.what());
			FEngineErrorsPump->AddError("TSMF", "UpdateAllAccessField",
				FSQLDB->GetLastDBErrorMessage());
			FEngineErrorsPump->AddError("TSMF", "UpdateAllAccessField",
				t_buffer);
			t_result = -1;
		}
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "UpdateAllAccessField",
			"Inappropriate use of this function; please read documentation");
	}

	return t_result;
} // UpdateAllAccessField
//---------------------------------------------------------------------------

int TSMF::RecompileEditLogic( const char* edit_tag )
{
	int t_result = 0,
		t_edit_pkey = 0,
		t_field_pkey,
		t_default_message_pkey = 0,
		t_how_many_errors,
		t_how_many_warnings;
	std::string t_edit_logic = "";

	std::string t_edit_tag = stringReplace(edit_tag, "'", "''");

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[100];
	std::snprintf(t_buffer, 100,
		"SELECT [PKey], [MessagePKey], [EditLogic] FROM [EE_Edits] WHERE [EditTag] = '%s';",
		t_edit_tag.c_str());

	try
	{
		// Set "check_is_parameterized" to false, because edit tags
		// can have whacko characters, including question marks
		t_qry->SetSQLText(t_buffer, false);
		if (t_qry->SetIsActive(true))
		{
			t_edit_pkey = t_qry->GetIntFieldByName("PKey");
			t_default_message_pkey = t_qry->GetIntFieldByName("MessagePKey");
			t_edit_logic.assign(t_qry->GetFieldByName("EditLogic"));

			t_qry->SetIsActive(false);
		}
		else
		{
			FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic",
				t_qry->LastErrorMessage());
			t_result = -1;
		}
	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", e.what());
		t_result = -1;
	}

	if (t_result != 0)
	{
		FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", t_buffer);
		return t_result;
	}

	t_result = CheckSyntax(t_edit_logic.c_str(),
		&t_how_many_errors, &t_how_many_warnings);

	if (t_result == 0 && t_edit_pkey > 0)
	{
		// Note: We can rebuild the messages and tables cross-references
		// entirely from what FEditCompiler discovered by compiling the edit logic,
		// but unfortunately the edit AUTHOR has to determine the field macro
		// order. We CAN ensure no field links linger if the logic doesn't still
		// reference them, and preserve a previously-established order by loading
		// this vector in EE_EditFieldsPivot.FieldOrder. Seems about right. If
		// any new fields emerge, we'll do what EditWriter does, and park them
		// at the bottom of the field macro order. The user can correct them if
		// necessary.
		std::vector<int> t_fields_list;
		char t_buffer[100];

		std::snprintf(t_buffer, 100,
			"SELECT [FieldPKey] FROM [EE_EditFieldsPivot] " \
			"WHERE [EditPKey] = %d ORDER BY [FieldOrder];",
			t_edit_pkey);

		t_qry->SetSQLText(t_buffer);
		if (t_qry->SetIsActive(true))
		{
			while (!t_qry->Eof())
			{
				t_field_pkey = t_qry->GetIntFieldByName("FieldPKey");
				t_fields_list.push_back(t_field_pkey);
				t_qry->FetchNext();
			}
			t_qry->SetIsActive(false);
		}
		else
		{
			FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic",
				t_qry->LastErrorMessage());

			return -1;
		}

		// Wrap a transaction around all these table updates...
		// use a savepoint in case the caller has already started a
		// transaction
		std::string t_savepoint_name = "_x_RecompileEditLogic_y_";
		if (!FSQLDB->SavePoint(t_savepoint_name.c_str()))
		{
			FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", "Failure to create savepoint");
			return -1;
		}

		try
		{
			if (t_how_many_errors == 0)
			{
				// Let's keep the pivot links for the sake
				// of the user interface; clear ONLY when we're going
				// to rewrite them from the information in the logic.
				FSMFWriter->ClearEditLogicPivots(t_edit_pkey);

				// FIRST: Write the compiled logic (and note this is NOT affecting
				// any other fields in EE_Edits... this allows a user without
				// access rights to keep all edits ready to run.
				std::string t_compiled_logic = FEditCompiler->GetCompiledCode();

				// escape embedded apostrophes... or they'll wreak havoc
				std::string t_sqlized = stringReplace(t_compiled_logic, "'", "''");

				std::string t_sql = "UPDATE [EE_Edits] SET [CompiledLogic] = '";
				t_sql += t_sqlized;
				t_sql += "' WHERE [EditTag] = '";
				//t_sql += stringReplace(edit_tag, "'", "''");
				t_sql += t_edit_tag;
				t_sql += "';";

				try
				{
					t_qry->SetSQLText(t_sql);
					if (!t_qry->ExecSQL())
					{
						FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic",
							t_qry->LastErrorMessage());
						t_result = -1;
					}
				}
				catch (std::runtime_error& e)
				{
					FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", e.what());
					t_result = -1;
				}
				if (t_result != 0)
				{
					FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", t_sql);
				}
				else
				{
					// NEXT: Update EE_EditFieldsPivot, using t_fields_list to obtain
					// the FieldOrder value.
					if (!update_edit_fields_pivot(t_edit_pkey, t_fields_list))
					{
						t_result = -1;
					}
					else if (!update_default_message_in_pivot(t_edit_pkey, t_default_message_pkey))
					{
						t_result = -1;
					}
					else if (!update_edit_messages_pivot(t_edit_pkey))
					{
						t_result = -1;
					}
					else if (!update_edit_errors_pivot(t_edit_pkey))
					{
						t_result = -1;
					}
					else if (!update_edit_tables_pivot(t_edit_pkey))
					{
						t_result = -1;
					}
				}
			} // if (t_how_many_errors == 0)
			else
			{
				// Clear the CompiledLogic field; the user may choose to save
				// her flawed edit logic for another time, but meanwhile, we don't
				// want the Engine to try to run bad code.
				std::string t_sql = "UPDATE [EE_Edits] SET [CompiledLogic] = ''";
				t_sql += " WHERE [EditTag] = '";
				t_sql += t_edit_tag;
				t_sql += "';";

				try
				{
					t_qry->SetSQLText(t_sql);
					if (!t_qry->ExecSQL())
					{
						FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic",
							t_qry->LastErrorMessage());
					}
				}
				catch (std::runtime_error& e)
				{
					FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", e.what());
				}

				// Return code -2 to indicate to caller that the syntax checker
				// reported errors
				t_result = -2;
			}

			if (t_result == 0 || t_result == -2)
			{
				// Release COMMITs and releases the savepoint
				FSQLDB->Release(t_savepoint_name.c_str());
			}
			else
			{
				// RollbackTo differs from RollbackTrans in that it
				// rolls back only to the named save point
				if (FSQLDB->RollbackTo(t_savepoint_name.c_str()))
				{
					FSQLDB->Release(t_savepoint_name.c_str());
				}

			}
		}
		catch (std::runtime_error& e)
		{
			FSQLDB->RollbackTo(t_savepoint_name.c_str());
			FEngineErrorsPump->AddError("TSMF", "RecompileEditLogic", e.what());
			t_result = -1;
		}
	} // if (t_result == 0 && t_edit_pkey > 0)

	// We may be modifying t_result based on whether we find any syntax warnings.
	if ((t_result != -1) && (t_how_many_warnings > 0))
	{
		if (t_result == 0)
			t_result = -3; // no errors, just warnings
		else
			t_result = -4; // BOTH errors AND warnings were issued
	}

	return t_result;
} // RecompileEditLogic
//---------------------------------------------------------------------------
bool TSMF::update_edit_fields_pivot(int edit_pkey, std::vector<int>& fields_list)
{
	bool t_result = true;
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[1000];
	int t_field_pkey,
		t_field_number,
		t_field_order;
	int t_how_many = FEditCompiler->GetLogicListCount('F');

	if (t_how_many > 0)
	{
		for (int t_idx=0; (t_idx < t_how_many) && t_result; t_idx++)
		{
			t_field_number = FEditCompiler->GetLogicFieldNumber(t_idx);
			std::snprintf(t_buffer, 1000,
				"SELECT [PKey] FROM [EE_Fields] WHERE [FieldNumber] = %d;",
				t_field_number);
			t_qry->SetSQLText(t_buffer);
			if (t_qry->SetIsActive(true))
			{
				t_field_pkey = t_qry->GetIntFieldByName("PKey");
				t_qry->SetIsActive(false);

				// Can we find this field PKey in our stored list?
				t_field_order = 0;
				for (int t_jdx=0; t_jdx < (int)fields_list.size(); t_jdx++)
				{
					int t_test = fields_list[t_jdx];
					if (t_test == t_field_pkey)
					{
						// The field was previously in the list,
						// at this macro order'th position
						t_field_order = t_jdx+1;
						break;
					}
				}
				// If we didn't find it in the list, it's new
				// (note: this should be a very rare occurrence
				// for this function; normally, EditWriter would
				// have compiled the edit already and correctly
				// created the fields list)
				if (t_field_order == 0)
				{
					fields_list.push_back(t_field_pkey);
					t_field_order = (int)fields_list.size();
				}

				// Note use of INSERT OR IGNORE to avoid hollering
				// from the metafile if we somehow manage to try to
				// load a duplicate. The OR IGNORE wll fail silently.
				std::snprintf(t_buffer, 1000,
					"INSERT OR IGNORE INTO [EE_EditFieldsPivot] ([EditPKey],[FieldPKey],[FieldOrder]) " \
					"VALUES (%d, %d, %d);",
					edit_pkey, t_field_pkey, t_field_order);
				t_qry->SetSQLText(t_buffer);
				if (!t_qry->ExecSQL())
				{
					FEngineErrorsPump->AddError("TSMF", "update_edit_fields_pivot",
						t_qry->LastErrorMessage());
					FEngineErrorsPump->AddError("TSMF", "update_edit_fields_pivot",
						t_buffer);
					t_result = false;
				}

			} // if (finding t_field_pkey)
		} // for (iterate over GetLogicFieldNumber)
	} // if (t_how_many > 0)

	if (t_result)
	{
		// Set EE_Edits.IsMultiField value here... 0 means "not multifield
		// (i.e., has zero or one field), 1 means "is multifield... has two
		// or more field references"
		std::snprintf(t_buffer, 1000,
			"UPDATE [EE_Edits] SET [IsMultiField] = %d WHERE [PKey] = %d;",
			(t_how_many > 1 ? 1 : 0), edit_pkey);
		t_qry->SetSQLText(t_buffer);
		if (!t_qry->ExecSQL())
		{
			FEngineErrorsPump->AddError("TSMF", "update_edit_fields_pivot",
				t_qry->LastErrorMessage());
			FEngineErrorsPump->AddError("TSMF", "update_edit_fields_pivot",
				t_buffer);
			t_result = false;
		}
	}
	return t_result;
} // update_edit_fields_pivot
//---------------------------------------------------------------------------
bool TSMF::update_default_message_in_pivot(int edit_pkey, int default_message_pkey)
{
	bool t_result = true;
	// This one is easy... the parameters give us what we need
	// NOTE: default_message_pkey of 1 is the 0-numbered message,
	// meaning "this edit has no default error message". We don't
	// create a cross-ref record for that.
	if (edit_pkey > 0 && default_message_pkey > 1)
	{
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		char t_message_func_token[2];
		t_message_func_token[0] = 'D';
		t_message_func_token[1] = '\x0';
		char t_buffer[1000];

		std::snprintf(t_buffer, 1000,
			"INSERT OR IGNORE INTO [EE_EditMessagesPivot] (" \
			"[EditPKey], [MessagePKey], [EditsFunc]) " \
			"VALUES (%d, %d, '%s');",
			edit_pkey, default_message_pkey, t_message_func_token);

		t_qry->SetSQLText(t_buffer);
		if (!t_qry->ExecSQL())
		{
			FEngineErrorsPump->AddError("TSMF", "update_default_message_in_pivot",
				t_qry->LastErrorMessage());
			FEngineErrorsPump->AddError("TSMF", "update_default_message_in_pivot",
				t_buffer);
			t_result = false;
		}
	}
	return t_result;
} // update_default_message_in_pivot
//---------------------------------------------------------------------------

bool TSMF::update_edit_messages_pivot(int edit_pkey)
{
	bool t_result = true;
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

	int t_message_pkey,
		t_message_number,
		t_token;
	char t_message_func_token[2];
	t_message_func_token[1] = '\x0';

	int t_how_many = FEditCompiler->GetLogicListCount('M');
	if (t_how_many > 0)
	{
		char t_buffer[1000];
		for (int t_idx=0; (t_idx < t_how_many) && t_result; t_idx++)
		{
			t_message_number =
				FEditCompiler->GetLogicMessageNumber(t_idx, &t_token);
			if (t_message_number == 0)
			{
				// See TLogicMessageObject in EditCompiler.h... it does double duty
				// for numbered messages and text messages
				continue;
			}
			t_message_func_token[0] = (char)t_token;
			std::snprintf(t_buffer, 1000,
				"SELECT [PKey] FROM [EE_Messages] WHERE [MessageNumber] = %d;",
				t_message_number);
			t_qry->SetSQLText(t_buffer);
			if (t_qry->SetIsActive(true))
			{
				t_message_pkey = t_qry->GetIntFieldByName("PKey");
				t_qry->SetIsActive(false);

				if (t_message_pkey > 0)
				{
					std::snprintf(t_buffer, 1000,
						"INSERT OR IGNORE INTO [EE_EditMessagesPivot] (" \
						"[EditPKey], [MessagePKey], [EditsFunc]) " \
						"VALUES (%d, %d, '%s');",
						edit_pkey, t_message_pkey, t_message_func_token);

					t_qry->SetSQLText(t_buffer);
					if (!t_qry->ExecSQL())
					{
						FEngineErrorsPump->AddError("TSMF", "update_edit_messages_pivot",
							t_qry->LastErrorMessage());
						FEngineErrorsPump->AddError("TSMF", "update_edit_messages_pivot",
							t_buffer);
						t_result = false;
					}
				}
			} // if (t_qry->SetIsActive())
		} // for (iterate over GetLogicMessageNumber)
	}
	return t_result;
} // update_edit_messages_pivot
//---------------------------------------------------------------------------
bool TSMF::update_edit_errors_pivot(int edit_pkey)
{
	bool t_result = true;
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

	int t_how_many = FEditCompiler->GetLogicListCount('E');
	if (t_how_many > 0)
	{
		std::string t_message;
		int t_token;
		char t_message_func_token[2];
		t_message_func_token[1] = '\x0';
		char t_buffer[1000];

		for (int t_idx=0; (t_idx < t_how_many) && t_result; t_idx++)
		{
			t_message = FEditCompiler->GetLogicErrorMessage(t_idx, &t_token);
			if (t_message.empty())
			{
				// See TLogicMessageObject in EditCompiler.h... it does double duty
				// for numbered messages and text messages
				continue;
			}
			t_message_func_token[0] = (char)t_token;
			std::snprintf(t_buffer, 1000,
				"INSERT OR IGNORE INTO [EE_EditErrorsPivot] (" \
				"[EditPKey], [ErrorText],[EditsFunc]) " \
				"VALUES (%d, '%s', '%s');",
				edit_pkey, stringReplace(t_message,"'","''").c_str(), t_message_func_token);
			t_qry->SetSQLText(t_buffer);
			if (!t_qry->ExecSQL())
			{
				FEngineErrorsPump->AddError("TSMF", "update_edit_messages_pivot",
					t_qry->LastErrorMessage());
				FEngineErrorsPump->AddError("TSMF", "update_edit_messages_pivot",
					t_buffer);
				t_result = false;
			}

		} // for (iterate over GetLogicErrorMessage)
	}
	return t_result;
} // update_edit_errors_pivot
//---------------------------------------------------------------------------
bool TSMF::update_edit_tables_pivot(int edit_pkey)
{
	bool t_result = true;
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

	int t_how_many = FEditCompiler->GetLogicListCount('T');
	if (t_how_many > 0)
	{
		int t_table_pkey;
		std::string t_table_name;
		char t_buffer[1000];
		for (int t_idx=0; (t_idx < t_how_many) && t_result; t_idx++)
		{
			t_table_name = FEditCompiler->GetLogicTableName(t_idx);
			std::snprintf(t_buffer, 1000,
				"SELECT [PKey] FROM [EE_LookupTables] WHERE [TableName] = '%s';",
				t_table_name.c_str());

			t_qry->SetSQLText(t_buffer);

			if (t_qry->SetIsActive(true))
			{
				t_table_pkey = t_qry->GetIntFieldByName("PKey");
				t_qry->SetIsActive(false);

				std::snprintf(t_buffer, 1000,
					"INSERT OR IGNORE INTO EE_EditTablesPivot (" \
					"[EditPKey], [LookupTablePKey]) "\
					"VALUES (%d, %d);",
					edit_pkey, t_table_pkey);

				t_qry->SetSQLText(t_buffer);
				if (!t_qry->ExecSQL())
				{
					FEngineErrorsPump->AddError("TSMF", "update_edit_tables_pivot",
						t_qry->LastErrorMessage());
					FEngineErrorsPump->AddError("TSMF", "update_edit_tables_pivot",
						t_buffer);
					t_result = false;
				}
			}
		} // for (iterate over GetLogicTableName
	}
	return t_result;
} // update_edit_tables_pivot
//---------------------------------------------------------------------------
int TSMF::CheckSyntax( const char* edit_logic, int* how_many_errors,
	int* how_many_warnings)
{
	int t_result = 0;

	// Initialize how_many_errors to zero
	*how_many_errors = 0;
	*how_many_warnings = 0;

	t_result = FEditCompiler->Compile(edit_logic, how_many_errors, how_many_warnings);

	return t_result;
} // CheckSyntax
//---------------------------------------------------------------------------

int TSMF::GetSyntaxError( const int error_idx, int* line_number,
	char* syntax_error, const int syntax_error_len)
{
	int t_result = 0;

	// Retrieve the error from TEditCompiler. If error_idx is greater than
	// or equal to how_many_errors returned from CheckSyntax, throw an
	// error from this function.
	int t_error_max_count = FEditCompiler->GetErrorCount();
	if (t_error_max_count > 0)
	{
		if ((error_idx >= 0) && (error_idx < t_error_max_count))
		{
			std::string t_msg = FEditCompiler->GetErrorMessage(error_idx);
			// It will look like this: "Line 123: message"
			// We want the "123" to convert to int and return in param line_number
			*line_number = std::atoi(t_msg.substr(6,3).c_str());
			std::strncpy(syntax_error, t_msg.c_str(), syntax_error_len);
		}
		else
		{
			char t_buffer[100];
			std::snprintf(t_buffer, 100,
				"error_idx must be between 0 and %d", t_error_max_count-1);
			FEngineErrorsPump->AddError("TSMF", "GetSyntaxError", t_buffer);
			t_result = -1;
		}
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "GetSyntaxError",
			"No errors were generated by the syntax checker");
		t_result = -1;
	}

	return t_result;
} // GetSyntaxError
//---------------------------------------------------------------------------

int TSMF::GetSyntaxWarning( const int warning_idx, int* line_number,
	char* syntax_warning, const int syntax_warning_len)
{
	int t_result = 0;

	// Retrieve the warning from TEditCompiler. If warning_idx is greater than
	// or equal to how_many_warnings returned from CheckSyntax, throw an
	// error from this function.
	int t_warning_max_count = FEditCompiler->GetWarningCount();
	if (t_warning_max_count > 0)
	{
		if ((warning_idx >= 0) && (warning_idx < t_warning_max_count))
		{
			std::string t_msg = FEditCompiler->GetWarningMessage(warning_idx);
			// It will look like this: "Line 123: message"
			// We want the "123" to convert to int and return in param line_number
			*line_number = std::atoi(t_msg.substr(6,3).c_str());
			std::strncpy(syntax_warning, t_msg.c_str(), syntax_warning_len);
		}
		else
		{
			char t_buffer[100];
			std::snprintf(t_buffer, 100,
				"warning_idx must be between 0 and %d", t_warning_max_count-1);
			FEngineErrorsPump->AddError("TSMF", "GetSyntaxWarning", t_buffer);
			t_result = -1;
		}
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "GetSyntaxWarning",
			"No warnings were generated by the syntax checker");
		t_result = -1;
	}
	return t_result;
} // GetSyntaxWarning
//---------------------------------------------------------------------------

#ifdef __SQLEW__

/****************************************************************************/
/*************        API Functions for EditWriter             **************/
/****************************************************************************/
int TSMF::Login( const char* agency_code, const char* password )
{
	if (FEw_AccessController)
	{
		return FEw_AccessController->Login(agency_code, password);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "Login",
			"FEw_AccessController is NULL (Engine programmer error)");
		return -1;
	}
} // Login
//---------------------------------------------------------------------------
int TSMF::GetLoginsCount( int* how_many_logins )
{
	if (FEw_AccessController)
	{
		return FEw_AccessController->GetLoginsCount(how_many_logins);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "GetLoginsCount",
			"FEw_AccessController is NULL (Engine programmer error)");
		return -1;
	}
} // GetLoginsCount
//---------------------------------------------------------------------------
int TSMF::GetLoginByIndex( const int idx, int* agency_pkey, char* agency_code,
	const int agency_code_size )
{
	if (FEw_AccessController)
	{
		return FEw_AccessController->GetLoginByIndex(idx, agency_pkey,
			agency_code, agency_code_size);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "GetLoginByIndex",
			"FEw_AccessController is NULL (Engine programmer error)");
		return -1;
	}
} // GetLoginByIndex
//---------------------------------------------------------------------------
int TSMF::ValidateUnique( const char* table_name, const int pkey,
	const char* field_name, const char* value, int* data_is_unique,
	int* dupe_pkey )
{
	int t_result = FSMFWriter->ValidateUnique(table_name, pkey,
		field_name, value, data_is_unique, dupe_pkey );

	return t_result;
}
//---------------------------------------------------------------------------
int TSMF::BeginTrans()
{
	int t_result = 0;

	if (!FSQLDB->BeginTrans())
	{
		FEngineErrorsPump->AddError("TSMF", "BeginTrans", FSQLDB->GetLastDBErrorMessage());
		t_result = -1;
	}

	return t_result;
} // BeginTrans()
//---------------------------------------------------------------------------

int TSMF::CommitTrans()
{
	int t_result = 0;

	if (!FSQLDB->CommitTrans())
	{
		FEngineErrorsPump->AddError("TSMF", "CommitTrans", FSQLDB->GetLastDBErrorMessage());
		t_result = -1;
	}

	return t_result;
} // CommitTrans()
//---------------------------------------------------------------------------

int TSMF::RollbackTrans()
{
	int t_result = 0;

	if (!FSQLDB->RollbackTrans())
	{
		FEngineErrorsPump->AddError("TSMF", "RollbackTrans", FSQLDB->GetLastDBErrorMessage());
		t_result = -1;
	}

	return t_result;
} // RollbackTrans()
//---------------------------------------------------------------------------

int TSMF::SavePoint(const char* savepoint_name)
{
	int t_result = 0;

	if (!FSQLDB->SavePoint(savepoint_name))
	{
		FEngineErrorsPump->AddError("TSMF", "SavePoint", FSQLDB->GetLastDBErrorMessage());
		t_result = -1;
	}

	return t_result;
} // SavePoint()
//---------------------------------------------------------------------------

int TSMF::ReleaseSavePoint(const char* savepoint_name)
{
	int t_result = 0;

	if (!FSQLDB->Release(savepoint_name))
	{
		FEngineErrorsPump->AddError("TSMF", "ReleaseSavePoint", FSQLDB->GetLastDBErrorMessage());
		t_result = -1;
	}

	return t_result;
} // ReleaseSavePoint()
//---------------------------------------------------------------------------

int TSMF::RollbackToSavePoint(const char* savepoint_name)
{
	int t_result = 0;

	if (!FSQLDB->RollbackTo(savepoint_name))
	{
		FEngineErrorsPump->AddError("TSMF", "RollbackToSavePoint", FSQLDB->GetLastDBErrorMessage());
		t_result = -1;
	}

	return t_result;
} // RollbackToSavePoint()
//---------------------------------------------------------------------------

int TSMF::Edit( const char* table_name, const int pkey )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->Edit(table_name, pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "Edit",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // Edit
//---------------------------------------------------------------------------

int TSMF::Insert( const char* table_name )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->Insert(table_name);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "Insert",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // Insert
//---------------------------------------------------------------------------

int TSMF::Delete( const char* table_name, const int delete_pkey)
{
	int t_result;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->Delete(table_name, delete_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "Delete",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // Delete
//---------------------------------------------------------------------------

int TSMF::Post( const char* table_name, int* new_pkey )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->Post(table_name, new_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "Post",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // Post
//---------------------------------------------------------------------------

int TSMF::Cancel( const char* table_name )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->Cancel(table_name);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "Cancel",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // Cancel
//---------------------------------------------------------------------------

int TSMF::AddEditSetsPivot(const int editset_pkey, const int edit_pkey)
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->AddEditSetsPivot(editset_pkey, edit_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "AddEditSetsPivot",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;

} // AddEditSetsPivot
//---------------------------------------------------------------------------

int TSMF::DeleteEditSetsPivot(const int editset_pkey, const int edit_pkey)
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->DeleteEditSetsPivot(editset_pkey, edit_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "DeleteEditSetsPivot",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;

} // DeleteEditSetsPivot
//---------------------------------------------------------------------------

int TSMF::AddLayoutsPivot(const int layout_pkey, const int field_pkey,
	const int start_pos)
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->AddLayoutsPivot(layout_pkey, field_pkey, start_pos);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "AddLayoutsPivot",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;

} // AddLayoutsPivot
//---------------------------------------------------------------------------

int TSMF::DeleteLayoutsPivot(const int layout_pkey, const int field_pkey)
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->DeleteLayoutsPivot(layout_pkey, field_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "DeleteLayoutsPivot",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;

} // DeleteLayoutsPivot
//---------------------------------------------------------------------------

int TSMF::SetFieldByName( const char* table_name, const char* field_name,
	const char* value )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->SetFieldByName(table_name, field_name, value);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "SetFieldByName",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // SetFieldByName
//---------------------------------------------------------------------------

int TSMF::SetFieldByNumber( const char* table_name, const int column_number,
	const char* value )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->SetFieldByNumber(table_name, column_number, value);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "SetFieldByNumber",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // SetFieldByNumber
//---------------------------------------------------------------------------
int TSMF::ClearEditLogicPivots( const int edit_pkey )
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->ClearEditLogicPivots(edit_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "ClearEditLogicPivots",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;

} // ClearEditLogicPivots
//---------------------------------------------------------------------------
int TSMF::GetLogicListCount(char which_list, int* how_many_items)
{
	int t_result = 0;

	try
	{
		*how_many_items = FEditCompiler->GetLogicListCount(which_list);
	}
	catch (...)
	{
		t_result = -1;
		// return a count of 0
		*how_many_items = 0;
	}

	return t_result;
} // GetLogicListCount
//---------------------------------------------------------------------------
int TSMF::GetLogicFieldNumber(const int idx, int* item)
{
	int t_result = 0;

	try
	{
		*item = FEditCompiler->GetLogicFieldNumber(idx);
	}
	catch (...)
	{
		t_result = -1;
	}

	return t_result;
} // GetLogicFieldNumber
//---------------------------------------------------------------------------
int TSMF::GetLogicTableName(const int idx, char* table_name,
	const int buff_size)
{
	int t_result = 0;
	try
	{
		std::string t_item = FEditCompiler->GetLogicTableName(idx);

		if (buff_size > 0)
		{
			std::strncpy(table_name, t_item.c_str(), buff_size);
		}
	}
	catch (...)
	{
		t_result = -1;
	}

	return t_result;
} // GetLogicTableName
//---------------------------------------------------------------------------
int TSMF::GetLogicErrorMessage(const int idx, int* msg_func_token,
	char* error_msg, const int buff_size)
{
	// This retrieves the TEXT error messages

	int t_result = 0;
	try
	{
		std::string t_item = FEditCompiler->GetLogicErrorMessage(idx, msg_func_token);

		if (buff_size > 0)
		{
			std::strncpy(error_msg, t_item.c_str(), buff_size);
		}
	}
	catch (...)
	{
		t_result = -1;
	}

	return t_result;
} // GetLogicErrorMessage
//---------------------------------------------------------------------------

int TSMF::GetLogicMessageNumber(const int idx, int* msg_func_token,
	int* msg_number)
{
	// This retrieves the TEXT error messages

	int t_result = 0;
	try
	{
		*msg_number = FEditCompiler->GetLogicMessageNumber(idx, msg_func_token);
	}
	catch (...)
	{
		t_result = -1;
	}

	return t_result;
} // GetLogicMessageNumber
//---------------------------------------------------------------------------
//***************************************************************************
// Support for Import Metafile
//***************************************************************************
int TSMF::OpenImportMetafile(const char* source_smf_name)
{
	int t_result = 0;

	if (FImportMetafile)
	{
		FEngineErrorsPump->AddError("TSMF", "OpenImportMetafile", "Import Metafile is already running");
		t_result = -1;
	}
	else
	{
		FImportMetafile = new TImportMetafile(this, FSQLDB, FEngineErrorsPump, FEw_AccessController, FTablesList);
		t_result = FImportMetafile->Open(source_smf_name);
	}

	return t_result;
} // OpenImportMetafile
//---------------------------------------------------------------------------
int TSMF::CloseImportMetafile()
{
	int t_result = 0;

	if (FImportMetafile)
	{
		delete FImportMetafile;
		FImportMetafile = NULL;
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "CloseImportMetafile", "FImportMetafile is NULL; call OpenImportMetafile first");
		t_result = -1;
	}

	return t_result;
}
//---------------------------------------------------------------------------
int TSMF::GetImportCandidates( const char* table_name,
	const char* match_on_field_name, void* owner, void* callback_func)
{
	int t_result = 0;

	if (FImportMetafile)
	{
		try
		{
			FImportMetafile->GetImportCandidates(
				table_name, match_on_field_name, owner, callback_func);
		}
		catch (std::runtime_error& e)
		{
			std::string t_errmsg = "Something bad has happened: ";
			t_errmsg += e.what();
			FEngineErrorsPump->AddError("TSMF", "GetImportCandidates", t_errmsg);
			t_result = -1;
		}
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "GetImportCandidates", "FImportMetafile is NULL; call OpenImportMetafile first");
		t_result = -1;
	}

	return t_result;
} // GetImportCandidates
//---------------------------------------------------------------------------
int TSMF::ImportLookupTable(const char* table_name)
{
	int t_result = 0;

	if (FImportMetafile)
	{
		try
		{
			t_result = FImportMetafile->ImportLookupTable(table_name);

			if (t_result == 0)
			{
				// Update TableStats
				if (!UpdateTableStats(table_name))
				{
					t_result = -1;
				}
				// This is just a pointer, which we won't be using except
				// for creating tables and the original call to build_tables_list.
				// It's a publicly-accessible member of this class so that the
				// callback functions for retrieving constraint information can
				// access its structure.
				FTableDef = NULL;
			} // if (t_result == 0)
		}
		catch (std::runtime_error& e)
		{
			std::string t_errmsg = "Something bad has happened: ";
			t_errmsg += e.what();
			FEngineErrorsPump->AddError("TSMF", "ImportLookupTable", t_errmsg);
			FTableDef = NULL;
			t_result = -1;
		}
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "ImportLookupTable", "FImportMetafile is NULL; call OpenImportMetafile first");
		t_result = -1;
	}

	return t_result;
} // ImportLookupTable
//---------------------------------------------------------------------------

int TSMF::ImportWritePassword( const int agency_pkey, const char* the_password )
{
	int t_result = 0;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[500];
	std::snprintf(t_buffer, 500,
		"UPDATE [EE_Agencies] SET [Password] = '%s' WHERE [PKey] = %d;",
		the_password, agency_pkey);

	try
	{
		t_qry->SetSQLText(t_buffer);
		if (!t_qry->ExecSQL())
		{
			FEngineErrorsPump->AddError("TSMF", "ImportWritePassword",
				t_qry->LastErrorMessage());
		}

	}
	catch (std::runtime_error& e)
	{
		FEngineErrorsPump->AddError("TSMF", "ImportWritePassword",
			e.what());
		FEngineErrorsPump->AddError("TSMF", "ImportWritePassword",
			FSQLDB->GetLastDBErrorMessage());
		FEngineErrorsPump->AddError("TSMF", "ImportWritePassword",
			t_buffer);
		t_result = -1;
	}

	return t_result;
} // ImportWritePassword
//---------------------------------------------------------------------------


int TSMF::RenameLookupTable(const char* old_table_name,
	const char* new_table_name, bool is_renaming_table,
	const int agency_pkey /* = 1 */)
{
	int t_result = 0;

	//int t_how_many, t_idx;

	TTableDef* t_old_table_def = FTablesList->GetTable(old_table_name);
	if (t_old_table_def)
	{
		std::string t_old_name,
					t_new_name,
					t_old_idx_name,
					t_new_idx_name,
					t_old_trig_name,
					t_new_trig_name,
					t_errmsg;

		t_old_name.assign(old_table_name);
		t_new_name.assign(new_table_name);

		// Index names take the form of "xCS_OBS_INDEX1" in the EMFtoSMF.dll
		// conversion. In order to keep them from getting too unwieldy, we'll
		// truncate the new name to the first 8 bytes; thus, if table
		// CS_OBS is renamed to Collaborative_Stage_Obsolete, we'll prefix
		// its indexes with "xCollabor_"
		if (t_old_name.length() > 8)
			t_old_idx_name = std::string("x") + t_old_name.substr(0,8) + std::string("_");
		else
			t_old_idx_name = std::string("x") + t_old_name + std::string("_");

		if (t_new_name.length() > 8)
			t_new_idx_name = std::string("x") + t_new_name.substr(0,8) + std::string("_");
		else
			t_new_idx_name = std::string("x") + t_new_name + std::string("_");

		uppercase(t_old_name);
		uppercase(t_new_name);

		if (t_old_name.compare(t_new_name) == 0)
		{
			t_errmsg.assign("Old table name and new table name are the same");
			FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);
			t_result = -1;
		}

		if (t_old_name.substr(0,3).compare("EE_") == 0)
		{
			if (is_renaming_table)
				t_errmsg.assign("You may not rename a metafile control table (starts with \"EE_\")");
			else
				t_errmsg.assign("You may not copy a metafile control table (starts with \"EE_\")");
			FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);
			t_result = -1;
		}

		if (t_new_name.substr(0,3).compare("EE_") == 0)
		{
			t_errmsg.assign("You may not start a look-up table name with \"EE_\"");
			FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);
			t_result = -1;
		}

		if (t_result == 0)
		{
			std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

			std::string t_sql,
						t_old_scrap,
						t_new_scrap;

			char t_buffer[500];

			try
			{
				/* EPIC FAIL: Although this statement renames triggers, it doesn't
				 * modify the trigger code, so the old table name remained embedded
				 * in the trigger. The only way to make this work is to create
				 * a new table, copy everything over, and drop the original.
				char t_buffer[500];
				std::snprintf(t_buffer, 500,
					"ALTER TABLE [%s] RENAME TO [%s];",
					old_table_name, new_table_name);

				*/

				std::vector<std::string> t_index_defs;
				std::vector<std::string> t_trigger_defs;

				while (t_result == 0)
				{
					// FIRST UP: Get all the index and trigger definitions...
					// they'll be wiped out once we hit the DROP TABLE instruction,
					// below
					std::snprintf(t_buffer, 500,
						"SELECT [sql] " \
						"FROM [sqlite_master] " \
						"WHERE [tbl_name] LIKE '%s' " \
						"AND [type] LIKE 'index';",
						old_table_name);

					t_qry->SetSQLText(t_buffer);
					t_qry->SetIsActive(true);
					while (!t_qry->Eof())
					{
						t_index_defs.push_back(t_qry->GetFieldByNumber(0));
						t_qry->FetchNext();
					}

					t_qry->SetIsActive(false);

					std::snprintf(t_buffer, 500,
						"SELECT [sql] " \
						"FROM [sqlite_master] " \
						"WHERE [tbl_name] LIKE '%s' " \
						"AND [type] LIKE 'trigger';",
						old_table_name);

					t_qry->SetSQLText(t_buffer);
					t_qry->SetIsActive(true);
					while (!t_qry->Eof())
					{
						t_trigger_defs.push_back(t_qry->GetFieldByNumber(0));
						t_qry->FetchNext();
					}

					t_qry->SetIsActive(false);

					// NB: When working with DDL (data definition language), there
					// can be no rolling back on failure, so transactions are pointless
					std::snprintf(t_buffer, 500,
						"SELECT [sql] " \
						"FROM [sqlite_master] " \
						"WHERE [tbl_name] LIKE '%s' " \
						"AND [type] LIKE 'table';",
						old_table_name);

					t_qry->SetSQLText(t_buffer);
					t_qry->SetIsActive(true);
					t_sql.assign(t_qry->GetFieldByNumber(0));
					t_qry->SetIsActive(false);

					t_old_scrap = std::string("CREATE TABLE ") + std::string(old_table_name);
					t_new_scrap = std::string("CREATE TABLE ") + std::string(new_table_name);
					t_sql = stringReplace(t_sql, t_old_scrap, t_new_scrap);

					t_qry->SetSQLText(t_sql.c_str());

					/////////////////////
					/// CREATE NEW TABLE
					/////////////////////
					if (!t_qry->ExecSQL())
					{
						t_errmsg.assign("CREATE TABLE under new name failed");
						FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);
						t_errmsg.assign("SQL: ");
						t_errmsg += t_qry->GetSQLText();
						FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

						t_result = -1;
						break;
					}

					std::snprintf(t_buffer, 500,
						"INSERT INTO %s " \
						"SELECT * FROM %s;",
						new_table_name, old_table_name);

					t_qry->SetSQLText(t_buffer);

					/////////////////////////////
					/// COPY DATA INTO NEW TABLE
					/////////////////////////////
					if (!t_qry->ExecSQL())
					{
						t_errmsg.assign("Failure to copy data from old table into new table");
						FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

						t_errmsg.assign(t_qry->LastErrorMessage());
						FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

						t_errmsg.assign("SQL: ");
						t_errmsg += t_qry->GetSQLText();
						FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

						t_result = -1;
						break;
					}

					if (is_renaming_table)
					{
						std::snprintf(t_buffer, 500,
							"UPDATE [EE_LookupTables] " \
							"SET [TableName] = '%s' " \
							"WHERE [TableName] = '%s';",
							new_table_name, old_table_name);

						t_qry->SetSQLText(t_buffer);

						/////////////////////////////
						/// CHANGE NAME IN EE_LOOKUPTABLES
						/////////////////////////////
						if (!t_qry->ExecSQL())
						{
							t_errmsg.assign("Failure to update table name in EE_LookupTables");
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign(t_qry->LastErrorMessage());
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign("SQL: ");
							t_errmsg += t_qry->GetSQLText();
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_result = -1;
							break;
						}
					}
					else
					{
						// copying table, not renaming
						char t_date_str[20];
						get_current_date_time(t_date_str, 20);
						std::snprintf(t_buffer, 500,
							"INSERT INTO [EE_LookupTables] " \
							"([AgencyPKey], [TableName], [RecCount], [LastModDate], [Comments]) " \
							"SELECT %d, '%s', a.[RecCount], '%s', a.[Comments] " \
							"FROM [EE_LookupTables] a " \
							"WHERE a.[TableName] = '%s';",
							agency_pkey, new_table_name, t_date_str, old_table_name);

						t_qry->SetSQLText(t_buffer);

						if (!t_qry->ExecSQL())
						{
							t_errmsg.assign("Failure to add new table to EE_LookupTables");
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign(t_qry->LastErrorMessage());
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign("SQL: ");
							t_errmsg += t_qry->GetSQLText();
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_result = -1;
							break;
						}

					} // if copying, not renaming, table

					////////////////////////////////
					/// CREATE INDEXES
					////////////////////////////////
					t_old_scrap = std::string("ON [") + std::string(old_table_name) + std::string("]");
					t_new_scrap = std::string("ON [") + std::string(new_table_name) + std::string("]");
					for (size_t t_idx = 0;
						t_idx < t_index_defs.size() && t_result == 0; t_idx++)
					{
						t_sql.assign(t_index_defs[t_idx]);
						// Note that we captured the old/new index names at the
						// top of this routine.
						t_sql = stringReplace(t_sql, t_old_idx_name, t_new_idx_name);
						t_sql = stringReplace(t_sql, t_old_scrap, t_new_scrap);

						t_qry->SetSQLText(t_sql.c_str());
						if (!t_qry->ExecSQL())
						{
							t_errmsg.assign("Failure to create index");
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign(t_qry->LastErrorMessage());
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign("SQL: ");
							t_errmsg += t_qry->GetSQLText();
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_result = -1;
							break;
						}
					}

					if (t_result != 0)
					{
						break;
					}

					////////////////////////////////
					/// CREATE TRIGGERS
					////////////////////////////////
					t_old_scrap = std::string("UPDATE ") + std::string(old_table_name);
					t_new_scrap = std::string("UPDATE ") + std::string(new_table_name);

					t_old_trig_name = std::string("ON ") + std::string(old_table_name);
					t_new_trig_name = std::string("ON ") + std::string(new_table_name);

					// Note that we captured the old/new index names at the
					// top of this routine. We're going to modify them slightly
					// for re-use as the trigger names, which take this form:
					// "update_CS_OBS_INDEX1" and "insert_CS_OBS_INDEX1"
					// t_old_idx_name presently looks like: "xCS_OBS_", so
					// we'll replace the leading "x" with "_" and it will
					// work nicely for the trigger renaming.
					t_old_idx_name = std::string("_") + t_old_idx_name.substr(1, t_old_idx_name.length());
					t_new_idx_name = std::string("_") + t_new_idx_name.substr(1, t_new_idx_name.length());

					for (size_t t_idx = 0;
						t_idx < t_trigger_defs.size() && t_result == 0; t_idx++)
					{
						t_sql.assign(t_trigger_defs[t_idx]);

						t_sql = stringReplace(t_sql, t_old_idx_name, t_new_idx_name);
						t_sql = stringReplace(t_sql, t_old_trig_name, t_new_trig_name);
						t_sql = stringReplace(t_sql, t_old_scrap, t_new_scrap);

						t_qry->SetSQLText(t_sql.c_str());
						if (!t_qry->ExecSQL())
						{
							t_errmsg.assign("Failure to create trigger");
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign(t_qry->LastErrorMessage());
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign("SQL: ");
							t_errmsg += t_qry->GetSQLText();
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_result = -1;
							break;
						}
					}

					if (t_result != 0)
					{
						break;
					}


					if (is_renaming_table)
					{
						std::snprintf(t_buffer, 500,
							"DROP TABLE IF EXISTS %s;",
							old_table_name);

						t_qry->SetSQLText(t_buffer);

						/////////////////////////////
						/// DROP OLD TABLE
						/////////////////////////////
						if (!t_qry->ExecSQL())
						{
							t_errmsg.assign("Failure to drop old table from the database (metafile)");
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign(t_qry->LastErrorMessage());
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_errmsg.assign("SQL: ");
							t_errmsg += t_qry->GetSQLText();
							FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

							t_result = -1;
							break;
						}
					}

					// We made it! The table has been recreated under a new name
					// and the old table has been dropped (if renaming ... not if copying)
					break;
				} // while (t_result == 0)

				// This loop just makes bailing out easier
				while (t_result == 0)
				{
					std::unique_ptr<TSQLQuery> t_struct_qry(new TSQLQuery(FSQLDB));

					if ( !create_table_def(new_table_name, t_struct_qry.get()) )
					{
						FTableDef = NULL;
						t_result = -1;
						break;
					}

					if (!add_table_def_indexes(t_new_name))
					{
						delete FTableDef;
						FTableDef = NULL;
						t_result = -1;
						break;
					}

					if (!add_table_def_constraints(t_new_name))
					{
						delete FTableDef;
						FTableDef = NULL;
						t_result = -1;
						break;
					}

					if (!add_table_triggers(t_new_name))
					{
						delete FTableDef;
						FTableDef = NULL;
						throw std::runtime_error("Failure in call to add_table_triggers");
					}

					// Use t_old_table_def to retrieve default constraints
					if (!copy_table_defaults(t_old_table_def, FTableDef))
					{
						delete FTableDef;
						FTableDef = NULL;
						throw std::runtime_error("Failure in call to copy_table_defaults");
					}

					// Use t_old_table_def to retrieve check constraints
					if (!copy_table_checks(t_old_table_def, FTableDef))
					{
						delete FTableDef;
						FTableDef = NULL;
						throw std::runtime_error("Failure in call to copy_table_checks");
					}

					FTablesList->AddTable(FTableDef);

					if (is_renaming_table)
					{
						FTablesList->RemoveTable(t_old_table_def);
					}

					break;
				}

			}
			catch (std::runtime_error& e)
			{
				t_errmsg.assign(e.what());
				FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);
				t_errmsg.assign(t_qry->LastErrorMessage());
				FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_errmsg);

				t_result = -1;
			}
		}

	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += old_table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMF", "RenameLookupTable", t_error);
		t_result = -1;
	}

	return t_result;
} // RenameLookupTable
//---------------------------------------------------------------------------

int TSMF::DefineTableStructure( const char* table_name, const int restructure_existing)
{
	int t_result = 0;
	if (FSMFWriter)
	{
		t_result = FSMFWriter->DefineTableStructure(table_name, restructure_existing);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "DefineTableStructure",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}
	return t_result;
} // DefineTableStructure
//---------------------------------------------------------------------------
int TSMF::AddColumnDef(
				 const char* col_name,
				 const char* copy_from_col_name,
				 const int data_type,
				 const int data_length,
				 const int is_nullable,
				 const int is_trigger_field,
				 const char* check_limit_to_list,
				 const char* check_range_low,
				 const char* check_range_high,
				 const char* default_value,
				 const int index_status)
{

	int t_result = 0;

	if (FSMFWriter)
	{
		t_result = FSMFWriter->AddColumnDef(col_name,
											copy_from_col_name,
											data_type,
											data_length,
											is_nullable,
											is_trigger_field,
											check_limit_to_list,
											check_range_low,
											check_range_high,
											default_value,
											index_status);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "AddColumnDef",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}

	return t_result;
} // AddColumnDef
//---------------------------------------------------------------------------

int TSMF::CreateTable(int* lookup_pkey)
{
	int t_result = 0;

	if (FSMFWriter)
	{
		t_result = FSMFWriter->CreateTable(lookup_pkey);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMF", "CreateTable",
			"Engine programmer error: FSMFWriter not instantiated");
		t_result = -1;
	}

	return t_result;
} // CreateTable
//---------------------------------------------------------------------------
#endif  // __SQLEW__


