//---------------------------------------------------------------------------

#ifndef SMFH
#define SMFH

#include <string>
#include <map>

#include "edit_external.h"
#include "TableStats.h"

//---------------------------------------------------------------------------

// NOTE: These functions are implemented in this unit for convenience;
// they are NOT member functions of class TSMF.
int CreateNewMetafile(const char* smf_name);
int CompactMetafile(const char* smf_name);
int GetEngineVersion(char* version_buffer, int buffer_length);
int GetMetafileStructureVersion(char* smf_structure_version, int smf_version_size);

/****************************************************************************
 ***              DEFAULT ERROR CODE FROM std::exception ERRORS
 *** 06/29/2004 KB: This is the last-ditch error code to return if the engine
 *** fails in some awful way that isn't expected or addressed by any of our
 *** documented error codes.
 ****************************************************************************/
#define NO_SUCH_ERROR_CODE        -1
#define BAD_SMFID_ERROR_CODE      -2
#define STD_EXCEPTION_ERROR_CODE  -3

// Forward declarations
class TSQLDB;
class TSQLQuery;
class TEngineErrorsPump;
class TRunEdits;
#ifdef __SQLEW__
class TEw_AccessController;
class TSMFWriter;
class TEditCompiler;
class TImportMetafile;
#endif

class TSMF
{
private:
	TSQLDB* FSQLDB;

	TTablesList* FTablesList;
	TTableDef* FTableDef;

	TEngineErrorsPump* FEngineErrorsPump;

	TRunEdits* FRunEdits;


	// FQueriesList will hold query objects where the SQLText is like this:
	// SELECT * FROM [table_name] WHERE PKey = [int_value];
	// That way, we can more efficiently run get_data_by_field_name and
	// get_data_by_column_number.
	std::map<std::string, TSQLQuery*> FQueriesList;

	// FCursorsList holds query objects where the SQLText is written
	// by the caller. It can be any SELECT statement that runs in
	// SQLite and references objects in the metafile. These queries
	// are for the cursor-access functions.
	std::map<std::string, TSQLQuery*> FCursorsList;

#ifdef __SQLEW__
	TEw_AccessController* FEw_AccessController;
	TSMFWriter* FSMFWriter;
	TImportMetafile* FImportMetafile;
#endif // __SQLEW__

	TEditCompiler* FEditCompiler;

	// Parameterized queries for build_tables_list
	TSQLQuery* FAddTableDefIndexesQry;
	TSQLQuery* FAddTableDefConstraintsQry;
	TSQLQuery* FAddTableDefTriggersQry;

	bool build_tables_list();
	bool create_table_def(const std::string table_name, TSQLQuery* qry);
	bool add_table_def_indexes(const std::string table_name);
	bool add_table_def_constraints(const std::string table_name);
	bool add_table_triggers(const std::string table_name);
	bool copy_table_defaults(TTableDef* old_table_def, TTableDef* new_table_def);
	bool copy_table_checks(TTableDef* old_table_def, TTableDef* new_table_def);

	std::string field_data_type_as_string(TFieldDef* fld);
	std::string field_nullable_as_string(TFieldDef* fld);

	// Support function for GetDataSizeByColumnOrder and GetDataByColumnOrder
	std::string get_data_by_column_order(const char* table_name, int pkey,	const int column_order);
	// Support function for GetDataSizeByFieldName and GetDataByFieldName
	std::string get_data_by_field_name(const char* table_name, int pkey, const char* field_name);

	// Support for TSMF(), verify that the metafile structure version
	// is AT LEAST the value the engine expects
	bool check_mf_structure_version(char* smf_init_error, const int error_buffer_size, int* version_compare);

	// Support for Edit_UpdateRecordByName and Edit_DisplayDataRecordByName
	std::string get_layout_tag_from_name(const char* layout_name);

	bool sql_is_valid(std::string& sql, std::string& err_msg);
	bool cursor_name_exists(std::string& cursor_name, std::string& err_msg);
	TSQLQuery* get_cursor_from_list(std::string& cursor_name);

	// support for RecompileEditLogic
	bool update_edit_fields_pivot(int edit_pkey, std::vector<int>& fields_list);
	bool update_default_message_in_pivot(int edit_pkey, int t_default_message_pkey);
	bool update_edit_messages_pivot(int edit_pkey);
	bool update_edit_errors_pivot(int edit_pkey);
	bool update_edit_tables_pivot(int edit_pkey);

public:
	TSMF();
	~TSMF();

	int InitializeMetafile(const char* smf_name, char* smf_init_error, const int error_buffer_size,
		int* version_compare, const int check_version = 1 );
	int UpdateMetafile(char* smf_init_error, const int error_buffer_size);

	int GetMetafileVersion(char* smf_version, int smf_version_size);
	int GetMetafileComment(char* smf_comment, int smf_comment_size);

	int ClearEngineErrorsPump();
	int GetEngineErrorsCount(int* how_many_errors);
	int GetEngineErrorByIndex(int idx, char* error_message, int message_length);


	// for use by callbacks invoked from build_tables_list()
	TTableDef* GetCurrentTableDef();
	// For calls generated within TSMF as well as from TSMFWriter, this needs to be public
	bool UpdateTableStats(const char* table_name);

	// Interface support functions
	int GetPKeys( const char* table_name, const char* order_by, void* owner, void* callback_func );

	// Check edit set against layout for edits that will skip silently
	int CheckEditSetAgainstLayout( const int edit_set_pkey, const int layout_pkey, void* owner, void* callback_func );
	// Get metafile table names, using callback
	int GetTableNames( void* owner, void* callback_func );
	// or get them without a callback
	int GetTablesCount(int* how_many_tables );
	int GetTableNameByIndex( const int idx, char* buffer, const int buff_size );

	// Retrieve index definitions
	int GetIndexCount(const char* table_name, int* how_many_indexes);
	int GetIndex( const char* table_name, const int idx, char* index_sql, const int buff_size );

	// Retrieve trigger definitions
	int GetTriggerCount(const char* table_name, int* how_many_triggers);
	int GetTrigger( const char* table_name, const int idx, char* trigger_sql, const int buff_size );

	// Get a table structure, using callback
	int GetTableStructure(const char* table_name, void* owner, void* callback_func );
	// or get the column list without a callback
	int GetColumnsCount(const char* table_name, int* how_many_columns );
	int GetColumnStructure( const char* table_name, int column_order, char* field_name, int fname_size, char* data_type, int type_size, char* nullability, int null_size, int* is_trigger_field);

	// Other service functions:
	// Give me a table and field name, I'll tell you its column order
	int GetColumnOrder(const char* table_name, const char* field_name, int* col_order );
	// Give me a table and column order, I'll tell you its field name
	int GetFieldName(const char* table_name, const int col_order, char* field_name, const int fldname_size );


	// Read data from any table in the metafile, using callback
	int ReadData(const char* table_name, int pkey, void* owner, void* callback_func );

	// Take a "fresh look" at the data (refresh the query) for calls to GetDataBy* functions
	int RefreshReadData( const char* table_name, const int pkey );

	// or get the data without a callback, referencing field by column order
	int GetDataSizeByColumnOrder( const char* table_name, const int pkey, const int column_order, int* data_size );
	int GetDataByColumnOrder( const char* table_name, const int pkey, const int column_order, char* buffer, const int buff_size );
	// get the data without a callback, referencing field by field name
	int GetDataSizeByFieldName( const char* table_name,	const int pkey, const char* field_name, int* data_size );
	int GetDataByFieldName( const char* table_name, const int pkey, const char* field_name, char* buffer, const int buff_size );

	// Cursor access functions
	int OpenCursor(const char* cursor_name, const char* sql);
	int CloseCursor(const char* cursor_name);
	int PrepareParameterizedCursor(const char* cursor_name, const char* sql);
	int BindIntParameter(const char* cursor_name, const int param_num, const int value);
	int BindTextParameter(const char* cursor_name, const int param_num, const char* value);
	int OpenParameterizedCursor(const char* cursor_name);
	int ResetParameterizedCursor(const char* cursor_name);
	int ReleaseParameterizedCursor(const char* cursor_name);
	int CursorRowCount(const char* cursor_name, int* row_count);
	int CursorColCount(const char* cursor_name, int* col_count);
	int CursorNext(const char* cursor_name );
	int CursorEof(const char* cursor_name, int* is_eof);
	int ReadCursorData(const char* cursor_name, void* owner, void* callback_func );
	int GetCursorDataSizeByNumber( const char* cursor_name, const int column_number, int* data_size );
	int GetCursorDataByNumber( const char* cursor_name, const int column_number, char* buffer, const int buff_size );
	int GetCursorDataSizeByName( const char* cursor_name, const char* field_name, int* data_size );
	int GetCursorDataByName( const char* cursor_name, const char* field_name, char* buffer, const int buff_size );
	int GetCursorIntByNumber( const char* cursor_name, const int column_number, int* data );
	int GetCursorIntByName( const char* cursor_name, const char* field_name, int* data );

// SUPPORT FOR RUNNING EDITS
	int UpdateRecordByTag( const char* layout_tag, const char* data);
	int UpdateRecordByName( const char* layout_name, const char* data);

	int DisplayDataRecordByTag( const char* layout_tag, void* owner, void* callback_func );
	int DisplayDataRecordByName( const char* layout_name, void* owner, void* callback_func );

	int DisplayFieldInMessagesAs( const int field_number, const char* display_name );

	int NeverSkipEmptyField( const int field_number, const int never_skip_empty );

	int FieldDataByNumber(
		const char* layout_tag, const int field_number, int* pkey,
		int* start_pos, int* field_length,
		char* field_name, const int field_name_len,
		char* display_name, const int display_name_len,
		char* data_type, const int data_type_len,
		char* value, const int value_len );

	int LoadEditSet( const char* edit_set_tag, const char* layout_tag );
	int UnloadEditSet( const char* edit_set_tag, const char* layout_tag );

	int SetEditOptions( const char* edit_set_tag, const char* edit_tag,
		const int suppress_edit, const int skip_if_empty, const int skip_if_failed,
		const int no_skips_override );
	int GetEditOptions( const char* edit_set_tag, const char* edit_tag,
		int* is_suppressed, int* is_skip_if_empty, int* is_skip_if_failed,
		int* is_no_skips_override );
	int ClearEditOptions( const char* edit_set_tag );

	int RunEdits(const char* edit_set_tag, const char* layout_tag, const char* data, const int edit_options, int* errors_count, void* owner, void* callback_func);
	int GetRunEditsSummary(const char* edit_set_tag, const char* layout_tag,
		const int include_zero_counts, int* total_errors, int* total_warnings, int* total_skips, int* total_messages,
		int* total_warning_messages, void* owner, void* callback_func);
	int GetEditDetails(const char* edit_tag, void* owner, void* callback_func);
	int DebugEdit(const char* edit_tag, const char* layout_tag,	const char* data, void* owner, void* logic_callback, void* variables_callback, void* fields_callback, void* notes_callback);

	int RecompileEditLogic( const char* edit_tag );
	int CheckSyntax( const char* edit_logic, int* how_many_errors, int* how_many_warnings);
	int GetSyntaxError( const int error_idx, int* line_number, char* syntax_error, const int syntax_error_len);
	int GetSyntaxWarning( const int warning_idx, int* line_number, char* syntax_warning, const int syntax_warning_len);

    int UpdateAllAccessField( const char* table_name, const int pkey, const char* field_name, const char* value );

#ifdef __SQLEW__
	int Login( const char* agency_code, const char* password );
	int GetLoginsCount( int* how_many_logins );
	int GetLoginByIndex( const int idx, int* agency_pkey, char* agency_code, const int agency_code_size );

	int ValidateUnique( const char* table_name, const int pkey, const char* field_name, const char* value, int* data_is_unique, int* dupe_pkey );

	int BeginTrans();
	int CommitTrans();
	int RollbackTrans();

	int SavePoint( const char* savepoint_name );
	int ReleaseSavePoint( const char* savepoint_name );
	int RollbackToSavePoint( const char* savepoint_name );

	int Edit( const char* table_name, const int pkey );
	int Insert( const char* table_name );
	int Delete( const char* table_name, const int delete_pkey);
	int Post( const char* table_name, int* new_pkey );
	int Cancel( const char* table_name );

	int AddEditSetsPivot(const int editset_pkey, const int edit_pkey);
	int DeleteEditSetsPivot(const int editset_pkey, const int edit_pkey);

	int AddLayoutsPivot(const int layout_pkey, const int field_pkey, const int start_pos);
	int DeleteLayoutsPivot(const int layout_pkey, const int field_pkey);

	int SetFieldByName( const char* table_name, const char* field_name, const char* value );
	int SetFieldByNumber( const char* table_name, const int column_number, const char* value );

	int GetLogicListCount(char which_list, int* how_many_items);
	int GetLogicFieldNumber(const int idx, int* item);
	int GetLogicTableName(const int idx, char* table_name, const int buff_size);
	int GetLogicErrorMessage(const int idx, int* msg_func_token, char* error_msg, const int buff_size);
	int GetLogicMessageNumber(const int idx, int* msg_func_token, int* msg_number);

	// Lookup table maintenance (Data Definition Language foo
	int DefineTableStructure(const char* table_name, const int restructure_existing);
	int AddColumnDef(const char* col_name,
					 const char* copy_from_col_name,
					 const int data_type,
					 const int data_length,
					 const int is_nullable,
					 const int is_trigger_field,
					 const char* check_limit_to_list,
					 const char* check_range_low,
					 const char* check_range_high,
					 const char* default_value,
					 const int index_status);
	int CreateTable(int* lookup_pkey);

	int ClearEditLogicPivots( const int edit_pkey );
	int RenameLookupTable(const char* old_table_name, const char* new_table_name,
	bool is_renaming_table, const int agency_pkey = 1);

	int OpenImportMetafile(const char* source_smf_name);
	int CloseImportMetafile();
	int ImportLookupTable(const char* table_name);
	int GetImportCandidates( const char* table_name, const char* match_on_field_name, void* owner, void* callback_func);
	int ImportWritePassword( const int agency_pkey, const char* the_password );


#endif // __SQLEW__

};

#endif
