//---------------------------------------------------------------------------
#include <algorithm>  // std::for_each
#include <cstdlib>    // std::atoi
#include <cstdio>     // std::snprintf
#include <cstring>    // std::strlen, std::memset
#include <memory>     // smart pointers
#include <stdexcept>

#include "SMFWriter.h"
#include "SMF.h"        // RecompileEditLogic
#include "SQLDB.h"
#include "SQLQuery.h"
#include "EngineErrorsPump.h"
#include "Ew_AccessController.h"
#include "Utility.h"
//---------------------------------------------------------------------------

#ifdef __SQLEW__

// TUpdateEditLogicObject exists for TSMFWriter::post_edit, which is watching
// for modifications to specific fields in specific MCO tables, which will
// require cascading the changes to edit logic and (for changes to EE_Fields.FieldLength)
// record layouts.

TUpdateEditLogicObject::TUpdateEditLogicObject(
					   const char* table_name,
					   const char* field_name,
					   const char* original_value,
					   const char* new_value)
{
	FTableName.assign(table_name);
	FFieldName.assign(field_name);
	FOriginalValue.assign(original_value);
	FValue.assign(new_value);
}
//---------------------------------------------------------------------------

const char* TUpdateEditLogicObject::GetTableName()
{
	return FTableName.c_str();
}
//---------------------------------------------------------------------------
const char* TUpdateEditLogicObject::GetFieldName()
{
	return FFieldName.c_str();
}
//---------------------------------------------------------------------------
const char* TUpdateEditLogicObject::GetOriginalValue()
{
	return FOriginalValue.c_str();
}
//---------------------------------------------------------------------------
const char* TUpdateEditLogicObject::GetNewValue()
{
	return FValue.c_str();
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// class TImportMetafile
TImportMetafile::TImportMetafile(TSMF* parent, TSQLDB* database,
	TEngineErrorsPump* pump, TEw_AccessController* access_controller,
	TTablesList* tables_list)
{
	FParent = parent;
	FSQLDB = database;
	FEngineErrorsPump = pump;
	FAccessController = access_controller;
	FTablesList = tables_list;

	FCallbackOwner = NULL;
	FCallbackFunction = NULL;

	FSourceName.clear();
	// See attach_to_metafile, ImportLookupTable, detach_from_metafile
	FAttachedDBAlias.clear();
} // ctor
//---------------------------------------------------------------------------
TImportMetafile::~TImportMetafile()
{
	Close();
} // dtor
//---------------------------------------------------------------------------
int TImportMetafile::Open(const char* source_metafile_name)
{
	int t_result = 0;
	// For now, let's prevent attaching more than one metafile at a time
	if (!FAttachedDBAlias.empty())
	{
		FEngineErrorsPump->AddError("TImportMetafile", "Open", "Another instance of TImportMetafile is active");
		t_result = -1;
	}
	else
	{
		FSourceName.assign(source_metafile_name);
		if (attach_to_metafile(FSourceName) == 0)
		{
			FAccessController->AddImportSuperUser(FWasAlreadySuperUser);
		}
	}

	return t_result;

} // Open
//---------------------------------------------------------------------------
int TImportMetafile::Close()
{
	// Remove the user's super-user privileges, unless she was already
	// logged in as the super-user
	FAccessController->RemoveImportSuperUser(FWasAlreadySuperUser);
	int t_result = detach_from_metafile();

	return t_result;
} // Close
//---------------------------------------------------------------------------

int TImportMetafile::attach_to_metafile(const std::string source_smf_name)
{
	int t_result = 0;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "pragma autocommit(false); ATTACH DATABASE '";
	t_sql += source_smf_name;
	t_sql += "' AS SourceDB;";

	try
	{
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->ExecSQL();

		// If we make it to here, we're attached. Save the alias name
		FAttachedDBAlias = "SourceDB";
	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot attach to metafile \"";
		t_error += source_smf_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TImportMetafile", "attach_to_metafile", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "attach_to_metafile", t_what);
		}
		t_error = "SQL: " + t_sql;
		FEngineErrorsPump->AddError("TImportMetafile", "attach_to_metafile", t_error);

		t_result = -1;
	}

	return t_result;
} // attach_to_metafile
//---------------------------------------------------------------------------
int TImportMetafile::detach_from_metafile()
{
	int t_result = 0;

	if (!FAttachedDBAlias.empty())
	{
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		std::string t_sql = "DETACH DATABASE ";
		t_sql += FAttachedDBAlias;
		t_sql += "; pragma autocommit(true);";

		try
		{
			t_qry->SetSQLText(t_sql.c_str());

			t_qry->ExecSQL();

			// If we make it to here, we're detached. Clear the alias name
			FAttachedDBAlias.clear();
		}
		catch (std::runtime_error& e)
		{
			std::string t_error = "Cannot detach from metafile aliased \"";
			t_error += FAttachedDBAlias;
			t_error += "\"";
			FEngineErrorsPump->AddError("TImportMetafile", "detach_from_metafile", t_error);
			std::string t_what = e.what();
			if (!t_what.empty())
			{
				FEngineErrorsPump->AddError("TImportMetafile", "detach_from_metafile", t_what);
			}
			t_error = "SQL: " + t_sql;
			FEngineErrorsPump->AddError("TImportMetafile", "detach_from_metafile", t_error);

			t_result = -1;
		}
	}
	else
	{
		t_result = -1;
		std::string t_error = "No attached metafile (did you call attach_to_metafile()?)";
		FEngineErrorsPump->AddError("TImportMetafile", "detach_from_metafile", t_error);
	}
	return t_result;
} // detach_from_metafile
//---------------------------------------------------------------------------
int TImportMetafile::GetImportCandidates( const char* table_name,
	const char* match_on_field_name, void* owner, void* callback_func)
{
	int t_result = 0;

	if (owner && callback_func)
	{
		FCallbackOwner = owner;
		FCallbackFunction = (CallbackImportMetafile)callback_func;

		try
		{
			if (posix_stricmp(table_name, "EE_Agencies") == 0)
			{
				t_result = get_agencies();
			}
			else if (posix_stricmp(table_name, "EE_Edits") == 0)
			{
				t_result = get_edits(match_on_field_name);
			}
			else if (posix_stricmp(table_name, "EE_EditSets") == 0)
			{
				t_result = get_edit_sets(match_on_field_name);
			}
			else if (posix_stricmp(table_name, "EE_Fields") == 0)
			{
				t_result = get_fields();
			}
			else if (posix_stricmp(table_name, "EE_Layouts") == 0)
			{
				t_result = get_layouts(match_on_field_name);
			}
			else if (posix_stricmp(table_name, "EE_Messages") == 0)
			{
				t_result = get_messages();
			}
			else if (posix_stricmp(table_name, "EE_LookupTables") == 0)
			{
				t_result = get_tables();
			}
			else
			{
				char t_buffer[100];
				std::snprintf(t_buffer, 100,
					"Table '%s' is not a Metafile Consituent Object",
					table_name);

				FEngineErrorsPump->AddError("TImportMetafile", "GetImportCandidates", t_buffer);

				t_result = -1;
			}

			FCallbackOwner = NULL;
			FCallbackFunction = NULL;

		}
		catch(...)
		{
			FCallbackOwner = NULL;
			FCallbackFunction = NULL;
		}
	}
	else
	{
		std::string t_error = "Callback owner and callback function must not be NULL";
		FEngineErrorsPump->AddError("TImportMetafile", "GetImportCandidates", t_error);

		t_result = -1;
	}

	return t_result;
} // GetImportCandidates
//---------------------------------------------------------------------------
int TImportMetafile::get_agencies()
{
	int t_result = 0;

	std::string t_table_name = "EE_Agencies";
	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, OrgName
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey as SrcPKey, src.OrgName, src.AdminCode ";
	t_sql += "FROM SourceDB.EE_Agencies src ";
	t_sql += "WHERE NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_Agencies targ ";
	t_sql += "WHERE targ.AdminCode = src.AdminCode ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey,
				t_record_key.c_str(), t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones
		t_status = 2; // Changed

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.OrgName, src.AdminCode " \
			"FROM SourceDB.EE_Agencies src, EE_Agencies targ " \
			"WHERE src.AdminCode = targ.AdminCode " \
			"AND (" \
			"   src.OrgName != targ.OrgName " \
			"OR src.Comments != targ.Comments);"
		);

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey,
				t_record_key.c_str(), t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Finally, identify the objects where a conflict exists on
		// AdminCode/OrgName
		t_status = 3;  // Conflict resolution required
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.OrgName, src.AdminCode " \
			"FROM SourceDB.EE_Agencies src, EE_Agencies targ " \
			"WHERE (src.OrgName = targ.OrgName " \
			"AND src.AdminCode != targ.AdminCode) OR " \
			"(src.AdminCode = targ.AdminCode " \
			"AND src.OrgName != targ.OrgName);");

		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey,
				t_record_key.c_str(), t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Agencies from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_agencies", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_agencies", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_agencies", t_error);

		t_result = -1;
	}
	return t_result;
} // get_agencies
//---------------------------------------------------------------------------
int TImportMetafile::get_edits(const char* match_on_field_name)
{
	int t_result = 0;

	std::string t_table_name = "EE_Edits";
	bool t_match_tag = (posix_stricmp(match_on_field_name, "EditTag") == 0);

	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	// 3 = New, but conflict on Name/Tag exists
	// 4 = Changed, but conflict on Name/Tag exists
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, EditName
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey as SrcPKey, src.EditName, sa.AdminCode ";
	t_sql += "FROM SourceDB.EE_Edits src, SourceDB.EE_Agencies sa ";
	t_sql += "WHERE src.AgencyPKey = sa.PKey ";
	t_sql += "AND NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_Edits targ ";
	if (t_match_tag)
		t_sql += "WHERE targ.EditTag = src.EditTag ); ";
	else
		t_sql += "WHERE targ.EditName = src.EditName ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones
		t_status = 2; // Changed

		if (t_match_tag)
		{
			t_sql = std::string(
				"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditName, sa.AdminCode " \
				"FROM SourceDB.EE_Edits src, SourceDB.EE_Agencies sa, SourceDB.EE_Messages sm, " \
				"	 EE_Edits targ, EE_Agencies ta, EE_Messages tm " \
				"WHERE sa.PKey = src.AgencyPKey " \
				"AND ta.PKey = targ.AgencyPKey " \
				"AND sm.PKey = src.MessagePKey " \
				"AND tm.PKey = targ.MessagePKey " \
				"AND src.EditTag = targ.EditTag " \
				"AND ( " \
				"	src.EditName != targ.EditName " \
				" OR sa.AdminCode != ta.AdminCode " \
				" OR sm.MessageNumber != tm.MessageNumber " \
				" OR src.Description != targ.Description " \
				" OR src.EditLogic != targ.EditLogic " \
				" OR src.AdminNotes != targ.AdminNotes " \
				");");
		 }
		 else
		 {
			t_sql = std::string(
				"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditName, sa.AdminCode " \
				"FROM SourceDB.EE_Edits src, SourceDB.EE_Agencies sa, SourceDB.EE_Messages sm, " \
				"	 EE_Edits targ, EE_Agencies ta, EE_Messages tm " \
				"WHERE sa.PKey = src.AgencyPKey " \
				"AND ta.PKey = targ.AgencyPKey " \
				"AND sm.PKey = src.MessagePKey " \
				"AND tm.PKey = targ.MessagePKey " \
				"AND src.EditName = targ.EditName " \
				"AND ( " \
				"	src.EditTag != targ.EditTag " \
				" OR sa.AdminCode != ta.AdminCode " \
				" OR sm.MessageNumber != tm.MessageNumber " \
				" OR src.Description != targ.Description " \
				" OR src.EditLogic != targ.EditLogic " \
				" OR src.AdminNotes != targ.AdminNotes " \
				");");
		 }

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Finally, identify the objects where a conflict exists on
		// EditName/EditTag
		t_status = 3; // Conflict resolution required
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditName, sa.AdminCode " \
			"FROM SourceDB.EE_Edits src, EE_Edits targ, SourceDB.EE_Agencies sa " \
			"WHERE (src.AgencyPKey = sa.PKey) " \
			"AND (src.EditTag = targ.EditTag " \
			"AND src.EditName != targ.EditName) OR " \
			"(src.EditName = targ.EditName " \
			"AND src.EditTag != targ.EditTag);");
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);
	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Edits from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_edits", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_edits", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_edits", t_error);

		t_result = -1;
	}
	return t_result;
} // get_edits
//---------------------------------------------------------------------------
int TImportMetafile::get_edit_sets(const char* match_on_field_name)
{
	int t_result = 0;

	std::string t_table_name = "EE_EditSets";
	bool t_match_tag = (posix_stricmp(match_on_field_name, "EditSetTag") == 0);

	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	// 3 = New, but conflict on Name/Tag exists
	// 4 = Changed, but conflict on Name/Tag exists
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, EditSetName
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey as SrcPKey, src.EditSetName, sa.AdminCode ";
	t_sql += "FROM SourceDB.EE_EditSets src, SourceDB.EE_Agencies sa ";
	t_sql += "WHERE src.AgencyPKey = sa.PKey ";
	t_sql += "AND NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_EditSets targ ";
	if (t_match_tag)
		t_sql += "WHERE targ.EditSetTag = src.EditSetTag ); ";
	else
		t_sql += "WHERE targ.EditSetName = src.EditSetName ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones
		t_status = 2;

		if (t_match_tag)
		{
			t_sql = std::string(
				"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditSetName, sa.AdminCode " \
				"FROM SourceDB.EE_EditSets src, SourceDB.EE_Agencies sa, " \
				"EE_EditSets targ, EE_Agencies ta " \
				"WHERE sa.PKey = src.AgencyPKey " \
				"AND ta.PKey = targ.AgencyPKey " \
				"AND src.EditSetTag = targ.EditSetTag " \
				"AND ( " \
				"   src.EditSetName != targ.EditSetName " \
				" OR sa.AdminCode != ta.AdminCode " \
				" OR src.Description != targ.Description);");
		}
		else
		{
			t_sql = std::string(
				"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditSetName, sa.AdminCode " \
				"FROM SourceDB.EE_EditSets src, SourceDB.EE_Agencies sa, " \
				"EE_EditSets targ, EE_Agencies ta " \
				"WHERE sa.PKey = src.AgencyPKey " \
				"AND ta.PKey = targ.AgencyPKey " \
				"AND src.EditSetName = targ.EditSetName " \
				"AND ( " \
				"   src.EditSetTag != targ.EditSetTag " \
				" OR sa.AdminCode != ta.AdminCode " \
				" OR src.Description != targ.Description);");
		}

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Finally, identify the objects where a conflict exists on
		// EditSetName/EditSetTag
		t_status = 3; // Conflict resolution required
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditSetName, sa.AdminCode " \
			"FROM SourceDB.EE_EditSets src, EE_EditSets targ, SourceDB.EE_Agencies sa " \
			"WHERE (src.AgencyPKey = sa.PKey) " \
			"AND (src.EditSetTag = targ.EditSetTag " \
			"AND src.EditSetName != targ.EditSetName) OR " \
			"(src.EditSetName = targ.EditSetName " \
			"AND src.EditSetTag != targ.EditSetTag);");
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// No, THIS is "finally": If a user imported the edit set without
		// dependencies, the foregoing tests would not find a discrepancy.
		// So for those where the EditSetName and EditSetTag DO match,
		// drill down into the lists of Edits to see if a discrepancy exists.

		t_status = 2;    // These will come up "Changed"
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.EditSetName, sa.AdminCode " \
			"FROM SourceDB.EE_EditSets src, EE_EditSets targ, SourceDB.EE_Agencies sa " \
			"WHERE src.AgencyPKey = sa.PKey " \
			"AND src.EditSetTag = targ.EditSetTag " \
			"AND src.EditSetName = targ.EditSetName;");

		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			if (!editset_edits_match(t_source_pkey, t_target_pkey, t_match_tag))
			{
				(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
					t_source_pkey, t_target_pkey, t_record_key.c_str(),
					t_admin_code.c_str(), t_status);
			}

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);
	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Edit Sets from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_edit_sets", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_edit_sets", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_edit_sets", t_error);

		t_result = -1;
	}


	return t_result;
} // get_edit_sets
//---------------------------------------------------------------------------
bool TImportMetafile::editset_edits_match(int src_editset_pkey,
	int targ_editset_pkey, bool match_on_tag)
{
	// Called by get_edit_sets:
	// Compare the Edits lists of the Edit Set in the Source and Target
	// databases. If they disagree on existence, the match is false.
	// NOTE: I am having to do the comparison for the edits looking at
	// BOTH match by EditTag AND match by EditName, because it is too much
	// of a hairball to modify the call into Ew_GetImportCandidates to take
	// that extra parameter for this one circumstance.

	bool t_result = true;

	int t_count = 0;
	char t_sql[1001];
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	if (match_on_tag)
	{
		// Matching Source and Target on EditSetTag
		t_count = std::snprintf(t_sql, 1000,
			"SELECT srced.EditTag, srced.EditName " \
			"FROM SourceDB.EE_EditSetEditsPivot srcpiv, " \
			"  SourceDB.EE_EditSets srcset, SourceDB.EE_Edits srced " \
			"WHERE srcset.PKey = srcpiv.EditSetPKey " \
			"AND srced.PKey=srcpiv.EditPKey " \
			"AND srcpiv.EditSetPKey = %d " \
			"AND NOT EXISTS ( " \
			"   SELECT targed.EditTag " \
			"   FROM EE_EditSetEditsPivot targpiv, " \
			"	  EE_EditSets targset, EE_Edits targed " \
			"   WHERE targset.PKey = targpiv.EditSetPKey " \
			"   AND targed.PKey=targpiv.EditPKey " \
			"   AND targset.EditSetTag = srcset.EditSetTag " \
			"   AND targed.EditTag = srced.EditTag " \
			") " \
			"AND NOT EXISTS ( " \
			"   SELECT targed.EditName " \
			"   FROM EE_EditSetEditsPivot targpiv, " \
			"	  EE_EditSets targset, EE_Edits targed " \
			"   WHERE targset.PKey = targpiv.EditSetPKey " \
			"   AND targed.PKey=targpiv.EditPKey " \
			"   AND targset.EditSetTag = srcset.EditSetTag " \
			"   AND targed.EditName = srced.EditName " \
			");",
			src_editset_pkey);
	}
	else
	{
		// Matching Source and Target on EditSetName
		t_count = std::snprintf(t_sql, 1000,
			"SELECT srced.EditTag, srced.EditName " \
			"FROM SourceDB.EE_EditSetEditsPivot srcpiv, " \
			"  SourceDB.EE_EditSets srcset, SourceDB.EE_Edits srced " \
			"WHERE srcset.PKey = srcpiv.EditSetPKey " \
			"AND srced.PKey=srcpiv.EditPKey " \
			"AND srcpiv.EditSetPKey = %d " \
			"AND NOT EXISTS ( " \
			"   SELECT targed.EditTag " \
			"   FROM EE_EditSetEditsPivot targpiv, " \
			"	  EE_EditSets targset, EE_Edits targed " \
			"   WHERE targset.PKey = targpiv.EditSetPKey " \
			"   AND targed.PKey=targpiv.EditPKey " \
			"   AND targset.EditSetName = srcset.EditSetName " \
			"   AND targed.EditTag = srced.EditTag " \
			") " \
			"AND NOT EXISTS ( " \
			"   SELECT targed.EditName " \
			"   FROM EE_EditSetEditsPivot targpiv, " \
			"	  EE_EditSets targset, EE_Edits targed " \
			"   WHERE targset.PKey = targpiv.EditSetPKey " \
			"   AND targed.PKey=targpiv.EditPKey " \
			"   AND targset.EditSetName = srcset.EditSetName " \
			"   AND targed.EditName = srced.EditName " \
			");",
			src_editset_pkey);
	}

	if (t_count > 0)
	{
		t_qry->SetSQLText(t_sql);
		t_qry->SetIsActive(true);

		if (t_qry->RowCount() > 0)
		{
			// Okay, the edit lists do NOT match. Look no farther.
			t_result = false;
		}
		else
		{
			// However, if the row count is 0, there still exists the
			// possibility that the source edit set REMOVED an edit that
			// had previously existed in the the target edit set.
			t_qry->SetIsActive(false);

			if (match_on_tag)
			{
				t_count = std::snprintf(t_sql, 1000,
					"SELECT targed.EditTag, targed.EditName " \
					"FROM EE_EditSetEditsPivot targpiv, " \
					"  EE_EditSets targset, EE_Edits targed " \
					"WHERE targset.PKey = targpiv.EditSetPKey " \
					"AND targed.PKey=targpiv.EditPKey " \
					"AND targpiv.EditSetPKey = %d " \
					"AND NOT EXISTS ( " \
					"   SELECT srced.EditTag " \
					"   FROM SourceDB.EE_EditSetEditsPivot srcpiv, " \
					"	  SourceDB.EE_EditSets srcset, SourceDB.EE_Edits srced " \
					"   WHERE srcset.PKey = srcpiv.EditSetPKey " \
					"   AND srced.PKey=srcpiv.EditPKey " \
					"   AND srcset.EditSetTag = targset.EditSetTag " \
					"   AND srced.EditTag = targed.EditTag " \
					") " \
					"AND NOT EXISTS ( " \
					"   SELECT srced.EditName " \
					"   FROM SourceDB.EE_EditSetEditsPivot srcpiv, " \
					"	  SourceDB.EE_EditSets srcset, SourceDB.EE_Edits srced " \
					"   WHERE srcset.PKey = srcpiv.EditSetPKey " \
					"   AND srced.PKey=srcpiv.EditPKey " \
					"   AND srcset.EditSetTag = targset.EditSetTag " \
					"   AND srced.EditName = targed.EditName " \
					");",
					targ_editset_pkey);
			}
			else
			{
				t_count = std::snprintf(t_sql, 1000,
					"SELECT targed.EditTag, targed.EditName " \
					"FROM EE_EditSetEditsPivot targpiv, " \
					"  EE_EditSets targset, EE_Edits targed " \
					"WHERE targset.PKey = targpiv.EditSetPKey " \
					"AND targed.PKey=targpiv.EditPKey " \
					"AND targpiv.EditSetPKey = %d " \
					"AND NOT EXISTS ( " \
					"   SELECT srced.EditTag " \
					"   FROM SourceDB.EE_EditSetEditsPivot srcpiv, " \
					"	  SourceDB.EE_EditSets srcset, SourceDB.EE_Edits srced " \
					"   WHERE srcset.PKey = srcpiv.EditSetPKey " \
					"   AND srced.PKey=srcpiv.EditPKey " \
					"   AND srcset.EditSetName = targset.EditSetName " \
					"   AND srced.EditTag = targed.EditTag " \
					") " \
					"AND NOT EXISTS ( " \
					"   SELECT srced.EditName " \
					"   FROM SourceDB.EE_EditSetEditsPivot srcpiv, " \
					"	  SourceDB.EE_EditSets srcset, SourceDB.EE_Edits srced " \
					"   WHERE srcset.PKey = srcpiv.EditSetPKey " \
					"   AND srced.PKey=srcpiv.EditPKey " \
					"   AND srcset.EditSetName = targset.EditSetName " \
					"   AND srced.EditName = targed.EditName " \
					");",
					targ_editset_pkey);
			} // if (match_on_tag)

			if (t_count > 0)
			{
				t_qry->SetSQLText(t_sql);
				t_qry->SetIsActive(true);

				if (t_qry->RowCount() > 0)
				{
					// Edits exist in the target edit set that are gone from
					// the source edit set; this is an import candidate
					t_result = false;
				}

				t_qry->SetIsActive(false);
			}
		}
	}
	return t_result;
} // editset_edits_match
//---------------------------------------------------------------------------
int TImportMetafile::get_fields()
{
	int t_result = 0;

	std::string t_table_name = "EE_Fields";

	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	// 3 = New, but conflict on LocalName/FieldNumber exists
	// 4 = Changed, but conflict on LocalName/FieldNumber exists
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, FieldName
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey, src.FieldName || ' [# ' || src.FieldNumber || ']' as RecordKey, ";
	t_sql += "sa.AdminCode ";
	t_sql += "FROM SourceDB.EE_Fields src, SourceDB.EE_Agencies sa ";
	t_sql += "WHERE src.AgencyPKey = sa.PKey ";
	t_sql += "AND NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_Fields targ ";
	t_sql += "WHERE targ.FieldNumber = src.FieldNumber ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones
		t_status = 2; // Changed

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, " \
			"src.FieldName || ' [# ' || src.FieldNumber || ']' as RecordKey, sa.AdminCode " \
			"FROM SourceDB.EE_Fields src, SourceDB.EE_Agencies sa, " \
			" EE_Fields targ, EE_Agencies ta " \
			"WHERE sa.PKey = src.AgencyPKey " \
			"AND ta.PKey = targ.AgencyPKey " \
			"AND src.FieldNumber = targ.FieldNumber " \
			"AND ( " \
			"	src.FieldName != targ.FieldName " \
			" OR sa.AdminCode != ta.AdminCode " \
			" OR src.FieldLength != targ.FieldLength " \
			" OR src.DataType != targ.DataType " \
			" OR src.LocalName != targ.LocalName " \
			" OR src.Comments != targ.Comments " \
			");");

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Finally, identify the objects where a conflict exists on
		// FieldName/FieldNumber OR LocalName/FieldNumber
		// OR the FieldLength has been changed.
		t_status = 3; // Conflict resolution required

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, " \
			"src.FieldName || ' [# ' || src.FieldNumber || ']' as RecordKey, sa.AdminCode " \
			"FROM SourceDB.EE_Fields src, EE_Fields targ, SourceDB.EE_Agencies sa " \
			"WHERE (src.AgencyPKey = sa.PKey) " \
			"AND (src.FieldName = targ.FieldName " \
			"AND src.FieldNumber != targ.FieldNumber) OR " \
			"(src.FieldNumber = targ.FieldNumber " \
			"AND src.FieldName != targ.FieldName) OR " \
			"(src.FieldNumber = targ.FieldNumber " \
			"AND src.FieldLength != targ.FieldLength);");
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, " \
			"src.FieldName || ' [# ' || src.FieldNumber || ']' as RecordKey, sa.AdminCode " \
			"FROM SourceDB.EE_Fields src, EE_Fields targ, SourceDB.EE_Agencies sa  " \
			"WHERE src.AgencyPKey = sa.PKey " \
			"AND src.LocalName = targ.LocalName " \
			"AND src.FieldNumber != targ.FieldNumber;");
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Fields from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_fields", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_fields", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_fields", t_error);

		t_result = -1;
	}
	return t_result;
} // get_fields
//---------------------------------------------------------------------------
int TImportMetafile::get_layouts(const char* match_on_field_name)
{
	int t_result = 0;

	std::string t_table_name = "EE_Layouts";
	bool t_match_tag = (posix_stricmp(match_on_field_name, "LayoutTag") == 0);

	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	// 3 = New, but conflict on Name/Tag exists
	// 4 = Changed, but conflict on Name/Tag exists
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, LayoutName
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey as SrcPKey, src.LayoutName, sa.AdminCode ";
	t_sql += "FROM SourceDB.EE_Layouts src, SourceDB.EE_Agencies sa ";
	t_sql += "WHERE src.AgencyPKey = sa.PKey ";
	t_sql += "AND NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_Layouts targ ";
	if (t_match_tag)
		t_sql += "WHERE targ.LayoutTag = src.LayoutTag ); ";
	else
		t_sql += "WHERE targ.LayoutName = src.LayoutName ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones
		t_status = 2;

		if (t_match_tag)
		{
			t_sql = std::string(
				"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.LayoutName, sa.AdminCode " \
				"FROM SourceDB.EE_Layouts src, SourceDB.EE_Agencies sa, " \
				"EE_Layouts targ, EE_Agencies ta " \
				"WHERE sa.PKey = src.AgencyPKey " \
				"AND ta.PKey = targ.AgencyPKey " \
				"AND src.LayoutTag = targ.LayoutTag " \
				"AND ( " \
				"   src.LayoutName != targ.LayoutName " \
				" OR sa.AdminCode != ta.AdminCode " \
				" OR src.Description != targ.Description);");
		}
		else
		{
			t_sql = std::string(
				"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.LayoutName, sa.AdminCode " \
				"FROM SourceDB.EE_Layouts src, SourceDB.EE_Agencies sa, " \
				"EE_Layouts targ, EE_Agencies ta " \
				"WHERE sa.PKey = src.AgencyPKey " \
				"AND ta.PKey = targ.AgencyPKey " \
				"AND src.LayoutName = targ.LayoutName " \
				"AND ( " \
				"   src.LayoutTag != targ.LayoutTag " \
				" OR sa.AdminCode != ta.AdminCode " \
				" OR src.Description != targ.Description);");
		}

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_record_key  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Finally, identify the objects where a conflict exists on
		// LayoutName/LayoutTag
		t_status = 3; // Conflict resolution required
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.LayoutName, sa.AdminCode " \
			"FROM SourceDB.EE_Layouts src, EE_Layouts targ, SourceDB.EE_Agencies sa " \
			"WHERE (src.AgencyPKey = sa.PKey) " \
			"AND (src.LayoutTag = targ.LayoutTag " \
			"AND src.LayoutName != targ.LayoutName) OR " \
			"(src.LayoutName = targ.LayoutName " \
			"AND src.LayoutTag != targ.LayoutTag);");
		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// No, THIS is "finally": If a user imported the layout without
		// dependencies, the foregoing tests would not find a discrepancy.
		// So for those where the LayoutName and LayoutTag DO match,
		// drill down into the Fields to see if a discrepancy exists.

		t_status = 2; // These will come up as "Changed"
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.LayoutName, sa.AdminCode " \
			"FROM SourceDB.EE_Layouts src, EE_Layouts targ, SourceDB.EE_Agencies sa " \
			"WHERE src.AgencyPKey = sa.PKey " \
			"AND src.LayoutTag = targ.LayoutTag " \
			"AND src.LayoutName = targ.LayoutName;");

		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			if (!layout_fields_match(t_source_pkey, t_target_pkey, t_match_tag))
			{
				(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
					t_source_pkey, t_target_pkey, t_record_key.c_str(),
					t_admin_code.c_str(), t_status);
			}

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);
	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Layouts from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_layouts", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_layouts", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_layouts", t_error);

		t_result = -1;
	}


	return t_result;
} // get_layouts
//---------------------------------------------------------------------------
bool TImportMetafile::layout_fields_match(int src_layout_pkey,
	int targ_layout_pkey, bool match_on_tag)
{
	// Called by get_layouts:
	// Compare the Fields lists of the Layout in the Source and Target
	// databases. If they disagree on existence, StartPos or FieldLength,
	// the match is false.
	bool t_result = true;
	int t_count = 0;
	char t_sql[1001];
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	if (match_on_tag)
	{
		// Matching Source and Target on LayoutTag
		t_count = std::snprintf(t_sql, 1000,
			"SELECT srcfld.FieldNumber " \
			"FROM SourceDB.EE_LayoutFieldsPivot srcpiv, " \
			"  SourceDB.EE_Layouts srclo, SourceDB.EE_Fields srcfld " \
			"WHERE srclo.PKey = srcpiv.LayoutPKey " \
			"AND srcfld.PKey=srcpiv.FieldPKey " \
			"AND srcpiv.LayoutPKey = %d " \
			"AND NOT EXISTS ( " \
			"   SELECT targfld.FieldNumber " \
			"   FROM EE_LayoutFieldsPivot targpiv, " \
			"	  EE_Layouts targlo, EE_Fields targfld " \
			"   WHERE targlo.PKey = targpiv.LayoutPKey " \
			"   AND targfld.PKey=targpiv.FieldPKey " \

			"   AND targlo.LayoutTag = srclo.LayoutTag " \

			"   AND targfld.FieldNumber = srcfld.FieldNumber " \
			"   AND targpiv.StartPos = srcpiv.StartPos " \
			"   AND targfld.FieldLength = srcfld.FieldLength " \
			");",
			src_layout_pkey);
	}
	else
	{
		// Matching Source and Target on LayoutName
		t_count = std::snprintf(t_sql, 1000,
			"SELECT srcfld.FieldNumber " \
			"FROM SourceDB.EE_LayoutFieldsPivot srcpiv, " \
			"  SourceDB.EE_Layouts srclo, SourceDB.EE_Fields srcfld " \
			"WHERE srclo.PKey = srcpiv.LayoutPKey " \
			"AND srcfld.PKey=srcpiv.FieldPKey " \
			"AND srcpiv.LayoutPKey = %d " \
			"AND NOT EXISTS ( " \
			"   SELECT targfld.FieldNumber " \
			"   FROM EE_LayoutFieldsPivot targpiv, " \
			"	  EE_Layouts targlo, EE_Fields targfld " \
			"   WHERE targlo.PKey = targpiv.LayoutPKey " \
			"   AND targfld.PKey=targpiv.FieldPKey " \

			"   AND targlo.LayoutName = srclo.LayoutName " \

			"   AND targfld.FieldNumber = srcfld.FieldNumber " \
			"   AND targpiv.StartPos = srcpiv.StartPos " \
			"   AND targfld.FieldLength = srcfld.FieldLength " \
			");",
			src_layout_pkey);
	}

	if (t_count > 0)
	{
		t_qry->SetSQLText(t_sql);
		t_qry->SetIsActive(true);

		if (t_qry->RowCount() > 0)
		{
			// Okay, the field lists do NOT match. Look no farther.
			t_result = false;
		}
		else
		{
			// However, if the row count is 0, there still exists the
			// possibility that a field that existed in the target layout
			// has been removed in the source layout. (And yes, this would
			// require the user to be meticulous about not messing up the
			// StartPos of subsequent fields, but that could happen.)

			t_qry->SetIsActive(false);
			if (match_on_tag)
			{
				t_count = std::snprintf(t_sql, 1000,
					"SELECT targfld.FieldNumber " \
					"FROM EE_LayoutFieldsPivot targpiv, " \
					"  EE_Layouts targlo, EE_Fields targfld " \
					"WHERE targlo.PKey = targpiv.LayoutPKey " \
					"AND targfld.PKey=targpiv.FieldPKey " \
					"AND targpiv.LayoutPKey = %d " \
					"AND NOT EXISTS ( " \
					"   SELECT srcfld.FieldNumber " \
					"   FROM SourceDB.EE_LayoutFieldsPivot srcpiv, " \
					"	  SourceDB.EE_Layouts srclo, SourceDB.EE_Fields srcfld " \
					"   WHERE srclo.PKey = srcpiv.LayoutPKey " \
					"   AND srcfld.PKey=srcpiv.FieldPKey " \

					"   AND srclo.LayoutTag = targlo.LayoutTag " \

					"   AND srcfld.FieldNumber = targfld.FieldNumber " \
					");",
					targ_layout_pkey);
			}
			else
			{
				t_count = std::snprintf(t_sql, 1000,
					"SELECT targfld.FieldNumber " \
					"FROM EE_LayoutFieldsPivot targpiv, " \
					"  EE_Layouts targlo, EE_Fields targfld " \
					"WHERE targlo.PKey = targpiv.LayoutPKey " \
					"AND targfld.PKey=targpiv.FieldPKey " \
					"AND targpiv.LayoutPKey = %d " \
					"AND NOT EXISTS ( " \
					"   SELECT srcfld.FieldNumber " \
					"   FROM SourceDB.EE_LayoutFieldsPivot srcpiv, " \
					"	  SourceDB.EE_Layouts srclo, SourceDB.EE_Fields srcfld " \
					"   WHERE srclo.PKey = srcpiv.LayoutPKey " \
					"   AND srcfld.PKey=srcpiv.FieldPKey " \

					"   AND srclo.LayoutName = targlo.LayoutName " \

					"   AND srcfld.FieldNumber = targfld.FieldNumber " \
					");",
					targ_layout_pkey);
			} // if (match_on_tag)

			if (t_count > 0)
			{
				t_qry->SetSQLText(t_sql);
				t_qry->SetIsActive(true);

				if (t_qry->RowCount() > 0)
				{
					// Fields exist in the target layout that are gone from
					// the source layout; this is an import candidate
					t_result = false;
				}

				t_qry->SetIsActive(false);
			}
		}

	}

	return t_result;
} // layout_fields_match
//---------------------------------------------------------------------------

int TImportMetafile::get_messages()
{
	int t_result = 0;

	std::string t_table_name = "EE_Messages";

	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, MessageNumber: Message
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey, src.MessageNumber || ': ' || src.Message as TheMessage, sa.AdminCode ";
	t_sql += "FROM SourceDB.EE_Messages src, SourceDB.EE_Agencies sa ";
	t_sql += "WHERE src.AgencyPKey = sa.PKey ";
	t_sql += "AND NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_Messages targ ";
	t_sql += "WHERE targ.MessageNumber = src.MessageNumber ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones
		t_status = 2;

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, " \
			"src.MessageNumber || ': ' || src.Message as TheMessage, sa.AdminCode " \
			"FROM SourceDB.EE_Messages src, SourceDB.EE_Agencies sa, " \
			"EE_Messages targ, EE_Agencies ta " \
			"WHERE sa.PKey = src.AgencyPKey " \
			"AND ta.PKey = targ.AgencyPKey " \
			"AND src.MessageNumber = targ.MessageNumber " \
			"AND ( " \
			"    sa.AdminCode != ta.AdminCode " \
			" OR src.Message != targ.Message);");

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Finally, identify the objects where a conflict exists on
		// Message text/Message Number
		t_status = 3;  // Conflict resolution required
		/* Let's make it a conflict issue ONLY if the same
		 * message text occurs for more than one message number
		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, " \
			"src.MessageNumber || ': ' || src.Message as TheMessage " \
			"FROM SourceDB.EE_Messages src, EE_Messages targ " \
			"WHERE (src.Message = targ.Message " \
			"AND src.MessageNumber != targ.MessageNumber) OR " \
			"(src.MessageNumber = targ.MessageNumber " \
			"AND src.Message != targ.Message);");
		*/

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, " \
			"src.MessageNumber || ': ' || src.Message as TheMessage, sa.AdminCode " \
			"FROM SourceDB.EE_Messages src, EE_Messages targ, SourceDB.EE_Agencies sa " \
			"WHERE src.AgencyPKey = sa.PKey " \
			"AND (src.Message = targ.Message " \
			"AND src.MessageNumber != targ.MessageNumber);");

		t_qry->SetSQLText(t_sql.c_str());
		t_qry->SetIsActive(true);

		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Messages from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_messages", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_messages", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_messages", t_error);

		t_result = -1;
	}

	return t_result;
} // get_messages
//---------------------------------------------------------------------------
int TImportMetafile::get_tables()
{
	int t_result = 0;

	std::string t_table_name = "EE_LookupTables";

	// status parameter to callback function will be one of:
	// 1 = New
	// 2 = Changed
	int t_status = 1;

	int t_source_pkey,
		t_target_pkey;

	std::string t_record_key; // in this case, TableName
	std::string t_admin_code;

	// First, find the new ones
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::string t_sql = "SELECT src.PKey as SrcPKey, src.TableName, sa.AdminCode ";
	t_sql += "FROM SourceDB.EE_LookupTables src, SourceDB.EE_Agencies sa ";
	t_sql += "WHERE src.AgencyPKey = sa.PKey ";
	t_sql += "AND NOT EXISTS ( ";
	t_sql += "SELECT targ.PKey FROM EE_LookupTables targ ";
	t_sql += "WHERE targ.TableName = src.TableName ); ";

	try
	{
		t_target_pkey = -1; // New ones have no match in the target metafile
		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_record_key  = t_qry->GetFieldByNumber(1);
			t_admin_code  = t_qry->GetFieldByNumber(2);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

		// Now find the changed ones.
		// NOTE: For LookupTables, we must always consider the LastModDate,
		// because it may be the only way to know that a row in the table
		// was modified, even if the RecCount didn't change.
		t_status = 2;

		t_sql = std::string(
			"SELECT src.PKey as SrcPKey, targ.PKey as TargPKey, src.TableName, sa.AdminCode " \
			"FROM SourceDB.EE_LookupTables src, SourceDB.EE_Agencies sa, " \
			"EE_LookupTables targ, EE_Agencies ta " \
			"WHERE sa.PKey = src.AgencyPKey " \
			"AND ta.PKey = targ.AgencyPKey " \
			"AND src.TableName = targ.TableName " \
			"AND ( " \
			"    sa.AdminCode != ta.AdminCode " \
			" OR src.RecCount != targ.RecCount " \
			" OR src.LastModDate != targ.LastModDate " \
			" OR src.Comments != targ.Comments);");

		t_qry->SetSQLText(t_sql.c_str());

		t_qry->SetIsActive(true);
		while (!t_qry->Eof())
		{
			t_source_pkey = t_qry->GetIntFieldByNumber(0);
			t_target_pkey = t_qry->GetIntFieldByNumber(1);
			t_record_key  = t_qry->GetFieldByNumber(2);
			t_admin_code  = t_qry->GetFieldByNumber(3);

			(*FCallbackFunction)(FCallbackOwner, t_table_name.c_str(),
				t_source_pkey, t_target_pkey, t_record_key.c_str(),
				t_admin_code.c_str(), t_status);

			t_qry->FetchNext();
		}

		t_qry->SetIsActive(false);

	}
	catch (std::runtime_error& e)
	{
		std::string t_error = "Cannot get Lookup Tables from source metafile ";
		FEngineErrorsPump->AddError("TImportMetafile", "get_tables", t_error);
		std::string t_what = e.what();
		if (!t_what.empty())
		{
			FEngineErrorsPump->AddError("TImportMetafile", "get_tables", t_what);
		}
		t_error = "SQL: ";
		t_error += t_qry->GetSQLText();
		FEngineErrorsPump->AddError("TImportMetafile", "get_tables", t_error);

		t_result = -1;
	}

	return t_result;
} // get_tables
//---------------------------------------------------------------------------

int TImportMetafile::ImportLookupTable(const std::string table_name)
{
	int t_result = 0;

	// NOTE: There is no "rollback" option when running data definition commands,
	// so if this fails midstream, we could end up with a mess in the destination
	// database.

	if (!FAttachedDBAlias.empty())
	{
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		std::string t_source_sql = "SELECT m.sql FROM ";
		t_source_sql += FAttachedDBAlias;
		t_source_sql += ".sqlite_master m WHERE m.tbl_name = '";
		t_source_sql += table_name;
		t_source_sql += "'";

		// Get the CREATE TABLE definition
		std::string t_sql = t_source_sql;
		t_sql += " AND m.type LIKE 'table%';";

		std::unique_ptr<TSQLQuery> t_create_qry(new TSQLQuery(FSQLDB));
		std::string t_create_table;
		std::string t_create_index;
		std::string t_create_trigger;
		try
		{
			t_qry->SetSQLText(t_sql.c_str());

			t_qry->SetIsActive(true);
			if (!t_qry->Eof())
			{
				t_create_table = t_qry->GetFieldByName("sql");

				t_create_qry->SetSQLText(t_create_table.c_str());
				t_create_qry->ExecSQL();
			}
			t_qry->SetIsActive(false);

			// Get index defs, if any exist
			t_sql = t_source_sql;
			t_sql += " AND m.type LIKE 'index%';";

			t_qry->SetSQLText(t_sql.c_str());

			t_qry->SetIsActive(true);
			while (!t_qry->Eof())
			{
				t_create_index = t_qry->GetFieldByName("sql");

				t_create_qry->SetSQLText(t_create_index.c_str());
				t_create_qry->ExecSQL();

				t_qry->FetchNext();
			}
			t_qry->SetIsActive(false);
/*
			// Now get trigger defs, if any exist
			t_sql = t_source_sql;
			t_sql += " AND m.type LIKE 'trigger%';";

			t_qry->SetSQLText(t_sql.c_str());

			t_qry->SetIsActive(true);
			while (!t_qry->Eof())
			{
				t_create_trigger = t_qry->GetFieldByName("sql");

				t_create_qry->SetSQLText(t_create_trigger.c_str());
				t_create_qry->ExecSQL();

				t_qry->FetchNext();
			}
			t_qry->SetIsActive(false);
*/
			// And finally, import the data
			t_sql  = "INSERT INTO ";
			t_sql += table_name;
			t_sql += " SELECT * FROM ";
			t_sql += FAttachedDBAlias;
			t_sql += ".";
			t_sql += table_name;
			t_sql += ";";

			t_qry->SetSQLText(t_sql.c_str());
			t_qry->ExecSQL();

			// Now get trigger defs, if any exist
			t_sql = t_source_sql;
			t_sql += " AND m.type LIKE 'trigger%';";

			t_qry->SetSQLText(t_sql.c_str());

			t_qry->SetIsActive(true);
			if (!t_qry->Eof())
			{
				int t_row_count = t_qry->RowCount();

				while (!t_qry->Eof())
				{
					t_create_trigger = t_qry->GetFieldByName("sql");

					t_create_qry->SetSQLText(t_create_trigger.c_str());
					t_create_qry->ExecSQL();

					t_qry->FetchNext();
				}

				t_qry->SetIsActive(false);

				if (t_row_count <= 4)
				{
					// Now update all trigger-maintained fields
					t_sql  = "UPDATE ";
					t_sql += table_name;
					t_sql += " SET PKey = PKey;";

					t_qry->SetSQLText(t_sql.c_str());
					t_qry->ExecSQL();
				}
				else
				{
					// If more than 4 triggers are defined on a table, the
					// one-swell-foop approach to updating bogs down horribly.
					// We have only one insane look-up table: CS_OBS with a
					// whopping 12 triggers (6 update, 6 insert). So for the
					// many edits that are associated with this table, we
					// have to hit each row individually.
// FIX ME: This is impossible! Have to rethink the whole stupid thing
					t_sql  = "SELECT PKey FROM ";
					t_sql += table_name;

					t_qry->SetSQLText(t_sql.c_str());
					t_qry->SetIsActive(true);
					if (!t_qry->Eof())
					{
						std::string t_savepoint_name = "_ImportLookupTable_";
						if (FSQLDB->SavePoint(t_savepoint_name.c_str()))
						{
							try
							{
								std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));

								char t_buffer[100];
								int t_pkey;
								while (!t_qry->Eof())
								{
									t_pkey = t_qry->GetIntFieldByName("PKey");
									std::snprintf(t_buffer, 100,
										"UPDATE %s SET PKey=PKey WHERE PKey=%d;",
										table_name.c_str(), t_pkey);
									t_update_qry->SetSQLText(t_buffer);
									t_update_qry->ExecSQL();

									t_qry->FetchNext();

								}
							}
							catch (...)
							{
								// Remember: rolling back to a savepoint still
								// requires releasing the savepoint.
								FSQLDB->RollbackTo(t_savepoint_name.c_str());
							}

							FSQLDB->Release(t_savepoint_name.c_str());
						}
					}

					t_qry->SetIsActive(false);
				}
			}

			t_qry->SetIsActive(false);

		}
		catch (std::runtime_error& e)
		{
			std::string t_error = "Cannot import lookup table \"";
			t_error += table_name;
			t_error += "\"";
			FEngineErrorsPump->AddError("TImportMetafile", "ImportLookupTable", t_error);
			std::string t_what = e.what();
			if (!t_what.empty())
			{
				FEngineErrorsPump->AddError("TImportMetafile", "ImportLookupTable", t_what);
			}
			t_error = "SQL: " + t_sql;
			FEngineErrorsPump->AddError("TImportMetafile", "ImportLookupTable", t_error);

			t_result = -1;
		}
	}
	else
	{
		t_result = -1;
		std::string t_error = "No attached metafile (did you call attach_to_metafile()?)";
		FEngineErrorsPump->AddError("TImportMetafile", "ImportLookupTable", t_error);
	}
	return t_result;
} // ImportLookupTable
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

// post_edit calls this to clean up any objects it created
void DeleteUpdateEditLogicObject(TUpdateEditLogicObject* obj)
{
	if (obj)
	{
		delete obj;
		obj = NULL;
	}
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

TSMFWriter::TSMFWriter(TSMF* parent, TSQLDB* database, TEngineErrorsPump* pump,
	TEw_AccessController* access_controller, TTablesList* tables_list)
{
	FParent = parent;
	FSQLDB = database;
	FEngineErrorsPump = pump;
	FAccessController = access_controller;
	FTablesList = tables_list;

	// for lookup table maintenance
	// we create only one of these and re-use it if needed by EditWriter's
	// Table maintenance form
	FDefineTable = new TDefineTable();

} // constructor
//---------------------------------------------------------------------------
TSMFWriter::~TSMFWriter()
{
	if (FDefineTable)
	{
		delete FDefineTable;
	}
} // destructor
//---------------------------------------------------------------------------
bool TSMFWriter::have_access_rights(TEE_MCO mco, const int pkey)
{
	bool t_result = true; // until shown otherwise
	int t_agency_pkey = 0;
	char t_buffer[500];
	t_buffer[0] = '\x0';

	switch (mco)
	{
		case _mco_Agencies:
			t_agency_pkey = pkey;
			t_result = FAccessController->IsLoggedIn(t_agency_pkey);
			break;

		case _mco_Edits:
			std::snprintf(t_buffer, 500,
				"SELECT [AgencyPKey] FROM [EE_Edits] WHERE [PKey] = %d;", pkey);
			break;

		case _mco_EditSets:
			std::snprintf(t_buffer, 500,
				"SELECT [AgencyPKey] FROM [EE_EditSets] WHERE [PKey] = %d;", pkey);
			break;

		case _mco_Fields:
			std::snprintf(t_buffer, 500,
				"SELECT [AgencyPKey] FROM [EE_Fields] WHERE [PKey] = %d;", pkey);
			break;

		case _mco_Layouts:
			std::snprintf(t_buffer, 500,
				"SELECT [AgencyPKey] FROM [EE_Layouts] WHERE [PKey] = %d;", pkey);
			break;

		case _mco_Messages:
			std::snprintf(t_buffer, 500,
				"SELECT [AgencyPKey] FROM [EE_Messages] WHERE [PKey] = %d;", pkey);
			break;

		case _mco_LookupTables:
			std::snprintf(t_buffer, 500,
				"SELECT [AgencyPKey] FROM [EE_LookupTables] WHERE [PKey] = %d;", pkey);
			break;

		case _mco_LayoutFieldsPivot:
			std::snprintf(t_buffer, 500,
				"SELECT b.[AgencyPKey] " \
				"FROM [EE_LayoutFieldsPivot] a, [EE_Layouts] b " \
				"WHERE a.[LayoutPKey] = b.[PKey] " \
				"AND a.[PKey] = %d;", pkey);
			break;

		case _mco_EditSetEditsPivot:
			std::snprintf(t_buffer, 500,
				"SELECT b.[AgencyPKey] " \
				"FROM [EE_EditSetEditsPivot] a, [EE_EditSets] b " \
				"WHERE a.[EditSetPKey] = b.[PKey] " \
				"AND a.[PKey] = %d;", pkey);
			break;

		case _mco_EditFieldsPivot:
			std::snprintf(t_buffer, 500,
				"SELECT b.[AgencyPKey] " \
				"FROM [EE_EditFieldsPivot] a, [EE_Edits] b " \
				"WHERE a.[EditPKey] = b.[PKey] " \
				"AND a.[PKey] = %d;", pkey);
			break;

		case _mco_EditTablesPivot:
			std::snprintf(t_buffer, 500,
				"SELECT b.[AgencyPKey] " \
				"FROM [EE_EditTablesPivot] a, [EE_Edits] b " \
				"WHERE a.[EditPKey] = b.[PKey] " \
				"AND a.[PKey] = %d;", pkey);
			break;

		case _mco_EditMessagesPivot:
			std::snprintf(t_buffer, 500,
				"SELECT b.[AgencyPKey] " \
				"FROM [EE_EditMessagesPivot] a, [EE_Edits] b " \
				"WHERE a.[EditPKey] = b.[PKey] " \
				"AND a.[PKey] = %d;", pkey);
			break;

		case _mco_EditErrorsPivot:
			std::snprintf(t_buffer, 500,
				"SELECT b.[AgencyPKey] " \
				"FROM [EE_EditErrorsPivot] a, [EE_Edits] b " \
				"WHERE a.[EditPKey] = b.[PKey] " \
				"AND a.[PKey] = %d;", pkey);
			break;

		default:
			// _mco_Other and _mco_Control don't require login access
			break;
	}

	if (std::strlen(t_buffer) > 0)
	{
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

		t_qry->SetSQLText(t_buffer);

		t_qry->SetIsActive(true);

		t_agency_pkey = t_qry->GetIntFieldByNumber(0);

		t_result = FAccessController->IsLoggedIn(t_agency_pkey);
	}

	return t_result;
} // have_access_rights
//---------------------------------------------------------------------------
int TSMFWriter::Edit(const std::string table_name, const int pkey)
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if ((t_table))
	{
		if (have_access_rights(t_table->GetMCO(), pkey))
		{
			t_table->Edit(pkey);

			TEw_DataSetState t_state = t_table->GetState();
			if (t_state != ewEdit)
			{
				std::string t_error = "Table named \"";
				t_error += table_name;
				t_error += "\" is NOT in Edit mode!";
				FEngineErrorsPump->AddError("TSMFWriter", "Edit", t_error);
				t_result = -1;
			}
		}
		else
		{
			std::string t_error = "You do not have access rights to modify this object";
			FEngineErrorsPump->AddError("TSMFWriter","Edit",t_error);
			t_result = -1;
		}
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMFWriter","Edit",t_error);
		t_result = -1;
	}

	return t_result;
} // Edit
//---------------------------------------------------------------------------
int TSMFWriter::Insert(const std::string table_name)
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		t_table->Insert();

		TEw_DataSetState t_state = t_table->GetState();
		if (t_state != ewInsert)
		{
			std::string t_error = "Table named \"";
			t_error += table_name;
			t_error += "\" is NOT in Insert mode!";
			FEngineErrorsPump->AddError("TSMFWriter", "Insert", t_error);
			t_result = -1;
		}

	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMFWriter","Insert",t_error);
		t_result = -1;
	}

	return t_result;
} // Insert()
//---------------------------------------------------------------------------
bool TSMFWriter::verify_delete_impact(std::string table_name, int delete_pkey)
{
	bool t_ok_to_delete = true;
	char t_buffer[200];

	////////////////
	/// EE_Agencies
	////////////////
	if (posix_stricmp((char*)table_name.c_str(), "EE_Agencies") == 0)
	{
		// Go through all of the MCO tables and change AgencyPKey to 1,
		// the "NONE" Agency.
		if (delete_pkey == 1)
		{
			FEngineErrorsPump->AddError("TSMFWriter","verify_delete_impact",
				"Agency \"NONE\" may not be deleted");
			t_ok_to_delete = false;
		}
		else
		{
			std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
			int t_update_counter = 1;
			try
			{
				while (t_update_counter < 7)
				{
					switch (t_update_counter)
					{
						case 1: // EE_Edits
							std::snprintf(t_buffer, 200,
								"UPDATE EE_Edits SET AgencyPKey = 1 WHERE AgencyPKey = %d;",
								delete_pkey);
							break;
						case 2: // EE_EditSets
							std::snprintf(t_buffer, 200,
								"UPDATE EE_EditSets SET AgencyPKey = 1 WHERE AgencyPKey = %d;",
								delete_pkey);
							break;
						case 3: // EE_Fields
							std::snprintf(t_buffer, 200,
								"UPDATE EE_Fields SET AgencyPKey = 1 WHERE AgencyPKey = %d;",
								delete_pkey);
							break;
						case 4: // EE_Layouts
							std::snprintf(t_buffer, 200,
								"UPDATE EE_Layouts SET AgencyPKey = 1 WHERE AgencyPKey = %d;",
								delete_pkey);
							break;
						case 5: // EE_Messages
							std::snprintf(t_buffer, 200,
								"UPDATE EE_Messages SET AgencyPKey = 1 WHERE AgencyPKey = %d;",
								delete_pkey);
							break;
						case 6: // EE_LookupTables
							std::snprintf(t_buffer, 200,
								"UPDATE EE_LookupTables SET AgencyPKey = 1 WHERE AgencyPKey = %d;",
								delete_pkey);
							break;
					}

					t_qry->SetSQLText(t_buffer);
					t_qry->ExecSQL();

					t_update_counter++;
				}
			}
			catch (...)
			{
				FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
					"Cannot update MCO table to change Agency to \"NONE\"");
				FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
					t_qry->GetSQLText());
				t_ok_to_delete = false;
			}
		}
	} // EE_Agencies

	////////////////
	/// EE_Fields
	////////////////
	else if (posix_stricmp((char*)table_name.c_str(), "EE_Fields") == 0)
	{
		std::unique_ptr<TSQLQuery> t_xref_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));
		int t_edit_pkey;

		// NOTE: Although deleting a Field object that is referenced from
		// a Layout will also remove the reference from that Layout, there
		// is no housekeeping to be performed for that issue. (The FK definition
		// from EE_LayoutFieldsPivot ON CASCADE DELETE will take care of that.)
		//
		// HOWEVER: Deleting a Field object that is referenced from EditLogic
		// means the logic will no longer compile. EditWriter can warn the user
		// about the side effect, but the Engine must ensure that the CompiledLogic
		// is cleared so that nobody attempts to "run edits" with corrupted logic.
		std::snprintf(t_buffer, 200,
			"SELECT EditPKey FROM EE_EditFieldsPivot WHERE FieldPKey = %d;",
			delete_pkey);

		t_xref_qry->SetSQLText(t_buffer);
		t_xref_qry->SetIsActive(true);

		while (!t_xref_qry->Eof())
		{
			t_edit_pkey = t_xref_qry->GetIntFieldByName("EditPKey");
			std::snprintf(t_buffer, 200,
				"UPDATE EE_Edits SET CompiledLogic = '' WHERE PKey = %d;",
				t_edit_pkey);

			try
			{
				t_update_qry->SetSQLText(t_buffer);
				t_update_qry->ExecSQL();
			}
			catch (...)
			{
				FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
					"Cannot clear CompiledLogic for Edit that references deleted field");
				FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
					t_update_qry->GetSQLText());

				t_ok_to_delete = false;
				break;
			}

			t_xref_qry->FetchNext();
		}

		t_xref_qry->SetIsActive(false);

	} // EE_Fields

	////////////////
	/// EE_Messages
	////////////////
	else if (posix_stricmp((char*)table_name.c_str(), "EE_Messages") == 0)
	{
		// Change the default error message (EE_Edits.MessagePKey)
		// to the "This edit has no default error message", MessageNumber=0,
		// EE_Messages.PKey == 1
		if (delete_pkey == 1)
		{
			FEngineErrorsPump->AddError("TSMFWriter","verify_delete_impact",
				"Message Number 0 may not be deleted");
			t_ok_to_delete = false;
		}
		else
		{
			// For all edits with a cross-reference row in EE_EditMessagesPivot,
			// clear the EE_Edits.CompiledLogic. NOTE: Some may become cleared
			// because of the 'D' (default) entry, even if the message was not
			// also referenced in the logic. It's a small price to pay... the
			// user can recompile all/affected edits to reinstate the compiled
			// logic. Note that if logic invokes a function that also references
			// the default error message, the only entry in EE_EditMessagesPivot
			// is the one marked 'D' for default, so we have no other (reasonable)
			// way to find the edits that don't necessarily have to clear their
			// CompiledLogic. Too bad, so sad... that's what you get for deleting
			// MCOs that are in use.
			std::unique_ptr<TSQLQuery> t_xref_qry(new TSQLQuery(FSQLDB));
			std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));
			int t_edit_pkey;

			std::snprintf(t_buffer, 200,
				"UPDATE EE_Edits SET MessagePKey = 1 WHERE MessagePKey = %d;",
				delete_pkey);

			t_update_qry->SetSQLText(t_buffer);
			if (!t_update_qry->ExecSQL())
			{
				FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
					"Cannot reset Default Message for Edit that references deleted message object");
				FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
					t_update_qry->GetSQLText());

				t_ok_to_delete = false;
			}
			else
			{
				std::snprintf(t_buffer, 200,
					"SELECT EditPKey FROM EE_EditMessagesPivot WHERE MessagePKey = %d;",
					delete_pkey);

				t_xref_qry->SetSQLText(t_buffer);
				t_xref_qry->SetIsActive(true);

				while (!t_xref_qry->Eof())
				{
					t_edit_pkey = t_xref_qry->GetIntFieldByName("EditPKey");

					std::snprintf(t_buffer, 200,
						"UPDATE EE_Edits SET CompiledLogic = '' WHERE PKey = %d;",
						t_edit_pkey);

					try
					{
						t_update_qry->SetSQLText(t_buffer);
						t_update_qry->ExecSQL();
					}
					catch (...)
					{
						FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
							"Cannot clear CompiledLogic for Edit that references deleted message object");
						FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
							t_update_qry->GetSQLText());

						t_ok_to_delete = false;
						break;
					}
					t_xref_qry->FetchNext();
				}

				t_xref_qry->SetIsActive(false);

			}
		}

	} // EE_Messages

	////////////////////
	/// EE_LookupTables
	////////////////////
	else if (posix_stricmp((char*)table_name.c_str(), "EE_LookupTables") == 0)
	{
		// This control block ensures that the TSQLQuery pointers are closed
		// and finalized before attempting to call DROP TABLE in the next control
		// block. SQLite will not allow Data Definition Language commands on an
		// object that is still being referenced from a not-finalized query.
		//
		// See this message on stackoverflow for more information:
		// https://stackoverflow.com/questions/21204431/database-is-locked-error-sqlite-ios/21216713#21216713

		{
			std::unique_ptr<TSQLQuery> t_xref_qry(new TSQLQuery(FSQLDB));
			std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));
			int t_edit_pkey;

			// Deleting a LookupTable object that is referenced from EditLogic
			// means the logic will no longer compile. EditWriter can warn the user
			// about the side effect, but the Engine must ensure that the CompiledLogic
			// is cleared so that nobody attempts to "run edits" with corrupted logic.
			std::snprintf(t_buffer, 200,
				"SELECT EditPKey FROM EE_EditTablesPivot WHERE LookupTablePKey = %d;",
				delete_pkey);

			t_xref_qry->SetSQLText(t_buffer);
			t_xref_qry->SetIsActive(true);

			while (!t_xref_qry->Eof())
			{
				t_edit_pkey = t_xref_qry->GetIntFieldByName("EditPKey");
				std::snprintf(t_buffer, 200,
					"UPDATE EE_Edits SET CompiledLogic = '' WHERE PKey = %d;",
					t_edit_pkey);

				try
				{
					t_update_qry->SetSQLText(t_buffer);
					t_update_qry->ExecSQL();
				}
				catch (...)
				{
					FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
						"Cannot clear CompiledLogic for Edit that references deleted lookup table");
					FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
						t_update_qry->GetSQLText());

					t_ok_to_delete = false;
					break;
				}

				t_xref_qry->FetchNext();
			}

			t_xref_qry->SetIsActive(false);
		}

		// The end of the preceding control block ensured the TSQLQuery objects
		// were destroyed (TSQLQuery class calls sqlite3_finalize in destructor).
		{
			std::unique_ptr<TSQLQuery> t_xref_qry(new TSQLQuery(FSQLDB));
			std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));

			// One more little thing... we have to DROP the table from the database
			// Remember: This action cannot be reversed by ROLLBACK TRANS.
			std::snprintf(t_buffer, 200,
				"SELECT TableName FROM EE_LookupTables WHERE PKey = %d;",
				delete_pkey);

			// re-use the query
			t_xref_qry->SetSQLText(t_buffer);
			t_xref_qry->SetIsActive(true);
			if (!t_xref_qry->Eof())
			{
				std::string t_lookup_table = t_xref_qry->GetFieldByName("TableName");
				std::snprintf(t_buffer, 200,
					"DROP TABLE IF EXISTS %s;",
					t_lookup_table.c_str());

				t_xref_qry->SetIsActive(false);

				t_update_qry->SetSQLText(t_buffer);
				if (t_update_qry->ExecSQL())
				{
					TTableDef* t_table_def = FTablesList->GetTable(t_lookup_table);
					if (t_table_def)
					{
						FTablesList->RemoveTable(t_table_def);
					}
				}
				else
				{
					// Yikes! Rollback the other changes
					FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
						"Cannot drop lookup table from metafile database");
					FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
						t_update_qry->LastErrorMessage());
					FEngineErrorsPump->AddError("TSMFWriter", "verify_delete_impact",
						t_update_qry->GetSQLText());

					t_ok_to_delete = false;
				}
			}

			t_xref_qry->SetIsActive(false);
		}
	} // EE_LookupTables

	//////////////////////////////
	/// all others are good to go
	//////////////////////////////
	return t_ok_to_delete;
} // verify_delete_impact
//---------------------------------------------------------------------------
int TSMFWriter::Delete(const std::string table_name, const int delete_pkey)
{
	int t_result = 0;
	std::string t_error;
	char t_buffer[200];

	std::string t_savepoint_name = "_Delete_";
	FSQLDB->SavePoint(t_savepoint_name.c_str());
	try
	{
		// Deleting from EE_Agencies means updating the AgencyPKey in
		// all linked MCOs.
		// Deleting from EE_Fields, EE_Messages or EE_Tables means
		// clearing EE_Edits.CompiledLogic for all refereneced edits.
		// If for any reason these side effects cannot be accommodated,
		// we'll roll back all changes.
		if ( verify_delete_impact(table_name, delete_pkey))
		{
			TTableDef* t_table = FTablesList->GetTable(table_name);
			if (t_table)
			{
				std::snprintf(t_buffer,200,
					"DELETE FROM [%s] WHERE [PKey] = %d;",
					table_name.c_str(),
					delete_pkey);

				std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

				try
				{
					t_qry->SetSQLText(t_buffer);

					t_qry->ExecSQL();
				}
				catch (...)
				{
					t_error = t_qry->LastErrorMessage();
					FEngineErrorsPump->AddError("TSMFWriter","Delete",t_error);
					std::snprintf(t_buffer,200,"SQL: %s", t_buffer);
					// Report the SQL that failed
					t_error.assign(t_buffer);
					FEngineErrorsPump->AddError("TSMFWriter","Delete",t_error);

					t_result = -1;
				}
			}
			else
			{
				t_error = "Cannot find table named \"";
				t_error += table_name;
				t_error += "\"";
				FEngineErrorsPump->AddError("TSMFWriter","Delete",t_error);
				t_result = -1;
			}

			FSQLDB->Release(t_savepoint_name.c_str());
		}
		else
		{
			FSQLDB->RollbackTo(t_savepoint_name.c_str());
			FSQLDB->Release(t_savepoint_name.c_str());
			// verify_delete_impact will add messages to the FEngineErrorsPump
			t_result = -1;
		} // verify_delete_impact
	}
	catch (std::runtime_error& e)
	{
		std::snprintf(t_buffer, 200,
			"Error in attempt to delete PKey %d from table %s: %s",
			delete_pkey, table_name.c_str(), e.what());
		FEngineErrorsPump->AddError("TSMFWriter","Delete",t_buffer);


		FSQLDB->RollbackTrans();
	}
	return t_result;
} // Delete
//---------------------------------------------------------------------------
bool TSMFWriter::is_system_table(TTableDef* table)
{
	bool t_result = false;
	std::string t_table_name = table->GetTableName();

	lowercase(t_table_name);

	if ( (t_table_name.compare("ee_edits") == 0) ||
		 (t_table_name.compare("ee_editsets") == 0) ||
		 (t_table_name.compare("ee_fields") == 0) ||
		 (t_table_name.compare("ee_layouts") == 0) ||
		 (t_table_name.compare("ee_messages") == 0) ||
		 (t_table_name.compare("ee_lookuptables") == 0))
	{
		t_result = true;
	}

	return t_result;
}
//---------------------------------------------------------------------------
int TSMFWriter::set_agency(TTableDef* table, std::string& old_admin_code,
	std::string& new_admin_code)
{
	int t_result = 0;
	// default the output parameters
	old_admin_code = "";
	new_admin_code = "";

	TFieldDef* t_fld = table->GetColumn("AgencyPKey");
	int t_current_agency_pkey,
		t_use_this_agency_pkey;

	// We're checking the state, because an ewInsert would not have any underlying
	// table to reference.
	if (table->GetState() == ewEdit)
	{
		// Have to get this from the underlying table... unless the super-user
		// set it to something specific. In any case, we'll need to know the "old"
		// AdminCode for the caller.
		t_current_agency_pkey = std::atoi(t_fld->GetValue());
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		char t_buffer[200];
		std::snprintf(t_buffer,200,
			"SELECT a.AgencyPKey, b.AdminCode FROM [%s] a, EE_Agencies b WHERE (b.PKey=a.AgencyPKey) AND a.PKey=%d;",
			table->GetTableName(),
			table->GetPostToPKey());

		t_qry->SetSQLText(t_buffer);
		t_qry->SetIsActive(true);
		if (!t_qry->Eof())
		{
			if (t_current_agency_pkey == 0)
			{
				t_current_agency_pkey = t_qry->GetIntFieldByName("AgencyPKey");
			}
			old_admin_code = t_qry->GetFieldByName("AdminCode");
		}
	}
	t_result = FAccessController->GetLoggedInAgencyPKey(t_current_agency_pkey,
		&t_use_this_agency_pkey);

	std::string t_value = intToString(t_use_this_agency_pkey);
	t_fld->SetValue(t_value.c_str());

	// Last thing: Fill in the new agency code so that the caller can use it.
	std::unique_ptr<TSQLQuery> t_after_qry(new TSQLQuery(FSQLDB));
	char t_buffer[200];
	std::snprintf(t_buffer,200,
		"SELECT AdminCode FROM EE_Agencies WHERE PKey=%d;",
		t_use_this_agency_pkey);

	t_after_qry->SetSQLText(t_buffer);
	t_after_qry->SetIsActive(true);
	if (!t_after_qry->Eof())
	{
		new_admin_code = t_after_qry->GetFieldByName("AdminCode");
	}

	return t_result;
} // set_agency
//---------------------------------------------------------------------------
int TSMFWriter::Post(const std::string table_name, int* new_pkey)
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		if (!t_table->IsEditing())
		{
			std::string t_error = "Table named \"";
			t_error += table_name;
			t_error += "\" is NOT in insert or edit mode; cannot post";
			FEngineErrorsPump->AddError("TSMFWriter","Post",t_error);
			return -1;
		}

		if (t_table->GetState() == ewInsert)
		{
			t_result = post_insert(t_table, new_pkey);
		}
		else
		{
			*new_pkey = t_table->GetPostToPKey();
			t_result = post_edit(t_table);
		}

		// Call t_table->Cancel() to take the table out of "editing mode"
		t_table->Cancel();

	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMFWriter","Post",t_error);
		t_result = -1;
	}

	return t_result;
} // Post()
//---------------------------------------------------------------------------
void TSMFWriter::get_original_value(TTableDef* table, TFieldDef* field,
	std::string& original_value)
{
	// called by post_edit
	int t_pkey = table->GetPostToPKey();

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[100];
	std::snprintf(t_buffer, 100,
		"SELECT [%s] FROM [%s] WHERE PKey = %d;",
		field->GetFieldName(), table->GetTableName(), t_pkey);

	t_qry->SetSQLText(t_buffer);
	try
	{
		if (t_qry->SetIsActive(true))
		{
			original_value.assign(t_qry->GetFieldByNumber(0));
			t_qry->SetIsActive(false);
		}
	}
	catch (...)
	{
		original_value.clear();
	}

} // get_original_value
//---------------------------------------------------------------------------
bool TSMFWriter::update_edit_fields(TUpdateEditLogicObject* obj,
	const int field_pkey)
{
	bool t_result = true;

	std::string t_field_to_update = obj->GetFieldName();

	if ((t_field_to_update.compare("FieldName") == 0) ||
		(t_field_to_update.compare("FieldNumber") == 0))
	{
		std::string t_original_value, t_new_value;

		std::unique_ptr<TSQLQuery> t_pivot_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_logic_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));

		std::vector<int> t_edit_pkeys;
		std::vector<std::string> t_edit_tags;

		int t_edit_pkey;
		char t_pivot_buffer[200];
		char t_logic_buffer[200];
		// I can make no assumptions about the length of EditLogic, so we'll
		// build a string
		std::string t_edit_tag, t_edit_logic, t_update_logic;

		std::snprintf(t_pivot_buffer, 200,
			"SELECT a.[EditPKey], b.[EditTag] FROM [EE_EditFieldsPivot] a, [EE_Edits] b " \
			"WHERE a.[EditPKey] = b.[PKey] " \
			"AND a.[FieldPKey] = %d;",
			field_pkey);

		t_pivot_qry->SetSQLText(t_pivot_buffer);
		try
		{
			t_pivot_qry->SetIsActive(true);
			while (!t_pivot_qry->Eof())
			{
				t_edit_pkeys.push_back(t_pivot_qry->GetIntFieldByNumber(0));
				t_edit_tags.push_back(t_pivot_qry->GetFieldByNumber(1));

				t_pivot_qry->FetchNext();
			}

			t_pivot_qry->SetIsActive(false);

			for (size_t t_idx = 0; t_idx < t_edit_pkeys.size(); t_idx++)
			{
				t_edit_pkey = t_edit_pkeys[t_idx];
				t_edit_tag.assign(t_edit_tags[t_idx]);

				if (t_field_to_update.compare("FieldName") == 0)
				{
					std::snprintf(t_logic_buffer, 200,
						"SELECT [EditLogic] FROM [EE_Edits] WHERE [PKey] = %d;",
						t_edit_pkey);

					t_logic_qry->SetSQLText(t_logic_buffer);
					t_logic_qry->SetIsActive(true);

					t_edit_logic.assign(t_logic_qry->GetFieldByNumber(0));

					t_original_value  = "#S\"";
					t_original_value += obj->GetOriginalValue();
					t_original_value += "\"";

					t_new_value  = "#S\"";
					t_new_value += obj->GetNewValue();
					t_new_value += "\"";

					t_edit_logic =
						ci_stringReplace(t_edit_logic, t_original_value, t_new_value);

					// SQLize it!
					t_edit_logic =
						stringReplace(t_edit_logic, "'", "''");

					t_update_logic  = "UPDATE [EE_Edits] SET [EditLogic] = '";
					t_update_logic += t_edit_logic;
					t_update_logic += "' WHERE [PKey] = ";
					t_update_logic += intToString(t_edit_pkey);

					t_update_qry->SetSQLText(t_update_logic.c_str());
					t_update_qry->ExecSQL();

				}

				FParent->RecompileEditLogic(t_edit_tag.c_str());

			}
		}
		catch (...)
		{
			// something bad has happened
			t_result = false;
		}
	} // checking we are dealing with FieldName or FieldNumber, nothing else
	return t_result;

} // update_edit_fields
//---------------------------------------------------------------------------
bool TSMFWriter::update_edit_tables(TUpdateEditLogicObject* obj,
	const int table_pkey)
{
	bool t_result = true;

	std::string t_field_to_update = obj->GetFieldName();

	if (t_field_to_update.compare("TableName") == 0)
	{
		std::string t_original_value, t_new_value;

		std::unique_ptr<TSQLQuery> t_pivot_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_logic_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));

		std::vector<int> t_edit_pkeys;
		std::vector<std::string> t_edit_tags;

		int t_edit_pkey;
		char t_pivot_buffer[200];
		char t_logic_buffer[200];
		// I can make no assumptions about the length of EditLogic, so we'll
		// build a string
		std::string t_edit_tag, t_edit_logic, t_update_logic;

		std::snprintf(t_pivot_buffer, 200,
			"SELECT a.[EditPKey], b.[EditTag] FROM [EE_EditTablesPivot] a, [EE_Edits] b " \
			"WHERE a.[EditPKey] = b.[PKey] " \
			"AND a.[LookupTablePKey] = %d;",
			table_pkey);

		t_pivot_qry->SetSQLText(t_pivot_buffer);
		try
		{
			t_pivot_qry->SetIsActive(true);
			while (!t_pivot_qry->Eof())
			{
				t_edit_pkeys.push_back(t_pivot_qry->GetIntFieldByNumber(0));
				t_edit_tags.push_back(t_pivot_qry->GetFieldByNumber(1));

				t_pivot_qry->FetchNext();
			}

			t_pivot_qry->SetIsActive(false);

			for (size_t t_idx = 0; t_idx < t_edit_pkeys.size(); t_idx++)
			{
				t_edit_pkey = t_edit_pkeys[t_idx];
				t_edit_tag.assign(t_edit_tags[t_idx]);

				std::snprintf(t_logic_buffer, 200,
					"SELECT [EditLogic] FROM [EE_Edits] WHERE [PKey] = %d;",
					t_edit_pkey);

				t_logic_qry->SetSQLText(t_logic_buffer);
				t_logic_qry->SetIsActive(true);

				t_edit_logic.assign(t_logic_qry->GetFieldByNumber(0));

				t_original_value  = "\"";
				t_original_value += obj->GetOriginalValue();

				t_new_value  = "\"";
				t_new_value += obj->GetNewValue();

				t_edit_logic =
					ci_stringReplace(t_edit_logic, t_original_value, t_new_value);
				// SQLize it!
				t_edit_logic =
					stringReplace(t_edit_logic, "'", "''");

				t_update_logic  = "UPDATE [EE_Edits] SET [EditLogic] = '";
				t_update_logic += t_edit_logic;
				t_update_logic += "' WHERE [PKey] = ";
				t_update_logic += intToString(t_edit_pkey);

				t_update_qry->SetSQLText(t_update_logic.c_str());
				if (t_update_qry->ExecSQL())
				{
					FParent->RecompileEditLogic(t_edit_tag.c_str());
				}
				else
				{
					t_result = false;
				}
			}
		}
		catch (...)
		{
			// something bad has happened
			t_result = false;
		}
	} // checking we are dealing with TABLENAME, nothing else

	return t_result;

} // update_edit_tables
//---------------------------------------------------------------------------
bool TSMFWriter::update_edit_messages(TUpdateEditLogicObject* obj, const int message_pkey)
{
	bool t_result = true;

	std::string t_field_to_update = obj->GetFieldName();

	if (t_field_to_update.compare("MessageNumber") == 0)
	{
		std::string t_original_value, t_new_value;

		std::unique_ptr<TSQLQuery> t_pivot_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_logic_qry(new TSQLQuery(FSQLDB));
		std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));

		std::vector<int> t_edit_pkeys;
		std::vector<std::string> t_edit_tags;

		int t_edit_pkey;
		char t_pivot_buffer[200];
		char t_logic_buffer[200];
		// I can make no assumptions about the length of EditLogic, so we'll
		// build a string
		std::string t_edit_tag, t_edit_logic, t_update_logic;

		std::snprintf(t_pivot_buffer, 200,
			"SELECT a.[EditPKey], b.[EditTag] FROM [EE_EditMessagesPivot] a, [EE_Edits] b " \
			"WHERE a.[EditPKey] = b.[PKey] " \
			"AND a.[MessagePKey] = %d;",
			message_pkey);

		t_pivot_qry->SetSQLText(t_pivot_buffer);
		try
		{
			t_pivot_qry->SetIsActive(true);
			while (!t_pivot_qry->Eof())
			{
				t_edit_pkeys.push_back(t_pivot_qry->GetIntFieldByNumber(0));
				t_edit_tags.push_back(t_pivot_qry->GetFieldByNumber(1));

				t_pivot_qry->FetchNext();
			}

			t_pivot_qry->SetIsActive(false);

			for (size_t t_idx = 0; t_idx < t_edit_pkeys.size(); t_idx++)
			{
				t_edit_pkey = t_edit_pkeys[t_idx];
				t_edit_tag.assign(t_edit_tags[t_idx]);

				std::snprintf(t_logic_buffer, 200,
					"SELECT [EditLogic] FROM [EE_Edits] WHERE [PKey] = %d;",
					t_edit_pkey);

				t_logic_qry->SetSQLText(t_logic_buffer);
				t_logic_qry->SetIsActive(true);

				t_edit_logic.assign(t_logic_qry->GetFieldByNumber(0));

				t_original_value  = obj->GetOriginalValue();

				t_new_value  = obj->GetNewValue();

///////////////////
				//t_edit_logic =
				//	ci_stringReplace(t_edit_logic, t_original_value, t_new_value);

				size_t t_pos,
					   t_paren_pos;
				std::string t_scrap,
							t_outstring = "";

				char* t_func_names[] = {(char*)"set_warning",
										(char*)"save_warning_text",
										(char*)"error_msg",
										(char*)"save_error_text"};

				for (size_t t_idx=0; t_idx < 4; t_idx++)
				{

					t_pos = ci_stringFind(t_edit_logic.substr(0, t_edit_logic.length()),
						t_func_names[t_idx], true);

					while (t_pos != std::string::npos)
					{
						t_scrap.assign(t_func_names[t_idx]);

						// add to t_outstring everything up to just past where
						// we found the function name
						t_outstring += t_edit_logic.substr(0, t_pos + t_scrap.length());

						// Reduce t_edit_logic to start just past the t_scrap
						t_edit_logic = t_edit_logic.substr(t_pos + t_scrap.length(), t_edit_logic.length());

						t_paren_pos = t_edit_logic.substr(0, t_edit_logic.length()).find(")");

						if (t_paren_pos != std::string::npos)
						{
							t_scrap.assign(t_edit_logic.substr(0, t_paren_pos+1));
							t_scrap = ci_stringReplace(t_scrap, t_original_value, t_new_value, true);
							t_outstring += t_scrap;

							t_edit_logic = t_edit_logic.substr(t_paren_pos+1, t_edit_logic.length());
						}

						t_pos = ci_stringFind(t_edit_logic.substr(0, t_edit_logic.length()),
							t_func_names[t_idx], true);

					} // while (t_pos != std::string::npos)
				}

				// When we get to here, t_outstring either has part of the revised
				// edit logic, or it has nothing. Either way, we want to finish
				// it off...
				t_outstring += t_edit_logic;

				// SQLize it!
				t_edit_logic =
					stringReplace(t_outstring, "'", "''");
///////////////////
				t_update_logic  = "UPDATE [EE_Edits] SET [EditLogic] = '";
				t_update_logic += t_edit_logic;
				t_update_logic += "' WHERE [PKey] = ";
				t_update_logic += intToString(t_edit_pkey);

				t_update_qry->SetSQLText(t_update_logic.c_str());
				if (t_update_qry->ExecSQL())
				{
                    t_edit_tag = stringReplace(t_edit_tag, "'", "''");
					FParent->RecompileEditLogic(t_edit_tag.c_str());
				}
				else
				{
					t_result = false;
				}
			}
		}
		catch (...)
		{
			// something bad has happened
			t_result = false;
		}


	} // checking we are dealing with MESSAGENUMBER, nothing else
	return t_result;
} // update_edit_messages
//---------------------------------------------------------------------------
int TSMFWriter::post_edit(TTableDef* table)
{
	int t_result = 0;

	std::string t_savepoint_name = "__smfwriter_post_edit__";

	FSQLDB->SavePoint(t_savepoint_name.c_str());

	std::vector<TUpdateEditLogicObject*> t_update_edit_logic_objects;

	try
	{
		std::string t_table_name = table->GetTableName();
		std::string t_field_name;

		// Build a SQL string from the fields that are flagged
		// "modified", then write it to the database
		std::string t_data, t_scrap;
		int t_max_cols = table->GetColumnCount();
		for (int t_col = 0; t_col < t_max_cols; t_col++)
		{
			TFieldDef* t_fld = table->GetColumn(t_col);
			if (t_fld)
			{
				if (t_fld->GetIsModified())
				{
					// We pause here to consider whether a change to a critical
					// field in specific MCO tables will require us to cascade
					// the change into edit logic. The evaluation and processing
					// will occur at the end of this function.

					std::string t_original_value;
					t_field_name.assign(t_fld->GetFieldName());

					if (t_table_name.compare("EE_Fields") == 0)
					{
						if ((t_field_name.compare("FieldName") == 0) ||
							(t_field_name.compare("FieldNumber") == 0))
						{
							// We won't bother doing anything unless
							// the new name differs from the old
							get_original_value(table, t_fld, t_original_value);
							std::string t_new_value = t_fld->GetValue();
							if (!t_original_value.empty())
							{
								TUpdateEditLogicObject* t_obj =
									new TUpdateEditLogicObject(t_table_name.c_str(),
															   t_field_name.c_str(),
															   t_original_value.c_str(),
															   t_new_value.c_str());

								t_update_edit_logic_objects.push_back(t_obj);
							}
						}

					}
					else if ((t_table_name.compare("EE_Messages") == 0) &&
							 (t_field_name.compare("MessageNumber") == 0))
					{
						get_original_value(table, t_fld, t_original_value);
						std::string t_new_value = t_fld->GetValue();
						if (!t_original_value.empty())
						{
							TUpdateEditLogicObject* t_obj =
								new TUpdateEditLogicObject(t_table_name.c_str(),
														   t_field_name.c_str(),
														   t_original_value.c_str(),
														   t_new_value.c_str());

							t_update_edit_logic_objects.push_back(t_obj);
						}
					}

					else if ((t_table_name.compare("EE_LookupTables") == 0) &&
							 (t_field_name.compare("TableName") == 0))
					{
						get_original_value(table, t_fld, t_original_value);
						std::string t_new_value = t_fld->GetValue();
						if (!t_original_value.empty())
						{
							TUpdateEditLogicObject* t_obj =
								new TUpdateEditLogicObject(t_table_name.c_str(),
														   t_field_name.c_str(),
														   t_original_value.c_str(),
														   t_new_value.c_str());

							t_update_edit_logic_objects.push_back(t_obj);
						}
					}

					// Now back to our processing...
					t_scrap.assign(t_fld->GetValue());

					// WATCH OUT for Lookup Tables that use the LimitToList or Between
					// CHECK constraint feature. If the column is defined to
					// be NULLable, and a user wants to eliminate a value
					// previously loaded into it, set it to NULL.
					if ((t_scrap.empty()) &&
						(t_fld->HasLimitToList() ||
						 t_fld->HasCheckBetween()) &&
						(t_fld->GetIsNullable()))
					{
						t_scrap.assign("NULL");
					}

					// Escape embedded apostrophes
					t_scrap = stringReplace(t_scrap, "'", "''");

					if (t_data.length() > 0)
						t_data += ", ";

					TEditDataType t_data_type = t_fld->GetDataType();
					if ((t_data_type == edtInteger) ||
						(t_data_type == edtDouble))
					{
						t_data +=
							(std::string)"[" +
							(std::string)t_fld->GetFieldName() +
							(std::string)"] = " +
							t_scrap ;

					}
					else
					{
						// This is also for the LimitToList/NULLable issue
						// described, above.
						if (t_scrap.compare("NULL") == 0)
						{
							t_data +=
								(std::string)"[" +
								(std::string)t_fld->GetFieldName() +
								(std::string)"] = NULL";
						}
						else
						{
							t_data +=
								(std::string)"[" +
								(std::string)t_fld->GetFieldName() +
								(std::string)"] = '" +
								t_scrap +
								(std::string)"'";
						}
					}
				}
			}
			else
			{
				// I should probably throw an error here...
				std::string t_error = "Mayhem in the tables list! Column ";
				t_error += intToString(t_col);
				t_error += " in table \"";
				t_error += table->GetTableName();
				t_error += "\" out of range";

				FEngineErrorsPump->AddError("TSMFWriter","post_edit",t_error);

				if (!t_update_edit_logic_objects.empty())
				{
					std::for_each(t_update_edit_logic_objects.begin(),
								  t_update_edit_logic_objects.end(),
								  DeleteUpdateEditLogicObject);
				}

				return -1;
			}
		}

		if (t_data.length() > 0)
		{
			int t_pkey = table->GetPostToPKey();

			std::string t_sql = "UPDATE [" ;
			t_sql += table->GetTableName() ;
			t_sql += "] ";
			t_sql += (std::string)"SET " + t_data;
			t_sql += " WHERE PKey = ";
			t_sql += intToString(t_pkey) + (std::string)";";

			std::unique_ptr<TSQLQuery> t_update_qry(new TSQLQuery(FSQLDB));

			try
			{
				t_update_qry->SetSQLText(t_sql.c_str());

				if (!t_update_qry->ExecSQL())
				{
					std::string t_error = t_update_qry->LastErrorMessage();
					FEngineErrorsPump->AddError("TSMFWriter","post_edit",t_error);
					t_error = (std::string)"SQL: " + t_sql;
					FEngineErrorsPump->AddError("TSMFWriter","post_edit",t_error);

					t_result = -1;
				}

				// This table is no longer in insert/update mode
				table->ClearWriteFields();
			}
			catch (...)
			{
				std::string t_error = t_update_qry->LastErrorMessage();
				FEngineErrorsPump->AddError("TSMFWriter","post_edit",t_error);
				t_result = -1;
			}

		}

		// Cascade changes, and clean up
		if (t_result == 0 && !t_update_edit_logic_objects.empty())
		{
			if (t_table_name.compare("EE_Fields") == 0)
			{
				int t_field_pkey = table->GetPostToPKey();

				// The field object could have modified FieldName and/or
				// FieldNumber, so we have to iterate the
				// vector for each of these.
				// Note: function update_edit_fields will recompile each affected
				// edit every time it is called. 
				for (size_t t_idx=0; t_idx < t_update_edit_logic_objects.size(); t_idx++)
				{
					TUpdateEditLogicObject* t_obj = t_update_edit_logic_objects[t_idx];
					t_scrap.assign(t_obj->GetFieldName());
					if ((t_scrap.compare("FieldName") == 0) ||
						(t_scrap.compare("FieldNumber") == 0))
					{
						if (!update_edit_fields(t_obj, t_field_pkey))
						{
							t_result = -1;
							break;
						}
					}
				}
			}
			else if (t_table_name.compare("EE_LookupTables") == 0)
			{
				int t_table_pkey = table->GetPostToPKey();
				// The only concern we have here is the table name, for which
				// a change must be cascaded into edit logic. So grab the one-and-
				// only entry in the vector and process it.
				TUpdateEditLogicObject* t_obj = t_update_edit_logic_objects[0];
				// abundance of caution
				t_scrap.assign(t_obj->GetFieldName());
				if (t_scrap.compare("TableName") == 0)
				{
					if (FParent->RenameLookupTable(t_obj->GetOriginalValue(),
												   t_obj->GetNewValue(), true) == 0)
					{
						if (!update_edit_tables(t_obj, t_table_pkey))
						{
							t_result = -1;
						}
					}
					else
					{
						t_result = -1;
					}

					if (t_result != 0)
					{
						std::string t_error = "DANGER! Failed attempt to rename table may have corrupted this metafile. Close it WITHOUT saving!";
						FEngineErrorsPump->AddError("TSMFWriter","post_edit",t_error);
					}
				}
			}
			else if (t_table_name.compare("EE_Messages") == 0)
			{
				int t_message_pkey = table->GetPostToPKey();
				// If the message number was modified, it could affect edit logic
				TUpdateEditLogicObject* t_obj = t_update_edit_logic_objects[0];
				t_scrap.assign(t_obj->GetFieldName());
				if (t_scrap.compare("MessageNumber") == 0)
				{
					if (!update_edit_messages(t_obj, t_message_pkey))
					{
						t_result = -1;
					}
				}
			}

		} // if (t_result == 0 && !t_update_edit_logic_objects.empty())

		// If we make it to here, commit the savpoint transaction
		if (t_result != 0)
		{
			FSQLDB->RollbackTo(t_savepoint_name.c_str());
		}
		// even if rolled back, we have to release the savepoint or it wreaks havoc
		FSQLDB->Release(t_savepoint_name.c_str());
	}
	catch (...)
	{
		FSQLDB->RollbackTo(t_savepoint_name.c_str());
		FSQLDB->Release(t_savepoint_name.c_str());
		t_result = -1;
	}

	std::for_each(t_update_edit_logic_objects.begin(),
				  t_update_edit_logic_objects.end(),
				  DeleteUpdateEditLogicObject);

	return t_result;
} // post_edit
//---------------------------------------------------------------------------
int TSMFWriter::post_insert(TTableDef* table, int* new_pkey)
{
	int t_result = 0;

	// default the "new primary key" value to 0
	*new_pkey = 0;

	// Build a SQL string from the fields that are flagged
	// "modified", then write it to the database
	std::string t_fld_list, t_data, t_scrap;
	int t_max_cols = table->GetColumnCount();
	for (int t_col = 0; t_col < t_max_cols; t_col++)
	{
		TFieldDef* t_fld = table->GetColumn(t_col);
		if (t_fld)
		{
			if (t_fld->GetIsModified())
			{
				if (t_fld_list.length() > 0)
					t_fld_list += ", ";
				if (t_data.length() > 0)
					t_data += ", ";
				t_fld_list +=
					(std::string)"[" +
					(std::string)t_fld->GetFieldName() +
					(std::string)"]";

				t_scrap = stringReplace(t_fld->GetValue(), "'", "''");

				TEditDataType t_data_type = t_fld->GetDataType();
				if ((t_data_type == edtInteger) ||
					(t_data_type == edtDouble))
				{
					t_data += t_scrap;
				}
				else
				{
					t_data +=
						(std::string)"'" +
						t_scrap +
						(std::string)"'";
				}
			}
		}
		else
		{
			std::string t_error = "Mayhem in the tables list! Column ";
			t_error += intToString(t_col);
			t_error += " in table \"";
			t_error += table->GetTableName();
			t_error += "\" out of range";

			FEngineErrorsPump->AddError("TSMFWriter","post_insert",t_error);

			return -1;
		}
	} // for (t_col < t_max_cols)

	if (t_data.length() > 0)
	{
		std::string t_sql = "INSERT INTO [" ;
		t_sql += table->GetTableName() ;
		t_sql += "] (";
		t_sql += t_fld_list;
		t_sql += ") VALUES ( ";
		t_sql += t_data;
		t_sql += ");";

		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));

		try
		{
			t_qry->SetSQLText(t_sql.c_str());

			if (t_qry->ExecSQL())
			{
				// Now get the new primary key
				t_sql = "SELECT PKey FROM [";
				t_sql += table->GetTableName();
				t_sql += "] WHERE rowid = last_insert_rowid();";

				t_qry->SetSQLText(t_sql.c_str());
				t_qry->SetIsActive(true);
				try
				{
					if (!t_qry->Eof())
					{
						*new_pkey = t_qry->GetIntFieldByNumber(0);
					}
				}
				catch(...)
				{
					t_qry->SetIsActive(false);
					throw;
				}

				t_qry->SetIsActive(false);
			}
			else
			{
				std::string t_error = t_qry->LastErrorMessage();
				FEngineErrorsPump->AddError("TSMFWriter","post_insert",t_error);
				t_error = "SQL: " + t_sql;
				FEngineErrorsPump->AddError("TSMFWriter","post_insert",t_error);
				t_result = -1;
			}

			// This table is no longer in insert/update mode
			table->ClearWriteFields();
		}
		catch (...)
		{
			std::string t_error = t_qry->LastErrorMessage();
			FEngineErrorsPump->AddError("TSMFWriter","post_insert",t_error);

			t_result = -1;
		}
	}

	/* Example:
	insert into [DPLACE] ([CODE],[LABEL])
	values ('996','Nobody died');
	// To get the pkey (or any other field, for that matter) added in the
	// last insert into this table, use this:
	SELECT PKey FROM [table_name]
	WHERE rowid = last_insert_rowid();
	*/

	return t_result;
} // post_insert
//---------------------------------------------------------------------------

int TSMFWriter::Cancel(const std::string table_name)
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		t_table->Cancel();
		/* I don't think we need to raise an error in cases of
		   cancelling edit/insert mode unnecessarily. No harm, no foul.
		if (!t_table->Cancel())
		{
			std::string t_error = "Table named \"";
			t_error += table_name;
			t_error += "\" is NOT in insert or edit mode; nothing to cancel";
			FEngineErrorsPump->AddError("TSMFWriter","Cancel",t_error);
			t_result = -1;
		}
		*/
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMFWriter","Cancel",t_error);
		t_result = -1;
	}

	return t_result;
} // Cancel()
//---------------------------------------------------------------------------
int TSMFWriter::AddEditSetsPivot( const int editset_pkey, const int edit_pkey)
{
	int t_result = 0;

	if (have_access_rights(_mco_EditSets, editset_pkey))
	{
		char t_buffer[200];
		std::snprintf(t_buffer, 200,
			"SELECT PKey FROM EE_EditSetEditsPivot WHERE EditSetPKey = %d AND EditPKey = %d;",
			editset_pkey, edit_pkey);

		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		try
		{
			t_qry->SetSQLText(t_buffer);
			t_qry->SetIsActive(true);
			// If we do NOT get an empty result set, this record is already in
			// the table. No harm, no foul... just return t_result;
			if (t_qry->Eof())
			{
				t_qry->SetIsActive(false);
				std::snprintf(t_buffer, 200,
					"INSERT INTO EE_EditSetEditsPivot (EditSetPKey, EditPKey) VALUES (%d, %d);",
					editset_pkey, edit_pkey);

				t_qry->SetSQLText(t_buffer);
				t_qry->ExecSQL();
			}
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMFWriter","AddEditSetsPivot",
				"Failure to add Edit Set/Edit pivot record");
			FEngineErrorsPump->AddError("TSMFWriter", "AddEditSetsPivot",
				t_qry->LastErrorMessage());
			FEngineErrorsPump->AddError("TSMFWriter", "AddEditSetsPivot",
				(std::string)"SQL: " + (std::string)t_buffer);

			t_result = -1;
		}
	}
	else
	{
		// User does not have access rights to this edit set
		FEngineErrorsPump->AddError("TSMFWriter","AddEditSetsPivot",
			"You do not have access rights to modify this edit set");
		t_result = -1;
	}
	return t_result;
} // AddEditSetsPivot
//---------------------------------------------------------------------------
int TSMFWriter::DeleteEditSetsPivot( const int editset_pkey, const int edit_pkey)
{
	int t_result = 0;

	if (have_access_rights(_mco_EditSets, editset_pkey))
	{
		char t_buffer[200];
		std::snprintf(t_buffer, 200,
			"SELECT PKey FROM EE_EditSetEditsPivot WHERE EditSetPKey = %d AND EditPKey = %d;",
			editset_pkey, edit_pkey);

		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		try
		{
			t_qry->SetSQLText(t_buffer);
			t_qry->SetIsActive(true);
			// If we do NOT get an empty result set, this record is already in
			// the table. No harm, no foul... just return t_result;
			if (!t_qry->Eof())
			{
				int t_piv_pkey = t_qry->GetIntFieldByName("PKey");
				t_qry->SetIsActive(false);
				std::snprintf(t_buffer, 200,
					"DELETE FROM EE_EditSetEditsPivot WHERE PKey=%d;",
					t_piv_pkey);

				t_qry->SetSQLText(t_buffer);
				t_qry->ExecSQL();
			}
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMFWriter","DeleteEditSetsPivot",
				"Failure to delete Edit Set/Edit pivot record");
			FEngineErrorsPump->AddError("TSMFWriter", "DeleteEditSetsPivot",
				t_qry->LastErrorMessage());
			FEngineErrorsPump->AddError("TSMFWriter", "DeleteEditSetsPivot",
				(std::string)"SQL: " + (std::string)t_buffer);


			t_result = -1;
		}

	}
	else
	{
		// User does not have access rights to this edit set
		FEngineErrorsPump->AddError("TSMFWriter","DeleteEditSetsPivot",
			"You do not have access rights to modify this edit set");
		t_result = -1;
	}
	return t_result;
} // DeleteEditSetsPivot
//---------------------------------------------------------------------------

int TSMFWriter::AddLayoutsPivot( const int layout_pkey, const int field_pkey,
	const int start_pos)
{
	int t_result = 0;

	if (have_access_rights(_mco_Layouts, layout_pkey))
	{
		char t_buffer[200];
		std::snprintf(t_buffer, 200,
			"SELECT PKey FROM EE_LayoutFieldsPivot WHERE LayoutPKey = %d AND FieldPKey = %d;",
			layout_pkey, field_pkey);

		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		try
		{
			t_qry->SetSQLText(t_buffer);
			t_qry->SetIsActive(true);
			// If we do NOT get an empty result set, this record is already in
			// the table. No harm, no foul... just return t_result;
			if (t_qry->Eof())
			{
				t_qry->SetIsActive(false);
				std::snprintf(t_buffer, 200,
					"INSERT INTO EE_LayoutFieldsPivot (LayoutPKey, FieldPKey, StartPos) VALUES (%d, %d, %d);",
					layout_pkey, field_pkey, start_pos);

				t_qry->SetSQLText(t_buffer);
				t_qry->ExecSQL();
			}
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMFWriter","AddLayoutsPivot",
				"Failure to add Layout/Field pivot record");
			FEngineErrorsPump->AddError("TSMFWriter", "AddLayoutsPivot",
				t_qry->LastErrorMessage());
			FEngineErrorsPump->AddError("TSMFWriter", "AddLayoutsPivot",
				(std::string)"SQL: " + (std::string)t_buffer);

			t_result = -1;
		}
	}
	else
	{
		// User does not have access rights to this layout
		FEngineErrorsPump->AddError("TSMFWriter","AddLayoutsPivot",
			"You do not have access rights to modify this layout");
		t_result = -1;
	}

	return t_result;
} // AddLayoutsPivot
//---------------------------------------------------------------------------
int TSMFWriter::DeleteLayoutsPivot( const int layout_pkey, const int field_pkey)
{
	int t_result = 0;

	if (have_access_rights(_mco_Layouts, layout_pkey))
	{
		char t_buffer[200];
		std::snprintf(t_buffer, 200,
			"SELECT PKey FROM EE_LayoutFieldsPivot WHERE LayoutPKey = %d AND FieldPKey = %d;",
			layout_pkey, field_pkey);

		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		try
		{
			t_qry->SetSQLText(t_buffer);
			t_qry->SetIsActive(true);
			// If we do NOT get an empty result set, this record is already in
			// the table. No harm, no foul... just return t_result;
			if (!t_qry->Eof())
			{
				int t_piv_pkey = t_qry->GetIntFieldByName("PKey");
				t_qry->SetIsActive(false);
				std::snprintf(t_buffer, 200,
					"DELETE FROM EE_LayoutFieldsPivot WHERE PKey=%d;",
					t_piv_pkey);

				t_qry->SetSQLText(t_buffer);
				t_qry->ExecSQL();
			}
		}
		catch (...)
		{
			FEngineErrorsPump->AddError("TSMFWriter","DeleteLayoutsPivot",
				"Failure to delete Layout/Field pivot record");
			FEngineErrorsPump->AddError("TSMFWriter", "DeleteLayoutsPivot",
				t_qry->LastErrorMessage());
			FEngineErrorsPump->AddError("TSMFWriter", "DeleteLayoutsPivot",
				(std::string)"SQL: " + (std::string)t_buffer);


			t_result = -1;
		}
	}
	else
	{
		// User does not have access rights to this layout
		FEngineErrorsPump->AddError("TSMFWriter","DeleteLayoutsPivot",
			"You do not have access rights to modify this layout");
		t_result = -1;
	}

	return t_result;
} // DeleteLayoutsPivot
//---------------------------------------------------------------------------
int TSMFWriter::ClearEditLogicPivots( const int edit_pkey )
{
	int t_result = 0;

	char t_buffer[200];

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	try
	{
		std::snprintf(t_buffer, 200,
			"DELETE FROM EE_EditFieldsPivot WHERE EditPKey = %d;", edit_pkey);
		t_qry->SetSQLText(t_buffer);
		t_qry->ExecSQL();

		std::snprintf(t_buffer, 200,
			"DELETE FROM EE_EditTablesPivot WHERE EditPKey = %d;", edit_pkey);
		t_qry->SetSQLText(t_buffer);
		t_qry->ExecSQL();

		std::snprintf(t_buffer, 200,
			"DELETE FROM EE_EditMessagesPivot WHERE EditPKey = %d;", edit_pkey);
		t_qry->SetSQLText(t_buffer);
		t_qry->ExecSQL();

		std::snprintf(t_buffer, 200,
			"DELETE FROM EE_EditErrorsPivot WHERE EditPKey = %d;", edit_pkey);
		t_qry->SetSQLText(t_buffer);
		t_qry->ExecSQL();
	}
	catch (...)
	{
		FEngineErrorsPump->AddError("TSMFWriter","ClearEditLogicPivots",
			"Failure to delete Layout/Field pivot record");
		FEngineErrorsPump->AddError("TSMFWriter", "ClearEditLogicPivots",
			t_qry->LastErrorMessage());
		FEngineErrorsPump->AddError("TSMFWriter", "ClearEditLogicPivots",
			(std::string)"SQL: " + (std::string)t_buffer);


		t_result = -1;
	}
	return t_result;
} // ClearEditLogicPivots
//---------------------------------------------------------------------------
void TSMFWriter::compare_old_new_field_length(int field_pkey, int new_field_length)
{
	// Called by SetFieldByName and SetFieldByNumber, this function
	// checks to see if EE_Fields.FieldLength is being changed for
	// an existing Field object, and if so, delete references that may
	// exist for it in Layouts.

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[100];
	int t_old_field_length;

	std::snprintf(t_buffer, 100,
		"SELECT FieldLength FROM EE_Fields WHERE PKey=%d;",
		field_pkey);

	t_qry->SetSQLText(t_buffer);

	t_qry->SetIsActive(true);
	t_old_field_length = t_qry->GetIntFieldByName("PKey");

	t_qry->SetIsActive(false);

	if (t_old_field_length != new_field_length)
	{
		std::snprintf(t_buffer, 100,
			"DELETE FROM EE_LayoutFieldsPivot WHERE FieldPKey=%d;",
				field_pkey);

		t_qry->SetSQLText(t_buffer);
		t_qry->ExecSQL();
	}

} // compare_old_new_field_length
//---------------------------------------------------------------------------
int TSMFWriter::SetFieldByName(const std::string table_name,
	const std::string field_name, const std::string value)
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		if (!t_table->IsEditing())
		{
			std::string t_error = "Table named \"";
			t_error += table_name;
			t_error += "\" is NOT in insert or edit mode; cannot modify data";
			FEngineErrorsPump->AddError("TSMFWriter","SetFieldByName",t_error);
			return -1;
		}

		TFieldDef* t_fld = t_table->GetColumn(field_name.c_str());
		if (!t_fld)
		{
			std::string t_error = "Cannot find field named \"";
			t_error += field_name;
			t_error += "\" in table named \"";
			t_error += table_name;
			t_error += "\"";
			FEngineErrorsPump->AddError("TSMFWriter","SetFieldByName",t_error);
			return -1;
		}

		// Special handling for passwords
		if ((table_name.compare("EE_Agencies") == 0) &&
			(field_name.compare("Password") == 0))
		{
			// Translate the plain-text password into munged
			char t_munged_pw[11];
			std::memset(t_munged_pw, '\0', 11);
			// Munge everything except the NULL terminator
			FAccessController->MungedPassword(value.c_str(), t_munged_pw, 10);

			t_fld->SetValue(t_munged_pw);
		}
		// Special handling for EditLogic
		else if ((table_name.compare("EE_Edits") == 0) &&
				 (field_name.compare("EditLogic") == 0))
		{
			t_fld->SetValue(value.c_str());
			// Clear the CompiledLogic field
			t_fld = t_table->GetColumn("CompiledLogic");
			if (t_fld)
			{
				t_fld->SetValue("");
			}
		}
		// Issue error message for attempts to change CompiledLogic
		else if ((table_name.compare("EE_Edits") == 0) &&
				 (field_name.compare("CompiledLogic") == 0))
		{
			std::string t_error = "Programmer error: It is illegal to write ";
			t_error += "directly to EE_Edits.CompiledLogic.";
			FEngineErrorsPump->AddError("TSMFWriter","SetFieldByName",t_error);
			t_error += "Use Edit_RecompileEditLogic instead";
			FEngineErrorsPump->AddError("TSMFWriter","SetFieldByName",t_error);
			return -1;
		}
		// Special handling for changes to Field Length
		else if ((table_name.compare("EE_Fields") == 0) &&
				 (field_name.compare("FieldLength") == 0) &&
				 (t_table->GetState() == ewEdit))
		{
			int t_field_pkey = t_table->GetPostToPKey();

			if (t_field_pkey > 0)
			{
				int t_new_field_length = std::atoi(value.c_str());
				compare_old_new_field_length(t_field_pkey, t_new_field_length);
			}

			t_fld->SetValue(value.c_str());

		}
		else
		{
			t_fld->SetValue(value.c_str());
		}
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMFWriter","SetFieldByName",t_error);
		t_result = -1;
	}

	return t_result;
} // SetFieldByName
//---------------------------------------------------------------------------
int TSMFWriter::SetFieldByNumber(const std::string table_name,
	const int column_number, const std::string value)
{
	int t_result = 0;

	TTableDef* t_table = FTablesList->GetTable(table_name);
	if (t_table)
	{
		if (!t_table->IsEditing())
		{
			std::string t_error = "Table named \"";
			t_error += table_name;
			t_error += "\" is NOT in insert or edit mode; cannot modify data";
			FEngineErrorsPump->AddError("TSMFWriter","SetFieldByNumber",t_error);
			return -1;
		}

		TFieldDef* t_fld = t_table->GetColumn(column_number -1);
		if (!t_fld)
		{
			std::string t_error = "Column number ";
			t_error += intToString(column_number);
			t_error += "\" in table named \"";
			t_error += table_name;
			t_error += "\" is out of range; assume 1-based numbering";
			FEngineErrorsPump->AddError("TSMFWriter","SetFieldByNumber",t_error);
			return -1;
		}

		// Watch out for EE_Agencies.Password,
		// EE_Edits.EditLogic/EE_Edits.CompiledLogic
		// and EE_Fields.FieldLength
		if ((table_name.compare("EE_Agencies") == 0) ||
			(table_name.compare("EE_Edits") == 0) ||
			(table_name.compare("EE_Fields") == 0))
		{
			std::string t_field_name = t_fld->GetFieldName();

			if (t_field_name.compare("Password") == 0)
			{
				// Translate the plain-text password into munged
				char t_munged_pw[11];
				std::memset(t_munged_pw, '\0', 11);
				// Munge everything except the NULL terminator
				FAccessController->MungedPassword(value.c_str(), t_munged_pw, 10);

				t_fld->SetValue(t_munged_pw);

				return t_result;
			}
			else if (t_field_name.compare("EditLogic") == 0)
			{
				t_fld->SetValue(value.c_str());
				// Clear the CompiledLogic field
				t_fld = t_table->GetColumn("CompiledLogic");
				if (t_fld)
				{
					t_fld->SetValue("");
				}

				return t_result;
			}
			else if (t_field_name.compare("CompiledLogic") == 0)
			{
				std::string t_error = "Programmer error: It is illegal to write ";
				t_error += "directly to EE_Edits.CompiledLogic.";
				FEngineErrorsPump->AddError("TSMFWriter","SetFieldByNumber",t_error);
				t_error += "Use Edit_RecompileEditLogic instead";
				FEngineErrorsPump->AddError("TSMFWriter","SetFieldByNumber",t_error);
				return -1;
			}
			else if ((table_name.compare("EE_Fields") == 0) &&
					 (t_field_name.compare("FieldLength") == 0) &&
					 (t_table->GetState() == ewEdit))
			{
				int t_field_pkey = t_table->GetPostToPKey();

				if (t_field_pkey > 0)
				{
					int t_new_field_length = std::atoi(value.c_str());
					compare_old_new_field_length(t_field_pkey, t_new_field_length);
				}

				t_fld->SetValue(value.c_str());

				return t_result;

			}
		}

		t_fld->SetValue(value.c_str());
	}
	else
	{
		std::string t_error = "Cannot find table named \"";
		t_error += table_name;
		t_error += "\"";
		FEngineErrorsPump->AddError("TSMFWriter","SetFieldByNumber",t_error);
		t_result = -1;
	}

	return t_result;
} // SetFieldByNumber
//---------------------------------------------------------------------------
int TSMFWriter::ValidateUnique( const char* table_name, const int pkey,
	const char* field_name, const char* value, int* data_is_unique,
	int* dupe_pkey )
{
	int t_result = 0;

	// Handle embedded apostrophes in the value
	std::string t_value = stringReplace(value, "'", "''");

	char t_buffer[5000];
	std::snprintf(t_buffer, 5000,
		"SELECT PKey FROM [%s] WHERE ([%s] = '%s') AND PKey <> %d;",
		table_name,
		field_name,
		t_value.c_str(),
		pkey);
	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	try
	{
		t_qry->SetSQLText(t_buffer);
		t_qry->SetIsActive(true);
		if (t_qry->Eof())
		{
			*data_is_unique = 1; // true, data is unique
			*dupe_pkey = 0;      // no dupes found
		}
		else
		{
			*data_is_unique = 0; // false, data is NOT unique
			*dupe_pkey = t_qry->GetIntFieldByName("PKey");
		}
	}
	catch (...)
	{
		FEngineErrorsPump->AddError("TSMFWriter", "ValidateUnique", t_qry->LastErrorMessage());
		FEngineErrorsPump->AddError("TSMFWriter", "ValidateUnique", (std::string)"SQL: " +
			(std::string)t_buffer);
		t_result = -1;
	}


	return t_result;
} // ValidateUnique
//---------------------------------------------------------------------------

int TSMFWriter::BeginTrans()
{
	int t_result = 0;
	try
	{
		FSQLDB->BeginTrans();
	}
	catch (...)
	{
		std::string t_error = FSQLDB->GetLastDBErrorMessage();
		if (t_error.length() == 0)
		{
			t_error = "Unable to begin transaction; reason not known";
		}

		FEngineErrorsPump->AddError("TSMFWriter","BeginTrans",t_error);

		t_result = -1;
	}

	return t_result;
} // BeginTrans
//---------------------------------------------------------------------------
int TSMFWriter::CommitTrans()
{
	int t_result = 0;
	try
	{
		FSQLDB->CommitTrans();
	}
	catch (...)
	{
		std::string t_error = FSQLDB->GetLastDBErrorMessage();
		if (t_error.length() == 0)
		{
			t_error = "Unable to commit transaction; reason not known";
		}

		FEngineErrorsPump->AddError("TSMFWriter","CommitTrans",t_error);

		t_result = -1;
	}

	return t_result;
} // CommitTrans
//---------------------------------------------------------------------------
int TSMFWriter::RollbackTrans()
{
	int t_result = 0;
	try
	{
		FSQLDB->RollbackTrans();
	}
	catch (...)
	{
		std::string t_error = FSQLDB->GetLastDBErrorMessage();
		if (t_error.length() == 0)
		{
			t_error = "Unable to roll back transaction; reason not known";
		}

		FEngineErrorsPump->AddError("TSMFWriter","RollbackTrans",t_error);

		t_result = -1;
	}

	return t_result;
} // RollbackTrans
//---------------------------------------------------------------------------

int TSMFWriter::DefineTableStructure( const char* table_name,
	const int restructure_existing)
{
	int t_result = 0;
	if (FDefineTable)
	{
		// Start with a fresh slate
		FDefineTable->ClearAll();

		FDefineTable->SetTableName(table_name);
		bool t_restructure = (restructure_existing != 0);
		FDefineTable->SetIsRestructuring(t_restructure);
	}
	else
	{
		FEngineErrorsPump->AddError("TSMFWriter", "DefineTableStructure",
			"Engine programmer error: FDefineTable not instantiated");
		t_result = -1;
	}
	return t_result;
} // DefineTableStructure
//---------------------------------------------------------------------------
int TSMFWriter::AddColumnDef(
				 const char* col_name,
				 const char* copy_from_col_name,
				 const int data_type,
				 const int data_length,
				 const int is_nullable,
				 const int is_trigger_field,
				 const char* check_limit_to_list,
				 const char* check_range_low,
				 const char* check_range_high,
				 const char* default_value,
				 const int index_status)
{

	int t_result = 0;

	if (FDefineTable)
	{
		// create TDefineColumn object and add to FDefineTable
		// First translate the incoming parameters into a string that
		// describes a column definition, e.g.:
		//
		// [Foo] varchar(10) NOT NULL collate nocase check(Foo in ('Foo','Bar')) default 'Foo'
		//
		std::string t_lb = "[", t_rb = "]";
		// column name
		std::string t_spec = t_lb + std::string(col_name) + t_rb;

		// data type / is_trigger_field
		if (data_type == 0) // character field
		{
			if (data_length > 0)
			{
				// Note: We use a trick to identify trigger-maintained
				// fields in EditWriter: If they are defined as "varchar(1)",
				// we know a trigger is attached. Any other character field
				// of fewer than 7 characters will be defined as "char(<n>)"
				if ((is_trigger_field != 0) || (data_length > 6))
				{
					t_spec += std::string(" varchar(");
				}
				else
				{
					t_spec += std::string(" char(");
				}

				t_spec += intToString(data_length) + std::string(") ");
			}
			else
			{
				FEngineErrorsPump->AddError("TSMFWriter", "AddColumnDef",
					"Character fields cannot be defined with 0 length");
				t_result = -1;
			}
		}
		else
		{
			t_spec += std::string(" integer ");
		}

		// nullable, collate sequence for character
		if (is_nullable == 0)
		{
			t_spec += std::string("NOT NULL ");
		}
		else
		{
			t_spec += std::string("NULL ");
		}

		if (data_type == 0) // character
		{
			// case-insensitive collation
			t_spec += std::string("collate nocase ");

			// check constraints; can be only one or the other type
			if (check_limit_to_list && std::strlen(check_limit_to_list) > 0)
			{
				t_spec += std::string("check(") + std::string(col_name);
				t_spec += std::string(" in (") + std::string(check_limit_to_list);
				t_spec += std::string(")) ");
			}
			else if (check_range_low && check_range_high)
			{
				if (std::strlen(check_range_low) > 0 &&
					std::strlen(check_range_high) > 0)
				{
					t_spec += std::string("check(") + std::string(col_name);
					t_spec += std::string(" between '") + std::string(check_range_low);
					t_spec += std::string("' and '") + std::string(check_range_high);
					t_spec += "') ";
				}
			}
		}
		else
		{
			// ensure that integer columns always get integer data
			t_spec += std::string("check(typeof(") + std::string(col_name) +
					  std::string(") = 'integer') ");

			if (check_range_low && check_range_high)
			{
				if (std::strlen(check_range_low) > 0 &&
					std::strlen(check_range_high) > 0)
				{
					t_spec += std::string("check(") + std::string(col_name);
					t_spec += std::string(" between ") + std::string(check_range_low);
					t_spec += std::string(" and ") + std::string(check_range_high);
					t_spec += ") ";
				}
			}
		}

		// default constraint; force it if column is to be NOT NULL
		if (default_value || (is_nullable == 0))
		{
			if (default_value && std::strlen(default_value) > 0)
			{
				if (data_type == 0) // character, surround with single quotes
				{
					t_spec += std::string("default '") + std::string(default_value);
					t_spec += std::string("'");
				}
				else
				{
					t_spec += std::string("default ") + std::string(default_value);
				}
			}
			else
			{
				if (data_type == 0) // character
				{
					t_spec += std::string("default ''");
				}
				else                // integer
				{
					t_spec += std::string("default 0");
				}
			}
		}
/*
		if (default_value &&
			((is_nullable == 0) || (std::strlen(default_value) > 0)))
		{
			if (data_type == 0) // character, surround with single quotes
			{
				t_spec += std::string("default '") + std::string(default_value);
				t_spec += std::string("'");
			}
			else
			{
				t_spec += std::string("default ") + std::string(default_value);
			}
		}
*/
		// All systems go?
		if (t_result == 0)
		{
			TDefineColumn* t_col = new TDefineColumn(col_name, t_spec.c_str(), copy_from_col_name);
			t_col->SetIsTriggerMaintained(is_trigger_field);
			FDefineTable->AddColumnDef(t_col);

			if (index_status > 0)
			{
				// Long table names need to be truncated for use in the index names
				std::string t_idx_name = std::string(FDefineTable->GetTableName());
				if (t_idx_name.length() > 8)
					t_idx_name = std::string("x") + t_idx_name.substr(0,8) + std::string("_");
				else
					t_idx_name = std::string("x") + t_idx_name + std::string("_");

				t_idx_name += std::string(col_name);
				char t_test_idx_name[51];
				std::strncpy(t_test_idx_name, t_idx_name.c_str(), 50);

				size_t t_counter = 0;

				while (index_name_exists(t_test_idx_name))
				{
					if (t_counter > 9)
						break;

					std::snprintf(t_test_idx_name, 50,
						"%s%d",
						t_idx_name.c_str(), (int)t_counter++);

				}

				if (t_counter <= 9)
				{
					t_idx_name.assign(t_test_idx_name);

					// Note the differences, please:
					if (index_status == 1 )
					{
						// index_status 1 means "indexed, allow dupes", but we want
						// to make the index more efficient by adding the unique
						// identifier PKey to the spec.
						t_spec = std::string("CREATE INDEX ") + t_idx_name +
								 std::string(" ON [") +
								 std::string(FDefineTable->GetTableName()) +
								 std::string("] ([") + std::string(col_name) +
								 std::string("], [PKey])");
					}
					else
					{
						// index_status 2 means "indexed, dupes NOT allowed"
						// we'll use the UNIQUE constraint, and this time we
						// do NOT want to include the PKey in the spec
						t_spec = std::string("CREATE UNIQUE INDEX ") + t_idx_name +
								 std::string(" ON [") +
								 std::string(FDefineTable->GetTableName()) +
								 std::string("] ([") + std::string(col_name) +
								 std::string("])");
					}

					FDefineTable->AddIndexDef(t_spec);
				}
				else
				{
					// We can't find a unique name! Reject this attempt to
					// create a index until the user changes something about
					// the name.
					char t_err_msg[500];
					std::snprintf(t_err_msg, 499,
						"Index name \"%s\" is already in use, and 10 attempts to modify it have failed." \
						"Change either the first 8 characters of the Table name to something unique," \
						"or change the column name and try again.",
						t_idx_name.c_str());

					FEngineErrorsPump->AddError("TSMFWriter", "AddColumnDef", t_err_msg);

					t_result = -1;
				} // if (t_counter <= 9)

			} // if (index_status > 0)
		} // if (t_result == 0)
	}
	else
	{
		FEngineErrorsPump->AddError("TSMFWriter", "AddColumnDef",
			"Engine programmer error: FDefineTable not instantiated");
		t_result = -1;
	}

	return t_result;
} // AddColumnDef
//---------------------------------------------------------------------------
bool TSMFWriter::index_name_exists(const char* index_name)
{
	bool t_result = false;

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	char t_buffer[301];

	std::snprintf(t_buffer, 300,
		"SELECT [name] " \
		"FROM sqlite_master " \
		"WHERE [type] LIKE 'index' " \
		"AND [name] LIKE '%s';",
		index_name);

	try
	{
		t_qry->SetSQLText(t_buffer);
		t_qry->SetIsActive(true);

		t_result = (t_qry->RowCount() > 0);
	}
	catch (...)
	{
		FEngineErrorsPump->AddError("TSMFWriter", "index_name_exists",
			FSQLDB->GetLastDBErrorMessage());
	}
	return t_result;
} //index_name_exists
//---------------------------------------------------------------------------
int TSMFWriter::CreateTable(int* lookup_pkey)
{
	int t_result = 0;

	if (FDefineTable)
	{
		std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
		int t_count;

		if (FDefineTable->GetIsRestructuring())
		{
			// Retrieve trigger definitions, if any exist, on the original table
			// and park them in FDefineTable's storage locker. When the original
			// table is renamed to a backup name, SQLite helps by renaming
			// related objects, including portions of trigger definitions,
			// and we want to retrieve them as they used to be
			TTableDef* t_original = FTablesList->GetTable(FDefineTable->GetTableName());
			if (t_original)
			{
				t_count = (int)t_original->GetTriggerCount();
				for (int t_idx = 0; t_idx < t_count; t_idx++)
				{
					FDefineTable->AddTriggerDef(t_original->GetTrigger(t_idx));
				}
			}

			// Rename the source table to a backup name
			char t_buffer[500];
			std::snprintf(t_buffer, 500,
				"ALTER TABLE [%s] RENAME TO [%s_bak];",
				FDefineTable->GetTableName(), FDefineTable->GetTableName());

			t_qry->SetSQLText(t_buffer);
			try
			{
				t_qry->ExecSQL();
			}
			catch(...)
			{
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"Unable to rename original table to a back-up name");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"WARNING: This metafile may now be corrupted; close without saving");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					FSQLDB->GetLastDBErrorMessage());
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					t_buffer);

				t_result = -1;
			}
			if (t_result != 0)
			{
				return t_result;
			}
		} // if (FDefineTable->GetIsRestructuring())

		t_count = (int)FDefineTable->GetColumnsCount();
		if (t_count == 0)
		{
			FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
				"No columns have been defined; inappropriate use of Ew_CreateTable");
			t_result = -1;
			return t_result;
		}

		// Ready, Hezzie? We're going to try to create the new table structure now
		std::string t_spec =
			std::string("CREATE TABLE ") + std::string(FDefineTable->GetTableName()) +
			std::string("( [PKey] integer NOT NULL PRIMARY KEY, \n") ;
		for (int t_idx = 0; t_idx < t_count; t_idx++)
		{
			TDefineColumn* t_col = FDefineTable->GetColumnDef((size_t)t_idx);
			if (t_col)
			{
				t_spec += std::string(rtrim(t_col->GetFormattedDef()));
				if (t_idx < t_count-1)
					t_spec += std::string(", \n");
			}
		}

		t_spec += std::string(");");

		t_qry->SetSQLText(t_spec);
		try
		{
			t_qry->ExecSQL();
		}
		catch(...)
		{
			FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
				"Unable to create table as defined");
			FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
				"WARNING: This metafile may now be corrupted; close without saving");
			FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
				FSQLDB->GetLastDBErrorMessage());
			FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
				t_spec);

			t_result = -1;
		}
		if (t_result != 0)
		{
			return t_result;
		}

		// If we're restructuring, port data from old to new, wherever
		// we have a pairing of the new column name to the old (source) column name.
		if (FDefineTable->GetIsRestructuring())
		{
			std::string t_field_list = "",
						t_source_list = "";

			for (int t_idx = 0; t_idx < t_count; t_idx++)
			{
				TDefineColumn* t_col = FDefineTable->GetColumnDef((size_t)t_idx);
				// t_col->GetColumnSourceName() will be non-empty whenever
				// a new column is added to the table structure. Note, however,
				// that it is the duty of the application programmer to provide
				// this important piece of information when invoking
				// Ew_AddColumnDef!
				if (t_col && std::strlen(t_col->GetColumnSourceName()) > 0)
				{
					if (!t_field_list.empty())
					{
						t_field_list += std::string(", ");
						t_source_list += std::string(", ");
					}

					t_field_list += std::string("[") +
									std::string(t_col->GetColumnName()) +
									std::string("]");

					t_source_list += std::string("[") +
									std::string(t_col->GetColumnSourceName()) +
									std::string("]");
				}
			}

			t_spec = std::string("INSERT INTO [") +
					 std::string(FDefineTable->GetTableName()) +
					 std::string("] (") +
					 t_field_list +
					 std::string(") SELECT ") +
					 t_source_list + std::string(" FROM [") +
					 std::string(FDefineTable->GetTableName()) + std::string("_bak];");

			t_qry->SetSQLText(t_spec);
			try
			{
				t_qry->ExecSQL();
			}
			catch(...)
			{
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"Unable to port data from old table structure to new");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"WARNING: This metafile may now be corrupted; close without saving");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					FSQLDB->GetLastDBErrorMessage());
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					t_spec);

				t_result = -1;
			}
			if (t_result != 0)
			{
				return t_result;
			}

			// Have to drop the _bak table now
			// DROP TABLE IF EXISTS AJC5SIGP_bak;
			t_spec = std::string("DROP TABLE IF EXISTS ") +
					 std::string(FDefineTable->GetTableName()) +
					 std::string("_bak;");

			t_qry->SetSQLText(t_spec);
			try
			{
				t_qry->ExecSQL();
			}
			catch(...)
			{
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"Unable to drop old/back-up table");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"WARNING: This metafile may now be corrupted; close without saving");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					FSQLDB->GetLastDBErrorMessage());
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					t_spec);

				t_result = -1;
			}
			if (t_result != 0)
			{
				return t_result;
			}

			// Apply triggers for any columns that need them
			size_t t_tx_ctr = FDefineTable->GetTriggerCount();
			if (t_tx_ctr > 0)
			{
				std::string t_set_col,
							t_trigger;
				bool t_applied_trigger = false;
				// We don't just apply all the old triggers... one or more
				// may belong to columns that were deleted in restructuring.
				// Look at the columns we DO have to decide.
				for (int t_idx = 0; t_idx < t_count; t_idx++)
				{
					TDefineColumn* t_col = (TDefineColumn*)FDefineTable->GetColumnDef((size_t)t_idx);
					if (t_col && t_col->GetIsTriggerMaintained())
					{
						t_set_col = std::string("SET [") +
									std::string(t_col->GetColumnName()) +
									std::string("] ");
						for (size_t t_jdx = 0; t_jdx < t_tx_ctr; t_jdx++)
						{
							t_trigger = FDefineTable->GetTriggerDef(t_jdx);
							if (t_trigger.find(t_set_col) != std::string::npos)
							{
								t_qry->SetSQLText(t_trigger);
								try
								{
									if (t_qry->ExecSQL())
									{
										t_applied_trigger = true;
									}
								}
								catch (...)
								{
									FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
										"Unable to reapply trigger to restructured table");
									FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
										"WARNING: This metafile may now be corrupted; close without saving");
									FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
										FSQLDB->GetLastDBErrorMessage());
									FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
										t_spec);

									t_result = -1;
									break;
								}
							} // found a trigger def
						} // for (iterate trigger defs...
						if (t_result != 0)
						{
							break;
						}
					} // if (t_col...
				} // for (iterate columns...
				if (t_result != 0)
				{
					return t_result;
				}

				if (t_applied_trigger)
				{
					// At least one column's triggers were applied. Now we
					// have to fire them.
					t_spec = std::string("UPDATE ") +
							 std::string(FDefineTable->GetTableName()) +
							 std::string(" SET PKey = PKey;");
					t_qry->SetSQLText(t_spec);
					if (!t_qry->ExecSQL())
					{
						FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
							"Unable to fire trigger(s) in restructured table");
						FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
							"WARNING: This metafile may now be corrupted; close without saving");
						FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
							FSQLDB->GetLastDBErrorMessage());
						FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
							t_spec);

						t_result = -1;
						return t_result;
					}
				}
			} // if (t_tx_ctr > 0)

		} // if (FDefineTable->GetIsRestructuring())

		// Create the indexes
		size_t t_ix_ctr = FDefineTable->GetIndexCount();
		if (t_ix_ctr > 0)
		{
			for (size_t t_jdx=0; t_jdx < t_ix_ctr; t_jdx++)
			{
				t_qry->SetSQLText(FDefineTable->GetIndexDef(t_jdx));
				if (!t_qry->ExecSQL())
				{
					FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
						"Unable to create index for restructured table");
					FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
						"WARNING: This metafile may now be corrupted; close without saving");
					FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
						FSQLDB->GetLastDBErrorMessage());
					FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
						t_spec);

					t_result = -1;
					return t_result;
				}
			}
		}

		// If this is a new table, add it to EE_LookupTables
		if (!FDefineTable->GetIsRestructuring())
		{
			bool t_added = false;
			if (Insert("EE_LookupTables") == 0)
			{
				SetFieldByName("EE_LookupTables", "TableName", FDefineTable->GetTableName());
				SetFieldByName("EE_LookupTables", "RecCount", "0");
				if (Post("EE_LookupTables", lookup_pkey) == 0)
				{
					t_added = true;
				}
			}

			if (!t_added)
			{
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"Unable to add newly-created table to EE_LookupTables");
				FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
					"WARNING: This metafile may now be corrupted; close without saving");

				t_result = -1;
				return t_result;
			}
		}

		if (t_result == 0)
		{
			// If we make it to here, presume success. However,
			// we still have to update the FTablesList definition for the
			// new or restructured table.
			TTableDef* t_table_def = FTablesList->GetTable(FDefineTable->GetTableName());
			if (t_table_def)
			{
				FTablesList->RemoveTable(t_table_def);
			}

			// create a new table def...
			FParent->UpdateTableStats(FDefineTable->GetTableName());
		}
	} // if (FDefineTable)
	else
	{
		FEngineErrorsPump->AddError("TSMFWriter", "CreateTable",
			"Engine programmer error: FDefineTable not instantiated");
		t_result = -1;
	}

	return t_result;
} // CreateTable
//---------------------------------------------------------------------------


#endif // __SQLEW__
