//---------------------------------------------------------------------------
#include <algorithm> // std::transform
#include <stdexcept>

#ifdef EE_WINDOWS
#include <windows.h> // Sleep()
#else
#include <unistd.h>  // sleep()
#endif

#include "SQLDB.h"
/*****************************************
 ***  TSQLDB
 *****************************************/
TSQLDB::TSQLDB()
{
	FDatabase = NULL;
	FDBName.clear();

	sqlite3_initialize();
}
//---------------------------------------------------------------------------
TSQLDB::~TSQLDB()
{
	CloseDatabase();

	sqlite3_shutdown();
}
//---------------------------------------------------------------------------
sqlite3* TSQLDB::Database()
{
	return FDatabase;
}
//---------------------------------------------------------------------------
bool TSQLDB::OpenDatabase(const char* dbname)
{
	bool t_result = true;
	if (!FDatabase)
	{
		int t_sql_result = sqlite3_open(dbname, &FDatabase);

		if (t_sql_result == SQLITE_OK)
		{
            FDBName.assign(dbname);

			// This is how to set pragmas in SQLite. It's the way to
			// control database behavior, overriding defaults.
			// Support for foreign keys is OFF by default (ack!)
			// so we want to turn it back on.
			sqlite3_exec(FDatabase, "PRAGMA foreign_keys = 1", 0,0,0);
			// This turned out not to be needed, but if we ever DO
			// want to suppress check constraints, set the value to 1
			// (it's 0 by default).
			//sqlite3_exec(FDatabase, "PRAGMA ignore_check_constraints = 0", 0,0,0);
		}
		else
		{
			t_result = false;
		}
	}
	// If the sqlite3 *db is already assigned, just return true
	return t_result;
}
//---------------------------------------------------------------------------

bool TSQLDB::CloseDatabase()
{
	if (FDatabase)
    {
        int t_result = sqlite3_close(FDatabase);
        if (t_result == SQLITE_OK)
        {
            FDatabase = NULL;
            return true;
        }
        else
        {
			// Debugging - "unable to close due to unfinalized statements or unfinished backups"
			// Somewhere, a TSQLQuery was created but not destroyed. Happy hunting!
			//std::string t_err;
			//t_err.assign(GetLastDBErrorMessage());
			return false;
        }
    }
    return false;
}
//---------------------------------------------------------------------------
bool TSQLDB::Vacuum()
{
	if (FDatabase)
	{
		int t_result =
			sqlite3_exec(FDatabase,
						 "Vacuum;",
						 0,0,0);

		return (t_result == (SQLITE_OK));
	}

	return false;
}
//---------------------------------------------------------------------------
/*  sqlite3_exec will throw an error if there are problems with these
 *  transaction management calls
 */
bool TSQLDB::BeginTrans()
{
	int t_result = sqlite3_exec(FDatabase, "BEGIN", 0, 0, 0);
	return (t_result == SQLITE_OK);
}
//---------------------------------------------------------------------------
bool TSQLDB::CommitTrans()
{
	int t_result = sqlite3_exec(FDatabase, "COMMIT", 0, 0, 0);
	return (t_result == SQLITE_OK);
}
//---------------------------------------------------------------------------
bool TSQLDB::RollbackTrans()
{
	int t_result = sqlite3_exec(FDatabase, "ROLLBACK", 0, 0, 0);
	return (t_result == SQLITE_OK);
}
//---------------------------------------------------------------------------
bool TSQLDB::SavePoint(const char* savepoint_name)
{
	char t_command[100];
	std::snprintf(t_command, 100, "SAVEPOINT %s", savepoint_name);
	int t_result = sqlite3_exec(FDatabase, t_command, 0, 0, 0);
	return (t_result == SQLITE_OK);
}
//---------------------------------------------------------------------------
bool TSQLDB::Release(const char* savepoint_name)
{
	char t_command[100];
	std::snprintf(t_command, 100, "RELEASE %s", savepoint_name);
	int t_result = sqlite3_exec(FDatabase, t_command, 0, 0, 0);
	return (t_result == SQLITE_OK);
}
//---------------------------------------------------------------------------

bool TSQLDB::RollbackTo(const char* savepoint_name)
{
	char t_command[100];
	std::snprintf(t_command, 100, "ROLLBACK TO %s", savepoint_name);
	int t_result = sqlite3_exec(FDatabase, t_command, 0, 0, 0);
	return (t_result == SQLITE_OK);
}
//---------------------------------------------------------------------------

char * TSQLDB::GetLastDBErrorMessage()
{
    if (FDatabase)
    {
        return (char *)sqlite3_errmsg(FDatabase);
    }

    return (char*)"";
}
//---------------------------------------------------------------------------
char * TSQLDB::GetDBName()
{
    return (char*)FDBName.c_str();
}
//---------------------------------------------------------------------------
/* See SQLiteBrowser TQueryForm for example of use
static int PragmaCallback(void* callback_owner, int col_count,
	char** col_values, char** col_names)
{
	std::string t_values;
	std::string t_names;

	for (int t_idx = 0; t_idx < col_count; t_idx++)
	{

	}
	return 0;
}
//---------------------------------------------------------------------------
*/
int TSQLDB::Pragma(const char* sql, void* arg_to_callback, void* callback_func)
{

	char* errmsg = NULL;
	int t_result = 0;

	if (callback_func)
	{
		t_result = sqlite3_exec(FDatabase, sql, (CallbackPragma)callback_func, arg_to_callback, &errmsg);
		// If sqlite3_exec populated errmsg, then it allocated the memory for errmsg;
		// we have to release it.
		sqlite3_free(errmsg);
	}

	return t_result;
} // Pragma
//---------------------------------------------------------------------------

