#include <algorithm> // std::transform
#include <cctype>    // std::toupper
#include <cstring>   // std::strlen
#include <stdexcept>

#ifdef EE_WINDOWS
#include <windows.h> // Sleep()
#else
#include <unistd.h>  // sleep()
#endif

#include "SQLQuery.h"
#include "SQLDB.h" // for TSQLDB

//---------------------------------------------------------------------------
/*****************************************
 ***  TSQLQuery
 *****************************************/
TSQLQuery::TSQLQuery(TSQLDB* database)
{
	SQLDB = database;
	FSQLText.clear();

	FIsActive = false;
	FStatement = NULL;
	FColCount = 0;
	FRowCount = -1;

	FIsParameterized = false; // SetSQLText may change this to true
	FPreparedCount = 0;

} // constructor
//---------------------------------------------------------------------------
TSQLQuery::~TSQLQuery()
{
	force_finalize_query();
} // destructor
//---------------------------------------------------------------------------
void TSQLQuery::force_finalize_query()
{
	// called from the destructor
	int t_count = get_prepared_count();
	int t_rc = SQLITE_OK, t_idx;
	for (t_idx = 0; ((t_idx < t_count) && (t_rc == SQLITE_OK)); t_idx++)
	{
		try
		{
			t_rc = sqlite3_finalize(FStatement);
		}
		catch (...)
		{
			std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

			FLastErrorMessage.assign(t_err_msg);
			throw std::runtime_error("TSQLQuery::force_finalize_query");
		}
	}
} // force_finalize_query
//---------------------------------------------------------------------------
int TSQLQuery::get_prepared_count()
{
	return FPreparedCount;
}
//---------------------------------------------------------------------------
void TSQLQuery::incr_prepared_count()
{
	FPreparedCount += 1;
}
//---------------------------------------------------------------------------
void TSQLQuery::decr_prepared_count()
{
	if (FPreparedCount > 0)
	{
		FPreparedCount -= 1;
	}
}
//---------------------------------------------------------------------------
bool TSQLQuery::Eof()
{
	return FEof;
}
//---------------------------------------------------------------------------
int TSQLQuery::ColCount()
{
	return FColCount;
}
//---------------------------------------------------------------------------
int TSQLQuery::RowCount()
{
	// We'll get this only if the caller actually needs it; it's too
	// expensive otherwise
	if (FRowCount < 0)
	{
		std::string t_parse_sql = get_ucase_sql();
		FRowCount = get_row_count(t_parse_sql);
	}
	return FRowCount;
}
//---------------------------------------------------------------------------
bool TSQLQuery::PrepareParameterizedQuery()
{
	if (FSQLText.empty())
	{
		FLastErrorMessage.assign("SQLText is empty");
	}
	else if (!SQLDB)
	{
		FLastErrorMessage.assign("SQLDB not assigned for this query");
	}
	else if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: use open_query instead");
	}
	else
	{

		std::string t_column_name;
		std::string t_parse_sql = get_ucase_sql();

		if ((t_parse_sql.length() > 6) &&
			(t_parse_sql.substr(0,6) != "SELECT"))
		{
			// We found that this is NOT a SELECT statement. Tell the caller
			// to use ExecSQL instead.
			FLastErrorMessage.assign("SetActive expects SELECT statement; use ExecSQL instead");
		}
		else
		{
			if (!reset_statement())
			{
				FLastErrorMessage.assign("Failure to reset statement object");
			}
			else
			{
				FColCount = 0;
				FRowCount = -1;
				FEof = false;
				const char* t_tail;
				int t_rc;

				// NB: See comments in TSQLQuery::close_query about
				// the requirement to balance calls to sqlite3_prepare_v2
				// with corresponding calls to sqlite3_finalize... lest
				// you leave the database in a state that it will refuse to
				// close.
				t_rc = sqlite3_prepare_v2(
					SQLDB->Database(),        // Database handle
					(char*)FSQLText.c_str(),  // SQL text
					(int)FSQLText.length(),   // Length of SQL Text in bytes
					&FStatement,              // OUT: statement handle
					&t_tail);                 // OUT: pointer to unused portion of SQL Text

				if (t_rc == SQLITE_OK)
				{
                    incr_prepared_count();
					return true;
				}
				else
				{
					std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

					FLastErrorMessage.assign(t_err_msg);
				}
			}
		} // if (!reset_statement())
	}
	// If we got to here, we failed
	throw std::runtime_error("TSQLQuery::PrepareParameterizedQuery");
}  // PrepareParameterizedQuery
//---------------------------------------------------------------------------
bool TSQLQuery::reset_statement()
{
	// NOTE: This is called for BOTH parameterized and non-parameterized
	// queries. For the former, we are handling these calls internally, so
	// we don't need to bark at the user... see note in ResetParameterizedQuery
	// Clear this for re-use
	int t_rc = sqlite3_reset(FStatement);

	if (t_rc != SQLITE_OK)
	{
		FLastErrorMessage.assign("Failure to reset statement object");
		throw std::runtime_error("TSQLQuery::reset_statement");
	}

	return true;

} // reset_statement
//---------------------------------------------------------------------------
bool TSQLQuery::ResetParameterizedQuery()
{
	// The caller uses this to reset a previously-prepared query so that
	// new parameter values may be bound and it can be re-run. If the
	// query is not parameterized, there is no need to call this. True,
	// there is no harm, either, but I want programmers to think of these
	// functions in the context of sets and use them appropriately.
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: inappropriate call");
		throw std::runtime_error("TSQLQuery::ResetParameterizedQuery");
	}

	return reset_statement();

} // ResetParameterizedQuery
//---------------------------------------------------------------------------
bool TSQLQuery::BindIntParameter(const int param_num, const int value)
{
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: nothing to bind");
		throw std::runtime_error("TSQLQuery::BindIntParameter");
	}

	int t_rc = sqlite3_bind_int(FStatement, param_num, value);

	if (t_rc == SQLITE_OK)
	{
		return true;
	}
	else if (t_rc == SQLITE_MISUSE)
	{
		std::string t_err_msg = "unknown error (SQLITE_MISUSE)";
		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TSQLQuery::BindIntParameter");
	}
	else
	{
		std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TSQLQuery::BindIntParameter");
	}
} // BindIntParameter
//---------------------------------------------------------------------------
bool TSQLQuery::BindTextParameter(const int param_num, const char* value)
{
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: nothing to bind");
		throw std::runtime_error("TSQLQuery::BindTextParameter");
	}

	std::string t_value;
	t_value.assign(value);

	int t_rc = sqlite3_bind_text(FStatement, param_num,
		t_value.c_str(), t_value.length(), SQLITE_TRANSIENT);

	if (t_rc == SQLITE_OK)
	{
		return true;
	}
	else if (t_rc == SQLITE_MISUSE)
	{
		std::string t_err_msg = "SQLITE_MISUSE: ";
		t_err_msg += SQLDB->GetLastDBErrorMessage();
		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TSQLQuery::BindTextParameter");
	}
	else
	{
		std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

		FLastErrorMessage.assign(t_err_msg);
		throw std::runtime_error("TSQLQuery::BindTextParameter");
	}
} // BindTextParameter
//---------------------------------------------------------------------------
std::string TSQLQuery::get_ucase_sql()
{
	std::string t_ucase_sql;
	if (!FSQLText.empty())
	{
		t_ucase_sql.assign(FSQLText);
#ifdef BCB
		std::transform(t_ucase_sql.begin(),
					   t_ucase_sql.end(),
					   t_ucase_sql.begin(),
					   std::toupper);
#else
		// from linuxquestions.org: for g++ to be happy, it wants the toupper that
		// lives in the global namespace.
		std::transform(t_ucase_sql.begin(),
					   t_ucase_sql.end(),
					   t_ucase_sql.begin(),
					   ::toupper);
#endif
	}

	return t_ucase_sql;

} // get_ucase_sql
//---------------------------------------------------------------------------
bool TSQLQuery::OpenParameterizedQuery()
{
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: use open_query instead");
		throw std::runtime_error("TSQLQuery::OpenParameterizedQuery");
	}

	std::string t_column_name;
	int t_idx;

	int t_param_count = sqlite3_bind_parameter_count(FStatement);
	if (t_param_count > 0)
	{
		// Position on the first row in the result set
		int t_rc = sqlite3_step(FStatement);

		if (t_rc == SQLITE_ROW)
		{
			FEof = false;
		}
		else if (t_rc == SQLITE_DONE)
		{
			FEof = true;
		}
		// else { should probably yelp }

		// Regardless whether we have an empty result set,
		// get the number of columns in the result set
		FColCount = sqlite3_column_count(FStatement);

		// Clear any previous error message
		FLastErrorMessage.clear();

		// Set up the FColumnNames map so that GetFieldByName can work
		FColumnNames.clear();

		for (t_idx = 0; t_idx < FColCount; t_idx++)
		{
			t_column_name = sqlite3_column_name(FStatement, t_idx);
			FColumnNamesByNumber[t_idx] = t_column_name;
			std::transform(t_column_name.begin(),
						   t_column_name.end(),
						   t_column_name.begin(),
						   toupper);
			FColumnNames[t_column_name] = t_idx;
		}

		// Set the public flag
		FIsActive = true;

		return true;
	}
	// t_param_count is not greater than 0
	FLastErrorMessage.assign("Parameterized query has no bound parameters");
	throw std::runtime_error("TSQLQuery::OpenParameterizedQuery");
} // OpenParameterizedQuery
//---------------------------------------------------------------------------
bool TSQLQuery::ReleaseParameterizedQuery()
{
	// The caller uses this to finalize a previously-prepared query. If the
	// query is not parameterized, there is no need to call this. True,
	// there is no harm, either, but I want programmers to think of these
	// functions in the context of sets and use them appropriately.
	// Therefore, even though this function itself calls SetIsActive(false),
	// it will bark at the programmer to use the functions that belong
	// together.
	if (!GetIsParameterized())
	{
		FLastErrorMessage.assign("Not a parameterized query: use SetIsActive instead");
		throw std::runtime_error("TSQLQuery::ReleaseParameterizedQuery");
	}

	// This function exists ONLY to provide a "set"... goes with
	// PrepareParameterizedQuery
	return finalize_query();
}
//---------------------------------------------------------------------------
bool TSQLQuery::open_query()
{
	if (FSQLText.empty())
	{
		FLastErrorMessage.assign("SQLText is empty");
	}
	else if (!SQLDB)
	{
		FLastErrorMessage.assign("SQLDB not assigned for this query");
	}
	else if (GetIsParameterized())
	{
		FLastErrorMessage.assign("Parameterized query: Use OpenParameterizedQuery instead");
	}
	else
	{

		std::string t_column_name;
		int t_idx;
		std::string t_parse_sql = get_ucase_sql();

		if ((t_parse_sql.length() > 6) &&
			(t_parse_sql.substr(0,6) != "SELECT"))
		{
			// We found that this is NOT a SELECT statement. Tell the caller
			// to use ExecSQL instead.
			FLastErrorMessage.assign("Active property expects SELECT statement; use ExecSQL instead");
		}
		else
		{
			// Clear this for re-use
			int t_rc = sqlite3_reset(FStatement);

			if (t_rc != SQLITE_OK)
			{
				FLastErrorMessage.assign("TSQLQuery::open_query failed to reset statement object");
				return false;
			}

			FColCount = 0;
			FRowCount = -1;
			FEof = false;
			const char* t_tail;

			// NB: See comments in TSQLQuery::close_query about
			// the requirement to balance calls to sqlite3_prepare_v2
			// with corresponding calls to sqlite3_finalize... lest
			// you leave the database in a state that it will refuse to
			// close.
			t_rc = sqlite3_prepare_v2(
				SQLDB->Database(),        // Database handle
				(char*)FSQLText.c_str(),  // SQL text
				(int)FSQLText.length(),   // Length of SQL Text in bytes
				&FStatement,              // OUT: statement handle
				&t_tail);                 // OUT: pointer to unused portion of SQL Text

			if (t_rc == SQLITE_OK)
			{

                incr_prepared_count();

				// Position on the first row in the result set
				t_rc = sqlite3_step(FStatement);

				if (t_rc == SQLITE_ROW)
				{
					FEof = false;             // another row is waiting to be fetched
				}
				else if (t_rc == SQLITE_DONE) // empty result set/end of file
				{
					FEof = true;
				}

				// Regardless whether we have an empty result set,
				// get the number of columns in the result set
				FColCount = sqlite3_column_count(FStatement);

				// Clear any previous error message
				FLastErrorMessage.clear();

				// Set up the FColumnNames map so that GetFieldByName can work
				FColumnNames.clear();

				for (t_idx = 0; t_idx < FColCount; t_idx++)
				{
					t_column_name = sqlite3_column_name(FStatement, t_idx);
					FColumnNamesByNumber[t_idx] = t_column_name;
					std::transform(t_column_name.begin(),
								   t_column_name.end(),
								   t_column_name.begin(),
								   toupper);
					FColumnNames[t_column_name] = t_idx;
				}

				// Set the public flag
				FIsActive = true;

				return true;
			}
			else
			{
				std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

				FLastErrorMessage.assign(t_err_msg);
				throw std::runtime_error("TSQLQuery.open_query");
			}
		}

	}
	// If we got to here, we do not have an active query
	FIsActive = false;
	return false;
} // open_query
//---------------------------------------------------------------------------
bool TSQLQuery::close_query()
{
	if (GetIsParameterized())
	{
		FLastErrorMessage.assign("Parameterized query: Use ReleaseParameterizedQuery instead");
		return false;
	}

	return finalize_query();

} // close_query
//---------------------------------------------------------------------------
bool TSQLQuery::finalize_query()
{
	// NOTE: Every call to sqlite3_prepare_v2 MUST HAVE a corresponding
	// call to sqlite3_finalize. If not, the "close database" call to
	// sqlite3_close will fail with an error message mumbling about
	// unfinalized statements.
	//
	// The trick was figuring out how to balance these without trashing
	// the FStatement while a cursor query is open.

	// Assume the best...
	int t_rc = SQLITE_OK;

	if (FIsActive)
	{
		FIsActive = false;
		/* Refer to documents for sqlite3_finalize, and note particularly
		   this:
		   It is a grievous error for the application to try to use a prepared
		   statement after it has been finalized. Any use of a prepared
		   statement after it has been finalized can result in undefined and
		   undesirable behavior such as segfaults and heap corruption.

		   It is therefore critical to assign FStatement = NULL after
		   invoking sqlite3_finalize.
		*/
		try
		{
			t_rc = sqlite3_finalize(FStatement);
			if (t_rc == SQLITE_OK)
			{
				decr_prepared_count();
			}
		}
		catch (...)
		{
			FStatement = NULL;

			std::string t_err_msg = SQLDB->GetLastDBErrorMessage();

			FLastErrorMessage.assign(t_err_msg);
			throw std::runtime_error("TSQLQuery::finalize_query");
		}

		FStatement = NULL;
	}
	return (t_rc == SQLITE_OK);
} // finalize_query
//---------------------------------------------------------------------------
int TSQLQuery::get_row_count(std::string ucase_sql)
{
	if (FIsParameterized)
	{
		// Parameterized queries contain wildcards, so a row count
		// will always yield no results. Save the nanoseconds this
		// call would otherwise take to reach this verdict:
		return 0;
	}
	sqlite3_stmt* t_rows_statement = NULL;
	std::string t_rowcount_sql;
	const char* t_rows_tail;
	int t_row_count = 0;
	int t_rc;

	try
	{
		t_rowcount_sql.assign(ucase_sql);

		size_t t_pos = t_rowcount_sql.find("FROM ");
		if (t_pos != std::string::npos)
        {
            t_rowcount_sql = t_rowcount_sql.substr(t_pos, t_rowcount_sql.length());
            // remove ORDER BY, if one exists
            t_pos = t_rowcount_sql.find("ORDER BY");
            if (t_pos != std::string::npos)
            {
                t_rowcount_sql = t_rowcount_sql.substr( 0, t_pos-1);
            }
            t_rowcount_sql.assign("SELECT COUNT(*) " + t_rowcount_sql);

            sqlite3_reset(t_rows_statement);
			t_rc = sqlite3_prepare_v2(
                SQLDB->Database(),
                (char*)t_rowcount_sql.c_str(),
                (int)t_rowcount_sql.length(),
                &t_rows_statement,
                &t_rows_tail);

            if (t_rc == SQLITE_OK)
            {
				incr_prepared_count();

				// Position on the first row in the result set
                t_rc = sqlite3_step(t_rows_statement);

                t_row_count = sqlite3_column_int(t_rows_statement, 0);
            }
        }
    }
    catch (...)
    {
		t_rc =sqlite3_finalize(t_rows_statement);
		if (t_rc == SQLITE_OK)
		{
			decr_prepared_count();
		}
		return 0;
	}

	t_rc = sqlite3_finalize(t_rows_statement);
	if (t_rc == SQLITE_OK)
	{
		decr_prepared_count();
	}
	return t_row_count;
} // get_row_count
//---------------------------------------------------------------------------
char* TSQLQuery::GetColumnName(int col_idx)
{
    if (col_idx >= 0 && col_idx < FColCount)
		return (char*)FColumnNamesByNumber[col_idx].c_str();
    else
		return (char*)"GetColumName error: col_idx out of bounds";
}
//---------------------------------------------------------------------------
char* TSQLQuery::GetColumnDataType( int col_idx )
{
    if (col_idx >= 0 && col_idx < FColCount)
    {
        char* t_data_type =
			(char*)sqlite3_column_decltype(FStatement, col_idx);
		// Preventing a crash
		if (t_data_type)
			return t_data_type;
		else
			return (char*)"";
    }
    else
		return (char*)"GetColumDataType error: col_idx out of bounds";
}
//---------------------------------------------------------------------------
bool TSQLQuery::GetColumnConstraints( const char* table_name, const char* column_name,
	int& is_nullable, int& is_primary_key, int& is_autoinc)
{
	// NOTE:
	// SQLITE_ENABLE_COLUMN_METADATA must be defined IN THE PROJECT CONDITIONALS
	// in order for sqlite3_table_column_metadata to work

	int t_not_null,
		t_primary_key,
		t_auto_inc;

	int t_result =
		sqlite3_table_column_metadata(
			SQLDB->Database(),
			NULL, // database name may be null
			table_name,
			column_name,
			NULL, // char const **pzDataType,
			NULL, // char const **pzCollSeq,
			&t_not_null,
			&t_primary_key,
			&t_auto_inc);

	if (t_result == SQLITE_OK)
	{
		is_nullable = (t_not_null == 0);
		is_primary_key = t_primary_key;
		is_autoinc = t_auto_inc;

		return true;
	}


	return false;

} // GetColumnIsNullable
//---------------------------------------------------------------------------

bool TSQLQuery::FetchNext()
{
    int t_rc = -1; // initialize to meaningless value
    int t_retries = 0;

    FLastErrorMessage.clear();

    if (FEof)
    {
        FLastErrorMessage.assign("End of result set; no more records");
        return false;
    }
    else
    {
		while ((t_rc != SQLITE_ROW) && (!FEof) && (t_retries < 3))
        {
            t_rc = sqlite3_step(FStatement);

            switch (t_rc)
            {
                case SQLITE_DONE:
                    //FLastErrorMessage.assign("End of result set; no more records");
                    FEof = true;
					break;

                case SQLITE_BUSY:
#ifdef EE_WINDOWS
                    Sleep(1000); // sleep for 1 second
#else
		    sleep(1);
#endif
                    if (++t_retries > 3)
                        FLastErrorMessage.assign("Database busy; cannot fetch next record");
                    break;

                case SQLITE_MISUSE:
                    FLastErrorMessage.assign("SQLITE_MISUSE: Statement is bad");
                    // force out of this loop
                    t_retries = 4;
                    break;

                case SQLITE_ERROR:
					FLastErrorMessage.assign(SQLDB->GetLastDBErrorMessage());
                    // force out of this loop
                    t_retries = 4;
                    break;

                default:
                    // I'll assume (for starters) that we get here because
                    // t_rc == SQLITE_ROW
                    if (t_rc != SQLITE_ROW)
                    {
                        FLastErrorMessage.assign(SQLDB->GetLastDBErrorMessage());
                        // force out of this loop
                        t_retries = 4;
                    }
                    break;
            }  // switch (t_rc)
        } // while ((t_rc != SQLITE_ROW) && (!FEof) && (t_retries < 3))
    } // if (!FEof)

	return (t_rc == SQLITE_ROW || t_rc == SQLITE_DONE);
}  // FetchNext()
//---------------------------------------------------------------------------
char* TSQLQuery::GetFieldByName(std::string fldname)
{
	const char* t_value = "";
	if (FIsActive)
	{
		std::string t_fld_name;
		t_fld_name.assign(fldname);
		std::transform(t_fld_name.begin(),
					   t_fld_name.end(),
					   t_fld_name.begin(),
					   toupper);
		int t_col_number = FColumnNames[t_fld_name];
		if ((t_col_number >= 0) && (t_col_number < FColCount))
		{
			t_value = GetFieldByNumber(t_col_number);
		}
	}
	return (char*)t_value;
} // GetFieldByName
//---------------------------------------------------------------------------
char* TSQLQuery::GetFieldByNumber(int column_number)
{
	const char* t_value = "";
    if (FIsActive)
    {
		t_value = (char*)sqlite3_column_text(FStatement, column_number);
		// sqlite3_column_text will return NULL if there is no data in the
		// column, which can lead to an exception when a calling routine
		// assumes the value will be a NULL-terminated string. Check this
		// value, and fix it if necessary.
		if (!t_value)
			t_value = "";
    }
	return (char*)t_value;
} // GetFieldByNumber
//---------------------------------------------------------------------------
int TSQLQuery::GetIntFieldByName(std::string fldname)
{
    int t_value = 0;
    if (FIsActive)
    {
		std::string t_fld_name;
		t_fld_name.assign(fldname);
		std::transform(t_fld_name.begin(),
					   t_fld_name.end(),
					   t_fld_name.begin(),
					   toupper);

		int t_col_number = FColumnNames[t_fld_name];

		if ((t_col_number >= 0) && (t_col_number < FColCount))
		{
			t_value = GetIntFieldByNumber(t_col_number);
		}
	}
    return t_value;
} // GetIntFieldByName
//---------------------------------------------------------------------------
int TSQLQuery::GetIntFieldByNumber(int column_number)
{
    int t_value = 0;
    if (FIsActive)
    {
        t_value = sqlite3_column_int(FStatement, column_number);
    }
    return t_value;
} // GetFieldByNumber
//---------------------------------------------------------------------------
bool TSQLQuery::GetIsActive()
{
	return FIsActive;
}
//---------------------------------------------------------------------------
bool TSQLQuery::SetIsActive(bool is_active)
{
    if (is_active)
    {
        // if the query is already active, close it and reopen it
        if (FIsActive)
            close_query();

		return open_query();
	}
	else
	{
		return close_query();
	}
}
//---------------------------------------------------------------------------
bool TSQLQuery::GetIsParameterized()
{
	return FIsParameterized;
}
//---------------------------------------------------------------------------
void TSQLQuery::set_is_parameterized(bool is_parameterized)
{
	FIsParameterized = is_parameterized;
}
//---------------------------------------------------------------------------
char* TSQLQuery::GetSQLText()
{
	return (char*)FSQLText.c_str();
}
//---------------------------------------------------------------------------
void TSQLQuery::SetSQLText(std::string sql_text, bool check_parameterized /* = true */)
{
	// Until proven otherwise...
	set_is_parameterized(false);

	if (!sql_text.empty())
	{
		if (check_parameterized)
		{
			// Check to see if this is a parameterized query, and set the
			// FIsParameterized property accordingly. Watch out for WHERE
			// clauses with literal strings that embed question marks,
			// e.g., "WHERE f.FieldName LIKE '%available?%'"
			// (we encountered a custom metafile that did exactly that).

			// Note that check_parameterized is false when called from
			// TLibFuncs::internal_lookup. We may find other reasons
			// why we should make this option more public, but let's
			// wait to see...
			size_t t_pos = sql_text.find("?");
			if (t_pos != std::string::npos)
			{
				size_t t_open_quote = std::string::npos,
					   t_close_quote = std::string::npos,
					   t_test = t_pos;
				// find the nearest open quote, if any
				while (t_test > 0)
				{
					if (sql_text[t_test] == 39)  // ASCII apostrophe
					{
						t_open_quote = t_test;
						break;
					}
					t_test--;
				}

				if (t_open_quote != std::string::npos)
				{
					t_test = t_pos;
					// now look for the nearest closing quote
					while (t_test < sql_text.length())
					{
						if (sql_text[t_test] == 39)  // ASCII apostrophe
						{
							t_close_quote = t_test;
							break;
						}

						t_test++;
					}

					// If we located an opening quote before the question mark,
					// and a closing quote after the question mark, I'm calling
					// this a string with a literal question mark embedded.
					// Otherwise, set_is_parameterized to true.
					if (t_close_quote == std::string::npos)
					{
						set_is_parameterized(true);
					}
				}
				else
				{
					// If no opening quotes, that question mark indicates a
					// parameter position.
					set_is_parameterized(true);
				}
			}
		}
	}
	// Empty string or not, assign the string the user provided.
	FSQLText.assign(sql_text);
} // SetSQLText
//---------------------------------------------------------------------------
char* TSQLQuery::LastErrorMessage()
{
	// Support for property LastErrorMessage, a read-only property
	return (char*)FLastErrorMessage.c_str();
}
//---------------------------------------------------------------------------
bool TSQLQuery::ExecSQL()
{
    // ExecSQL is used for SQL statements that do NOT return a
    // result set.
    if (FSQLText.empty())
    {
        FLastErrorMessage.assign("SQLText is empty");
    }
    else if (!SQLDB)
    {
        FLastErrorMessage.assign("SQLDB not assigned for this query");
    }
    else
    {
        std::string t_find_select;

        t_find_select.assign(FSQLText);

#ifdef BCB
        std::transform(t_find_select.begin(),
                       t_find_select.end(),
                       t_find_select.begin(),
                       std::toupper);
#else
        // g++ wants toupper in global namespace
        std::transform(t_find_select.begin(),
                       t_find_select.end(),
                       t_find_select.begin(),
                       ::toupper);
#endif
        if ((t_find_select.length() > 6) &&
            (t_find_select.substr(0,6) == "SELECT"))
        {
            // We found that this is a SELECT statement. Tell the caller
            // to use property Active instead.
            FLastErrorMessage.assign("SELECT statement should not use ExecSQL(); use SetActive(true), instead");
        }
        else
		{
            try
            {
                char* t_errmsg;

                int t_result =
                    sqlite3_exec(SQLDB->Database(),
                    (char*)FSQLText.c_str(),
                    NULL, // no callback
                    NULL, // no argument to callback
                    &t_errmsg);

                if (t_result == SQLITE_OK)
                {
                    // Clear any previous error message
                    FLastErrorMessage.clear();
                    // and return success flag
                    return true;
                }
                else
                {
                    if (t_errmsg)
                    {
                        FLastErrorMessage.assign(t_errmsg);
                    }
                }
            }
            catch(std::exception& e)
            {
				std::string t_what = e.what();
				if (t_what.empty())
				{
					t_what.assign("undefined error in TSQLDB::ExecSQL");
				}
				FLastErrorMessage.assign(t_what);
            }
        } // if not a SELECT statement
    }
    return false;
}  // ExecSQL()
//---------------------------------------------------------------------------
