//---------------------------------------------------------------------------
#include <string>
#include <map>
#include <vector>

#ifndef TableStatsH
#define TableStatsH
//---------------------------------------------------------------------------

enum TEditDataType { edtUnknown,
                     edtInteger,
                     edtDouble,
                     edtVarchar,
                     edtChar,
                     edtMemo,
					 edtDateTime };

// Metafile MCO types
enum TEE_MCO {
    _mco_Other,               // ( 0) Anything that doesn't begin "EE_"
	_mco_Agencies,            // ( 1) EE_Agencies
	_mco_Edits,               // ( 2) EE_Edits
	_mco_EditSets,            // ( 3) EE_EditSets
	_mco_Fields,              // ( 4) EE_Fields
	_mco_Layouts,             // ( 5) EE_Layouts
	_mco_Messages,            // ( 6) EE_Messages
	_mco_LookupTables,        // ( 7) EE_LookupTables
	_mco_LayoutFieldsPivot,   // ( 8) EE_LayoutFieldsPivot
	_mco_EditSetEditsPivot,   // ( 9) EE_EditSetEditsPivot
	_mco_EditFieldsPivot,     // (10) EE_EditFieldsPivot
	_mco_EditTablesPivot,     // (11) EE_EditTablesPivot
	_mco_EditMessagesPivot,   // (12) EE_EditMessagesPivot
	_mco_EditErrorsPivot,     // (13) EE_EditErrorsPivot
	_mco_Control              // (14) EE_Control
};


#ifdef __SQLEW__
	// The data set state helps us manage putting a particular tuple
	// in the table into "ready mode" for inserting or updating data.
	enum TEw_DataSetState { ewInactive, // not open in a cursor
							ewBrowse,   // open and cursor may be browsed
							ewEdit,     // pending a Post() for update
							ewInsert    // pending a Post() for insert
						  };
#endif


class TFieldDef
{
private:
	std::string FFieldName;
	int FColumnOrder;
	TEditDataType FDataType;

	int FColumnSize;       // applies only to varchar and char data types,
						   // all others default to 0
	bool FIsNullable;
	bool FIsPrimaryKey;
	bool FIsAutoInc;      // field defined to auto-increment
	bool FIsTriggerField; // field contents maintained by trigger(s)
	std::string FDefaultAsString; // depending on FDataType, it will be cast appropriately
	std::string FLimitToList;  // a CHECK constraint like CHECK(ThisColumn IN('XXX','YYY','ZZZ'))
	std::string FCheckBetween; // another CHECK constraint, like CHECK(PrimarySite BETWEEN 'C000' AND 'C999');


#ifdef __SQLEW__
	bool FIsModified;     // if the field is modified, its value should be posted to the table
	std::string FValue;   // current value
#endif
public:
	TFieldDef(const char* field_name, const int column_order,
		const TEditDataType data_type, int column_size,
		bool is_nullable, bool is_primary_key, bool is_autoinc);
	~TFieldDef();

	char* GetFieldName();
	void SetFieldName(const char* field_name);
	int GetColumnOrder();
	TEditDataType GetDataType();
	int GetColumnSize();
	bool GetIsNullable();
	bool GetIsPrimaryKey();
	bool GetIsAutoInc();
	void SetIsTriggerField(bool is_triggered);
	bool GetIsTriggerField();

	bool HasDefaultValue();
	char* GetDefaultValue();
	void SetDefaultValue(const char* default_value);

	bool HasLimitToList();
	char* GetLimitToList();
	void SetLimitToList(const char* limit_to_list);

	bool HasCheckBetween();
	char* GetCheckBetween();
	void SetCheckBetween(const char* check_between);

#ifdef __SQLEW__
	// These properties come into play only for EditWriter-type applications
	bool GetIsModified();
	char* GetValue();
	void SetValue(const char* value);

	void ClearWriteProperties();
#endif

};


class TTableDef
{
private:
	std::string FTableName;
	std::map<int, TFieldDef*> FFieldsByColumnOrder;
	std::map<std::string, TFieldDef*> FFieldsByName;

	std::vector<std::string> FIndexDefs;   // index name, sql
	std::vector<std::string> FTriggerDefs; // trigger name, sql
	std::vector<std::string> FForeignKeys; // constructed foreign key constraints
	std::vector<std::string> FCheckConstraints;  // check constraints from sqlite_master
	std::vector<std::string> FDefaultConstraints; // defaults from pragma table_info

	TEE_MCO FMCO; // see enum, above

#ifdef __SQLEW__
	TEw_DataSetState FState;
	int FPostToPKey;    // identifier for tables with pending Post() for Edit()

#endif // __SQLEW__


public:
	TTableDef(const char* table_name);
	~TTableDef();

	char* GetTableName();
	TEE_MCO GetMCO();
	int GetColumnCount();
	int AddColumn(TFieldDef* field_def);
	TFieldDef* GetColumn(int column_order);
	TFieldDef* GetColumn(const char* field_name);

	void AddIndex(const char* sql);
	int GetIndexCount();
	char* GetIndex(const int idx);

	void AddTrigger(const char* sql);
	int GetTriggerCount();
	char* GetTrigger(const int idx);

	void AddForeignKey(const char* fkey);
	int GetForeignKeyCount();
	char* GetForeignKey(const int idx);

	void AddCheck(const char* check_constraint);
	int GetCheckCount();
	char* GetCheck(const int idx);

	void AddDefault(const char* default_constraint);
	int GetDefaultCount();
	char* GetDefault(const int idx);

#ifdef __SQLEW__
	// This information is used to marshall up all that will be required
	// to construct an UPDATE or INSERT SQL statement for this table.
	// Refer also to the properties of TFieldDef, above.

	bool Edit(int pkey);
	bool Insert();
	//void Post(); Not here! This goes in TSMFWriter
	bool Cancel();
	bool IsEditing();
	TEw_DataSetState GetState();
	int GetPostToPKey();

	void ClearWriteFields();

#endif


};

class TTablesList
{
private:
    std::map<std::string, TTableDef*> FTablesList;
public:
    TTablesList();
    ~TTablesList();

	int AddTable(TTableDef* table_def);
	int RemoveTable(TTableDef* table_def);
    TTableDef* GetTable(std::string table_name);
    int Count();
    std::map<std::string, TTableDef*> TablesList();

};

/****************************************************************************
 *** Support for defining new lookup tables and restructuring existing ones
 ****************************************************************************/
#ifdef __SQLEW__
class TDefineColumn
{
private:
	std::string FColDstName;     // column name to copy data to new structure
	std::string FFormattedDef;   // fully formatted, ready to go
	std::string FColSrcName;     // column name to copy data from old structure
	int FIsTriggerMaintained;
public:
	TDefineColumn(const char* col_name, const char* formatted_def,
		const char* copy_from_col_name);
	~TDefineColumn() {}; // no dtor needed

	char* GetColumnName() { return (char*)FColDstName.c_str(); }
	char* GetFormattedDef() { return (char*)FFormattedDef.c_str(); }
	char* GetColumnSourceName() { return (char*)FColSrcName.c_str(); }

	void SetIsTriggerMaintained(const int is_trigger_maintained);
	int GetIsTriggerMaintained() { return FIsTriggerMaintained; }
};

class TDefineTable
{
private:
	std::string FTableName;
	bool FIsRestructuring;
	std::vector<TDefineColumn*> FColumnDefs;
	std::vector<std::string> FIndexDefs;
	std::vector<std::string> FTriggerDefs;
public:
	TDefineTable();
	~TDefineTable();

	void ClearAll();

	void SetTableName(const char* table_name);
	char* GetTableName() { return (char*)FTableName.c_str(); }

	void SetIsRestructuring(bool is_restructuring);
	bool GetIsRestructuring() { return FIsRestructuring; }

	size_t GetColumnsCount();
	void AddColumnDef(TDefineColumn* col_def);
	TDefineColumn* GetColumnDef(size_t idx);

	size_t GetIndexCount();
	void AddIndexDef(std::string index_def);
	char* GetIndexDef(size_t idx);

	size_t GetTriggerCount();
	void AddTriggerDef(std::string trigger_def);
	char* GetTriggerDef(size_t idx);

};



#endif // __SQLEW__


#endif  // TableStatsH
