//---------------------------------------------------------------------------

#ifndef TokenMasterH
#define TokenMasterH
//---------------------------------------------------------------------------
#include <string>
#include <vector>
#include <stdint.h> // intptr_t

#include "CompilerTokens.h" // used by TTokenMaster and TEditCompiler

// forward declarations
class TEngineErrorsPump;

/*  NOTE: Compiler option "Integer-sized enums", which defaults to true,
 *  is overridden in this project to set it to false. From the CodeGear help:

	Allocates a whole word (a four-byte int for 32-bit programs) for enumeration
	types (variables of type enum).

	When this option is off (-bi-), the compiler allocates the smallest integer
	that can hold the enumeration values: the compiler allocates an unsigned or
	signed char if the values of the enumeration are within the range of 0
	through 255 (minimum) or -128 through 127 (maximum), or an unsigned or
	signed short if the values of the enumeration are within the following
	ranges: 0..65535 or -32768..32767.

	The compiler allocates a four-byte int (32-bit) to represent the enumeration
	values if any value is out of these ranges.

	Default = True
 */

class TToken
{
private:
	std::string FOriginal;
	TTokenID FTokenID;
	TTokenType FTokenType;
	char FCompiled;
public:
	TToken(const char* original, TTokenID token_id, TTokenType token_type, char compiled);

	// Getters
	char* GetOriginal() { return (char*)FOriginal.c_str(); }
	TTokenID GetTokenID() { return FTokenID; }
	TTokenType GetTokenType() { return FTokenType; }
	char GetCompiled() { return FCompiled; }

	void SetOriginal(const char* original) { FOriginal.assign(original); }
	void SetTokenID(TTokenID token_id) { FTokenID = token_id; }
	void SetTokenType(TTokenType token_type) { FTokenType = token_type; }
	void SetCompiled(char compiled) { FCompiled = compiled; }
};

class TCargo
{
private:
	char* FCargoString;
	size_t FCargoSize;
public:

	TCargo(size_t string_size);
	~TCargo();

	size_t GetCargoSize();
	intptr_t GetCargoAddress();
	void SetCargoString( const char* cargo_string, size_t cargo_len );
};

class TTokenMaster
{
private:
	const char* FCompiledLogic; // assigned in TRunEdits::RunEdits and DebugEdit
	size_t FCompiledLength;
	size_t FPos;
	size_t FOldPos;

	// FStrings vector never clears its contents (until the TTokenMaster dtor)
	std::vector<TCargo*> FStrings;
	size_t FStringsVectorSize;  // starts at 0, increments as needed by 20
	size_t FTopOfStringsStack;  // reset to 0 in call from TRunEdits::interpret_block

	// FOneChar holds the one-character cargo, such as the tokens identifying
	// libfuncs. These don't have to live as long as the FStrings do, so we
	// can just keep overwriting for every call into GetCompiledToken
	// Note that although we're interested in just the single character, we
	// put it into a 2-byte string for easier use by the callers.
	char FOneChar[2];


	TEngineErrorsPump* FEngineErrorsPump;

	TToken* FTokensByCompiled[127];

	void set_up_tokens();
	void add_token( const char* original, const TTokenID token_id,
		const TTokenType token_type, const unsigned char compiled_char);
	void extract_field_number();

	void expand_strings_vector();

public:
	TTokenMaster(TEngineErrorsPump* pump);
	~TTokenMaster();

	void ResetTopOfStringsStack();
	intptr_t CopyToCargo(const char* cargo);
	intptr_t GetCallCargo();

	intptr_t GetOneChar();

	void SetCompiledLogic(const char* logic);
	size_t GetCompiledLength();
	const char* GetCompiledLogic();
	void SavePos();
	void RestorePos();
	size_t GetPos();
	void MovePosBy( int move_by );
	void MovePosTo( size_t move_to );
	void SkipNewLines();

	TToken* GetCompiledToken();
	TToken* GetHardToken(const TTokenID token_id );
};

#endif
